Package: tall
Title: Text Analysis for All
Version: 1.0.0
Authors@R: c(
    person(given = "Massimo", 
           family = "Aria",
           email = "aria@unina.it", 
           role = c("aut", "cre", "cph"),
           comment= c(ORCID = "0000-0002-8517-9411")),
    person(given = "Maria", 
           family = "Spano", 
           email = "maria.spano@unina.it", 
           role = c("aut"),
           comment= c(ORCID = "0000-0002-3103-2342")),
    person(given ="Luca", 
           family = "D'Aniello", 
           email = "luca.daniello@unina.it", 
           role = c("aut"),
           comment= c(ORCID = "0000-0003-1019-9212")),
    person(given = "Corrado", 
           family = "Cuccurullo", 
           email = "corrado.cuccurullo@unicampania.it", 
           role = c("ctb"),
           comment= c(ORCID = "0000-0002-7401-8575")),
    person(given = "Michelangelo", 
           family = "Misuraca", 
           email = "mmisuraca@unisa.it", 
           role = c("ctb"),
           comment= c(ORCID = "0000-0002-8794-966X"))
    )
Description: An R 'shiny' app designed for diverse text analysis tasks, offering a wide range of methodologies tailored to Natural Language Processing (NLP) needs. 
             It is a versatile, general-purpose tool for analyzing textual data. 
             'tall' features a comprehensive workflow, including data cleaning, preprocessing, statistical analysis, and visualization, all integrated for effective text analysis.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://github.com/massimoaria/tall,
        https://massimoaria.github.io/tall-app/
BugReports: https://github.com/massimoaria/tall/issues
Depends: R (>= 3.5.0), shiny
Imports: base64enc, ca, chromote, curl (>= 6.3.0), doParallel, dplyr
        (>= 1.1.0), DT, fontawesome, future, ggplot2, ggraph,
        ggwordcloud, httr2, igraph, jsonlite, later, Matrix, openxlsx,
        pagedown, pdftools (>= 3.6.0), plotly, promises, purrr, ranger,
        Rcpp (>= 1.0.3), readr, readtext, readxl, rlang, RSpectra,
        shinycssloaders (>= 1.1.0), shinydashboardPlus, shinyFiles,
        shinyjs, shinyWidgets, sparkline, stringr, strucchange,
        textrank, tidygraph, tidyr, stm, topicmodels, udpipe, umap,
        visNetwork, word2vec
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
LazyData: true
LinkingTo: Rcpp
NeedsCompilation: yes
Packaged: 2026-04-15 13:42:15 UTC; massimoaria
Author: Massimo Aria [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-8517-9411>),
  Maria Spano [aut] (ORCID: <https://orcid.org/0000-0002-3103-2342>),
  Luca D'Aniello [aut] (ORCID: <https://orcid.org/0000-0003-1019-9212>),
  Corrado Cuccurullo [ctb] (ORCID:
    <https://orcid.org/0000-0002-7401-8575>),
  Michelangelo Misuraca [ctb] (ORCID:
    <https://orcid.org/0000-0002-8794-966X>)
Maintainer: Massimo Aria <aria@unina.it>
Repository: CRAN
Date/Publication: 2026-04-15 14:20:02 UTC
Built: R 4.4.3; x86_64-apple-darwin20; 2026-04-17 08:24:23 UTC; unix
Archs: tall.so.dSYM

Package: tall
Title: Text Analysis for All
Version: 1.0.0
Authors@R: c(
    person(given = "Massimo", 
           family = "Aria",
           email = "aria@unina.it", 
           role = c("aut", "cre", "cph"),
           comment= c(ORCID = "0000-0002-8517-9411")),
    person(given = "Maria", 
           family = "Spano", 
           email = "maria.spano@unina.it", 
           role = c("aut"),
           comment= c(ORCID = "0000-0002-3103-2342")),
    person(given ="Luca", 
           family = "D'Aniello", 
           email = "luca.daniello@unina.it", 
           role = c("aut"),
           comment= c(ORCID = "0000-0003-1019-9212")),
    person(given = "Corrado", 
           family = "Cuccurullo", 
           email = "corrado.cuccurullo@unicampania.it", 
           role = c("ctb"),
           comment= c(ORCID = "0000-0002-7401-8575")),
    person(given = "Michelangelo", 
           family = "Misuraca", 
           email = "mmisuraca@unisa.it", 
           role = c("ctb"),
           comment= c(ORCID = "0000-0002-8794-966X"))
    )
Description: An R 'shiny' app designed for diverse text analysis tasks, offering a wide range of methodologies tailored to Natural Language Processing (NLP) needs. 
             It is a versatile, general-purpose tool for analyzing textual data. 
             'tall' features a comprehensive workflow, including data cleaning, preprocessing, statistical analysis, and visualization, all integrated for effective text analysis.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://github.com/massimoaria/tall,
        https://massimoaria.github.io/tall-app/
BugReports: https://github.com/massimoaria/tall/issues
Depends: R (>= 3.5.0), shiny
Imports: base64enc, ca, chromote, curl (>= 6.3.0), doParallel, dplyr
        (>= 1.1.0), DT, fontawesome, future, ggplot2, ggraph,
        ggwordcloud, httr2, igraph, jsonlite, later, Matrix, openxlsx,
        pagedown, pdftools (>= 3.6.0), plotly, promises, purrr, ranger,
        Rcpp (>= 1.0.3), readr, readtext, readxl, rlang, RSpectra,
        shinycssloaders (>= 1.1.0), shinydashboardPlus, shinyFiles,
        shinyjs, shinyWidgets, sparkline, stringr, strucchange,
        textrank, tidygraph, tidyr, stm, topicmodels, udpipe, umap,
        visNetwork, word2vec
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
LazyData: true
LinkingTo: Rcpp
NeedsCompilation: yes
Packaged: 2026-04-15 13:42:15 UTC; massimoaria
Author: Massimo Aria [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-8517-9411>),
  Maria Spano [aut] (ORCID: <https://orcid.org/0000-0002-3103-2342>),
  Luca D'Aniello [aut] (ORCID: <https://orcid.org/0000-0003-1019-9212>),
  Corrado Cuccurullo [ctb] (ORCID:
    <https://orcid.org/0000-0002-7401-8575>),
  Michelangelo Misuraca [ctb] (ORCID:
    <https://orcid.org/0000-0002-8794-966X>)
Maintainer: Massimo Aria <aria@unina.it>
Repository: CRAN
Date/Publication: 2026-04-15 14:20:02 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-04-20 23:45:34 UTC; unix
Archs: tall.so.dSYM

# Generated by roxygen2: do not edit by hand

export(calculate_ngram_is)
export(compute_syntactic_complexity)
export(extract_noun_phrases)
export(extract_svo_triplets)
export(process_multiwords_fast)
export(reinPlot)
export(reinSummary)
export(reinert)
export(tall)
export(term_per_cluster)
export(txt_recode_fast)
export(txt_recode_ngram_fast)
import(Rcpp)
import(curl)
import(doParallel)
import(dplyr)
import(fontawesome)
import(ggraph)
import(ggwordcloud)
import(openxlsx)
import(pagedown)
import(pdftools)
import(purrr)
import(readtext)
import(readxl)
import(shiny)
import(shinyFiles)
import(shinycssloaders)
import(shinydashboardPlus)
import(stm)
import(stringr)
import(tidygraph)
import(tidyr)
import(topicmodels)
import(udpipe)
import(visNetwork)
importFrom(DT,DTOutput)
importFrom(DT,datatable)
importFrom(DT,formatRound)
importFrom(DT,formatStyle)
importFrom(DT,renderDT)
importFrom(Matrix,Matrix)
importFrom(RSpectra,svds)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(base64enc,dataURI)
importFrom(ca,ca)
importFrom(chromote,default_chromote_object)
importFrom(future,plan)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggsave)
importFrom(httr2,req_perform)
importFrom(httr2,request)
importFrom(httr2,resp_body_json)
importFrom(igraph,centr_betw)
importFrom(igraph,cluster_walktrap)
importFrom(igraph,graph_from_data_frame)
importFrom(igraph,membership)
importFrom(jsonlite,fromJSON)
importFrom(later,later)
importFrom(plotly,add_annotations)
importFrom(plotly,add_markers)
importFrom(plotly,add_polygons)
importFrom(plotly,add_trace)
importFrom(plotly,config)
importFrom(plotly,event_data)
importFrom(plotly,event_register)
importFrom(plotly,layout)
importFrom(plotly,plot_ly)
importFrom(plotly,plotlyOutput)
importFrom(plotly,renderPlotly)
importFrom(promises,promise_all)
importFrom(ranger,ranger)
importFrom(readr,write_csv)
importFrom(rlang,sym)
importFrom(shinyWidgets,alert)
importFrom(shinyWidgets,prettyCheckboxGroup)
importFrom(shinyWidgets,updatePrettyCheckboxGroup)
importFrom(shinyjs,useShinyjs)
importFrom(sparkline,sparkline)
importFrom(sparkline,spk_chr)
importFrom(strucchange,breakpoints)
importFrom(textrank,textrank_sentences)
importFrom(umap,umap)
importFrom(word2vec,word2vec)
useDynLib(tall)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# TALL - Text Analysis for ALL

<img src="images/tall_logo.jpg" width="685" />

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/tall)](https://CRAN.R-project.org/package=tall)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/tall)](https://CRAN.R-project.org/package=tall)
<!-- badges: end -->

TALL (Text Analysis for ALL) is an interactive R Shiny application
designed for exploring, modeling, and visualizing textual data. It
provides a comprehensive, code-free environment for Natural Language
Processing, enabling researchers without extensive programming skills to
perform sophisticated text analyses through an intuitive graphical
interface.

TALL integrates state-of-the-art NLP techniques — tokenization,
lemmatization, Part-of-Speech tagging, dependency parsing, topic
modeling, sentiment analysis, and more — into a unified, reproducible
workflow.

------------------------------------------------------------------------

## Reference Paper

> **Aria, M., Spano, M., D’Aniello, L., Cuccurullo, C., & Misuraca, M.**
> (2026). TALL: Text analysis for all — an interactive R-shiny
> application for exploring, modeling, and visualizing textual data.
> *SoftwareX*, 34, 102590.

**[Read the full paper (Open
Access)](https://doi.org/10.1016/j.softx.2026.102590)** \|
**[Supplementary
material](https://ars.els-cdn.com/content/image/1-s2.0-S2352711026000841-mmc1.pdf)**

> ⚠️ **Citation policy.** TALL is open source and distributed under the
> MIT license. However, **whenever results obtained with TALL are used
> in a publication, proper citation of the reference paper above is
> mandatory**. Failure to properly cite the software is considered a
> violation of the license.

------------------------------------------------------------------------

## Setup

### System Requirements

Before installing TALL, ensure you have:

1.  **R version 4.2.0 or higher** — Download from
    [CRAN](https://cran.r-project.org/)
2.  **RStudio** (recommended) — Available at
    [Posit](https://posit.co/products/open-source/rstudio)
3.  **Active internet connection** for downloads and dependencies
4.  **Additional tools for the development version:**
    - Windows: [Rtools](https://cran.r-project.org/bin/windows/Rtools/)
    - Mac: Xcode Command Line Tools

### Stable Version (CRAN)

``` r
install.packages("tall")
library(tall)
tall()
```

### Development Version (GitHub)

First, verify your build tools:

``` r
if (!require("pak", quietly = TRUE)) install.packages("pak")
pkgbuild::check_build_tools(debug = TRUE)
```

Then install from [GitHub](https://github.com/massimoaria/tall):

``` r
if (!require("remotes", quietly = TRUE)) install.packages("remotes")
remotes::install_github("massimoaria/tall")
library(tall)
tall()
```

The development version includes the latest features but may contain
occasional bugs.

For detailed installation instructions, visit: **[Download &
Install](https://massimoaria.github.io/tall-app/download.html)**

An interactive tutorial is also available: **[View
tutorial](https://www.k-synth.com/tall/TALL_tutorial.html)**

------------------------------------------------------------------------

## Overview

Researchers across disciplines face the challenge of analyzing large
volumes of textual data — research articles, social media posts,
customer reviews, survey responses, legal documents, and literary works.
While programming languages such as R and Python offer powerful NLP
capabilities, not all researchers have the time or expertise to use them
effectively.

TALL bridges this gap by providing a general-purpose, code-free text
analysis platform built on the R ecosystem. It combines the statistical
rigor of established R packages with the accessibility of a modern web
interface, enabling researchers to conduct reproducible analyses from
import through visualization without writing a single line of code.

------------------------------------------------------------------------

## Workflow

TALL follows a structured analytical workflow that guides users from raw
text to interpretable results:

![](images/workflow.png)

The workflow consists of three main stages:

### 1. Import and Manipulation

TALL supports multiple input formats (plain text, CSV, Excel, PDF,
Biblioshiny exports) and provides tools for corpus splitting, random
sampling, and integration of external metadata. Analysis sessions can be
saved and reloaded as `.tall` files for full reproducibility.

### 2. Pre-processing and Cleaning

Linguistic pre-processing is powered by **UDPipe** with updated
Universal Dependencies v2.15 language models, supporting 60+ languages.
The pre-processing pipeline includes:

- **Tokenization** — splitting raw text into words and sentences
- **Lemmatization** — reducing words to their dictionary base form
- **Part-of-Speech tagging** — assigning grammatical categories (noun,
  verb, adjective, etc.)
- **Dependency parsing** — identifying syntactic relationships between
  words
- **Special entity detection** — tagging emails, URLs, hashtags,
  mentions, emojis, and IP addresses
- **Multi-word expression extraction** — six methods including syntactic
  dependency parsing, RAKE, PMI, Mutual Dependency, Log-Frequency Biased
  MD, and IS Index
- **Custom term lists and synonym merging** — domain-specific vocabulary
  management

### 3. Statistical Text Analysis and Dynamic Visualization

TALL offers a rich set of analytical methods organized across three
sections: **Overview**, **Words**, and **Documents**.

------------------------------------------------------------------------

## Analytical Methods

### Overview

Corpus-level descriptive statistics provide a quantitative profile of
the text collection:

- **Corpus metrics** — document count, tokens, types, lemmas, sentences,
  with averages and standard deviations
- **Lexical richness indices** — Type-Token Ratio (TTR), Hapax Legomena,
  Guiraud Index, Yule’s K
- **Stylistic indices** — Lexical Density, Nominal Ratio, Gini Index
- **Frequency distributions** — word frequency, TF-IDF rankings, Zipf’s
  law visualization, word clouds
- **Morphological features** — distribution of grammatical features
  (Tense, Mood, Number, Person, VerbForm, Degree, Gender, Case, Voice)
  extracted from Universal Dependencies annotation
- **Dependency tree viewer** — interactive visualization of syntactic
  parse trees for individual sentences, with color-coded Part-of-Speech
  tags and labeled dependency arcs

### Words

Word-level analyses reveal the internal structure and thematic
organization of the corpus:

- **Keyness analysis** — statistical comparison of word frequencies
  across groups using Chi-squared and Log-Likelihood Ratio tests, with
  frequency-context plots and word clouds
- **Words in Context (KWIC)** — concordance analysis with customizable
  context windows, co-occurrence network visualization, and distribution
  insights
- **Correspondence Analysis** — factorial mapping of word-document
  associations with hierarchical clustering and dimensional
  interpretation
- **Co-occurrence Network** — weighted word co-occurrence networks with
  multiple normalization indices (Association, Jaccard, Inclusion,
  Salton, Equivalence), Louvain community detection, and adjustable
  community repulsion for visual cluster separation
- **Thematic Map** — strategic diagram positioning topics along Callon’s
  Centrality and Density dimensions (motor themes, basic themes, niche
  themes, emerging/declining themes)
- **Word Embeddings** — Word2Vec model training with similarity network
  analysis, UMAP dimensionality reduction, and cluster-based semantic
  exploration

### Documents

Document-level analyses operate on entire texts and their structural
properties:

- **Topic Modeling** — Latent Dirichlet Allocation (LDA), Correlated
  Topic Models (CTM), and Structural Topic Models (STM), with automated
  K selection using four complementary metrics (CaoJuan 2009, Arun 2010,
  Deveaud 2014, Perplexity/Lower Bound), consensus recommendation, model
  diagnostics, and covariate effects analysis
- **Syntactic Complexity** — dependency-based metrics including Mean
  Dependency Distance (Liu, 2008), parse tree depth, clauses per
  sentence, subordination ratio, and branching factor, computed per
  document with corpus-level summaries and distribution plots
- **SVO Triplet Extraction** — Subject-Verb-Object relationship
  extraction using dependency parsing, with frequency tables, Sankey
  flow diagrams, and verb frequency analysis for content and narrative
  structure analysis
- **Polarity Detection** — lexicon-based sentiment analysis with
  contextual valence shifters (negators, amplifiers, diminishers),
  supporting Hu & Liu, Loughran & McDonald, and NRC lexicons.
  Document-level polarity scoring with distribution visualization
- **Emotion Analysis** — eight-emotion detection (Anger, Anticipation,
  Disgust, Fear, Joy, Sadness, Surprise, Trust) using the NRC
  Word-Emotion Association Lexicon (EmoLex), with emotion distribution
  charts, radar plots, per-document heatmaps, and top contributing words
  per emotion
- **Text Summarization** — extractive summarization via the TextRank
  algorithm (graph-based sentence ranking with PageRank scoring), and
  abstractive summarization powered by Google Gemini AI

### TALL AI

TALL integrates **Google Gemini AI** as an intelligent assistant that
provides automated interpretation of analytical results. Available
across most analysis tabs (Overview, KWIC, Correspondence Analysis,
Co-occurrence Network, Thematic Map, Word Embeddings, Topic Modeling,
Polarity Detection, Emotion Analysis, Syntactic Complexity, SVO
Triplets), TALL AI examines the visual and numerical outputs and
generates contextual, academically-grounded interpretations. AI calls
run asynchronously, keeping the application responsive during
processing.

### Reporting

All analyses can be exported to an Excel workbook with embedded plots,
enabling reproducible reporting. Individual plots can be exported as
high-resolution PNG images with configurable DPI settings. Network
visualizations use native canvas capture for crisp, DPI-aware rendering.

------------------------------------------------------------------------

## Screenshots

### Import text from multiple file formats

![](images/import.png)

### Edit, divide, and add external information

![](images/edit.png)

### Automatic Lemmatization and PoS-Tagging

![](images/Tokenization&PoSTagging.png)

### Language, Model, and Analysis Term Selection

![](images/Language,%20Model,%20and%20Analysis%20Term%20Selection.png)

### Special Entity Tagging

![](images/SpecEntitiesTagging1.png)
![](images/SpecEntitiesTagging2.png)

### Multi-word Expression Extraction

![](images/automatic_multiword.png) ![](images/bylist_customlist.png)

### Overview — Descriptive statistics, concordance analysis, word frequency distributions

![](images/overview.png)

### Words — Topic detection, correspondence analysis, co-occurrence networks

![](images/words.png)

### Documents — Topic modeling, sentiment analysis, syntactic analysis

![](images/documents.png)

------------------------------------------------------------------------

## Authors

### Creators

- **Massimo Aria** — University of Naples Federico II
  - email: <aria@unina.it> \| ORCID:
    [0000-0002-8517-9411](https://orcid.org/0000-0002-8517-9411)
- **Maria Spano** — University of Naples Federico II
  - email: <maria.spano@unina.it> \| ORCID:
    [0000-0002-3103-2342](https://orcid.org/0000-0002-3103-2342)
- **Luca D’Aniello** — University of Naples Federico II
  - email: <luca.daniello@unina.it> \| ORCID:
    [0000-0003-1019-9212](https://orcid.org/0000-0003-1019-9212)
- **Corrado Cuccurullo** — University of Campania Luigi Vanvitelli
  - email: <corrado.cuccurullo@unicampania.it> \| ORCID:
    [0000-0002-7401-8575](https://orcid.org/0000-0002-7401-8575)

### Contributors

- **Michelangelo Misuraca** — University of Salerno
  - email: <mmisuraca@unisa.it> \| ORCID:
    [0000-0002-8794-966X](https://orcid.org/0000-0002-8794-966X)

### Maintainer

Massimo Aria (<aria@unina.it>)

------------------------------------------------------------------------

## License

MIT License. Copyright 2023-2026 Massimo Aria.

See [LICENSE](LICENSE) for details.

# Acknowledgment
# This function is based on the source code of the webshot2 package.
# We would like to acknowledge and express our gratitude to the authors
# and contributors of webshot2 for their valuable work in developing tools
# for web content capture.
#
# In this implementation, we have adapted and modified parts of the original
# webshot2 code to address an issue where PNG files were being generated as
# empty images when used on Windows systems. Our modifications aim to improve
# compatibility and ensure reliable output across different platforms.
#
# Reference:
# webshot2 package: https://github.com/rstudio/webshot2

tallShot <- function(
    url = NULL,
    file = "tallShot.png",
    vwidth = 992,
    vheight = 744,
    selector = NULL,
    cliprect = NULL,
    expand = NULL,
    delay = 0.3,
    zoom = 1,
    useragent = NULL,
    max_concurrent = getOption("tallShot.concurrent", default = 6),
    verbose = FALSE) {
  if (length(url) == 0) {
    stop("Need url.")
  }

  url <- vapply(
    url,
    function(x) {
      if (!is_url(x)) {
        file_url(x)
      } else {
        x
      }
    },
    character(1)
  )

  if (!is.null(cliprect) && !is.list(cliprect)) cliprect <- list(cliprect)
  if (!is.null(selector) && !is.list(selector)) selector <- list(selector)
  if (!is.null(expand) && !is.list(expand)) expand <- list(expand)

  if (is.null(selector)) {
    selector <- "html"
  }

  if (length(url) > 1 && length(file) == 1) {
    file <- vapply(1:length(url), FUN.VALUE = character(1), function(i) {
      replacement <- sprintf("%03d.\\1", i)
      gsub("\\.(.{3,4})$", replacement, file)
    })
  }

  args_all <- list(
    url = url,
    file = file,
    vwidth = vwidth,
    vheight = vheight,
    selector = selector,
    cliprect = cliprect,
    expand = expand,
    delay = delay,
    zoom = zoom,
    useragent = useragent,
    verbose = verbose
  )

  n_urls <- length(url)
  args_all <- mapply(args_all, names(args_all),
                     FUN = function(arg, name) {
                       if (length(arg) == 0) {
                         return(vector(mode = "list", n_urls))
                       } else if (length(arg) == 1) {
                         return(rep(arg, n_urls))
                       } else if (length(arg) == n_urls) {
                         return(arg)
                       } else {
                         stop("Argument `", name, "` should be NULL, length 1, or same length as `url`.")
                       }
                     },
                     SIMPLIFY = FALSE
  )

  args_all <- long_to_wide(args_all)

  if (!requireNamespace("chromote", quietly = TRUE)) {
    stop(
      "Package 'chromote' is required to export plots as PNG. ",
      "Please install it with: install.packages('chromote')",
      call. = FALSE
    )
  }
  cm <- chromote::default_chromote_object()

  # A list of promises for the screenshots
  res <- lapply(
    args_all,
    function(args) {
      new_session_screenshot(
        cm,
        args$url, args$file, args$vwidth, args$vheight, args$selector,
        args$cliprect, args$expand, args$delay, args$zoom, args$useragent,
        verbose
      )
    }
  )

  p <- promises::promise_all(.list = res)
  res <- cm$wait_for(p)
  res <- structure(unlist(res), class = "webshot")
  res
}


new_session_screenshot <- function(
    chromote,
    url,
    file,
    vwidth,
    vheight,
    selector,
    cliprect,
    expand,
    delay,
    zoom,
    useragent,
    verbose = FALSE) {
  filetype <- tolower(tools::file_ext(file))
  if (filetype != "png" && filetype != "pdf") {
    stop("File extension must be 'png' or 'pdf'")
  }

  if (is.null(selector)) {
    selector <- "html"
  }

  if (is.character(cliprect)) {
    if (cliprect == "viewport") {
      cliprect <- c(0, 0, vwidth, vheight)
    } else {
      stop("Invalid value for cliprect: ", cliprect)
    }
  } else {
    if (!is.null(cliprect) && !(is.numeric(cliprect) && length(cliprect) == 4)) {
      stop("`cliprect` must be a vector with four numbers, or a list of such vectors")
    }
  }

  s <- NULL

  p <- chromote$new_session(
    wait_ = FALSE,
    width = vwidth,
    height = vheight
  )$
    then(function(session) {
      s <<- session

      if (!is.null(useragent)) {
        s$Network$setUserAgentOverride(userAgent = useragent)
      }
      res <- s$Page$loadEventFired(wait_ = FALSE)
      s$Page$navigate(url, wait_ = FALSE)
      res
    })$
    then(function(value) {
      if (delay > 0) {
        promises::promise(function(resolve, reject) {
          later::later(
            function() {
              resolve(value)
            },
            delay
          )
        })
      } else {
        value
      }
    })$
    then(function(value) {
      if (filetype == "png") {
        s$screenshot(
          filename = file, selector = selector, cliprect = cliprect,
          expand = expand, scale = zoom,
          show = FALSE, wait_ = FALSE
        )
      } else if (filetype == "pdf") {
        s$screenshot_pdf(filename = file, wait_ = FALSE)
      }
    })$
    then(function(value) {
      if (verbose) message(url, " screenshot completed")
      normalizePath(value)
    })$
    finally(function() {
      s$close()
    })

  p
}

is_url <- function(x) {
  grepl("^[a-zA-Z]+://", x)
}

# Given the path to a file, return a file:// URL.
file_url <- function(filename) {
  if (is_windows()) {
    paste0("file://", normalizePath(filename, mustWork = TRUE))
  } else {
    enc2utf8(paste0("file:///", normalizePath(filename, winslash = "/", mustWork = TRUE)))
  }
}

is_windows <- function() .Platform$OS.type == "windows"

is_mac <- function() Sys.info()[["sysname"]] == "Darwin"

is_linux <- function() Sys.info()[["sysname"]] == "Linux"

long_to_wide <- function(x) {
  if (length(x) == 0) {
    return(x)
  }

  lapply(seq_along(x[[1]]), function(n) {
    lapply(stats::setNames(names(x), names(x)), function(name) {
      x[[name]][[n]]
    })
  })
}

vocabulary <- function(dfTag, term) {
  dictFreq <- LemmaSelection(dfTag) %>%
    dplyr::filter(docSelected) %>%
    mutate(token = ifelse(upos == "MULTIWORD", lemma, token))

  if (term == "lemma") {
    dictFreq <- dictFreq %>%
      group_by(upos, lemma) %>%
      summarize(n = n()) %>%
      arrange(desc(n)) %>%
      rename(
        Lemma = lemma,
        Frequency = n,
        "Part of Speech" = upos
      ) %>%
      relocate("Part of Speech", .after = last_col())
  } else {
    dictFreq <- dictFreq %>%
      group_by(upos, token) %>%
      summarize(n = n()) %>%
      arrange(desc(n)) %>%
      rename(
        Token = token,
        Frequency = n,
        "Part of Speech" = upos
      ) %>%
      relocate("Part of Speech", .after = last_col())
  }
  return(dictFreq)
}

# tfidf calculation
tfidfTable <- function(dfTag, term) {
  tfidfDATA <- LemmaSelection(dfTag) %>%
    dplyr::filter(docSelected) %>%
    tfidf(term = term)

  if (term == "lemma") {
    tfidfDATA <- tfidfDATA %>%
      rename(
        "Lemma" = term,
        "TF-IDF" = TFIDF
      )
  } else {
    tfidfDATA <- tfidfDATA %>%
      rename(
        "Token" = term,
        "TF-IDF" = TFIDF
      )
  }
  return(tfidfDATA)
}

# Term Frequency Distributions
freqByPos <- function(df, term = "lemma", pos = "NOUN") {
  obj <- df %>%
    dplyr::filter(upos %in% pos) %>%
    count(term = .[[term]]) %>%
    arrange(desc(n))

  if ("URL" %in% pos) {
    obj$term <- paste0(
      '<a href=\"',
      obj$term,
      '\" target=\"_blank\">',
      obj$term,
      "</a>"
    )
  }
  return(obj)
}

# freqPlotly ----
freqPlotly <- function(
  dfPlot,
  x,
  y,
  n = 10,
  xlabel,
  ylabel,
  scale = c("identity", "log"),
  topicmodel = FALSE,
  color = "#4F7942",
  decimal = 0
) {
  # function to build and plot plotly horizontal barplot
  dfPlot <- dfPlot %>% dplyr::slice_head(n = n)
  xmax <- max(dfPlot[[x]])

  switch(scale, log = {
    dfPlot$n <- log(dfPlot$n)
  })

  if (isTRUE(topicmodel)) {
    fig1 <- plot_ly(
      data = dfPlot,
      x = round(dfPlot[[x]], decimal),
      y = dfPlot[[y]],
      source = "A",
      type = "bar",
      orientation = "h",
      marker = list(
        color = paste0(color, "60"),
        line = list(color = color, width = 1)
      ),
      hovertemplate = "<b>Word</b>: <i>%{y}</i> <br><b><i>Value: %{x}</i></b><extra></extra>"
    )
  } else {
    fig1 <- plot_ly(
      data = dfPlot,
      x = dfPlot[[x]],
      y = ~ reorder(dfPlot[[y]], dfPlot[[x]]),
      source = "A",
      type = "bar",
      orientation = "h",
      marker = list(
        color = paste0(color, "60"),
        line = list(color = color, width = 1)
      ),
      hovertemplate = "<b><i>Word: %{y}</i></b> <br> <b><i>Value: %{x}</i></b><extra></extra>"
    )
  }

  fig1 <- fig1 %>%
    layout(
      yaxis = list(
        title = ylabel,
        showgrid = FALSE,
        showline = FALSE,
        showticklabels = TRUE,
        domain = c(0, 1)
      ),
      xaxis = list(
        title = xlabel,
        zeroline = FALSE,
        showline = FALSE,
        showticklabels = TRUE,
        showgrid = FALSE
      ),
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)"
    )

  if (decimal > 0) {
    ann_text <- format(round(dfPlot[[x]], decimal), nsmall = decimal)
  } else {
    ann_text <- dfPlot[[x]]
  }

  fig1 <- fig1 %>%
    add_annotations(
      xref = "x1",
      yref = "y",
      x = dfPlot[[x]] + xmax * 0.015,
      y = dfPlot[[y]],
      text = ann_text,
      font = list(family = "Arial", size = 12, color = color),
      showarrow = FALSE
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )
  # %>%
  #   event_register("plotly_selecting")

  fig1
}

# ValueBoxes Indices ----
valueBoxesIndices <- function(x) {
  x <- x %>%
    filter(!upos %in% c("PUNCT", "SYM", "NUM", "X"))

  # 1. # of documents
  nDoc <- length(unique(x$doc_id))

  # 2. # of tokens
  nTokens <- nrow(x)

  # 3. Dictionary: # di termini differenti senza ripetizione
  nDictionary <- length(unique(x$token))

  # 4. # of lemmas
  nLemmas <- length(unique(x$lemma))

  # 5. # of sentences
  # nSentences <- x %>%
  #   group_by(doc_id) %>%
  #   reframe(nSent = max(sentence_id)) %>%
  #   ungroup() %>%
  #   select(nSent) %>%
  #   reframe(nSent = sum(nSent)) %>%
  #   as.numeric()
  nSentences <- x %>%
    select(doc_id, sentence) %>%
    distinct() %>%
    nrow()

  # 6. # avg document length
  # avgDocLength <- x %>%
  #   group_by(doc_id) %>%
  #   select(doc_id, sentence) %>%
  #   reframe(
  #     nTokens = n(),
  #     nChars = nchar(paste(sentence, collapse = " "))
  #   ) %>%
  #   ungroup() %>%
  #   reframe(
  #     avgChars = round(mean(nChars), 0),
  #     sdChars = round(sd(nChars), 0),
  #     avgTokens = round(mean(nTokens), 0),
  #     sdTokens = round(sd(nTokens), 0)
  #   )

  avgDocLength <- x %>%
    group_by(doc_id) %>%
    summarise(
      char_length = sum(nchar(token), na.rm = TRUE),
      token_length = n()
    ) %>%
    summarise(
      avgChars = mean(char_length, na.rm = TRUE),
      sdChars = sd(char_length, na.rm = TRUE),
      avgTokens = mean(token_length, na.rm = TRUE),
      sdTokens = sd(token_length, na.rm = TRUE)
    )

  # 7. # avg length sentence
  # avgSentLength <- x %>%
  #   group_by(doc_id, sentence_id) %>%
  #   reframe(
  #     sentLength = n(),
  #     nChars = nchar(sentence)
  #   ) %>%
  #   ungroup() %>%
  #   reframe(
  #     avgTokens = round(mean(sentLength), 1),
  #     sdTokens = round(sd(sentLength), 1),
  #     avgChars = round(mean(nChars), 1),
  #     sdChars = round(sd(nChars), 1)
  #   )
  avgSentLength <- x %>%
    group_by(doc_id, sentence_id) %>%
    summarise(
      char_length = sum(nchar(token), na.rm = TRUE),
      token_length = n(),
      .groups = "drop"
    ) %>%
    summarise(
      avgChars = mean(char_length, na.rm = TRUE),
      sdChars = sd(char_length, na.rm = TRUE),
      avgTokens = mean(token_length, na.rm = TRUE),
      sdTokens = sd(token_length, na.rm = TRUE)
    )

  # 8. TTR: il rapporto tra la varietà del dizionario (Dictionary) e il numero totale di token in una raccolta testuale (# terms); in altre parole, misura la diversità lessicale in un corpus
  TTR <- round(nDictionary / nTokens * 100, 2)

  # 9.  %hapax
  hapax <- x %>%
    group_by(token) %>%
    count() %>%
    filter(n == 1) %>%
    ungroup() %>%
    reframe(n = sum(n)) %>%
    as.numeric() /
    nDictionary *
    100

  # 10. Guiraud
  guiraud <- round(nDictionary / sqrt(nTokens), 1)

  # 11. Lexical density
  lexical_words <- x %>%
    filter(upos %in% c("NOUN", "VERB", "ADJ", "ADV")) # Parole di contenuto

  lexical_density <- (nrow(lexical_words) / nrow(x)) * 100

  # 12. Nominal Ratio (Rapporto Nominale)
  num_nouns <- sum(x$upos == "NOUN", na.rm = TRUE) # Numero di sostantivi
  num_verbs <- sum(x$upos == "VERB", na.rm = TRUE) # Numero di verbi

  nominal_ratio <- ifelse(num_verbs > 0, num_nouns / num_verbs, NA) # Evita divisione per zero

  # 13. Gini Index sui Token (Disomogeneità della Distribuzione)
  token_counts <- x %>%
    count(token, sort = TRUE) %>%
    pull(n) # Ottieni solo il conteggio delle occorrenze

  gini_index <- Gini(token_counts) # Calcola l'indice di Gini sulla distribuzione delle parole

  # 14 Yule’s K (Misura della diversità lessicale)
  words <- x$token[x$token != ""] # Remove empty strings
  word_freq <- table(words)
  M1 <- length(word_freq)
  M2 <- sum(word_freq^2)
  K <- 10000 * (M2 - M1) / (length(words)^2)

  obj <- list(
    nDoc = nDoc,
    nTokens = nTokens,
    nDictionary = nDictionary,
    nLemmas = nLemmas,
    nSentences = nSentences,
    avgDocLengthChars = round(avgDocLength$avgChars, 0),
    avgDocLengthCharsSD = round(avgDocLength$sdChars, 0),
    avgDocLengthTokens = round(avgDocLength$avgTokens, 0),
    avgDocLengthTokensSD = round(avgDocLength$sdTokens, 0),
    avgSentLengthTokens = round(avgSentLength$avgTokens, 0),
    avgSentLengthTokensSD = round(avgSentLength$sdTokens, 0),
    avgSentLengthChars = round(avgSentLength$avgChars, 0),
    avgSentLengthCharsSD = round(avgSentLength$sdChars, 0),
    TTR = TTR,
    hapax = round(hapax, 1),
    guiraud = guiraud,
    lexical_density = round(lexical_density, 1),
    nominal_ratio = round(nominal_ratio, 2),
    gini_index = round(gini_index, 2),
    yule_k = round(K, 1)
  )
}

Gini <- function(x, corr = FALSE, na.rm = TRUE) {
  if (!na.rm && any(is.na(x))) {
    return(NA_real_)
  }
  x <- as.numeric(na.omit(x))
  n <- length(x)
  x <- sort(x)
  G <- sum(x * 1L:n)
  G <- 2 * G / sum(x) - (n + 1L)
  if (corr) G / (n - 1L) else G / n
}

## wordcloud2vis

# wordcloud2vis <- function(nodes, labelsize = 7, opacity = 1) {
#   nodes <- nodes %>%
#     mutate(id = row_number())
#   # size scaling
#   scalemin <- 20 * (1 + labelsize / 5)
#   scalemax <- 100 * (1 + labelsize / 5)
#   N <- nrow(nodes)
#
#   colorlists <- colorlist()
#   colorlists <- sample(colorlists, N, replace = TRUE)
#
#   opacity.min <- 0.6
#   shape <- "text"
#   layout <- "layout_nicely"
#
#   nodes <- nodes %>%
#     mutate(
#       font.color = colorlists,
#       id = row_number(),
#       shape = shape,
#       color = colorlists,
#       title = paste(
#         "<strong>",
#         label,
#         "</strong>",
#         "<br><h5>freq = ",
#         value,
#         "</h5>",
#         sep = ""
#       )
#     )
#
#   nodes$font.size <- log(nodes$value)
#   Min <- min(nodes$font.size)
#   Max <- max(nodes$font.size)
#   if (Max > Min) {
#     size <- (nodes$font.size - Min) / (Max - Min) * 15 * labelsize + 10
#   } else {
#     size <- 10 * labelsize
#   }
#   size[size < scalemin] <- scalemin
#   size[size > scalemax] <- scalemax
#   nodes$font.size <- size
#
#   if (shape %in% c("dot", "square")) {
#     nodes$font.vadjust <- -0.7 * nodes$font.size
#   } else {
#     nodes$font.vadjust <- 0
#   }
#
#   ## opacity for label
#   opacity_font <- sqrt(
#     (nodes$font.size - min(nodes$font.size)) / diff(range(nodes$font.size))
#   ) *
#     opacity +
#     opacity.min +
#     0.1
#   if (is.nan(opacity_font[1])) {
#     opacity_font <- rep(opacity.min, length(opacity_font))
#   }
#
#   # node colors
#   nodes$opacity.nodes <- (opacity_font - min(opacity_font)) /
#     (diff(range(opacity_font))) *
#     0.5 +
#     opacity.min
#   nodes$opacity.nodes[is.nan(nodes$opacity.nodes)] <- 0.5
#
#   VIS <-
#     visNetwork::visNetwork(
#       nodes = nodes,
#       edges = NULL,
#       type = "full",
#       smooth = TRUE,
#       physics = TRUE
#     ) %>%
#     visNetwork::visNodes(
#       shadow = FALSE,
#       shape = nodes$shape,
#       font = list(
#         color = nodes$font.color,
#         size = nodes$font.size,
#         vadjust = nodes$font.vadjust
#       )
#     ) %>%
#     visNetwork::visOptions(
#       highlightNearest = list(enabled = T, hover = T, degree = 1),
#       nodesIdSelection = T
#     ) %>%
#     visNetwork::visInteraction(
#       dragNodes = TRUE,
#       navigationButtons = F,
#       hideEdgesOnDrag = TRUE,
#       zoomSpeed = 0.2
#     ) %>%
#     visEvents(
#       click = "function(nodes){
#                   Shiny.onInputChange('click', nodes.nodes[0]);
#                   ;}"
#     ) %>%
#     visNetwork::visOptions(
#       manipulation = FALSE,
#       height = "100%",
#       width = "100%"
#     )
#   return(VIS)
# }

## wordcloud function
# wordcloud2a <- function(
#   data,
#   size = 1,
#   minSize = 0,
#   gridSize = 0,
#   fontFamily = "Segoe UI",
#   fontWeight = "bold",
#   color = "random-dark",
#   backgroundColor = "transparent",
#   minRotation = -pi / 4,
#   maxRotation = pi / 4,
#   shuffle = TRUE,
#   rotateRatio = 0.4,
#   shape = "circle",
#   ellipticity = 0.65,
#   widgetsize = NULL,
#   figPath = NULL,
#   hoverFunction = NULL
# ) {
#   if ("table" %in% class(data)) {
#     dataOut <- data.frame(name = names(data), freq = as.vector(data))
#   } else {
#     data <- as.data.frame(data)
#     dataOut <- data[, 1:2]
#     names(dataOut) <- c("name", "freq")
#   }
#   if (!is.null(figPath)) {
#     if (!file.exists(figPath)) {
#       stop("cannot find fig in the figPath")
#     }
#     spPath <- strsplit(figPath, "\\.")[[1]]
#     len <- length(spPath)
#     figClass <- spPath[len]
#     if (!figClass %in% c("jpeg", "jpg", "png", "bmp", "gif")) {
#       stop("file should be a jpeg, jpg, png, bmp or gif file!")
#     }
#     base64 <- base64enc::base64encode(figPath)
#     base64 <- paste0(
#       "data:image/",
#       figClass,
#       ";base64,",
#       base64
#     )
#   } else {
#     base64 <- NULL
#   }
#   weightFactor <- size * 180 / max(dataOut$freq)
#   settings <- list(
#     word = dataOut$name,
#     freq = dataOut$freq,
#     fontFamily = fontFamily,
#     fontWeight = fontWeight,
#     color = color,
#     minSize = minSize,
#     weightFactor = weightFactor,
#     backgroundColor = backgroundColor,
#     gridSize = gridSize,
#     minRotation = minRotation,
#     maxRotation = maxRotation,
#     shuffle = shuffle,
#     rotateRatio = rotateRatio,
#     shape = shape,
#     ellipticity = ellipticity,
#     figBase64 = base64,
#     hover = htmlwidgets::JS(hoverFunction)
#   )
#   chart <- htmlwidgets::createWidget(
#     "wordcloud2",
#     settings,
#     width = widgetsize[1],
#     height = widgetsize[2],
#     sizingPolicy = htmlwidgets::sizingPolicy(
#       viewer.padding = 0,
#       browser.padding = 0,
#       browser.fill = TRUE
#     )
#   )
#   chart
# }

## TFIDF functions ----
tfidf <- function(dfTag, term = "lemma", document = "doc_id") {
  # calculate tfidf
  dtm <- dfTag %>% dplyr::filter(POSSelected)
  dtm <- document_term_frequencies(dtm, document = document, term = term)
  dtm <- document_term_matrix(dtm)
  tfidf <- dtm_tfidf(dtm)
  tibble(term = names(tfidf), TFIDF = as.numeric(tfidf)) %>%
    arrange(desc(tfidf))
}

### WORDCLOUD ----
#' Create a word cloud using ggwordcloud
#'
#' This function creates customizable word clouds using the ggwordcloud package.
#' It supports various shapes, colors, rotations, and faceting for comparative visualizations.
#'
#' @param data A data frame with at least two columns: the first column contains words,
#'   the second column contains the size/frequency of words. Additional columns can be
#'   used for faceting.
#' @param col_names Optional character vector of length 2 specifying column names
#'   for words and sizes. If NULL, uses first two columns (default: NULL)
#' @param max_size Maximum font size for words (default: 20)
#' @param min_size Minimum font size for words (default: 2)
#' @param eccentricity Eccentricity of the spiral (default: 0.65).
#'   Values > 1 make it wider, < 1 make it taller
#' @param shape Shape of the word cloud. One of "circle", "cardioid", "diamond",
#'   "square", "triangle-forward", "triangle-upright", "pentagon", "star" (default: "circle")
#' @param rm_outside Remove words that don't fit (default: FALSE)
#' @param rot_per Proportion of words with 90 degree rotation (default: 0.1)
#' @param colors Color palette or single color for words (default: "black").
#'   When faceting and length(colors) equals the number of facets, each facet
#'   will use a single color. Otherwise, colors are assigned to individual words.
#' @param color_by_size If TRUE, color words by their size using a gradient (default: FALSE)
#' @param color_gradient_low Low end of color gradient when color_by_size = TRUE (default: "darkblue")
#' @param color_gradient_high High end of color gradient when color_by_size = TRUE (default: "lightblue")
#' @param seed Random seed for reproducibility (default: NA)
#' @param area_corr Use area correction - font size proportional to sqrt of size (default: TRUE)
#' @param rstep Radial step size for spiral (default: 0.01)
#' @param tstep Angular step size for spiral (default: 0.02)
#' @param grid_size Grid size for collision detection (default: 4)
#' @param max_steps Maximum number of steps to find position for each word (default: 10)
#' @param xlim X-axis limits as c(min, max) (default: c(NA, NA))
#' @param ylim Y-axis limits as c(min, max) (default: c(NA, NA))
#' @param mask Optional mask image (PNG array) for custom shapes (default: NA)
#' @param show_boxes Show bounding boxes for debugging (default: FALSE)
#' @param background_color Background color of the plot (default: "white")
#' @param facet_by Optional column name for faceting. Creates separate word clouds
#'   for each unique value in this column (default: NULL)
#' @param facet_ncol Number of columns in facet layout (default: NULL, automatic)
#' @param facet_nrow Number of rows in facet layout (default: NULL, automatic)
#' @param facet_scales Control facet scales: "fixed", "free", "free_x", or "free_y" (default: "fixed")
#' @param facet_labeller Labeller for facet titles (default: "label_value")
#' @param facet_text_size Font size for facet titles (default: 16)
#' @param facet_text_face Font face for facet titles: "plain", "bold", "italic", or "bold.italic" (default: "bold")
#' @param facet_text_color Color for facet titles (default: "black")
#' @param ... Additional arguments passed to geom_text_wordcloud_area or geom_text_wordcloud
#'
#' @return A ggplot object with the word cloud visualization
#'
#' @examples
#' library(ggwordcloud)
#' library(dplyr)
#'
#' # Simple word cloud
#' df <- data.frame(
#'   word = c("statistics", "machine learning", "data", "analysis", "visualization"),
#'   freq = c(50, 45, 40, 30, 25)
#' )
#' wordcloud(df)
#'
#' # With custom colors and rotation
#' wordcloud(df,
#'           colors = "darkblue",
#'           rot_per = 0.3,
#'           shape = "diamond",
#'           max_size = 30)
#'
#' # With color gradient by size
#' wordcloud(df,
#'           color_by_size = TRUE,
#'           color_gradient_low = "blue",
#'           color_gradient_high = "red",
#'           seed = 42)
#'
#' # Multiple colors for individual words
#' colors <- c("#E41A1C", "#377EB8", "#4DAF4A", "#984EA3", "#FF7F00")
#' wordcloud(df,
#'           colors = colors,
#'           seed = 123)
#'
#' # With faceting - comparing two groups
#' df_facet <- data.frame(
#'   word = c("research", "analysis", "study", "data", "method",
#'            "treatment", "patient", "disease", "clinical", "therapy"),
#'   freq = c(100, 85, 75, 60, 55, 90, 80, 70, 65, 50),
#'   category = c(rep("Research", 5), rep("Clinical", 5))
#' )
#'
#' # One color per facet
#' wordcloud(df_facet,
#'           facet_by = "category",
#'           facet_ncol = 2,
#'           colors = c("#4575B4", "#D73027"),  # 2 colors for 2 facets
#'           facet_text_size = 16,
#'           seed = 42)
#'
#' # With larger facet titles
#' wordcloud(df_facet,
#'           facet_by = "category",
#'           facet_nrow = 1,
#'           colors = c("#E41A1C", "#377EB8"),
#'           facet_text_size = 18,
#'           facet_text_face = "bold",
#'           facet_text_color = "darkblue",
#'           max_size = 25,
#'           seed = 42)
#'
#' # Different shapes and eccentricity
#' wordcloud(df,
#'           shape = "star",
#'           eccentricity = 1.5,
#'           rot_per = 0.2,
#'           colors = "#FF7F00",
#'           max_size = 35,
#'           seed = 42)
#'
#' # Using bind_rows for comparative analysis
#' corpus1 <- data.frame(word = c("land", "restoration", "ecosystem"),
#'                       freq = c(50, 45, 40),
#'                       corpus = "Corpus 1")
#' corpus2 <- data.frame(word = c("disease", "treatment", "patient"),
#'                       freq = c(55, 50, 45),
#'                       corpus = "Corpus 2")
#'
#' wordcloud(bind_rows(corpus1, corpus2),
#'           facet_by = "corpus",
#'           facet_ncol = 2,
#'           colors = c("#1B9E77", "#D95F02"),
#'           facet_text_size = 16,
#'           max_size = 30,
#'           seed = 123)

posSel <- function(dfTag, pos) {
  dfTag <- dfTag %>% mutate(POSSelected = ifelse(upos %in% pos, TRUE, FALSE))
}

# remove Hapax and lowwer and higher lemmas
removeHapaxFreq <- function(dfTag, hapax, singleChar) {
  ## reset noHapax column
  dfTag <- dfTag %>%
    mutate(noHapax = TRUE)

  ## Hapax
  if (is.null(hapax)) {
    H <- dfTag %>%
      group_by(lemma) %>%
      count() %>%
      filter(n == 1) %>%
      select(lemma)
    H <- unique(H$lemma)
    dfTag <- dfTag %>%
      mutate(noHapax = ifelse(lemma %in% H, FALSE, TRUE))
  }

  ## Single Char
  if (is.null(singleChar)) {
    dfTag <- dfTag %>%
      mutate(
        noSingleChar = ifelse(nchar(lemma) > 1, TRUE, FALSE),
        noSingleChar = ifelse(
          upos %in% c("EMOJI", "MENTION", "HASH", "IP_ADDRESS", "URL", "EMAIL"),
          TRUE,
          noSingleChar
        )
      )
  }

  return(dfTag)
}

# Select lemmas by PoS Tags, Hapax and Frequency Range #
LemmaSelection <- function(dfTag) {
  if (!"noHapax" %in% names(dfTag)) {
    dfTag <- dfTag %>%
      mutate(noHapax = TRUE)
  }
  if (!"noSingleChar" %in% names(dfTag)) {
    dfTag <- dfTag %>%
      mutate(noSingleChar = TRUE)
  }

  dfTag <- dfTag %>%
    dplyr::filter(POSSelected, noHapax, noSingleChar) %>%
    arrange(doc_id, paragraph_id, sentence_id)

  return(dfTag)
}


# ## Highlight function ----
highlight_segments <- function(tc, n) {
  segments <- tc$segments
  n <- 10
  id <- tc$terms %>%
    slice_max(order_by = chi_square, n = n)
  id <- id$term
  for (i in 1:length(id)) {
    segments$segment <- stringi::stri_replace_all_fixed(
      segments$segment,
      id[i],
      paste0("<mark><strong>", id[i], "</strong></mark>")
    )
  }
  tc$segments <- segments
  return(tc)
}

highlight_word <- function(input_string, target_word, upos) {
  # Controllo della validità della parola target
  if (
    is.na(target_word) ||
      target_word == "" ||
      upos %in%
        c("DET", "PART", "PUNCT", "X", "SYM", "INTJ", "NUM", "NGRAM_MERGED")
  ) {
    return(input_string)
  }

  # Escape dei caratteri speciali
  target_word_escaped <- gsub(
    "([.\\^$*+?()\\[\\]|\\\\])",
    "\\\\\\1",
    target_word,
    perl = TRUE
  )

  # Controllo se la parola contiene caratteri non alfanumerici
  if (grepl("\\W", target_word, perl = TRUE)) {
    word_boundary <- "" # Nessun confine di parola se contiene caratteri speciali
  } else {
    word_boundary <- "\\b"
  }

  # Sostituzione con evidenziazione HTML
  highlighted_string <- gsub(
    paste0(word_boundary, target_word_escaped, word_boundary),
    paste0("<mark><strong>", target_word, "</strong></mark>"),
    input_string,
    perl = TRUE
  )

  return(highlighted_string)
}

# highlight <- function(df){
#   df <- df %>%
#     mutate(sentence_hl = mapply(highlight_word, sentence, token, upos))
# }

highlight <- function(df, term = "lemma", upos = NULL) {
  # Ensure term is valid
  if (!term %in% c("lemma", "token")) {
    stop("Invalid term. Use 'lemma' or 'token'.")
  }

  # Dynamically select the column based on term argument
  term_col <- sym(term)

  if (is.null(upos)) {
    df <- df %>%
      mutate(sentence_hl = mapply(highlight_word, sentence, !!term_col, upos))
  } else {
    dfUpos <- df %>%
      mutate(id = row_number()) %>%
      filter(upos %in% upos) %>%
      mutate(
        sentence_hl = mapply(highlight_word, sentence, !!term_col, upos)
      ) %>%
      select("id", "sentence_hl")

    df <- df %>%
      mutate(id = row_number()) %>%
      left_join(dfUpos, by = "id") %>%
      mutate(sentence_hl = coalesce(sentence_hl.y, sentence_hl.x)) %>%
      select(-sentence_hl.x, -sentence_hl.y, -id) # Remove the extra columns
  }

  return(df)
}


## saveTall function ----

################################################################################
# COLLOCATION ANALYSIS MODULE
# Implements AntConc-like collocation analysis features:
# - Plot: Distribution of target word across documents with document viewer
# - Collocate: Words that co-occur with target word (with all statistics: MI, LogLik, T-Score)
################################################################################

# ============================================================================
# UI COMPONENT
# ============================================================================

collocationUI <- function() {
  tabItem(
    tabName = "kwic",
    fluidRow(
      column(
        12,
        h2(icon("magnifying-glass"), strong("KWIC Analysis"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
      )
    ),
    fluidRow(
      column(
        12,
        box(
          width = 12,
          status = "success",
          solidHeader = FALSE,
          collapsible = FALSE,
          tabsetPanel(
            id = "collocationTabs",

            # ==========================================
            # TAB 1: PLOT - Word Distribution
            # ==========================================
            tabPanel(
              title = "In-Document Plot",
              icon = icon("chart-area"),
              fluidPage(
                br(),
                fluidRow(
                  column(
                    12,
                    wellPanel(
                      style = "background: white; border: 1px solid #ddd; padding: 15px;",
                      fluidRow(
                        column(
                          3,
                          h4(
                            strong("Search Settings"),
                            style = "color: #4F7942; margin-top: 0;"
                          ),

                          # Search query
                          textInput(
                            "collocPlotQuery",
                            "Target Word:",
                            placeholder = "Enter word to search..."
                          ),

                          # Case sensitive option
                          checkboxInput(
                            "collocPlotCase",
                            "Case Sensitive",
                            value = FALSE
                          )
                        ),
                        column(
                          3,
                          h4(
                            strong("Display Options"),
                            style = "color: #4F7942; margin-top: 0;"
                          ),

                          # Sort options
                          selectInput(
                            "collocPlotSort",
                            "Sort Documents By:",
                            choices = c(
                              #"Row Number" = "row",
                              "Document ID" = "doc_id",
                              "Frequency" = "freq",
                              "Dispersion" = "dispersion"
                            ),
                            selected = "freq"
                          )
                        ),
                        column(
                          6,
                          # Results summary
                          htmlOutput("collocPlotSummary")
                        )
                      ),
                      fluidRow(
                        column(
                          12,
                          # Search button
                          actionButton(
                            "collocPlotSearch",
                            "Search",
                            icon = icon("search"),
                            class = "btn-success",
                            style = "width: 200px; margin-top: 10px;"
                          )
                        )
                      )
                    )
                  )
                ),

                fluidRow(
                  column(
                    12,
                    # Documents table with inline plots
                    h4(strong("Documents"), style = "color: #4F7942;"),
                    shinycssloaders::withSpinner(
                      DT::DTOutput("collocPlotTable"),
                      color = "#4F7942"
                    )
                  )
                )
              )
            ),

            # ==========================================
            # TAB 2: COLLOCATE
            # ==========================================
            tabPanel(
              title = "Collocation",
              icon = icon("table"),
              fluidPage(
                br(),
                fluidRow(
                  column(
                    12,
                    wellPanel(
                      style = "background: white; border: 1px solid #ddd; padding: 15px;",
                      fluidRow(
                        column(
                          3,
                          h4(
                            strong("Search Settings"),
                            style = "color: #4F7942; margin-top: 0;"
                          ),

                          # Search query
                          textInput(
                            "collocCollocQuery",
                            "Target Word:",
                            placeholder = "Enter word to search..."
                          ),

                          # Case sensitive
                          checkboxInput(
                            "collocCollocCase",
                            "Case Sensitive",
                            value = FALSE
                          )
                        ),
                        column(
                          3,
                          h4(
                            strong("Window Settings"),
                            style = "color: #4F7942; margin-top: 0;"
                          ),

                          # Window span
                          fluidRow(
                            column(
                              6,
                              numericInput(
                                "collocCollocLeft",
                                "Left Span:",
                                value = 2,
                                min = 0,
                                max = 10,
                                step = 1
                              )
                            ),
                            column(
                              6,
                              numericInput(
                                "collocCollocRight",
                                "Right Span:",
                                value = 2,
                                min = 0,
                                max = 10,
                                step = 1
                              )
                            )
                          )
                        ),
                        column(
                          3,
                          h4(
                            strong("Filter Settings"),
                            style = "color: #4F7942; margin-top: 0;"
                          ),

                          # Minimum frequency
                          numericInput(
                            "collocCollocMinFreq",
                            "Minimum Frequency:",
                            value = 2,
                            min = 1,
                            max = 100,
                            step = 1
                          )
                        ),
                        column(
                          3,
                          # Results summary
                          htmlOutput("collocCollocSummary")
                        )
                      ),
                      fluidRow(
                        column(
                          12,
                          # Search button
                          actionButton(
                            "collocCollocSearch",
                            "Search",
                            icon = icon("search"),
                            class = "btn-success",
                            style = "width: 200px; margin-top: 10px;"
                          )
                        )
                      )
                    )
                  )
                ),

                fluidRow(
                  column(
                    12,
                    # Collocates table
                    h4(strong("Collocates"), style = "color: #4F7942;"),
                    shinycssloaders::withSpinner(
                      DT::DTOutput("collocCollocTable"),
                      color = "#4F7942"
                    )
                  )
                )
              )
            ),

            # ==========================================
            # TAB 3: KWIC Network
            # ==========================================
            tabPanel(
              title = "Network",
              icon = icon("diagram-project"),
              fluidPage(
                br(),
                fluidRow(
                  column(
                    12,
                    wellPanel(
                      style = "background: white; border: 1px solid #ddd; padding: 15px;",
                      fluidRow(
                        column(
                          3,
                          h4(
                            strong("Search Settings"),
                            style = "color: #4F7942; margin-top: 0;"
                          ),

                          # Search query
                          selectizeInput(
                            inputId = "wordsContSearch",
                            label = "Search word(s) in text",
                            choices = NULL
                          )
                        ),
                        column(
                          3,
                          h4(
                            strong("Window Settings"),
                            style = "color: #4F7942; margin-top: 0;"
                          ),

                          # Window span
                          fluidRow(
                            column(
                              6,
                              numericInput(
                                "wordsContBefore",
                                "Left Span:",
                                value = 5,
                                min = 0,
                                max = 20,
                                step = 1
                              )
                            ),
                            column(
                              6,
                              numericInput(
                                "wordsContAfter",
                                "Right Span:",
                                value = 5,
                                min = 0,
                                max = 20,
                                step = 1
                              )
                            )
                          )
                        ),
                        column(
                          2,
                          h4(
                            strong("Network Settings"),
                            style = "color: #4F7942; margin-top: 0;"
                          ),

                          # Window span
                          fluidRow(
                            column(
                              9,
                              numericInput(
                                "wordsContN",
                                "Words",
                                value = 50,
                                min = 10,
                                max = 200,
                                step = 1
                              )
                            )
                          )
                        ),
                        column(
                          3,
                          # Results summary
                          htmlOutput("collocNetworkSummary")
                        )
                      ),
                      fluidRow(
                        column(
                          2,
                          # Search button
                          actionButton(
                            "wordsContApply",
                            "Search",
                            icon = icon("search"),
                            class = "btn-success",
                            style = "width: 200px; margin-top: 10px;"
                          )
                        ),
                        column(
                          2,
                          # Search button
                          actionButton(
                            "wordsContReset",
                            "Reset",
                            icon = icon("xmark"),
                            class = "btn-error",
                            style = "width: 200px; margin-top: 10px;"
                          )
                        )
                      )
                    )
                  )
                ),
                fluidRow(
                  tabsetPanel(
                    type = "tabs",
                    tabPanel(
                      title = "Words in Context",
                      icon = icon("table"),
                      br(),
                      fluidRow(
                        column(
                          12,
                          fluidRow(
                            column(11),
                            div(
                              title = "Export Table as Excel",
                              column(
                                1,
                                do.call(
                                  "actionButton",
                                  c(
                                    export_bttn,
                                    list(
                                      inputId = "wordsContSave"
                                    )
                                  )
                                ),
                                br()
                              )
                            )
                          )
                        )
                      ),
                      fluidRow(
                        column(
                          12,
                          div(
                            style = "height: 450px; overflow-y: scroll; border: 1px solid #ccc; padding: 10px; background-color: #f9f9f9;",
                            shinycssloaders::withSpinner(
                              uiOutput("wordsContHtml"),
                              color = getOption(
                                "spinner.color",
                                default = "#4F7942"
                              )
                            )
                          )
                        )
                      )
                    ),
                    tabPanel(
                      title = "Network Plot",
                      icon = icon("diagram-project"),
                      br(),
                      fluidRow(
                        column(10),
                        div(
                          title = t_export,
                          column(
                            1,
                            do.call(
                              "actionButton",
                              c(
                                export_bttn,
                                list(
                                  inputId = "wordContNetExport"
                                )
                              )
                            )
                          )
                        ),
                        div(
                          title = t_report,
                          column(
                            1,
                            do.call(
                              "actionButton",
                              c(
                                report_bttn,
                                list(
                                  inputId = "wordContNetReport"
                                )
                              )
                            )
                          )
                        )
                      ),
                      fluidRow(
                        shinycssloaders::withSpinner(
                          visNetworkOutput(
                            "wordsContNetwork",
                            width = "auto",
                            height = "58vh"
                          ),
                          color = getOption(
                            "spinner.color",
                            default = "#4F7942"
                          )
                        )
                      )
                    ),
                    tabPanel(
                      title = "TALL AI",
                      icon = icon("robot"),
                      fluidPage(
                        br(),
                        fluidRow(
                          column(
                            12,
                            br(),
                            shinycssloaders::withSpinner(
                              htmlOutput("ContextGeminiUI"),
                              caption = HTML(
                                "<br><strong>Thinking...</strong>"
                              ),
                              image = "ai_small2.gif",
                              color = "#4F7942"
                            )
                          )
                        )
                      )
                    )
                  )
                )
              )
            ),

            # ==========================================
            # TAB 4: INFO & REFERENCES
            # ==========================================
            tabPanel(
              title = "Info & References",
              icon = icon("circle-info"),
              fluidPage(
                br(),
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(
                      "<h3 style='color: #4F7942;'>Collocation Analysis</h3>
                        <p>This module implements collocation analysis features inspired by AntConc:</p>
                        <ul>
                          <li><b>Plot:</b> Visualizes the distribution of a target word across documents
                          in a 'barcode' format, showing where hits appear in each document. Click 'View' button
                          to see the full document with highlighted target words.</li>
                          <li><b>Collocate:</b> Identifies words that frequently appear near the target
                          word, using multiple statistical measures (MI score, Log-Likelihood, and T-Score)
                          displayed simultaneously in the results table.</li>
                        </ul>

                        <h4 style='color: #4F7942;'>Statistical Measures</h4>
                        <ul>
                          <li><b>MI Score (Mutual Information):</b> Measures the strength of association
                          between words. Higher values indicate stronger collocations. Values > 3 are typically
                          considered significant.</li>
                          <li><b>Log-Likelihood:</b> Statistical test for significance of co-occurrence.
                          Values > 3.84 are significant (p < 0.05), values > 15.13 are highly significant (p < 0.0001).</li>
                          <li><b>T-Score:</b> Measures the confidence in the collocation. Values > 2
                          are generally considered significant. Unlike MI, T-Score is less sensitive to low frequencies.</li>
                        </ul>

                        <h4 style='color: #4F7942;'>References</h4>
                        <ul>
                          <li>Anthony, L. (2024). AntConc (Version 4.3.1) [Computer Software].
                          Tokyo, Japan: Waseda University.</li>
                          <li>Evert, S. (2005). The Statistics of Word Cooccurrences:
                          Word Pairs and Collocations. University of Stuttgart.</li>
                          <li>McEnery, T., & Hardie, A. (2012). Corpus Linguistics:
                          Method, Theory and Practice. Cambridge University Press.</li>
                        </ul>"
                    )
                  ),
                  column(1)
                )
              )
            )
          )
        )
      )
    )
  )
  # )
}

# ============================================================================
# SERVER LOGIC
# ============================================================================

collocationServer <- function(input, output, session, values, statsValues) {
  # Reactive values to store results
  collocResults <- reactiveValues(
    plot_data = NULL,
    collocate_data = NULL
  )

  # ============================================================================
  # TAB 1: PLOT - Word Distribution
  # ============================================================================

  # Perform plot search
  observeEvent(input$collocPlotSearch, {
    req(values$dfTag)
    req(input$collocPlotQuery)
    req(values$generalTerm)

    # Get search parameters
    query <- input$collocPlotQuery
    case_sensitive <- input$collocPlotCase

    # Calculate plot data
    collocResults$plot_data <- calculatePlotDistribution(
      dfTag = values$dfTag,
      term_col = values$generalTerm,
      query = query,
      case_sensitive = case_sensitive
    )
  })

  # Display plot summary
  output$collocPlotSummary <- renderUI({
    req(collocResults$plot_data)

    data <- collocResults$plot_data
    total_hits <- sum(data$freq, na.rm = TRUE)
    total_docs <- nrow(data)
    docs_with_hits <- sum(data$freq > 0, na.rm = TRUE)

    HTML(paste0(
      "<div style='background: #e8f5e9; padding: 15px; border-radius: 5px; border-left: 4px solid #4F7942; margin-top: 20px;'>",
      "<h4 style='margin-top: 0; color: #4F7942;'>Search Results</h4>",
      "<p style='margin: 5px 0;'><b>Target Word:</b> ",
      input$collocPlotQuery,
      "</p>",
      "<p style='margin: 5px 0;'><b>Total Hits:</b> ",
      format(total_hits, big.mark = ","),
      "</p>",
      "<p style='margin: 5px 0;'><b>Documents with Hits:</b> ",
      docs_with_hits,
      " / ",
      total_docs,
      " (",
      round(docs_with_hits / total_docs * 100, 1),
      "%)</p>",
      "</div>"
    ))
  })

  # Display plot table with inline plots
  output$collocPlotTable <- DT::renderDT(
    {
      req(collocResults$plot_data)

      # Sort data
      data <- collocResults$plot_data
      sort_by <- input$collocPlotSort

      if (sort_by == "freq") {
        data <- data %>% arrange(desc(freq))
      } else if (sort_by == "dispersion") {
        data <- data %>% arrange(desc(dispersion))
      } else if (sort_by == "doc_id") {
        data <- data %>% arrange(doc_id)
      }

      # Add row numbers
      data <- data %>%
        mutate(
          Row = row_number(),
          # Calculate normalized frequency (per million tokens)
          norm_freq = round((freq / doc_length) * 1000000, 2),
          # Create inline plot HTML
          Plot = createInlinePlot(
            hit_positions,
            doc_length,
            300
            # input$collocPlotZoom
          ),
          # Add View button
          View = sprintf(
            '<button class="btn btn-sm btn-primary" onclick="Shiny.setInputValue(\'collocPlotViewDoc\', \'%s\', {priority: \'event\'});" style="padding: 2px 8px; font-size: 11px;">
            <i class="fa fa-eye"></i> View
          </button>',
            doc_id
          )
        )

      # Prepare display data
      display_data <- data %>%
        select(
          Row,
          doc_id,
          doc_length,
          freq,
          norm_freq,
          dispersion,
          Plot,
          View
        ) %>%
        rename(
          "Doc ID" = doc_id,
          "Doc Tokens" = doc_length,
          "Freq" = freq,
          "Norm Freq" = norm_freq,
          "Dispersion" = dispersion
        ) %>%
        dplyr::filter(Freq > 0) # Show only documents with hits

      # Create DT table
      DT::datatable(
        display_data,
        selection = "none",
        extensions = "Buttons",
        escape = FALSE, # Important: allow HTML in Plot and View columns
        rownames = FALSE,
        options = list(
          pageLength = 10,
          lengthMenu = c(10, 20, 50),
          scrollX = TRUE,
          dom = "Blfrtip",
          buttons = list(
            list(
              extend = 'excel',
              text = '<i class="fa fa-file-excel"></i> Excel',
              className = 'btn btn-success',
              exportOptions = list(
                modifier = list(page = "all") # Esporta tutte le pagine, non solo quella corrente
              ),
              filename = paste0("In_Documents_Table_", Sys.Date())
            )
          ),
          columnDefs = list(
            list(width = paste0(input$collocPlotZoom, "px"), targets = 6), # Plot column
            list(width = "80px", targets = 7), # View column
            list(className = "dt-center", targets = 0:7)
          ),
          initComplete = JS(
            "function(settings, json) {",
            "$(this.api().table().header()).css({'background-color': '#4F7942', 'color': '#fff'});",
            "}"
          )
        ),
        class = "display compact"
      ) %>%
        DT::formatStyle(
          columns = 1:8,
          fontSize = "12px"
        )
    },
    server = FALSE
  )

  # Handle View button click - show document in modal
  observeEvent(input$collocPlotViewDoc, {
    req(collocResults$plot_data)
    req(input$collocPlotViewDoc)
    req(values$dfTag)

    doc_id <- input$collocPlotViewDoc

    # Get document data
    doc_df <- values$dfTag %>%
      dplyr::filter(docSelected, doc_id == !!doc_id) %>%
      arrange(sentence_id, term_id)

    if (nrow(doc_df) == 0) {
      showNotification("Document not found", type = "error")
      return()
    }

    # Get search parameters
    term_col <- values$generalTerm
    query <- input$collocPlotQuery
    case_sensitive <- input$collocPlotCase

    # Prepare search term
    search_col <- doc_df[[term_col]]
    if (!case_sensitive) {
      search_col <- tolower(search_col)
      query_lower <- tolower(query)
    } else {
      query_lower <- query
    }

    # Mark matches
    doc_df <- doc_df %>%
      mutate(
        is_match = search_col == query_lower
      )

    # Build HTML with highlighted terms
    html_content <- buildDocumentHTML(doc_df, term_col)

    # Show modal
    showModal(
      modalDialog(
        title = tags$div(
          style = "color: #4F7942;",
          tags$h4(
            tags$i(class = "fa fa-file-text-o"),
            " Document View: ",
            doc_id,
            style = "color: #FFFFFF;"
          )
        ),
        tags$div(
          style = "max-height: 70vh; overflow-y: auto; padding: 20px;
                   background: #fff; border: 1px solid #ddd; border-radius: 5px;",
          HTML(html_content)
        ),
        size = "l",
        easyClose = TRUE,
        footer = modalButton("Close")
      )
    )
  })

  # ============================================================================
  # TAB 2: DOCUMENT VIEW
  # ============================================================================

  # ============================================================================
  # TAB 2: COLLOCATE
  # ============================================================================

  # Perform collocate search
  observeEvent(input$collocCollocSearch, {
    req(values$dfTag)
    req(input$collocCollocQuery)
    req(values$generalTerm)

    # Get search parameters
    term_col <- values$generalTerm
    query <- input$collocCollocQuery
    case_sensitive <- input$collocCollocCase
    left_span <- input$collocCollocLeft
    right_span <- input$collocCollocRight
    min_freq <- input$collocCollocMinFreq

    # Calculate collocates
    collocResults$collocate_data <- calculateCollocates(
      dfTag = values$dfTag,
      term_col = term_col,
      query = query,
      case_sensitive = case_sensitive,
      left_span = left_span,
      right_span = right_span,
      min_freq = min_freq
    )
  })

  # Display collocate summary
  output$collocCollocSummary <- renderUI({
    req(collocResults$collocate_data)

    data <- collocResults$collocate_data
    n_collocates <- nrow(data)

    HTML(paste0(
      "<div style='background: #f8f9fa; padding: 15px; border-radius: 5px; border-left: 4px solid #4F7942;'>",
      "<h4 style='margin-top: 0; color: #4F7942;'>Collocate Results</h4>",
      "<p style='margin: 5px 0;'><b>Target Word:</b> ",
      input$collocCollocQuery,
      "</p>",
      "<p style='margin: 5px 0;'><b>Collocates Found:</b> ",
      format(n_collocates, big.mark = ","),
      "</p>",
      "<p style='margin: 5px 0;'><b>Window:</b> ",
      input$collocCollocLeft,
      "L - ",
      input$collocCollocRight,
      "R</p>",
      "<hr style='margin: 10px 0; border: none; border-top: 1px solid #ddd;'>",
      "<p style='margin: 5px 0; font-size: 11px;'><b>Legend:</b> ",
      "<br>",
      "<span style='background: #d4edda; padding: 2px 6px; border-radius: 3px; font-weight: bold;'Significant</span> ",
      "MI > 3 | LogLik > 3.84 | T-Score > 2 &nbsp;&nbsp; ",
      "<br>",
      "<span style='background: #a8d5ba; padding: 2px 6px; border-radius: 3px; font-weight: bold;'Highly Significant</span> ",
      "LogLik > 15.13",
      "</p>",
      "</div>"
    ))
  })

  # Display collocates table
  output$collocCollocTable <- DT::renderDT(
    {
      req(collocResults$collocate_data)

      data <- collocResults$collocate_data
      if ("LogLik" %in% names(data)) {
        data <- data %>% arrange(desc(LogLik))
      }

      # Create DT table with custom formatting
      dt <- DT::datatable(
        data,
        rownames = FALSE,
        extensions = "Buttons",
        options = list(
          pageLength = 15,
          lengthMenu = c(10, 20, 50, 100),
          scrollX = TRUE,
          dom = "Blfrtip",
          buttons = list(
            list(
              extend = 'excel',
              text = '<i class="fa fa-file-excel"></i> Excel',
              className = 'btn btn-success',
              exportOptions = list(
                modifier = list(page = "all") # Esporta tutte le pagine, non solo quella corrente
              ),
              filename = paste0("Collocation_Table_", Sys.Date())
            )
          ),
          columnDefs = list(
            list(className = "dt-center", targets = 1:7)
          ),
          initComplete = JS(
            "function(settings, json) {",
            "$(this.api().table().header()).css({'background-color': '#4F7942', 'color': '#fff'});",
            "}"
          )
        ),
        class = "display compact"
      ) %>%
        # Format MI column: > 3 is significant
        DT::formatStyle(
          "MI",
          backgroundColor = styleInterval(
            cuts = c(3),
            values = c("white", "#d4edda") # Light green for significant
          ),
          fontWeight = styleInterval(
            cuts = c(3),
            values = c("normal", "bold")
          )
        ) %>%
        # Format LogLik column: > 3.84 significant, > 15.13 highly significant
        DT::formatStyle(
          "LogLik",
          backgroundColor = styleInterval(
            cuts = c(3.84, 15.13),
            values = c("white", "#d4edda", "#a8d5ba") # Light green -> darker green
          ),
          fontWeight = styleInterval(
            cuts = c(3.84),
            values = c("normal", "bold")
          )
        ) %>%
        # Format TScore column: > 2 is significant
        DT::formatStyle(
          "TScore",
          backgroundColor = styleInterval(
            cuts = c(2),
            values = c("white", "#d4edda") # Light green for significant
          ),
          fontWeight = styleInterval(
            cuts = c(2),
            values = c("normal", "bold")
          )
        ) %>%
        DT::formatStyle(
          columns = 0:7,
          fontSize = "12px"
        )

      return(dt)
    },
    server = FALSE
  )

  # ============================================================================
  # TAB 3: KWIC NETWORK
  # ============================================================================

  ## KWIC Network ----
  observe({
    req(values$dfTag)

    term <- values$generalTerm
    words <- values$dfTag %>%
      dplyr::filter(docSelected) %>%
      LemmaSelection() %>%
      select(all_of(term)) %>%
      pull() %>%
      unique() %>%
      sort() %>%
      tolower()
    updateSelectizeInput(
      session,
      "wordsContSearch",
      choices = c("", words),
      selected = "",
      server = TRUE
    )
  })

  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$wordsContReset
    },
    handlerExpr = {
      req(values$dfTag)
      values$wordInContest <- data.frame()
      term <- values$generalTerm
      words <- values$dfTag %>%
        dplyr::filter(docSelected) %>%
        LemmaSelection() %>%
        select(all_of(term)) %>%
        pull() %>%
        unique() %>%
        sort() %>%
        tolower()
      updateSelectizeInput(
        session,
        "wordsContSearch",
        choices = c("", words),
        selected = "",
        server = TRUE
      )
    }
  )

  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$wordsContSave
    },
    handlerExpr = {
      req(values$wordInContext)
      file_path <- destFolder(
        paste("Tall-WordsInContext-", sys.time(), ".xlsx", sep = ""),
        values$wdTall
      )
      openxlsx::write.xlsx(
        x = values$wordInContext,
        file = file_path,
        colNames = TRUE
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  wordsInContextMenu <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$wordsContApply
    },
    valueExpr = {
      if (input$wordsContSearch != "") {
        word_search <- req(tolower(trimws(input$wordsContSearch)))
        dfFiltered <- values$dfTag %>% dplyr::filter(docSelected)
        values$wordInContext <- get_context_window(
          dfFiltered,
          target_word = word_search,
          n_left = input$wordsContBefore,
          n_right = input$wordsContAfter,
          term = values$generalTerm
        )
        if (nrow(values$wordInContext) > 1) {
          values$contextNetwork <- contextNetwork(
            df = values$wordInContext,
            dfTag = dfFiltered,
            target_word = word_search,
            n = input$wordsContN,
            seed = values$random_seed
          )
        } else {
          values$contextNetwork <- NULL
        }
      } else {
        values$wordInContest <- NULL
      }
    }
  )

  # Display collocate summary
  output$collocNetworkSummary <- renderUI({
    req(values$wordInContext)

    n_collocates <- nrow(values$wordInContext)

    HTML(paste0(
      "<div style='background: #f8f9fa; padding: 15px; border-radius: 5px; border-left: 4px solid #4F7942;'>",
      "<h4 style='margin-top: 0; color: #4F7942;'>KWIC Network Results</h4>",
      "<p style='margin: 5px 0;'><b>Target Word:</b> ",
      tolower(trimws(input$wordsContSearch)),
      "</p>",
      "<p style='margin: 5px 0;'><b>Collocates Found:</b> ",
      format(n_collocates, big.mark = ","),
      "</p>",
      "<p style='margin: 5px 0;'><b>Window:</b> ",
      input$wordsContBefore,
      "L - ",
      input$wordsContAfter,
      "R</p>",
      "</div>"
    ))
  })

  output$wordsContHtml <- renderUI({
    wordsInContextMenu()
    req(values$wordInContext)
    if (nrow(values$wordInContext) == 0) {
      return(HTML("<p>No results found.</p>"))
    }

    content <- lapply(1:nrow(values$wordInContext), function(i) {
      row <- values$wordInContext[i, ]

      div(
        style = "display: flex; align-items: center; margin-bottom: 5px; border-bottom: 1px solid #eee; padding: 5px 0; width: 100%;",
        # Document ID
        span(
          style = "color: darkblue; width: 150px; font-weight: bold; flex-shrink: 0; white-space: nowrap; overflow: hidden; text-overflow: ellipsis;",
          row$doc_id
        ),
        # Context Before
        span(
          style = "color: gray; text-align: right; flex: 1; white-space: nowrap; overflow: hidden; text-overflow: ellipsis; padding-right: 10px;",
          paste0(unlist(row$context_before), collapse = " ")
        ),
        # Target Word
        span(
          style = "color: #4F7942; font-weight: bold; flex-shrink: 0; padding: 0 10px; min-width: 80px; text-align: center;",
          row$target_word
        ),
        # Context After
        span(
          style = "color: gray; text-align: left; flex: 1; white-space: nowrap; overflow: hidden; text-overflow: ellipsis; padding-left: 10px;",
          paste0(unlist(row$context_after), collapse = " ")
        )
      )
    })

    do.call(tagList, content)
  })

  # output$wordsContHtml <- renderUI({
  #   wordsInContextMenu()
  #   req(values$wordInContext)
  #   if (nrow(values$wordInContext) == 0) {
  #     return(HTML("<p>No results found.</p>"))
  #   }
  #
  #   content <- lapply(1:nrow(values$wordInContext), function(i) {
  #     row <- values$wordInContext[i, ]
  #     before <- paste(unlist(row$context_before), collapse = " ")
  #     div(
  #       style = "display: flex; justify-content: space-between; align-items: center; margin-bottom: 10px;",
  #       span(
  #         style = "color: darkblue; text-align: left; width: 150px; font-weight: bold;",
  #         row$doc_id
  #       ),
  #       span(
  #         style = "color: gray; text-align: right; flex: 1;",
  #         paste0(unlist(row$context_before), collapse = " ")
  #       ),
  #       span(
  #         style = "color: #4F7942; font-weight: bold; padding: 0 10px;",
  #         row$target_word
  #       ),
  #       span(
  #         style = "color: gray; text-align: left; flex: 1;",
  #         paste0(unlist(row$context_after), collapse = " ")
  #       )
  #     )
  #   })
  #
  #   do.call(tagList, content)
  # })

  output$wordsContNetwork <- renderVisNetwork({
    wordsInContextMenu()
    req(values$contextNetwork)
    values$contextNetwork
  })

  output$ContextGeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "Gemini AI", content = values$contextGemini, values)
  })

  ## export Network button
  observeEvent(
    eventExpr = {
      input$wordContNetExport
    },
    handlerExpr = {
      file <- paste("KWICNetwork-", sys.time(), ".png", sep = "")
      file <- destFolder(file, values$wdTall)
      plot2png(values$contextNetwork, filename = file, type = "vis", dpi = values$dpi, height = values$h)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report
  observeEvent(input$wordContNetReport, {
    if (!is.null(values$wordInContext)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "KWICNetwork"
      list_df <- list(
        values$wordInContext
      )
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$filenetVis <- plot2png(
        values$contextNetwork,
        filename = "KWIC-Network.png",
        type = "vis",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$filenetVis, res$col)
      )
      popUp(title = "KWIC Network Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })
}

# ----------------------------------------------------------------------------
# Build HTML for document view with highlighted target words
# ----------------------------------------------------------------------------
buildDocumentHTML <- function(doc_df, term_col) {
  # Group by paragraph and sentence
  sentences <- doc_df %>%
    group_by(paragraph_id, sentence_id) %>%
    summarise(
      sentence_tokens = list(token),
      is_match_list = list(is_match),
      .groups = "drop"
    ) %>%
    arrange(paragraph_id, sentence_id)

  # Build HTML for each sentence
  sentences <- sentences %>%
    rowwise() %>%
    mutate(
      sentence_html = {
        tokens <- sentence_tokens
        matches <- is_match_list

        # Create HTML for each token
        token_html <- sapply(1:length(tokens), function(j) {
          token <- tokens[j]
          is_match <- matches[j]

          if (is_match) {
            # Highlight matched token
            paste0(
              '<span style="background-color: #ffeb3b; font-weight: bold; padding: 2px 4px; border-radius: 3px;">',
              htmltools::htmlEscape(token),
              '</span>'
            )
          } else {
            htmltools::htmlEscape(token)
          }
        })

        # Join tokens with spaces
        paste(token_html, collapse = " ")
      }
    ) %>%
    ungroup()

  # Group by paragraph and combine sentences
  paragraphs <- sentences %>%
    group_by(paragraph_id) %>%
    summarise(
      paragraph_html = paste(sentence_html, collapse = " "),
      .groups = "drop"
    ) %>%
    arrange(paragraph_id)

  # Build final HTML with proper paragraph spacing
  html_paragraphs <- sapply(1:nrow(paragraphs), function(i) {
    paste0(
      '<p style="line-height: 1.8; margin-bottom: 20px; text-align: justify;">',
      paragraphs$paragraph_html[i],
      '</p>'
    )
  })

  # Combine all paragraphs
  full_html <- paste0(
    '<div style="font-family: \'Georgia\', serif; font-size: 14px; color: #333;">',
    paste(html_paragraphs, collapse = "\n"),
    '</div>'
  )

  return(full_html)
}

# ============================================================================
# ANALYSIS FUNCTIONS
# ============================================================================

# ----------------------------------------------------------------------------
# Calculate plot distribution of target word across documents
# ----------------------------------------------------------------------------
calculatePlotDistribution <- function(dfTag, term_col, query, case_sensitive) {
  # Filter selected documents
  df <- dfTag %>% dplyr::filter(docSelected)

  # Prepare search term
  if (!case_sensitive) {
    df[[term_col]] <- tolower(df[[term_col]])
    query <- tolower(query)
  }

  # Find matches
  df <- df %>%
    mutate(is_match = .data[[term_col]] == query)

  # Calculate document-level statistics
  # Exclude PUNCT, NUM, SYM, X from document length count
  doc_stats <- df %>%
    group_by(doc_id) %>%
    summarise(
      freq = sum(is_match),
      # Count only content words (exclude punctuation, numbers, symbols)
      doc_length = sum(!upos %in% c("PUNCT", "NUM", "SYM", "X")),
      # Get actual token positions (not just indices)
      hit_positions = list(which(is_match)),
      first_pos = ifelse(any(is_match), min(which(is_match)), NA),
      .groups = "drop"
    ) %>%
    rowwise() %>%
    mutate(
      # Calculate dispersion (within-document distribution evenness)
      # Using Juilland's D-like measure: how evenly distributed hits are in the document
      dispersion = {
        pos <- hit_positions
        n_hits <- freq
        doc_len <- doc_length

        # If no hits or only one hit, dispersion is 0
        if (is.na(n_hits) || n_hits <= 1) {
          0
        } else {
          # Divide document into segments (e.g., 10 parts)
          n_segments <- min(10, doc_len)
          segment_size <- doc_len / n_segments

          # Count hits in each segment
          segment_counts <- sapply(1:n_segments, function(seg) {
            seg_start <- (seg - 1) * segment_size
            seg_end <- seg * segment_size
            sum(pos > seg_start & pos <= seg_end)
          })

          # Expected count per segment if perfectly distributed
          expected_count <- n_hits / n_segments

          # Calculate coefficient of variation
          if (expected_count > 0) {
            observed_variance <- sum((segment_counts - expected_count)^2) /
              n_segments
            cv <- sqrt(observed_variance) / expected_count

            # Normalize to 0-1 scale (lower CV = higher dispersion)
            # Perfect dispersion (CV=0) -> D=1, High concentration (high CV) -> D=0
            dispersion_value <- 1 / (1 + cv)
          } else {
            dispersion_value <- 0
          }

          round(dispersion_value, 3)
        }
      }
    ) %>%
    ungroup()

  return(doc_stats)
}

# ----------------------------------------------------------------------------
# Create inline plot HTML for table display (mini barcode plot)
# ----------------------------------------------------------------------------
createInlinePlot <- function(hit_positions_list, doc_length, width = 150) {
  sapply(seq_along(hit_positions_list), function(i) {
    positions <- hit_positions_list[[i]]
    length <- doc_length[i]

    if (length(positions) == 0 || length == 0) {
      # No hits - show empty bar
      return(paste0(
        '<svg width="',
        width,
        '" height="20" style="background: #f0f0f0; border: 1px solid #ddd;">',
        '</svg>'
      ))
    }

    # Normalize positions to 0-width scale (0-1 first, then to pixel width)
    normalized_pos <- (positions / length) * width

    # Ensure positions are within bounds
    normalized_pos <- pmin(pmax(normalized_pos, 0), width)

    # Create SVG with hit markers
    svg <- paste0(
      '<svg width="',
      width,
      '" height="20" style="background: #f0f0f0; border: 1px solid #ccc;">',
      # Add a light gray background rectangle
      '<rect x="0" y="0" width="',
      width,
      '" height="20" fill="#f0f0f0"/>',
      # Add border
      '<rect x="0" y="0" width="',
      width,
      '" height="20" fill="none" stroke="#ccc" stroke-width="1"/>',
      # Add vertical lines for each hit
      paste(
        sapply(normalized_pos, function(pos) {
          paste0(
            '<line x1="',
            pos,
            '" y1="2" x2="',
            pos,
            '" y2="18" ',
            'stroke="#4F7942" stroke-width="2"/>'
          )
        }),
        collapse = ""
      ),
      '</svg>'
    )

    return(svg)
  })
}

# ----------------------------------------------------------------------------
# Calculate collocates
# ----------------------------------------------------------------------------
calculateCollocates <- function(
  dfTag,
  term_col,
  query,
  case_sensitive,
  left_span,
  right_span,
  min_freq
) {
  # Filter selected documents and exclude punctuation, numbers, symbols
  df <- dfTag %>%
    dplyr::filter(docSelected) %>%
    filter(!upos %in% c("PUNCT", "NUM", "SYM", "X")) %>%
    arrange(doc_id, sentence_id, token_id)

  # Prepare search term
  if (!case_sensitive) {
    df[[term_col]] <- tolower(df[[term_col]])
    query <- tolower(query)
  }

  # Find target word positions
  target_positions <- which(df[[term_col]] == query)

  if (length(target_positions) == 0) {
    return(data.frame(
      collocate = character(),
      upos = character(),
      freq = numeric(),
      left_freq = numeric(),
      right_freq = numeric(),
      MI = numeric(),
      LogLik = numeric(),
      TScore = numeric()
    ))
  }

  # Extract collocates within window
  collocates_list <- lapply(target_positions, function(pos) {
    # Get window boundaries
    left_start <- max(1, pos - left_span)
    left_end <- pos - 1
    right_start <- pos + 1
    right_end <- min(nrow(df), pos + right_span)

    # Extract collocates
    left_collocates <- if (left_start <= left_end) {
      data.frame(
        collocate = df[[term_col]][left_start:left_end],
        upos = df[["upos"]][left_start:left_end],
        position = "left",
        stringsAsFactors = FALSE
      )
    } else {
      data.frame(collocate = character(), position = character())
    }

    right_collocates <- if (right_start <= right_end) {
      data.frame(
        collocate = df[[term_col]][right_start:right_end],
        upos = df[["upos"]][right_start:right_end],
        position = "right",
        stringsAsFactors = FALSE
      )
    } else {
      data.frame(collocate = character(), position = character())
    }

    rbind(left_collocates, right_collocates)
  })

  # upos selected
  pos_selected <- df %>%
    dplyr::filter(POSSelected) %>%
    dplyr::pull(upos) %>%
    unique()

  # Combine all collocates
  collocates <- bind_rows(collocates_list) %>%
    dplyr::filter(upos %in% pos_selected)

  # Calculate frequencies
  colloc_stats <- collocates %>%
    dplyr::filter(!is.na(collocate), collocate != query) %>%
    group_by(collocate, upos) %>%
    summarise(
      freq = n(),
      left_freq = sum(position == "left"),
      right_freq = sum(position == "right"),
      .groups = "drop"
    ) %>%
    dplyr::filter(freq >= min_freq)

  # Calculate ALL statistical measures
  if (nrow(colloc_stats) > 0) {
    # Corpus statistics for calculations
    N <- nrow(df) # Total tokens
    f_target <- length(target_positions) # Target frequency

    # Pre-compute corpus frequencies via left_join (avoids O(n*m) full scans)
    term_freq <- df %>%
      dplyr::count(.data[[term_col]], name = "f_colloc") %>%
      dplyr::rename(collocate = 1)

    colloc_stats <- colloc_stats %>%
      left_join(term_freq, by = "collocate") %>%
      mutate(
        f_colloc = ifelse(is.na(f_colloc), 0L, f_colloc),
        # Calculate contingency table values
        O11 = freq, # Observed co-occurrence
        O12 = pmax(0, f_target - freq), # Target without collocate
        O21 = pmax(0, f_colloc - freq), # Collocate without target
        O22 = pmax(0, N - f_target - f_colloc + freq), # Neither
        # Expected frequencies
        E11 = (O11 + O12) * (O11 + O21) / N,
        E12 = (O11 + O12) * (O12 + O22) / N,
        E21 = (O21 + O22) * (O11 + O21) / N,
        E22 = (O21 + O22) * (O12 + O22) / N,
        # MI Score
        MI = ifelse(is.na(E11) | E11 == 0, 0, log2(O11 / E11)),
        # Log-Likelihood (vectorized with safe log)
        LogLik = 2 * (
          ifelse(!is.na(O11) & !is.na(E11) & O11 > 0 & E11 > 0, O11 * log(O11 / E11), 0) +
          ifelse(!is.na(O12) & !is.na(E12) & O12 > 0 & E12 > 0, O12 * log(O12 / E12), 0) +
          ifelse(!is.na(O21) & !is.na(E21) & O21 > 0 & E21 > 0, O21 * log(O21 / E21), 0) +
          ifelse(!is.na(O22) & !is.na(E22) & O22 > 0 & E22 > 0, O22 * log(O22 / E22), 0)
        ),
        # T-Score
        TScore = ifelse(is.na(E11) | E11 == 0 | is.na(O11) | O11 == 0,
          0, (O11 - E11) / sqrt(O11))
      ) %>%
      select(
        collocate,
        upos,
        freq,
        left_freq,
        right_freq,
        MI,
        LogLik,
        TScore
      ) %>%
      mutate(
        MI = round(MI, 3),
        LogLik = round(LogLik, 3),
        TScore = round(TScore, 3)
      ) %>%
      arrange(desc(MI))
  } else {
    # Ensure expected columns exist even when empty
    colloc_stats$MI <- numeric(0)
    colloc_stats$LogLik <- numeric(0)
    colloc_stats$TScore <- numeric(0)
  }

  return(colloc_stats)
}

# ----------------------------------------------------------------------------
# Calculate statistical measures for collocates

# ----------------------------------------------------------------------------

# ----------------------------------------------------------------------------
# DTformat helper function (if not already defined in tallFunctions.R)
# ----------------------------------------------------------------------------
# This function should already exist in the TALL project
# Included here for completeness
if (!exists("DTformat")) {
  DTformat <- function(data, nrow = 20, filename = "data", ...) {
    DT::datatable(
      data,
      options = list(
        pageLength = nrow,
        scrollX = TRUE,
        dom = "Bfrtip",
        buttons = list(
          list(extend = "csv", filename = filename),
          list(extend = "excel", filename = filename)
        )
      ),
      extensions = "Buttons",
      ...
    )
  }
}

# UI

#  Function source ----
files <- c(
  "libraries.R",
  "tallUtils.R",
  "tallTextIO.R",
  "tallNLP.R",
  "tallOverview.R",
  "tallNetwork.R",
  "tallTopicModel.R",
  "tallEmbeddings.R",
  "tallSentiment.R",
  "tallReport.R",
  "tallVisualization.R",
  "tallLanguages.R",
  "helpContent.R",
  "cssTags.R",
  "header.R",
  "home.R",
  "import.R",
  "edit.R",
  "preprocessing.R",
  "filters_groups.R",
  "overview.R",
  "featureroles.R",
  "keyness.R",
  "words.R",
  "collocation.R",
  "documents.R",
  "doc_classification.R",
  "report.R",
  "settings.R"
)

for (file in files) {
  source(file, local = TRUE)
}

libraries()

htmltools::findDependencies(selectizeInput(
  "dummy",
  label = NULL,
  choices = NULL
))

## Language model list
languages <- langrepo()
label_lang <- unique(languages$language_name)
names(label_lang) <- gsub("_", " ", label_lang)

### input scale choices
choices <- paste0(seq(from = 0, to = 100, by = 1), "%")

## button style and contents

style_bttn <- "border-radius: 20px; border-width: 1px; font-size: 15px; text-align: center; color: #ffff; padding-left: 7px; padding-right: 20px"
style_opt <- "border-radius: 20px; border-width: 1px; font-size: 15px; margin-top: 15px" # (option button)
style_start <- "border-radius: 15px; border-width: 3px; font-size: 15px; width:100% " # (start button)
# style_bttn <- "border-radius: 15px; border-width: 3px; font-size: 15px; margin-top: 15px;" # (action buttons)
t_report <- "Add Results to the Report"
t_export <- "Export Plot as PNG"
t_run <- "Run the Analysis"
t_view <- "View document"
t_save <- "Save the Analysis"
t_back <- "Back to the original text(s)"

run_bttn <- list(
  label = NULL,
  style = "display:block; height: 37px; width: 37px; border-radius: 50%; border: 3px; margin-top: 15px",
  icon = icon("play")
)

view_bttn <- list(
  label = NULL,
  style = "display:block; height: 37px; width: 37px; border-radius: 50%; border: 3px; margin-top: 15px",
  icon = fa_i(name = "magnifying-glass", prefer_type = "solid")
  # icon = icon("magnifying-glass")
)

export_bttn <- list(
  label = NULL,
  style = "display:block; height: 37px; width: 37px; border-radius: 50%; border: 3px; margin-top: 15px",
  icon = icon("download")
)
report_bttn <- list(
  label = NULL,
  style = "display:block; height: 37px; width: 37px; border-radius: 50%; border: 3px; margin-top: 15px",
  icon = icon("plus")
)
save_bttn <- list(
  label = NULL,
  style = "display:block; height: 37px; width: 37px; border-radius: 50%; border: 1px;", # margin-top: 15px",
  icon = icon("floppy-disk")
)
back_bttn <- list(
  label = NULL,
  style = "display:block; height: 37px; width: 37px; border-radius: 50%; border: 3px; margin-top: 15px",
  icon = icon("rotate")
)
x_bttn <- list(
  label = NULL,
  style = "display:block; height: 37px; width: 37px; border-radius: 50%; border: 3px; margin-top: 15px",
  icon = icon("xmark")
)

## HEADER ----

title_tall <- tags$link(
  tags$a(
    href = "https://tall-app.com",
    target = "_blank",
    tags$img(src = "logo_white.jpg", height = "30", width = "30")
  ),
  strong(" TALL", style = "font-size:17px;")
)

infoTexts <- helpContent()

header <- shinydashboardPlus::dashboardHeader(
  title = title_tall,
  titleWidth = 250,
  controlbarIcon = NULL,
  .list = headerUI()
)

## SIDEBAR ----
sidebar <- dashboardSidebar(
  width = 250,
  sidebarMenu(
    id = "sidebarmenu",
    # shinyjs::useShinyjs(),
    # style = "position: relative; overflow: visible;",
    menuItem(
      "TALL",
      tabName = "tall",
      icon = fa_i(name = "house-user")
    ),
    menuItemOutput("rest_of_sidebar")
  )
)

## BODY ----

# Build each composite UI once, then extract sub-elements
edit_ui <- editUI()
preprocessing_ui <- preprocessingUI()
words_ui <- wordsUI()
documents_ui <- documentsUI()
fg_ui <- filters_groupsUI()

body <- dashboardBody(
  customTheme(),
  cssTags(), ## CSS tags
  tabItems(
    ## TALL PAGE ----
    homeUI(),

    ## IMPORT ----
    importUI(),

    ## EDIT ----

    ### Split ----
    edit_ui$split,

    ### Random selection ----
    edit_ui$randomText,

    ### EXTERNAL INFORMATION ----
    edit_ui$extInfo,

    ## PRE-PROCESSING ----

    ### Tokenization & PoS Tagging -----
    preprocessing_ui$tokpos,

    ### Special Entities Tagging ----
    preprocessing_ui$specialentities,

    ### Multi-Word Units ----
    preprocessing_ui$mwcreation,
    preprocessing_ui$mwlist,

    ### Custom Term List ----
    preprocessing_ui$custompos,

    ### Synonyms Merging ----
    preprocessing_ui$synonyms,

    ### PoS Selection ----
    preprocessing_ui$postagselection,

    ## FILTER ----
    fg_ui$filters,

    ## GROUPS ----
    fg_ui$groups,

    ## FEATURE ROLES ----
    featureRolesUI(),

    ## OVERVIEW ----
    overviewUI(),

    ## KEYNESS ----

    keynessUI(),

    ## WORDS ----

    ### Collocation Analysis ----
    collocationUI(),

    ### KWIC ----
    #words_ui$kwic,

    ### Reinert Clustering ----
    words_ui$reinert,

    ### Correspondence Analysis ----
    words_ui$ca,

    ### Word Network Analysis ----
    words_ui$coword,

    ### Thematic Analysis ----
    words_ui$tm,

    ### Embedding Training ----
    words_ui$we_training,

    ### Embedding Similarity ----
    words_ui$we_similarity,

    ### SVO Triplets ----
    documents_ui$svo_analysis,

    ## DOCUMENTS ----

    ### Syntactic Complexity ----
    documents_ui$syntactic_complexity,

    ### Topic Modeling ----
    #### TM K Choices ----
    documents_ui$tm_k,

    #### TM Analysis ----
    documents_ui$tm_analysis,

    #### Doc Classification ----
    docClassificationUI(),

    ### Polarity Detection ----
    documents_ui$polarity,

    ### Emotion Detection ----
    documents_ui$emotion,

    ### Abstractive Summarization ----
    documents_ui$abs_summ,

    ### Extractive Summarization ----
    documents_ui$ext_summ,

    ## REPORT ----
    reportUI(),

    ## SETTINGS ----
    settingsUI()
  )
) # END DASHBOARDBODY


## UserInterface ####
ui <- tagList(
  dashboardPage(header, sidebar, body),
  scrollToTopButton()
)

### FEATURE ROLES MODULE ####
# UI and Server logic for Feature Roles functionality

#### UI Function ####
featureRolesUI <- function() {
  tabItem(
    tabName = "feature_roles",
    fluidPage(
      # Page Header
      fluidRow(
        column(
          12,
          div(
            h2(
              icon("tags"),
              strong("Feature Roles"),
              style = "color: #4F7942; text-align: center; margin-bottom: 20px;"
            ),
            p(
              "Assign roles to features for different types of text analysis.",
              style = "text-align: center; font-size: 16px; color: #666; margin-bottom: 30px;"
            )
          )
        )
      ),

      fluidRow(
        tabsetPanel(
          type = "tabs",

          # Main Panel - Feature Role Assignment
          tabPanel(
            title = "Role Assignment",
            icon = icon("sliders"),
            br(),
            fluidRow(
              # Main Content Area
              column(
                9,
                div(
                  style = "margin-top: 20px;",

                  # Time Variable Box
                  box(
                    title = strong(
                      icon("clock"),
                      " Time Variable",
                      style = "font-size: 18px;"
                    ),
                    width = 12,
                    status = "primary",
                    solidHeader = TRUE,
                    collapsible = TRUE,

                    p(
                      "Select a numeric or date variable to represent time for diachronic text analysis (e.g., longitudinal topic models).",
                      style = "color: #666; margin-bottom: 15px;"
                    ),

                    fluidRow(
                      column(
                        4,
                        uiOutput("timeVarSelect")
                      ),
                      column(
                        3,
                        uiOutput("timeAggregationSelect")
                      ),
                      column(
                        2,
                        div(
                          style = "margin-top: 25px;",
                          actionBttn(
                            inputId = "previewTimeVar",
                            label = "Preview",
                            style = "bordered",
                            color = "primary",
                            size = "sm",
                            icon = icon("eye")
                          )
                        )
                      ),
                      column(
                        2,
                        div(
                          style = "margin-top: 25px;",
                          actionBttn(
                            inputId = "clearTimeVar",
                            label = "Clear",
                            style = "bordered",
                            color = "danger",
                            size = "sm",
                            icon = icon("xmark")
                          )
                        )
                      )
                    ),

                    # Display current selection
                    conditionalPanel(
                      condition = "output.timeVarSelected",
                      div(
                        style = "margin-top: 15px; padding: 10px; background-color: #e8f5e9; border-left: 4px solid #4caf50; border-radius: 4px;",
                        icon(
                          "check",
                          style = "color: #4caf50;"
                        ),
                        strong(" Current selection: "),
                        textOutput("timeVarDisplay", inline = TRUE)
                      )
                    )
                  ),

                  # Label Variable Box
                  box(
                    title = strong(
                      icon("tag"),
                      " Label Variable",
                      style = "font-size: 18px;"
                    ),
                    width = 12,
                    status = "info",
                    solidHeader = TRUE,
                    collapsible = TRUE,

                    p(
                      "Select a variable that represents the response in supervised text classification models (e.g., Random Forest, SVM).",
                      style = "color: #666; margin-bottom: 15px;"
                    ),

                    fluidRow(
                      column(
                        6,
                        uiOutput("labelVarSelect")
                      ),
                      column(
                        3,
                        div(
                          style = "margin-top: 25px;",
                          actionBttn(
                            inputId = "previewLabelVar",
                            label = "Preview Values",
                            style = "bordered",
                            color = "primary",
                            size = "sm",
                            icon = icon("eye")
                          )
                        )
                      ),
                      column(
                        3,
                        div(
                          style = "margin-top: 25px;",
                          actionBttn(
                            inputId = "clearLabelVar",
                            label = "Clear",
                            style = "bordered",
                            color = "danger",
                            size = "sm",
                            icon = icon("xmark")
                          )
                        )
                      )
                    ),

                    # Display current selection
                    conditionalPanel(
                      condition = "output.labelVarSelected",
                      div(
                        style = "margin-top: 15px; padding: 10px; background-color: #e3f2fd; border-left: 4px solid #2196f3; border-radius: 4px;",
                        icon(
                          "check",
                          style = "color: #2196f3;"
                        ),
                        strong(" Current selection: "),
                        textOutput("labelVarDisplay", inline = TRUE)
                      )
                    )
                  ),

                  # Keyness Group Variable Box
                  box(
                    title = strong(
                      icon("right-left"),
                      " Keyness Group Variable",
                      style = "font-size: 18px;"
                    ),
                    width = 12,
                    status = "warning",
                    solidHeader = TRUE,
                    collapsible = TRUE,

                    p(
                      "Select a variable that divides the corpus into two subcorpora for keyness analysis and comparison.",
                      style = "color: #666; margin-bottom: 15px;"
                    ),

                    fluidRow(
                      column(
                        6,
                        uiOutput("keynessVarSelect")
                      ),
                      column(
                        3,
                        div(
                          style = "margin-top: 25px;",
                          actionBttn(
                            inputId = "previewKeynessVar",
                            label = "Preview Values",
                            style = "bordered",
                            color = "primary",
                            size = "sm",
                            icon = icon("eye")
                          )
                        )
                      ),
                      column(
                        3,
                        div(
                          style = "margin-top: 25px;",
                          actionBttn(
                            inputId = "clearKeynessVar",
                            label = "Clear",
                            style = "bordered",
                            color = "danger",
                            size = "sm",
                            icon = icon("xmark")
                          )
                        )
                      )
                    ),

                    # Display current selection
                    conditionalPanel(
                      condition = "output.keynessVarSelected",
                      div(
                        style = "margin-top: 15px; padding: 10px; background-color: #fff3e0; border-left: 4px solid #ff9800; border-radius: 4px;",
                        icon(
                          "check",
                          style = "color: #ff9800;"
                        ),
                        strong(" Current selection: "),
                        textOutput("keynessVarDisplay", inline = TRUE)
                      )
                    ),

                    # Group assignment UI (shown only if variable has more than 2 categories)
                    uiOutput("keynessGroupAssignment")
                  )
                )
              ),

              # Sidebar - Summary Panel
              column(
                3,
                div(
                  style = "margin-top: 20px;",
                  box(
                    width = 12,
                    solidHeader = TRUE,
                    status = "success",

                    div(
                      class = "box-header with-border",
                      h4(
                        icon("rectangle-list"),
                        strong("Summary"),
                        style = "margin: 0; color: #4F7942;"
                      )
                    ),

                    div(
                      class = "box-body",

                      # Available variables count
                      div(
                        style = "margin-bottom: 20px; padding: 15px; background-color: #f5f5f5; border-radius: 5px;",
                        h5(
                          strong("Available Features"),
                          style = "margin-top: 0; color: #333;"
                        ),
                        h3(
                          textOutput("availableVarsCount"),
                          style = "margin: 10px 0; color: #4F7942; text-align: center;"
                        )
                      ),

                      hr(style = "border-color: #e0e0e0; margin: 20px 0;"),

                      # Quick summary
                      div(
                        h5(
                          strong("Current Assignments"),
                          style = "color: #666; margin-bottom: 15px;"
                        ),

                        div(
                          style = "padding: 10px; background-color: #fff; border: 1px solid #e0e0e0; border-radius: 5px;",

                          div(
                            style = "margin-bottom: 10px;",
                            icon(
                              "clock",
                              style = "color: #3f51b5;"
                            ),
                            strong(" Time: "),
                            br(),
                            span(
                              uiOutput("summaryTimeVar", inline = TRUE),
                              style = "color: #666; font-size: 13px;"
                            )
                          ),

                          div(
                            style = "margin-bottom: 10px;",
                            icon(
                              "tag",
                              style = "color: #00bcd4;"
                            ),
                            strong(" Label: "),
                            br(),
                            span(
                              uiOutput("summaryLabelVar", inline = TRUE),
                              style = "color: #666; font-size: 13px;"
                            )
                          ),

                          div(
                            icon(
                              "right-left",
                              style = "color: #ff9800;"
                            ),
                            strong(" Keyness: "),
                            br(),
                            span(
                              uiOutput("summaryKeynessVar", inline = TRUE),
                              style = "color: #666; font-size: 13px;"
                            )
                          )
                        )
                      ),

                      hr(style = "border-color: #e0e0e0; margin: 20px 0;"),

                      # Action buttons
                      div(
                        actionBttn(
                          inputId = "applyVarRoles",
                          label = strong("Apply Roles"),
                          style = "pill",
                          color = "success",
                          size = "md",
                          block = TRUE,
                          icon = icon("check")
                        ),
                        br(),
                        actionBttn(
                          inputId = "resetVarRoles",
                          label = strong("Reset All"),
                          style = "pill",
                          color = "danger",
                          size = "sm",
                          block = TRUE,
                          icon = icon("arrows-rotate")
                        )
                      )
                    )
                  )
                )
              )
            )
          ),

          # Info & References Panel
          tabPanel(
            title = "Info & References",
            icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  div(
                    style = "padding: 30px; background: white; border-radius: 8px;
                       box-shadow: 0 2px 4px rgba(0,0,0,0.05); margin-top: 20px;",
                    HTML(infoTexts$featureroles)
                  )
                )
              )
            )
          )
        )
      )
    )
  )
}


#### Server Function ####
featureRolesServer <- function(input, output, session, values) {
  # Get available columns (excluding technical columns)
  availableColumns <- reactive({
    req(values$dfTag)
    noGroupLabels(names(values$dfTag))
  })

  # Get available TIME columns (only numeric or date types)
  availableTimeColumns <- reactive({
    req(values$dfTag)
    all_cols <- availableColumns()

    # Filter for numeric or date types
    time_cols <- sapply(all_cols, function(col) {
      col_data <- values$dfTag[[col]]
      is.numeric(col_data) ||
        inherits(col_data, c("Date", "POSIXct", "POSIXlt"))
    })

    all_cols[time_cols]
  })

  # Render available features count
  output$availableVarsCount <- renderText({
    length(availableColumns())
  })

  # Time Variable Selection UI - FILTERED
  output$timeVarSelect <- renderUI({
    choices <- availableTimeColumns()

    if (length(choices) == 0) {
      div(
        style = "padding: 10px; background-color: #fff3cd; border: 1px solid #ffc107; border-radius: 4px;",
        icon("triangle-exclamation", style = "color: #856404;"),
        span(
          " No numeric or date variables available",
          style = "color: #856404; margin-left: 5px;"
        )
      )
    } else {
      selectInput(
        inputId = "timeVarInput",
        label = "Select time variable:",
        choices = c("None" = "", choices),
        selected = values$timeVariable,
        width = "100%"
      )
    }
  })

  # Time Aggregation Selection UI (only for Date variables)
  output$timeAggregationSelect <- renderUI({
    req(input$timeVarInput, input$timeVarInput != "")
    col_data <- values$dfTag[[input$timeVarInput]]

    if (inherits(col_data, c("Date", "POSIXct", "POSIXlt"))) {
      selectInput(
        inputId = "timeAggregation",
        label = "Aggregate by:",
        choices = c(
          "Day" = "day",
          "Week" = "week",
          "Month" = "month",
          "Quarter" = "quarter",
          "Year" = "year"
        ),
        selected = if (!is.null(values$timeAggregation)) values$timeAggregation else "year",
        width = "100%"
      )
    } else {
      # Numeric variable: no aggregation needed
      div(
        style = "margin-top: 25px; padding: 8px; background-color: #f0f0f0; border-radius: 4px; text-align: center;",
        span("Numeric variable", style = "color: #888; font-size: 12px;")
      )
    }
  })

  # Label Variable Selection UI
  output$labelVarSelect <- renderUI({
    choices <- availableColumns()

    selectInput(
      inputId = "labelVarInput",
      label = "Select label variable:",
      choices = c("None" = "", choices),
      selected = values$labelVariable,
      width = "100%"
    )
  })

  # Keyness Variable Selection UI
  output$keynessVarSelect <- renderUI({
    choices <- availableColumns()

    selectInput(
      inputId = "keynessVarInput",
      label = "Select keyness group variable:",
      choices = c("None" = "", choices),
      selected = values$keynessVariable,
      width = "100%"
    )
  })

  # Display outputs for conditional panels
  output$timeVarSelected <- reactive({
    !is.null(values$timeVariable) && values$timeVariable != ""
  })
  outputOptions(output, "timeVarSelected", suspendWhenHidden = FALSE)

  output$labelVarSelected <- reactive({
    !is.null(values$labelVariable) && values$labelVariable != ""
  })
  outputOptions(output, "labelVarSelected", suspendWhenHidden = FALSE)

  output$keynessVarSelected <- reactive({
    !is.null(values$keynessVariable) && values$keynessVariable != ""
  })
  outputOptions(output, "keynessVarSelected", suspendWhenHidden = FALSE)

  # Display current selections
  output$timeVarDisplay <- renderText({
    if (!is.null(values$timeVariable) && values$timeVariable != "") {
      if (!is.null(values$timeAggregation)) {
        agg_label <- switch(
          values$timeAggregation,
          "day" = "Day", "week" = "Week", "month" = "Month",
          "quarter" = "Quarter", "year" = "Year"
        )
        paste0(values$timeVariable, " (aggregated by ", agg_label, ")")
      } else {
        values$timeVariable
      }
    } else {
      "None"
    }
  })

  output$labelVarDisplay <- renderText({
    if (!is.null(values$labelVariable) && values$labelVariable != "") {
      values$labelVariable
    } else {
      "None"
    }
  })

  output$keynessVarDisplay <- renderText({
    if (!is.null(values$keynessVariable) && values$keynessVariable != "") {
      values$keynessVariable
    } else {
      "None"
    }
  })

  # Summary outputs - Changed to renderUI for dynamic updates
  output$summaryTimeVar <- renderUI({
    if (!is.null(values$timeVariable) && values$timeVariable != "") {
      agg_info <- ""
      if (!is.null(values$timeAggregation)) {
        agg_label <- switch(
          values$timeAggregation,
          "day" = "Day", "week" = "Week", "month" = "Month",
          "quarter" = "Quarter", "year" = "Year"
        )
        agg_info <- paste0(
          "<br><small style='color: #666;'>Aggregated by: <strong>",
          agg_label, "</strong></small>"
        )
      }
      HTML(paste0(
        "<span style='color: #4F7942; font-weight: bold;'>",
        values$timeVariable,
        "</span>",
        agg_info
      ))
    } else {
      HTML("<span style='color: #999;'>Not assigned</span>")
    }
  })

  output$summaryLabelVar <- renderUI({
    if (!is.null(values$labelVariable) && values$labelVariable != "") {
      HTML(paste0(
        "<span style='color: #4F7942; font-weight: bold;'>",
        values$labelVariable,
        "</span>"
      ))
    } else {
      HTML("<span style='color: #999;'>Not assigned</span>")
    }
  })

  output$summaryKeynessVar <- renderUI({
    if (!is.null(values$keynessVariable) && values$keynessVariable != "") {
      keyness_text <- values$keynessVariable

      # Add group information if groups are defined
      if (!is.null(values$keynessGroup1) && !is.null(values$keynessGroup2)) {
        group1_text <- paste(values$keynessGroup1, collapse = ", ")
        group2_text <- paste(values$keynessGroup2, collapse = ", ")

        keyness_text <- paste0(
          "<span style='color: #4F7942; font-weight: bold;'>",
          values$keynessVariable,
          "</span><br>",
          "<small style='color: #666;'>",
          "<strong>Group 1:</strong> ",
          group1_text,
          "<br>",
          "<strong>Group 2:</strong> ",
          group2_text,
          "</small>"
        )
      } else {
        keyness_text <- paste0(
          "<span style='color: #4F7942; font-weight: bold;'>",
          keyness_text,
          "</span>"
        )
      }

      HTML(keyness_text)
    } else {
      HTML("<span style='color: #999;'>Not assigned</span>")
    }
  })

  # Helper: aggregate date variable by selected period
  aggregateTimeVar <- function(date_values, aggregation) {
    switch(
      aggregation,
      "day" = as.character(date_values),
      "week" = format(date_values, "%Y-W%V"),
      "month" = format(date_values, "%Y-%m"),
      "quarter" = paste0(format(date_values, "%Y"), "-", quarters(date_values)),
      "year" = format(date_values, "%Y"),
      as.character(date_values)
    )
  }

  # Preview Time Variable - IMPROVED LAYOUT
  observeEvent(input$previewTimeVar, {
    req(input$timeVarInput, input$timeVarInput != "")

    # Aggregate by document (doc_id) to get one value per document
    # Apply docSelected filter to respect filters from Filters menu
    doc_level_data <- values$dfTag %>%
      filter(docSelected) %>%
      group_by(doc_id) %>%
      summarise(
        var_value = first(.data[[input$timeVarInput]]),
        .groups = "drop"
      )

    var_data <- doc_level_data$var_value
    total_docs <- nrow(doc_level_data)
    is_date <- inherits(var_data, c("Date", "POSIXct", "POSIXlt"))

    if (is_date) {
      # Date variable: show aggregated distribution
      aggregation <- if (!is.null(input$timeAggregation)) input$timeAggregation else "year"
      agg_label <- switch(
        aggregation,
        "day" = "Day", "week" = "Week", "month" = "Month",
        "quarter" = "Quarter", "year" = "Year"
      )

      agg_values <- aggregateTimeVar(var_data, aggregation)
      freq_table <- sort(table(agg_values, useNA = "ifany"))
      unique_periods <- length(freq_table)

      # Create table rows for time periods
      category_rows <- paste(
        sapply(seq_along(freq_table), function(i) {
          bg_color <- if (i %% 2 == 0) "#f8f9fa" else "white"
          paste0(
            "<tr style='background-color: ", bg_color, ";'>",
            "<td style='padding: 6px; color: #333; font-weight: 500;'>", names(freq_table)[i], "</td>",
            "<td style='padding: 6px; text-align: right; font-weight: bold; color: #4F7942;'>",
            format(freq_table[i], big.mark = ","), "</td>",
            "<td style='padding: 6px; text-align: right;'>",
            "<span style='background-color: #667eea; color: white; padding: 2px 8px; border-radius: 10px; font-size: 11px;'>",
            round(100 * freq_table[i] / sum(freq_table), 1), "%</span></td>",
            "</tr>"
          )
        }),
        collapse = ""
      )

      preview_text <- paste0(
        "<div style='text-align: left; padding: 15px;'>",
        "<div style='background: linear-gradient(135deg, #667eea 0%, #764ba2 100%); color: white; padding: 15px; border-radius: 8px; margin-bottom: 20px;'>",
        "<h3 style='margin: 0; font-size: 18px;'><i class='fa fa-clock'></i> ", input$timeVarInput, "</h3>",
        "<p style='margin: 5px 0 0 0; opacity: 0.9; font-size: 13px;'>Time Variable Preview (aggregated by ", agg_label, ")</p>",
        "</div>",
        "<div style='background-color: #f8f9fa; padding: 12px; border-radius: 6px; margin-bottom: 15px;'>",
        "<table style='width: 100%; border-collapse: collapse;'>",
        "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Variable Type:</td>",
        "<td style='padding: 6px; text-align: right;'><span style='background-color: #667eea; color: white; padding: 3px 10px; border-radius: 12px; font-size: 12px;'>Date</span></td></tr>",
        "<tr style='background-color: white;'><td style='padding: 6px; color: #666; font-weight: 600;'>Aggregation:</td>",
        "<td style='padding: 6px; text-align: right;'><span style='background-color: #764ba2; color: white; padding: 3px 10px; border-radius: 12px; font-size: 12px;'>", agg_label, "</span></td></tr>",
        "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Total Documents:</td>",
        "<td style='padding: 6px; text-align: right; font-weight: bold; color: #4F7942;'>", format(total_docs, big.mark = ","), "</td></tr>",
        "<tr style='background-color: white;'><td style='padding: 6px; color: #666; font-weight: 600;'>Date Range:</td>",
        "<td style='padding: 6px; text-align: right; font-weight: bold;'>", format(min(var_data, na.rm = TRUE)), " to ", format(max(var_data, na.rm = TRUE)), "</td></tr>",
        "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Time Periods:</td>",
        "<td style='padding: 6px; text-align: right; font-weight: bold; color: #667eea;'>", format(unique_periods, big.mark = ","), "</td></tr>",
        "</table></div>",
        "<div style='background-color: #fff; border: 2px solid #e0e0e0; padding: 12px; border-radius: 6px; max-height: 400px; overflow-y: auto;'>",
        "<h4 style='margin: 0 0 10px 0; color: #333; font-size: 14px; border-bottom: 2px solid #667eea; padding-bottom: 5px;'><i class='fa fa-calendar'></i> Documents per ", agg_label, "</h4>",
        "<table style='width: 100%; border-collapse: collapse;'>",
        "<thead><tr style='background-color: #f5f5f5; border-bottom: 2px solid #ddd;'>",
        "<th style='padding: 6px; text-align: left; color: #666; font-weight: 600;'>Period</th>",
        "<th style='padding: 6px; text-align: right; color: #666; font-weight: 600;'>Documents</th>",
        "<th style='padding: 6px; text-align: right; color: #666; font-weight: 600;'>Percentage</th>",
        "</tr></thead><tbody>",
        category_rows,
        "</tbody></table></div></div>"
      )
    } else if (is.numeric(var_data)) {
      # Numeric variable: show descriptive statistics
      summary_stats <- summary(var_data)
      unique_vals <- length(unique(var_data[!is.na(var_data)]))

      preview_text <- paste0(
        "<div style='text-align: left; padding: 15px;'>",

        # Header section
        "<div style='background: linear-gradient(135deg, #667eea 0%, #764ba2 100%); color: white; padding: 15px; border-radius: 8px; margin-bottom: 20px;'>",
        "<h3 style='margin: 0; font-size: 18px;'><i class='fa fa-clock'></i> ", input$timeVarInput, "</h3>",
        "<p style='margin: 5px 0 0 0; opacity: 0.9; font-size: 13px;'>Time Variable Preview</p>",
        "</div>",

        # Basic info
        "<div style='background-color: #f8f9fa; padding: 12px; border-radius: 6px; margin-bottom: 15px;'>",
        "<table style='width: 100%; border-collapse: collapse;'>",
        "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Variable Type:</td><td style='padding: 6px; text-align: right;'><span style='background-color: #667eea; color: white; padding: 3px 10px; border-radius: 12px; font-size: 12px;'>Numeric</span></td></tr>",
        "<tr style='background-color: white;'><td style='padding: 6px; color: #666; font-weight: 600;'>Total Documents:</td><td style='padding: 6px; text-align: right; font-weight: bold; color: #4F7942;'>",
        format(total_docs, big.mark = ","), "</td></tr>",
        "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Unique Values:</td><td style='padding: 6px; text-align: right; font-weight: bold; color: #667eea;'>",
        format(unique_vals, big.mark = ","), "</td></tr>",
        "</table></div>",

        # Statistics
        "<div style='background-color: #fff; border: 2px solid #e0e0e0; padding: 12px; border-radius: 6px;'>",
        "<h4 style='margin: 0 0 10px 0; color: #333; font-size: 14px; border-bottom: 2px solid #667eea; padding-bottom: 5px;'><i class='fa fa-chart-bar'></i> Descriptive Statistics</h4>",
        "<table style='width: 100%; border-collapse: collapse;'>",
        "<tr><td style='padding: 6px; color: #666;'>Minimum:</td><td style='padding: 6px; text-align: right; font-weight: bold;'>",
        format(round(summary_stats["Min."], 2), big.mark = ","), "</td></tr>",
        "<tr style='background-color: #f8f9fa;'><td style='padding: 6px; color: #666;'>1st Quartile:</td><td style='padding: 6px; text-align: right; font-weight: bold;'>",
        format(round(summary_stats["1st Qu."], 2), big.mark = ","), "</td></tr>",
        "<tr><td style='padding: 6px; color: #666;'>Median:</td><td style='padding: 6px; text-align: right; font-weight: bold; color: #667eea;'>",
        format(round(summary_stats["Median"], 2), big.mark = ","), "</td></tr>",
        "<tr style='background-color: #f8f9fa;'><td style='padding: 6px; color: #666;'>Mean:</td><td style='padding: 6px; text-align: right; font-weight: bold; color: #4F7942;'>",
        format(round(summary_stats["Mean"], 2), big.mark = ","), "</td></tr>",
        "<tr><td style='padding: 6px; color: #666;'>3rd Quartile:</td><td style='padding: 6px; text-align: right; font-weight: bold;'>",
        format(round(summary_stats["3rd Qu."], 2), big.mark = ","), "</td></tr>",
        "<tr style='background-color: #f8f9fa;'><td style='padding: 6px; color: #666;'>Maximum:</td><td style='padding: 6px; text-align: right; font-weight: bold;'>",
        format(round(summary_stats["Max."], 2), big.mark = ","), "</td></tr>",
        "</table></div></div>"
      )
    } else {
      preview_text <- "<p>Unsupported variable type for time variable.</p>"
    }

    sendSweetAlert(
      session = session,
      title = NULL,
      text = HTML(preview_text),
      html = TRUE,
      type = NULL,
      width = "600px"
    )
  })

  # Preview Label Variable - IMPROVED LAYOUT
  observeEvent(input$previewLabelVar, {
    req(input$labelVarInput, input$labelVarInput != "")

    # Aggregate by document (doc_id) to get one value per document
    # Apply docSelected filter to respect filters from Filters menu
    doc_level_data <- values$dfTag %>%
      filter(docSelected) %>%
      group_by(doc_id) %>%
      summarise(
        var_value = first(.data[[input$labelVarInput]]),
        .groups = "drop"
      )

    var_data <- doc_level_data$var_value
    total_docs <- nrow(doc_level_data)
    freq_table <- table(var_data, useNA = "ifany")

    # Create table rows for categories
    category_rows <- paste(
      sapply(seq_along(freq_table), function(i) {
        bg_color <- if (i %% 2 == 0) "#f8f9fa" else "white"
        paste0(
          "<tr style='background-color: ",
          bg_color,
          ";'>",
          "<td style='padding: 8px; color: #333; font-weight: 500;'>",
          names(freq_table)[i],
          "</td>",
          "<td style='padding: 8px; text-align: right; font-weight: bold; color: #4F7942;'>",
          format(freq_table[i], big.mark = ","),
          "</td>",
          "<td style='padding: 8px; text-align: right;'>",
          "<span style='background-color: #00bcd4; color: white; padding: 2px 8px; border-radius: 10px; font-size: 11px;'>",
          round(100 * freq_table[i] / sum(freq_table), 1),
          "%",
          "</span></td>",
          "</tr>"
        )
      }),
      collapse = ""
    )

    preview_text <- paste0(
      "<div style='text-align: left; padding: 15px;'>",

      # Header section
      "<div style='background: linear-gradient(135deg, #4facfe 0%, #00f2fe 100%); color: white; padding: 15px; border-radius: 8px; margin-bottom: 20px;'>",
      "<h3 style='margin: 0; font-size: 18px;'><i class='fa fa-tag'></i>",
      input$labelVarInput,
      "</h3>",
      "<p style='margin: 5px 0 0 0; opacity: 0.9; font-size: 13px;'>Label Variable Preview</p>",
      "</div>",

      # Basic info
      "<div style='background-color: #f8f9fa; padding: 12px; border-radius: 6px; margin-bottom: 15px;'>",
      "<table style='width: 100%; border-collapse: collapse;'>",
      "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Variable Type:</td><td style='padding: 6px; text-align: right;'><span style='background-color: #00bcd4; color: white; padding: 3px 10px; border-radius: 12px; font-size: 12px;'>",
      class(var_data)[1],
      "</span></td></tr>",
      "<tr style='background-color: white;'><td style='padding: 6px; color: #666; font-weight: 600;'>Total Documents:</td><td style='padding: 6px; text-align: right; font-weight: bold; color: #4F7942;'>",
      format(total_docs, big.mark = ","),
      "</td></tr>",
      "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Total Categories:</td><td style='padding: 6px; text-align: right; font-weight: bold; color: #00bcd4;'>",
      length(unique(var_data[!is.na(var_data)])),
      "</td></tr>",
      "</table>",
      "</div>",

      # Category distribution
      "<div style='background-color: #fff; border: 2px solid #e0e0e0; padding: 12px; border-radius: 6px; max-height: 400px; overflow-y: auto;'>",
      "<h4 style='margin: 0 0 10px 0; color: #333; font-size: 14px; border-bottom: 2px solid #00bcd4; padding-bottom: 5px;'><i class='fa fa-list'></i> Category Distribution</h4>",
      "<table style='width: 100%; border-collapse: collapse;'>",
      "<thead>",
      "<tr style='background-color: #f5f5f5; border-bottom: 2px solid #ddd;'>",
      "<th style='padding: 8px; text-align: left; color: #666; font-weight: 600;'>Category</th>",
      "<th style='padding: 8px; text-align: right; color: #666; font-weight: 600;'>Documents</th>",
      "<th style='padding: 8px; text-align: right; color: #666; font-weight: 600;'>Percentage</th>",
      "</tr>",
      "</thead>",
      "<tbody>",
      category_rows,
      "</tbody>",
      "</table>",
      "</div>",

      "</div>"
    )

    sendSweetAlert(
      session = session,
      title = NULL,
      text = HTML(preview_text),
      html = TRUE,
      type = NULL,
      width = "600px"
    )
  })

  # Preview Keyness Variable - IMPROVED LAYOUT
  observeEvent(input$previewKeynessVar, {
    req(input$keynessVarInput, input$keynessVarInput != "")

    # Aggregate by document (doc_id) to get one value per document
    # Apply docSelected filter to respect filters from Filters menu
    doc_level_data <- values$dfTag %>%
      filter(docSelected) %>%
      group_by(doc_id) %>%
      summarise(
        var_value = first(.data[[input$keynessVarInput]]),
        .groups = "drop"
      )

    var_data <- doc_level_data$var_value
    total_docs <- nrow(doc_level_data)
    freq_table <- table(var_data, useNA = "ifany")

    # Create table rows for categories
    category_rows <- paste(
      sapply(seq_along(freq_table), function(i) {
        bg_color <- if (i %% 2 == 0) "#f8f9fa" else "white"
        paste0(
          "<tr style='background-color: ",
          bg_color,
          ";'>",
          "<td style='padding: 8px; color: #333; font-weight: 500;'>",
          names(freq_table)[i],
          "</td>",
          "<td style='padding: 8px; text-align: right; font-weight: bold; color: #4F7942;'>",
          format(freq_table[i], big.mark = ","),
          "</td>",
          "<td style='padding: 8px; text-align: right;'>",
          "<span style='background-color: #ff9800; color: white; padding: 2px 8px; border-radius: 10px; font-size: 11px;'>",
          round(100 * freq_table[i] / sum(freq_table), 1),
          "%",
          "</span></td>",
          "</tr>"
        )
      }),
      collapse = ""
    )

    preview_text <- paste0(
      "<div style='text-align: left; padding: 15px;'>",

      # Header section
      "<div style='background: linear-gradient(135deg, #fa709a 0%, #fee140 100%); color: white; padding: 15px; border-radius: 8px; margin-bottom: 20px;'>",
      "<h3 style='margin: 0; font-size: 18px;'><i class='fa fa-right-left'></i>",
      input$keynessVarInput,
      "</h3>",
      "<p style='margin: 5px 0 0 0; opacity: 0.9; font-size: 13px;'>Keyness Group Variable Preview</p>",
      "</div>",

      # Basic info
      "<div style='background-color: #f8f9fa; padding: 12px; border-radius: 6px; margin-bottom: 15px;'>",
      "<table style='width: 100%; border-collapse: collapse;'>",
      "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Variable Type:</td><td style='padding: 6px; text-align: right;'><span style='background-color: #ff9800; color: white; padding: 3px 10px; border-radius: 12px; font-size: 12px;'>",
      class(var_data)[1],
      "</span></td></tr>",
      "<tr style='background-color: white;'><td style='padding: 6px; color: #666; font-weight: 600;'>Total Documents:</td><td style='padding: 6px; text-align: right; font-weight: bold; color: #4F7942;'>",
      format(total_docs, big.mark = ","),
      "</td></tr>",
      "<tr><td style='padding: 6px; color: #666; font-weight: 600;'>Total Categories:</td><td style='padding: 6px; text-align: right; font-weight: bold; color: #ff9800;'>",
      length(unique(var_data[!is.na(var_data)])),
      "</td></tr>",
      "</table>",
      "</div>",

      # Category distribution
      "<div style='background-color: #fff; border: 2px solid #e0e0e0; padding: 12px; border-radius: 6px; max-height: 400px; overflow-y: auto;'>",
      "<h4 style='margin: 0 0 10px 0; color: #333; font-size: 14px; border-bottom: 2px solid #ff9800; padding-bottom: 5px;'><i class='fa fa-list'></i> Category Distribution</h4>",
      "<table style='width: 100%; border-collapse: collapse;'>",
      "<thead>",
      "<tr style='background-color: #f5f5f5; border-bottom: 2px solid #ddd;'>",
      "<th style='padding: 8px; text-align: left; color: #666; font-weight: 600;'>Category</th>",
      "<th style='padding: 8px; text-align: right; color: #666; font-weight: 600;'>Documents</th>",
      "<th style='padding: 8px; text-align: right; color: #666; font-weight: 600;'>Percentage</th>",
      "</tr>",
      "</thead>",
      "<tbody>",
      category_rows,
      "</tbody>",
      "</table>",
      "</div>",

      "</div>"
    )

    sendSweetAlert(
      session = session,
      title = NULL,
      text = HTML(preview_text),
      html = TRUE,
      type = NULL,
      width = "600px"
    )
  })

  # Keyness Group Assignment UI (for variables with more than 2 categories)
  output$keynessGroupAssignment <- renderUI({
    req(input$keynessVarInput, input$keynessVarInput != "")

    # Get unique values at document level
    # Apply docSelected filter to respect filters from Filters menu
    doc_level_data <- values$dfTag %>%
      filter(docSelected) %>%
      group_by(doc_id) %>%
      summarise(
        var_value = first(.data[[input$keynessVarInput]]),
        .groups = "drop"
      )

    categories <- sort(unique(doc_level_data$var_value[
      !is.na(doc_level_data$var_value)
    ]))

    if (length(categories) > 2) {
      div(
        style = "margin-top: 20px; padding: 15px; background-color: #fff8e1; border: 1px solid #ffeb3b; border-radius: 5px;",

        h5(
          icon("circle-info", style = "color: #f57c00;"),
          strong(" Group Assignment Required"),
          style = "color: #f57c00; margin-top: 0;"
        ),

        p(
          "The selected variable has more than 2 categories. Please assign categories to Group 1 and Group 2, then click 'Apply Roles' to create the keyness groups.",
          style = "color: #666; margin-bottom: 15px;"
        ),

        fluidRow(
          column(
            6,
            div(
              style = "padding: 10px; background-color: #e3f2fd; border-radius: 5px;",
              h5(strong("Group 1"), style = "color: #1976d2; margin-top: 0;"),
              pickerInput(
                inputId = "keynessGroup1Categories",
                label = NULL,
                choices = as.character(categories),
                selected = values$keynessGroup1,
                multiple = TRUE,
                options = list(
                  `actions-box` = TRUE,
                  `selected-text-format` = "count > 2",
                  `count-selected-text` = "{0} categories selected"
                )
              )
            )
          ),
          column(
            6,
            div(
              style = "padding: 10px; background-color: #ffebee; border-radius: 5px;",
              h5(strong("Group 2"), style = "color: #c62828; margin-top: 0;"),
              pickerInput(
                inputId = "keynessGroup2Categories",
                label = NULL,
                choices = as.character(categories),
                selected = values$keynessGroup2,
                multiple = TRUE,
                options = list(
                  `actions-box` = TRUE,
                  `selected-text-format` = "count > 2",
                  `count-selected-text` = "{0} categories selected"
                )
              )
            )
          )
        )
      )
    } else {
      NULL
    }
  })

  # Clear individual selections
  observeEvent(input$clearTimeVar, {
    values$timeVariable <- NULL
    values$timeAggregation <- NULL
    if ("time_agg" %in% names(values$dfTag)) {
      values$dfTag$time_agg <- NULL
    }
    updateSelectInput(session, "timeVarInput", selected = "")
  })

  observeEvent(input$clearLabelVar, {
    values$labelVariable <- NULL
    updateSelectInput(session, "labelVarInput", selected = "")
  })

  observeEvent(input$clearKeynessVar, {
    values$keynessVariable <- NULL
    values$keynessGroup1 <- NULL
    values$keynessGroup2 <- NULL

    # Remove keyness_group variable if it exists
    if ("keyness_group" %in% names(values$dfTag)) {
      values$dfTag$keyness_group <- NULL
    }

    updateSelectInput(session, "keynessVarInput", selected = "")
  })

  # Apply feature roles - UNIFIED WORKFLOW
  observeEvent(input$applyVarRoles, {
    success_message <- ""

    # Apply Time Variable
    if (!is.null(input$timeVarInput) && input$timeVarInput != "") {
      values$timeVariable <- input$timeVarInput
      col_data <- values$dfTag[[input$timeVarInput]]
      is_date <- inherits(col_data, c("Date", "POSIXct", "POSIXlt"))

      if (is_date && !is.null(input$timeAggregation)) {
        values$timeAggregation <- input$timeAggregation
        values$dfTag$time_agg <- aggregateTimeVar(col_data, input$timeAggregation)
        agg_label <- switch(
          input$timeAggregation,
          "day" = "Day", "week" = "Week", "month" = "Month",
          "quarter" = "Quarter", "year" = "Year"
        )
        success_message <- paste0(
          success_message,
          "Time variable assigned (Date aggregated by ", agg_label, ").<br>"
        )
      } else {
        values$timeAggregation <- NULL
        # Remove previous aggregation column if exists
        if ("time_agg" %in% names(values$dfTag)) {
          values$dfTag$time_agg <- NULL
        }
        success_message <- paste0(success_message, "Time variable assigned.<br>")
      }
    }

    # Apply Label Variable
    if (!is.null(input$labelVarInput) && input$labelVarInput != "") {
      values$labelVariable <- input$labelVarInput
      success_message <- paste0(success_message, "Label variable assigned.<br>")
    }

    # Apply Keyness Variable and create groups
    if (!is.null(input$keynessVarInput) && input$keynessVarInput != "") {
      values$keynessVariable <- input$keynessVarInput

      var_data <- values$dfTag[[input$keynessVarInput]]
      categories <- unique(var_data[!is.na(var_data)])

      if (length(categories) == 2) {
        # Automatically create keyness_group for binary variables
        keyness_group <- as.numeric(factor(var_data, levels = sort(categories)))
        values$dfTag$keyness_group <- keyness_group

        # Count documents per group
        doc_groups <- values$dfTag %>%
          group_by(doc_id, keyness_group) %>%
          summarise(.groups = "drop") %>%
          count(keyness_group)

        success_message <- paste0(
          success_message,
          "Keyness variable assigned and groups created:<br>",
          "<small>Group 1: ",
          sort(categories)[1],
          " (",
          doc_groups$n[doc_groups$keyness_group == 1],
          " documents)<br>",
          "Group 2: ",
          sort(categories)[2],
          " (",
          doc_groups$n[doc_groups$keyness_group == 2],
          " documents)</small><br>"
        )
      } else if (length(categories) > 2) {
        # Check if groups have been assigned
        if (
          is.null(input$keynessGroup1Categories) ||
            length(input$keynessGroup1Categories) == 0 ||
            is.null(input$keynessGroup2Categories) ||
            length(input$keynessGroup2Categories) == 0
        ) {
          sendSweetAlert(
            session = session,
            title = "Group Assignment Required",
            text = "Please assign categories to both Group 1 and Group 2 before applying roles.",
            type = "warning"
          )
          return()
        }

        # Check for overlap
        overlap <- intersect(
          input$keynessGroup1Categories,
          input$keynessGroup2Categories
        )

        if (length(overlap) > 0) {
          sendSweetAlert(
            session = session,
            title = "Error",
            text = paste(
              "Categories cannot belong to both groups. Overlapping categories:",
              paste(overlap, collapse = ", ")
            ),
            type = "error"
          )
          return()
        }

        # Check if all categories are assigned
        assigned_categories <- c(
          input$keynessGroup1Categories,
          input$keynessGroup2Categories
        )
        unassigned <- setdiff(as.character(categories), assigned_categories)

        # Create keyness_group variable
        keyness_group <- rep(NA, nrow(values$dfTag))
        keyness_group[
          values$dfTag[[input$keynessVarInput]] %in%
            input$keynessGroup1Categories
        ] <- 1
        keyness_group[
          values$dfTag[[input$keynessVarInput]] %in%
            input$keynessGroup2Categories
        ] <- 2

        # Add to dfTag
        values$dfTag$keyness_group <- keyness_group

        # Store group assignments in values
        values$keynessGroup1 <- input$keynessGroup1Categories
        values$keynessGroup2 <- input$keynessGroup2Categories

        # Count documents per group
        doc_groups <- values$dfTag %>%
          filter(!is.na(keyness_group)) %>%
          group_by(doc_id, keyness_group) %>%
          summarise(.groups = "drop") %>%
          count(keyness_group)

        n_group1 <- ifelse(
          1 %in% doc_groups$keyness_group,
          doc_groups$n[doc_groups$keyness_group == 1],
          0
        )
        n_group2 <- ifelse(
          2 %in% doc_groups$keyness_group,
          doc_groups$n[doc_groups$keyness_group == 2],
          0
        )

        success_message <- paste0(
          success_message,
          "Keyness variable assigned and groups created:<br>",
          "<small>Group 1: ",
          n_group1,
          " documents<br>",
          "Group 2: ",
          n_group2,
          " documents"
        )

        if (length(unassigned) > 0) {
          success_message <- paste0(
            success_message,
            "<br>Unassigned categories (excluded): ",
            paste(unassigned, collapse = ", ")
          )
        }

        success_message <- paste0(success_message, "</small><br>")
      }
    }

    # Show confirmation message
    if (success_message == "") {
      sendSweetAlert(
        session = session,
        title = "No Changes",
        text = "Please select at least one variable role to apply.",
        type = "info"
      )
    } else {
      sendSweetAlert(
        session = session,
        title = "Success!",
        text = HTML(success_message),
        html = TRUE,
        type = "success"
      )
    }
  })

  # Reset all feature roles
  observeEvent(input$resetVarRoles, {
    ask_confirmation(
      inputId = "confirmResetRoles",
      title = "Reset Feature Roles",
      text = "Are you sure you want to reset all feature role assignments? This will also remove the keyness_group variable if it exists.",
      type = "warning",
      btn_labels = c("Cancel", "Reset")
    )
  })

  observeEvent(input$confirmResetRoles, {
    if (isTRUE(input$confirmResetRoles)) {
      values$timeVariable <- NULL
      values$timeAggregation <- NULL
      if ("time_agg" %in% names(values$dfTag)) {
        values$dfTag$time_agg <- NULL
      }
      values$labelVariable <- NULL
      values$keynessVariable <- NULL
      values$keynessGroup1 <- NULL
      values$keynessGroup2 <- NULL

      # Remove keyness_group variable if it exists
      if ("keyness_group" %in% names(values$dfTag)) {
        values$dfTag$keyness_group <- NULL
      }

      # Update UI
      updateSelectInput(session, "timeVarInput", selected = "")
      updateSelectInput(session, "labelVarInput", selected = "")
      updateSelectInput(session, "keynessVarInput", selected = "")

      sendSweetAlert(
        session = session,
        title = "Reset Complete",
        text = "All feature roles have been cleared.",
        type = "info"
      )
    }
  })
}

loadSentimentLanguage <- function(language) {
  home <- homeFolder()

  # setting up the main directory
  path_tall <- file.path(home, "tall")
  path_language_model <- file.path(path_tall, "language_models")
  # check if sub directory exists
  if (file.exists(path_tall)) {
    if (!file.exists(path_language_model)) dir.create(path_language_model)
  } else {
    dir.create(path_tall)
    dir.create(path_language_model)
  }

  # check if the file model already exists
  file_lang <- dir(path_language_model, pattern = paste0(language, ".lexicon"))[
    1
  ]

  if (is.na(file_lang)) {
    switch(
      Sys.info()[["sysname"]],
      Windows = {
        download.file(
          url = paste0(
            "https://raw.githubusercontent.com/massimoaria/tall.language.models/main/lexicon.data/",
            language,
            ".lexicon"
          ),
          destfile = paste0(path_language_model, "/", language, ".lexicon"),
          mode = "wb"
        )
      },
      {
        download.file(
          url = paste0(
            "https://raw.githubusercontent.com/massimoaria/tall.language.models/main/lexicon.data/",
            language,
            ".lexicon"
          ),
          destfile = paste0(path_language_model, "/", language, ".lexicon")
        )
      }
    )
  }

  load(file = paste0(path_language_model, "/", language, ".lexicon"))

  return(sentimentData)
}

polarity_colors <- function() {
  c("#FF6666", "#FFB266", "#FFFF66", "#66FF66", "#00FF00")
}

## polarity unit choice ###
ids <- function(dfTag, type) {
  if (is.null(type)) {
    type <- "Documents"
  }
  if (type == "Documents" & "ungroupDoc_id" %in% names(dfTag)) {
    dfTag <- backToOriginalGroups(dfTag)
  }
  unique(dfTag$doc_id[dfTag$docSelected])
}

freqPlotlySentiment <- function(
  dfPlot,
  x,
  y,
  xlabel,
  ylabel,
  scale = c("identity", "log"),
  decimal = 0
) {
  polarity_colors <- polarity_colors()

  # function to build and plot plotly horizontal barplot
  dfPlot <- dfPlot %>%
    group_by(lemma) %>%
    mutate(tot = sum(n)) %>%
    ungroup() %>%
    arrange(tot, lemma, doc_pol_clas)

  xmax <- max(dfPlot[[x]])

  switch(scale, log = {
    # dfPlot$scale <- log(obj$n)
    dfPlot$n <- log(dfPlot$n)
  })

  fig1 <- plot_ly(
    data = dfPlot,
    x = dfPlot[[x]],
    y = ~ reorder(dfPlot[[y]], dfPlot[["tot"]]),
    type = "bar",
    orientation = "h",
    hovertext = ~doc_pol_clas,
    marker = list(
      color = ~ paste0(polarity_colors[as.numeric(doc_pol_clas)], "60"),
      line = list(
        color = ~ polarity_colors[as.numeric(doc_pol_clas)],
        width = 1
      )
    ),
    hovertemplate = "<b><i>Word: %{hovertext}</i></b> <br> <b><i>N. Docs: %{x}</i></b><extra></extra>"
  )

  fig1 <- fig1 %>%
    layout(
      yaxis = list(
        title = ylabel,
        showgrid = FALSE,
        showline = FALSE,
        showticklabels = TRUE,
        domain = c(0, 1)
      ),
      xaxis = list(
        title = xlabel,
        zeroline = FALSE,
        showline = FALSE,
        showticklabels = TRUE,
        showgrid = FALSE
      ),
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)"
    )

  fig1 <- fig1 %>%
    # add_annotations(xref = 'x1', yref = 'y',
    #                                x = dfPlot[[x]] + xmax*0.015,  y = dfPlot[[y]],
    #                                text = ann_text,
    #                                font = list(family = 'Arial', size = 12, color = color),
    #                                showarrow = FALSE) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        #' toImage',
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )
  # %>%
  #   event_register("plotly_selecting")

  fig1
}

sentimentAnalysis <- function(
  dfTag,
  language = "english",
  lexicon_model = "huliu"
) {
  # lexicon for polarity detection
  if (
    !language %in%
      c(
        "ancient_greek",
        "classical_chinese",
        "coptic",
        "gothic",
        "north_sami",
        "old_church_slavonic",
        "old_french",
        "old_russian",
        "scottish_gaelic",
        "wolof"
      )
  ) {
    sentimentData <- loadSentimentLanguage(language)
  } else {
    return(NA)
  }

  amplifiers <- attr(sentimentData, "amplifiers")
  de_amplifiers <- attr(sentimentData, "de_amplifiers")
  negators <- attr(sentimentData, "negators")

  if (language == "english") {
    sentimentData <- sentimentData %>%
      filter(lexicon == lexicon_model)
  }

  polarity_terms <- data.frame(
    term = sentimentData$Word,
    polarity = sentimentData$sentiment
  )

  sentiment <- txt_sentiment(
    dfTag,
    term = "lemma",
    polarity_terms = polarity_terms,
    polarity_negators = negators,
    polarity_amplifiers = amplifiers,
    polarity_deamplifiers = de_amplifiers,
    amplifier_weight = 0.8,
    n_before = 4,
    n_after = 2,
    constrain = TRUE
  )

  s_data <- sentiment$data
  s_overall <- sentiment$overall

  s_data <- s_data %>%
    left_join(
      s_overall %>%
        select(doc_id, sentiment_polarity) %>%
        rename(doc_polarity = sentiment_polarity),
      by = "doc_id"
    ) %>%
    filter(!is.na(polarity)) %>%
    mutate(
      doc_pol_clas = cut(
        .data$doc_polarity,
        breaks = c(-1, -0.6, -0.2, 0.2, 0.6, 1),
        labels = c(
          "Very Negative",
          "Negative",
          "Neutral",
          "Positive",
          "Very Positive"
        ),
        # labels=c("Very Positive", "Positive", "Neutral", "Negative", "Very Negative"),
        include.lowest = T,
        ordered_result = TRUE
      )
    )

  s_overall <- s_overall %>%
    mutate(
      doc_pol_clas = cut(
        .data$sentiment_polarity,
        breaks = c(-1, -0.6, -0.2, 0.2, 0.6, 1),
        labels = c(
          "Very Negative",
          "Negative",
          "Neutral",
          "Positive",
          "Very Positive"
        ),
        include.lowest = T,
        ordered_result = TRUE
      )
    )

  results <- list(sent_data = s_data, sent_overall = s_overall)
  return(results)
}

sentimentWordPlot <- function(sent_data, n = 10) {
  # sent_dist <- sent_data %>%
  #   group_by(doc_id, doc_pol_clas) %>%
  #   distinct(doc_id) %>%
  #   ungroup() %>%
  #   count(doc_pol_clas, name = "doc_N")

  top_words <- sent_data %>%
    group_by(doc_id, polarity) %>%
    distinct(lemma) %>%
    ungroup() %>%
    group_by(polarity) %>%
    count(lemma) %>%
    slice_max(n, n = 10)

  voc_sent <- sent_data %>%
    group_by(polarity, doc_pol_clas, doc_id) %>%
    distinct(lemma) %>%
    ungroup() %>%
    group_by(polarity, doc_pol_clas) %>%
    count(lemma, sort = TRUE) %>%
    filter(lemma %in% top_words$lemma) # %>%
  # left_join(sent_dist, by="doc_pol_clas") %>%
  # mutate(perc = n/doc_N*100)

  fig_pos <- voc_sent %>%
    dplyr::filter(polarity == 1) %>%
    freqPlotlySentiment(
      x = "n",
      y = "lemma",
      xlabel = "Polarized words count",
      ylabel = "word",
      scale = "identity",
      decimal = 0
    )

  fig_neg <- voc_sent %>%
    dplyr::filter(polarity == -1) %>%
    freqPlotlySentiment(
      x = "n",
      y = "lemma",
      xlabel = "Polarized words count",
      ylabel = "word",
      scale = "identity",
      decimal = 0
    )

  plots <- list(positive = fig_pos, negative = fig_neg)
  return(plots)
}

sentimentPieChart <- function(df) {
  plotly::plot_ly(
    data = df,
    values = ~n,
    labels = ~ factor(Polarity),
    sort = FALSE,
    marker = list(colors = paste0(polarity_colors(), "60")),
    textposition = "outside",
    type = "pie",
    hole = 0.4,
    domain = list(x = c(0, 1), y = c(0, 1))
  ) %>%
    layout(
      legend = list(x = -0.1, y = 0.9),
      xaxis = list(
        ticktext = list(
          "Very Negative",
          "Negative",
          "Neutral",
          "Positive",
          "Very Positive"
        ),
        tickvals = list(-0.8, -0.4, 0, 0.4, 0.8),
        tickmode = "array"
      )
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        #' toImage',
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )
  # %>%
  #   event_register("plotly_click")
}

sentimentDensityPlot <- function(x, from = -1, to = 1) {
  fit <- density(x, from = from, to = to)

  plot_ly(
    x = fit$x,
    y = fit$y,
    type = "scatter",
    mode = "lines",
    color = I("#6CC283"),
    fill = "tozeroy",
    text = NULL,
    hoverinfo = "text"
  ) %>%
    layout(
      xaxis = list(
        ticktext = list(
          "Very Negative",
          "Negative",
          "Neutral",
          "Positive",
          "Very Positive"
        ),
        tickvals = list(-0.8, -0.4, 0, 0.4, 0.8),
        tickmode = "array",
        zeroline = FALSE
      ),
      yaxis = list(domain = c(0, 0.90)),
      annotations = list(
        text = "Density Plot",
        xref = "paper",
        x = 0.5,
        yref = "paper",
        y = 0.95,
        yshift = 30,
        showarrow = FALSE,
        font = list(size = 20, color = "gray30")
      )
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        #' toImage',
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )
}

sentimentBoxPlot <- function(sent_overall) {
  plot_ly(
    data = sent_overall,
    x = ~ round(sentiment_polarity, 4),
    y = "",
    type = "box",
    hoverinfo = "x",
    boxpoints = "all",
    jitter = 0.3,
    color = I("#6CC283"),
    pointpos = -1.8
  ) %>%
    layout(
      yaxis = list(
        zeroline = FALSE,
        showgrid = TRUE,
        showline = FALSE,
        showticklabels = TRUE,
        domain = c(0, 0.9)
      ),
      xaxis = list(
        title = "",
        zeroline = FALSE,
        showgrid = TRUE,
        showline = FALSE,
        showticklabels = TRUE,
        range = c(-1, 1),
        ticktext = list(
          "Very Negative",
          "Negative",
          "Neutral",
          "Positive",
          "Very Positive"
        ),
        tickvals = list(-0.8, -0.4, 0, 0.4, 0.8),
        tickmode = "array"
      ),
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)",
      annotations = list(
        text = "Box Plot",
        xref = "paper",
        x = 0.5,
        yref = "paper",
        y = 0.95,
        yshift = 30,
        showarrow = FALSE,
        font = list(size = 20, color = "gray30")
      )
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        #' toImage',
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )
}

### EMOTION ANALYSIS: NRC EmoLex -----

emotion_colors <- function() {
  c(
    anger = "#D32F2F",
    anticipation = "#FF9800",
    disgust = "#6D4C41",
    fear = "#7B1FA2",
    joy = "#FFD600",
    sadness = "#5C6BC0",
    surprise = "#FF6F00",
    trust = "#388E3C"
  )
}

loadEmotionLexicon <- function(language) {
  emotion_names <- c(
    "anger", "anticipation", "disgust", "fear",
    "joy", "sadness", "surprise", "trust"
  )

  if (language == "english") {
    # English lexicon lacks emotion columns; extract from Italian lexicon
    # which has English Word -> emotion mapping from NRC EmoLex
    sentimentData <- loadSentimentLanguage("italian")
    if (!"English Word" %in% names(sentimentData)) return(NULL)
    emotion_lexicon <- sentimentData %>%
      select(`English Word`, all_of(emotion_names)) %>%
      rename(Word = `English Word`) %>%
      distinct(Word, .keep_all = TRUE) %>%
      filter(rowSums(across(all_of(emotion_names))) > 0)
  } else {
    sentimentData <- loadSentimentLanguage(language)
    if (!all(emotion_names %in% names(sentimentData))) return(NULL)
    emotion_lexicon <- sentimentData %>%
      select(Word, all_of(emotion_names)) %>%
      distinct(Word, .keep_all = TRUE) %>%
      filter(rowSums(across(all_of(emotion_names))) > 0)
  }
  return(emotion_lexicon)
}

emotionAnalysis <- function(dfTag, language = "english") {
  unsupported <- c(
    "ancient_greek", "classical_chinese", "coptic", "gothic",
    "north_sami", "old_church_slavonic", "old_french",
    "old_russian", "scottish_gaelic", "wolof"
  )
  if (language %in% unsupported) return(NA)

  emotion_names <- c(
    "anger", "anticipation", "disgust", "fear",
    "joy", "sadness", "surprise", "trust"
  )

  emotion_lexicon <- loadEmotionLexicon(language)
  if (is.null(emotion_lexicon)) return(NA)

  # Get lemmas from tagged data
  df <- dfTag %>% filter(docSelected)

  # Join lemmas with emotion lexicon
  word_emotions <- df %>%
    select(doc_id, lemma) %>%
    inner_join(emotion_lexicon, by = c("lemma" = "Word"))

  if (nrow(word_emotions) == 0) return(NA)

  # Per-document emotion counts
  doc_emotions <- word_emotions %>%
    group_by(doc_id) %>%
    summarise(across(all_of(emotion_names), sum), .groups = "drop") %>%
    mutate(total = rowSums(across(all_of(emotion_names)))) %>%
    mutate(
      across(
        all_of(emotion_names),
        ~ ifelse(total > 0, . / total, 0),
        .names = "{.col}_prop"
      )
    )

  # Corpus-level proportions
  total_count <- sum(colSums(doc_emotions[emotion_names]))
  corpus_emotions <- data.frame(
    emotion = emotion_names,
    count = as.numeric(colSums(doc_emotions[emotion_names])),
    stringsAsFactors = FALSE
  ) %>%
    mutate(proportion = if (total_count > 0) count / total_count else 0)

  # Word-level emotion counts (long format for word plots)
  word_emotion_long <- word_emotions %>%
    group_by(lemma) %>%
    summarise(across(all_of(emotion_names), sum), .groups = "drop") %>%
    pivot_longer(
      cols = all_of(emotion_names),
      names_to = "emotion",
      values_to = "count"
    ) %>%
    filter(count > 0)

  # Doc emotions in long format for heatmap
  doc_emotions_long <- doc_emotions %>%
    select(doc_id, all_of(paste0(emotion_names, "_prop"))) %>%
    pivot_longer(cols = -doc_id, names_to = "emotion", values_to = "proportion") %>%
    mutate(emotion = gsub("_prop$", "", emotion))

  return(list(
    doc_emotions = doc_emotions,
    doc_emotions_long = doc_emotions_long,
    corpus_emotions = corpus_emotions,
    word_emotions = word_emotion_long
  ))
}

emotionBarChart <- function(corpus_emotions) {
  colors <- emotion_colors()

  fig <- plot_ly(
    data = corpus_emotions,
    x = ~emotion,
    y = ~proportion,
    type = "bar",
    marker = list(
      color = colors[corpus_emotions$emotion],
      line = list(color = "white", width = 1)
    ),
    hovertemplate = paste0(
      "<b>%{x}</b><br>",
      "Proportion: %{y:.3f}<br>",
      "Count: ", corpus_emotions$count,
      "<extra></extra>"
    )
  ) %>%
    layout(
      xaxis = list(
        title = "Emotion",
        categoryorder = "total descending"
      ),
      yaxis = list(title = "Proportion"),
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)"
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        "sendDataToCloud", "pan2d", "select2d", "lasso2d",
        "toggleSpikelines", "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )

  fig
}

emotionWordPlot <- function(word_emotions, emotion_sel, n = 10) {
  colors <- emotion_colors()

  df <- word_emotions %>%
    filter(emotion == emotion_sel) %>%
    slice_max(count, n = n) %>%
    arrange(count)

  if (nrow(df) == 0) return(NULL)

  fig <- plot_ly(
    data = df,
    x = ~count,
    y = ~ reorder(lemma, count),
    type = "bar",
    orientation = "h",
    marker = list(
      color = paste0(colors[emotion_sel], "80"),
      line = list(color = colors[emotion_sel], width = 1)
    ),
    hovertemplate = "<b>%{y}</b><br>Count: %{x}<extra></extra>"
  ) %>%
    layout(
      xaxis = list(title = "Count"),
      yaxis = list(title = ""),
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)"
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        "sendDataToCloud", "pan2d", "select2d", "lasso2d",
        "toggleSpikelines", "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )

  fig
}

emotionRadarPlot <- function(corpus_emotions) {
  colors <- emotion_colors()

  # Close the polygon by repeating the first point
  df <- rbind(corpus_emotions, corpus_emotions[1, ])

  fig <- plot_ly(
    type = "scatterpolar",
    r = df$proportion,
    theta = df$emotion,
    fill = "toself",
    fillcolor = "rgba(74, 124, 89, 0.25)",
    line = list(color = "#4a7c59", width = 2),
    marker = list(
      color = colors[df$emotion],
      size = 8,
      line = list(color = "white", width = 1)
    ),
    hovertemplate = paste0(
      "<b>%{theta}</b><br>",
      "Proportion: %{r:.3f}<br>",
      "Count: ", df$count,
      "<extra></extra>"
    )
  ) %>%
    layout(
      polar = list(
        radialaxis = list(
          visible = TRUE,
          range = c(0, max(corpus_emotions$proportion) * 1.15),
          showticklabels = TRUE,
          tickfont = list(size = 10, color = "gray50")
        ),
        angularaxis = list(
          tickfont = list(size = 13, color = "gray30")
        )
      ),
      showlegend = FALSE,
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)"
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        "sendDataToCloud", "pan2d", "select2d", "lasso2d",
        "toggleSpikelines", "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )

  fig
}

emotionHeatmap <- function(doc_emotions_long) {
  fig <- plot_ly(
    data = doc_emotions_long,
    x = ~emotion,
    y = ~doc_id,
    z = ~proportion,
    type = "heatmap",
    colorscale = list(c(0, "#F7F9FA"), c(1, "#4a7c59")),
    hovertemplate = paste0(
      "<b>Doc:</b> %{y}<br>",
      "<b>Emotion:</b> %{x}<br>",
      "<b>Proportion:</b> %{z:.3f}",
      "<extra></extra>"
    )
  ) %>%
    layout(
      xaxis = list(title = ""),
      yaxis = list(title = "", autorange = "reversed"),
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)"
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        "sendDataToCloud", "pan2d", "select2d", "lasso2d",
        "toggleSpikelines", "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )

  fig
}

### TEXT SUMMARIZATION: TEXTRANK -----

highlightSentences <- function(dfTag, id) {
  df <- dfTag %>%
    filter(doc_id == id)

  n <- max(3, round(0.05 * max(df$sentence_id)))

  sentences <- df %>%
    select(sentence_id, sentence) %>%
    distinct()

  terminology <- df %>%
    filter(POSSelected) %>%
    select(sentence_id, lemma)

  tr <- textrank_sentences(data = sentences, terminology = terminology)
  s <- tr$sentences %>%
    arrange(desc(textrank))

  n <- min(n, nrow(s))

  s$h <- c(rep(1, n), rep(0, nrow(s) - n))
  s <- s %>%
    left_join(
      df %>% select(paragraph_id, sentence_id) %>% distinct(),
      by = c("textrank_id" = "sentence_id")
    ) %>%
    mutate(
      sentence = ifelse(
        h == 1,
        paste0("<mark><strong>", sentence, "</strong></mark>"),
        sentence
      )
    ) %>%
    arrange(textrank_id) %>%
    group_by(paragraph_id) %>%
    summarize(
      paragraph = paste(sentence, collapse = " "),
      highlighted = ifelse(sum(h) > 0, "Yes", "No")
    ) %>%
    filter(highlighted == "Yes") %>%
    arrange(paragraph_id) %>%
    select(paragraph_id, paragraph) %>%
    rename(
      "Paragraph ID" = paragraph_id,
      "Paragraph" = paragraph
    )
  return(s)
}

textrankDocument <- function(dfTag, id) {
  df <- dfTag[dfTag$doc_id == id, ]

  # n <- max(3,round(0.05*max(df$sentence_id)))

  sentences <- df %>%
    select(sentence_id, sentence) %>%
    distinct()

  terminology <- df %>%
    filter(POSSelected) %>%
    select(sentence_id, lemma)

  tr <- textrank_sentences(data = sentences, terminology = terminology)
  s <- tr$sentences %>%
    arrange(desc(textrank))

  s <- s %>%
    left_join(
      df %>% select(paragraph_id, sentence_id, sentence) %>% distinct(),
      by = c("sentence")
    )
  results <- list(
    s = s,
    id = id,
    sentences = tr$sentences %>% arrange(desc(textrank))
  )
  return(results)
}

abstractingDocument <- function(s, n, id) {
  switch(
    n,
    "More Concise" = {
      n <- "5%"
    },
    "Less Concise" = {
      n <- "100%"
    },
    {
      n <- n
    }
  )
  n <- as.numeric(gsub("%", "", n))
  n <- ceiling(n * nrow(s) / 100) ## calculate n from %
  n <- min(n, nrow(s))

  s$h <- c(rep(1, n), rep(0, nrow(s) - n))
  # s <- s %>%
  #   left_join(df %>% select(paragraph_id,sentence_id) %>% distinct(), by = c("textrank_id"="sentence_id"))

  abstract <- s %>%
    filter(h == 1) %>%
    group_by(paragraph_id) %>%
    arrange(sentence_id, .by_group = TRUE) %>%
    summarize(paragraph = paste(sentence, collapse = " ")) %>%
    ungroup() %>%
    summarize(
      text = paste(paragraph, collapse = "<br><br>&nbsp&nbsp&nbsp&nbsp&nbsp")
    ) %>%
    mutate(
      text = paste0(
        "<h3>Document: <strong>",
        id,
        "</strong></h3><hr><br><em>",
        text,
        "</em>"
      )
    )

  s <- s %>%
    mutate(
      sentence = ifelse(
        h == 1,
        paste0("<mark><strong>", sentence, "</strong></mark>"),
        sentence
      )
    ) %>%
    arrange(textrank_id) %>%
    group_by(paragraph_id) %>%
    summarize(
      paragraph = paste(sentence, collapse = " "),
      highlighted = ifelse(sum(h) > 0, "Yes", "No")
    ) %>%
    # filter(highlighted=="Yes") %>%
    arrange(paragraph_id) %>%
    select(paragraph_id, paragraph) %>%
    rename(
      "Paragraph ID" = paragraph_id,
      "Paragraph" = paragraph
    )

  results <- list(document = s, abstract = abstract$text[1])
  return(results)
}

# Helper function to create styled HTML box for abstract
create_abstract_box <- function(abstract_text) {
  html_content <- paste0(
    "<div style='",
    "background: linear-gradient(135deg, #f5f7fa 0%, #c3cfe2 100%);",
    "border-radius: 12px;",
    "padding: 30px;",
    "box-shadow: 0 8px 16px rgba(0,0,0,0.1);",
    "border-left: 5px solid #4CAF50;",
    "font-family: \"Georgia\", \"Times New Roman\", serif;",
    "line-height: 1.8;",
    "color: #2c3e50;",
    "max-height: 600px;",
    "overflow-y: auto;",
    "'>",
    "<div style='",
    "font-size: 1.1em;",
    "text-align: justify;",
    "'>",
    abstract_text,
    "</div>",
    "</div>"
  )
  return(html_content)
}

# Helper function to create document HTML with highlighted sentences
create_document_box <- function(
  document_df,
  doc_id,
  summarization_type = "extractive"
) {
  if (summarization_type != "original_text") {
    # Extract paragraphs and combine them
    paragraphs <- document_df$Paragraph

    # Replace <mark><strong> tags with styled span for highlighting
    paragraphs <- gsub(
      "<mark><strong>(.*?)</strong></mark>",
      "<span style='background-color: #ffeb3b; padding: 2px 4px; border-radius: 3px; font-weight: 500;'>\\1</span>",
      paragraphs
    )

    # Create paragraph HTML
    paragraph_html <- paste0(
      "<p style='margin-bottom: 20px; text-indent: 30px;'>",
      paragraphs,
      "</p>"
    )

    full_text <- paste(paragraph_html, collapse = "\n")
  } else {
    full_text <- document_df
  }

  if (summarization_type == "extractive") {
    legend = paste0(
      "<span style='",
      "display: inline-block;",
      "background-color: #ffeb3b;",
      "padding: 4px 8px;",
      "border-radius: 3px;",
      "margin-right: 8px;",
      "'>Highlighted</span> ",
      "= Sentences selected for summarization"
    )
  } else {
    legend = ""
  }

  html_content <- paste0(
    "<div style='",
    "background: #ffffff;",
    "border-radius: 12px;",
    "padding: 30px;",
    "box-shadow: 0 4px 12px rgba(0,0,0,0.08);",
    "border: 1px solid #e0e0e0;",
    "max-height: 600px;",
    "overflow-y: auto;",
    "'>",
    "<h3 style='",
    "color: #2c3e50;",
    "border-bottom: 2px solid #4CAF50;",
    "padding-bottom: 10px;",
    "margin-bottom: 25px;",
    "font-family: \"Arial\", sans-serif;",
    "'>Document: <strong>",
    doc_id,
    "</strong></h3>",
    "<div style='",
    "font-family: \"Georgia\", \"Times New Roman\", serif;",
    "font-size: 1.05em;",
    "line-height: 1.8;",
    "color: #34495e;",
    "text-align: justify;",
    "'>",
    full_text,
    "</div>",
    "<div style='",
    "margin-top: 20px;",
    "padding-top: 15px;",
    "border-top: 1px solid #e0e0e0;",
    "font-size: 0.9em;",
    "color: #7f8c8d;",
    "'>",
    legend,
    # "<span style='",
    # "display: inline-block;",
    # "background-color: #ffeb3b;",
    # "padding: 4px 8px;",
    # "border-radius: 3px;",
    # "margin-right: 8px;",
    # "'>Highlighted</span> ",
    # "= Sentences selected for summarization",
    "</div>",
    "</div>"
  )

  return(html_content)
}


### ABSTRACTIVE TEXT SUMMARIZATION: ----

abstractive_summary <- function(
  values,
  input,
  id,
  nL = 250,
  maxTokens = 16384,
  api_key = NULL,
  model = "2.0-flash",
  retry_attempts = 5
) {
  # Input validation
  if (missing(values) || missing(id)) {
    stop("Both 'values' and 'id' parameters are required")
  }

  if (!is.data.frame(values$dfTag)) {
    stop("values$txt must be a data frame")
  }

  # Validate numeric parameters
  if (!is.numeric(nL) || nL <= 0) {
    stop("nL must be a positive number")
  }

  if (!is.numeric(maxTokens) || maxTokens <= 0) {
    stop("maxTokens must be a positive number")
  }

  # Extract document text based on provided ID
  doc_data <- values$dfTag %>%
    filter(doc_id == !!id) %>%
    rebuild_documents()

  # doc_data <- values$txt %>%
  #   filter(doc_id == !!id)

  # Check if document exists
  if (nrow(doc_data) == 0) {
    warning(paste("Document with ID", id, "not found"))
    return(NA)
  }

  # Extract text content
  doc <- doc_data %>%
    pull(text)

  # Handle case where multiple documents have same ID (take first one)
  if (length(doc) > 1) {
    warning(paste(
      "Multiple documents found for ID",
      id,
      "- using first occurrence"
    ))
    doc <- doc[1]
  }

  # Check for empty or missing text
  if (is.na(doc) || nchar(trimws(doc)) == 0) {
    warning(paste("Document", id, "contains no text content"))
    return(NA)
  }

  # Estimate token count for the document
  tryCatch(
    {
      n_tokens <- estimate_gemini_tokens(doc)
    },
    error = function(e) {
      warning(paste("Token estimation failed for document", id, ":", e$message))
      return(NA)
    }
  )

  # Check if document exceeds token limit
  if (n_tokens > maxTokens) {
    return(paste(
      "Document",
      id,
      "too long (",
      n_tokens,
      "tokens >",
      maxTokens,
      "limit), skipping summarization"
    ))
  }

  # Construct detailed prompt for summarization
  prompt <- paste0(
    "Create a comprehensive abstractive summary of the following text. ",
    "Requirements:\n",
    "- Capture all main points and key details\n",
    "- Maintain clarity and readability\n",
    "- Preserve important context and nuances\n",
    "- Target length: approximately ",
    nL,
    " words\n",
    "- Use clear, concise language\n",
    "- Maintain the original tone when appropriate\n\n",
    "Text to summarize:\n\n",
    doc
  )

  if (!is.null(input$abstractivePrompt)) {
    prompt <- paste(input$abstractivePrompt, prompt, sep = "\n\n")
  }

  # Handle API key configuration
  if (is.null(api_key)) {
    # Try to get from environment variable
    api_key <- Sys.getenv("GEMINI_API_KEY", unset = NA)
    if (is.na(api_key)) {
      stop(
        "API key must be provided either as parameter or GEMINI_API_KEY environment variable"
      )
    }
  }

  res <- gemini_ai(
    image = NULL,
    prompt = prompt,
    model = model,
    type = "text",
    retry_503 = retry_attempts,
    api_key = api_key,
    outputSize = "medium"
  )
  return(res)
}

### EXCEL REPORT FUNCTIONS ----

langrepo <- function() {
  language_name <- c(
    "afrikaans",
    "ancient_greek",
    "ancient_greek",
    "arabic",
    "armenian",
    "basque",
    "belarusian",
    "bulgarian",
    "catalan",
    "chinese",
    "chinese",
    "classical_chinese",
    "coptic",
    "croatian",
    "czech",
    "czech",
    "czech",
    "czech",
    "danish",
    "dutch",
    "dutch",
    "english",
    "english",
    "english",
    "english",
    "estonian",
    "estonian",
    "finnish",
    "finnish",
    "french",
    "french",
    "french",
    "galician",
    "galician",
    "german",
    "german",
    "gothic",
    "greek",
    "hebrew",
    "hindi",
    "hungarian",
    "indonesian",
    "irish",
    "italian",
    "italian",
    "italian",
    "italian",
    "italian",
    "japanese",
    "korean",
    "korean",
    "latin",
    "latin",
    "latin",
    "latvian",
    "lithuanian",
    "lithuanian",
    "maltese",
    "marathi",
    "norwegian",
    "norwegian",
    "old_church_slavonic",
    "persian",
    "polish",
    "polish",
    "portuguese",
    "portuguese",
    "romanian",
    "romanian",
    "russian",
    "russian",
    "russian",
    "scottish_gaelic",
    "serbian",
    "slovak",
    "slovenian",
    "slovenian",
    "spanish",
    "spanish",
    "swedish",
    "swedish",
    "tamil",
    "telugu",
    "turkish",
    "ukrainian",
    "urdu",
    "uyghur",
    "vietnamese",
    "wolof"
  )

  treebank <- c(
    "AfriBooms",
    "Perseus",
    "PROIEL",
    "PADT",
    "ArmTDP",
    "BDT",
    "HSE",
    "BTB",
    "AnCora",
    "GSD",
    "GSDSimp",
    "Kyoto",
    "Scriptorium",
    "SET",
    "CAC",
    "CLTT",
    "FicTree",
    "PDT",
    "DDT",
    "Alpino",
    "LassySmall",
    "EWT",
    "GUM",
    "LinES",
    "ParTUT",
    "EDT",
    "EWT",
    "FTB",
    "TDT",
    "GSD",
    "ParTUT",
    "Sequoia",
    "CTG",
    "TreeGal",
    "GSD",
    "HDT",
    "PROIEL",
    "GDT",
    "HTB",
    "HDTB",
    "Szeged",
    "GSD",
    "IDT",
    "ISDT",
    "ParTUT",
    "PoSTWITA",
    "TWITTIRO",
    "VIT",
    "GSD",
    "GSD",
    "Kaist",
    "ITTB",
    "Perseus",
    "PROIEL",
    "LVTB",
    "ALKSNIS",
    "HSE",
    "MUDT",
    "UFAL",
    "Bokmaal",
    "Nynorsk",
    "PROIEL",
    "Seraji",
    "LFG",
    "PDB",
    "Bosque",
    "GSD",
    "Nonstandard",
    "RRT",
    "GSD",
    "SynTagRus",
    "Taiga",
    "ARCOSG",
    "SET",
    "SNK",
    "SSJ",
    "SST",
    "AnCora",
    "GSD",
    "LinES",
    "Talbanken",
    "TTB",
    "MTG",
    "IMST",
    "IU",
    "UDTB",
    "UDT",
    "VTB",
    "WTB"
  )

  description <- c(
    "UD Afrikaans-AfriBooms is a conversion of the AfriBooms Dependency Treebank, originally annotated with a simplified PoS set and dependency relations according to a subset of the Stanford tag set. The corpus consists of public government documents.",
    "This Universal Dependencies Ancient Greek Treebank consists of an automatic conversion of a selection of passages from the Ancient Greek and Latin Dependency Treebank 2.1.",
    "UD_Ancient_Greek-PROIEL is converted from the Ancient Greek data in the PROIEL treebank, and consists of the New Testament plus selections from Herodotus.",
    "The Arabic-PADT UD treebank is based on the",
    "A Universal Dependencies treebank for Eastern Armenian developed for UD originally by the ArmTDP team led by Marat M. Yavrumyan at the Yerevan State University.",
    "The Basque UD treebank is based on a automatic conversion from part of the Basque Dependency Treebank (BDT), created at the University of of the Basque Country by the IXA NLP research group. The treebank consists of 8.993 sentences (121.443 tokens) and covers mainly literary and journalistic texts.",
    "The Belarusian UD treebank is based on a sample of the news texts included in the Belarusian-Russian parallel subcorpus of the Russian National Corpus,",
    "UD_Bulgarian-BTB is based on the HPSG-based BulTreeBank.",
    "Catalan data from the [AnCora](http://clic.ub.edu/corpus/) corpus.",
    "Traditional Chinese Universal Dependencies Treebank annotated and converted by Google.",
    "Simplified Chinese Universal Dependencies dataset converted from the GSD (traditional) dataset with manual corrections.",
    "Classical Chinese Universal Dependencies Treebank annotated and converted by Institute for Research in Humanities, Kyoto University.",
    "UD Coptic contains manually annotated Sahidic Coptic texts, including Biblical texts, sermons, letters, and hagiography.",
    "The Croatian UD treebank is based on the extension of the SETimes-HR corpus, the [hr500k](http://hdl.handle.net/11356/1183) corpus.",
    "The UD_Czech-CAC treebank is based on the Czech Academic Corpus 2.0 (CAC) created at Charles University in Prague.",
    "The UD_Czech-CLTT treebank is based on the Czech Legal Text Treebank 2.0,",
    "FicTree is a treebank of Czech fiction, automatically converted into the UD format. The treebank was built at Charles University in Prague.",
    "The Czech-PDT UD treebank is based on the Prague Dependency Treebank – Consolidated 1.0, created at the Charles University in Prague.",
    "The Danish UD treebank is a conversion of the Danish Dependency Treebank.",
    "This corpus consists of samples from various treebanks annotated at the University of Groningen using the Alpino annotation tools and guidelines.",
    "This corpus contains sentences from the Wikipedia section of the Lassy Small Treebank.",
    "A Gold Standard Universal Dependencies Corpus for English, built over the source material of the English Web Treebank LDC2012T13 (https://catalog.ldc.upenn.edu/LDC2012T13).",
    "Universal Dependencies syntax annotations from the GUM corpus (https://gucorpling.org/gum/)",
    "UD English_LinES is the English half of the LinES Parallel Treebank with the original dependency annotation first automatically converted into Universal Dependencies and then partially reviewed. Its contents cover literature, an online manual and Europarl data.",
    "UD_English-ParTUT is a conversion of a multilingual parallel treebank developed at the University of Turin, and consisting of a variety of text genres, including talks, legal texts and Wikipedia articles, among others.",
    "UD Estonian is a converted version of the Estonian Dependency Treebank (EDT), originally annotated in the Constraint Grammar (CG) annotation scheme, and consisting of genres of fiction, newspaper texts and scientific texts. The treebank contains 30,972 trees, 437,769 tokens.",
    "UD EWT treebank consists of different genres of new media. The treebank contains 7,190 trees, 90,585 tokens.",
    "FinnTreeBank 1 consists of manually annotated grammatical examples from VISK. The UD version of FinnTreeBank 1 was converted from a native annotation model with a script and later manually revised.",
    "UD_Finnish-TDT is based on the Turku Dependency Treebank (TDT), a broad-coverage dependency treebank of general Finnish covering numerous genres. The conversion to UD was followed by extensive manual checks and corrections, and the treebank closely adheres to the UD guidelines.",
    "The UD_French-GSD was converted in 2015 from the content head version of the universal dependency treebank v2.0 (https://github.com/ryanmcd/uni-dep-tb). It is updated since 2015 independently from the previous source.",
    "UD_French-ParTUT is a conversion of a multilingual parallel treebank developed at the University of Turin, and consisting of a variety of text genres, including talks, legal texts and Wikipedia articles, among others.",
    "**UD_French-Sequoia** is an automatic conversion of the [SUD_French-Sequoia](https://github.com/surfacesyntacticud/SUD_French-Sequoia) treebank, which comes from the former corpus [French Sequoia corpus](http://deep-sequoia.inria.fr).",
    "The Galician UD treebank is based on the automatic parsing of the Galician Technical Corpus (http://sli.uvigo.gal/CTG) created at the University of Vigo by the the TALG NLP research group.",
    "The Galician-TreeGal is a treebank for Galician developed at LyS Group (Universidade da Coruña) and at CiTIUS (Universidade de Santiago de Compostela).",
    "The German UD is converted from the content head version of the [universal dependency treebank v2.0 (legacy)](https://github.com/ryanmcd/uni-dep-tb).",
    "UD German-HDT is a conversion of the Hamburg Dependency Treebank, created at the University of Hamburg through manual annotation in conjunction with a standard for morphologically and syntactically annotating sentences as well as a constraint-based parser.",
    "The UD Gothic treebank is based on the Gothic data from the PROIEL treebank, and consists of Wulfila's Bible translation.",
    "The Greek UD treebank (UD_Greek-GDT) is derived from the Greek Dependency Treebank.",
    "A Universal Dependencies Corpus for Hebrew.",
    "The Hindi UD treebank is based on the Hindi Dependency Treebank (HDTB), created at IIIT Hyderabad, India.",
    "The Hungarian UD treebank is derived from the Szeged Dependency Treebank (Vincze et al. 2010).",
    "The Indonesian-GSD treebank was originally converted from the content head version of the [universal dependency treebank v2.0 (legacy)](https://github.com/ryanmcd/uni-dep-tb) in 2015. In order to comply with the latest Indonesian annotation guidelines, the treebank has undergone a major revision between UD releases v2.8 and v2.9 (2021).",
    "A Universal Dependencies 4910-sentence treebank for modern Irish.",
    "The Italian corpus annotated according to the UD annotation scheme was obtained by conversion from ISDT (Italian Stanford Dependency Treebank), released for the dependency parsing shared task of Evalita-2014 (Bosco et al. 2014).",
    "UD_Italian-ParTUT is a conversion of a multilingual parallel treebank developed at the University of Turin, and consisting of a variety of text genres, including talks, legal texts and Wikipedia articles, among others.",
    "PoSTWITA-UD is a collection of Italian tweets annotated in Universal Dependencies that can be exploited for the training of NLP systems to enhance their performance on social media texts.",
    "TWITTIRÒ-UD is a collection of ironic Italian tweets annotated in Universal Dependencies. The treebank can be exploited for the training of NLP systems to enhance their performance on social media texts, and in particular, for irony detection purposes.",
    "The UD_Italian-VIT corpus was obtained by conversion from VIT (Venice Italian Treebank), developed at the Laboratory of Computational Linguistics of the Università Ca' Foscari in Venice (Delmonte et al. 2007; Delmonte 2009; http://rondelmo.it/resource/VIT/Browser-VIT/index.htm).",
    "This Universal Dependencies (UD) Japanese treebank is based on the definition of UD Japanese convention described in the UD documentation.  The original sentences are from Google UDT 2.0.",
    "The Google Korean Universal Dependency Treebank is first converted from the [Universal Dependency Treebank v2.0 (legacy)](https://github.com/ryanmcd/uni-dep-tb), and then enhanced by Chun et al., 2018.",
    "The KAIST Korean Universal Dependency Treebank is generated by Chun et al., 2018 from the constituency trees in the [KAIST Tree-Tagging Corpus](http://semanticweb.kaist.ac.kr/home/index.php/Corpus4).",
    "Latin data from the _Index Thomisticus_ Treebank. Data are taken from the _Index Thomisticus_ corpus by Roberto Busa SJ, which contains the complete work by Thomas Aquinas (1225–1274; Medieval Latin) and by 61 other authors related to Thomas.",
    "This Universal Dependencies Latin Treebank consists of an automatic conversion of a selection of passages from the Ancient Greek and Latin Dependency Treebank 2.1",
    "The Latin PROIEL treebank is based on the Latin data from the PROIEL treebank, and contains most of the Vulgate New Testament translations plus selections from Caesar's Gallic War, Cicero's Letters to Atticus, Palladius' Opus Agriculturae and the first book of Cicero's De officiis.",
    "Latvian UD Treebank is based on Latvian Treebank ([LVTB](http://sintakse.korpuss.lv)), being created at University of Latvia, Institute of Mathematics and Computer Science, [Artificial Intelligence Laboratory](http://ailab.lv).",
    "The Lithuanian dependency treebank ALKSNIS v3.0 (Vytautas Magnus University).",
    "Lithuanian treebank annotated manually (dependencies) using the Morphological Annotator by CCL, Vytautas Magnus University (http://tekstynas.vdu.lt/) and manual disambiguation.",
    "MUDT (Maltese Universal Dependencies Treebank) is a manually annotated treebank of Maltese, a Semitic language of Malta descended from North African Arabic with a significant amount of Italo-Romance influence.",
    "UD Marathi is a manually annotated treebank consisting primarily of stories from Wikisource, and parts of an article on Wikipedia.",
    "The Norwegian UD treebank is based on the Bokmål section of the Norwegian Dependency Treebank (NDT), which is a syntactic treebank of Norwegian. The current version of NDT has been automatically converted to the UD scheme by Ingerid Løyning Dale, Per Erik Solberg and Andre Kåsen at the Norwegian Language Bank at the National Library of Norway. This conversion builds to a large extent on previous conversions by Lilja Øvrelid at the University of Oslo.",
    "The Norwegian UD treebank is based on the Nynorsk section of the Norwegian Dependency Treebank (NDT), which is a syntactic treebank of Norwegian. NDT has been automatically converted to the UD scheme by Lilja Øvrelid at the University of Oslo.",
    "The Old Church Slavonic (OCS) UD treebank is based on canonical Old Church Slavonic data from the PROIEL and TOROT treebanks.",
    "The Persian Universal Dependency Treebank (Seraji) is based on Uppsala Persian Dependency Treebank (UPDT).",
    "The LFG Enhanced UD treebank of Polish is based on a corpus of LFG (Lexical Functional Grammar) syntactic structures generated by an LFG grammar of Polish, POLFIE, and manually disambiguated by human annotators.",
    "The Polish PDB-UD treebank is automatically converted from the Polish Dependency Bank 2.0 (PDB 2.0). Both treebanks were created at the [Institute of Computer Science, Polish Academy of Sciences](https://ipipan.waw.pl/en/) in Warsaw (Poland).",
    "This Universal Dependencies (UD) Portuguese treebank is based on the Constraint Grammar converted version of the Bosque, which is part of the Floresta Sintá(c)tica treebank. It contains both European (CETEMPúblico) and Brazilian (CETENFolha) variants.",
    "The Brazilian Portuguese UD is converted from the [Google Universal Dependency",
    "The Romanian Non-standard UD treebank (called UAIC-RoDia) is based on UAIC-RoDia Treebank. UAIC-RoDia = ISLRN 156-635-615-024-0",
    "The Romanian UD treebank (called RoRefTrees) (Barbu Mititelu et al., 2016) is the reference treebank in UD format for standard Romanian.",
    "Russian Universal Dependencies Treebank annotated and converted by Google.",
    "Russian data from the SynTagRus corpus.",
    "Universal Dependencies treebank is based on data samples extracted from Taiga Corpus and MorphoRuEval-2017 and GramEval-2020 shared tasks collections.",
    "A treebank of Scottish Gaelic based on the [Annotated Reference Corpus Of Scottish Gaelic (ARCOSG)](https://github.com/Gaelic-Algorithmic-Research-Group/ARCOSG).",
    "The Serbian UD treebank is based on the [SETimes-SR](http://hdl.handle.net/11356/1200) corpus and additional news documents from the Serbian web.",
    "The Slovak UD treebank is based on data originally annotated as part of the Slovak National Corpus, following the annotation style of the Prague Dependency Treebank.",
    "The SSJ treebank is the reference UD treebank for Slovenian, consisting of approximately 13,000 sentences and 267,097 tokens from fiction, non-fiction, periodical and Wikipedia texts in standard modern Slovenian. As of UD release 2.10 in May 2022, the original version of the SSJ UD treebank has been partially manually revised and extended with new manually annotated data.",
    "The Spoken Slovenian Treebank (SST) is a manually annotated collection of transcribed audio recordings featuring spontaneous speech in various everyday situations. It includes 344 unique speech events (documents) amounting to approximately 10 hours of speech, encompassing a total of 6,108 utterances and 98,396 tokens.",
    "Spanish data from the [AnCora](http://clic.ub.edu/corpus/) corpus.",
    "The Spanish UD is converted from the content head version of the [universal dependency treebank v2.0 (legacy)](https://github.com/ryanmcd/uni-dep-tb).",
    "UD Swedish_LinES is the Swedish half of the LinES Parallel Treebank with UD annotations. All segments are translations from English and the sources cover literary genres, online manuals and Europarl data.",
    "The Swedish-Talbanken treebank is based on Talbanken, a treebank developed at Lund University in the 1970s.",
    "The UD Tamil treebank is based on the Tamil Dependency Treebank created at the Charles University in Prague by Loganathan Ramasamy.",
    "The Telugu UD treebank is created in UD based on manual annotations of sentences from a grammar book.",
    "The UD Turkish Treebank, also called the IMST-UD Treebank, is a semi-automatic conversion of the IMST Treebank (Sulubacak&Eryiğit, 2018; Sulubacak et al., 2016).",
    "Gold standard Universal Dependencies corpus for Ukrainian, developed for UD originally, by [Institute for Ukrainian](https://mova.institute), NGO.   [[українською](https://mova.institute/золотий_стандарт)]",
    "The Urdu Universal Dependency Treebank was automatically converted from Urdu Dependency Treebank (UDTB) which is part of an ongoing effort of creating multi-layered treebanks for Hindi and Urdu.",
    "The Uyghur UD treebank is based on the Uyghur Dependency Treebank (UDT), created at the Xinjiang University, China.",
    "The Vietnamese UD treebank is a conversion of the constituent treebank created in the VLSP project (https://vlsp.hpda.vn/).",
    "UD_Wolof-WTB is a natively manual developed treebank for Wolof. Sentences were collected from encyclopedic, fictional, biographical, religious texts and news."
  )

  contributors <- c(
    "Peter Dirix, Liesbeth Augustinus, Daniel van Niekerk",
    "Giuseppe G. A. Celano, Daniel Zeman",
    "Dag Haug",
    "Daniel Zeman, Zdeněk Žabokrtský, Shadi Saleh",
    "Marat M. Yavrumyan",
    "Maria Jesus Aranzabe, Aitziber Atutxa, Kepa Bengoetxea, Arantza Diaz de Ilarraza, Iakes Goenaga, Koldo Gojenola, Larraitz Uria",
    "Olga Lyashevskaya, Angelika Peljak-Łapińska, Daria Petrova, Yana Shishkina",
    "Kiril Simov, Petya Osenova, Martin Popel",
    "Héctor Martínez Alonso, Elena Pascual, Daniel Zeman",
    "Mo Shen, Ryan McDonald, Daniel Zeman, Peng Qi",
    "Peng Qi, Koichi Yasuoka",
    "Koichi Yasuoka, Christian Wittern, Tomohiko Morioka, Takumi Ikeda, Naoki Yamazaki, Yoshihiro Nikaido, Shingo Suzuki, Shigeki Moro, Yuan Li, Hiroyuki Shirasu, Kazunori Fujita",
    "Mitchell Abrams, Elizabeth Davidson, Amir Zeldes",
    "Željko Agić, Nikola Ljubešić, Daniel Zeman",
    "Barbora Hladká, Daniel Zeman",
    "Barbora Hladká, Daniel Zeman, Martin Popel",
    "Tomáš Jelínek, Daniel Zeman",
    "Daniel Zeman, Jan Hajič",
    "Anders Johannsen, Héctor Martínez Alonso, Barbara Plank",
    "Daniel Zeman, Zdeněk Žabokrtský, Gosse Bouma, Gertjan van Noord",
    "Gosse Bouma, Gertjan van Noord",
    "Natalia Silveira, Timothy Dozat, Christopher Manning, Sebastian Schuster, Ethan Chi, John Bauer, Miriam Connor, Marie-Catherine de Marneffe, Nathan Schneider, Sam Bowman, Hanzhi Zhu, Daniel Galbraith, John Bauer",
    "Siyao Peng, Amir Zeldes",
    "Lars Ahrenberg",
    "Cristina Bosco, Manuela Sanguinetti",
    "Kadri Muischnek, Kaili Müürisep, Tiina Puolakainen, Andriela Rääbis, Liisi Torga",
    "Kadri Muischnek, Kaili Müürisep, Tiina Puolakainen, Dage Särg, Sandra Eiche, Andriela Rääbis",
    "Jussi Piitulainen, Hanna Nurmi, Jack Rueter",
    "Filip Ginter, Jenna Kanerva, Veronika Laippala, Niko Miekka, Anna Missilä, Stina Ojala, Sampo Pyysalo",
    "Marie-Catherine de Marneffe, Bruno Guillaume, Ryan McDonald, Alane Suhr, Joakim Nivre, Matias Grioni, Carly Dickerson, Guy Perrier",
    "Cristina Bosco, Manuela Sanguinetti",
    "Marie Candito, Djamé Seddah, Guy Perrier, Bruno Guillaume",
    "Xavier Gómez Guinovart",
    "Marcos Garcia, Xulia Sánchez-Rodríguez",
    "Slav Petrov, Wolfgang Seeker, Ryan McDonald, Joakim Nivre, Daniel Zeman, Adriane Boyd",
    "Emanuel Borges Völker, Felix Hennig, Arne Köhn, Maximilan Wendt, Verena Blaschke, Nina Böbel, Leonie Weissweiler",
    "Dag Haug",
    "Prokopis Prokopidis",
    "Yoav Goldberg, Reut Tsarfaty, Amir More, Shoval Sadde, Victoria Basmov, Yuval Pinter",
    "Riyaz Ahmad Bhat, Daniel Zeman",
    "Richárd Farkas, Katalin Simkó, Zsolt Szántó, Viktor Varga, Veronika Vincze",
    "Ryan McDonald, Joakim Nivre, Daniel Zeman, Septina Dian Larasati, Ika Alfina",
    "Teresa Lynn, Jennifer Foster, Sarah McGuinness, Abigail Walsh, Jason Phelan, Kevin Scannell",
    "Cristina Bosco, Alessandro Lenci, Simonetta Montemagni, Maria Simi",
    "Cristina Bosco, Manuela Sanguinetti",
    "Cristina Bosco, Manuela Sanguinetti",
    "Alessandra T. Cignarella, Cristina Bosco, Manuela Sanguinetti",
    "Fabio Tamburini, Maria Simi, Cristina Bosco",
    "Mai Omura, Yusuke Miyao, Hiroshi Kanayama, Hiroshi Matsuda, Aya Wakasa, Kayo Yamashita, Masayuki Asahara, Takaaki Tanaka, Yugo Murawaki, Yuji Matsumoto, Shinsuke Mori, Sumire Uematsu, Ryan McDonald, Joakim Nivre, Daniel Zeman",
    "Ryan McDonald, Joakim Nivre, Daniel Zeman, Jinho Choi, Na-Rae Han, Jena Hwang, Jayeol Chun",
    "Jinho Choi, Na-Rae Han, Jena Hwang, Jayeol Chun",
    "Marco Passarotti, Marinella Testori, Daniel Zeman, Berta González Saavedra, Flavio Massimiliano Cecchini",
    "Giuseppe G. A. Celano, Daniel Zeman, Federica Gamba",
    "Dag Haug",
    "Lauma Pretkalniņa, Laura Rituma, Gunta Nešpore-Bērzkalne, Baiba Saulīte, Artūrs Znotiņš, Normunds Grūzītis",
    "Andrius Utka, Erika Rimkutė, Agnė Bielinskienė, Jolanta Kovalevskaitė, Loïc Boizou, Gabrielė Aleksandravičiūtė, Kristina Brokaitė, Daniel Zeman, Natalia Perkova, Bernadeta Griciūtė",
    "Olga Lyashevskaya, Dmitri Sitchinava",
    "Slavomír Čéplö, Daniel Zeman",
    "Vinit Ravishankar",
    "Lilja Øvrelid, Fredrik Jørgensen, Petter Hohle, Ingerid Løyning Dale, Per Erik Solberg, Andre Kåsen",
    "Lilja Øvrelid, Fredrik Jørgensen, Petter Hohle, Ingerid Løyning Dale, Per Erik Solberg, Andre Kåsen",
    "Dag Haug",
    "Mojgan Seraji, Filip Ginter, Joakim Nivre, Martin Popel, Daniel Zeman",
    "Agnieszka Patejuk, Adam Przepiórkowski",
    "Alina Wróblewska, Daniel Zeman, Jan Mašek, Rudolf Rosa",
    "Alexandre Rademaker, Cláudia Freitas, Elvis de Souza, Aline Silveira, Tatiana Cavalcanti, Wograine Evelyn, Luisa Rocha, Isabela Soares-Bastos, Eckhard Bick, Fabricio Chalub, Guilherme Paulino-Passos, Livy Real, Valeria de Paiva, Daniel Zeman, Martin Popel, David Mareček, Natalia Silveira, André Martins",
    "Alexandre Rademaker, Ryan McDonald, Joakim Nivre, Daniel Zeman, Fabricio Chalub, Carlos Ramisch, Juan Belieni, Vanessa Berwanger Wille, Rodrigo Pintucci",
    "Cătălina Mărănduc, Cenel-Augusto Perez, Victoria Bobicev, Cătălin Mititelu, Florinel Hociung, Valentin Roșca, Roman Untilov, Petru Rebeja",
    "Verginica Barbu Mititelu, Elena Irimia, Cenel-Augusto Perez, Radu Ion, Radu Simionescu, Martin Popel",
    "Ryan McDonald, Vitaly Nikolaev, Olga Lyashevskaya",
    "Kira Droganova, Olga Lyashevskaya, Daniel Zeman",
    "Olga Lyashevskaya, Olga Rudina, Natalia Vlasova, Anna Zhuravleva",
    "Colin Batchelor",
    "Tanja Samardžić, Nikola Ljubešić",
    "Katarína Gajdošová, Mária Šimková, Daniel Zeman",
    "Kaja Dobrovoljc, Tomaž Erjavec, Simon Krek",
    "Kaja Dobrovoljc, Joakim Nivre",
    "Héctor Martínez Alonso, Daniel Zeman",
    "Miguel Ballesteros, Héctor Martínez Alonso, Ryan McDonald, Elena Pascual, Natalia Silveira, Daniel Zeman, Joakim Nivre",
    "Lars Ahrenberg",
    "Joakim Nivre, Aaron Smith, Victor Norrman",
    "Loganathan Ramasamy, Daniel Zeman",
    "Taraka Rama, Sowmya Vajjala",
    "Utku Türk, Şaziye Betül Özateş, Büşra Marşan, Salih Furkan Akkurt, Çağrı Çöltekin, Gülşen Cebiroğlu Eryiğit, Memduh Gökırmak, Hüner Kaşıkara, Umut Sulubacak, Francis Tyers",
    "Natalia Kotsyba, Bohdan Moskalevskyi, Mykhailo Romanenko",
    "Riyaz Ahmad Bhat, Daniel Zeman",
    "Marhaba Eli, Daniel Zeman, Francis Tyers",
    "Lương Nguyễn Thị, Linh Hà Mỹ, Phương Lê Hồng, Huyền Nguyễn Thị Minh",
    "Bamba Dione"
  )

  file <- c(
    "af_afribooms",
    "grc_perseus",
    "grc_proiel",
    "ar_padt",
    "hy_armtdp",
    "eu_bdt",
    "be_hse",
    "bg_btb",
    "ca_ancora",
    "zh_gsd",
    "zh_gsdsimp",
    "lzh_kyoto",
    "cop_scriptorium",
    "hr_set",
    "cs_cac",
    "cs_cltt",
    "cs_fictree",
    "cs_pdt",
    "da_ddt",
    "nl_alpino",
    "nl_lassysmall",
    "en_ewt",
    "en_gum",
    "en_lines",
    "en_partut",
    "et_edt",
    "et_ewt",
    "fi_ftb",
    "fi_tdt",
    "fr_gsd",
    "fr_partut",
    "fr_sequoia",
    "gl_ctg",
    "gl_treegal",
    "de_gsd",
    "de_hdt",
    "got_proiel",
    "el_gdt",
    "he_htb",
    "hi_hdtb",
    "hu_szeged",
    "id_gsd",
    "ga_idt",
    "it_isdt",
    "it_partut",
    "it_postwita",
    "it_twittiro",
    "it_vit",
    "ja_gsd",
    "ko_gsd",
    "ko_kaist",
    "la_ittb",
    "la_perseus",
    "la_proiel",
    "lv_lvtb",
    "lt_alksnis",
    "lt_hse",
    "mt_mudt",
    "mr_ufal",
    "no_bokmaal",
    "no_nynorsk",
    "cu_proiel",
    "fa_seraji",
    "pl_lfg",
    "pl_pdb",
    "pt_bosque",
    "pt_gsd",
    "ro_nonstandard",
    "ro_rrt",
    "ru_gsd",
    "ru_syntagrus",
    "ru_taiga",
    "gd_arcosg",
    "sr_set",
    "sk_snk",
    "sl_ssj",
    "sl_sst",
    "es_ancora",
    "es_gsd",
    "sv_lines",
    "sv_talbanken",
    "ta_ttb",
    "te_mtg",
    "tr_imst",
    "uk_iu",
    "ur_udtb",
    "ug_udt",
    "vi_vtb",
    "wo_wtb"
  )

  hub_link <- paste0(
    "https://universaldependencies.org/treebanks/",
    file,
    "/index.html"
  )

  tokens <- c(
    "22369",
    "202989",
    "214005",
    "93712",
    "52220",
    "121443",
    "305417",
    "156149",
    "537767",
    "123291",
    "123291",
    "433168",
    "26837",
    "2247925",
    "494142",
    "36007",
    "166602",
    "1526323",
    "506231",
    "208745",
    "297486",
    "251536",
    "208446",
    "93200",
    "49601",
    "438312",
    "50473",
    "159314",
    "202209",
    "389367",
    "27638",
    "68593",
    "126011",
    "23479",
    "287721",
    "3399390",
    "87266",
    "61773",
    "114648",
    "351704",
    "1182501",
    "119830",
    "115990",
    "278461",
    "51614",
    "119334",
    "28384",
    "259625",
    "193654",
    "80322",
    "350090",
    "450480",
    "28868",
    "328476",
    "328305",
    "70049",
    "5356",
    "44162",
    "13089",
    "310221",
    "301353",
    "553686",
    "151625",
    "130967",
    "347319",
    "210958",
    "296169",
    "1896343",
    "218522",
    "97994",
    "1517881",
    "197001",
    "86089",
    "97673",
    "365490",
    "267097",
    "98393",
    "555670",
    "423346",
    "90961",
    "96859",
    "8635",
    "6465",
    "56422",
    "122934",
    "138077",
    "40236",
    "58069",
    "42832"
  )

  words <- c(
    "25017",
    "202989",
    "214005",
    "94390",
    "52585",
    "121443",
    "305417",
    "156149",
    "553638",
    "123291",
    "123291",
    "433168",
    "57098",
    "2252453",
    "495219",
    "36023",
    "167226",
    "1529074",
    "506232",
    "208746",
    "297486",
    "254865",
    "212035",
    "94217",
    "49632",
    "438313",
    "50486",
    "159612",
    "202453",
    "400391",
    "28576",
    "70545",
    "138837",
    "25548",
    "292769",
    "3455580",
    "88934",
    "63441",
    "160195",
    "351704",
    "1183884",
    "122019",
    "115990",
    "298375",
    "55558",
    "124515",
    "29602",
    "280153",
    "193654",
    "80322",
    "350090",
    "450517",
    "29221",
    "328476",
    "328305",
    "70049",
    "5356",
    "44162",
    "13089",
    "310221",
    "301353",
    "553686",
    "152920",
    "130967",
    "349978",
    "227827",
    "318666",
    "1896343",
    "218522",
    "97994",
    "1517881",
    "197001",
    "89958",
    "97673",
    "365490",
    "267097",
    "98393",
    "568249",
    "431600",
    "90961",
    "96859",
    "9581",
    "6465",
    "58096",
    "122983",
    "138077",
    "40236",
    "58069",
    "44258"
  )

  sentences <- c(
    "1975",
    "13919",
    "17082",
    "4800",
    "2500",
    "8993",
    "25231",
    "11138",
    "16678",
    "4997",
    "4997",
    "86239",
    "2203",
    "127794",
    "24709",
    "1121",
    "12760",
    "87907",
    "30723",
    "13603",
    "17120",
    "16622",
    "12146",
    "5243",
    "2090",
    "30930",
    "2829",
    "18723",
    "15136",
    "16342",
    "1020",
    "3099",
    "3993",
    "1000",
    "15590",
    "189928",
    "4328",
    "2521",
    "6143",
    "16649",
    "53564",
    "5598",
    "4910",
    "14167",
    "2090",
    "6712",
    "1424",
    "10087",
    "8100",
    "6339",
    "27363",
    "26977",
    "2273",
    "19387",
    "19367",
    "3642",
    "263",
    "2074",
    "1144",
    "20044",
    "17575",
    "40085",
    "5997",
    "17246",
    "22152",
    "9357",
    "12020",
    "116324",
    "9524",
    "5030",
    "87336",
    "17872",
    "4741",
    "4384",
    "19543",
    "13435",
    "6108",
    "17662",
    "16014",
    "5243",
    "6026",
    "600",
    "1328",
    "5635",
    "7092",
    "5130",
    "3456",
    "3323",
    "2107"
  )

  folder <- paste0("UD_", language_name, "-", treebank) # Generazione dinamica

  chapter <- c(
    "Hoofstuk", # Afrikaans
    "κεφάλαιο", # Ancient Greek
    "κεφάλαιο", # Ancient Greek (duplicated)
    "فصل", # Arabic
    "Գլուխ", # Armenian
    "Kapitulua", # Basque
    "Глава", # Belarusian
    "Глава", # Bulgarian
    "Capítol", # Catalan
    "章节", # Chinese (Simplified)
    "章", # Chinese (Traditional)
    "章节", # Classical Chinese
    "ⲕⲁⲡⲓⲧⲓⲗⲟⲛ", # Coptic
    "Poglavlje", # Croatian
    "Kapitola", # Czech
    "Kapitola", # Czech (duplicated)
    "Kapitola", # Czech (duplicated)
    "Kapitola", # Czech (duplicated)
    "Kapitel", # Danish
    "Hoofdstuk", # Dutch
    "Hoofdstuk", # Dutch (duplicated)
    "Chapter", # English
    "Chapter", # English (duplicated)
    "Chapter", # English (duplicated)
    "Chapter", # English (duplicated)
    "Peatükk", # Estonian
    "Peatükk", # Estonian (duplicated)
    "Luku", # Finnish
    "Luku", # Finnish (duplicated)
    "Chapitre", # French
    "Chapitre", # French (duplicated)
    "Chapitre", # French (duplicated)
    "Capítulo", # Galician
    "Capítulo", # Galician (duplicated)
    "Kapitel", # German
    "Kapitel", # German (duplicated)
    "𐍉𐌼𐌿𐌽𐌳𐌰𐌽𐌾𐌾𐌰", # Gothic
    "Κεφάλαιο", # Greek
    "פרק", # Hebrew
    "अध्याय", # Hindi
    "Fejezet", # Hungarian
    "Bab", # Indonesian
    "Caibidil", # Irish
    "Capitolo", # Italian
    "Capitolo", # Italian (duplicated)
    "Capitolo", # Italian (duplicated)
    "Capitolo", # Italian (duplicated)
    "Capitolo", # Italian (duplicated)
    "章", # Japanese
    "장", # Korean
    "장", # Korean (duplicated)
    "Capitulum", # Latin
    "Capitulum", # Latin (duplicated)
    "Capitulum", # Latin (duplicated)
    "Nodaļa", # Latvian
    "Skyrius", # Lithuanian
    "Skyrius", # Lithuanian (duplicated)
    "Kapitel", # Maltese
    "अध्याय", # Marathi
    "Kapittel", # Norwegian
    "Kapittel", # Norwegian (duplicated)
    "Глава", # Old Church Slavonic
    "فصل", # Persian
    "Rozdział", # Polish
    "Rozdział", # Polish (duplicated)
    "Capítulo", # Portuguese
    "Capítulo", # Portuguese (duplicated)
    "Capitol", # Romanian
    "Capitol", # Romanian (duplicated)
    "Глава", # Russian
    "Глава", # Russian (duplicated)
    "Глава", # Russian (duplicated)
    "Caibideil", # Scottish Gaelic
    "Глава", # Serbian
    "Kapitola", # Slovak
    "Poglavje", # Slovenian
    "Poglavje", # Slovenian (duplicated)
    "Capítulo", # Spanish
    "Capítulo", # Spanish (duplicated)
    "Kapitel", # Swedish
    "Kapitel", # Swedish (duplicated)
    "அத்தியாயம்", # Tamil
    "అధ్యాయం", # Telugu
    "Bölüm", # Turkish
    "Розділ", # Ukrainian
    "باب", # Urdu
    "باب", # Uyghur
    "Chương", # Vietnamese
    "Kàll" # Wolof
  )

  flag <- c(
    "AF.svg",
    "GRC.svg",
    "GRC.svg",
    "AR.svg",
    "HY.svg",
    "EU.svg",
    "BE.svg",
    "BG.svg",
    "CA.svg",
    "ZH.svg",
    "ZH.svg",
    "LZH.svg",
    "COP.svg",
    "HR.svg",
    "CS.svg",
    "CS.svg",
    "CS.svg",
    "CS.svg",
    "DA.svg",
    "NL.svg",
    "NL.svg",
    "EN.svg",
    "EN.svg",
    "EN.svg",
    "EN.svg",
    "ET.svg",
    "ET.svg",
    "FI.svg",
    "FI.svg",
    "FR.svg",
    "FR.svg",
    "FR.svg",
    "GL.svg",
    "GL.svg",
    "DE.svg",
    "DE.svg",
    "GOT.svg",
    "EL.svg",
    "HE.svg",
    "HI.svg",
    "HU.svg",
    "ID.svg",
    "GA.svg",
    "IT.svg",
    "IT.svg",
    "IT.svg",
    "IT.svg",
    "IT.svg",
    "JA.svg",
    "KO.svg",
    "KO.svg",
    "LA.svg",
    "LA.svg",
    "LA.svg",
    "LV.svg",
    "LT.svg",
    "LT.svg",
    "MT.svg",
    "MR.svg",
    "NO.svg",
    "NO.svg",
    "CU.svg",
    "FA.svg",
    "PL.svg",
    "PL.svg",
    "PT.svg",
    "PT.svg",
    "RO.svg",
    "RO.svg",
    "RU.svg",
    "RU.svg",
    "RU.svg",
    "GD.svg",
    "SR.svg",
    "SK.svg",
    "SL.svg",
    "SL.svg",
    "ES.svg",
    "ES.svg",
    "SV.svg",
    "SV.svg",
    "TA.svg",
    "TE.svg",
    "TR.svg",
    "UK.svg",
    "UR.svg",
    "UG.svg",
    "VI.svg",
    "WO.svg"
  )

  ranking <- c(
    21,
    22,
    23,
    24,
    25,
    26,
    27,
    28,
    29,
    19,
    20,
    30,
    31,
    32,
    33,
    34,
    35,
    36,
    37,
    38,
    39,
    2,
    1,
    3,
    4,
    40,
    41,
    42,
    43,
    10,
    11,
    12,
    44,
    45,
    16,
    15,
    46,
    47,
    48,
    49,
    50,
    51,
    52,
    5,
    6,
    7,
    8,
    9,
    53,
    54,
    55,
    56,
    57,
    58,
    59,
    60,
    61,
    62,
    63,
    64,
    65,
    66,
    67,
    68,
    69,
    17,
    18,
    70,
    71,
    72,
    73,
    74,
    75,
    76,
    77,
    78,
    79,
    13,
    14,
    80,
    81,
    82,
    83,
    84,
    85,
    86,
    87,
    88,
    89
  )

  df <- data.frame(
    language_name = language_name,
    treebank = treebank,
    contributors = contributors,
    description = description,
    file = file,
    folder = folder,
    hub_page_link = hub_link,
    flag = flag,
    tokens = tokens,
    words = words,
    sentences = sentences,
    chapter = toupper(chapter),
    index = ranking
  )
  # df <- df[index,]
  df <- df %>%
    arrange(index) %>%
    select(-index) %>%
    filter(!file %in% c("gl_treegal", "la_perseus"))

  return(df)
}

model_accuracy <- function() {
  # Creazione dei vettori stringa per ciascuna colonna
  Language <- c(
    "Afrikaans",
    "Arabic",
    "Belarusian",
    "Bulgarian",
    "Catalan",
    "Coptic",
    "Czech",
    "Czech",
    "Czech",
    "Czech",
    "Old Church Slavonic",
    "Danish",
    "German",
    "Greek",
    "English",
    "English",
    "English",
    "English",
    "Spanish",
    "Spanish",
    "Estonian",
    "Estonian",
    "Basque",
    "Persian",
    "Finnish",
    "Finnish",
    "French",
    "French",
    "French",
    "Irish",
    "Scottish Gaelic",
    "Galician",
    "Gothic",
    "Ancient Greek",
    "Ancient Greek",
    "Hebrew",
    "Hindi",
    "Croatian",
    "Hungarian",
    "Armenian",
    "Indonesian",
    "Italian",
    "Italian",
    "Italian",
    "Italian",
    "Italian",
    "Japanese",
    "Korean",
    "Korean",
    "Latin",
    "Latin",
    "Lithuanian",
    "Lithuanian",
    "Latvian",
    "Classical Chinese",
    "Marathi",
    "Maltese",
    "Dutch",
    "Dutch",
    "Norwegian",
    "Norwegian",
    "Polish",
    "Polish",
    "Portuguese",
    "Portuguese",
    "Romanian",
    "Romanian",
    "Russian",
    "Russian",
    "Russian",
    "Slovak",
    "Slovenian",
    "Slovenian",
    "Serbian",
    "Swedish",
    "Swedish",
    "Tamil",
    "Telugu",
    "Turkish",
    "Uyghur",
    "Ukrainian",
    "Urdu",
    "Vietnamese",
    "Wolof",
    "Chinese",
    "Chinese",
    "German"
  )

  Treebank <- c(
    "AfriBooms",
    "PADT",
    "HSE",
    "BTB",
    "AnCora",
    "Scriptorium",
    "CAC",
    "CLTT",
    "FicTree",
    "PDT",
    "PROIEL",
    "DDT",
    "GSD",
    "GDT",
    "EWT",
    "GUM",
    "LinES",
    "ParTUT",
    "AnCora",
    "GSD",
    "EDT",
    "EWT",
    "BDT",
    "Seraji",
    "FTB",
    "TDT",
    "GSD",
    "ParTUT",
    "Sequoia",
    "IDT",
    "ARCOSG",
    "CTG",
    "PROIEL",
    "Perseus",
    "PROIEL",
    "HTB",
    "HDTB",
    "SET",
    "Szeged",
    "ArmTDP",
    "GSD",
    "ISDT",
    "ParTUT",
    "PoSTWITA",
    "TWITTIRO",
    "VIT",
    "GSD",
    "GSD",
    "Kaist",
    "ITTB",
    "PROIEL",
    "ALKSNIS",
    "HSE",
    "LVTB",
    "Kyoto",
    "UFAL",
    "MUDT",
    "Alpino",
    "LassySmall",
    "Bokmaal",
    "Nynorsk",
    "LFG",
    "PDB",
    "Bosque",
    "GSD",
    "Nonstandard",
    "RRT",
    "GSD",
    "SynTagRus",
    "Taiga",
    "SNK",
    "SSJ",
    "SST",
    "SET",
    "LinES",
    "Talbanken",
    "TTB",
    "MTG",
    "IMST",
    "UDT",
    "IU",
    "UDTB",
    "VTB",
    "WTB",
    "GSD",
    "GSDSimp",
    "HDT"
  )

  Words <- c(
    99.93,
    94.51,
    99.28,
    99.92,
    99.94,
    75.38,
    99.96,
    99.37,
    99.99,
    99.91,
    100.00,
    99.93,
    99.56,
    99.89,
    99.14,
    99.57,
    99.94,
    99.78,
    99.94,
    99.72,
    99.88,
    98.95,
    99.98,
    99.65,
    99.99,
    99.68,
    98.89,
    99.41,
    99.09,
    99.67,
    97.39,
    99.14,
    100.00,
    99.98,
    100.00,
    85.08,
    99.99,
    99.91,
    99.69,
    99.38,
    99.45,
    99.80,
    99.70,
    99.36,
    99.02,
    99.75,
    95.77,
    99.85,
    99.99,
    99.97,
    99.99,
    99.95,
    97.65,
    99.23,
    97.70,
    95.27,
    99.79,
    99.59,
    99.83,
    99.82,
    99.90,
    99.86,
    99.85,
    99.67,
    99.27,
    98.88,
    99.68,
    99.53,
    99.57,
    98.74,
    100.00,
    99.89,
    100.00,
    100.00,
    99.96,
    99.91,
    94.93,
    100.00,
    97.60,
    99.74,
    99.71,
    99.99,
    85.11,
    99.16,
    89.89,
    89.05,
    99.9
  )

  Sentences <- c(
    99.65,
    81.43,
    84.80,
    94.67,
    99.08,
    30.20,
    99.76,
    93.79,
    97.56,
    91.91,
    34.43,
    89.92,
    78.69,
    89.51,
    86.17,
    95.07,
    87.88,
    99.02,
    97.92,
    94.54,
    89.81,
    75.16,
    99.42,
    98.00,
    86.93,
    85.59,
    93.45,
    98.64,
    85.01,
    97.57,
    56.07,
    94.80,
    28.25,
    98.73,
    45.89,
    99.69,
    97.69,
    93.54,
    94.48,
    97.28,
    90.47,
    96.59,
    99.02,
    24.37,
    28.57,
    96.73,
    99.45,
    93.63,
    100.00,
    92.53,
    28.48,
    87.75,
    92.04,
    99.06,
    42.03,
    81.19,
    82.32,
    89.99,
    81.93,
    97.34,
    92.74,
    99.83,
    96.06,
    89.56,
    87.99,
    96.92,
    94.01,
    97.41,
    97.55,
    84.37,
    77.73,
    99.07,
    93.73,
    93.84,
    86.83,
    94.20,
    95.08,
    98.98,
    97.51,
    82.17,
    96.71,
    97.12,
    97.02,
    90.27,
    98.80,
    98.80,
    92.6
  )

  UPOS <- c(
    95.79,
    88.91,
    96.07,
    98.01,
    98.22,
    72.69,
    98.56,
    94.87,
    97.60,
    98.63,
    91.36,
    95.24,
    64.12,
    95.13,
    94.10,
    95.77,
    95.07,
    94.30,
    98.25,
    95.49,
    95.45,
    88.85,
    92.43,
    96.02,
    92.01,
    94.28,
    96.24,
    95.26,
    96.42,
    93.43,
    91.38,
    96.23,
    93.73,
    81.44,
    95.34,
    80.84,
    95.84,
    96.67,
    90.96,
    92.39,
    93.22,
    97.24,
    96.73,
    94.25,
    90.37,
    96.38,
    93.50,
    91.40,
    92.95,
    98.42,
    94.53,
    90.37,
    71.39,
    95.23,
    87.70,
    78.79,
    93.79,
    94.18,
    94.79,
    96.39,
    95.92,
    96.86,
    97.30,
    96.12,
    95.58,
    94.94,
    96.91,
    95.00,
    97.42,
    93.42,
    92.98,
    96.78,
    94.83,
    97.36,
    94.67,
    95.71,
    78.41,
    90.29,
    88.85,
    88.80,
    94.85,
    91.53,
    76.10,
    91.73,
    82.84,
    82.28,
    97.8
  )

  XPOS <- c(
    91.39,
    82.35,
    95.40,
    94.72,
    95.04,
    72.55,
    90.88,
    82.33,
    90.26,
    93.12,
    91.63,
    99.92,
    64.29,
    95.13,
    93.23,
    95.32,
    94.08,
    93.86,
    94.92,
    99.72,
    96.80,
    91.47,
    99.98,
    95.78,
    90.74,
    95.43,
    98.89,
    94.69,
    99.09,
    92.20,
    83.27,
    95.80,
    94.20,
    71.34,
    95.65,
    80.84,
    94.91,
    89.99,
    99.69,
    99.38,
    91.90,
    97.11,
    96.37,
    93.88,
    89.60,
    95.37,
    92.80,
    81.07,
    80.41,
    93.26,
    94.73,
    81.26,
    70.36,
    87.53,
    87.29,
    95.27,
    93.49,
    91.51,
    92.21,
    97.70,
    97.30,
    87.18,
    88.49,
    99.67,
    82.76,
    90.06,
    95.94,
    94.74,
    99.57,
    98.74,
    77.04,
    90.26,
    88.51,
    90.76,
    91.95,
    93.90,
    76.99,
    90.29,
    88.54,
    90.12,
    84.03,
    89.00,
    75.07,
    91.42,
    83.56,
    82.98,
    97.4
  )

  UFeats <- c(
    93.67,
    82.44,
    88.30,
    95.77,
    97.68,
    72.88,
    89.58,
    82.64,
    90.96,
    93.21,
    82.50,
    94.48,
    53.65,
    89.96,
    94.37,
    95.10,
    94.94,
    94.21,
    97.94,
    95.97,
    93.71,
    86.94,
    87.17,
    95.92,
    92.21,
    90.54,
    95.34,
    92.58,
    94.91,
    84.15,
    85.18,
    98.92,
    86.76,
    84.62,
    87.71,
    78.57,
    90.17,
    90.59,
    87.20,
    83.70,
    94.17,
    97.12,
    96.40,
    94.26,
    90.00,
    96.33,
    95.75,
    99.50,
    99.99,
    93.98,
    86.89,
    82.42,
    66.42,
    91.82,
    88.09,
    70.06,
    99.79,
    93.28,
    93.95,
    95.06,
    94.93,
    87.48,
    88.84,
    94.07,
    91.78,
    88.31,
    96.09,
    85.57,
    90.50,
    86.33,
    79.32,
    90.53,
    88.67,
    91.10,
    90.11,
    94.18,
    79.37,
    98.89,
    84.31,
    83.63,
    83.67,
    80.26,
    85.11,
    90.89,
    88.57,
    87.79,
    91.4
  )

  AllTags <- c(
    91.26,
    82.16,
    87.69,
    94.39,
    94.83,
    71.87,
    89.45,
    82.23,
    90.14,
    92.83,
    80.73,
    93.22,
    50.34,
    88.75,
    91.61,
    93.69,
    91.84,
    92.45,
    94.54,
    93.55,
    92.15,
    83.72,
    84.68,
    95.31,
    89.00,
    89.50,
    94.51,
    91.12,
    94.17,
    81.76,
    82.49,
    95.38,
    84.55,
    71.24,
    86.28,
    77.75,
    87.68,
    89.95,
    85.84,
    82.79,
    86.65,
    96.33,
    95.19,
    92.50,
    87.00,
    94.18,
    92.30,
    78.74,
    80.41,
    92.63,
    85.59,
    80.88,
    60.98,
    87.48,
    84.79,
    67.88,
    93.20,
    90.78,
    91.41,
    93.74,
    93.29,
    86.56,
    87.88,
    92.66,
    82.58,
    87.67,
    95.83,
    84.50,
    90.11,
    85.42,
    76.77,
    89.80,
    87.49,
    90.73,
    86.66,
    92.43,
    72.93,
    90.29,
    81.58,
    76.78,
    82.86,
    75.43,
    75.04,
    88.62,
    82.30,
    81.76,
    91.0
  )

  Lemma <- c(
    96.40,
    87.36,
    88.77,
    92.88,
    98.19,
    73.30,
    96.24,
    92.93,
    95.60,
    97.50,
    81.73,
    94.46,
    86.07,
    92.74,
    95.15,
    96.55,
    95.78,
    94.76,
    98.17,
    94.08,
    87.93,
    84.30,
    91.09,
    92.74,
    85.76,
    84.52,
    96.78,
    95.84,
    96.60,
    91.96,
    91.33,
    95.18,
    89.51,
    79.90,
    90.77,
    79.01,
    98.05,
    93.30,
    87.29,
    90.83,
    87.32,
    96.98,
    96.35,
    93.84,
    89.43,
    96.52,
    94.21,
    86.96,
    88.12,
    98.04,
    91.73,
    86.89,
    68.95,
    93.35,
    97.46,
    83.15,
    99.79,
    90.82,
    92.16,
    96.04,
    95.47,
    92.56,
    94.28,
    96.50,
    91.72,
    91.61,
    95.16,
    89.90,
    95.24,
    89.54,
    83.32,
    95.04,
    94.84,
    93.40,
    94.50,
    95.13,
    82.41,
    100.00,
    85.77,
    83.47,
    91.39,
    91.47,
    83.66,
    92.81,
    89.83,
    89.00,
    92.3
  )

  UAS <- c(
    82.55,
    67.64,
    77.08,
    87.98,
    86.97,
    50.32,
    84.81,
    72.06,
    84.95,
    83.49,
    66.17,
    77.57,
    71.43,
    83.36,
    81.03,
    81.52,
    81.11,
    82.15,
    85.23,
    83.40,
    77.08,
    66.96,
    74.57,
    79.98,
    77.89,
    76.79,
    86.39,
    85.33,
    84.69,
    80.52,
    74.56,
    75.72,
    70.88,
    58.70,
    68.94,
    59.74,
    90.44,
    82.08,
    70.59,
    75.17,
    80.88,
    87.46,
    86.26,
    72.70,
    70.77,
    81.76,
    84.20,
    60.94,
    72.96,
    79.31,
    63.36,
    65.80,
    43.34,
    78.23,
    65.90,
    62.55,
    75.31,
    78.21,
    80.11,
    85.36,
    83.85,
    90.64,
    84.74,
    83.83,
    82.18,
    81.26,
    82.69,
    81.58,
    83.19,
    72.77,
    78.84,
    84.17,
    70.58,
    84.90,
    81.31,
    81.51,
    59.66,
    87.93,
    61.37,
    67.06,
    77.54,
    81.25,
    47.83,
    75.18,
    58.96,
    58.27,
    77.0
  )

  LAS <- c(
    77.97,
    62.60,
    73.45,
    83.52,
    83.75,
    48.26,
    81.13,
    67.99,
    80.79,
    80.18,
    59.74,
    73.77,
    65.53,
    79.58,
    77.64,
    78.18,
    76.60,
    78.75,
    81.52,
    79.41,
    73.13,
    60.45,
    69.35,
    75.19,
    72.65,
    72.74,
    83.00,
    82.11,
    81.63,
    72.83,
    68.31,
    72.53,
    63.47,
    51.74,
    63.73,
    56.38,
    85.86,
    77.39,
    64.95,
    67.58,
    74.29,
    84.47,
    83.08,
    67.74,
    64.11,
    77.58,
    82.36,
    51.64,
    61.60,
    75.58,
    57.80,
    59.65,
    31.05,
    73.52,
    60.15,
    51.15,
    68.95,
    73.20,
    75.79,
    82.30,
    80.62,
    86.73,
    80.83,
    79.97,
    78.52,
    75.55,
    77.96,
    76.41,
    79.37,
    67.12,
    74.61,
    80.51,
    65.41,
    80.63,
    76.54,
    77.23,
    49.62,
    75.45,
    52.36,
    53.01,
    72.78,
    74.23,
    40.71,
    69.40,
    55.27,
    54.47,
    81.4
  )

  df <- data.frame(
    language_name = gsub(" ", "_", tolower(Language)),
    treebank = Treebank,
    Words = Words,
    Sentences = Sentences,
    UPOS = UPOS,
    XPOS = XPOS,
    UFeats = UFeats,
    AllTags = AllTags,
    Lemma = Lemma,
    UAS = UAS,
    LAS = LAS,
    stringsAsFactors = FALSE # Per mantenere le colonne di testo come stringhe
  )

  return(df)
}

##  Server ####
files <- c(
  "libraries.R",
  "tallAI.R",
  "tallShot.R",
  "tallUtils.R",
  "tallTextIO.R",
  "tallNLP.R",
  "tallOverview.R",
  "tallNetwork.R",
  "tallTopicModel.R",
  "tallEmbeddings.R",
  "tallSentiment.R",
  "tallReport.R",
  "tallVisualization.R",
  "tallLanguages.R",
  "helpContent.R",
  "cssTags.R",
  "header.R",
  "home.R",
  "import.R",
  "edit.R",
  "preprocessing.R",
  "filters_groups.R",
  "overview.R",
  "featureroles.R",
  "keyness.R",
  "collocation.R",
  "words.R",
  "documents.R",
  "doc_classification.R",
  "report.R",
  "settings.R"
)

for (file in files) {
  source(file, local = TRUE)
}

## file upload max size
maxUploadSize <- 1000 # default value
maxUploadSize <- getShinyOption("maxUploadSize", maxUploadSize)
options(shiny.maxRequestSize = maxUploadSize * 1024^2)

param_stay_page <- FALSE

server <- function(input, output, session) {
  ## suppress summarise message
  options(dplyr.summarise.inform = FALSE)

  ## Load analysis packages after UI has rendered
  libraries_analysis()

  ## Setup async execution for AI calls
  future::plan(future::multisession, workers = 2)

  ## Inject JS for DPI-aware visNetwork canvas capture (same approach as biblioshiny)
  session$onFlushed(function() {
    shinyjs::runjs('
      window._captureVisCanvas = function(networkId, dpi, callback) {
        var widget = HTMLWidgets.find("#" + networkId);
        if (!widget || !widget.network) return;
        var network = widget.network;
        var canvas = network.canvas.frame.canvas;
        var targetScale = (dpi || 96) / 96;
        Object.defineProperty(window, "devicePixelRatio", {
          value: targetScale, configurable: true
        });
        network.redraw();
        var dataURL = canvas.toDataURL("image/png");
        delete window.devicePixelRatio;
        network.redraw();
        callback(dataURL);
      };
      window.captureVisExport = function(networkId, filename, dpi) {
        _captureVisCanvas(networkId, dpi, function(dataURL) {
          var link = document.createElement("a");
          link.href = dataURL;
          link.download = filename;
          document.body.appendChild(link);
          link.click();
          document.body.removeChild(link);
        });
      };
      window.captureVisReport = function(networkId, dpi) {
        _captureVisCanvas(networkId, dpi, function(dataURL) {
          Shiny.setInputValue("vis_canvas_report", {
            data: dataURL
          }, {priority: "event"});
        });
      };
    ')
  }, once = TRUE)

  ### Initial values ----
  values <- resetValues()
  statsValues <- updateStats(NULL, "token")

  ## Setting plot values
  values$dpi <- 300
  values$report_dpi <- 72
  values$h <- 7
  values$aspect <- 1.5  # Publication (3:2) default; Wide = 2.0

  ## Load saved graph settings
  home <- homeFolder()
  path_graph_settings <- file.path(home, "tall", ".tall_graph_settings.txt")
  if (file.exists(path_graph_settings)) {
    graph_settings <- readLines(path_graph_settings, warn = FALSE)
    if (length(graph_settings) >= 3) {
      saved_dpi <- suppressWarnings(as.numeric(graph_settings[1]))
      saved_report_dpi <- suppressWarnings(as.numeric(graph_settings[2]))
      saved_h <- suppressWarnings(as.numeric(graph_settings[3]))
      if (!is.na(saved_dpi)) values$dpi <- saved_dpi
      if (!is.na(saved_report_dpi)) values$report_dpi <- saved_report_dpi
      if (!is.na(saved_h)) values$h <- saved_h
      if (length(graph_settings) >= 4) {
        saved_aspect <- suppressWarnings(as.numeric(graph_settings[4]))
        if (!is.na(saved_aspect) && saved_aspect %in% c(1.5, 2.0)) {
          values$aspect <- saved_aspect
        }
      }
    }
  }

  ## Render settings sliders with saved values
  output$dpi_slider <- renderUI({
    sliderTextInput(
      inputId = "dpi",
      label = NULL,
      grid = TRUE,
      force_edges = TRUE,
      choices = c("75", "150", "300", "600"),
      width = "100%",
      selected = as.character(as.integer(isolate(values$dpi)))
    )
  })
  output$report_dpi_slider <- renderUI({
    sliderTextInput(
      inputId = "report_dpi",
      label = NULL,
      grid = TRUE,
      force_edges = TRUE,
      choices = c("72", "150", "300"),
      width = "100%",
      selected = as.character(as.integer(isolate(values$report_dpi)))
    )
  })
  output$h_slider <- renderUI({
    sliderTextInput(
      inputId = "h",
      label = NULL,
      grid = TRUE,
      force_edges = TRUE,
      choices = seq(5, 15),
      width = "100%",
      selected = as.character(as.integer(isolate(values$h)))
    )
  })
  output$aspect_radio <- renderUI({
    radioGroupButtons(
      inputId = "aspect",
      label = NULL,
      choices = c("Publication (3:2)" = "1.5", "Wide (2:1)" = "2"),
      selected = as.character(isolate(values$aspect)),
      justified = TRUE,
      size = "sm",
      status = "primary"
    )
  })

  ## Define JS functions for visNetwork canvas capture (DPI-aware)
  session$onFlushed(
    function() {
      shinyjs::runjs(
        '
      window._captureVisCanvas = function(networkId, dpi, callback) {
        var widget = HTMLWidgets.find("#" + networkId);
        if (!widget || !widget.network) return;
        var network = widget.network;
        var canvas = network.canvas.frame.canvas;
        var targetScale = (dpi || 96) / 96;

        Object.defineProperty(window, "devicePixelRatio", {
          value: targetScale, configurable: true
        });

        network.redraw();
        var dataURL = canvas.toDataURL("image/png");

        delete window.devicePixelRatio;
        network.redraw();

        callback(dataURL);
      };

      window.captureVisExport = function(networkId, filename, dpi) {
        _captureVisCanvas(networkId, dpi, function(dataURL) {
          var link = document.createElement("a");
          link.href = dataURL;
          link.download = filename;
          document.body.appendChild(link);
          link.click();
          document.body.removeChild(link);
        });
      };

      window.captureVisReport = function(networkId, dpi) {
        _captureVisCanvas(networkId, dpi, function(dataURL) {
          Shiny.setInputValue("vis_canvas_report", {
            data: dataURL
          }, {priority: "event"});
        });
      };
    '
      )
    },
    once = TRUE
  )

  saved_message <- "Done!"

  ### SIDEBARMENU ----
  homeServer(input, output, session, values, statsValues)

  ### IMPORT ----
  importServer(input, output, session, values, statsValues)

  ### EDIT ----
  editServer(input, output, session, values, statsValues)

  ### PRE-PROCESSING ----
  preprocessingServer(input, output, session, values, statsValues)

  ### FILTERS & GROUPS ----
  filters_groupsServer(input, output, session, values, statsValues)

  ## FEATURE ROLES ----
  featureRolesServer(input, output, session, values)

  ## OVERVIEW ----
  overviewServer(input, output, session, values, statsValues)

  ### KEYNESS ----
  keynessServer(input, output, session, values)

  ## WORDS ----
  wordsServer(input, output, session, values, statsValues)

  ### COLLOCATION ----
  collocationServer(input, output, session, values, statsValues)

  ## DOCUMENTS ----
  documentsServer(input, output, session, values, statsValues)

  ### Doc Supervised Classification ----
  docClassificationServer(input, output, session, values)

  ## REPORT ----
  reportServer(input, output, session, values)

  ## SETTINGS & UTILITY ----
  settingsServer(input, output, session, values, statsValues)

  ## Header Object ----
  headerServer(input, output, session, values)
} # END SERVER

reportUI <- function() {
  tabItem(
    tabName = "report",
    fluidPage(
      fluidRow(
        h2(icon("file-pdf"), strong("Report"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;"),
        br(),
      ),
      fluidRow(
        column(
          6,
          offset = 1,
          box(
            title = strong(
              "Select results to include in the Report",
              style = "font-size:20px;color:white;"
            ),
            status = "primary",
            width = 11,
            solidHeader = TRUE,
            tags$style(HTML(
              "
                         .box.box-solid.box-primary>.box-header {
                         background:#4F7942;
                         }
                         .box.box-solid.box-primary{
                         border-bottom-color:black;
                         border-left-color:black;
                         border-right-color:black;
                         border-top-color:black;
                         border-width:2px;
                                         }"
            )),
            br(),
            uiOutput("reportSheetsUI"),
            tags$style("#reportSheets {font-size:18px;}")
          )
        ), # column(1),
        column(
          2,
          div(
            style = "border-radius: 10px; border-width: 3px; font-size: 10px;",
            align = "center",
            actionBttn(
              inputId = "allSheets",
              label = strong("Select All"),
              icon = icon("circle-check"),
              style = "pill",
              color = "primary",
              block = TRUE
            ),
            br(),
            actionBttn(
              inputId = "noSheets",
              label = strong("Deselect All"),
              icon = icon("circle-xmark"),
              style = "pill",
              color = "primary",
              block = TRUE
            ),
            br(),
            hr(),
            actionBttn(
              inputId = "report.save",
              label = strong("Export Report"),
              style = "pill",
              color = "success",
              size = "md",
              block = TRUE,
              no_outline = TRUE,
              icon = icon("download")
            ),
            br(),
            hr(),
            actionBttn(
              inputId = "deleteAll",
              label = strong("Delete Report"),
              icon = icon("circle-exclamation"),
              style = "pill",
              color = "danger",
              block = TRUE
            )
          )
        )
      )
    )
  )
}

reportServer <- function(input, output, session, values) {
  ## REPORT ----
  ### Report Save xlsx ----
  observeEvent(
    eventExpr = {
      input$report.save
    },
    handlerExpr = {
      file <- paste("TallReport-", sys.time(), ".xlsx", sep = "")
      file <- destFolder(file, values$wdTall)

      wb_export <- copyWorkbook(values$wb)
      if (nrow(values$list_file) > 0) {
        wb_export <- addScreenWb(df = values$list_file, wb = wb_export) # , width=10, height=7, dpi=300)
      }
      sheetToRemove <- setdiff(sheets(wb_export), input$reportSheets)
      if (length(sheetToRemove) > 0) {
        for (i in sheetToRemove) {
          removeWorksheet(wb_export, i)
        }
      }
      sheetToAdd <- sheets(wb_export)
      for (i in sheetToAdd) {
        setColWidths(
          wb_export,
          sheet = i,
          cols = 1,
          widths = 30,
          hidden = FALSE
        )
      }
      openxlsx::saveWorkbook(wb_export, file = file)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ### Report UI elements
  observe({
    output$reportSheetsUI <- renderUI({
      prettyCheckboxGroup(
        inputId = "reportSheets",
        label = NULL, # short2long(df=values$dfLabel, myC=values$myChoices),
        choices = short2long(df = values$dfLabel, myC = values$myChoices),
        selected = values$myChoices,
        icon = icon("check"),
        animation = "pulse",
        status = "primary",
        bigger = T,
        fill = TRUE
      )
    })
  })

  observe({
    updatePrettyCheckboxGroup(
      session = getDefaultReactiveDomain(),
      inputId = "reportSheets",
      # label = short2long(df=values$dfLabel, myC=values$myChoices),
      choices = short2long(df = values$dfLabel, myC = values$myChoices),
      selected = if (!input$noSheets) values$myChoices,
      prettyOptions = list(
        animation = "pulse",
        status = "info",
        bigger = T
      )
    )
  })

  observe({
    updatePrettyCheckboxGroup(
      session = getDefaultReactiveDomain(),
      inputId = "reportSheets",
      choices = short2long(df = values$dfLabel, myC = values$myChoices),
      selected = if (input$allSheets) values$myChoices,
      prettyOptions = list(
        animation = "pulse",
        status = "info",
        bigger = T
      )
    )
  })

  observeEvent(input$deleteAll, {
    ask_confirmation(
      inputId = "delete_confirmation",
      title = "Want to confirm?",
      text = "All the results will be removed from the report",
      type = "warning",
      btn_labels = c("CANCEL", "CONFIRM"),
    )
  })

  observeEvent(
    input$delete_confirmation,
    {
      if (isTRUE(input$delete_confirmation)) {
        values$myChoices <- "Empty Report"
        values$list_file <- data.frame(sheet = NULL, file = NULL, n = NULL)
        values$wb <- openxlsx::createWorkbook()
      }
    },
    ignoreNULL = TRUE
  )
}

headerUI <- function() {
  list(
    tags$li(class = "dropdown", tags$a(uiOutput("dataGroupedBy"))), # Rimosso HTML/paste
    tags$li(class = "dropdown", tags$a(uiOutput("dataFilteredBy"))),
    # tags$li(class = "dropdown", tags$a(uiOutput("termSelected"))),
    tags$li(
      class = "dropdown",
      actionButton(
        inputId = "go_to_postagging",
        label = tags$a(
          HTML(
            paste(uiOutput("termSelected"))
          ),
          style = "color: #fff;"
        ),
        # icon = icon("gear", lib = "font-awesome"),
        style = "background: transparent;
               border: none;
               color: #fff;
               font-size: 16px;  /* Increased from 16px to 20px */
               margin-top: 7px;
               cursor: pointer;",
        title = "PoS Tagging"
      )
    ),
    tags$li(
      class = "dropdown",
      tags$a(
        title = "Total downloads from CRAN",
        icon("cloud-arrow-down", lib = "font-awesome"),
        uiOutput("header_download_count", inline = TRUE)
      )
    ),
    tags$li(
      class = "dropdown",
      actionButton(
        "show_donate",
        label = NULL,
        icon = icon("heart"),
        style = "background: none; border: none; color: white; font-size: 15px; padding: 15px; margin: 0; transition: background 0.3s;",
        onmouseover = "this.style.background='rgba(0,0,0,0.1)';",
        onmouseout = "this.style.background='none';"
      )
    ),

    # Author CARD ----
    tags$li(
      class = "dropdown",
      actionButton(
        "show_team",
        label = NULL,
        icon = icon("users"),
        style = "background: none; border: none; color: white; font-size: 15px; padding: 15px; margin: 0;"
      )
    ),

    tags$li(
      class = "dropdown",
      actionButton(
        "show_credits",
        label = NULL,
        icon = icon("cube"),
        style = "background: none; border: none; color: white; font-size: 15px; padding: 15px; margin: 0; transition: background 0.3s;",
        onmouseover = "this.style.background='rgba(0,0,0,0.1)';",
        onmouseout = "this.style.background='none';"
      )
    ),

    # Settings Button - uses actionLink to trigger server-side tab change
    tags$li(
      class = "dropdown",
      actionButton(
        inputId = "go_to_settings",
        label = NULL,
        icon = icon("gear", lib = "font-awesome"),
        style = "background: transparent;
               border: none;
               color: #fff;
               font-size: 20px;  /* Increased from 16px to 20px */
               margin-top: 7px;
               cursor: pointer;",
        title = "Settings"
      )
    ),
    tags$li(
      class = "dropdown",
      tags$style(".main-header .logo {height: 53px}")
    )
  )
}

headerServer <- function(input, output, session, values) {
  ## CRAN download count (async, non-blocking) ----
  output$header_download_count <- renderUI({
    downloads <- suppressWarnings(total_downloads("tall"))
    tags$span(
      HTML(format_abbreviated(downloads)),
      style = "margin-left: 5px; font-weight: bold;"
    )
  })

  ## Team Card ----
  observeEvent(input$show_team, {
    showModal(modalDialog(
      title = div(
        style = "text-align: center; margin-bottom: 10px;",
        icon("users", style = "color: #3c8dbc; font-size: 26px;"),
        span(
          " TALL Team",
          style = "font-size: 24px; font-weight: bold; color: #2c3e50; margin-left: 10px;"
        )
      ),
      size = "l",
      easyClose = TRUE,
      footer = modalButton("Close"),

      # Creators Section
      div(
        style = "padding: 20px;",

        div(
          style = "margin-bottom: 35px;",
          h4(
            icon("star", style = "color: #f39c12;"),
            " Creators",
            style = "color: #2c3e50; margin-bottom: 20px; border-bottom: 2px solid #3c8dbc; padding-bottom: 10px; font-weight: bold;"
          ),

          div(
            style = "display: grid; grid-template-columns: repeat(auto-fit, minmax(200px, 1fr)); gap: 25px; margin-bottom: 10px;",

            createAuthorCard(
              name = "Massimo Aria",
              title = "Full Professor of Statistics for Social Sciences",
              affiliation = "University of Naples Federico II",
              url = "https://www.massimoaria.com",
              photo = "images/team/massimo_aria.jpg",
              scholar = FALSE
            ),

            createAuthorCard(
              name = "Maria Spano",
              title = "Associate Professor of Statistics for Social Sciences",
              affiliation = "University of Naples Federico II",
              url = "https://scholar.google.com/citations?user=kh_hGT0AAAAJ&hl=it&oi=ao",
              photo = "images/team/maria_spano.jpg",
              scholar = TRUE
            ),

            createAuthorCard(
              name = "Luca D'Aniello",
              title = "Assistant Professor of Statistics for Social Sciences",
              affiliation = "University of Naples Federico II",
              url = "https://scholar.google.com/citations?user=IXJxh0MAAAAJ&hl=it&oi=ao",
              photo = "images/team/luca_daniello.jpg",
              scholar = TRUE
            )
          )
        ),

        # Contributors Section
        div(
          h4(
            icon("hands-helping", style = "color: #95a5a6;"),
            " Contributors",
            style = "color: #2c3e50; margin-bottom: 20px; border-bottom: 2px solid #95a5a6; padding-bottom: 10px; font-weight: bold;"
          ),

          div(
            style = "display: grid; grid-template-columns: repeat(auto-fit, minmax(200px, 1fr)); gap: 25px;",

            createAuthorCard(
              name = "Corrado Cuccurullo",
              title = "Full Professor of Corporate Governance",
              affiliation = "University of Campania Luigi Vanvitelli",
              url = "https://www.corradocuccurullo.com/",
              photo = "images/team/corrado_cuccurullo.jpg",
              scholar = FALSE
            ),

            createAuthorCard(
              name = "Michelangelo Misuraca",
              title = "Associate Professor of Statistics for Social Sciences",
              affiliation = "University of Salerno",
              url = "https://scholar.google.com/citations?user=WdivjAUAAAAJ&hl=it",
              photo = "images/team/michelangelo_misuraca.jpg",
              scholar = TRUE
            )
          )
        )
      )
    ))
  })

  ## Link Card ----
  observeEvent(input$show_credits, {
    showModal(modalDialog(
      title = div(
        style = "text-align: center; margin-bottom: 10px;",
        icon("cube", style = "color: #667eea; font-size: 26px;"),
        span(
          " Credits",
          style = "font-size: 24px; font-weight: bold; color: #2c3e50; margin-left: 10px;"
        )
      ),
      size = "m",
      easyClose = TRUE,
      footer = modalButton("Close"),

      div(
        style = "padding: 20px;",
        # K-Synth Card
        tags$a(
          href = "https://www.k-synth.com",
          target = "_blank",
          style = "text-decoration: none; display: block; margin-bottom: 20px;",
          div(
            style = "background: linear-gradient(135deg, #667eea 0%, #764ba2 100%); padding: 25px; border-radius: 12px; box-shadow: 0 5px 20px rgba(102, 126, 234, 0.3); transition: all 0.3s; cursor: pointer;",
            onmouseover = "this.style.transform='translateY(-5px)'; this.style.boxShadow='0 8px 25px rgba(102, 126, 234, 0.4)';",
            onmouseout = "this.style.transform='translateY(0)'; this.style.boxShadow='0 5px 20px rgba(102, 126, 234, 0.3)';",

            div(
              style = "display: flex; align-items: center; margin-bottom: 15px;",
              div(
                style = "width: 60px; height: 60px; border-radius: 50%; background: white; display: flex; align-items: center; justify-content: center; margin-right: 20px; box-shadow: 0 3px 10px rgba(0,0,0,0.1);",
                icon(
                  "watchman-monitoring",
                  style = "color: #667eea; font-size: 28px;"
                )
              ),
              div(
                div(
                  "K-Synth srl",
                  style = "color: white; font-size: 22px; font-weight: bold; margin-bottom: 5px;"
                ),
                div(
                  "Academic Spin-Off of University of Naples Federico II",
                  style = "color: rgba(255,255,255,0.9); font-size: 14px;"
                )
              )
            ),

            div(
              style = "color: rgba(255,255,255,0.95); font-size: 13px; line-height: 1.6; margin-bottom: 12px;",
              'A science-centric information & intelligence Specialist Firm, specialized in the research, consulting, and production of "knowledge", extracted and analyzed from a heterogeneous plurality of large volumes of data, through quantitative, statistical, and data science methods.'
            ),

            div(
              icon("external-link-alt", style = "margin-right: 5px;"),
              "Visit K-Synth",
              style = "color: white; font-size: 13px; font-weight: 600; display: inline-block; padding: 8px 16px; background: rgba(255,255,255,0.2); border-radius: 6px;"
            )
          )
        ),

        # GitHub Card
        tags$a(
          href = "https://github.com/massimoaria/tall",
          target = "_blank",
          style = "text-decoration: none; display: block;",
          div(
            style = "background: linear-gradient(135deg, #24292e 0%, #000000 100%); padding: 25px; border-radius: 12px; box-shadow: 0 5px 20px rgba(0, 0, 0, 0.3); transition: all 0.3s; cursor: pointer;",
            onmouseover = "this.style.transform='translateY(-5px)'; this.style.boxShadow='0 8px 25px rgba(0, 0, 0, 0.4)';",
            onmouseout = "this.style.transform='translateY(0)'; this.style.boxShadow='0 5px 20px rgba(0, 0, 0, 0.3)';",

            div(
              style = "display: flex; align-items: center; margin-bottom: 15px;",
              div(
                style = "width: 60px; height: 60px; border-radius: 50%; background: white; display: flex; align-items: center; justify-content: center; margin-right: 20px; box-shadow: 0 3px 10px rgba(0,0,0,0.1);",
                icon("github", style = "color: #24292e; font-size: 28px;")
              ),
              div(
                div(
                  "GitHub",
                  style = "color: white; font-size: 22px; font-weight: bold; margin-bottom: 5px;"
                ),
                div(
                  "Open Source Repository",
                  style = "color: rgba(255,255,255,0.9); font-size: 14px;"
                )
              )
            ),

            div(
              style = "color: rgba(255,255,255,0.95); font-size: 13px; line-height: 1.6; margin-bottom: 12px;",
              "Access the source code, report issues, contribute to development, and stay updated with the latest releases."
            ),

            div(
              icon("external-link-alt", style = "margin-right: 5px;"),
              "View Repository",
              style = "color: white; font-size: 13px; font-weight: 600; display: inline-block; padding: 8px 16px; background: rgba(255,255,255,0.2); border-radius: 6px;"
            )
          )
        )
      )
    ))
  })

  ## Donation Card ----
  observeEvent(input$show_donate, {
    showModal(modalDialog(
      title = div(
        style = "text-align: center; margin-bottom: 10px;",
        icon("heart", style = "color: #e74c3c; font-size: 26px;"),
        span(
          " Support TALL",
          style = "font-size: 24px; font-weight: bold; color: #2c3e50; margin-left: 10px;"
        )
      ),
      size = "m",
      easyClose = TRUE,
      footer = modalButton("Close"),

      div(
        style = "padding: 20px;",

        # Messaggio principale
        div(
          style = "text-align: center; margin-bottom: 30px;",

          div(
            style = "font-size: 16px; color: #2c3e50; line-height: 1.8; margin-bottom: 20px;",
            "TALL is a free and open-source tool developed with passion by our research team.",
            tags$br(),
            "Your support helps us maintain and improve the project."
          ),

          div(
            style = "font-size: 18px; color: #e74c3c; font-weight: bold; margin-bottom: 25px;",
            icon("heart", style = "margin-right: 8px;"),
            "Help us keep TALL free for everyone"
          )
        ),

        # Card donazione principale
        tags$a(
          href = "https://www.bibliometrix.org/home/index.php/donation",
          target = "_blank",
          style = "text-decoration: none; display: block;",
          div(
            style = "background: linear-gradient(135deg, #e74c3c 0%, #c0392b 100%); padding: 30px; border-radius: 12px; box-shadow: 0 5px 20px rgba(231, 76, 60, 0.3); transition: all 0.3s; cursor: pointer;",
            onmouseover = "this.style.transform='translateY(-5px)'; this.style.boxShadow='0 8px 25px rgba(231, 76, 60, 0.4)';",
            onmouseout = "this.style.transform='translateY(0)'; this.style.boxShadow='0 5px 20px rgba(231, 76, 60, 0.3)';",

            div(
              style = "text-align: center;",

              # Icona grande
              div(
                style = "width: 80px; height: 80px; border-radius: 50%; background: white; display: flex; align-items: center; justify-content: center; margin: 0 auto 20px; box-shadow: 0 4px 15px rgba(0,0,0,0.1);",
                icon(
                  "hand-holding-heart",
                  style = "color: #e74c3c; font-size: 35px;"
                )
              ),

              # Titolo
              div(
                "Make a Donation",
                style = "color: white; font-size: 24px; font-weight: bold; margin-bottom: 12px;"
              ),

              # Descrizione
              div(
                "Support the development and maintenance of TALL",
                style = "color: rgba(255,255,255,0.95); font-size: 14px; margin-bottom: 20px; line-height: 1.6;"
              ),

              # Button
              div(
                icon("external-link-alt", style = "margin-right: 8px;"),
                "Donate Now",
                style = "color: #e74c3c; font-size: 16px; font-weight: 700; display: inline-block; padding: 12px 30px; background: white; border-radius: 8px; box-shadow: 0 3px 10px rgba(0,0,0,0.1);"
              )
            )
          )
        ),

        # Info aggiuntive
        div(
          style = "margin-top: 25px; padding: 20px; background: #f8f9fa; border-radius: 8px; border-left: 4px solid #e74c3c;",

          div(
            icon("info-circle", style = "color: #3498db; margin-right: 8px;"),
            strong("Why donate?", style = "color: #2c3e50; font-size: 14px;")
          ),

          div(
            style = "margin-top: 10px; color: #7f8c8d; font-size: 13px; line-height: 1.6;",
            tags$ul(
              style = "margin: 10px 0; padding-left: 20px;",
              tags$li("Keep the tool free and accessible"),
              tags$li("Fund new features and improvements"),
              tags$li("Support open-source research software"),
              tags$li("Enable continuous maintenance and updates")
            )
          )
        )
      )
    ))
  })
}

## TALL functions ----

# Libraries ----

# Core packages needed for UI rendering at startup
libraries <- function() {
  # Framework and UI essentials
  suppressPackageStartupMessages(library(tall))
  suppressPackageStartupMessages(library(shiny))
  suppressPackageStartupMessages(library(shinydashboard))
  suppressPackageStartupMessages(library(shinydashboardPlus))
  suppressPackageStartupMessages(library(shinyWidgets))
  suppressPackageStartupMessages(library(shinycssloaders))
  suppressPackageStartupMessages(library(shinyjs))
  suppressPackageStartupMessages(library(shinyFiles))
  suppressPackageStartupMessages(library(fontawesome))
  # Core data manipulation
  suppressPackageStartupMessages(library(dplyr))
  suppressPackageStartupMessages(library(tidyr))
  suppressPackageStartupMessages(library(purrr))
  suppressPackageStartupMessages(library(tibble))
  suppressPackageStartupMessages(library(glue))
  # Core display
  suppressPackageStartupMessages(library(DT))
  suppressPackageStartupMessages(library(plotly))
  suppressPackageStartupMessages(library(visNetwork))
}

# Analysis packages loaded after UI renders (called from server.R)
libraries_analysis <- function() {
  suppressPackageStartupMessages(library(readr))
  suppressPackageStartupMessages(library(openxlsx))
  suppressPackageStartupMessages(library(readxl))
  suppressPackageStartupMessages(library(readtext))
  suppressPackageStartupMessages(library(pdftools))
  suppressPackageStartupMessages(library(jsonlite))
  suppressPackageStartupMessages(library(udpipe))
  suppressPackageStartupMessages(library(ca))
  suppressPackageStartupMessages(library(topicmodels))
  suppressPackageStartupMessages(library(textrank))
  suppressPackageStartupMessages(library(igraph))
  suppressPackageStartupMessages(library(tidygraph))
  suppressPackageStartupMessages(library(strucchange))
  suppressPackageStartupMessages(library(sparkline))
  if (requireNamespace("chromote", quietly = TRUE)) suppressPackageStartupMessages(library(chromote))
  suppressPackageStartupMessages(library(word2vec))
  suppressPackageStartupMessages(library(umap))
  suppressPackageStartupMessages(library(httr2))
  suppressPackageStartupMessages(library(ggwordcloud))
  suppressPackageStartupMessages(library(future))
  suppressPackageStartupMessages(library(promises))
  suppressPackageStartupMessages(library(ranger))
}

messageItem2 <- function(
  from,
  message,
  icon = shiny::icon("user"),
  time = NULL,
  href = NULL,
  inputId = NULL
) {
  if (is.null(href)) {
    href <- "#"
  }
  shiny::tags$li(shiny::a(
    id = inputId,
    class = if (!is.null(inputId)) {
      "action-button"
    },
    href = href,
    target = "_blank",
    icon,
    shiny::h4(
      from,
      if (!is.null(time)) {
        shiny::tags$small(shiny::icon("clock-o"), time)
      }
    ),
    shiny::p(message)
  ))
}

messageItemCustom <- function(
  from,
  message,
  icon = shiny::icon("user"),
  time = NULL,
  href = NULL,
  inputId = NULL
) {
  style_icon <- paste0(
    "
  background: url('",
    icon,
    ".svg') ;
  background-size: cover;
  background-position: left center;
  background-repeat: no-repeat;
  height: 30px;
  width: 30px;
  display: block;
  font-family: Font Awesome 6 Free;
  position: sticky;
  text-align: center;
  line-height: inherit;
  left: 1.35em;
  padding: 0.2em 0.25em 0.15em;
  border-radius: var(--fa-border-radius, 0.1em);
  border-style: var(--fa-border-style, solid);
  border-width: var(--fa-border-width, 0em);
  "
  )

  icon <- icon(
    name = NULL,
    style = style_icon
  )

  from <- paste0("&nbsp;&nbsp;", from)
  if (is.null(href)) {
    href <- "#"
  }
  shiny::tags$li(shiny::a(
    id = inputId,
    class = if (!is.null(inputId)) {
      "action-button"
    },
    href = href,
    target = "_blank",
    icon,
    shiny::h4(
      HTML(from),
      style = "font-style: inherit; text-align: left;",
      if (!is.null(time)) {
        shiny::tags$small(shiny::icon("clock-o"), time)
      }
    ),
    shiny::p(message)
  ))
}


# Custom Theme ----
customTheme <- function() {
  shinyDashboardThemeDIY(
    appFontFamily = "Helvetica",
    appFontColor = "rgb(0,0,0)",
    primaryFontColor = "rgb(0,0,0)",
    infoFontColor = "rgb(0,0,0)",
    successFontColor = "rgb(0,0,0)",
    warningFontColor = "rgb(0,0,0)",
    dangerFontColor = "rgb(0,0,0)",
    bodyBackColor = "rgb(255,255,255)",
    logoBackColor = "#4F7942", # "rgb(88,101,185)"
    headerButtonBackColor = "#4F7942", # "rgb(88,101,185)"
    headerButtonIconColor = "rgb(248,248,248)",
    headerButtonBackColorHover = "#4F7942", # "rgb(75,90,179)"
    headerButtonIconColorHover = "rgb(248,248,248)",
    headerBackColor = "#4F7942", # "rgb(88,101,185)"
    headerBoxShadowColor = "rgb(210,210,210)",
    headerBoxShadowSize = "2px 2px 2px",
    sidebarBackColor = cssGradientThreeColors(
      direction = "down",
      colorStart = "#4F7942", # "rgb(88,101,185)"
      colorMiddle = "#6CC283", # "rgb(29,143,225)"
      colorEnd = "#9ED69C", # "rgb(34,220,253)"
      colorStartPos = 0,
      colorMiddlePos = 55,
      colorEndPos = 100
    ),
    sidebarPadding = 0,
    sidebarMenuBackColor = "transparent",
    sidebarMenuPadding = 0,
    sidebarMenuBorderRadius = 0,
    sidebarShadowRadius = "0px 0px 0px",
    sidebarShadowColor = "#aaaaaa",
    sidebarUserTextColor = "rgb(255,255,255)",
    sidebarSearchBackColor = cssGradientThreeColors(
      direction = "right",
      colorStart = "#9ED69C", # "rgb(34,220,253)"
      colorMiddle = "#6CC283", # "rgb(29,143,225)"
      colorEnd = "#4F7942", # "rgb(88,101,185)"
      colorStartPos = 0,
      colorMiddlePos = 75,
      colorEndPos = 100
    ),
    sidebarSearchIconColor = "rgb(255,255,255)",
    sidebarSearchBorderColor = "#6CC283", # "rgb(29,143,225)"
    sidebarTabTextColor = "rgb(255,255,255)",
    sidebarTabTextSize = 15,
    sidebarTabBorderStyle = "none none solid none",
    sidebarTabBorderColor = "rgb(35,106,135)",
    sidebarTabBorderWidth = 1,
    sidebarTabBackColorSelected = cssGradientThreeColors(
      direction = "right",
      colorStart = "#9ED69C", # "rgb(34,220,253)"
      colorMiddle = "#6CC283", # "rgb(29,143,225)"
      colorEnd = "#4F7942", # "rgb(88,101,185)"
      colorStartPos = 0,
      colorMiddlePos = 75,
      colorEndPos = 100
    ),
    sidebarTabTextColorSelected = "rgb(255,255,255)",
    sidebarTabRadiusSelected = "0px 0px 0px 0px",
    sidebarTabBackColorHover = "rgb(255,255,255)",
    sidebarTabTextColorHover = "rgb(50,50,50)",
    sidebarTabBorderStyleHover = "none none solid none",
    sidebarTabBorderColorHover = "rgb(75,126,151)",
    sidebarTabBorderWidthHover = 1,
    sidebarTabRadiusHover = "0px 0px 0px 0px",
    boxBackColor = "#F5F5F560", # "rgb(255,255,255)"
    boxBorderRadius = 5,
    boxShadowSize = "0px 0px 0px",
    boxShadowColor = "rgba(0,0,0,.1)",
    boxTitleSize = 1,
    boxDefaultColor = "rgb(210,214,220)",
    boxPrimaryColor = "rgba(44,222,235,1)",
    boxInfoColor = "rgb(210,214,220)",
    boxSuccessColor = "#6CC283", # "rgb(88,101,185)"
    boxWarningColor = "rgb(244,156,104)",
    boxDangerColor = "rgb(255,88,55)",
    tabBoxTabColor = "rgb(255,255,255)",
    tabBoxTabTextSize = 14,
    tabBoxTabTextColor = "rgb(0,0,0)",
    tabBoxTabTextColorSelected = "rgb(0,0,0)",
    tabBoxBackColor = "rgb(255,255,255)",
    tabBoxHighlightColor = "rgba(44,222,235,1)",
    tabBoxBorderRadius = 5,
    buttonBackColor = cssGradientThreeColors(
      direction = "right",
      colorStart = "#9ED69C", # "rgb(34,220,253)"
      colorMiddle = "#6CC283", # "rgb(29,143,225)"
      colorEnd = "#4F7942", # "rgb(88,101,185)"
      colorStartPos = 0,
      colorMiddlePos = 55,
      colorEndPos = 100
    ),
    buttonTextColor = "#4F7942", # rgb(255,255,255)"
    buttonBorderColor = "rgb(200,200,200)",
    buttonBorderRadius = 5,
    buttonBackColorHover = "#4F7942", # "rgb(255,255,255)"
    buttonTextColorHover = "rgb(0,0,0)",
    buttonBorderColorHover = "rgb(200,200,200)",
    textboxBackColor = "rgb(255,255,255)",
    textboxBorderColor = "rgb(200,200,200)",
    textboxBorderRadius = 5,
    textboxBackColorSelect = "rgb(255,255,255)",
    textboxBorderColorSelect = "rgb(200,200,200)",
    tableBackColor = "rgb(255,255,255)",
    tableBorderColor = "rgb(240,240,240)",
    tableBorderTopSize = 1,
    tableBorderRowSize = 1
  )
}

shinyDashboardThemeDIY <- function(
  appFontFamily,
  appFontColor,
  logoBackColor,
  bodyBackColor,
  headerButtonBackColor,
  headerButtonIconColor,
  headerButtonBackColorHover,
  headerButtonIconColorHover,
  headerBackColor,
  headerBoxShadowColor,
  headerBoxShadowSize,
  sidebarBackColor,
  sidebarPadding,
  sidebarShadowRadius,
  sidebarShadowColor,
  sidebarMenuBackColor,
  sidebarMenuPadding,
  sidebarMenuBorderRadius,
  sidebarUserTextColor,
  sidebarSearchBackColor,
  sidebarSearchIconColor,
  sidebarSearchBorderColor,
  sidebarTabTextColor,
  sidebarTabTextSize,
  sidebarTabBorderStyle,
  sidebarTabBorderColor,
  sidebarTabBorderWidth,
  sidebarTabBackColorSelected,
  sidebarTabTextColorSelected,
  sidebarTabRadiusSelected,
  sidebarTabTextColorHover,
  sidebarTabBackColorHover,
  sidebarTabBorderStyleHover,
  sidebarTabBorderColorHover,
  sidebarTabBorderWidthHover,
  sidebarTabRadiusHover,
  boxBackColor,
  boxBorderRadius,
  boxShadowSize,
  boxShadowColor,
  boxTitleSize,
  boxDefaultColor,
  boxPrimaryColor,
  boxSuccessColor,
  boxWarningColor,
  boxDangerColor,
  tabBoxTabColor,
  tabBoxTabTextSize,
  tabBoxTabTextColor,
  tabBoxTabTextColorSelected,
  tabBoxBackColor,
  tabBoxHighlightColor,
  tabBoxBorderRadius,
  buttonBackColor,
  buttonTextColor,
  buttonBorderColor,
  buttonBorderRadius,
  buttonBackColorHover,
  buttonTextColorHover,
  buttonBorderColorHover,
  buttonHeight = "auto",
  buttonPadding = "6px 12px",
  textboxBackColor,
  textboxBorderColor,
  textboxBorderRadius,
  textboxBackColorSelect,
  textboxBorderColorSelect,
  textboxHeight = "auto",
  textboxPadding = "6px 12px",
  tableBackColor,
  tableBorderColor,
  tableBorderTopSize,
  tableBorderRowSize,
  primaryFontColor = "auto",
  successFontColor = "auto",
  warningFontColor = "auto",
  dangerFontColor = "auto",
  infoFontColor = "auto",
  boxInfoColor = "auto"
) {
  cssCode <- paste0(
    '
    /* font */
    body, label, input, button, select, box,
    .h1, .h2, .h3, .h4, .h5, h1, h2, h3, h4, h5 {
      font-family: "',
    appFontFamily,
    '";
      color: ',
    appFontColor,
    ';
    }

    /* font: fix for h6 */
    /* messes up sidebar user section if included above */
    .h6, h6 {
      font-family: "',
    appFontFamily,
    '";
    }

    /* sidebar: logo */
    .skin-blue .main-header .logo {
      background: ',
    logoBackColor,
    ";
    }

    /* sidebar: logo hover */
    .skin-blue .main-header .logo:hover {
      background: ",
    logoBackColor,
    ";
    }

    /* sidebar: collapse button  */
    .skin-blue .main-header .navbar .sidebar-toggle {
      background: ",
    headerButtonBackColor,
    ";
      color:",
    headerButtonIconColor,
    ";
    }

    /* sidebar: collapse button hover */
    .skin-blue .main-header .navbar .sidebar-toggle:hover {
      background: ",
    headerButtonBackColorHover,
    ";
      color:",
    headerButtonIconColorHover,
    ";
    }

    /* header */
    .skin-blue .main-header .navbar {
      background: ",
    headerBackColor,
    ";
      box-shadow: ",
    headerBoxShadowSize,
    " ",
    headerBoxShadowColor,
    ";
    }

    /* sidebar*/
    .skin-blue .main-sidebar {
      background: ",
    sidebarBackColor,
    ";
      box-shadow: ",
    sidebarShadowRadius,
    " ",
    sidebarShadowColor,
    ";
      padding-left: ",
    sidebarPadding,
    "px;
      padding-right: ",
    sidebarPadding,
    "px;
      /* padding-top: 60px; */
    }

    /* sidebar menu */
    .main-sidebar .user-panel, .sidebar-menu, .sidebar-menu>li.header {
      white-space: nowrap;
      background: ",
    sidebarMenuBackColor,
    ";
      padding: ",
    sidebarMenuPadding,
    "px;
      border-radius: ",
    sidebarMenuBorderRadius,
    "px;
    }

    /* fix for user panel */
    .user-panel>.info>p, .skin-blue .user-panel>.info {
      color: ",
    sidebarUserTextColor,
    ";
      font-size: 12px;
      font-weight: normal;
    }
    section.sidebar .user-panel {
      padding: 10px;
    }

    /* sidebar: tabs */
    .skin-blue .main-sidebar .sidebar .sidebar-menu a {
      color: ",
    sidebarTabTextColor,
    ";
      font-size: ",
    sidebarTabTextSize,
    "px;
      border-style: ",
    sidebarTabBorderStyle,
    ";
      border-color: ",
    sidebarTabBorderColor,
    ";
      border-width: ",
    sidebarTabBorderWidth,
    "px;
    }

    /* sidebar: tab selected */
    .skin-blue .main-sidebar .sidebar .sidebar-menu .active > a {
      color: ",
    sidebarTabTextColorSelected,
    ";
      font-size: ",
    sidebarTabTextSize,
    "px;
      border-radius: ",
    sidebarTabRadiusSelected,
    ";
      border-style: ",
    sidebarTabBorderStyleHover,
    ";
      border-color: ",
    sidebarTabBorderColorHover,
    ";
      border-width: ",
    sidebarTabBorderWidthHover,
    "px;
    }
    .skin-blue .sidebar-menu > li:hover > a,
    .skin-blue .sidebar-menu > li.active > a {
      color: ",
    sidebarTabTextColorSelected,
    ";
      background: ",
    sidebarTabBackColorSelected,
    ";
      border-radius: ",
    sidebarTabRadiusHover,
    ";
    }

    /* sidebar: tab hovered */
    .skin-blue .main-sidebar .sidebar .sidebar-menu a:hover {
      background: ",
    sidebarTabBackColorHover,
    ";
      color: ",
    sidebarTabTextColorHover,
    ";
      font-size: ",
    sidebarTabTextSize,
    "px;
      border-style: ",
    sidebarTabBorderStyleHover,
    ";
      border-color: ",
    sidebarTabBorderColorHover,
    ";
      border-width: ",
    sidebarTabBorderWidthHover,
    "px;
      border-radius: ",
    sidebarTabRadiusHover,
    ";
    }

    /* sidebar: subtab */
    .skin-blue .sidebar-menu > li > .treeview-menu {
      margin: 0px;
      background: ",
    sidebarMenuBackColor,
    ";
    }
    .skin-blue .treeview-menu > li > a {
      background: ",
    sidebarMenuBackColor,
    ";
    }
    /* sidebar: subtab selected */
    .skin-blue .treeview-menu > li.active > a,
    .skin-blue .treeview-menu > li > a:hover {
      background: ",
    sidebarTabBackColorSelected,
    ";
    }

    /* sidebar: search text area */
    .skin-blue .sidebar-form input[type=text] {
      background: ",
    sidebarSearchBackColor,
    ";
      color: ",
    appFontColor,
    ";
      border-radius: ",
    textboxBorderRadius,
    "px 0px 0px ",
    textboxBorderRadius,
    "px;
      border-color: ",
    sidebarSearchBorderColor,
    ";
      border-style: solid none solid solid;
    }

    /* sidebar: search button */
    .skin-blue .input-group-btn > .btn {
      background: ",
    sidebarSearchBackColor,
    ";
      color: ",
    sidebarSearchIconColor,
    ";
      border-radius: 0px ",
    textboxBorderRadius,
    "px ",
    textboxBorderRadius,
    "px 0px;
      border-style: solid solid solid none;
      border-color: ",
    sidebarSearchBorderColor,
    ";
    }

    /* sidebar form */
    .skin-blue .sidebar-form {
      border-radius: 0px;
      border: 0px none rgb(255,255,255);
      margin: 10px;
    }

    /* body */
    .content-wrapper, .right-side {
      background: ",
    bodyBackColor,
    ";
    }

    /* box */
    .box {
      background: ",
    boxBackColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
      box-shadow: ",
    boxShadowSize,
    " ",
    boxShadowColor,
    ";
    }

    /* box: title */
    .box-header .box-title {
      font-size: ",
    boxTitleSize,
    "px;
    }

    /* tabbox: title */
    .nav-tabs-custom>.nav-tabs>li.header {
      color: ",
    appFontColor,
    ";
      font-size: ",
    boxTitleSize,
    "px;
    }

    /* tabbox: tab color */
    .nav-tabs-custom, .nav-tabs-custom .nav-tabs li.active:hover a,
    .nav-tabs-custom .nav-tabs li.active a {
      background: ",
    tabBoxTabColor,
    ";
      color: ",
    appFontColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }

    .nav-tabs-custom {
      box-shadow: ",
    boxShadowSize,
    " ",
    boxShadowColor,
    ";
    }

    /* tabbox: active tab bg */
    .nav-tabs-custom>.nav-tabs>li.active {
      border-radius: ",
    tabBoxBorderRadius,
    "px;
      border-top-color: ",
    tabBoxHighlightColor,
    ";
      # box-shadow: ",
    boxShadowSize,
    " ",
    boxShadowColor,
    ";
    }

    /* tabbox: font color */
    .nav-tabs-custom>.nav-tabs>li.active:hover>a, .nav-tabs-custom>.nav-tabs>li.active>a {
      border-bottom-color: ",
    tabBoxTabColor,
    ";
      border-top-color: ",
    tabBoxHighlightColor,
    ";
      border-right-color: ",
    tabBoxHighlightColor,
    ";
      border-left-color: ",
    tabBoxHighlightColor,
    ";
      color: ",
    tabBoxTabTextColorSelected,
    ";
      font-size: ",
    tabBoxTabTextSize,
    "px;
      border-radius: ",
    tabBoxBorderRadius,
    "px;
    }

    /* tabbox: inactive tabs background */
    .nav-tabs-custom>.nav-tabs>li>a {
      color: ",
    tabBoxTabTextColor,
    ";
      font-size: ",
    tabBoxTabTextSize,
    "px;
    }

    /* tabbox: top area back color */
    .nav-tabs-custom, .nav-tabs-custom>.tab-content, .nav-tabs-custom>.nav-tabs {
      border-bottom-color: ",
    tabBoxHighlightColor,
    ";
      background: ",
    tabBoxBackColor,
    ";
    }

    /* tabbox: top area rounded corners */
    .nav-tabs-custom>.nav-tabs {
      margin: 0;
      border-radius: ",
    tabBoxBorderRadius,
    "px;
    }

    /* infobox */
    .info-box {
      background: ",
    boxBackColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
      box-shadow: ",
    boxShadowSize,
    " ",
    boxShadowColor,
    ";
    }

    /* valuebox */
    .small-box {
      border-radius: ",
    boxBorderRadius,
    "px;
      box-shadow: ",
    boxShadowSize,
    " ",
    boxShadowColor,
    ";
    }

    /* valuebox: main font color */
    .small-box h3, .small-box p {
      color: rgb(255,255,255)
    }

    /* box: default color */
    .box.box-solid>.box-header, .box>.box-header {
      color: ",
    appFontColor,
    ";
    }
    .box.box-solid>.box-header {
      border-radius: ",
    boxBorderRadius,
    "px;
    }
    .box.box-solid, .box {
      border-radius: ",
    boxBorderRadius,
    "px;
      border-top-color: ",
    boxDefaultColor,
    ";
    }

    /* box: info color */
    .box.box-solid.box-info>.box-header h3, .box.box-info>.box-header h3 {
      color: ",
    infoFontColor,
    ";
    }
    .box.box-solid.box-info>.box-header {
      background: ",
    boxInfoColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }
    .box.box-solid.box-info, .box.box-info {
      border-color: ",
    boxInfoColor,
    ";
      border-left-color: ",
    boxInfoColor,
    ";
      border-right-color: ",
    boxInfoColor,
    ";
      border-top-color: ",
    boxInfoColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }

    /* box: primary color */
    .box.box-solid.box-primary>.box-header h3, .box.box-primary>.box-header h3 {
      color: ",
    primaryFontColor,
    ";
    }
    .box.box-solid.box-primary>.box-header {
      background: ",
    boxPrimaryColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }
    .box.box-solid.box-primary, .box.box-primary {
      border-color: ",
    boxPrimaryColor,
    ";
      border-left-color: ",
    boxPrimaryColor,
    ";
      border-right-color: ",
    boxPrimaryColor,
    ";
      border-top-color: ",
    boxPrimaryColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }

    /* box: success color */
    .box.box-solid.box-success>.box-header h3, .box.box-success>.box-header h3 {
      color: ",
    successFontColor,
    ";
    }
    .box.box-solid.box-success>.box-header {
      background: ",
    boxSuccessColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }
    .box.box-solid.box-success, .box.box-success {
      border-color: ",
    boxSuccessColor,
    ";
      border-left-color: ",
    boxSuccessColor,
    ";
      border-right-color: ",
    boxSuccessColor,
    ";
      border-top-color: ",
    boxSuccessColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }

    /* box: warning color */
    .box.box-solid.box-warning>.box-header h3, .box.box-warning>.box-header h3 {
      color: ",
    warningFontColor,
    ";
    }
    .box.box-solid.box-warning>.box-header {
      background: ",
    boxWarningColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }
    .box.box-solid.box-warning, .box.box-warning {
      border-color: ",
    boxWarningColor,
    ";
      border-left-color: ",
    boxWarningColor,
    ";
      border-right-color: ",
    boxWarningColor,
    ";
      border-top-color: ",
    boxWarningColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }

    /* box: danger color */
    .box.box-solid.box-danger>.box-header h3, .box.box-danger>.box-header h3 {
      color: ",
    dangerFontColor,
    ";
    }
    .box.box-solid.box-danger>.box-header {
      background: ",
    boxDangerColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }
    .box.box-solid.box-danger, .box.box-danger {
      border-color: ",
    boxDangerColor,
    ";
      border-left-color: ",
    boxDangerColor,
    ";
      border-right-color: ",
    boxDangerColor,
    ";
      border-top-color: ",
    boxDangerColor,
    ";
      border-radius: ",
    boxBorderRadius,
    "px;
    }

    /* button */
    .btn-default {
      background: ",
    buttonBackColor,
    ";
      color: ",
    buttonTextColor,
    ";
      border-color: ",
    buttonBorderColor,
    ";
      border-radius: ",
    buttonBorderRadius,
    "px;
      height: ",
    buttonHeight,
    "px;
      padding: ",
    buttonPadding,
    ";
    }

    /* button: hover */
    .btn-default:hover {
      background: ",
    buttonBackColorHover,
    ";
      color: ",
    buttonTextColorHover,
    ";
      border-color: ",
    buttonBorderColorHover,
    ";
    }

    /* button: focus */
    .btn-default:focus, .action-button:focus {
      background: ",
    buttonBackColor,
    ";
      color: ",
    buttonTextColor,
    ";
      border-color: ",
    buttonBorderColor,
    ";
    }

    /* button: active */
    .btn-default:active, .action-button:active {
      background: ",
    buttonBackColor,
    ";
      color: ",
    buttonTextColor,
    ";
      border-color: ",
    buttonBorderColor,
    ";
    }

    /* button: visited */
    .btn-default:visited {
      background: ",
    buttonBackColor,
    ";
      color: ",
    buttonTextColor,
    ";
      border-color: ",
    buttonBorderColor,
    ";
    }

    /* textbox */
    .form-control, .selectize-input, .selectize-control.single .selectize-input {
      background: ",
    textboxBackColor,
    ";
      color: ",
    appFontColor,
    ";
      border-color: ",
    textboxBorderColor,
    ";
      border-radius: ",
    textboxBorderRadius,
    "px;
      height: ",
    textboxHeight,
    "px;
      min-height: ",
    textboxHeight,
    "px;
      padding: ",
    textboxPadding,
    ";
    }

    /* textbox: selected */
    .form-control:focus, .selectize-input.focus {
      color: ",
    appFontColor,
    ";
      background: ",
    textboxBackColorSelect,
    ";
      border-color: ",
    textboxBorderColorSelect,
    ";
      -webkit-box-shadow: inset 0px 0px 0px, 0px 0px 0px;
      box-shadow: inset 0px 0px 0px, 0px 0px 0px;
    }

    /* multi-row selectize input */
    .selectize-control.multi .selectize-input.has-items {
      height: auto;
    }

    /* verbatim text output */
    .qt pre, .qt code {
      font-family: ",
    appFontFamily,
    " !important;
    }
    pre {
      color: ",
    appFontColor,
    ";
      background-color: ",
    textboxBackColor,
    ";
      border: 1px solid ",
    textboxBorderColor,
    ";
      border-radius: ",
    textboxBorderRadius,
    "px;
    }

    /* drop-down menu */
    .selectize-dropdown, .selectize-dropdown.form-control {
      background: ",
    textboxBackColor,
    ";
      border-radius: 4px;
    }

    /* table */
    .table {
      background: ",
    tableBackColor,
    ";
      border-radius: ",
    textboxBorderRadius,
    "px;
    }

    /* table: row border color*/
    .table>tbody>tr>td, .table>tbody>tr>th, .table>tfoot>tr>td,
    .table>tfoot>tr>th, .table>thead>tr>td, .table>thead>tr>th {
      border-top: ",
    tableBorderRowSize,
    "px solid ",
    tableBorderColor,
    ";
    }

    /* table: top border color*/
    .table>thead>tr>th {
      border-bottom: ",
    tableBorderTopSize,
    "px solid ",
    tableBorderColor,
    ";
    }

    /* table: hover row */
    .table-hover>tbody>tr:hover {
    background-color: ",
    tableBorderColor,
    ";
    }

    /* table: stripe row */
    .table-striped>tbody>tr:nth-of-type(odd) {
      background-color: ",
    tableBorderColor,
    ";
    }

    /* table: body colour */
    table.dataTable tbody tr {
      background-color: ",
    tableBackColor,
    " !important;
    }

    /* table: text and footer border colour */
    table.dataTable {
      color: ",
    appFontColor,
    " !important;
      border: 0px !important;
    }

    /* datatable: selected row */
    table.dataTable tr.selected td, table.dataTable td.selected {
      background-color: ",
    boxSuccessColor,
    " !important;
      color: rgb(0,0,0) !important;
    }

    /* datatable: hover row */
    table.dataTable tr.hover td, table.dataTable td.hover {
      background-color: ",
    tableBorderColor,
    " !important;
    }
    table.dataTable.hover tbody tr:hover, table.dataTable.display tbody tr:hover {
      background-color: ",
    tableBorderColor,
    " !important;
    }
    table.dataTable.row-border tbody th, table.dataTable.row-border tbody td,
    table.dataTable.display tbody th, table.dataTable.display tbody td {
      border-top: 1px solid ",
    tableBorderColor,
    " !important;
    }

    /* datatable: stripe row */
    table.dataTable.stripe tbody tr.odd, table.dataTable.display tbody tr.odd {
      background-color: ",
    tableBorderColor,
    " !important;
    }

    /* datatable: page control */
    .dataTables_wrapper .dataTables_paginate .paginate_button {
      color: ",
    appFontColor,
    " !important;
    }

    /* datatable: table info */
    .dataTables_wrapper .dataTables_paginate .paginate_button.disabled,
    .dataTables_wrapper .dataTables_paginate .paginate_button.disabled:hover,
    .dataTables_wrapper .dataTables_paginate .paginate_button.disabled:active {
      color: ",
    appFontColor,
    " !important;
    }
    .dataTables_wrapper .dataTables_length, .dataTables_wrapper .dataTables_filter,
    .dataTables_wrapper .dataTables_info, .dataTables_wrapper .dataTables_processing,
    .dataTables_wrapper .dataTables_paginate {
      color: ",
    appFontColor,
    " !important;
    }
    .dataTables_wrapper .dataTables_length, .dataTables_wrapper .dataTables_filter,
    .dataTables_wrapper .dataTables_info, .dataTables_wrapper .dataTables_processing,
    .dataTables_wrapper .dataTables_paginate {
      color: ",
    appFontColor,
    " !important;
    }

    /* datatable search box */
    .dataTables_wrapper .dataTables_filter input {
      background-color: ",
    textboxBackColor,
    ";
      border: 1px solid ",
    textboxBorderColor,
    ";
      border-radius: ",
    textboxBorderRadius,
    "px;
    }

    /* notification and progress bar */
    .progress-bar {
      background-color: ",
    boxSuccessColor,
    ";
    }
    .shiny-notification {
      height: 80px;
      font-family: ",
    appFontFamily,
    ";
      font-size: 15px;
      border-radius: 10px;
      margin-left: -450px !important;
    }

    /* horizontal divider line */
    hr {
      border-top: 1px solid rgb(215,215,215);
    }

    /* modal */
    .modal-body {
      background-color: ",
    boxBackColor,
    ";
    }

    .modal-footer {
      background-color: ",
    boxBackColor,
    ";
    }

    .modal-header {
      background-color: ",
    boxBackColor,
    ";
    }

    "
  )

  # removing new line symbols and formatting spacing
  cssCode <- gsub(pattern = "\n", replacement = "", x = cssCode)
  cssCode <- gsub(pattern = "[[:space:]]{2,3}", replacement = "", x = cssCode)

  htmlCode <- htmltools::tags$head(
    htmltools::tags$style(
      htmltools::HTML(
        text = cssCode
      )
    )
  )

  return(htmlCode)
}


cssGradientThreeColors <- function(
  direction = "down",
  colorStart,
  colorMiddle,
  colorEnd,
  colorStartPos = 0,
  colorMiddlePos = 50,
  colorEndPos = 100
) {
  # handling direction types
  if (direction == "down") {
    colorStartSide <- "top"
    colorStartSideCorner <- "left top"
    colorEndSideCorner <- "left bottom"
    colorDirection <- "to bottom"
  } else if (direction == "right") {
    colorStartSide <- "left"
    colorStartSideCorner <- "left top"
    colorEndSideCorner <- "right top"
    colorDirection <- "to right"
  } else {
    stop("The chosen direction isn't supported.")
  }

  # building up css code
  cssCode <- paste0(
    colorStart,
    ";
  background: -moz-linear-gradient(",
    colorStartSide,
    ", ",
    colorStart,
    " ",
    colorStartPos,
    "%, ",
    colorMiddle,
    " ",
    colorMiddlePos,
    "%, ",
    colorEnd,
    " ",
    colorEndPos,
    "%);
  background: -webkit-gradient(",
    colorStartSideCorner,
    ",",
    colorEndSideCorner,
    ",
    color-stop(",
    colorStartPos,
    "%, ",
    colorStart,
    "),
    color-stop(",
    colorMiddlePos,
    "%, ",
    colorMiddle,
    "),
    color-stop(",
    colorEndPos,
    "%, ",
    colorEnd,
    "));
  background: -webkit-linear-gradient(",
    colorStartSide,
    ", ",
    colorStart,
    " ",
    colorStartPos,
    "%, ",
    colorMiddle,
    " ",
    colorMiddlePos,
    "%, ",
    colorEnd,
    " ",
    colorEndPos,
    "%);
  background: -o-linear-gradient(",
    colorStartSide,
    ", ",
    colorStart,
    " ",
    colorStartPos,
    "%, ",
    colorMiddle,
    " ",
    colorMiddlePos,
    "%, ",
    colorEnd,
    " ",
    colorEndPos,
    "%);
  background: -ms-linear-gradient(",
    colorStartSide,
    ", ",
    colorStart,
    " ",
    colorStartPos,
    "%, ",
    colorMiddle,
    " ",
    colorMiddlePos,
    "%, ",
    colorEnd,
    " ",
    colorEndPos,
    "%);
  background: linear-gradient(",
    colorDirection,
    ", ",
    colorStart,
    " ",
    colorStartPos,
    "%, ",
    colorMiddle,
    " ",
    colorMiddlePos,
    "%, ",
    colorEnd,
    " ",
    colorEndPos,
    "%)"
  )

  # removing new line symbols and formatting spacing
  cssCode <- gsub(pattern = "\n", replacement = "", x = cssCode)
  cssCode <- gsub(pattern = "[[:space:]]{2,3}", replacement = "", x = cssCode)

  return(cssCode)
}

settingsUI <- function() {
  ### SETTINGS ----
  tabItem(
    tabName = "settings",
    fluidPage(
      # Header
      fluidRow(
        column(
          12,
          h2(icon("cog"), strong("Settings"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;"),
          h5(
            "Configure global settings for plots, analysis reproducibility, and AI features.",
            align = "center",
            style = "color: #666;"
          ),
          br()
        )
      ),

      # ROW 1: Working Folder + Reproducibility + Language Models
      fluidRow(
        column(
          6,
          wellPanel(
            style = "background-color: #f8f9fa; border: 1px solid #dee2e6;",
            h4(icon("folder-open"), strong("Working Folder")),
            p(
              "All outputs (plots, reports, exported files) will be saved here.",
              style = "color: #666; font-size: 13px; margin-bottom: 15px;"
            ),
            uiOutput("wdFolderDisplay"),
            br(),
            div(
              style = "display: flex; gap: 10px; align-items: center;",
              shinyDirButton(
                "workingfolder",
                label = "Change Folder",
                title = "Select a Working Folder",
                icon = icon("folder-open"),
                style = "color: white; border-radius: 20px; padding: 8px 20px;"
              ),
              actionButton(
                inputId = "cache",
                style = "color:white; border-radius: 20px; padding: 8px 20px;",
                label = "Clean Model Cache",
                icon = icon("database")
              )
            )
          )
        ),
        column(
          6,
          wellPanel(
            style = "background-color: #e8f5e9; border: 1px solid #c8e6c9;",
            h4(icon("random"), strong("Reproducibility Settings")),
            br(),
            fluidRow(
              column(
                9,
                numericInput(
                  "random_seed",
                  "Random Seed",
                  value = 1234,
                  min = 1,
                  max = 99999,
                  step = 1
                )
              ),
              column(
                3,
                br(),
                actionButton(
                  "randomize_seed",
                  "Randomize",
                  icon = icon("random"),
                  style = "color:white; margin-top: 5px;"
                )
              )
            ),
            div(
              style = "background-color: #d1f2d5; padding: 10px; border-radius: 5px; border-left: 4px solid #4caf50;",
              icon("info-circle", style = "color: #2e7d32;"),
              span(
                " The seed ensures reproducible results for topic models, clustering, and all randomized analyses.",
                style = "color: #2e7d32;"
              )
            )
          )
        )
      ),

      # ROW 2: Graph Export Settings + TALL AI
      fluidRow(
        column(
          6,
          wellPanel(
            style = "background-color: #e3f2fd; border: 1px solid #bbdefb;",
            h4(icon("image"), strong("Graph Export Settings")),
            br(),
            div(
              style = "margin-bottom: 25px;",
              tags$label(
                "Export Resolution (DPI)",
                style = "font-weight: 600; color: #2E86AB; margin-bottom: 8px; display: block;"
              ),
              uiOutput("dpi_slider"),
              helpText(
                "Resolution for exported images (download button). Higher DPI = better quality but larger files.",
                style = "margin-top: 5px; color: #666; font-size: 12px;"
              )
            ),
            div(
              style = "margin-bottom: 25px;",
              tags$label(
                "Report Resolution (DPI)",
                style = "font-weight: 600; color: #2E86AB; margin-bottom: 8px; display: block;"
              ),
              uiOutput("report_dpi_slider"),
              helpText(
                "Resolution for images added to the Excel report. Lower DPI keeps the report file size small.",
                style = "margin-top: 5px; color: #666; font-size: 12px;"
              )
            ),
            div(
              style = "margin-bottom: 25px;",
              tags$label(
                "Plot Height (inches)",
                style = "font-weight: 600; color: #2E86AB; margin-bottom: 8px; display: block;"
              ),
              uiOutput("h_slider"),
              helpText(
                "Adjust the height of exported plots. Width is automatically calculated to maintain aspect ratio.",
                style = "margin-top: 5px; color: #666; font-size: 12px;"
              )
            ),
            div(
              style = "margin-bottom: 10px;",
              tags$label(
                "Plot Aspect Ratio",
                style = "font-weight: 600; color: #2E86AB; margin-bottom: 8px; display: block;"
              ),
              uiOutput("aspect_radio"),
              helpText(
                "Publication (3:2) is ideal for papers and books. Wide (2:1) is better for presentations and dashboards.",
                style = "margin-top: 5px; color: #666; font-size: 12px;"
              )
            )
          )
        ),
        column(
          6,
          wellPanel(
            style = "background-color: #fff3e0; border: 1px solid #ffe0b2;",
            h4(icon("robot"), strong("Tall AI - Google Gemini Integration")),
            p(
              "Enable AI-powered features by providing your Google Gemini API Key. Generate one at ",
              tags$a(
                "AI Studio",
                href = "https://aistudio.google.com/app/apikey",
                target = "_blank"
              ),
              ".",
              style = "color: #666; font-size: 13px; margin-bottom: 15px;"
            ),

            h5(strong("API Key")),
            passwordInput(
              "api_key",
              label = NULL,
              value = "",
              width = "100%"
            ),
            uiOutput("apiStatus"),
            br(),
            fluidRow(
              column(
                6,
                actionButton(
                  "set_key",
                  "Set API Key",
                  icon = icon("check"),
                  style = "color:white;",
                  width = "100%"
                )
              ),
              column(
                6,
                actionButton(
                  "remove_key",
                  "Remove Key",
                  icon = icon("trash"),
                  style = "color:white;",
                  width = "100%"
                )
              )
            ),
            hr(),
            fluidRow(
              column(
                6,
                h5(strong("Model")),
                uiOutput("geminiModelChoice")
              ),
              column(
                6,
                h5(strong("Output Size")),
                uiOutput("geminiOutputSize")
              )
            )
          )
        )
      )
    )
  )
}


settingsServer <- function(input, output, session, values, statsValues) {
  ## UTILITY ----

  observeEvent(input$d_abstractiveView, {
    showModal(showDocumentModal(
      session,
      input$Abst_document_selection,
      input
    ))
  })

  observeEvent(input$d_summarizationView, {
    showModal(showDocumentModal(
      session,
      input$document_selection,
      input
    ))
  })

  showDocumentModal <- function(session, docID, input) {
    ns <- session$ns
    if (
      input$sidebarmenu %in%
        c("import_tx", "split_tx", "extInfo", "randomText")
    ) {
      text <- create_document_box(
        values$txt %>% filter(doc_id == docID) %>% pull(text),
        docID,
        summarization_type = "original_text"
      )
    } else {
      text <- create_document_box(
        values$dfTag %>%
          filter(doc_id == !!docID) %>%
          select(doc_id, paragraph_id, sentence_id, sentence) %>%
          distinct() %>%
          group_by(doc_id, paragraph_id) %>%
          summarise(Paragraph = paste(sentence, collapse = "\n ")) %>%
          ungroup(),
        docID,
        summarization_type = "abstractive"
      )
    }

    modalDialog(
      tags$style(HTML(
        "
      .modal-dialog {
        width: 70vw !important;
        max-width: 70vw !important;
      }
      .modal-content {
        height: 85vh !important;
      }
      .modal-body {
        overflow-y: auto !important;
        max-height: calc(85vh - 120px) !important;
      }
    "
      )),
      div(
        h3(strong("Document corpus")),
        br(),
        HTML(text)
      ),
      size = "l",
      easyClose = TRUE,
      footer = modalButton("Close")
    )
  }

  ## table click button ----
  observeEvent(input$button_id, {
    if (input$button_id != "null") {
      showModal(showDocumentModal(session, input$button_id, input))
    }
  })

  observeEvent(input$button_id_del, {
    if (
      input$sidebarmenu %in% c("import_tx", "split_tx", "extInfo", "randomText")
    ) {
      values$txt <- values$txt %>%
        mutate(
          doc_selected = ifelse(
            doc_id == input$button_id_del,
            FALSE,
            doc_selected
          )
        )
    }
    if (!is.null(values$dfTag)) {
      values$dfTag <- values$dfTag %>%
        mutate(docSelected = ifelse(doc_id == input$button_id_del, FALSE, docSelected))
    }
  })

  ### SETTINGS ----

  ## Graph Settings ----
  saveGraphSettings <- function() {
    home <- homeFolder()
    path_tall <- file.path(home, "tall")
    if (!file.exists(path_tall)) {
      dir.create(path_tall)
    }
    writeLines(
      c(
        as.character(values$dpi),
        as.character(values$report_dpi),
        as.character(values$h),
        as.character(values$aspect)
      ),
      file.path(path_tall, ".tall_graph_settings.txt")
    )
  }

  observeEvent(input$dpi, {
    values$dpi <- as.numeric(input$dpi)
    saveGraphSettings()
  }, ignoreInit = TRUE)

  observeEvent(input$report_dpi, {
    values$report_dpi <- as.numeric(input$report_dpi)
    saveGraphSettings()
  }, ignoreInit = TRUE)

  observeEvent(input$h, {
    values$h <- as.numeric(input$h)
    saveGraphSettings()
  }, ignoreInit = TRUE)

  observeEvent(input$aspect, {
    values$aspect <- as.numeric(input$aspect)
    saveGraphSettings()
  }, ignoreInit = TRUE)

  observeEvent(input$cache, {
    deleteCache()
  })

  ## Choose Working folder in Setting Menu
  roots <- c(home = homeFolder())

  observe({
    shinyDirChoose(input, "workingfolder", roots = roots, filetypes = c(""))
  })

  observeEvent(
    eventExpr = input$workingfolder,
    handlerExpr = {
      wdTall <- parseDirPath(roots = roots, input$workingfolder)

      if (length(wdTall) == 0 || is.null(wdTall)) {
        if (is.null(wdFolder())) {
          values$menu <- -2
        }
      } else {
        # setting up the main directory
        home <- homeFolder()
        path_tall <- file.path(home, "tall")
        # check if sub directory exists
        if (!file.exists(path_tall)) {
          dir.create(path_tall)
        }
        writeLines(wdTall, con = paste0(path_tall, "/tallWD.tall"))
        if (values$menu == -2) {
          values$menu <- -1
        }
        values$wdTall <- wdTall
      }
    },
    ignoreNULL = TRUE
  )

  output$wdFolder <- renderText({
    values$wdTall
  })

  output$wdFolderDisplay <- renderUI({
    folder <- values$wdTall
    if (is.null(folder) || length(folder) == 0 || folder == "") {
      div(
        style = "padding: 12px; background-color: #fff3cd; border-left: 4px solid #ffc107; border-radius: 4px;",
        icon("exclamation-triangle", style = "color: #856404;"),
        span(
          " No working folder selected. Please select a folder to save your outputs.",
          style = "color: #856404; font-size: 13px;"
        )
      )
    } else {
      div(
        style = "padding: 12px; background-color: #e8f5e9; border-left: 4px solid #4caf50; border-radius: 4px;",
        div(
          style = "display: flex; align-items: center; gap: 8px;",
          icon("check-circle", style = "color: #4caf50; font-size: 18px;"),
          div(
            span("Current folder:", style = "color: #666; font-size: 12px;"),
            br(),
            strong(folder, style = "font-size: 14px; color: #2e7d32; word-break: break-all;")
          )
        )
      )
    }
  })

  # Random Seed Management
  observeEvent(input$random_seed, {
    if (!is.null(input$random_seed)) {
      set.seed(input$random_seed)
      values$random_seed <- input$random_seed
    }
  })

  observeEvent(input$randomize_seed, {
    new_seed <- sample(1:99999, 1)
    updateNumericInput(session, "random_seed", value = new_seed)
  })

  output$apiStatus <- renderUI({
    if (values$geminiAPI) {
      last <- showGeminiAPI()
      output$status <- renderText(paste0("✅ API key has been set: ", last))
    }
  })

  output$geminiOutputSize <- renderUI({
    list(
      selectInput(
        inputId = "gemini_output_size",
        label = "Max Output (in tokens)",
        selected = ifelse(
          is.null(values$gemini_output_size),
          "medium",
          values$gemini_output_size
        ),
        choices = c("Medium" = "medium", "Large" = "large")
      ),
      conditionalPanel(
        condition = "input.gemini_output_size == 'medium'",
        helpText(strong("Free Tier Output:")),
        helpText(em("Medium -> 16384 Tokens"))
      ),
      conditionalPanel(
        condition = "input.gemini_output_size == 'large'",
        helpText(strong("Free Tier Output:")),
        helpText(em("Large -> 32768 Tokens"))
      )
    )
  })

  output$geminiModelChoice <- renderUI({
    list(
      selectInput(
        inputId = "gemini_api_model",
        label = "Select the Gemini Model",
        choices = c(
          "Gemini 2.5 Flash" = "2.5-flash",
          "Gemini 2.5 Flash Lite" = "2.5-flash-lite",
          "Gemini 3.0 Flash" = "3-flash-preview",
          "Gemini 3.1 Pro" = "3.1-pro-preview",
          "Gemini Pro Latest" = "pro-latest"
        ),
        selected = ifelse(
          is.null(values$gemini_api_model),
          "2.5-flash-lite",
          values$gemini_api_model
        )
      ),
      conditionalPanel(
        condition = "input.gemini_api_model == '2.5-flash'",
        helpText(strong("Free Tier Rate Limits:")),
        helpText(em(
          "Request per Minutes: 10",
          tags$br(),
          "Requests per Day: 500",
          tags$br(),
          "Latency time: High"
        ))
      ),
      conditionalPanel(
        condition = "input.gemini_api_model == '2.5-flash-lite'",
        helpText(strong("Free Tier Rate Limits:")),
        helpText(em(
          "Request per Minutes: 15",
          tags$br(),
          "Requests per Day: 500",
          tags$br(),
          "Latency time: Low"
        ))
      ),
      conditionalPanel(
        condition = "input.gemini_api_model == '3-flash-preview'",
        helpText(strong("Free Tier Rate Limits:")),
        helpText(em(
          "Request per Minutes: 10",
          tags$br(),
          "Requests per Day: 500",
          tags$br(),
          "Latency time: Medium"
        ))
      ),
      conditionalPanel(
        condition = "input.gemini_api_model == '3.1-pro-preview' || input.gemini_api_model == 'pro-latest'",
        helpText(strong(style = "color: #d9534f;", "Paid API Key Required")),
        helpText(em(
          "Pro models require a paid (non-free tier) API key.",
          tags$br(),
          "Free tier API keys will not work with these models."
        ))
      )
    )
  })

  observeEvent(input$gemini_api_model, {
    if (!is.null(input$gemini_api_model)) {
      saveGeminiModel(
        model = c(input$gemini_api_model, input$gemini_output_model),
        file = paste0(homeFolder(), "/.tall_gemini_model.txt", collapse = "")
      )
      values$gemini_api_model <- input$gemini_api_model
      values$gemini_output_size <- input$gemini_output_size

      # Alert for Pro models requiring paid API key
      if (input$gemini_api_model %in% c("3.1-pro-preview", "pro-latest")) {
        showModal(modalDialog(
          title = "Paid API Key Required",
          tags$p("The selected Pro model requires a ", tags$strong("paid (non-free tier)"), " Google AI API key."),
          tags$p("Free tier API keys do not support Pro models. If you are using a free tier key, the API calls will fail."),
          tags$p("You can upgrade your API key at: ", tags$a(href = "https://aistudio.google.com/apikey", target = "_blank", "Google AI Studio")),
          easyClose = TRUE,
          footer = modalButton("OK")
        ))
      }
    }
  })

  observeEvent(input$gemini_output_size, {
    if (!is.null(input$gemini_output_size)) {
      saveGeminiModel(
        model = c(input$gemini_api_model, input$gemini_output_size),
        file = paste0(homeFolder(), "/.tall_gemini_model.txt", collapse = "")
      )
      values$gemini_api_model <- input$gemini_api_model
      values$gemini_output_size <- input$gemini_output_size
    }
  })

  observeEvent(input$set_key, {
    key <- input$api_key

    # Quick sync validations
    if (is.null(key) || !is.character(key) || nchar(key) == 0) {
      output$apiStatus <- renderUI({
        output$status <- renderText("\u274c API key must be a non-empty string.")
      })
      values$geminiAPI <- FALSE
      return()
    }
    if (nchar(key) < 10) {
      output$apiStatus <- renderUI({
        output$status <- renderText("\u274c API key seems too short. Please verify your key.")
      })
      values$geminiAPI <- FALSE
      return()
    }

    output$apiStatus <- renderUI({
      output$status <- renderText("\u231b Validating API key...")
    })

    # Async: validate key via API call in background
    promises::future_promise({
      apiCheck <- gemini_ai(
        image = NULL, prompt = "Hello", model = "2.5-flash",
        type = "png", retry_503 = 5, api_key = key
      )
      contains_error <- grepl("HTTP\\s*[1-5][0-9]{2}", apiCheck)
      list(valid = !contains_error, key = key)
    }, seed = TRUE) %...>%
      (function(result) {
        if (!result$valid) {
          output$apiStatus <- renderUI({
            output$status <- renderText(
              "\u274c API key seems be not valid! Please, check it or your connection."
            )
          })
          values$geminiAPI <- FALSE
        } else {
          Sys.setenv(GEMINI_API_KEY = result$key)
          last4 <- substr(
            result$key,
            max(1, nchar(result$key) - 3),
            nchar(result$key)
          )
          masked <- paste0(
            paste0(rep("*", nchar(result$key) - 4), collapse = ""),
            last4
          )
          output$apiStatus <- renderUI({
            output$status <- renderText(
              paste0("\u2705 API key has been set: ", masked)
            )
          })
          values$geminiAPI <- TRUE
          home <- homeFolder()
          path_gemini_key <- paste0(
            home, "/tall/.tall_gemini_key.txt", collapse = ""
          )
          writeLines(result$key, path_gemini_key)
        }
      }) %...!%
      (function(err) {
        output$apiStatus <- renderUI({
          output$status <- renderText(
            paste("Error validating key:", conditionMessage(err))
          )
        })
        values$geminiAPI <- FALSE
      })

    NULL
  })

  observeEvent(input$remove_key, {
    if (values$geminiAPI) {
      home <- homeFolder()
      path_gemini_key <- paste0(
        home,
        "/tall/.tall_gemini_key.txt",
        collapse = ""
      )
      file.remove(path_gemini_key)
      values$geminiAPI <- FALSE
      output$apiStatus <- renderUI({
        output$status <- renderText(paste0("❌ API key has been removed"))
      })
    }
  })

  ## SOLUTION FOR DT BUG ----
  observeEvent(
    input$sidebarmenu,
    {
      # Lista completa di tutti gli output DT per ogni menu
      dt_outputs <- list(
        # ===== DATA MANAGEMENT =====
        import_tx = c("dataImported"),
        split_tx = c("splitTextData"),
        randomText = c("randomTextData"),
        extInfo = c("extInfoData"),

        # ===== PRE-PROCESSING =====
        tokPos = c("tokPosTagData"),
        posSpecial = c("posSpecialTags", "posSpecialData", "specialEntityFreq"),
        custTermList = c("customPosTagData", "customListData"),
        multiwordCreat = c("multiwordList", "multiwordData"),
        multiwordByList = c("multiwordList2", "multiwordData2"),
        posTagSelect = c("posTagSelectData"),

        # ===== DATA FILTERING & GROUPING =====
        filter_text = c("filterData"),
        defineGroups = c("defineGroupsData", "groupData"),

        # ===== OVERVIEW & VOCABULARY =====
        overview = c("overviewData", "dictionaryData", "tfidfData"),

        # ===== WORDS ANALYSIS =====
        wordCont = c(
          "wFreqTable",
          "wordInContext",
          "docInContext",
          "docInContextHigh"
        ),
        keyness = c("keyness_table"),
        w_pos = c("posTable"),
        w_clustering = c("w_clusteringTable", "wordInContextDend"),
        w_reinclustering = c(
          "w_ReinSummaryTable",
          "w_ReinClusteringTableTerms",
          "w_ReinClusteringTableSegments",
          "wordInContextRein"
        ),
        ca = c(
          "caSingularValueTable",
          "caCoordTable",
          "caContribTable",
          "caCosineTable"
        ),

        # ===== NETWORKS =====
        w_networkCooc = c(
          "w_networkCoocNodesTable",
          "w_networkCoocEdgesTable",
          "wordInContextNet"
        ),
        w_networkTM = c("w_networkTMClusterTable", "w_networkTMWordTable"),
        w_word2vec = c("w_word2vecTable"),
        # ===== DOCUMENTS ANALYSIS =====
        d_tm_select = c("d_tm_selectTable"),
        d_tm_estim = c("d_tm_estimBpTable", "d_tm_estimTpTable"),
        d_syntactic = c("d_syntacticTable"),
        d_svo = c("d_svoTable"),
        d_polDet = c("d_polDetTable"),
        d_summarization = c("RelSentData")
      )

      # Ottieni il menu corrente
      current_menu <- input$sidebarmenu

      # Ottieni gli output DT del menu corrente
      current_outputs <- dt_outputs[[current_menu]]
      if (is.null(current_outputs)) {
        current_outputs <- character(0)
      }

      # Ottieni tutti gli output DT
      all_outputs <- unique(unlist(dt_outputs, use.names = FALSE))

      # Identifica gli output da sospendere (tutti tranne quelli del menu corrente)
      outputs_to_suspend <- setdiff(all_outputs, current_outputs)

      # Sospendi il rendering degli output che non appartengono al menu corrente
      # Questo libera risorse senza distruggere gli output
      lapply(outputs_to_suspend, function(x) {
        tryCatch(
          {
            outputOptions(output, x, suspendWhenHidden = TRUE)
          },
          error = function(e) {
            # Ignora errori per output che potrebbero non essere ancora stati creati
          }
        )
      })

      # Riattiva gli output del menu corrente
      lapply(current_outputs, function(x) {
        tryCatch(
          {
            outputOptions(output, x, suspendWhenHidden = FALSE)
          },
          error = function(e) {
            # Ignora errori per output che potrebbero non essere ancora stati creati
          }
        )
      })
    },
    priority = 100
  )
}

documentsUI <- function() {
  ### DOCUMENTS ----

  ### Topic Modeling ----
  ### K choice ----

  tm_k <- tabItem(
    tabName = "d_tm_select",
    fluidPage(
      fluidRow(
        column(
          9,
          h2(icon("chart-line"), strong("Topic Modeling: Optimal selection of topic number"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "d_tm_selectApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "d_tm_selectExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "d_tm_selectReport"
                )
              )
            )
          )
        )
      ),
      fluidRow(
        column(
          9,
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Tuning Plot", icon = icon("chart-line"),
              br(),
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "d_tm_selectPlot",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Multi-Metric Comparison", icon = icon("chart-column"),
              br(),
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "d_tm_multiMetricPlot",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "K Recommendation", icon = icon("lightbulb"),
              br(),
              shinycssloaders::withSpinner(
                uiOutput("d_tm_kRecommendationUI"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Table", icon = icon("table"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("d_tm_selectTable"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & References", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(infoTexts$tmkchoice)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          3,
          div(
            box(
              width = 12,
              div(
                h3(strong(em("Find optimal K"))),
                style = "margin-top:-57px"
              ),
              tags$hr(),
              style = "text-align: left; text-color: #989898",
              selectInput(
                inputId = "tmMethodK",
                label = "Model",
                choices = c(
                  "LDA" = "LDA",
                  "CTM (Correlated)" = "CTM",
                  "STM (Structural)" = "STM"
                ),
                selected = "LDA"
              ),
              selectInput(
                inputId = "groupTm",
                label = "Topics",
                choices = c(
                  "Groups" = "Groups",
                  "Docs" = "doc_id",
                  "Sentences" = "sentence_id"
                ),
                selected = "doc_id"
              ),
              selectInput(
                "metric",
                "Metric for model tuning",
                choices = c(
                  "CaoJuan-2009" = "CaoJuan2009",
                  "Deveaud-2014" = "Deveaud2014",
                  "Arun-2010" = "Arun2010",
                  "Perplexity" = "Perplexity"
                ),
                selected = "CaoJuan2009"
              ),
              fluidRow(
                column(6),
                column(
                  6,
                  numericInput(
                    "nTm",
                    label = "N. of terms",
                    value = 100,
                    min = 1,
                    step = 1
                  )
                )
              ),
              selectInput(
                "top_by",
                "Terms selection by:",
                choices = c(
                  "Frequency" = "freq",
                  "TF-IDF" = "tfidf"
                ),
                selected = "freq"
              ),
              fluidRow(
                column(
                  4,
                  numericInput(
                    "minK",
                    label = "K min",
                    value = 2,
                    min = 2,
                    step = 1
                  )
                ),
                column(
                  4,
                  numericInput(
                    "maxK",
                    label = "K max",
                    value = 20,
                    min = 3,
                    step = 1
                  )
                ),
                column(
                  4,
                  numericInput(
                    "Kby",
                    label = "K by:",
                    value = 1,
                    min = 1,
                    step = 1
                  )
                )
              )
            ),
            style = "margin-top:40px"
          )
        )
      )
    )
  )

  ### Model estimation ----

  tm_analysis <- tabItem(
    tabName = "d_tm_estim",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("diagram-project"), strong("Topic Modeling: Model estimation"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "d_tm_estimApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "d_tm_estimExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "d_tm_estimReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                selectInput(
                  inputId = "tmMethod",
                  label = "Model",
                  choices = c(
                    "LDA" = "LDA",
                    "CTM (Correlated)" = "CTM",
                    "STM (Structural)" = "STM"
                  ),
                  selected = "LDA"
                ),
                conditionalPanel(
                  "input.tmMethod == 'STM'",
                  uiOutput("stmPrevalenceSelect"),
                  helpText(
                    "Prevalence covariates affect topic proportions across documents.",
                    style = "font-size: 11px; color: #888; margin-top: -5px;"
                  )
                ),
                uiOutput("tmKSelectionUI"),
                selectInput(
                  inputId = "groupTmEstim",
                  label = "Topics",
                  choices = c(
                    "Groups" = "Groups",
                    "Docs" = "doc_id",
                    "Sentences" = "sentence_id"
                  ),
                  selected = "doc_id"
                )
              ),

              # Terms Selection Section
              tags$details(
                class = "filter-section",
                tags$summary(
                  div(
                    class = "filter-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("filter"),
                      " Terms Selection"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  numericInput(
                    "nTmEstim",
                    label = "N. of terms",
                    value = 100,
                    min = 1,
                    step = 1
                  ),
                  selectInput(
                    "top_byEstim",
                    "Terms selection by:",
                    choices = c(
                      "Frequency" = "freq",
                      "TF-IDF" = "tfidf"
                    ),
                    selected = "freq"
                  )
                )
              ),

              # Display Options Section
              tags$details(
                class = "params-section",
                tags$summary(
                  div(
                    class = "params-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("eye"),
                      " Display Options"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  numericInput(
                    "nTopicPlot",
                    label = "Word/Docs per Topic",
                    value = 10,
                    min = 2,
                    step = 1
                  )
                )
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "320px"
            )
          ),
          style = style_opt
        )
      ),
      tabsetPanel(
        type = "tabs",
        tabPanel(
          title = "Topic by Words Plot", icon = icon("chart-column"),
          br(),
          fluidRow(
            column(
              4,
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "d_tm_estimTPlot1",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            column(
              4,
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "d_tm_estimTPlot2",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            column(
              4,
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "d_tm_estimTPlot3",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            )
          ),
          uiOutput("TMplotArrowsUI")
        ),
        tabPanel(
          title = "Topic by Docs Plot", icon = icon("chart-column"),
          br(),
          fluidRow(
            column(
              4,
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "d_tm_DocPlot1",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            column(
              4,
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "d_tm_DocPlot2",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            column(
              4,
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "d_tm_DocPlot3",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            )
          ),
          uiOutput("TMdocArrowsUI")
        ),
        tabPanel(
          title = "Beta Probability", icon = icon("table"),
          br(),
          shinycssloaders::withSpinner(
            DT::DTOutput("d_tm_estimBpTable"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "Theta Probability", icon = icon("table"),
          br(),
          shinycssloaders::withSpinner(
            DT::DTOutput("d_tm_estimTpTable"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "Topic Correlation", icon = icon("diagram-project"),
          br(),
          shinycssloaders::withSpinner(
            plotlyOutput(
              outputId = "d_tm_networkPlot",
              height = "75vh",
              width = "98.9%"
            ),
            # visNetworkOutput("d_tm_networkPlot", width="auto", height = "75vh"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "Model Diagnostics", icon = icon("stethoscope"),
          br(),
          shinycssloaders::withSpinner(
            uiOutput("d_tm_diagnosticsUI"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "Covariate Effects", icon = icon("chart-line"),
          br(),
          shinycssloaders::withSpinner(
            uiOutput("d_tm_covariateUI"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "TALL AI", icon = icon("robot"),
          br(),
          fluidPage(
            fluidRow(
              column(
                12,
                br(),
                shinycssloaders::withSpinner(
                  htmlOutput("d_tm_GeminiUI"),
                  caption = HTML("<br><strong>Thinking...</strong>"),
                  image = "ai_small2.gif",
                  color = "#4F7942"
                )
              )
            )
          )
        ),
        tabPanel(
          title = "Info & References", icon = icon("circle-info"),
          br(),
          fluidPage(
            fluidRow(
              column(1),
              column(
                10,
                br(),
                HTML(infoTexts$tmmodelestimation)
              ),
              column(1)
            )
          )
        )
      )
    )
  )

  ### Syntactic Complexity ----

  syntactic_complexity <- tabItem(
    tabName = "d_syntactic",
    fluidPage(
      fluidRow(
        column(
          9,
          h2(icon("code-branch"), strong("Syntactic Complexity"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(inputId = "d_syntacticApply")
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(inputId = "d_syntacticExport")
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(inputId = "d_syntacticReport")
              )
            )
          )
        )
      ),
      tabsetPanel(
        type = "tabs",
        tabPanel(
          title = "Document Metrics", icon = icon("table"),
          br(),
          shinycssloaders::withSpinner(
            DT::DTOutput("d_syntacticTable"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "Corpus Summary", icon = icon("list"),
          br(),
          shinycssloaders::withSpinner(
            uiOutput("d_syntacticSummaryUI"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "Distributions", icon = icon("chart-area"),
          br(),
          shinycssloaders::withSpinner(
            plotlyOutput(
              outputId = "d_syntacticDistPlot",
              height = "80vh",
              width = "98.9%"
            ),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "TALL AI", icon = icon("robot"),
          br(),
          fluidPage(
            fluidRow(
              column(
                12,
                br(),
                shinycssloaders::withSpinner(
                  htmlOutput("d_syntactic_GeminiUI"),
                  caption = HTML("<br><strong>Thinking...</strong>"),
                  image = "ai_small2.gif",
                  color = "#4F7942"
                )
              )
            )
          )
        ),
        tabPanel(
          title = "Info & References", icon = icon("circle-info"),
          br(),
          fluidPage(
            fluidRow(
              column(1),
              column(
                10,
                br(),
                HTML(infoTexts$syntacticcomplexity)
              ),
              column(1)
            )
          )
        )
      )
    )
  )

  ### SVO Triplets ----

  svo_analysis <- tabItem(
    tabName = "d_svo",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("language"), strong("SVO Triplet Extraction"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(inputId = "d_svoApply")
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(inputId = "d_svoExport")
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(inputId = "d_svoReport")
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Configuration"
                ),
                numericInput(
                  "svoFreqMin",
                  label = "Min. Frequency",
                  value = 2,
                  min = 1,
                  step = 1
                ),
                numericInput(
                  "svoTopN",
                  label = "Top N Triplets",
                  value = 50,
                  min = 10,
                  step = 10
                )
              ),
              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "320px"
            )
          ),
          style = style_opt
        )
      ),
      tabsetPanel(
        type = "tabs",
        tabPanel(
          title = "SVO Table", icon = icon("table"),
          br(),
          shinycssloaders::withSpinner(
            DT::DTOutput("d_svoTable"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "SVO Network", icon = icon("diagram-project"),
          br(),
          shinycssloaders::withSpinner(
            plotlyOutput(
              outputId = "d_svoNetwork",
              height = "80vh",
              width = "98.9%"
            ),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "Verb Frequency", icon = icon("chart-bar"),
          br(),
          shinycssloaders::withSpinner(
            plotlyOutput(
              outputId = "d_svoVerbPlot",
              height = "75vh",
              width = "98.9%"
            ),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "TALL AI", icon = icon("robot"),
          br(),
          fluidPage(
            fluidRow(
              column(
                12,
                br(),
                shinycssloaders::withSpinner(
                  htmlOutput("d_svo_GeminiUI"),
                  caption = HTML("<br><strong>Thinking...</strong>"),
                  image = "ai_small2.gif",
                  color = "#4F7942"
                )
              )
            )
          )
        ),
        tabPanel(
          title = "Info & References", icon = icon("circle-info"),
          br(),
          fluidPage(
            fluidRow(
              column(1),
              column(
                10,
                br(),
                HTML(infoTexts$svo)
              ),
              column(1)
            )
          )
        )
      )
    )
  )

  ### Polarity detection ----

  polarity <- tabItem(
    tabName = "d_polDet",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("face-smile"), strong("Polarity Detection"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "d_polDetApply"
                )
              )
            )
          )
        ),
        div(
          # style=style_bttn,
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "d_polDetExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "d_polDetReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                selectInput(
                  inputId = "groupPolarity",
                  label = "Polarity of",
                  choices = c(
                    "Groups" = "Groups",
                    "Docs" = "doc_id"
                  ),
                  selected = "doc_id"
                ),
                uiOutput("lexiconD_polarity")
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "300px"
            )
          ),
          style = style_opt
        )
      ),
      fluidRow(
        tabsetPanel(
          type = "tabs",
          tabPanel(
            title = "Document Polarity Distribution", icon = icon("chart-area"),
            br(),
            fluidRow(
              column(
                6,
                shinycssloaders::withSpinner(
                  plotlyOutput(
                    outputId = "d_polPiePlot",
                    height = "75vh",
                    width = "98.9%"
                  ),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              ),
              column(
                6,
                shinycssloaders::withSpinner(
                  plotlyOutput(
                    outputId = "d_polDensPlot",
                    height = "37vh",
                    width = "98.9%"
                  ),
                  color = getOption("spinner.color", default = "#4F7942")
                ), # ),
                shinycssloaders::withSpinner(
                  plotlyOutput(
                    outputId = "d_polBoxPlot",
                    height = "37vh",
                    width = "98.9%"
                  ),
                  color = getOption("spinner.color", default = "#4F7942")
                ) # )
              )
            )
          ),
          tabPanel(
            title = "Top Words", icon = icon("chart-bar"),
            br(),
            fluidRow(
              column(
                6,
                align = "center",
                h4("Top Positive Words by Document Polarity Distribution"),
                shinycssloaders::withSpinner(
                  plotlyOutput(
                    outputId = "d_polDetPlotPos",
                    height = "75vh",
                    width = "98.9%"
                  ),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              ),
              column(
                6,
                align = "center",
                h4("Top Negative Words by Document Polarity Distribution"),
                shinycssloaders::withSpinner(
                  plotlyOutput(
                    outputId = "d_polDetPlotNeg",
                    height = "75vh",
                    width = "98.9%"
                  ),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              )
            )
          ),
          tabPanel(
            title = "Table", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("d_polDetTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "TALL AI", icon = icon("robot"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  br(),
                  shinycssloaders::withSpinner(
                    htmlOutput("d_polDet_GeminiUI"),
                    caption = HTML("<br><strong>Thinking...</strong>"),
                    image = "ai_small2.gif",
                    color = "#4F7942"
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "Info & References", icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  br(),
                  HTML(infoTexts$polaritydetection)
                ),
                column(1)
              )
            )
          )
        )
      )
    )
  )

  ### Emotion Detection ----

  emotion <- tabItem(
    tabName = "d_emo",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("heart-pulse"), strong("Emotion Detection"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "d_emoApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "d_emoExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "d_emoReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                selectInput(
                  inputId = "groupEmotion",
                  label = "Emotion Detection of",
                  choices = c(
                    "Groups" = "Groups",
                    "Docs" = "doc_id"
                  ),
                  selected = "doc_id"
                )
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "300px"
            )
          ),
          style = style_opt
        )
      ),
      fluidRow(
        tabsetPanel(
          type = "tabs",
          tabPanel(
            title = "Emotion Distribution", icon = icon("chart-area"),
            br(),
            fluidRow(
              column(
                6,
                shinycssloaders::withSpinner(
                  plotlyOutput(
                    outputId = "d_emoBarPlot",
                    height = "75vh",
                    width = "98.9%"
                  ),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              ),
              column(
                6,
                shinycssloaders::withSpinner(
                  plotlyOutput(
                    outputId = "d_emoRadarPlot",
                    height = "75vh",
                    width = "98.9%"
                  ),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              )
            )
          ),
          tabPanel(
            title = "Top Words by Emotion", icon = icon("chart-bar"),
            br(),
            fluidRow(
              column(
                4,
                selectInput(
                  inputId = "d_emoSelectEmotion",
                  label = "Select Emotion",
                  choices = c(
                    "anger", "anticipation", "disgust", "fear",
                    "joy", "sadness", "surprise", "trust"
                  ),
                  selected = "joy"
                )
              )
            ),
            shinycssloaders::withSpinner(
              plotlyOutput(
                outputId = "d_emoWordPlot",
                height = "75vh",
                width = "98.9%"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Document Heatmap", icon = icon("th"),
            br(),
            shinycssloaders::withSpinner(
              plotlyOutput(
                outputId = "d_emoHeatmap",
                height = "75vh",
                width = "98.9%"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Table", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("d_emoTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "TALL AI", icon = icon("robot"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  br(),
                  shinycssloaders::withSpinner(
                    htmlOutput("d_emo_GeminiUI"),
                    caption = HTML("<br><strong>Thinking...</strong>"),
                    image = "ai_small2.gif",
                    color = "#4F7942"
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "Info & References", icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  br(),
                  HTML(infoTexts$emotionanalysis)
                ),
                column(1)
              )
            )
          )
        )
      )
    )
  )

  ### Abstractive Summarization ----

  abs_summ <- tabItem(
    tabName = "d_astractive",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("wand-magic-sparkles"), strong("Abstractive Summarization"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      fluidRow(
        column(
          9,
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Abstract", icon = icon("file-lines"),
              br(),
              shinycssloaders::withSpinner(
                uiOutput("summaryData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Full Document", icon = icon("file-alt"),
              br(),
              shinycssloaders::withSpinner(
                uiOutput("documentData2"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & References", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(infoTexts$summarization)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          3,
          div(
            box(
              width = 12,
              h5(
                ("TALL generates a coherent and concise summary by interpreting and paraphrasing the main ideas from the original text"),
                style = "text-align: left; text-color: #989898"
              ),
              tags$hr(),
              fluidRow(column(
                12,
                uiOutput("optionsUnitAbstractive"),
                uiOutput("optionsAbstractive"),
                numericInput(
                  "summaryLength",
                  label = "Summary Length (in words)",
                  value = 250,
                  min = 10,
                  step = 10,
                  max = 16384
                ),
                uiOutput("abstractivePromptUI"),
              )),
              hr(),
              div(
                fluidRow(
                  column(
                    4,
                    div(
                      align = "center",
                      title = t_run,
                      do.call(
                        "actionButton",
                        c(
                          run_bttn,
                          list(
                            inputId = "d_abstractiveApply"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    4,
                    div(
                      align = "center",
                      title = t_view,
                      do.call(
                        "actionButton",
                        c(
                          view_bttn,
                          list(
                            inputId = "d_abstractiveView"
                          )
                        )
                      )
                    )
                  ),
                  column(4)
                ),
                style = "margin-top:-15px"
              )
            ),
            style = "margin-top:40px"
          )
        )
      )
    )
  )

  ### Extractive Summarization ----

  ext_summ <- tabItem(
    tabName = "d_summarization",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("scissors"), strong("Extractive Summarization"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      fluidRow(
        column(
          9,
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Abstract", icon = icon("file-lines"),
              br(),
              fluidRow(
                #style = "height:65vh",
                shinycssloaders::withSpinner(
                  uiOutput("abstractData"),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              ),
              fluidRow(
                br(),
                align = "center",
                shinycssloaders::withSpinner(
                  uiOutput("sliderAbstractData"),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              )
            ),
            tabPanel(
              title = "Full Document", icon = icon("file-alt"),
              br(),
              shinycssloaders::withSpinner(
                # DT::DTOutput("documentData"),
                uiOutput("documentData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Sentence Ranking", icon = icon("list-ol"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("RelSentData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & References", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(infoTexts$summarization)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          3,
          div(
            box(
              width = 12,
              h5(
                ("TALL selects and reorders the most relevant sentences from the original text to generate a coherent and concise summary"),
                style = "text-align: left; text-color: #989898"
              ),
              tags$hr(),
              fluidRow(column(
                12,
                uiOutput("optionsUnitSummarization"),
                uiOutput("optionsSummarization")
              )),
              hr(),
              div(
                fluidRow(
                  column(
                    4,
                    div(
                      align = "center",
                      title = t_run,
                      do.call(
                        "actionButton",
                        c(
                          run_bttn,
                          list(
                            inputId = "d_summarizationApply"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    4,
                    div(
                      align = "center",
                      title = t_view,
                      do.call(
                        "actionButton",
                        c(
                          view_bttn,
                          list(
                            inputId = "d_summarizationView"
                          )
                        )
                      )
                    )
                  ),
                  column(4)
                ),
                style = "margin-top:-15px"
              )
            ),
            style = "margin-top:40px"
          )
        )
      )
    )
  )

  return(list(
    tm_k = tm_k,
    tm_analysis = tm_analysis,
    syntactic_complexity = syntactic_complexity,
    svo_analysis = svo_analysis,
    polarity = polarity,
    emotion = emotion,
    abs_summ = abs_summ,
    ext_summ = ext_summ
  ))
}

documentsServer <- function(input, output, session, values, statsValues) {
  ## DOCUMENTS ----

  ## STM prevalence covariate selector ----
  output$stmPrevalenceSelect <- renderUI({
    req(values$dfTag)
    all_cols <- noGroupLabels(names(values$dfTag))
    # Filter to numeric, date, or factor/character types suitable as covariates
    cov_cols <- sapply(all_cols, function(col) {
      col_data <- values$dfTag[[col]]
      is.numeric(col_data) ||
        inherits(col_data, c("Date", "POSIXct", "POSIXlt")) ||
        is.character(col_data) || is.factor(col_data)
    })
    cov_choices <- all_cols[cov_cols]

    selectInput(
      inputId = "stmPrevalence",
      label = "Prevalence Covariates:",
      choices = cov_choices,
      selected = NULL,
      multiple = TRUE,
      width = "100%"
    )
  })

  ## K selection UI for Model Estimation panel ----
  output$tmKSelectionUI <- renderUI({
    # Check if K estimation was already performed
    k_estimated <- !is.null(values$TMKresult)

    if (k_estimated) {
      # Compute consensus K from previous estimation
      consensus_k <- tmConsensusK(values$TMKresult$metrics, method = input$tmMethod)

      tagList(
        div(
          style = "padding: 10px; background-color: #e8f5e9; border-left: 4px solid #4caf50; border-radius: 4px; margin-bottom: 10px;",
          icon("check-circle", style = "color: #4caf50;"),
          strong(" K estimation available"),
          p(
            paste0("Recommended K = ", consensus_k),
            style = "margin: 5px 0 0 0; color: #333; font-size: 13px;"
          )
        ),
        numericInput(
          "KEstim",
          label = "N. of Topics (K)",
          value = consensus_k,
          min = 2,
          step = 1
        )
      )
    } else {
      tagList(
        div(
          style = "padding: 10px; background-color: #fff3cd; border-left: 4px solid #ffc107; border-radius: 4px; margin-bottom: 10px;",
          icon("info-circle", style = "color: #856404;"),
          span(
            " No K estimation found. Run the K selection analysis first, or set K manually.",
            style = "color: #856404; font-size: 12px;"
          )
        ),
        numericInput(
          "KEstim",
          label = "N. of Topics (K)",
          value = 5,
          min = 2,
          step = 1
        )
      )
    }
  })

  ## Topic Modeling ----
  ## K choice (async) ----

  observeEvent(input$d_tm_selectApply, {
    # Sync: data preparation (filtering, DTM creation)
    switch(
      input$groupTm,
      Groups = {
        groupTm <- "doc_id"
      },
      {
        groupTm <- input$groupTm
      }
    )
    ## check to verify if groups exist or not
    filtered <- LemmaSelection(values$dfTag) %>% dplyr::filter(docSelected)
    if (
      input$groupTm == "doc_id" & "ungroupDoc_id" %in% names(values$dfTag)
    ) {
      filtered <- backToOriginalGroups(filtered)
    }

    # Clear previous results and show waiting state
    values$TMKresult <- NULL
    values$df <- NULL
    values$TMKcomputing <- TRUE

    showNotification(
      ui = tagList(
        icon("cogs", class = "fa-spin"),
        " Computing optimal K... The app remains usable while models are being fitted."
      ),
      id = "tmk_computing",
      duration = NULL,
      type = "message",
      closeButton = FALSE
    )

    # Build DTM synchronously (uses sourced functions: unique_identifier, etc.)
    ClusterRange <- sort(c(input$minK, input$maxK))
    minK <- max(ClusterRange[1], 2)
    maxK <- min(ClusterRange[2], length(unique(filtered$doc_id)))
    if (maxK < minK) {
      removeNotification("tmk_computing")
      values$TMKcomputing <- FALSE
      showNotification(
        paste0("K choice error: not enough documents (need at least ", minK, " unique docs)"),
        type = "error", duration = 10
      )
      return(NULL)
    }
    filtered$topic_level_id <- unique_identifier(filtered, fields = groupTm)
    dtf <- document_term_frequencies(
      filtered, document = "topic_level_id", term = "lemma"
    )
    dtm <- document_term_matrix(x = dtf)

    # DTM filtering (depends on method and top_by)
    method <- input$tmMethodK
    top_by <- input$top_by
    n <- input$nTm
    Kby <- input$Kby
    seed <- values$random_seed
    k_seq <- seq(from = minK, to = maxK, by = Kby)

    if (method == "STM") {
      switch(
        top_by,
        freq = {
          dtm <- dtm_remove_lowfreq(dtm, minfreq = 1, maxterms = n)
        },
        tfidf = {
          dtm <- dtm_remove_tfidf(dtm, top = n)
        }
      )
      # Async: STM model fitting in background
      stmTuningAsync_fn <- stmTuningAsync
      promises::future_promise({
        stmTuningAsync_fn(dtm, minK, maxK, Kby, seed)
      }, globals = list(
        stmTuningAsync_fn = stmTuningAsync_fn,
        dtm = dtm, minK = minK, maxK = maxK, Kby = Kby, seed = seed
      ), packages = c("stm", "tm"), seed = TRUE) %...>%
        (function(result) {
          removeNotification("tmk_computing")
          values$TMKresult <- result
          values$df <- result$metrics %>%
            dplyr::arrange(k) %>%
            dplyr::rename(topics = k) %>%
            dplyr::mutate(dplyr::across(
              .cols = -topics,
              .fns = ~ (. - min(.)) / (max(.) - min(.)),
              .names = "{.col}_Normalized"
            ))
          values$TMKcomputing <- FALSE
          showNotification("K choice completed!", type = "message", duration = 3)
        }) %...!%
        (function(err) {
          removeNotification("tmk_computing")
          values$TMKresult <- NULL
          values$TMKcomputing <- FALSE
          showNotification(
            paste("K choice error:", conditionMessage(err)),
            type = "error", duration = 10
          )
        })
    } else {
      # LDA/CTM: convert DTM and run in background
      switch(
        top_by,
        freq = {
          dtm <- dtm_remove_lowfreq(dtm, minfreq = 1, maxterms = n)
          dtm <- tm::as.DocumentTermMatrix(dtm, weighting = tm::weightTf)
        },
        tfidf = {
          dtm <- dtm_remove_tfidf(dtm, top = n)
          dtm <- tm::as.DocumentTermMatrix(dtm, weighting = tm::weightTfIdf)
        }
      )
      # Async: LDA/CTM model fitting in background
      tmTuningAsync_fn <- tmTuningAsync
      promises::future_promise({
        tmTuningAsync_fn(dtm, k_seq, seed, method)
      }, globals = list(
        tmTuningAsync_fn = tmTuningAsync_fn,
        dtm = dtm, k_seq = k_seq, seed = seed, method = method
      ), packages = c("topicmodels", "slam"), seed = TRUE) %...>%
        (function(result) {
          removeNotification("tmk_computing")
          values$TMKresult <- result
          values$df <- result$metrics %>%
            dplyr::arrange(k) %>%
            dplyr::rename(topics = k) %>%
            dplyr::mutate(dplyr::across(
              .cols = -topics,
              .fns = ~ (. - min(.)) / (max(.) - min(.)),
              .names = "{.col}_Normalized"
            ))
          values$TMKcomputing <- FALSE
          showNotification("K choice completed!", type = "message", duration = 3)
        }) %...!%
        (function(err) {
          removeNotification("tmk_computing")
          values$TMKresult <- NULL
          values$TMKcomputing <- FALSE
          showNotification(
            paste("K choice error:", conditionMessage(err)),
            type = "error", duration = 10
          )
        })
    }

    NULL
  })

  output$d_tm_selectPlot <- renderPlotly({
    req(values$TMKresult)
    values$TMKplot <- tmTuningPlot(values$TMKresult, metric = input$metric)
    values$TMKplot
  })

  ## Multi-Metric Comparison Plot ----
  output$d_tm_multiMetricPlot <- renderPlotly({
    req(values$TMKresult)
    tmMultiMetricPlot(values$TMKresult, method = input$tmMethodK)
  })

  ## K Recommendation Panel ----
  output$d_tm_kRecommendationUI <- renderUI({
    req(values$TMKresult)

    metrics_df <- values$TMKresult$metrics
    is_stm <- input$tmMethodK == "STM"

    # Compute optimal K per metric
    if (is_stm) {
      metric_list <- list(
        list(name = "Exclusivity", col = "CaoJuan2009", decreasing = TRUE,
             desc = "Measures how exclusive top words are to each topic. Higher = more distinct topics."),
        list(name = "Semantic Coherence", col = "Arun2010", decreasing = FALSE,
             desc = "Measures co-occurrence of top words within topics. Less negative = more coherent."),
        list(name = "Excl. + Coherence", col = "Deveaud2014", decreasing = TRUE,
             desc = "Combined score balancing exclusivity and coherence. Higher = better trade-off."),
        list(name = "Lower Bound", col = "logLik", decreasing = FALSE,
             desc = "Variational lower bound on log-likelihood. Higher (less negative) = better fit.")
      )
    } else {
      metric_list <- list(
        list(name = "CaoJuan 2009", col = "CaoJuan2009", decreasing = TRUE,
             desc = "Average cosine similarity between topics. Lower = more distinct topics."),
        list(name = "Arun 2010", col = "Arun2010", decreasing = FALSE,
             desc = "KL divergence between word-topic and doc-topic distributions. Lower = better fit."),
        list(name = "Deveaud 2014", col = "Deveaud2014", decreasing = TRUE,
             desc = "Jensen-Shannon divergence between topic pairs. Lower = more separated topics."),
        list(name = "Perplexity", col = "Perplexity", decreasing = TRUE,
             desc = "Predictive performance on held-out data. Lower = better generalization.")
      )
    }

    # Find optimal K for each metric
    recommendations <- lapply(metric_list, function(m) {
      k_opt <- find_elbow(
        metrics_df$k,
        metrics_df[[m$col]],
        decreasing = m$decreasing,
        plot = FALSE
      )
      list(name = m$name, k = k_opt, desc = m$desc)
    })

    # Consensus K (mode of recommendations)
    k_votes <- sapply(recommendations, function(r) r$k)
    k_table <- sort(table(k_votes), decreasing = TRUE)
    consensus_k <- as.integer(names(k_table)[1])
    vote_count <- k_table[1]
    total_metrics <- length(recommendations)

    # Build UI cards
    rec_cards <- lapply(recommendations, function(r) {
      is_consensus <- (r$k == consensus_k)
      div(
        style = paste0(
          "padding: 15px; border-radius: 8px; margin-bottom: 10px; border-left: 5px solid ",
          if (is_consensus) "#4F7942" else "#667eea", "; background-color: #f8f9fa;"
        ),
        fluidRow(
          column(
            8,
            strong(r$name, style = "font-size: 15px;"),
            if (is_consensus) span(" (consensus)", style = "color: #4F7942; font-weight: 600; font-size: 12px;"),
            p(r$desc, style = "color: #666; font-size: 12px; margin: 5px 0 0 0;")
          ),
          column(
            4,
            div(
              style = "text-align: center; padding: 8px;",
              span(
                paste0("K = ", r$k),
                style = paste0(
                  "font-size: 24px; font-weight: 700; color: ",
                  if (is_consensus) "#4F7942" else "#667eea", ";"
                )
              )
            )
          )
        )
      )
    })

    tagList(
      # Consensus header
      div(
        style = "padding: 20px; background: linear-gradient(135deg, #4F7942 0%, #6CC283 100%); color: white; border-radius: 10px; margin-bottom: 20px; text-align: center;",
        h3(
          icon("bullseye"),
          paste0("Recommended K = ", consensus_k),
          style = "margin: 0; font-weight: 700;"
        ),
        p(
          paste0(
            "Consensus from ", vote_count, " out of ", total_metrics,
            " metrics | Method: ", input$tmMethodK
          ),
          style = "margin: 8px 0 0 0; opacity: 0.9;"
        ),
        if (as.integer(vote_count) < total_metrics) {
          p(
            paste0("All recommendations: K = ", paste(sort(unique(k_votes)), collapse = ", ")),
            style = "margin: 5px 0 0 0; opacity: 0.8; font-size: 13px;"
          )
        }
      ),

      # Per-metric recommendations
      rec_cards
    )
  })

  output$d_tm_selectTable <- renderDataTable({
    req(values$df)
    DTformat(
      values$df,
      numeric = c(2, 3),
      round = 2,
      nrow = nrow(values$df),
      size = "110%"
    )
  })

  observeEvent(
    eventExpr = {
      input$d_tm_selectExport
    },
    handlerExpr = {
      file <- paste("TMTopicSelection-", sys.time(), ".png", sep = "")
      file <- destFolder(file, values$wdTall)
      plot2png(values$TMKplot, filename = file, type = "plotly", dpi = values$dpi, height = values$h)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report

  observeEvent(input$d_tm_selectReport, {
    if (!is.null(values$df)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "KChoice"
      list_df <- list(values$df)
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileKchoice <- plot2png(
        values$TMKplot,
        filename = "kchoiche.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileKchoice, res$col)
      )
      popUp(title = "K choice Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Model estimation ----

  netTMestim <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$d_tm_estimApply
    },
    valueExpr = {
      req(input$groupTmEstim, input$KEstim, input$tmMethod)
      values$TMplotIndex <- 1
      values$TMdocIndex <- 1

      switch(
        input$groupTmEstim,
        Groups = {
          groupTmEstim <- "doc_id"
        },
        {
          groupTmEstim <- input$groupTmEstim
        }
      )

      ## check to verify if groups exist or not
      filtered <- LemmaSelection(values$dfTag) %>% dplyr::filter(docSelected)
      if (
        input$groupTmEstim == "doc_id" &
          "ungroupDoc_id" %in% names(values$dfTag)
      ) {
        tm_data <- backToOriginalGroups(filtered)
      } else {
        tm_data <- filtered
      }

      # Get prevalence covariates for STM
      stm_prevalence <- NULL
      if (input$tmMethod == "STM" && !is.null(input$stmPrevalence) && length(input$stmPrevalence) > 0) {
        stm_prevalence <- input$stmPrevalence
      }

      values$tmK <- input$KEstim
      values$TMplotList <- split(
        1:values$tmK,
        ceiling(seq_along(1:values$tmK) / 3)
      )
      values$TMestim_result <- tmEstimate(
        tm_data,
        K = values$tmK,
        group = groupTmEstim,
        term = values$generalTerm,
        n = input$nTmEstim,
        top_by = input$top_byEstim,
        method = input$tmMethod,
        prevalence = stm_prevalence,
        seed = values$random_seed
      )
      ## End check ###

      ### BETA PROBABILITY
      values$beta <- values$TMestim_result$beta

      names(values$beta)[2:ncol(values$beta)] <- paste0(
        "Topic ",
        1:(ncol(values$beta) - 1)
      )
      values$tmHeatmap <- tmHeatmap(values$beta)

      ### THETA PROBABILITY
      values$theta <- values$TMestim_result$theta
      names(values$theta)[2:ncol(values$theta)] <- paste0(
        "Topic ",
        1:(ncol(values$theta) - 1)
      )
    }
  )

  output$d_tm_networkPlot <- renderPlotly({
    netTMestim()
    values$tmHeatmap$Hplot
  })

  ## Model Diagnostics output ----
  output$d_tm_diagnosticsUI <- renderUI({
    netTMestim()
    req(values$TMestim_result)

    model <- values$TMestim_result$topicModel
    K <- values$tmK
    method <- input$tmMethod

    # Common metrics
    if (method == "STM") {
      # STM diagnostics
      semcoh <- stm::semanticCoherence(model, documents = values$TMestim_result$stm_documents)
      excl <- stm::exclusivity(model)
      diag_df <- data.frame(
        Topic = paste("Topic", 1:K),
        `Semantic Coherence` = round(semcoh, 3),
        Exclusivity = round(excl, 3),
        check.names = FALSE
      )
      bound <- round(max(model$convergence$bound), 2)
      its <- model$convergence$its
      convergence_text <- paste0(
        "Approximate variational lower bound: <strong>", bound, "</strong><br>",
        "Iterations to convergence: <strong>", its, "</strong>"
      )
    } else {
      # LDA / CTM diagnostics
      ll <- round(as.numeric(logLik(model)), 2)
      n_terms <- model@Dim[2]
      n_docs <- model@Dim[1]

      # Per-topic word entropy
      beta_mat <- exp(model@beta)
      entropy <- apply(beta_mat, 1, function(row) {
        p <- row[row > 0]
        -sum(p * log(p))
      })
      # Topic size (mean theta per topic)
      topic_prop <- colMeans(model@gamma)
      diag_df <- data.frame(
        Topic = paste("Topic", 1:K),
        `Topic Share` = paste0(round(topic_prop * 100, 1), "%"),
        `Word Entropy` = round(entropy, 3),
        `Top Word Prob.` = round(apply(beta_mat, 1, max), 4),
        check.names = FALSE
      )
      convergence_text <- paste0(
        "Log-Likelihood: <strong>", format(ll, big.mark = ","), "</strong><br>",
        "Documents: <strong>", n_docs, "</strong> | ",
        "Vocabulary: <strong>", n_terms, "</strong>"
      )
    }

    tagList(
      fluidRow(
        column(
          6,
          div(
            style = "padding: 20px; background: linear-gradient(135deg, #667eea 0%, #764ba2 100%); color: white; border-radius: 10px; margin-bottom: 20px;",
            h4(icon("chart-line"), strong(paste0("Model Diagnostics — ", method)),
              style = "margin-top: 0; color: white;"),
            h5(paste0("K = ", K, " topics"), style = "opacity: 0.9; color: white;")
          ),
          div(
            style = "padding: 15px; background-color: #f8f9fa; border-radius: 8px; border-left: 4px solid #667eea; margin-bottom: 20px;",
            h5(strong("Global Metrics"), style = "margin-top: 0;"),
            HTML(convergence_text)
          )
        ),
        column(
          6,
          div(
            style = "padding: 15px; background-color: #fff; border: 1px solid #e0e0e0; border-radius: 8px;",
            h5(strong("Per-Topic Quality"), style = "margin-top: 0;"),
            DT::renderDT(
              DT::datatable(
                diag_df,
                options = list(dom = "t", pageLength = 50, scrollY = "350px"),
                rownames = FALSE,
                class = "compact stripe"
              )
            )
          )
        )
      ),
      if (method == "STM") {
        fluidRow(
          column(
            12,
            br(),
            div(
              style = "padding: 15px; background-color: #fff; border: 1px solid #e0e0e0; border-radius: 8px;",
              h5(strong("Semantic Coherence vs Exclusivity"), style = "margin-top: 0;"),
              renderPlotly({
                plot_ly(
                  diag_df,
                  x = ~`Semantic Coherence`,
                  y = ~Exclusivity,
                  text = ~Topic,
                  type = "scatter",
                  mode = "markers+text",
                  textposition = "top center",
                  marker = list(size = 12, color = "#667eea"),
                  textfont = list(size = 10, color = "gray40")
                ) %>%
                  layout(
                    xaxis = list(title = "Semantic Coherence"),
                    yaxis = list(title = "Exclusivity"),
                    paper_bgcolor = "white",
                    plot_bgcolor = "#fafafa"
                  ) %>%
                  config(displaylogo = FALSE)
              })
            )
          )
        )
      }
    )
  })

  ## Covariate Effects output (STM only) ----
  output$d_tm_covariateUI <- renderUI({
    netTMestim()
    req(values$TMestim_result)

    if (input$tmMethod != "STM") {
      return(
        div(
          style = "text-align: center; padding: 60px; color: #999;",
          icon("info-circle", style = "font-size: 48px; margin-bottom: 15px;"),
          h4("Covariate effects are available only for STM models."),
          p("Select STM (Structural) as the model type to use prevalence covariates.")
        )
      )
    }

    model <- values$TMestim_result$topicModel
    stm_meta <- values$TMestim_result$stm_meta
    K <- values$tmK
    prevalence_vars <- input$stmPrevalence

    if (is.null(prevalence_vars) || length(prevalence_vars) == 0 || is.null(stm_meta)) {
      return(
        div(
          style = "text-align: center; padding: 60px; color: #999;",
          icon("info-circle", style = "font-size: 48px; margin-bottom: 15px;"),
          h4("No prevalence covariates were specified."),
          p("Select covariates in the Options panel and re-run the analysis.")
        )
      )
    }

    # Rename metadata columns to safe names for estimateEffect
    safe_names <- gsub("[^a-zA-Z0-9_]", "_", prevalence_vars)
    meta_safe <- stm_meta
    for (i in seq_along(prevalence_vars)) {
      if (prevalence_vars[i] %in% names(meta_safe)) {
        names(meta_safe)[names(meta_safe) == prevalence_vars[i]] <- safe_names[i]
      }
    }

    # Build covariate effect plots
    cov_panels <- lapply(seq_along(prevalence_vars), function(idx) {
      cov_var <- prevalence_vars[idx]
      safe_var <- safe_names[idx]

      prep <- tryCatch({
        stm::estimateEffect(
          as.formula(paste("1:", K, "~", safe_var)),
          stmobj = model,
          metadata = meta_safe
        )
      }, error = function(e) NULL)

      if (is.null(prep)) {
        return(div(
          style = "padding: 20px; background-color: #fff3cd; border-radius: 8px; margin-bottom: 20px;",
          icon("exclamation-triangle", style = "color: #856404;"),
          strong(cov_var), ": unable to estimate covariate effects."
        ))
      }

      # Get covariate data to determine type
      cov_data <- meta_safe[[safe_var]]
      is_continuous <- is.numeric(cov_data)

      plot_output_id <- paste0("stm_cov_plot_", gsub("[^a-zA-Z0-9]", "_", cov_var))

      # Render the plot
      output[[plot_output_id]] <- renderPlot({
        if (is_continuous) {
          plot(
            prep,
            covariate = safe_var,
            topics = 1:K,
            model = model,
            method = "continuous",
            xlab = cov_var,
            main = paste0("Effect of ", cov_var, " on Topic Prevalence"),
            ci.level = 0.95,
            labeltype = "custom",
            custom.labels = paste("Topic", 1:K)
          )
        } else {
          plot(
            prep,
            covariate = safe_var,
            topics = 1:K,
            model = model,
            method = "pointestimate",
            xlab = "Expected Topic Proportion",
            main = paste0("Effect of ", cov_var, " on Topic Prevalence"),
            ci.level = 0.95,
            labeltype = "custom",
            custom.labels = paste("Topic", 1:K)
          )
        }
      }, height = max(400, K * 40))

      # Summary table
      summary_prep <- summary(prep)
      coef_list <- lapply(seq_len(K), function(t) {
        tbl <- summary_prep$tables[[t]]
        data.frame(
          Topic = paste("Topic", t),
          Coefficient = rownames(tbl),
          Estimate = round(tbl[, "Estimate"], 4),
          `Std. Error` = round(tbl[, "Std. Error"], 4),
          `t value` = round(tbl[, "t value"], 3),
          `Pr(>|t|)` = format.pval(tbl[, "Pr(>|t|)"], digits = 3),
          check.names = FALSE,
          stringsAsFactors = FALSE
        )
      })
      coef_df <- do.call(rbind, coef_list)
      # Keep only rows related to this covariate (not intercept)
      coef_df <- coef_df[!grepl("^\\(Intercept\\)$", coef_df$Coefficient), ]

      coef_output_id <- paste0("stm_cov_table_", gsub("[^a-zA-Z0-9]", "_", cov_var))
      output[[coef_output_id]] <- DT::renderDT({
        DT::datatable(
          coef_df,
          options = list(dom = "t", pageLength = 100, scrollY = "300px"),
          rownames = FALSE,
          class = "compact stripe"
        )
      })

      # UI element for this covariate
      div(
        style = "margin-bottom: 30px;",
        div(
          style = "padding: 12px; background: linear-gradient(135deg, #4F7942 0%, #6CC283 100%); color: white; border-radius: 8px 8px 0 0;",
          h4(icon("chart-area"), strong(cov_var),
            if (is_continuous) span(" (continuous)", style = "opacity: 0.8; font-size: 13px; color: white;")
            else span(" (categorical)", style = "opacity: 0.8; font-size: 13px; color: white;"),
            style = "margin: 0; color: white;")
        ),
        fluidRow(
          column(
            7,
            div(
              style = "padding: 15px; background-color: #fff; border: 1px solid #e0e0e0; border-radius: 0 0 0 8px;",
              plotOutput(plot_output_id, height = paste0(max(400, K * 40), "px"))
            )
          ),
          column(
            5,
            div(
              style = "padding: 15px; background-color: #f8f9fa; border: 1px solid #e0e0e0; border-radius: 0 0 8px 0;",
              h5(strong("Regression Coefficients"), style = "margin-top: 0;"),
              DT::DTOutput(coef_output_id)
            )
          )
        )
      )
    })

    tagList(
      div(
        style = "padding: 20px; background: linear-gradient(135deg, #4F7942 0%, #6CC283 100%); color: white; border-radius: 10px; margin-bottom: 20px;",
        h4(icon("project-diagram"), strong("STM Covariate Effects"),
          style = "margin-top: 0; color: white;"),
        h5(paste0("Prevalence covariates: ", paste(prevalence_vars, collapse = ", ")),
          style = "opacity: 0.9; color: white;")
      ),
      cov_panels
    )
  })

  output$TMplotArrowsUI <- renderUI({
    req(values$TMestim_result)
    fluidRow(
      actionButton("TMplotLeft", icon("chevron-left")),
      actionButton("TMplotRight", icon("chevron-right")),
      align = "center"
    )
  })

  observeEvent(input$TMplotRight, {
    if (values$TMplotIndex < ceiling(req(values$tmK) / 3)) {
      values$TMplotIndex <- values$TMplotIndex + 1
    }
  })

  observeEvent(input$TMplotLeft, {
    if (req(values$TMplotIndex) > 1) {
      values$TMplotIndex <- values$TMplotIndex - 1
    }
  })

  output$d_tm_estimTPlot1 <- renderPlotly({
    netTMestim()
    if (!values$TMplotIndex %in% 1:length(values$TMplotList)) {
      values$TMplotIndex <- 1
    }
    topic1 <- values$TMplotList[[values$TMplotIndex]]
    values$TMestim_plot1 <- tmTopicPlot(
      values$TMestim_result$beta,
      topic = topic1[[1]],
      nPlot = input$nTopicPlot
    )
    values$TMestim_plot1
  })

  output$d_tm_estimTPlot2 <- renderPlotly({
    topic2 <- values$TMplotList[[values$TMplotIndex]]
    if (length(topic2) >= 2) {
      values$TMestim_plot2 <- tmTopicPlot(
        values$TMestim_result$beta,
        topic = topic2[[2]],
        nPlot = input$nTopicPlot
      )
      values$TMestim_plot2
    }
  })

  output$d_tm_estimTPlot3 <- renderPlotly({
    topic3 <- values$TMplotList[[values$TMplotIndex]]
    if (length(topic3) == 3) {
      values$TMestim_plot3 <- tmTopicPlot(
        values$TMestim_result$beta,
        topic = topic3[[3]],
        nPlot = input$nTopicPlot
      )
      values$TMestim_plot3
    }
  })

  output$d_tm_estimBpTable <- renderDataTable(server = FALSE, {
    netTMestim()
    DTformat(
      values$beta,
      left = 1,
      numeric = c(2:ncol(values$TMestim_result$beta)),
      round = 4,
      nrow = 10,
      size = "85%",
      filename = "TopicModel_BetaTable"
    )
  })

  output$TMdocArrowsUI <- renderUI({
    req(values$TMestim_result)
    fluidRow(
      actionButton("TMdocLeft", icon("chevron-left")),
      actionButton("TMdocRight", icon("chevron-right")),
      align = "center"
    )
  })

  observeEvent(input$TMdocRight, {
    if (values$TMdocIndex < ceiling(req(values$tmK) / 3)) {
      values$TMdocIndex <- values$TMdocIndex + 1
    }
  })

  observeEvent(input$TMdocLeft, {
    if (req(values$TMdocIndex) > 1) {
      values$TMdocIndex <- values$TMdocIndex - 1
    }
  })

  output$d_tm_DocPlot1 <- renderPlotly({
    netTMestim()
    if (!values$TMdocIndex %in% 1:length(values$TMplotList)) {
      values$TMdocIndex <- 1
    }
    topic1 <- values$TMplotList[[values$TMdocIndex]]
    values$TMdoc_plot1 <- tmDocPlot(
      values$TMestim_result$theta,
      topic = topic1[[1]],
      nPlot = input$nTopicPlot
    )
    values$TMdoc_plot1
  })

  output$d_tm_DocPlot2 <- renderPlotly({
    topic2 <- values$TMplotList[[values$TMdocIndex]]
    if (length(topic2) >= 2) {
      values$TMdoc_plot2 <- tmDocPlot(
        values$TMestim_result$theta,
        topic = topic2[[2]],
        nPlot = input$nTopicPlot
      )
      values$TMdoc_plot2
    }
  })

  output$d_tm_DocPlot3 <- renderPlotly({
    topic3 <- values$TMplotList[[values$TMdocIndex]]
    if (length(topic3) == 3) {
      values$TMdoc_plot3 <- tmDocPlot(
        values$TMestim_result$theta,
        topic = topic3[[3]],
        nPlot = input$nTopicPlot
      )
      values$TMdoc_plot3
    }
  })

  output$d_tm_estimTpTable <- renderDataTable(server = FALSE, {
    netTMestim()
    DTformat(
      values$theta,
      left = 1,
      numeric = c(2:ncol(values$TMestim_result$theta)),
      round = 4,
      nrow = 10,
      size = "85%",
      filename = "TopicModel_ThetaTable"
    )
  })

  output$d_tm_GeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "Gemini AI", content = values$tmGemini, values)
  })

  observeEvent(
    eventExpr = {
      input$d_tm_estimExport
    },
    handlerExpr = {
      file1 <- paste("TMCorrPlots-", sys.time(), ".png", sep = "")
      file1 <- destFolder(file1, values$wdTall)
      file2 <- paste("TMTermPlots-", sys.time(), ".png", sep = "")
      file2 <- destFolder(file2, values$wdTall)
      file3 <- paste("TMDocPlots-", sys.time(), ".png", sep = "")
      file3 <- destFolder(file3, values$wdTall)
      values$tmGplotBeta <- topicGplot(
        values$TMestim_result$beta,
        nPlot = input$nTopicPlot,
        type = "beta"
      )
      values$tmGplotTheta <- topicGplot(
        values$TMestim_result$theta,
        nPlot = input$nTopicPlot,
        type = "theta"
      )
      ggsave(
        filename = file1,
        plot = values$tmHeatmap$HplotStatic,
        dpi = values$dpi,
        height = values$h,
        width = values$h * values$aspect,
        bg = "transparent"
      )
      ggsave(
        filename = file2,
        plot = values$tmGplotBeta,
        dpi = values$dpi,
        height = values$h,
        width = values$h * values$aspect,
        bg = "transparent"
      )
      ggsave(
        filename = file3,
        plot = values$tmGplotTheta,
        dpi = values$dpi,
        height = values$h,
        width = values$h * values$aspect,
        bg = "transparent"
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report

  observeEvent(input$d_tm_estimReport, {
    if (!is.null(values$TMestim_result$beta)) {
      popUp(title = NULL, type = "waiting")
      values$tmGplotBeta <- topicGplot(
        values$TMestim_result$beta,
        nPlot = input$nTopicPlot,
        type = "beta"
      )
      values$tmGplotTheta <- topicGplot(
        values$TMestim_result$theta,
        nPlot = input$nTopicPlot,
        type = "theta"
      )
      Gem <- values$tmGemini %>% string_to_sentence_df()
      list_df <- list(
        values$tmGemini %>% string_to_sentence_df(),
        values$beta,
        values$theta
      )
      list_plot <- list(
        values$tmGplotBeta,
        values$tmGplotTheta,
        values$tmHeatmap$HplotStatic
      )
      wb <- addSheetToReport(
        list_df,
        list_plot,
        sheetname = "ModelEstim",
        wb = values$wb,
        dpi = values$report_dpi,
        startRow = nrow(Gem) + 1
      )
      values$wb <- wb
      popUp(title = "Model Estimation Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Syntactic Complexity ----

  syntacticFunction <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$d_syntacticApply
    },
    valueExpr = {
      filtered <- values$dfTag %>% dplyr::filter(docSelected)
      values$syntacticResults <- computeSyntacticComplexity(filtered)
    }
  )

  output$d_syntacticTable <- renderDT(server = FALSE, {
    syntacticFunction()
    req(values$syntacticResults)
    df <- values$syntacticResults %>%
      rename(
        "Document" = doc_id,
        "Sentences" = n_sentences,
        "Mean Sent. Length" = mean_sent_length,
        "Mean Tree Depth" = mean_tree_depth,
        "Max Tree Depth" = max_tree_depth,
        "Mean Dep. Distance" = mean_dep_distance,
        "Clauses/Sent." = mean_clauses_per_sent,
        "Subordinate/Sent." = mean_subordinate,
        "Coordinate/Sent." = mean_coordinate,
        "Subordination Ratio" = subordination_ratio,
        "Branching Factor" = mean_branching_factor
      )
    DTformat(
      df,
      size = "100%",
      filename = "SyntacticComplexity",
      pagelength = TRUE,
      dom = TRUE,
      filter = "top"
    )
  })

  output$d_syntacticSummaryUI <- renderUI({
    syntacticFunction()
    req(values$syntacticResults)
    df <- values$syntacticResults
    n_docs <- nrow(df)

    metrics <- list(
      list(name = "Mean Sentence Length", val = mean(df$mean_sent_length), desc = "words per sentence (excl. punctuation)"),
      list(name = "Mean Tree Depth", val = mean(df$mean_tree_depth), desc = "levels of syntactic nesting"),
      list(name = "Mean Dependency Distance", val = mean(df$mean_dep_distance), desc = "tokens between head and dependent"),
      list(name = "Mean Clauses per Sentence", val = mean(df$mean_clauses_per_sent), desc = "main + subordinate + coordinate"),
      list(name = "Subordination Ratio", val = mean(df$subordination_ratio), desc = "proportion of subordinate clauses"),
      list(name = "Mean Branching Factor", val = mean(df$mean_branching_factor), desc = "children per non-leaf node")
    )

    cards <- lapply(metrics, function(m) {
      column(
        4,
        div(
          style = "padding: 20px; background-color: #f8f9fa; border-radius: 10px; border-left: 4px solid #4F7942; margin-bottom: 15px; text-align: center;",
          h3(strong(round(m$val, 2)), style = "color: #4F7942; margin: 0;"),
          h5(strong(m$name), style = "margin: 5px 0;"),
          p(m$desc, style = "color: #888; font-size: 12px; margin: 0;")
        )
      )
    })

    tagList(
      div(
        style = "padding: 15px; background: linear-gradient(135deg, #667eea 0%, #764ba2 100%); color: white; border-radius: 10px; margin-bottom: 20px;",
        h4(icon("chart-bar"), strong("Corpus-Level Syntactic Profile"),
          style = "margin-top: 0; color: white;"),
        h5(paste0(n_docs, " documents analyzed"), style = "opacity: 0.9; color: white;")
      ),
      fluidRow(cards[1:3]),
      fluidRow(cards[4:6])
    )
  })

  output$d_syntacticDistPlot <- renderPlotly({
    syntacticFunction()
    req(values$syntacticResults)
    df <- values$syntacticResults

    # Multi-panel distribution plot
    metrics_to_plot <- list(
      list(col = "mean_sent_length", title = "Sentence Length"),
      list(col = "mean_tree_depth", title = "Tree Depth"),
      list(col = "mean_dep_distance", title = "Dependency Distance"),
      list(col = "mean_clauses_per_sent", title = "Clauses per Sentence"),
      list(col = "subordination_ratio", title = "Subordination Ratio"),
      list(col = "mean_branching_factor", title = "Branching Factor")
    )

    colors <- c("#4575B4", "#4F7942", "#D73027", "#FF7F00", "#984EA3", "#E41A1C")

    fig <- plot_ly()
    for (i in seq_along(metrics_to_plot)) {
      m <- metrics_to_plot[[i]]
      fig <- fig %>%
        add_trace(
          x = df[[m$col]],
          type = "histogram",
          name = m$title,
          marker = list(color = colors[i], line = list(color = "white", width = 1)),
          opacity = 0.8,
          visible = if (i == 1) TRUE else "legendonly"
        )
    }

    values$syntacticDistPlot <- fig %>%
      layout(
        title = list(
          text = "<b>Distribution of Syntactic Complexity Metrics</b>",
          font = list(size = 16, color = "gray30"),
          x = 0.5
        ),
        xaxis = list(title = "Value"),
        yaxis = list(title = "Number of Documents"),
        barmode = "overlay",
        paper_bgcolor = "white",
        plot_bgcolor = "white",
        legend = list(orientation = "h", x = 0.5, xanchor = "center", y = -0.12)
      ) %>%
      config(displaylogo = FALSE)

    values$syntacticDistPlot
  })

  ## Syntactic Complexity Export (PNG)
  observeEvent(input$d_syntacticExport, {
    req(values$syntacticDistPlot)
    file <- paste("SyntacticComplexity-", sys.time(), ".png", sep = "")
    file <- destFolder(file, values$wdTall)
    plot2png(values$syntacticDistPlot, filename = file, type = "plotly", dpi = values$dpi, height = values$h)
    popUp(title = "Saved in your working folder", type = "saved")
  })

  ## Syntactic Complexity Report
  observeEvent(input$d_syntacticReport, {
    if (!is.null(values$syntacticResults) && nrow(values$syntacticResults) > 0) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "SyntacticComplexity"
      list_df <- list(values$syntacticResults)
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileSyntacticDist <- plot2png(
        values$syntacticDistPlot,
        filename = "SyntacticDist.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileSyntacticDist, res$col)
      )
      popUp(title = "Syntactic Complexity Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Syntactic Complexity TALL AI
  output$d_syntactic_GeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "Gemini AI", content = values$d_syntactic_Gemini, values)
  })

  ## SVO Triplet Extraction ----

  svoFunction <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$d_svoApply
    },
    valueExpr = {
      filtered <- LemmaSelection(values$dfTag) %>% dplyr::filter(docSelected)

      values$svoResults <- extractSVO(
        filtered,
        term = values$generalTerm,
        freq.min = input$svoFreqMin
      )

      if (nrow(values$svoResults) > 0) {
        # Top N for display
        values$svoTop <- values$svoResults %>%
          slice_head(n = input$svoTopN)
      }
    }
  )

  output$d_svoTable <- renderDT(server = FALSE, {
    svoFunction()
    req(values$svoResults)
    DTformat(
      values$svoResults %>%
        rename(
          Subject = subject,
          Verb = verb,
          Object = object,
          Frequency = freq,
          "Relation Type" = rel_type
        ),
      size = "100%",
      filename = "SVO_Triplets",
      pagelength = TRUE,
      dom = TRUE,
      filter = "top"
    )
  })

  output$d_svoNetwork <- renderPlotly({
    svoFunction()
    req(values$svoTop)
    df <- values$svoTop %>% filter(object != "")

    if (nrow(df) == 0) return(plotly_empty())

    # Build a Sankey-like network: Subject -> Verb -> Object
    subjects <- unique(df$subject)
    verbs <- unique(df$verb)
    objects <- unique(df$object[df$object != ""])
    all_nodes <- c(subjects, verbs, objects)

    # Source: subject -> verb
    sv_source <- match(df$subject, all_nodes) - 1
    sv_target <- match(df$verb, all_nodes) - 1
    sv_value <- df$freq

    # Target: verb -> object
    vo_source <- match(df$verb, all_nodes) - 1
    vo_target <- match(df$object, all_nodes) - 1
    vo_value <- df$freq

    # Node colors
    node_colors <- c(
      rep("#4575B4", length(subjects)),
      rep("#4F7942", length(verbs)),
      rep("#D73027", length(objects))
    )

    values$svoSankeyPlot <- plot_ly(
      type = "sankey",
      orientation = "h",
      node = list(
        label = all_nodes,
        color = node_colors,
        pad = 15,
        thickness = 20,
        line = list(color = "black", width = 0.5)
      ),
      link = list(
        source = c(sv_source, vo_source),
        target = c(sv_target, vo_target),
        value = c(sv_value, vo_value),
        color = "rgba(200,200,200,0.4)"
      )
    ) %>%
      layout(
        title = list(
          text = "<b>SVO Triplet Flow</b><br><sub>Subject (blue) -> Verb (green) -> Object (red)</sub>",
          font = list(size = 16, color = "gray30"),
          x = 0.5
        ),
        font = list(size = 11),
        paper_bgcolor = "white"
      ) %>%
      config(displaylogo = FALSE)

    values$svoSankeyPlot
  })

  output$d_svoVerbPlot <- renderPlotly({
    svoFunction()
    req(values$svoResults)

    verb_freq <- values$svoResults %>%
      group_by(verb) %>%
      summarise(freq = sum(freq), .groups = "drop") %>%
      arrange(desc(freq)) %>%
      slice_head(n = 30) %>%
      mutate(verb = factor(verb, levels = rev(verb)))

    values$svoVerbPlot <- plot_ly(
      verb_freq,
      x = ~freq,
      y = ~verb,
      type = "bar",
      orientation = "h",
      marker = list(color = "#4F7942")
    ) %>%
      layout(
        title = list(
          text = "<b>Most Frequent Verbs in SVO Triplets</b>",
          font = list(size = 16, color = "gray30"),
          x = 0.5
        ),
        xaxis = list(title = "Frequency"),
        yaxis = list(title = ""),
        paper_bgcolor = "white",
        plot_bgcolor = "white",
        margin = list(l = 120)
      ) %>%
      config(displaylogo = FALSE)

    values$svoVerbPlot
  })

  ## SVO Export (PNG images)
  observeEvent(input$d_svoExport, {
    req(values$svoResults)
    file1 <- paste("SVO-Network-", sys.time(), ".png", sep = "")
    file1 <- destFolder(file1, values$wdTall)
    file2 <- paste("SVO-VerbFreq-", sys.time(), ".png", sep = "")
    file2 <- destFolder(file2, values$wdTall)
    plot2png(values$svoSankeyPlot, filename = file1, type = "plotly", dpi = values$dpi, height = values$h)
    plot2png(values$svoVerbPlot, filename = file2, type = "plotly", dpi = values$dpi, height = values$h)
    popUp(title = "Saved in your working folder", type = "saved")
  })

  ## SVO Report (tables + images)
  observeEvent(input$d_svoReport, {
    if (!is.null(values$svoResults) && nrow(values$svoResults) > 0) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "SVOTriplets"
      list_df <- list(values$svoResults)
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileSVOSankey <- plot2png(
        values$svoSankeyPlot,
        filename = "SVONetwork.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$fileSVOVerb <- plot2png(
        values$svoVerbPlot,
        filename = "SVOVerbFreq.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileSVOSankey, res$col),
        c(sheetname = res$sheetname, values$fileSVOVerb, res$col)
      )
      popUp(title = "SVO Triplet Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## SVO TALL AI
  output$d_svo_GeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "Gemini AI", content = values$d_svo_Gemini, values)
  })

  ## Polarity detection ----

  output$lexiconD_polarity <- renderUI({
    if (values$language == "english") {
      selectInput(
        inputId = "lexiconD_polarity",
        label = "Select lexicon",
        choices = c(
          "huliu",
          "loughran_mcdonald",
          "nrc"
        ),
        selected = "huliu"
      )
    }
    # )
  })

  ## Model estimation ----
  docPolarityEstim <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$d_polDetApply
    },
    valueExpr = {
      choices <- c(
        "english",
        "italian",
        "french",
        "german",
        "spanish",
        "afrikaans",
        "arabic",
        "armenian",
        "basque",
        "belarusian",
        "bulgarian",
        "catalan",
        "chinese",
        "croatian",
        "czech",
        "danish",
        "dutch",
        "estonian",
        "finnish",
        "galician",
        "greek",
        "hebrew",
        "hindi",
        "hungarian",
        "indonesian",
        "irish",
        "japanese",
        "korean",
        "latin",
        "latvian",
        "lithuanian",
        "maltese",
        "marathi",
        "norwegian",
        "persian",
        "polish",
        "portuguese",
        "romanian",
        "russian",
        "serbian",
        "slovak",
        "slovenian",
        "swedish",
        "tamil",
        "telugu",
        "turkish",
        "ukrainian",
        "urdu",
        "uyghur",
        "vietnamese"
      )
      if (values$language %in% choices) {
        if (is.null(input$lexiconD_polarity)) {
          lexiconD_polarity <- "huliu"
        } else {
          lexiconD_polarity <- input$lexiconD_polarity
        }

        ## check to verify if groups exist or not
        if (
          input$groupPolarity == "doc_id" &
            "ungroupDoc_id" %in% names(values$dfTag)
        ) {
          values$docPolarity <- sentimentAnalysis(
            backToOriginalGroups(values$dfTag) %>% filter(docSelected),
            language = values$language,
            lexicon_model = lexiconD_polarity
          )
        } else {
          values$docPolarity <- sentimentAnalysis(
            values$dfTag %>% filter(docSelected),
            language = values$language,
            lexicon_model = lexiconD_polarity
          )
        }
        values$docPolPlots <- sentimentWordPlot(
          values$docPolarity$sent_data,
          n = 10
        )
      }

      values$docPolarityOverallData <- values$docPolarity$sent_overall %>%
        select(
          doc_id,
          sentiment_polarity,
          doc_pol_clas,
          terms_positive,
          terms_negative
        ) %>%
        rename(
          Polarity = sentiment_polarity,
          "Polarity Category" = doc_pol_clas,
          "Positive Words" = terms_positive,
          "Negative Words" = terms_negative
        )
      # Pie chart
      values$sentimentPieChart <- sentimentPieChart(
        values$docPolarity$sent_overall %>%
          count(doc_pol_clas) %>%
          rename("Polarity" = doc_pol_clas)
      )
      # Density plot
      values$sentimentDensityPlot <- sentimentDensityPlot(
        values$docPolarity$sent_overall$sentiment_polarity,
        from = -1,
        to = 1
      )
      # Box plot
      values$sentimentBoxPlot <- sentimentBoxPlot(
        values$docPolarity$sent_overall
      )
    }
  )

  output$d_polPiePlot <- renderPlotly({
    docPolarityEstim()
    values$sentimentPieChart
  })

  output$d_polDensPlot <- renderPlotly({
    docPolarityEstim()
    values$sentimentDensityPlot
  })

  output$d_polBoxPlot <- renderPlotly({
    docPolarityEstim()
    values$sentimentBoxPlot
  })

  output$d_polDetPlotPos <- renderPlotly({
    docPolarityEstim()
    values$docPolPlots$positive
  })
  output$d_polDetPlotNeg <- renderPlotly({
    docPolarityEstim()
    values$docPolPlots$negative
  })

  output$d_polDet_GeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "Gemini AI", content = values$d_polDet_Gemini, values)
  })

  output$d_polDetTable <- renderDT(server = FALSE, {
    docPolarityEstim()
    DTformat(
      values$docPolarityOverallData,
      filename = "DocPolarity",
      left = c(2, 4, 5, 6),
      numeric = 3,
      round = 4,
      button = TRUE
    )
  })

  observeEvent(
    eventExpr = {
      input$d_polDetExport
    },
    handlerExpr = {
      file1 <- paste("PieChart-", sys.time(), ".png", sep = "")
      file1 <- destFolder(file1, values$wdTall)
      file2 <- paste("DensDensity-", sys.time(), ".png", sep = "")
      file2 <- destFolder(file2, values$wdTall)
      file3 <- paste("BoxPlot-", sys.time(), ".png", sep = "")
      file3 <- destFolder(file3, values$wdTall)
      file4 <- paste("Positive-", sys.time(), ".png", sep = "")
      file4 <- destFolder(file4, values$wdTall)
      file5 <- paste("Negative-", sys.time(), ".png", sep = "")
      file5 <- destFolder(file5, values$wdTall)

      plot2png(values$sentimentPieChart, filename = file1, type = "plotly", dpi = values$dpi, height = values$h)
      plot2png(
        values$sentimentDensityPlot,
        filename = file2,
        type = "plotly", dpi = values$dpi, height = values$h
      )
      plot2png(values$sentimentBoxPlot, filename = file3, type = "plotly", dpi = values$dpi, height = values$h)
      plot2png(
        values$docPolPlots$positive,
        filename = file4,
        type = "plotly", dpi = values$dpi, height = values$h
      )
      plot2png(
        values$docPolPlots$negative,
        filename = file5,
        type = "plotly", dpi = values$dpi, height = values$h
      )

      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report

  observeEvent(input$d_polDetReport, {
    if (!is.null(values$docPolarityOverallData)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "PolarityDetection"

      Gem <- values$d_polDet_Gemini %>% string_to_sentence_df()

      list_df <- list(Gem, values$docPolarityOverallData)
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      files <- c(
        "PieChart.png",
        "DensDensity.png",
        "BoxPlot.png",
        "Positive.png",
        "Negative.png"
      )
      values$filePieChart <- plot2png(
        values$sentimentPieChart,
        filename = files[1],
        type = "plotly", dpi = values$report_dpi, height = values$h
      )
      values$fileDensityPlot <- plot2png(
        values$sentimentDensityPlot,
        filename = files[2],
        type = "plotly", dpi = values$report_dpi, height = values$h
      )
      values$fileBoxPlot <- plot2png(
        values$sentimentBoxPlot,
        filename = files[3],
        type = "plotly", dpi = values$report_dpi, height = values$h
      )
      values$filedocPolPos <- plot2png(
        values$docPolPlots$positive,
        filename = files[4],
        type = "plotly", dpi = values$report_dpi, height = values$h
      )
      values$filedocPolNeg <- plot2png(
        values$docPolPlots$negative,
        filename = files[5],
        type = "plotly", dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$filePieChart, res$col),
        c(sheetname = res$sheetname, values$fileDensityPlot, res$col),
        c(sheetname = res$sheetname, values$fileBoxPlot, res$col),
        c(sheetname = res$sheetname, values$filedocPolPos, res$col),
        c(sheetname = res$sheetname, values$filedocPolNeg, res$col)
      )
      popUp(title = "Polarity Detection Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Emotion Detection ----

  docEmotionEstim <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$d_emoApply
    },
    valueExpr = {
      choices <- c(
        "english",
        "italian",
        "french",
        "german",
        "spanish",
        "afrikaans",
        "arabic",
        "armenian",
        "basque",
        "belarusian",
        "bulgarian",
        "catalan",
        "chinese",
        "croatian",
        "czech",
        "danish",
        "dutch",
        "estonian",
        "finnish",
        "galician",
        "greek",
        "hebrew",
        "hindi",
        "hungarian",
        "indonesian",
        "irish",
        "japanese",
        "korean",
        "latin",
        "latvian",
        "lithuanian",
        "maltese",
        "marathi",
        "norwegian",
        "persian",
        "polish",
        "portuguese",
        "romanian",
        "russian",
        "serbian",
        "slovak",
        "slovenian",
        "swedish",
        "tamil",
        "telugu",
        "turkish",
        "ukrainian",
        "urdu",
        "uyghur",
        "vietnamese"
      )
      if (values$language %in% choices) {
        ## check to verify if groups exist or not
        if (
          input$groupEmotion == "doc_id" &
            "ungroupDoc_id" %in% names(values$dfTag)
        ) {
          values$docEmotionData <- emotionAnalysis(
            backToOriginalGroups(values$dfTag) %>% filter(docSelected),
            language = values$language
          )
        } else {
          values$docEmotionData <- emotionAnalysis(
            values$dfTag %>% filter(docSelected),
            language = values$language
          )
        }

        if (!is.list(values$docEmotionData)) return(NULL)

        values$emotionBarChart <- emotionBarChart(
          values$docEmotionData$corpus_emotions
        )
        values$emotionRadarPlot <- emotionRadarPlot(
          values$docEmotionData$corpus_emotions
        )
        values$emotionHeatmap <- emotionHeatmap(
          values$docEmotionData$doc_emotions_long
        )

        # Prepare table data
        emotion_names <- c(
          "anger", "anticipation", "disgust", "fear",
          "joy", "sadness", "surprise", "trust"
        )
        values$docEmotionTableData <- values$docEmotionData$doc_emotions %>%
          select(doc_id, all_of(emotion_names), total) %>%
          rename(Total = total)
      }
    }
  )

  output$d_emoBarPlot <- renderPlotly({
    docEmotionEstim()
    values$emotionBarChart
  })

  output$d_emoRadarPlot <- renderPlotly({
    docEmotionEstim()
    values$emotionRadarPlot
  })

  output$d_emoWordPlot <- renderPlotly({
    docEmotionEstim()
    req(values$docEmotionData)
    emotion_sel <- input$d_emoSelectEmotion
    if (is.null(emotion_sel)) emotion_sel <- "joy"
    emotionWordPlot(values$docEmotionData$word_emotions, emotion_sel, n = 10)
  })

  output$d_emoHeatmap <- renderPlotly({
    docEmotionEstim()
    values$emotionHeatmap
  })

  output$d_emo_GeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "Gemini AI", content = values$d_emo_Gemini, values)
  })

  output$d_emoTable <- renderDT(server = FALSE, {
    docEmotionEstim()
    DTformat(
      values$docEmotionTableData,
      filename = "DocEmotion",
      numeric = 2:9,
      round = 0
    )
  })

  observeEvent(
    eventExpr = {
      input$d_emoExport
    },
    handlerExpr = {
      file1 <- paste("EmotionDistribution-", sys.time(), ".png", sep = "")
      file1 <- destFolder(file1, values$wdTall)
      file2 <- paste("EmotionRadar-", sys.time(), ".png", sep = "")
      file2 <- destFolder(file2, values$wdTall)
      file3 <- paste("EmotionHeatmap-", sys.time(), ".png", sep = "")
      file3 <- destFolder(file3, values$wdTall)

      plot2png(values$emotionBarChart, filename = file1, type = "plotly", dpi = values$dpi, height = values$h)
      plot2png(values$emotionRadarPlot, filename = file2, type = "plotly", dpi = values$dpi, height = values$h)
      plot2png(values$emotionHeatmap, filename = file3, type = "plotly", dpi = values$dpi, height = values$h)

      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Emotion Report
  observeEvent(input$d_emoReport, {
    if (!is.null(values$docEmotionTableData)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "EmotionAnalysis"

      Gem <- values$d_emo_Gemini %>% string_to_sentence_df()

      list_df <- list(Gem, values$docEmotionTableData)
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      files <- c(
        "EmotionDistribution.png",
        "EmotionRadar.png",
        "EmotionHeatmap.png"
      )
      values$fileEmotionBar <- plot2png(
        values$emotionBarChart,
        filename = files[1],
        type = "plotly", dpi = values$report_dpi, height = values$h
      )
      values$fileEmotionRadar <- plot2png(
        values$emotionRadarPlot,
        filename = files[2],
        type = "plotly", dpi = values$report_dpi, height = values$h
      )
      values$fileEmotionHeatmap <- plot2png(
        values$emotionHeatmap,
        filename = files[3],
        type = "plotly", dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileEmotionBar, res$col),
        c(sheetname = res$sheetname, values$fileEmotionRadar, res$col),
        c(sheetname = res$sheetname, values$fileEmotionHeatmap, res$col)
      )
      popUp(title = "Emotion Detection Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Abstractive Summarization ----

  output$optionsAbstractive <- renderUI({
    selectizeInput(
      inputId = "Abst_document_selection",
      label = "Select Document",
      choices = values$dfTag %>%
        dplyr::filter(docSelected) %>%
        select(doc_id) %>%
        unique() %>%
        pull(doc_id),
      multiple = FALSE,
      width = "100%"
    )
  })

  output$abstractivePromptUI <- renderUI({
    textAreaInput(
      inputId = "abstractivePrompt",
      label = "Provide any additional prompts for the AI Assistant:",
      placeholder = "Example: Provide the summary in bullet points; or Provide the summary in Italian language; etc.",
      value = values$abstractivePrompt,
      rows = 8,
      width = "100%"
    )
  })

  observeEvent(input$d_abstractiveApply, {
    values$abstractivePrompt <- input$abstractivePrompt
    values$abstractiveSumm <- "\u231b Thinking..."

    # Snapshot inputs for the future worker
    dfTag <- values$dfTag
    abstractivePrompt <- input$abstractivePrompt
    doc_id <- input$Abst_document_selection
    nL <- input$summaryLength
    model <- values$gemini_api_model
    api_key <- Sys.getenv("GEMINI_API_KEY", unset = NA)

    # Sync: prepare the prompt (same logic as abstractive_summary, without gemini_ai)
    doc_data <- dfTag %>%
      dplyr::filter(doc_id == !!doc_id) %>%
      rebuild_documents()

    if (nrow(doc_data) == 0) {
      values$abstractiveSumm <- paste("Document", doc_id, "not found")
      return()
    }

    doc <- doc_data %>% dplyr::pull(text)
    if (length(doc) > 1) doc <- doc[1]

    if (is.na(doc) || nchar(trimws(doc)) == 0) {
      values$abstractiveSumm <- paste("Document", doc_id, "contains no text content")
      return()
    }

    n_tokens <- estimate_gemini_tokens(doc)
    if (n_tokens > 16384) {
      values$abstractiveSumm <- paste(
        "Document", doc_id, "too long (", n_tokens, "tokens > 16384 limit)"
      )
      return()
    }

    if (is.na(api_key)) {
      values$abstractiveSumm <- "\u26a0\ufe0f API key not set. Please configure it in Settings."
      return()
    }

    prompt <- paste0(
      "Create a comprehensive abstractive summary of the following text. ",
      "Requirements:\n",
      "- Capture all main points and key details\n",
      "- Maintain clarity and readability\n",
      "- Preserve important context and nuances\n",
      "- Target length: approximately ", nL, " words\n",
      "- Use clear, concise language\n",
      "- Maintain the original tone when appropriate\n\n",
      "Text to summarize:\n\n", doc
    )
    if (!is.null(abstractivePrompt) && nchar(trimws(abstractivePrompt)) > 0) {
      prompt <- paste(abstractivePrompt, prompt, sep = "\n\n")
    }

    # Async: only the API call runs in background
    promises::future_promise({
      gemini_ai(
        image = NULL, prompt = prompt, model = model,
        type = "text", api_key = api_key, outputSize = "medium"
      )
    }, seed = TRUE) %...>%
      (function(result) {
        values$abstractiveSumm <- result
      }) %...!%
      (function(err) {
        values$abstractiveSumm <- paste("\u26a0\ufe0f Error:", conditionMessage(err))
      })

    NULL
  })

  output$summaryData <- renderUI({
    req(values$abstractiveSumm)
    HTML(
      create_abstract_box(
        gemini_to_html(values$abstractiveSumm, type = "summary")
      )
    )
  })

  output$documentData2 <- renderUI({
    req(input$d_abstractiveApply)
    ## collapse sentences into paragraphs
    df_paragraphs <- values$dfTag %>%
      filter(doc_id == !!input$Abst_document_selection) %>%
      select(doc_id, paragraph_id, sentence_id, sentence) %>%
      distinct() %>%
      group_by(doc_id, paragraph_id) %>%
      summarise(Paragraph = paste(sentence, collapse = "\n ")) %>%
      ungroup()

    HTML(create_document_box(
      df_paragraphs,
      input$Abst_document_selection,
      summarization_type = "abstractive"
    ))
  })

  ## Report

  observeEvent(input$d_astractiveReport, {
    if (!is.null(values$abstractiveSumm)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "AbstractiveSummarization"

      list_df <- list(
        data.frame(
          Abstract = paste0(
            "Document: ",
            input$Abst_document_selection,
            "\n\n Abstract:\n\n",
            values$abstractiveSumm
          )
        )
      )
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      popUp(title = "Summarization Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Extractive Summarization ----

  output$optionsUnitSummarization <- renderUI({
    selectInput(
      inputId = "unit_selection",
      label = "Summarize ",
      choices = c("Groups", "Documents"),
      selected = "Documents",
      multiple = FALSE,
      width = "100%"
    )
  })

  output$optionsSummarization <- renderUI({
    selectizeInput(
      inputId = "document_selection",
      label = ifelse(
        input$unit_selection == "Documents",
        "Select Document",
        "Select Group"
      ),
      choices = ids(
        values$dfTag %>%
          group_by(doc_id) %>%
          mutate(n_sentences = max(sentence_id)) %>%
          ungroup() %>%
          dplyr::filter(docSelected, n_sentences > 1),
        type = input$unit_selection
      ),
      multiple = FALSE,
      width = "100%"
    )
  })

  docExtraction <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$d_summarizationApply
    },
    valueExpr = {
      values$docExtracted <- textrankDocument(
        values$dfTag,
        id = input$document_selection
      )
      values$docExtraction <- abstractingDocument(
        values$docExtracted$s,
        n = "5%",
        id = input$document_selection
      )
      values$docExtraction$sentences <- values$docExtracted$sentences %>%
        rename(S_id = textrank_id, Ranking = textrank)
    }
  )

  output$sliderAbstractData <- renderUI({
    docExtraction()
    choices <- c(
      "More Concise",
      paste0(seq(from = 10, to = 95, by = 5), "%"),
      "Less Concise"
    )
    sliderTextInput(
      inputId = "sliderAbstractData",
      label = "Summarization",
      choices = choices,
      selected = choices[1],
      grid = FALSE,
      hide_min_max = FALSE,
      animate = TRUE
    )
  })

  docExtractionVisualize <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$sliderAbstractData
    },
    valueExpr = {
      req(values$docExtracted)
      values$docExtraction <- abstractingDocument(
        values$docExtracted$s,
        n = input$sliderAbstractData,
        id = input$document_selection
      )
      values$docExtraction$sentences <- values$docExtracted$sentences %>%
        rename(S_id = textrank_id, Ranking = textrank)
    }
  )

  # Styled output for abstract
  output$abstractData <- renderUI({
    docExtractionVisualize()
    HTML(create_abstract_box(values$docExtraction$abstract))
  })

  output$RelSentData <- renderDT(server = FALSE, {
    docExtraction()
    DTformat(
      values$docExtraction$sentences,
      nrow = 10,
      size = "85%",
      title = paste0("Doc_id: ", input$document_selection),
      left = 1:2,
      numeric = 3,
      round = 4
    )
  })

  # Styled output for full document with highlighting
  output$documentData <- renderUI({
    docExtraction()
    HTML(create_document_box(
      values$docExtraction$document,
      input$document_selection,
      summarization_type = "extractive"
    ))
  })

  ## Report

  observeEvent(input$d_summarizationReport, {
    if (!is.null(values$docExtraction$sentences)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "ExtractiveSummarization"

      values$docExtraction$abstractData <- data.frame(
        "Abstract" = values$docExtraction$abstract
      )
      values$docExtraction$abstractData <- values$docExtraction$abstractData %>%
        mutate(Abstract = gsub("<.*?>", "", Abstract))

      list_df <- list(
        as.data.frame(values$docExtraction$abstractData),
        values$docExtraction$sentences,
        values$docExtraction$document
      )
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      popUp(title = "Summarization Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })
}

overviewUI <- function() {
  ### OVERVIEW ----
  tabItem(
    tabName = "overview",
    fluidPage(
      fluidRow(
        tabsetPanel(
          type = "tabs",
          id = "maininfo",
          # OVERVIEW TAB - NEW LAYOUT
          tabPanel(
            title = "Overview", icon = icon("chart-column"),
            br(),
            fluidRow(
              column(
                11,
                h2(icon("chart-pie"), strong("Overview"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
              ),
              div(
                title = t_report,
                style = style_opt,
                column(
                  1,
                  do.call(
                    "actionButton",
                    c(
                      report_bttn,
                      list(
                        inputId = "overviewReport"
                      )
                    )
                  )
                )
              )
            ),

            # GROUP 1: CORPUS SIZE & STRUCTURE
            fluidRow(
              column(
                12,
                div(
                  class = "group-header group-header-green",
                  icon("folder-open", class = "fa-lg"),
                  " Corpus Size & Structure"
                )
              )
            ),
            fluidRow(
              class = "overview-row",
              tagList(
                useShinyjs(),
                column(
                  2,
                  offset = 0,
                  div(
                    id = "clickbox1",
                    title = "Number of Documents",
                    valueBoxOutput("nDoc", width = NULL)
                  )
                ),
                column(
                  2,
                  offset = 0,
                  div(
                    id = "clickbox4",
                    title = "Number of Sentences",
                    valueBoxOutput("nSentences", width = NULL)
                  )
                ),
                column(
                  3,
                  offset = 0,
                  div(
                    id = "clickbox8",
                    title = "Number of Tokens",
                    valueBoxOutput("nTokens", width = NULL)
                  )
                ),
                column(
                  3,
                  offset = 0,
                  div(
                    id = "clickbox7",
                    title = "Number of Types",
                    valueBoxOutput("nDictionary", width = NULL)
                  )
                ),
                column(
                  2,
                  offset = 0,
                  div(
                    id = "clickbox9",
                    title = "Number of Lemma",
                    valueBoxOutput("nLemmas", width = NULL)
                  )
                )
              )
            ),

            # GROUP 2: AVERAGE LENGTH METRICS
            fluidRow(
              column(
                12,
                div(
                  class = "group-header group-header-blue",
                  style = "margin-top: 20px;",
                  icon("text-height", class = "fa-lg"),
                  " Average Length Metrics"
                )
              )
            ),
            fluidRow(
              class = "overview-row",
              tagList(
                column(
                  3,
                  div(
                    id = "clickbox2",
                    title = "Average Document's Length by Characters",
                    valueBoxOutput("avgDocLengthChar", width = NULL)
                  )
                ),
                column(
                  3,
                  div(
                    id = "clickbox3",
                    title = "Average Document's Length by Tokens",
                    valueBoxOutput("avgDocLengthTokens", width = NULL)
                  )
                ),
                column(
                  3,
                  div(
                    id = "clickbox5",
                    title = "Average Sentence's Length by Characters",
                    valueBoxOutput("avgSentLengthChar", width = NULL)
                  )
                ),
                column(
                  3,
                  div(
                    id = "clickbox6",
                    title = "Average Sentence's Length by Tokens",
                    valueBoxOutput("avgSentLengthTokens", width = NULL)
                  )
                )
              )
            ),

            # GROUP 3: LEXICAL METRICS (TUTTE E 7 SULLA STESSA RIGA)
            fluidRow(
              column(
                12,
                div(
                  class = "group-header group-header-orange",
                  style = "margin-top: 20px;",
                  icon("bar-chart", class = "fa-lg"),
                  " Lexical Metrics"
                )
              )
            ),
            div(
              class = "lexical-row",
              tagList(
                div(
                  id = "clickbox10",
                  title = "Type-Token Ratio",
                  valueBoxOutput("TTR", width = NULL)
                ),
                div(
                  id = "clickbox11",
                  title = "Percentage of Hapax Legomena",
                  valueBoxOutput("hapax", width = NULL)
                ),
                div(
                  id = "clickbox12",
                  title = "Guiraud Index",
                  valueBoxOutput("guiraud", width = NULL)
                ),
                div(
                  id = "clickbox13",
                  title = "Lexical Density",
                  valueBoxOutput("lexicalDensity", width = NULL)
                ),
                div(
                  id = "clickbox14",
                  title = "Nominal Ratio",
                  valueBoxOutput("nominalRatio", width = NULL)
                ),
                div(
                  id = "clickbox15",
                  title = "Gini Index",
                  valueBoxOutput("giniIndex", width = NULL)
                ),
                div(
                  id = "clickbox16",
                  title = "Yule's K Index",
                  valueBoxOutput("yuleK", width = NULL)
                )
              )
            )
          ),
          tabPanel(
            title = "Table", icon = icon("table"),
            br(),
            div(
              shinycssloaders::withSpinner(
                DT::DTOutput(outputId = "overviewData", width = 700),
                color = getOption("spinner.color", default = "#4F7942")
              ),
              align = "center"
            )
          ),
          tabPanel(
            title = "Vocabulary", icon = icon("book"),
            br(),
            column(
              12,
              div(
                shinycssloaders::withSpinner(
                  DT::DTOutput(outputId = "dictionaryData", width = 700),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              ),
              align = "center"
            )
            # ,column(1)
          ),
          tabPanel(
            title = "TF-IDF", icon = icon("table"),
            br(),
            column(
              12,
              div(
                shinycssloaders::withSpinner(
                  DT::DTOutput(outputId = "tfidfData", width = 700),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              ),
              align = "center"
            )
          ),
          tabPanel(
            title = "WordCloud", icon = icon("cloud"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  8,
                  h2(icon("cloud"), strong("WordCloud"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
                ),
                div(
                  title = t_run,
                  column(
                    1,
                    do.call(
                      "actionButton",
                      c(
                        run_bttn,
                        list(
                          inputId = "wcApply"
                        )
                      )
                    )
                  )
                ),
                div(
                  title = t_export,
                  column(
                    1,
                    do.call(
                      "actionButton",
                      c(
                        export_bttn,
                        list(
                          inputId = "wcSave"
                        )
                      )
                    ),
                  )
                ),
                div(
                  title = t_report,
                  column(
                    1,
                    do.call(
                      "actionButton",
                      c(
                        report_bttn,
                        list(
                          inputId = "wcReport"
                        )
                      )
                    )
                  )
                ),
                div(
                  column(
                    1,
                    dropdown(
                      h4(strong("Options: ")),
                      br(),

                      # Main Configuration
                      div(
                        class = "config-section",
                        div(
                          class = "config-section-header",
                          icon("gear"),
                          "Main Configuration"
                        ),
                        numericInput(
                          "nWC",
                          label = "Number of words",
                          value = 100,
                          min = 10,
                          max = 500,
                          step = 1
                        ),
                        numericInput(
                          "labelsizeWC",
                          label = "Text Size",
                          value = 5,
                          min = 1,
                          max = 20,
                          step = 1
                        )
                      ),
                      style = "gradient",
                      right = TRUE,
                      animate = TRUE,
                      circle = TRUE,
                      tooltip = tooltipOptions(title = "Options"),
                      icon = icon("sliders", lib = "font-awesome"),
                      width = "300px"
                    )
                  ),
                  style = style_opt
                )
              ),
              fluidRow(
                br(),
                shinycssloaders::withSpinner(
                  plotOutput(
                    outputId = "wordcloudPlot",
                    height = "700px",
                    width = "100%"
                  ),
                  color = getOption("spinner.color", default = "#4F7942")
                )
              )
            )
          ),
          tabPanel(
            title = "Frequency", icon = icon("chart-bar"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  8,
                  h2(icon("chart-bar"), strong("Word Frequency by PoS"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
                ),
                div(
                  title = t_run,
                  column(
                    1,
                    do.call(
                      "actionButton",
                      c(
                        run_bttn,
                        list(
                          inputId = "wFreqApply"
                        )
                      )
                    )
                  )
                ),
                div(
                  title = t_export,
                  column(
                    1,
                    do.call(
                      "actionButton",
                      c(
                        export_bttn,
                        list(
                          inputId = "wFreqExport"
                        )
                      )
                    ),
                  )
                ),
                div(
                  title = t_report,
                  column(
                    1,
                    do.call(
                      "actionButton",
                      c(
                        report_bttn,
                        list(
                          inputId = "wFreqReport"
                        )
                      )
                    )
                  )
                ),
                div(
                  column(
                    1,
                    dropdown(
                      h4(strong("Options: ")),
                      br(),

                      # Main Configuration
                      div(
                        class = "config-section",
                        div(
                          class = "config-section-header",
                          icon("gear"),
                          "Main Configuration"
                        ),
                        numericInput(
                          "wFreqN",
                          label = "Number of words",
                          value = 20,
                          min = 1,
                          step = 1
                        ),
                        uiOutput("posSelectionFreq"),
                        helpText("Select a PoS category (upos) from the options before running the analysis.")
                      ),
                      style = "gradient",
                      right = TRUE,
                      animate = TRUE,
                      circle = TRUE,
                      tooltip = tooltipOptions(title = "Options"),
                      icon = icon("sliders", lib = "font-awesome"),
                      width = "300px"
                    )
                  ),
                  style = style_opt
                )
              ),
              fluidRow(
                column(
                  6,
                  shinycssloaders::withSpinner(
                    plotlyOutput(
                      outputId = "wFreqPlot",
                      height = "75vh",
                      width = "98.9%"
                    ),
                    color = getOption("spinner.color", default = "#4F7942")
                  )
                ),
                column(
                  6,
                  br(),
                  shinycssloaders::withSpinner(
                    DT::DTOutput("wFreqTable", width = "90.0%"),
                    color = getOption("spinner.color", default = "#4F7942")
                  ),
                  align = "center"
                )
              )
            )
          ),
          tabPanel(
            title = "Morphological Features", icon = icon("language"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  h2(icon("puzzle-piece"), strong("Morphological Features"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;"),
                  p(
                    "Distribution of morphological features extracted from the Universal Dependencies annotation.",
                    style = "text-align: center; color: #666; margin-bottom: 20px;"
                  )
                )
              ),
              fluidRow(
                column(
                  12,
                  uiOutput("morphFeatureDesc")
                )
              ),
              fluidRow(
                column(
                  3,
                  selectInput(
                    "morphFeature",
                    "Select Feature:",
                    choices = c(
                      "Tense" = "Tense",
                      "Mood" = "Mood",
                      "Number" = "Number",
                      "Person" = "Person",
                      "VerbForm" = "VerbForm",
                      "Degree" = "Degree",
                      "Gender" = "Gender",
                      "Case" = "Case",
                      "Voice" = "Voice",
                      "Definite" = "Definite",
                      "PronType" = "PronType"
                    ),
                    selected = "Tense"
                  )
                ),
                column(
                  9,
                  shinycssloaders::withSpinner(
                    plotlyOutput(
                      outputId = "morphFeaturePlot",
                      height = "55vh",
                      width = "98.9%"
                    ),
                    color = getOption("spinner.color", default = "#4F7942")
                  )
                )
              ),
              fluidRow(
                column(
                  12,
                  br(),
                  h5(strong("Cross-tabulation: Feature Values by Part of Speech"),
                    style = "color: #555; margin-bottom: 10px;"),
                  shinycssloaders::withSpinner(
                    DT::DTOutput("morphFeatureTable"),
                    color = getOption("spinner.color", default = "#4F7942")
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "Dependency Tree", icon = icon("sitemap"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  9,
                  h2(icon("diagram-project"), strong("Dependency Tree Viewer"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
                ),
                div(
                  title = t_export,
                  column(
                    1,
                    do.call(
                      "actionButton",
                      c(
                        export_bttn,
                        list(inputId = "depTreeExport")
                      )
                    )
                  )
                ),
                div(
                  title = t_report,
                  column(
                    1,
                    do.call(
                      "actionButton",
                      c(
                        report_bttn,
                        list(inputId = "depTreeReport")
                      )
                    )
                  )
                )
              ),
              fluidRow(
                column(
                  3,
                  textInput(
                    "depTreeSearch",
                    "Search word:",
                    value = "",
                    placeholder = "Type a word to filter sentences..."
                  )
                ),
                column(
                  3,
                  uiOutput("depTreeDocSelect")
                ),
                column(
                  3,
                  uiOutput("depTreeSentSelect")
                ),
                column(
                  3,
                  div(
                    style = "margin-top: 25px;",
                    actionBttn(
                      inputId = "depTreeApply",
                      label = "Show Tree",
                      style = "bordered",
                      color = "success",
                      size = "sm",
                      icon = icon("sitemap")
                    )
                  )
                )
              ),
              fluidRow(
                column(
                  12,
                  uiOutput("depTreeSentenceText")
                )
              ),
              fluidRow(
                column(
                  12,
                  br(),
                  shinycssloaders::withSpinner(
                    uiOutput("depTreePlotUI"),
                    color = getOption("spinner.color", default = "#4F7942")
                  )
                )
              ),
              fluidRow(
                column(
                  12,
                  br(),
                  h5(strong("Token Details"), style = "color: #555; margin-bottom: 10px;"),
                  shinycssloaders::withSpinner(
                    DT::DTOutput("depTreeTable"),
                    color = getOption("spinner.color", default = "#4F7942")
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "TALL AI", icon = icon("robot"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  br(),
                  shinycssloaders::withSpinner(
                    htmlOutput("OverviewGeminiUI"),
                    caption = HTML("<br><strong>Thinking...</strong>"),
                    image = "ai_small2.gif",
                    color = "#4F7942"
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "Info & References", icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  br(),
                  HTML(infoTexts$overview)
                ),
                column(1)
              )
            )
          )
        ) # , align="center"
      )
    )
  )
}

overviewServer <- function(input, output, session, values, statsValues) {
  ## BOX ----

  # Eagerly recompute value box indices whenever dfTag changes (e.g. after filtering)
  observe({
    req(values$dfTag)
    req("docSelected" %in% names(values$dfTag))
    values$vb <- valueBoxesIndices(values$dfTag %>% filter(docSelected))
    values$VbData <- data.frame(
      Description = c(
        "Documents",
        "Tokens",
        "Types",
        "Lemma",
        "Sentences",
        "Docs Avg Length in Chars",
        "Docs SD Length in Chars",
        "Doc Avg Length in Tokens",
        "Doc SD Length in Tokens",
        "Sent Avg Length in Tokens",
        "Sent SD Length in Tokens",
        "Sent Avg Length in Chars",
        "Sent SD Length in Chars",
        "TTR (%)",
        "Hapax (%)",
        "Guiraud Index",
        "Lexical Density",
        "Nominal Ratio",
        "Gini Index",
        "Yule's K"
      ),
      Values = unlist(values$vb)
    )
  })

  # ============================================
  # CORPUS SIZE & STRUCTURE (GREEN)
  # ============================================

  #### box1 - Documents ---------------
  output$nDoc <- renderValueBox({
    valueBox(
      value = strong(values$vb$nDoc),
      subtitle = "Documents",
      icon = icon("file-text", class = "fa-2x", lib = "font-awesome"),
      color = "green",
      width = NULL
    )
  })

  #### box4 - Sentences ---------------
  output$nSentences <- renderValueBox({
    valueBox(
      value = strong(values$vb$nSentences),
      subtitle = "Sentences",
      icon = icon("align-left", class = "fa-2x"),
      color = "green",
      width = NULL
    )
  })

  #### box8 - Tokens ---------------
  output$nTokens <- renderValueBox({
    valueBox(
      value = strong(values$vb$nTokens),
      subtitle = "Tokens",
      icon = icon("font", class = "fa-2x"),
      color = "green",
      width = NULL
    )
  })

  #### box7 - Types ----------------
  output$nDictionary <- renderValueBox({
    valueBox(
      value = strong(values$vb$nDictionary),
      subtitle = "Types",
      icon = icon("rectangle-list", class = "fa-2x"),
      color = "green",
      width = NULL
    )
  })

  #### box9 - Lemma ---------------
  output$nLemmas <- renderValueBox({
    valueBox(
      value = strong(values$vb$nLemmas),
      subtitle = "Lemma",
      icon = icon("book", class = "fa-2x", lib = "font-awesome"),
      color = "green",
      width = NULL
    )
  })

  # ============================================
  # AVERAGE LENGTH METRICS (BLUE)
  # ============================================

  #### box2 - Doc Avg Length in Chars ---------------
  output$avgDocLengthChar <- renderValueBox({
    valueBox(
      value = strong(paste0(
        values$vb$avgDocLengthChars,
        " ± ",
        values$vb$avgDocLengthCharsSD
      )),
      subtitle = "Doc Avg Length in Chars",
      icon = icon("rectangle-list", class = "fa-2x"),
      color = "blue",
      width = NULL
    )
  })

  #### box3 - Doc Avg Length in Tokens ------------
  output$avgDocLengthTokens <- renderValueBox({
    valueBox(
      value = strong(paste0(
        values$vb$avgDocLengthTokens,
        " ± ",
        values$vb$avgDocLengthTokensSD
      )),
      subtitle = "Doc Avg Length in Tokens",
      icon = icon("text-height", class = "fa-2x", lib = "font-awesome"),
      color = "blue",
      width = NULL
    )
  })

  #### box5 - Sent Avg Length in Chars --------------------
  output$avgSentLengthChar <- renderValueBox({
    valueBox(
      value = strong(paste0(
        values$vb$avgSentLengthChars,
        " ± ",
        values$vb$avgSentLengthCharsSD
      )),
      subtitle = "Sent Avg Length in Chars",
      icon = icon("text-width", class = "fa-2x", lib = "font-awesome"),
      color = "blue",
      width = NULL
    )
  })

  #### box6 - Sent Avg Length in Tokens -------------
  output$avgSentLengthTokens <- renderValueBox({
    valueBox(
      value = strong(paste0(
        values$vb$avgSentLengthTokens,
        " ± ",
        values$vb$avgSentLengthTokensSD
      )),
      subtitle = "Sent Avg Length in Tokens",
      icon = icon("align-justify", class = "fa-2x"),
      color = "blue",
      width = NULL
    )
  })

  # ============================================
  # LEXICAL METRICS - TUTTE ARANCIONE
  # ============================================

  #### box10 - TTR ------------------
  output$TTR <- renderValueBox({
    valueBox(
      value = strong(values$vb$TTR),
      subtitle = "TTR (%)",
      icon = icon("percent", class = "fa-2x", lib = "font-awesome"),
      color = "orange",
      width = NULL
    )
  })

  #### box11 - Hapax ------
  output$hapax <- renderValueBox({
    valueBox(
      value = strong(round(values$vb$hapax, 0)),
      subtitle = "Hapax (%)",
      icon = icon("star", class = "fa-2x", lib = "font-awesome"),
      color = "orange",
      width = NULL
    )
  })

  #### box12 - Guiraud -------
  output$guiraud <- renderValueBox({
    valueBox(
      value = strong(values$vb$guiraud),
      subtitle = "Guiraud Index",
      icon = icon("bar-chart", class = "fa-2x", lib = "font-awesome"),
      color = "orange",
      width = NULL
    )
  })

  #### box13 - Lexical Density -------
  output$lexicalDensity <- renderValueBox({
    valueBox(
      value = strong(round(values$vb$lexical_density, 1)),
      subtitle = "Lexical Density",
      icon = icon("tachometer", class = "fa-2x", lib = "font-awesome"),
      color = "orange",
      width = NULL
    )
  })

  #### box14 - Nominal Ratio -------
  output$nominalRatio <- renderValueBox({
    valueBox(
      value = strong(round(values$vb$nominal_ratio, 2)),
      subtitle = "Nominal Ratio",
      icon = icon("balance-scale", class = "fa-2x", lib = "font-awesome"),
      color = "orange",
      width = NULL
    )
  })

  #### box15 - Gini Index -------
  output$giniIndex <- renderValueBox({
    valueBox(
      value = strong(round(values$vb$gini_index, 2)),
      subtitle = "Gini Index",
      icon = icon("signal", class = "fa-2x", lib = "font-awesome"),
      color = "orange",
      width = NULL
    )
  })

  #### box16 - Yule's K -------
  output$yuleK <- renderValueBox({
    valueBox(
      value = strong(round(values$vb$yule_k, 1)),
      subtitle = "Yule's K",
      icon = icon("line-chart", class = "fa-2x", lib = "font-awesome"),
      color = "orange",
      width = NULL
    )
  })
  ## Overview Table ----

  output$overviewData <- renderDT(server = FALSE, {
    DTformat(
      values$VbData,
      nrow = nrow(values$VbData),
      left = 1,
      right = 2,
      numeric = 2,
      pagelength = FALSE,
      dom = FALSE,
      size = "110%",
      filename = "Overview",
      col_to_remove = values$generalTerm
    )
  })

  output$OverviewGeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "Gemini AI", content = values$overviewGemini, values)
  })

  ## Report

  observeEvent(input$overviewReport, {
    if (!is.null(values$VbData)) {
      sheetname <- "Overview"
      list_df <- list(values$VbData)
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      values$wb <- res$wb
      popUp(title = "Overview Table", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }

    if (!is.null(values$dfTag) && "feats" %in% names(values$dfTag)) {
      morphFeatures <- c("Tense", "Mood", "Number", "Person", "VerbForm",
                          "Degree", "Gender", "Case", "Voice", "Definite", "PronType")
      morphData <- lapply(morphFeatures, function(feat) {
        vals <- parseMorphFeatures(values$dfTag$feats, feat)
        vals <- vals[!is.na(vals) & vals != ""]
        if (length(vals) > 0) {
          tab <- sort(table(vals), decreasing = TRUE)
          data.frame(Feature = feat, Value = names(tab), Count = as.integer(tab), stringsAsFactors = FALSE)
        }
      })
      morphData <- do.call(rbind, morphData)
      if (!is.null(morphData) && nrow(morphData) > 0) {
        list_df <- list(morphData)
        res <- addDataScreenWb(list_df, wb = values$wb, sheetname = "Morphological Features")
        values$wb <- res$wb
        values$myChoices <- sheets(values$wb)
      }
    }
  })

  ## WORDCLOUD ----

  wcData <- eventReactive(
    {
      input$wcApply
    },
    {
      N <- input$nWC # showing the first 100 lemma
      filtered <- LemmaSelection(values$dfTag) %>%
        dplyr::filter(docSelected)
      pos <- unique(filtered$upos)

      values$wcDfPlot <- freqByPos(
        filtered,
        term = values$generalTerm,
        pos = pos
      ) %>%
        slice_head(n = N) %>%
        rename(
          label = term,
          value = n
        )

      set.seed(values$random_seed)
      values$WCplot <- wordcloud(
        values$wcDfPlot,
        shape = "circle",
        rot_per = 0.2,
        eccentricity = 1.5,
        colors = sample(colorlist(), nrow(values$wcDfPlot), replace = T),
        seed = values$random_seed,
        max_size = input$labelsizeWC * 10
      )
    }
  )

  output$wordcloudPlot <- renderPlot({
    wcData()
    values$WCplot
  })

  ## export WordCloud button
  observeEvent(
    eventExpr = {
      input$wcSave
    },
    handlerExpr = {
      file <- paste("Wordcloud-", sys.time(), ".png", sep = "")
      file_path <- destFolder(file, values$wdTall)
      ggsave(
        filename = file_path,
        plot = values$WCplot,
        dpi = values$dpi,
        height = values$h,
        width = values$h * values$aspect,
        bg = "transparent"
      )

      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report
  observeEvent(input$wcReport, {
    if (!is.null(values$wcDfPlot)) {
      list_df <- list(values$wcDfPlot)
      list_plot <- list(values$WCplot)
      wb <- addSheetToReport(
        list_df,
        list_plot,
        sheetname = "WordCloud",
        wb = values$wb,
        dpi = values$report_dpi
      )
      values$wb <- wb
      popUp(
        title = paste0("WordCloud"),
        type = "success"
      )
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## VOCABULARY ----
  output$dictionaryData <- renderDT(server = FALSE, {
    # dictionary()
    values$dictFreq <- vocabulary(values$dfTag, term = values$generalTerm)
    DTformat(
      values$dictFreq,
      left = c(1, 2),
      nrow = 15,
      pagelength = TRUE,
      filename = "Dictionary",
      dom = TRUE,
      size = "110%",
      col_to_remove = values$generalTerm
    )
  })

  ## TF-IDF Table----
  output$tfidfData <- renderDT(server = FALSE, {
    values$tfidfDATA <- tfidfTable(values$dfTag, term = values$generalTerm)
    DTformat(
      values$tfidfDATA,
      left = 1,
      numeric = 2,
      round = 4,
      size = "110%",
      col_to_remove = values$generalTerm
    )
  })

  ## Frequency List ----

  ## Words Frequency by PoS ----

  output$posSelectionFreq <- renderUI({
    selectInput(
      inputId = "posSelectionFreq",
      "PoS Tag:",
      choices = posTagAll(values$dfTag)$pos,
      selected = "NOUN"
    )
  })

  wFreq <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$wFreqApply
    },
    valueExpr = {
      req(input$posSelectionFreq)
      values$wFreq <- freqByPos(
        values$dfTag %>% filter(docSelected),
        term = values$generalTerm,
        pos = input$posSelectionFreq
      )
      values$wFreqPlotly <- freqPlotly(
        values$wFreq,
        x = "n",
        y = "term",
        n = input$wFreqN,
        xlabel = "Frequency",
        ylabel = input$posSelectionFreq,
        scale = "identity"
      )

      values$wFreqData <- values$wFreq %>%
        rename(
          Word = term,
          Frequency = n
        )
    }
  )

  output$wFreqPlot <- renderPlotly({
    wFreq()
    values$wFreqPlotly
  })

  output$wFreqTable <- renderDT(server = FALSE, {
    wFreq()
    DTformat(
      values$wFreqData,
      nrow = 15,
      left = 1,
      right = 2,
      round = 0,
      numeric = 2,
      filename = "WordsFreqList",
      dom = FALSE,
      size = "90%",
      filter = "none",
      col_to_remove = values$generalTerm
    )
  })

  observeEvent(
    eventExpr = {
      input$wFreqExport
    },
    handlerExpr = {
      file <- paste(
        "WordsFrequency-",
        input$posSelectionFreq,
        "-",
        sys.time(),
        ".png",
        sep = ""
      )
      file_path <- destFolder(file, values$wdTall)
      g = values$wFreq
      save(g, file = "progaggplot.rdata")
      values$wFreqGgplot <- freqGgplot(
        values$wFreq,
        x = 2,
        y = 1,
        n = input$wFreqN,
        title = paste0("Words Frequency by ", input$posSelectionFreq)
      )
      ggsave(
        filename = file_path,
        plot = values$wFreqGgplot,
        dpi = values$dpi,
        height = values$h,
        width = values$h * values$aspect,
        bg = "transparent"
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report
  observeEvent(input$wFreqReport, {
    if (!is.null(values$wFreq)) {
      values$wFreqGgplot <- freqGgplot(
        values$wFreq,
        x = 2,
        y = 1,
        n = input$wFreqN,
        title = paste0("Words Frequency by ", input$posSelectionFreq)
      )
      list_df <- list(values$wFreqData)
      list_plot <- list(values$wFreqGgplot)
      wb <- addSheetToReport(
        list_df,
        list_plot,
        sheetname = "WordsFreq",
        wb = values$wb,
        dpi = values$report_dpi
      )
      values$wb <- wb
      popUp(
        title = paste0("Words Frequency by-", input$posSelectionFreq),
        type = "success"
      )
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## PART OF SPEECH ----

  posFreq <- eventReactive(
    eventExpr = {
      input$posApply
    },
    valueExpr = {
      values$freqPOS <- values$dfTag %>%
        filter(docSelected) %>%
        dplyr::filter(!upos %in% c("PUNCT", "SYM", "NUM", "NGRAM_MERGED")) %>%
        group_by(upos) %>%
        count() %>%
        arrange(desc(n)) %>%
        rename(PoS = upos)

      values$posPlotly <- freqPlotly(
        values$freqPOS,
        x = "n",
        y = "PoS",
        n = nrow(values$freqPOS),
        xlabel = "Frequency",
        ylabel = "Part of Speech",
        scale = "identity"
      )

      values$freqPOSData <- values$freqPOS %>%
        rename(Frequency = n)
    }
  )

  output$posPlot <- renderPlotly({
    posFreq()
    values$posPlotly
  })

  output$posTable <- renderDT(server = FALSE, {
    posFreq()
    DTformat(
      values$freqPOSData,
      left = 1,
      right = 2,
      round = 0,
      numeric = 2,
      filename = "POSFreqList",
      dom = FALSE,
      size = "110%",
      col_to_remove = values$generalTerm
    )
  })

  observeEvent(
    eventExpr = {
      input$posExport
    },
    handlerExpr = {
      file <- paste("PoSFrequency-", sys.time(), ".png", sep = "")
      file <- destFolder(file, values$wdTall)
      values$posGgplot <- freqGgplot(
        data.frame(values$freqPOS),
        x = 2,
        y = 1,
        n = length(values$freqPOS$PoS),
        title = "PoS Frequency"
      )
      ggsave(
        filename = file,
        plot = values$posGgplot,
        dpi = values$dpi,
        height = values$h,
        width = values$h * values$aspect,
        bg = "transparent"
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report

  observeEvent(input$posReport, {
    values$posGgplot <- freqGgplot(
      data.frame(values$freqPOS),
      x = 2,
      y = 1,
      n = length(values$freqPOS$PoS),
      title = "PoS Tag Frequency"
    )
    if (!is.null(values$freqPOS)) {
      list_df <- list(values$freqPOSData)
      list_plot <- list(values$posGgplot)
      wb <- addSheetToReport(
        list_df,
        list_plot,
        sheetname = "PoSFreq",
        wb = values$wb,
        dpi = values$report_dpi
      )
      values$wb <- wb
      popUp(title = "PoS Tag Frequency", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Dependency Tree Viewer ----

  # Reactive: filtered data based on search word
  depTreeFiltered <- reactive({
    req(values$dfTag)
    df <- values$dfTag %>% filter(docSelected)
    search <- trimws(tolower(input$depTreeSearch))
    if (!is.null(search) && nchar(search) > 0) {
      # Find doc_id + sentence_id pairs containing the search word
      matching <- df %>%
        filter(tolower(token) == search | tolower(lemma) == search) %>%
        distinct(doc_id, sentence_id)
      df <- df %>% semi_join(matching, by = c("doc_id", "sentence_id"))
    }
    df
  })

  output$depTreeDocSelect <- renderUI({
    df <- depTreeFiltered()
    docs <- unique(df$doc_id)
    if (length(docs) == 0) {
      return(div(
        style = "margin-top: 25px; color: #856404; font-size: 13px;",
        icon("exclamation-triangle"), " No matching sentences found."
      ))
    }
    selectInput(
      "depTreeDoc",
      "Document:",
      choices = docs,
      selected = docs[1],
      width = "100%"
    )
  })

  output$depTreeSentSelect <- renderUI({
    req(input$depTreeDoc)
    df <- depTreeFiltered()
    sents <- df %>%
      filter(doc_id == input$depTreeDoc) %>%
      pull(sentence_id) %>%
      unique() %>%
      sort()
    if (length(sents) == 0) return(NULL)

    # Build preview labels: sentence_id + first words
    sent_labels <- sapply(sents, function(s) {
      tokens <- df %>%
        filter(doc_id == input$depTreeDoc, sentence_id == s) %>%
        pull(token)
      preview <- paste(head(tokens, 8), collapse = " ")
      if (length(tokens) > 8) preview <- paste0(preview, "...")
      paste0("S", s, ": ", preview)
    })
    names(sents) <- sent_labels

    selectInput(
      "depTreeSent",
      "Sentence:",
      choices = sents,
      selected = sents[1],
      width = "100%"
    )
  })

  depTreeData <- eventReactive(input$depTreeApply, {
    req(input$depTreeDoc, input$depTreeSent)
    values$dfTag %>%
      filter(
        doc_id == input$depTreeDoc,
        sentence_id == as.integer(input$depTreeSent),
        !upos %in% c("NGRAM_MERGED")
      ) %>%
      mutate(
        tid = as.integer(token_id),
        hid = as.integer(head_token_id)
      ) %>%
      select(tid, hid, token, lemma, upos, dep_rel)
  })

  output$depTreeSentenceText <- renderUI({
    df <- depTreeData()
    req(nrow(df) > 0)
    sentence <- paste(df$token, collapse = " ")
    div(
      style = "padding: 12px 15px; background-color: #f8f9fa; border-left: 4px solid #4F7942; border-radius: 4px; margin-top: 10px; font-size: 15px; line-height: 1.6;",
      icon("quote-left", style = "color: #4F7942; margin-right: 5px;"),
      sentence
    )
  })

  output$depTreePlotUI <- renderUI({
    df <- depTreeData()
    req(nrow(df) > 0)
    n_tokens <- nrow(df)
    # Min 80px per token, min total 800px
    plot_width <- max(800, n_tokens * 80)
    plot_height <- max(400, min(700, n_tokens * 12 + 200))
    div(
      style = paste0("overflow-x: auto; overflow-y: hidden; border: 1px solid #eee; border-radius: 8px; background: white; padding: 10px;"),
      plotOutput(
        outputId = "depTreePlot",
        height = paste0(plot_height, "px"),
        width = paste0(plot_width, "px")
      )
    )
  })

  output$depTreePlot <- renderPlot({
    df <- depTreeData()
    req(nrow(df) > 0)

    n <- nrow(df)
    # x positions = token order
    df$x <- seq_len(n)

    # Build arc data: from dependent to head
    arcs <- df %>%
      filter(hid > 0) %>%
      mutate(
        x_from = x,
        x_to = match(hid, df$tid)
      ) %>%
      filter(!is.na(x_to))

    if (nrow(arcs) == 0) return(NULL)

    # Arc height proportional to distance
    arcs$height <- abs(arcs$x_from - arcs$x_to) * 0.5
    arcs$xmid <- (arcs$x_from + arcs$x_to) / 2

    # Color by relation category
    rel_colors <- c(
      nsubj = "#E41A1C", `nsubj:pass` = "#E41A1C", csubj = "#E41A1C",
      obj = "#377EB8", dobj = "#377EB8", iobj = "#377EB8",
      amod = "#4DAF4A", nmod = "#4DAF4A", `nmod:poss` = "#4DAF4A",
      advmod = "#FF7F00",
      det = "#999999", case = "#999999",
      compound = "#984EA3", flat = "#984EA3", `flat:name` = "#984EA3",
      conj = "#A65628", cc = "#A65628",
      punct = "#CCCCCC",
      root = "#000000"
    )
    arcs$color <- ifelse(
      arcs$dep_rel %in% names(rel_colors),
      rel_colors[arcs$dep_rel],
      "#666666"
    )

    # PoS colors for tokens
    pos_colors <- c(
      NOUN = "#4F7942", PROPN = "#2E5A1E",
      VERB = "#E41A1C", AUX = "#FB9A99",
      ADJ = "#377EB8", ADV = "#FF7F00",
      DET = "#999999", ADP = "#999999",
      PRON = "#984EA3", CCONJ = "#A65628", SCONJ = "#A65628",
      PUNCT = "#CCCCCC", NUM = "#666666"
    )
    df$pos_color <- ifelse(
      df$upos %in% names(pos_colors),
      pos_colors[df$upos],
      "#666666"
    )

    # Build arc curves manually
    arc_curves <- do.call(rbind, lapply(seq_len(nrow(arcs)), function(i) {
      x1 <- arcs$x_from[i]
      x2 <- arcs$x_to[i]
      h <- arcs$height[i]
      t_seq <- seq(0, pi, length.out = 50)
      data.frame(
        arc_id = i,
        x = x1 + (x2 - x1) * (1 - cos(t_seq)) / 2,
        y = h * sin(t_seq),
        dep_rel = arcs$dep_rel[i],
        color = arcs$color[i],
        stringsAsFactors = FALSE
      )
    }))

    # Build plot
    p <- ggplot() +
      # Arcs
      geom_path(
        data = arc_curves,
        aes(x = x, y = y, group = arc_id, color = color),
        linewidth = 0.7,
        show.legend = FALSE
      ) +
      scale_color_identity() +
      # Arc labels (dep_rel)
      geom_text(
        data = arcs,
        aes(x = xmid, y = height + 0.15, label = dep_rel),
        size = 2.8,
        color = arcs$color,
        fontface = "bold"
      ) +
      # Token boxes
      geom_label(
        data = df,
        aes(x = x, y = -0.3, label = token),
        fill = "white",
        color = df$pos_color,
        label.size = 0.5,
        size = 3.5,
        fontface = "bold",
        label.padding = unit(0.25, "lines")
      ) +
      # PoS labels
      geom_text(
        data = df,
        aes(x = x, y = -0.7, label = upos),
        size = 2.5,
        color = df$pos_color
      ) +
      # Styling
      theme_void() +
      theme(
        plot.margin = margin(20, 20, 20, 20),
        plot.background = element_rect(fill = "white", color = NA)
      ) +
      coord_cartesian(
        ylim = c(-1, max(arcs$height) + 0.8),
        clip = "off"
      )

    values$depTreeGgplot <- p
    p
  })

  ## Dep Tree Export (PNG)
  observeEvent(input$depTreeExport, {
    req(values$depTreeGgplot)
    df <- depTreeData()
    n_tokens <- nrow(df)
    w <- min(49, max(10, n_tokens * 1.2))
    h <- max(5, min(10, n_tokens * 0.15 + 3))
    file <- paste("DependencyTree-", sys.time(), ".png", sep = "")
    file <- destFolder(file, values$wdTall)
    ggsave(
      filename = file,
      plot = values$depTreeGgplot,
      dpi = values$dpi,
      width = w,
      height = h,
      bg = "white",
      limitsize = FALSE
    )
    popUp(title = "Saved in your working folder", type = "saved")
  })

  ## Dep Tree Report
  observeEvent(input$depTreeReport, {
    req(values$depTreeGgplot)
    df <- depTreeData()
    req(nrow(df) > 0)
    popUp(title = NULL, type = "waiting")

    display_df <- df %>%
      select(tid, token, lemma, upos, dep_rel, hid) %>%
      left_join(
        df %>% select(tid, token) %>% rename(head_token = token),
        by = c("hid" = "tid")
      ) %>%
      mutate(head_token = ifelse(hid == 0, "ROOT", head_token))

    sheetname <- "DependencyTree"
    n_tokens <- nrow(df)
    w <- max(10, n_tokens * 1.2)
    h <- max(5, min(10, n_tokens * 0.15 + 3))
    list_df <- list(display_df)
    list_plot <- list(values$depTreeGgplot)
    wb <- addSheetToReport(
      list_df,
      list_plot,
      sheetname = sheetname,
      wb = values$wb,
      dpi = values$report_dpi
    )
    values$wb <- wb
    popUp(title = "Dependency Tree Results", type = "success")
    values$myChoices <- sheets(values$wb)
  })

  output$depTreeTable <- renderDT(server = FALSE, {
    df <- depTreeData()
    req(nrow(df) > 0)

    display_df <- df %>%
      select(tid, token, lemma, upos, dep_rel, hid) %>%
      left_join(
        df %>% select(tid, token) %>% rename(head_token = token),
        by = c("hid" = "tid")
      ) %>%
      mutate(head_token = ifelse(hid == 0, "ROOT", head_token)) %>%
      rename(
        "ID" = tid,
        "Token" = token,
        "Lemma" = lemma,
        "PoS" = upos,
        "Relation" = dep_rel,
        "Head ID" = hid,
        "Head Token" = head_token
      )

    DTformat(
      display_df,
      size = "100%",
      filename = "DependencyTree",
      dom = FALSE
    )
  })

  ## Morphological Features ----

  # Human-readable labels for morphological feature values
  morphLabels <- list(
    Tense = c(Past = "Past", Pres = "Present", Fut = "Future", Imp = "Imperfect", Pqp = "Pluperfect"),
    Mood = c(Ind = "Indicative", Sub = "Subjunctive", Imp = "Imperative", Cnd = "Conditional"),
    Number = c(Sing = "Singular", Plur = "Plural", Dual = "Dual"),
    Person = c("1" = "1st Person", "2" = "2nd Person", "3" = "3rd Person"),
    VerbForm = c(Fin = "Finite", "Inf" = "Infinitive", Part = "Participle", Ger = "Gerund", Conv = "Converb", Sup = "Supine"),
    Degree = c(Pos = "Positive", Cmp = "Comparative", Sup = "Superlative", Abs = "Absolute Superlative"),
    Gender = c(Masc = "Masculine", Fem = "Feminine", Neut = "Neuter", Com = "Common"),
    Case = c(Nom = "Nominative", Acc = "Accusative", Gen = "Genitive", Dat = "Dative", Ins = "Instrumental", Loc = "Locative", Voc = "Vocative", Abl = "Ablative"),
    Voice = c(Act = "Active", Pass = "Passive", Mid = "Middle"),
    Definite = c(Def = "Definite", Ind = "Indefinite", Com = "Complex", Cons = "Construct"),
    PronType = c(Prs = "Personal", Dem = "Demonstrative", Rel = "Relative", Int = "Interrogative", Ind = "Indefinite", Neg = "Negative", Tot = "Total", Art = "Article", Exc = "Exclamative")
  )

  morphDescriptions <- c(
    Tense = "Verb tense distribution. Reveals whether the corpus is predominantly narrative (Past), descriptive/argumentative (Present), or forward-looking (Future).",
    Mood = "Verbal mood distribution. Indicative conveys facts; Subjunctive signals uncertainty, doubt, or desire; Imperative expresses commands; Conditional marks hypothetical situations.",
    Number = "Grammatical number distribution across nouns, pronouns, adjectives, and verbs.",
    Person = "Verb person distribution. 1st person signals personal/subjective style; 3rd person indicates impersonal/objective style; 2nd person suggests direct address.",
    VerbForm = "Verb form distribution. The ratio of finite to non-finite forms (infinitives, participles, gerunds) reflects syntactic complexity.",
    Degree = "Adjective/adverb degree distribution. High use of comparatives and superlatives may indicate evaluative or argumentative discourse.",
    Gender = "Grammatical gender distribution (language-dependent). Available for languages with grammatical gender marking.",
    Case = "Grammatical case distribution (language-dependent). Available for languages with case systems (e.g., German, Russian, Latin).",
    Voice = "Active vs passive voice distribution. High passive usage may indicate formal, scientific, or bureaucratic writing style.",
    Definite = "Definiteness marking. The ratio of definite to indefinite articles/determiners reveals information structure patterns.",
    PronType = "Pronoun and determiner type distribution. Shows the balance between personal, demonstrative, relative, and other pronoun types."
  )

  parseMorphFeatures <- function(feats_col, feature_name) {
    # Parse feats column (format: "Number=Plur|Tense=Past|Mood=Ind")
    # Extract values for the requested feature
    pattern <- paste0("(?:^|\\|)", feature_name, "=([^|]+)")
    matches <- regmatches(feats_col, regexpr(pattern, feats_col, perl = TRUE))
    values <- sub(paste0(".*", feature_name, "="), "", matches)
    values[matches == ""] <- NA
    # Handle non-matching rows
    result <- rep(NA_character_, length(feats_col))
    has_match <- nchar(matches) > 0
    result[has_match] <- values[has_match]
    return(result)
  }

  output$morphFeatureDesc <- renderUI({
    feature <- input$morphFeature
    desc <- morphDescriptions[feature]
    if (is.na(desc)) desc <- ""
    div(
      style = "padding: 10px; background-color: #f0f7ee; border-left: 4px solid #4F7942; border-radius: 4px; margin-bottom: 15px;",
      icon("info-circle", style = "color: #4F7942;"),
      span(desc, style = "color: #333; font-size: 13px;")
    )
  })

  output$morphFeaturePlot <- renderPlotly({
    req(values$dfTag)
    feature <- input$morphFeature

    df <- values$dfTag %>%
      dplyr::filter(docSelected) %>%
      mutate(feat_value = parseMorphFeatures(feats, feature)) %>%
      filter(!is.na(feat_value))

    if (nrow(df) == 0) return(plotly_empty())

    # Apply human-readable labels
    labels <- morphLabels[[feature]]
    if (!is.null(labels)) {
      df$feat_label <- ifelse(
        df$feat_value %in% names(labels),
        labels[df$feat_value],
        df$feat_value
      )
    } else {
      df$feat_label <- df$feat_value
    }

    freq_table <- df %>%
      count(feat_label) %>%
      arrange(desc(n)) %>%
      mutate(
        pct = round(100 * n / sum(n), 1),
        feat_label = factor(feat_label, levels = rev(feat_label))
      )

    values$morphPlot <- plot_ly(
      freq_table,
      x = ~n,
      y = ~feat_label,
      type = "bar",
      orientation = "h",
      marker = list(color = "#4F7942"),
      text = ~paste0(n, " (", pct, "%)"),
      textposition = "auto",
      hovertemplate = "<b>%{y}</b><br>Count: %{x}<br>%{text}<extra></extra>"
    ) %>%
      layout(
        title = list(
          text = paste0("<b>", feature, " Distribution</b>"),
          font = list(size = 16, color = "gray30"),
          x = 0.5
        ),
        xaxis = list(title = "Count"),
        yaxis = list(title = ""),
        paper_bgcolor = "white",
        plot_bgcolor = "white",
        margin = list(l = 150)
      ) %>%
      config(displaylogo = FALSE)

    values$morphPlot
  })

  output$morphFeatureTable <- renderDT(server = FALSE, {
    req(values$dfTag)
    feature <- input$morphFeature

    df <- values$dfTag %>%
      dplyr::filter(docSelected) %>%
      mutate(feat_value = parseMorphFeatures(feats, feature)) %>%
      filter(!is.na(feat_value))

    if (nrow(df) == 0) return(NULL)

    # Apply human-readable labels
    labels <- morphLabels[[feature]]
    if (!is.null(labels)) {
      df$feat_label <- ifelse(
        df$feat_value %in% names(labels),
        paste0(labels[df$feat_value], " (", df$feat_value, ")"),
        df$feat_value
      )
    } else {
      df$feat_label <- df$feat_value
    }

    # Cross-tabulation: feature value x PoS
    cross_tab <- df %>%
      count(feat_label, upos) %>%
      pivot_wider(names_from = upos, values_from = n, values_fill = 0) %>%
      mutate(Total = rowSums(across(where(is.numeric)))) %>%
      arrange(desc(Total)) %>%
      rename(!!feature := feat_label)

    DTformat(
      cross_tab,
      size = "100%",
      filename = paste0("MorphFeature_", feature)
    )
  })
}

wordsUI <- function() {
  ## WORDS ----

  ### Clustering ----
  clustering <- tabItem(
    tabName = "w_clustering",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("sitemap"), strong("Clustering"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "w_clusteringApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "w_clusteringExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "w_clusteringReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                selectInput(
                  "w_clusteringSimilarity",
                  label = "Words Similarity by:",
                  choices = c(
                    "None" = "none",
                    "Association Index" = "association",
                    "Cosine Similarity" = "cosine",
                    "Jaccard Index" = "jaccard"
                  ),
                  selected = "association"
                ),
                selectInput(
                  "w_clusteringMode",
                  label = "Cluster selection:",
                  choices = c(
                    "Auto" = "auto",
                    "Manual" = "manual"
                  ),
                  selected = "auto"
                ),
                conditionalPanel(
                  condition = "input.w_clusteringMode == 'manual'",
                  numericInput(
                    "w_nclusters",
                    label = "N. of Clusters",
                    value = 1,
                    min = 1,
                    step = 1
                  )
                )
              ),

              # Parameters Section
              tags$details(
                class = "params-section",
                tags$summary(
                  div(
                    class = "params-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("list"),
                      " Parameters"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "w_clusteringNMax",
                        label = "Words",
                        value = 50,
                        min = 2,
                        step = 1
                      )
                    ),
                    column(
                      6,
                      numericInput(
                        "w_clusteringLabelSize",
                        label = "Label Size",
                        value = 4,
                        min = 1,
                        step = 0.5
                      )
                    )
                  )
                )
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "320px"
            )
          ),
          style = style_opt
        )
      ),
      fluidRow(
        tabsetPanel(
          type = "tabs",
          tabPanel(
            title = "Dendrogram", icon = icon("sitemap"),
            br(),
            shinycssloaders::withSpinner(
              visNetworkOutput(
                "w_clusteringPlot",
                width = "auto",
                height = "75vh"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Table", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("w_clusteringTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          )
        )
      )
    )
  )

  ### Reinert Clustering ----
  reinert <- tabItem(
    tabName = "w_reinclustering",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("project-diagram"), strong("Reinert Clustering"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "w_reinclusteringApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "w_reinclusteringExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "w_reinclusteringReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                fluidRow(
                  column(
                    6,
                    numericInput(
                      "w_rein_k",
                      label = "Max N. of Clusters",
                      value = 10,
                      min = 1,
                      step = 1
                    )
                  ),
                  column(
                    6,
                    numericInput(
                      "w_rein_min_split_members",
                      label = "Min. Segments per Cluster",
                      value = 10,
                      min = 1,
                      step = 1
                    )
                  )
                )
              ),

              # Segment Parameters Section
              tags$details(
                class = "params-section",
                tags$summary(
                  div(
                    class = "params-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("scissors"),
                      " Segment Parameters"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "w_rein_segments_size",
                        label = "Segment Length",
                        value = 40,
                        min = 3,
                        step = 1
                      )
                    ),
                    column(
                      6,
                      numericInput(
                        "w_rein_min_segments",
                        label = "Min. Segment Length",
                        value = 5,
                        min = 3,
                        step = 1
                      )
                    )
                  )
                )
              ),

              # Feature Selection Section
              tags$details(
                class = "advanced-section",
                tags$summary(
                  div(
                    class = "advanced-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("circle-check"),
                      " Feature Selection Parameters"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "w_rein_cc_test",
                        label = "Contingency Coefficient Value",
                        value = 0.3,
                        min = 0.1,
                        step = 0.1,
                        max = 0.8
                      )
                    ),
                    column(
                      6,
                      numericInput(
                        "w_rein_tsj",
                        label = "Min freq",
                        value = 3,
                        min = 1,
                        step = 1
                      )
                    )
                  )
                )
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "320px"
            )
          ),
          style = style_opt
        )
      ),
      fluidRow(
        tabsetPanel(
          type = "tabs",
          tabPanel(
            title = "Dendrogram", icon = icon("sitemap"),
            br(),
            uiOutput("ReinCutree"),
            shinycssloaders::withSpinner(
              visNetworkOutput(
                "w_ReinClusteringPlot",
                width = "auto",
                height = "75vh"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Summary", icon = icon("list"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("w_ReinSummaryTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Terms by Cluster", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("w_ReinClusteringTableTerms"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Segments by Cluster", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("w_ReinClusteringTableSegments"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Info & References", icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  br(),
                  HTML(infoTexts$reinert)
                ),
                column(1)
              )
            )
          )
        )
      )
    )
  )

  ### Correspondence Analysis ----
  ca <- tabItem(
    tabName = "ca",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("braille"), strong("Correspondence Analysis"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "caApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "caExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "caReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                selectInput(
                  inputId = "groupCA",
                  label = "Unit of Analysis",
                  choices = c(
                    "Groups",
                    "Documents",
                    "Paragraphs",
                    "Sentences"
                  ),
                  selected = "Documents"
                ),
                numericInput(
                  "nCA",
                  label = "Words",
                  value = 50,
                  min = 2,
                  step = 1
                )
              ),

              # Clustering Parameters Section
              tags$details(
                class = "params-section",
                tags$summary(
                  div(
                    class = "params-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("chart-bar"),
                      " Clustering Parameters"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "nClustersCA",
                        label = "Clusters",
                        value = 1,
                        min = 1,
                        step = 1
                      )
                    ),
                    column(
                      6,
                      numericInput(
                        "nDimsCA",
                        label = "Dims for Clustering",
                        value = 2,
                        min = 1,
                        max = 10,
                        step = 1
                      )
                    )
                  )
                )
              ),

              # Graphical Options Section
              tags$details(
                class = "advanced-section",
                tags$summary(
                  div(
                    class = "advanced-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("image"),
                      " Graphical Options"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  selectInput(
                    "dimPlotCA",
                    "Select plane to plot:",
                    choices = c(
                      "1° Factorial Plane" = "1",
                      "2° Factorial Plane" = "2",
                      "3° Factorial Plane" = "3",
                      "4° Factorial Plane" = "4",
                      "5° Factorial Plane" = "5"
                    ),
                    selected = "1"
                  ),
                  numericInput(
                    "nDocCA",
                    label = "Top Contributing Docs/Groups",
                    value = 0,
                    min = 0,
                    step = 1
                  ),
                  numericInput(
                    "lim.contribCA",
                    label = "Max Contribution",
                    value = 2,
                    min = 0.01,
                    max = 10,
                    step = 0.01
                  ),
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "labelsizeCA",
                        label = "Label Size",
                        value = 16,
                        min = 2,
                        step = 1
                      )
                    ),
                    column(
                      6,
                      numericInput(
                        "sizeCA",
                        label = "Min. Dot Size",
                        value = 2,
                        min = 0,
                        max = 20,
                        step = 1
                      )
                    )
                  )
                )
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "320px"
            )
          ),
          style = style_opt
        )
      ),
      fluidRow(
        tabsetPanel(
          type = "tabs",
          tabPanel(
            title = "Factorial Plane", icon = icon("chart-column"),
            br(),
            shinycssloaders::withSpinner(
              plotlyOutput(
                outputId = "caPlot",
                height = "75vh",
                width = "98.9%"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Dendrogram", icon = icon("sitemap"),
            br(),
            shinycssloaders::withSpinner(
              visNetworkOutput(
                "caDendrogram",
                width = "auto",
                height = "75vh"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Singular Values", icon = icon("chart-line"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("caSingularValueTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Coordinates", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("caCoordTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Contributions", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("caContribTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Cosines Squared", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("caCosineTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "TALL AI", icon = icon("robot"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  br(),
                  shinycssloaders::withSpinner(
                    htmlOutput("caGeminiUI"),
                    caption = HTML("<br><strong>Thinking...</strong>"),
                    image = "ai_small2.gif",
                    color = "#4F7942"
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "Info & References", icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  br(),
                  HTML(infoTexts$correspondenceanalysis)
                ),
                column(1)
              )
            )
          )
        )
      )
    )
  )

  ### Network ----

  ## Word Network Analysis ----

  coword <- tabItem(
    tabName = "w_networkCooc",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("circle-nodes"), strong("Word Network Analysis"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "w_networkCoocApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "w_networkCoocExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "w_networkCoocReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                selectInput(
                  inputId = "w_networkType",
                  label = "Network Type",
                  choices = c(
                    "Co-occurrence" = "cooc",
                    "Dependency" = "dep"
                  ),
                  selected = "cooc"
                ),
                conditionalPanel(
                  'input.w_networkType == "cooc"',
                  selectInput(
                    inputId = "w_groupNet",
                    label = "Co-occurrences in",
                    choices = c(
                      "Groups",
                      "Documents",
                      "Paragraphs",
                      "Sentences"
                    ),
                    selected = "Sentences"
                  )
                ),
                conditionalPanel(
                  'input.w_networkType == "dep"',
                  selectInput(
                    inputId = "w_depRelFilter",
                    label = "Dependency Relations",
                    choices = c(
                      "All syntactic" = "all",
                      "Noun modifiers (amod, nmod, compound)" = "noun_mod",
                      "Subject-Verb-Object" = "svo",
                      "Custom" = "custom"
                    ),
                    selected = "all"
                  ),
                  conditionalPanel(
                    'input.w_depRelFilter == "custom"',
                    selectInput(
                      inputId = "w_depRelCustom",
                      label = "Select relations:",
                      choices = c("nsubj", "obj", "iobj", "obl", "amod", "nmod",
                        "compound", "flat", "conj", "appos", "acl", "advcl",
                        "xcomp", "ccomp", "advmod", "nummod"),
                      selected = c("nsubj", "obj", "amod", "nmod", "compound"),
                      multiple = TRUE
                    )
                  )
                ),
                selectInput(
                  "normalizationCooc",
                  label = "Normalization by:",
                  choices = c(
                    "None" = "none",
                    "Association Index" = "association",
                    "Cosine Similarity" = "cosine",
                    "Jaccard Index" = "jaccard"
                  ),
                  selected = "association"
                )
              ),

              # Network Options Section
              tags$details(
                class = "advanced-section",
                tags$summary(
                  div(
                    class = "advanced-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("link"),
                      " Network Options"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  materialSwitch(
                    inputId = "noOverlap",
                    label = "Avoid label overlap",
                    value = TRUE,
                    status = "success"
                  ),
                  materialSwitch(
                    inputId = "interLinks",
                    label = "Inter-group links",
                    value = TRUE,
                    status = "success"
                  ),
                  materialSwitch(
                    inputId = "removeIsolated",
                    label = "Delete isolated nodes",
                    value = FALSE,
                    status = "success"
                  )
                )
              ),

              # Graphical Parameters Section
              tags$details(
                class = "params-section",
                tags$summary(
                  div(
                    class = "params-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("eye"),
                      " Graphical Parameters"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "nMax",
                        label = "Words",
                        value = 100,
                        min = 2,
                        step = 1
                      ),
                      numericInput(
                        "labelSize",
                        label = "Label Size",
                        value = 4,
                        min = 1,
                        step = 0.5
                      )
                    ),
                    column(
                      6,
                      selectInput(
                        "minEdges",
                        label = "Top Link (%)",
                        choices = c("Auto", paste0(seq(10, 100, 10), "%")),
                        selected = "Auto"
                      ),
                      numericInput(
                        "opacity",
                        label = "Opacity",
                        value = 0.6,
                        min = 0,
                        max = 1,
                        step = 0.1
                      )
                    )
                  ),
                  fluidRow(
                    column(
                      12,
                      sliderInput(
                        "communityRepulsion",
                        label = "Community Repulsion",
                        value = 0.5,
                        min = 0,
                        max = 1,
                        step = 0.1
                      )
                    )
                  )
                )
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "320px"
            )
          ),
          style = style_opt
        )
      ),
      fluidRow(
        tabsetPanel(
          type = "tabs",
          tabPanel(
            title = "Network", icon = icon("diagram-project"),
            br(),
            shinycssloaders::withSpinner(
              visNetworkOutput(
                "w_networkCoocPlot",
                width = "auto",
                height = "75vh"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Words", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("w_networkCoocNodesTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Links", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("w_networkCoocEdgesTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "TALL AI", icon = icon("robot"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  br(),
                  shinycssloaders::withSpinner(
                    htmlOutput("w_networkGeminiUI"),
                    caption = HTML("<br><strong>Thinking...</strong>"),
                    image = "ai_small2.gif",
                    color = "#4F7942"
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "Info & References", icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  br(),
                  HTML(infoTexts$cowordanalysis)
                ),
                column(1)
              )
            )
          )
        )
      )
    )
  )

  ## Thematic Map ----

  tm <- tabItem(
    tabName = "w_networkTM",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("map"), strong("Thematic Map"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "w_networkTMApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "w_networkTMExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "w_networkTMReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                selectInput(
                  inputId = "w_groupTM",
                  label = "Co-occurrences in",
                  choices = c(
                    "Groups",
                    "Documents",
                    "Paragraphs",
                    "Sentences"
                  ),
                  selected = "Sentences"
                ),
                materialSwitch(
                  inputId = "noOverlapTM",
                  label = "Avoid label overlap",
                  value = TRUE,
                  status = "success"
                )
              ),

              # Graphical Parameters Section
              tags$details(
                class = "params-section",
                tags$summary(
                  div(
                    class = "params-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("eye"),
                      " Graphical Parameters"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "nMaxTM",
                        label = "Words",
                        value = 250,
                        min = 2,
                        step = 1
                      ),
                      numericInput(
                        "labelSizeTM",
                        label = "Label Size",
                        value = 4,
                        min = 1,
                        step = 0.5
                      )
                    ),
                    column(
                      6,
                      numericInput(
                        "n.labelsTM",
                        label = "Labels",
                        value = 3,
                        min = 0,
                        step = 1
                      ),
                      numericInput(
                        "opacityTM",
                        label = "Opacity",
                        value = 0.6,
                        min = 0,
                        max = 1,
                        step = 0.1
                      )
                    )
                  )
                )
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "320px"
            )
          ),
          style = style_opt
        )
      ),
      fluidRow(
        tabsetPanel(
          type = "tabs",
          tabPanel(
            title = "Thematic Map", icon = icon("map"),
            br(),
            shinycssloaders::withSpinner(
              plotlyOutput(
                "w_networkTMMapPlot",
                width = "auto",
                height = "75vh"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Network", icon = icon("diagram-project"),
            br(),
            shinycssloaders::withSpinner(
              visNetworkOutput(
                "w_networkTMNetPlot",
                width = "auto",
                height = "75vh"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Clusters", icon = icon("object-group"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("w_networkTMClusterTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "Words", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("w_networkTMWordTable"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            title = "TALL AI", icon = icon("robot"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  br(),
                  shinycssloaders::withSpinner(
                    htmlOutput("w_networkTMGeminiUI"),
                    caption = HTML("<br><strong>Thinking...</strong>"),
                    image = "ai_small2.gif",
                    color = "#4F7942"
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "Info & References", icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  br(),
                  HTML(infoTexts$thematicmap)
                ),
                column(1)
              )
            )
          )
        )
      )
    )
  )

  ## WORD EMBEDDINGS TRAINING----

  we_training <- tabItem(
    tabName = "w_word2vec",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("graduation-cap"), strong("Embedding Training"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      fluidRow(
        column(
          9,
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Embedding Distributions", icon = icon("chart-area"),
              br(),
              shinycssloaders::withSpinner(
                plotlyOutput(
                  "w_word2vecBoxplot",
                  height = "75vh",
                  width = "98.9%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Embedding Dimensions", icon = icon("chart-line"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("w_word2vecTable"),
                color = getOption("spinner.color", default = "#4F7942")
              ),
              shinycssloaders::withSpinner(
                plotlyOutput(
                  "w_word2vecPCA",
                  width = "auto",
                  height = "50vh"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & References", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(infoTexts$embeddingtrain)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          3,
          div(
            box(
              width = 12,
              div(
                h3(strong(em("Model Training"))),
                style = "margin-top:-57px"
              ),
              tags$hr(),
              style = "text-align: left; text-color: #989898",
              h4("Options:"),
              selectizeInput(
                inputId = "w2vMethod",
                label = "Embedding Method",
                choices = c(
                  "Continuous Bag of Words" = "cbow",
                  "Skip-Gram" = "skip-gram"
                ),
                selected = "cbow"
              ),
              fluidRow(
                column(
                  6,
                  numericInput(
                    inputId = "w2vDim",
                    label = "Dimensions",
                    value = 100,
                    min = 10,
                    step = 1
                  )
                ),
                column(
                  6,
                  numericInput(
                    inputId = "w2vIter",
                    label = "Iterations",
                    value = 20,
                    min = 5,
                    max = 100
                  )
                )
              ),
              fluidRow(
                column(
                  4,
                  align = "center",
                  do.call(
                    "actionButton",
                    c(
                      run_bttn,
                      list(
                        inputId = "w2vApply"
                      )
                    )
                  )
                ),
                column(
                  4,
                  align = "center",
                  do.call(
                    "actionButton",
                    c(
                      export_bttn,
                      list(
                        inputId = "w2vSave"
                      )
                    )
                  )
                ),
                column(
                  4,
                  align = "center",
                  do.call(
                    "actionButton",
                    c(
                      report_bttn,
                      list(
                        inputId = "w2vReport"
                      )
                    )
                  )
                )
              )
            ),
            style = "margin-top:40px"
          )
        )
      )
    )
  )

  ## WORD EMBEDDING SIMILARITY ----
  we_similarity <- tabItem(
    tabName = "w_w2v_similarity",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("vector-square"), strong("Embedding Similarity"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "w_w2v_similarityApply"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "w_w2v_similarityExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "w_w2v_similarityReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                numericInput(
                  "w_w2v_similarityN",
                  label = "Top Words",
                  min = 1,
                  value = 100,
                  step = 1
                ),
                sliderInput(
                  "w_w2v_font_size",
                  "Font Size",
                  min = 8,
                  max = 30,
                  value = 14,
                  step = 1
                ),
                selectInput(
                  "w_w2v_overlap",
                  "Avoid Label Overlap",
                  choices = c(
                    "No" = "none",
                    "Hide" = "hide",
                    "Transparency" = "transparency"
                  ),
                  selected = "transparency"
                )
              ),

              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "300px"
            )
          ),
          style = style_opt
        )
      )
    ),
    fluidRow(
      column(12,
      tabsetPanel(
        type = "tabs",
        tabPanel(
          title = "Similarity Network", icon = icon("diagram-project"),
          br(),
          fluidRow(
            column(11, align = "right", uiOutput("w_w2v_Selected")),
            column(
              1,
              title = "Back to the full network",
              do.call(
                "actionButton",
                c(
                  x_bttn,
                  list(
                    inputId = "w_w2v_Back"
                  )
                )
              )
            )
          ),
          shinycssloaders::withSpinner(
            visNetworkOutput(
              "w_w2vNetworkplot",
              width = "auto",
              height = "75vh"
            ),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "UMAP", icon = icon("chart-area"),
          br(),
          shinycssloaders::withSpinner(
            plotlyOutput("w_w2vUMAPplot", width = "auto", height = "75vh"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        tabPanel(
          title = "TALL AI", icon = icon("robot"),
          br(),
          fluidPage(
            fluidRow(
              column(
                12,
                br(),
                shinycssloaders::withSpinner(
                  htmlOutput("w_w2vGeminiUI"),
                  caption = HTML("<br><strong>Thinking...</strong>"),
                  image = "ai_small2.gif",
                  color = "#4F7942"
                )
              )
            )
          )
        ),
        tabPanel(
          title = "Info & References", icon = icon("circle-info"),
          br(),
          fluidPage(
            fluidRow(
              column(1),
              column(
                10,
                br(),
                HTML(infoTexts$embeddingsimilarity)
              ),
              column(1)
            )
          )
        )
      ))
    )
  )

  return(list(
    clustering = clustering,
    reinert = reinert,
    ca = ca,
    coword = coword,
    tm = tm,
    we_training = we_training,
    we_similarity = we_similarity
  ))
}

wordsServer <- function(input, output, session, values, statsValues) {
  ### WORDS ----

  showDocumentModal <- function(session, docID, input) {
    ns <- session$ns
    if (
      input$sidebarmenu %in%
        c("import_tx", "split_tx", "extInfo", "randomText")
    ) {
      text <- create_document_box(
        values$txt %>% filter(doc_id == docID) %>% pull(text),
        docID,
        summarization_type = "original_text"
      )
    } else {
      text <- create_document_box(
        values$dfTag %>%
          filter(doc_id == !!docID) %>%
          select(doc_id, paragraph_id, sentence_id, sentence) %>%
          distinct() %>%
          group_by(doc_id, paragraph_id) %>%
          summarise(Paragraph = paste(sentence, collapse = "\n ")) %>%
          ungroup(),
        docID,
        summarization_type = "abstractive"
      )
    }

    modalDialog(
      tags$style(HTML(
        "
      .modal-dialog {
        width: 70vw !important;
        max-width: 70vw !important;
      }
      .modal-content {
        height: 85vh !important;
      }
      .modal-body {
        overflow-y: auto !important;
        max-height: calc(85vh - 120px) !important;
      }
    "
      )),
      div(
        h3(strong("Document corpus")),
        br(),
        HTML(text)
      ),
      size = "l",
      easyClose = TRUE,
      footer = modalButton("Close")
    )
  }

  ## Click on Plotly graphs: DOC IN CONTEXT ----
  observeEvent(event_data("plotly_click"), {
    d <- event_data("plotly_click")
    elementY <- d$y[1]
    if (!is.null(elementY)) {
      if (elementY %in% unique(values$dfTag$doc_id)) {
        showModal(showDocumentModal(
          session,
          elementY,
          input
        ))
      }
    }
  })

  ## Reinert Clustering ----
  dendReinFunction <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$w_reinclusteringApply
    },
    valueExpr = {
      values$reinert <- tall::reinert(
        x = values$dfTag,
        k = input$w_rein_k,
        term = values$generalTerm,
        segment_size = input$w_rein_segments_size,
        min_segment_size = input$w_rein_min_segments,
        min_split_members = input$w_rein_min_split_members,
        cc_test = input$w_rein_cc_test,
        tsj = input$w_rein_tsj
      )

      values$tc <- tall::term_per_cluster(
        values$reinert,
        cutree = NULL,
        k = unique(values$reinert$group)
      )
      values$reinertSummary <- tall::reinSummary(values$tc, 10)

      # groups <- tibble(uc=1:length(values$reinert$group), Cluster=values$reinert$group)
      values$tc$segments <- values$tc$segments %>%
        group_by(cluster) %>%
        arrange(uc, .by_group = TRUE) %>%
        select(-"uc")

      output$ReinCutree <- renderUI({
        req(input$w_rein_k)
        fluidRow(
          column(9),
          column(
            3,
            selectInput(
              "ReinCutree",
              label = "Dendrogram Pruning",
              choices = input$w_rein_k:1,
              selected = input$w_rein_k
            )
          )
        )
      })

      values$ReinertDendrogram <- dend2vis(
        values$reinert,
        labelsize = 10,
        nclusters = input$w_rein_k,
        community = FALSE
      )
    }
  )

  cutree_event <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$ReinCutree
    },
    valueExpr = {
      values$ReinertDendrogram <- dend2vis(
        values$reinert,
        labelsize = 10,
        nclusters = as.numeric(input$ReinCutree),
        community = FALSE
      )
    }
  )

  output$w_ReinClusteringPlot <- renderVisNetwork({
    dendReinFunction()
    cutree_event()
    values$ReinertDendrogram
  })

  output$w_ReinClusteringTableSegments <- renderDT({
    dendReinFunction()
    # find sentences containing the tokens/lemma
    DTformat(
      values$tc$segments,
      size = "100%",
      button = FALSE,
      col_to_remove = values$generalTerm
    )
  })
  output$w_ReinSummaryTable <- renderDT({
    dendReinFunction()
    # find sentences containing the tokens/lemma
    DTformat(
      values$reinertSummary,
      size = "100%",
      button = FALSE,
      filter = "none",
      col_to_remove = values$generalTerm
    )
  })

  output$w_ReinClusteringTableTerms <- renderDT({
    dendReinFunction()
    DTformat(
      values$tc$terms %>%
        mutate(
          freq = round(freq * 100, 1),
          chi_square = round(chi_square, 1),
          p_value = round(p_value, 4)
        ) %>%
        select(term, freq_true, freq, chi_square, p_value, sign, cluster) %>%
        rename(
          "Term" = term,
          "Freq. in Cluster" = freq_true,
          "% in Cluster" = freq,
          "Chi^2" = chi_square,
          "P-Value" = p_value,
          "Sign" = sign,
          "Cluster" = cluster
        ),
      size = "85%",
      button = FALSE,
      numeric = c(5),
      round = 3
    )
  })

  ## export CLustering button
  observeEvent(
    eventExpr = {
      input$w_reinclusteringExport
    },
    handlerExpr = {
      file <- paste("ReinertDendrogram-", sys.time(), ".png", sep = "")
      file <- destFolder(file, values$wdTall)
      plot2png(values$ReinertDendrogram, filename = file, type = "vis", dpi = values$dpi, height = values$h)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report

  observeEvent(input$w_reinclusteringReport, {
    if (!is.null(values$reinert)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "Reinert"
      list_df <- list(
        values$reinertSummary,
        values$tc$terms %>%
          mutate(freq = round(freq * 100, 1)) %>%
          select(term, freq, chi_square, p_value, sign, cluster) %>%
          rename(
            "Term" = term,
            "% in Cluster" = freq,
            "Chi^2" = chi_square,
            "P-Value" = p_value,
            "Sign" = sign,
            "Cluster" = cluster
          ),
        values$tc$segments
      )
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileReinertDendrogram <- plot2png(
        values$ReinertDendrogram,
        filename = "ReinertDendrogram.png",
        type = "vis",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileReinertDendrogram, res$col)
      )
      popUp(title = "Reinert Clustering Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Clustering ----

  ## Dendrogramm ----
  dendFunction <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$w_clusteringApply
    },
    valueExpr = {
      results <- clustering(
        LemmaSelection(values$dfTag) %>% dplyr::filter(docSelected),
        n = input$w_clusteringNMax,
        group = "doc_id",
        minEdges = 25,
        term = values$generalTerm,
        normalization = input$w_clusteringSimilarity
      )
      values$wordCluster <- results$cluster
      values$wordCluster <- values$wordCluster %>%
        rename(Word = word, Group = group, Frequency = frequency)
      values$wordComm <- results$comm
      if (input$w_clusteringMode == "auto") {
        nclusters <- max(values$wordComm$membership)
      } else {
        nclusters <- min(
          input$w_nclusters,
          length(values$wordComm$membership) - 1
        )
      }
      values$WordDendrogram <- dend2vis(
        values$wordComm,
        labelsize = input$w_clusteringLabelSize,
        nclusters = nclusters
      )
    }
  )

  output$w_clusteringPlot <- renderVisNetwork({
    dendFunction()
    values$WordDendrogram
  })

  output$w_clusteringTable <- renderDT(server = FALSE, {
    dendFunction()
    DTformat(
      values$wordCluster,
      size = "100%",
      filename = "ClusterWordsTable",
      pagelength = TRUE,
      left = 1,
      right = NULL,
      numeric = NULL,
      dom = TRUE,
      filter = "top",
      col_to_remove = values$generalTerm
    )
  })

  ## export CLustering button
  observeEvent(
    eventExpr = {
      input$w_clusteringExport
    },
    handlerExpr = {
      file <- paste("Dendrogram-", sys.time(), ".png", sep = "")
      file <- destFolder(file, values$wdTall)
      plot2png(values$WordDendrogram, filename = file, type = "vis", dpi = values$dpi, height = values$h)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report

  observeEvent(input$w_clusteringReport, {
    if (!is.null(values$wordCluster)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "Clustering"
      list_df <- list(values$wordCluster)
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileDend <- plot2png(
        values$WordDendrogram,
        filename = "Clustering.png",
        type = "vis",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileDend, res$col)
      )
      popUp(title = "Clustering Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Correspondence Analysis ----

  # CA plot
  caPlotFunction <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$caApply
    },
    valueExpr = {
      ## check to verify if groups exist or not
      filtered <- LemmaSelection(values$dfTag) %>% dplyr::filter(docSelected)
      if (
        input$groupCA == "Documents" & "ungroupDoc_id" %in% names(values$dfTag)
      ) {
        filtered <- backToOriginalGroups(filtered)
      }
      values$CA <- wordCA(
        filtered,
        n = input$nCA,
        term = values$generalTerm,
        group = input$groupCA
      )
      ##
      values$CA <- caClustering(
        values$CA,
        nclusters = input$nClustersCA,
        nDim = input$nDimsCA,
        lim.contr = input$lim.contribCA
      )
      values$CAdimY <- as.numeric(input$dimPlotCA) * 2
      values$CAdimX <- values$CAdimY - 1
      values$plotCA <- ca2plotly(
        values$CA,
        dimX = values$CAdimX,
        dimY = values$CAdimY,
        topWordPlot = input$nCA,
        topDocPlot = input$nDocCA,
        threshold = 0.03,
        labelsize = input$labelsizeCA,
        size = input$sizeCA,
        lim.contr = input$lim.contribCA
      )
      values$CADendrogram <- dend2vis(
        values$CA$clustering$h,
        labelsize = input$labelsizeCA,
        nclusters = input$nClustersCA,
        community = FALSE
      )

      # wordCoordData
      values$CA$wordCoordData <- values$CA$wordCoord %>%
        select(label, everything()) %>%
        left_join(
          data.frame(
            label = names(values$CA$clustering$groups),
            Group = values$CA$clustering$groups
          ),
          by = "label"
        ) %>%
        rename(Label = label) %>%
        select(Label, Group, everything()) %>%
        rename(Cluster = Group)

      # contribData
      values$CA$contribData <- values$CA$contrib %>%
        tibble::rownames_to_column() %>%
        rename(Label = rowname)

      # cosineData
      values$CA$cosineData <- values$CA$cosine %>%
        tibble::rownames_to_column() %>%
        rename(Label = rowname)

      # dfCA
      if (length(values$CA$ca$sv) < 10) {
        values$dfCA <- data.frame(
          dim = paste0("Dim ", 1:length(values$CA$ca$sv)),
          sv = (values$CA$ca$sv / sum(values$CA$ca$sv) * 100),
          svcorr = values$CA$ca$eigCorrectedNorm
        )
      } else {
        values$dfCA <- data.frame(
          dim = paste0("Dim ", 1:10),
          sv = (values$CA$ca$sv / sum(values$CA$ca$sv) * 100)[1:10],
          svcorr = values$CA$ca$eigCorrectedNorm[1:10]
        )
      }
      values$dfCA <- values$dfCA %>%
        rename(
          "Factorial Dimension" = dim,
          "Singular Values" = sv,
          "Corrected Explained Inertia" = svcorr
        )
    }
  )

  output$caPlot <- renderPlotly({
    caPlotFunction()
    values$plotCA
  })

  output$caDendrogram <- renderVisNetwork({
    caPlotFunction()
    values$CADendrogram
  })

  # gemini button for correspondence analysis
  output$caGeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "", content = values$caGemini, values)
  })

  # CA Table
  output$caCoordTable <- renderDT(server = FALSE, {
    caPlotFunction()
    DTformat(
      values$CA$wordCoordData,
      size = "100%",
      filename = "CAWordCoordinatesTable",
      pagelength = TRUE,
      left = 1,
      right = 2:ncol(values$CA$wordCoordData),
      numeric = 3:ncol(values$CA$wordCoordData),
      dom = TRUE,
      filter = "top",
      round = 3,
      col_to_remove = values$generalTerm
    )
  })

  output$caContribTable <- renderDT(server = FALSE, {
    caPlotFunction()
    DTformat(
      values$CA$contribData,
      size = "100%",
      filename = "CAWordContributesTable",
      pagelength = TRUE,
      left = 1, # right=2:(ncol(values$CA$contrib)+1),
      numeric = 2:(ncol(values$CA$contrib) + 1),
      dom = TRUE,
      filter = "top",
      round = 3,
      col_to_remove = values$generalTerm
    )
  })

  output$caCosineTable <- renderDT(server = FALSE, {
    caPlotFunction()
    DTformat(
      values$CA$cosineData,
      size = "100%",
      filename = "CAWordCosinesTable",
      pagelength = TRUE,
      left = 1, # right=2:(ncol(values$CA$cosine)+1),
      numeric = 2:(ncol(values$CA$cosine) + 1),
      dom = TRUE,
      filter = "top",
      round = 3,
      col_to_remove = values$generalTerm
    )
  })

  output$caSingularValueTable <- renderDT(server = FALSE, {
    caPlotFunction()
    DTformat(
      values$dfCA,
      size = "100%",
      filename = "CAWordSingualValueTable",
      pagelength = TRUE,
      left = 1, # right=2:3,
      numeric = 2:3,
      dom = TRUE,
      filter = "top",
      round = 2,
      col_to_remove = values$generalTerm
    )
  })

  observeEvent(
    eventExpr = {
      input$caExport
    },
    handlerExpr = {
      file1 <- paste("CAMap-", sys.time(), ".png", sep = "")
      file1 <- destFolder(file1, values$wdTall)
      file2 <- paste("CADendrogram-", sys.time(), ".png", sep = "")
      file2 <- destFolder(file2, values$wdTall)
      plot2png(values$plotCA, filename = file1, type = "plotly", dpi = values$dpi, height = values$h)
      plot2png(values$CADendrogram, filename = file2, type = "vis", dpi = values$dpi, height = values$h)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report

  observeEvent(input$caReport, {
    if (!is.null(values$CA)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "CorrespondenceAnalysis"

      Gem <- values$caGemini %>% string_to_sentence_df()
      list_df <- list(
        Gem,
        values$CA$wordCoordData,
        values$CA$contribData,
        values$CA$cosineData,
        values$dfCA
      )
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileplotCA <- plot2png(
        values$plotCA,
        filename = "CAMap.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$fileCADendrogram <- plot2png(
        values$CADendrogram,
        filename = "CADendrogram.png",
        type = "vis",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileplotCA, res$col),
        c(sheetname = res$sheetname, values$fileCADendrogram, res$col)
      )
      popUp(title = "Correspondence Analysis Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Network ----

  ## Word Network Analysis ----
  netFunction <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$w_networkCoocApply
    },
    valueExpr = {
      switch(
        input$w_groupNet,
        Groups = {
          group <- "doc_id"
        },
        Documents = {
          group <- "doc_id"
        },
        Paragraphs = {
          group <- c("doc_id", "paragraph_id")
        },
        Sentences = {
          group <- c("doc_id", "sentence_id")
        }
      )
      ## check to verify if groups exist or not

      communityRepulsion <- if (!is.null(input$communityRepulsion)) input$communityRepulsion else 0.5

      filtered <- LemmaSelection(values$dfTag) %>% dplyr::filter(docSelected)
      if (
        input$w_groupNet == "Documents" &
          "ungroupDoc_id" %in% names(values$dfTag)
      ) {
        filtered <- backToOriginalGroups(filtered)
      }

      # Choose network type
      net_type <- if (!is.null(input$w_networkType)) input$w_networkType else "cooc"

      if (net_type == "dep") {
        # Dependency-based network
        dep_filter <- if (!is.null(input$w_depRelFilter)) input$w_depRelFilter else "all"
        dep_custom <- if (!is.null(input$w_depRelCustom)) input$w_depRelCustom else NULL
        values$network <- network(
          filtered,
          term = values$generalTerm,
          group = group,
          n = input$nMax,
          minEdges = input$minEdges,
          labelsize = input$labelSize,
          opacity = input$opacity,
          interLinks = input$interLinks,
          normalization = input$normalizationCooc,
          remove.isolated = input$removeIsolated,
          community.repulsion = communityRepulsion,
          seed = values$random_seed,
          cluster = "louvain",
          cooc_type = "dep",
          dep_rel_filter = dep_filter,
          dep_rel_custom = dep_custom
        )
      } else {
        # Co-occurrence network (existing)
        values$network <- network(
          filtered,
          term = values$generalTerm,
          group = group,
          n = input$nMax,
          minEdges = input$minEdges,
          labelsize = input$labelSize,
          opacity = input$opacity,
          interLinks = input$interLinks,
          normalization = input$normalizationCooc,
          remove.isolated = input$removeIsolated,
          community.repulsion = communityRepulsion,
          seed = values$random_seed,
          cluster = "louvain"
        )
      }
      ## end check
      # net=values$network
      # save(net, file="network.rdata")

      values$netVis <- net2vis(
        nodes = values$network$nodes,
        edges = values$network$edges,
        click = TRUE,
        noOverlap = input$noOverlap
      )

      # network$nodes
      if (is.na(values$network$nodes)[1]) {
        values$network$nodesData <- data.frame(
          Word = "",
          Frequency = NA,
          Group = NA,
          "Color Group" = ""
        )
        values$network$edgesData <- data.frame(
          From = NA,
          To = NA,
          "Co-occurence" = 0,
          "Association Index" = 0,
          "Cosine Similarity" = 0,
          "Jaccard Index" = 0,
          "Group From" = 0,
          "Group To" = 0
        )
      } else {
        values$network$nodesData <- values$network$nodes %>%
          select(label, value, group, color) %>%
          rename(
            Word = label,
            Frequency = value,
            Group = group,
            "Color Group" = color
          )

        # network$edges
        values$network$edgesData <- values$network$edges %>%
          select(term_from, term_to, group_from, group_to, s, sA, sC, sJ) %>%
          rename(
            From = term_from,
            To = term_to,
            "Co-occurence" = s,
            "Association Index" = sA,
            "Cosine Similarity" = sC,
            "Jaccard Index" = sJ,
            "Group From" = group_from,
            "Group To" = group_to
          )
        values$gemini_model_parameters <- geminiParameterPrompt(
          values,
          input$sidebarmenu,
          input
        )
      }
    }
  )

  output$w_networkCoocPlot <- renderVisNetwork({
    netFunction()
    values$netVis
  })

  # gemini button for word network
  output$w_networkGeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "", content = values$w_networkGemini, values)
  })

  output$w_networkCoocNodesTable <- renderDT(server = FALSE, {
    netFunction()
    DTformat(
      values$network$nodesData,
      size = "100%",
      filename = "NetworkWordsTable",
      pagelength = TRUE,
      left = NULL,
      right = NULL,
      numeric = NULL,
      dom = TRUE,
      filter = "top",
      col_to_remove = values$generalTerm
    )
  })

  output$w_networkCoocEdgesTable <- renderDT(server = FALSE, {
    netFunction()
    DTformat(
      values$network$edgesData,
      size = "100%",
      filename = "NetworkLinksTable",
      numeric = 6:8,
      round = 4,
      col_to_remove = values$generalTerm
    )
  })

  ## export Network button (JS canvas capture for crisp DPI-aware rendering)
  observeEvent(input$w_networkCoocExport, {
    file <- paste0("Network-Docs-", sys.time(), ".png")
    shinyjs::runjs(sprintf(
      'captureVisExport("w_networkCoocPlot", "%s", %d);',
      file, values$dpi
    ))
  })

  ## Report

  observeEvent(input$w_networkCoocReport, {
    if (!is.null(values$network$nodes)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "WordNetwork"
      Gem <- values$w_networkGemini %>% string_to_sentence_df()
      list_df <- list(
        Gem,
        values$network$nodesData,
        values$network$edgesData
      )
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$filenetVis <- plot2png(
        values$netVis,
        filename = "WordNetwork.png",
        type = "vis",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$filenetVis, res$col)
      )
      popUp(title = "Word Network Analysis Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Click on visNetwork: WORDS IN CONTEXT ----
  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$click
    },
    handlerExpr = {
      if (input$click != "null") {
        showModal(plotModalTermNet(session))
      }
    }
  )

  plotModalTermNet <- function(session) {
    ns <- session$ns
    modalDialog(
      h3(strong(("Words in Context"))),
      DTOutput(ns("wordInContextNet")),
      size = "l",
      easyClose = FALSE,
      footer = tagList(
        actionButton(
          label = "Close",
          inputId = "closePlotModalTermNet",
          style = "color: #ffff;",
          icon = icon("xmark")
        )
      ),
    )
  }

  observeEvent(input$closePlotModalTermNet, {
    removeModal(session = getDefaultReactiveDomain())
    # session$sendCustomMessage("click", 'null') # reset input value to plot modal more times
    resetModalButtons(session = getDefaultReactiveDomain())
  })

  output$wordInContextNet <- renderDT(
    server = FALSE,
    {
      if (!is.null(input$click)) {
        id <- input$click
      } else {
        id <- NULL
      }
      switch(
        input$sidebarmenu,
        "overview" = {
          word_search <- values$WC2VIS$x$nodes$label[
            values$WC2VIS$x$nodes$id == id
          ]
          sentences <- values$dfTag %>%
            filter(docSelected) %>%
            filter(.data[[values$generalTerm]] %in% word_search) %>%
            ungroup() %>%
            select(doc_id, lemma, token, any_of(c("sentence_hl", "sentence")))
        },
        {
          word_search <- values$network$nodes$label[
            values$network$nodes$id %in% id # to check for warning
          ]
          sentences <- values$dfTag %>%
            filter(docSelected) %>%
            filter(lemma %in% word_search) %>%
            ungroup() %>%
            select(doc_id, lemma, token, any_of(c("sentence_hl", "sentence")))
        }
      )

      # find sentences containing the tokens/lemma
      DTformat(
        sentences,
        size = "100%",
        button = TRUE,
        col_to_remove = "lemma"
      )
    },
    escape = FALSE
  )

  ## Click on Dendrogram: WORDS IN CONTEXT ----
  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$click_dend
    },
    handlerExpr = {
      if (input$click_dend != "null") {
        showModal(plotModalTermDend(session))
      }
    }
  )

  plotModalTermDend <- function(session) {
    ns <- session$ns
    modalDialog(
      h3(strong(("Words in Context"))),
      DTOutput(ns("wordInContextDend")),
      size = "l",
      easyClose = TRUE,
      footer = tagList(
        actionButton(
          label = "Close",
          inputId = "closeplotModalTermDend",
          style = "color: #ffff;",
          icon = icon("xmark")
        )
      ),
    )
  }

  observeEvent(input$closeplotModalTermDend, {
    removeModal(session = getDefaultReactiveDomain())
    # session$sendCustomMessage("click_dend",'null') # reset input value to plot modal more times
    resetModalButtons(session = getDefaultReactiveDomain())
  })

  output$wordInContextDend <- renderDT(
    server = FALSE,
    {
      if (!is.null(input$click_dend)) {
        id <- unlist(input$click_dend)
      }
      word_search <- character(0)
      switch(
        input$sidebarmenu,
        "w_clustering" = {
          words_id <- c(
            id,
            unlist(values$WordDendrogram$x$nodes$neib[
              values$WordDendrogram$x$nodes$id == id
            ])
          )
          words <- unlist(values$WordDendrogram$x$nodes$label[
            values$WordDendrogram$x$nodes$id %in% words_id
          ])
          word_search <- words[!is.na(words)]
        },
        "ca" = {
          words_id <- c(
            id,
            unlist(values$CADendrogram$x$nodes$neib[
              values$CADendrogram$x$nodes$id == id
            ])
          )
          words <- unlist(values$CADendrogram$x$nodes$label[
            values$CADendrogram$x$nodes$id %in% words_id
          ])
          word_search <- words[!is.na(words)]
        }
      )

      req(length(word_search) > 0)

      sentences <- values$dfTag %>%
        filter(docSelected) %>%
        filter(lemma %in% word_search) %>%
        ungroup() %>%
        select(doc_id, lemma, token, any_of(c("sentence_hl", "sentence")))

      # find sentences containing the tokens/lemma
      DTformat(
        sentences,
        size = "100%",
        button = TRUE,
        col_to_remove = values$generalTerm
      )
    },
    escape = FALSE
  )

  ## Click on Reinert Dendrogram: WORDS IN CONTEXT ----
  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$click_rein
    },
    handlerExpr = {
      if (input$click_rein != "null") {
        id <- unlist(input$click_rein)
        words_id <- c(
          id,
          unlist(values$ReinertDendrogram$x$nodes$neib[
            values$ReinertDendrogram$x$nodes$id == id
          ])
        )
        words <- unlist(values$ReinertDendrogram$x$nodes$label[
          values$ReinertDendrogram$x$nodes$id %in% words_id
        ])
        word_search <- as.numeric(words[!is.na(words)])
        values$word_search_rein <- word_search

        if (length(word_search) > 0) {
          values$tc_k <- values$tc

          # remove duplicated terms when two or more clusters are aggregated
          values$tc_k$terms <- values$tc_k$terms %>%
            filter(cluster %in% word_search) %>%
            group_by(term) %>%
            slice_min(order_by = p_value, n = 1) %>%
            ungroup()

          values$tc_k$segments <- values$tc_k$segments %>%
            filter(cluster %in% word_search)
          # segments <- values$tc
          values$tc_k <- highlight_segments(values$tc_k, n = 10)

          # values$tc_k$segments <- values$tc_k$segments %>%
          #   group_by(doc_id) %>%
          #   arrange(uc, .by_group = TRUE) %>%
          #   select(doc_id, uc, segment, cluster)
        }
        showModal(plotModalTermRein(session))
      }
    }
  )

  plotModalTermRein <- function(session) {
    ns <- session$ns
    modalDialog(
      tabsetPanel(
        type = "tabs",
        tabPanel(
          title = "Terms by Cluster", icon = icon("table"),
          br(),
          h3(strong(
            (paste0(
              "Terms associated to Cluster(s): ",
              paste0(values$word_search_rein, collapse = ", "),
              collape = ""
            ))
          )),
          plotlyOutput(ns("plotInContextRein"))
        ),
        tabPanel(
          title = "Segments by Cluster", icon = icon("table"),
          br(),
          h3(strong(
            (paste0(
              "Segments associated to Cluster(s): ",
              paste0(values$word_search_rein, collapse = ", "),
              collape = ""
            ))
          )),
          DTOutput(ns("wordInContextRein"))
        ),
      ),
      size = "l",
      easyClose = TRUE,
      footer = tagList(
        actionButton(
          label = "Close",
          inputId = "closeplotModalTermRein",
          style = "color: #ffff;",
          icon = icon("xmark")
        )
      ),
    )
  }

  observeEvent(input$closeplotModalTermRein, {
    removeModal(session = getDefaultReactiveDomain())
    # session$sendCustomMessage("click_dend",'null') # reset input value to plot modal more times
    resetModalButtons(session = getDefaultReactiveDomain())
  })

  output$wordInContextRein <- renderDT(
    server = FALSE,
    {
      req(values$tc_k$segments)
      # find sentences containing the tokens/lemma
      DTformat(
        values$tc_k$segments,
        nrow = 5,
        size = "80%",
        button = TRUE,
        col_to_remove = values$generalTerm
      )
    },
    escape = FALSE
  )

  output$plotInContextRein <- renderPlotly({
    tall::reinPlot(values$tc_k$terms, nPlot = 10)
  })

  ## Thematic Map ----
  TMFunction <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$w_networkTMApply
    },
    valueExpr = {
      switch(
        input$w_groupTM,
        Groups = {
          group <- "doc_id"
        },
        Documents = {
          group <- "doc_id"
        },
        Paragraphs = {
          group <- c("doc_id", "paragraph_id")
        },
        Sentences = {
          group <- c("doc_id", "sentence_id")
        }
      )
      ## check to verify if groups exist or not

      filtered <- LemmaSelection(values$dfTag) %>% dplyr::filter(docSelected)
      if (
        input$w_groupTM == "Documents" &
          "ungroupDoc_id" %in% names(values$dfTag)
      ) {
        filtered <- backToOriginalGroups(filtered)
      }
      values$TM <- tallThematicmap(
        filtered,
        term = values$generalTerm,
        group = group,
        n = input$nMaxTM,
        labelsize = input$labelSizeTM,
        n.labels = input$n.labelsTM,
        opacity = input$opacityTM,
        seed = values$random_seed
      )

      values$TMvis <- net2vis(
        nodes = values$TM$net$nodes,
        edges = values$TM$net$edges,
        click = FALSE,
        noOverlap = TRUE
      )
      values$TMmap <- plotTM(values$TM$df, size = input$labelSizeTM / 10)
      values$TM$ClusterTable <- values$TM$df %>%
        rename(
          "Label" = name,
          "Cluster Frequency" = freq,
          "Num. of Words" = n,
          "Centrality" = centrality,
          "Density" = density,
          "Cluster" = groups,
          "Color" = color
        ) %>%
        select(
          "Label",
          "Cluster",
          "Cluster Frequency",
          "Num. of Words",
          "Centrality",
          "Density",
          "Color"
        )
    }
  )

  output$w_networkTMMapPlot <- renderPlotly({
    TMFunction()
    values$TMmap
  })

  # gemini button for word in context
  output$w_networkTMGeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "", content = values$w_networkTMGemini, values)
  })

  output$w_networkTMNetPlot <- renderVisNetwork({
    TMFunction()
    values$TMvis
  })

  output$w_networkTMClusterTable <- renderDT(server = FALSE, {
    TMFunction()
    df <- values$TM$ClusterTable
    if ("Color" %in% names(df)) {
      df <- df %>% mutate(Color = paste0('<span style="display:inline-block;width:14px;height:14px;background-color:', Color, ';border-radius:2px;margin-right:6px;vertical-align:middle;"></span>', Color))
    }
    DTformat(
      df,
      size = "100%",
      filename = "TMClusterTable",
      pagelength = TRUE,
      left = NULL,
      right = NULL,
      numeric = NULL,
      dom = TRUE,
      filter = "top",
      col_to_remove = values$generalTerm,
      escape = FALSE
    )
  })

  output$w_networkTMWordTable <- renderDT(server = FALSE, {
    TMFunction()
    df <- values$TM$df_lab %>% select(-Cluster_Frequency)
    if ("Color" %in% names(df)) {
      df <- df %>% mutate(Color = paste0('<span style="display:inline-block;width:14px;height:14px;background-color:', Color, ';border-radius:2px;margin-right:6px;vertical-align:middle;"></span>', Color))
    }
    DTformat(
      df,
      size = "100%",
      filename = "TMWordsTable",
      pagelength = TRUE,
      left = NULL,
      right = NULL,
      numeric = NULL,
      dom = TRUE,
      filter = "top",
      col_to_remove = values$generalTerm,
      escape = FALSE
    )
  })

  ## export TM button
  observeEvent(
    eventExpr = {
      input$w_networkTMExport
    },
    handlerExpr = {
      file1 <- paste("TAMap-", sys.time(), ".png", sep = "")
      file1 <- destFolder(file1, values$wdTall)
      file2 <- paste("TANetwork-", sys.time(), ".png", sep = "")
      file2 <- destFolder(file2, values$wdTall)
      plot2png(values$TMmap, filename = file1, type = "plotly", dpi = values$dpi, height = values$h)
      plot2png(values$TMvis, filename = file2, type = "vis", dpi = values$dpi, height = values$h)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report

  observeEvent(input$w_networkTMReport, {
    if (!is.null(values$TM)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "ThematicMap"
      Gem <- values$w_networkTMGemini %>% string_to_sentence_df()
      list_df <- list(
        Gem,
        values$TM$ClusterTable,
        values$TM$df_lab %>% select(-Cluster_Frequency)
      )
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileplotTM <- plot2png(
        values$TMmap,
        filename = "TMMap.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$fileTMNetwork <- plot2png(
        values$TMvis,
        filename = "TMNetwork.png",
        type = "vis",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileplotTM, res$col),
        c(sheetname = res$sheetname, values$fileTMNetwork, res$col)
      )
      popUp(title = "Thematic Map Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## word2vec TRAINING ----

  w2vTrainingFunction <- eventReactive(input$w2vApply, valueExpr = {
    values$w2v_model <- w2vTraining(
      values$dfTag %>% filter(docSelected),
      term = values$generalTerm,
      dim = input$w2vDim,
      iter = input$w2vIter
    )
    values$w2v_stats <- list()
    values$w2v_stats$stats <- summary_stats_embeddings(as.matrix(
      values$w2v_model
    ))
    # values$w2v_stats$distances <- distance_similarity_stats(as.matrix(values$w2v_model))
    values$w2v_stats$pca <- pca_analysis_embeddings(as.matrix(values$w2v_model))
    values$df_EmbeddingDims <- as.matrix(values$w2v_model) %>%
      as.data.frame() %>%
      tibble::rownames_to_column(var = "Word") %>%
      tidyr::pivot_longer(
        cols = -Word,
        names_to = "Dimension",
        values_to = "Value"
      )
    values$df_EmbeddingDims$Dimension <- rep(
      sprintf(
        "D%03d",
        seq_len(
          nrow(values$df_EmbeddingDims) /
            length(unique(values$df_EmbeddingDims$Word))
        )
      ),
      length(unique(values$df_EmbeddingDims$Word))
    )
    values$w2vBoxplot <- plot_ly(
      data = values$df_EmbeddingDims,
      y = ~Value,
      x = ~Dimension,
      type = "box",
      boxpoints = "outliers",
      hoverinfo = "x+y"
    ) %>%
      layout(
        title = "Distribution of embedding values by dimension",
        xaxis = list(title = "Dimension", tickangle = -45),
        yaxis = list(title = "Value")
      )
    values$w2vPCA <- plot_ly(
      x = sprintf("PC%03d", seq_along(values$w2v_stats$pca)),
      y = values$w2v_stats$pca * 100,
      type = "bar"
    ) %>%
      layout(
        title = "Variance Explained by Principal Components",
        xaxis = list(title = "Principal Components"),
        yaxis = list(title = "Variance Proportion"),
        bargap = 0.2
      )
  })

  output$w_word2vecBoxplot <- renderPlotly({
    w2vTrainingFunction()
    values$w2vBoxplot
  })

  output$w_word2vecPCA <- renderPlotly({
    w2vTrainingFunction()
    values$w2vPCA
  })

  output$w_word2vecTable <- renderDT(server = FALSE, {
    w2vTrainingFunction()
    DTformat(
      values$w2v_stats$stats,
      size = "80%",
      filename = "WordEmbeddingTable",
      pagelength = TRUE,
      left = NULL,
      right = NULL,
      numeric = NULL,
      dom = TRUE,
      filter = "none"
    )
  })

  ## export Embedding button
  observeEvent(
    eventExpr = {
      input$w2vSave
    },
    handlerExpr = {
      file1 <- paste("WEmatrix-", sys.time(), ".csv", sep = "")
      file1 <- destFolder(file1, values$wdTall)
      file2 <- paste("WEboxplot-", sys.time(), ".png", sep = "")
      file2 <- destFolder(file2, values$wdTall)
      file3 <- paste("WEpca-", sys.time(), ".png", sep = "")
      file3 <- destFolder(file3, values$wdTall)
      write.csv(as.matrix(values$w2v_model), file = file1)
      plot2png(values$w2vBoxplot, filename = file2, type = "plotly", dpi = values$dpi, height = values$h)
      plot2png(values$w2vPCA, filename = file3, type = "plotly", dpi = values$dpi, height = values$h)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report
  observeEvent(input$w2vReport, {
    if (!is.null(values$w2v_model)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "EmbeddingTraining"
      list_df <- list(values$w2v_stats$stats)
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileplotw2vBoxplot <- plot2png(
        values$w2vBoxplot,
        filename = "w2vBoxplot.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$fileplotw2vPCA <- plot2png(
        values$w2vPCA,
        filename = "w2vPCA.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileplotw2vBoxplot, res$col),
        c(sheetname = res$sheetname, values$fileplotw2vPCA, res$col)
      )
      popUp(title = "Word Embedding Training Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## word2vec SIMILARITY ----

  w2vSimilarity <- eventReactive(input$w_w2v_similarityApply, {
    w2vTrainingFunction()
    values$w2vNetwork <- w2vNetwork(
      values$w2v_model,
      values$dfTag,
      term = values$generalTerm,
      n = input$w_w2v_similarityN
    )
    values$umapDf <- w2vUMAP(
      values$w2v_model,
      top_words = values$w2vNetwork$top_words
    )
    df_adj <- adjust_labels_iterative_with_opacity(
      values$umapDf,
      min_dist = 0.3,
      max_iter = 50,
      shift_step = 0.05,
      alpha_low = 0.4
    )
    values$w2vUMAPplot <- plot_ly(
      data = df_adj,
      x = ~x,
      y = ~y,
      type = "scatter",
      mode = "text",
      text = ~word,
      textfont = list(
        size = input$w_w2v_font_size
      ),
      textposition = "top center"
    ) %>%
      style(
        textfont = list(color = df_adj$text_color, size = input$w_w2v_font_size)
      ) %>%
      layout(
        # title = list(text = "CBOW Embeddings Visualization (UMAP)", x = 0.5),
        xaxis = list(
          title = "UMAP Dimension 1",
          zeroline = FALSE,
          showgrid = FALSE
        ),
        yaxis = list(
          title = "UMAP Dimension 2",
          zeroline = FALSE,
          showgrid = FALSE
        ),
        hovermode = "closest"
      )
  })

  output$w_w2v_Selected <- renderUI({
    w2vTrainingFunction()
    #nodesId <- sort(values$w2vNetwork$top_words)
    nodesId <- sort(values$w2vNetwork$nodes$id)
    selectInput(
      "w2v_selected_node",
      "Select word to highlight:",
      choices = c("", nodesId),
      selected = ""
    )
  })

  output$w_w2vNetworkplot <- renderVisNetwork({
    w2vSimilarity()
    values$w2vNetworkPlot <- w2v2Vis(
      nodes = values$w2vNetwork$nodes,
      edges = values$w2vNetwork$edges,
      size = 20,
      labelsize = input$w_w2v_font_size,
      overlap = input$w_w2v_overlap
    )
    values$w2vNetworkPlot
  })

  # gemini button for word embedding similarity
  output$w_w2vGeminiUI <- renderUI({
    values$gemini_model_parameters <- geminiParameterPrompt(
      values,
      input$sidebarmenu,
      input
    )
    geminiOutput(title = "", content = values$w_w2vGemini, values)
  })

  observe({
    visNetworkProxy("w_w2vNetworkplot") %>%
      visSelectNodes(id = input$w2v_selected_node) %>%
      visFit(nodes = input$w2v_selected_node)
  })

  observeEvent(
    eventExpr = {
      input$w_w2v_Back
    },
    handlerExpr = {
      visNetworkProxy("w_w2vNetworkplot") %>%
        visSelectNodes(id = values$w2vNetwork$top_words) %>%
        visFit(nodes = NULL)
    }
  )

  output$w_w2vUMAPplot <- renderPlotly({
    w2vSimilarity()
    values$w2vUMAPplot
  })

  ## export Embedding Similarity button
  observeEvent(
    eventExpr = {
      input$w_w2v_similarityExport
    },
    handlerExpr = {
      file1 <- paste("WENetplot-", sys.time(), ".png", sep = "")
      file1 <- destFolder(file1, values$wdTall)
      file2 <- paste("WEumap-", sys.time(), ".png", sep = "")
      file2 <- destFolder(file2, values$wdTall)
      plot2png(values$w2vNetworkPlot, filename = file1, type = "vis", dpi = values$dpi, height = values$h)
      plot2png(values$w2vUMAPplot, filename = file2, type = "plotly", dpi = values$dpi, height = values$h)
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Report
  observeEvent(input$w_w2v_similarityReport, {
    if (!is.null(values$umapDf)) {
      popUp(title = NULL, type = "waiting")
      sheetname <- "EmbeddingSimilarity"
      Gem <- values$w_w2vGemini %>% string_to_sentence_df()
      list_df <- list(
        Gem,
        values$w2vNetwork$edges,
        values$umapDf
      )
      res <- addDataScreenWb(list_df, wb = values$wb, sheetname = sheetname)
      # values$wb <- res$wb
      owd <- setwd(tempdir())
      on.exit(setwd(owd))
      values$fileplotw2vNet <- plot2png(
        values$w2vNetworkPlot,
        filename = "w2vNetworkPlot.png",
        type = "vis",
        dpi = values$report_dpi, height = values$h
      )
      values$fileplotw2vUMAP <- plot2png(
        values$w2vUMAPplot,
        filename = "w2vUMAPPlot.png",
        type = "plotly",
        dpi = values$report_dpi, height = values$h
      )
      values$list_file <- rbind(
        values$list_file,
        c(sheetname = res$sheetname, values$fileplotw2vNet, res$col),
        c(sheetname = res$sheetname, values$fileplotw2vUMAP, res$col)
      )
      popUp(title = "Word Embedding Similarity Results", type = "success")
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

}

importUI <- function() {
  tabItem(
    tabName = "import_tx",
    fluidPage(
      # Page Header
      fluidRow(
        column(
          12,
          div(
            h2(
              icon("file-lines"),
              strong("Import Texts"),
              style = "color: #4F7942; text-align: center; margin-bottom: 20px;"
            )
          )
        )
      ),

      fluidRow(
        tabsetPanel(
          type = "tabs",
          tabPanel(
            title = "Corpus",
            icon = icon("database"),
            br(),
            fluidRow(
              # Main Content Area - Data Table
              column(
                9,
                div(
                  style = "margin-top: 20px;",
                  shinycssloaders::withSpinner(
                    DT::DTOutput("dataImported"),
                    color = getOption("spinner.color", default = "#4F7942")
                  )
                )
              ),

              # Sidebar - Control Panel
              column(
                3,
                div(
                  style = "margin-top: 20px;",
                  box(
                    width = 12,
                    solidHeader = TRUE,
                    status = "success",

                    # Box Header
                    div(
                      class = "box-header with-border",
                      h4(
                        icon("upload"),
                        strong("Import Controls"),
                        style = "margin: 0; color: #4F7942;"
                      )
                    ),

                    # Box Body
                    div(
                      class = "box-body",

                      # Import Button Area
                      div(
                        style = "margin-bottom: 20px;",
                        uiOutput("runButton")
                      ),

                      # Action Buttons (shown after import)
                      conditionalPanel(
                        condition = "input.runImport > 0",
                        tags$hr(
                          style = "border-color: #e0e0e0; margin: 20px 0;"
                        ),

                        div(
                          h5(
                            strong("Actions"),
                            style = "color: #666; margin-bottom: 15px;"
                          ),

                          fluidRow(
                            column(
                              6,
                              div(
                                align = "center",
                                div(
                                  title = "Export raw text(s) in Excel",
                                  actionButton(
                                    inputId = "collection.save",
                                    label = NULL,
                                    icon = icon(
                                      "download"
                                    ),
                                    style = "display: block; height: 45px; width: 45px;
                                         border-radius: 50%; border: 2px solid #4F7942;
                                         background: linear-gradient(135deg, #6CC283, #4F7942);
                                         color: white; transition: all 0.3s ease;
                                         box-shadow: 0 2px 4px rgba(0,0,0,0.1);"
                                  )
                                )
                              )
                            ),

                            column(
                              6,
                              div(
                                align = "center",
                                div(
                                  title = "Back to imported text(s)",
                                  do.call(
                                    "actionButton",
                                    c(
                                      back_bttn,
                                      list(
                                        inputId = "importTextBack",
                                        style = "display: block; height: 45px; width: 45px;
                                             border-radius: 50%; border: 2px solid #4F7942;
                                             background: linear-gradient(135deg, #6CC283, #4F7942);
                                             color: white; transition: all 0.3s ease;
                                             box-shadow: 0 2px 4px rgba(0,0,0,0.1);"
                                      )
                                    )
                                  )
                                )
                              )
                            )
                          )
                        )
                      )
                    )
                  )
                )
              )
            )
          ),
          tabPanel(
            title = "Info & References",
            icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(
                  12,
                  div(
                    style = "padding: 30px; background: white; border-radius: 8px;
                         box-shadow: 0 2px 4px rgba(0,0,0,0.05); margin-top: 20px;",
                    HTML(infoTexts$importmenu)
                  )
                )
              )
            )
          )
        )
      )
    )
  )
}


importServer <- function(input, output, session, values, statsValues) {
  ## Code to reset shiny app
  reset_rv <- reactiveVal(value = 0L)
  session$onSessionEnded(function() {
    # x <- Inf
    x <- isolate(reset_rv())

    if (!is.null(x)) {
      if (x == 0) {
        stopApp()
      }
    }
  })
  ###

  output$resetButton <- renderUI({
    reset_bttn <- list(
      label = NULL,
      style = "margin-top: -8px; font-size: 8px; border-radius:2%",
      # style ="display:block; height: 37px; width: 37px; border-radius: 50%; border: 3px; margin-top: 15px",
      icon = icon("arrows-rotate")
    )
    do.call(
      "actionButton",
      c(
        reset_bttn,
        list(
          inputId = "resetApp"
        )
      )
    )
  })

  observeEvent(input$resetApp, {
    ask_confirmation(
      inputId = "reset_confirmation",
      title = "Restart TALL",
      text = HTML(
        "Restarting TAll will result in the loss of all analyses currently in progress<br><br>
                  <b>Do you want to confirm?</b>"
      ),
      html = TRUE,
      type = "warning",
      btn_labels = c("CANCEL", "CONFIRM")
    )
  })

  observeEvent(input$reset_confirmation, {
    if (isTRUE(input$reset_confirmation)) {
      reset_rv(input$resApp)
      session$reload()
      param_stay_page <<- TRUE
    }
  })

  if (param_stay_page) {
    updateTabItems(session, "sidebarmenu", "import_tx")
    param_stay_page_newPT <<- FALSE
  }

  ## observe Gemini copy2clipboard button
  observeEvent(input$copy_btn, {
    content <- geminiSave(values, input$sidebarmenu)
    copy_to_clipboard(content)
  })

  ## TALL AI Save button (browser download)
  output$save_btn <- downloadHandler(
    filename = function() {
      paste0("TallAI_", input$sidebarmenu, "_", format(Sys.time(), "%Y-%m-%d_%H-%M-%S"), ".txt")
    },
    content = function(file) {
      txtOutput <- geminiSave(values, input$sidebarmenu)
      readr::write_lines(txtOutput, file = file)
    }
  )

  ## observe gemini generate button (async)
  observeEvent(input$gemini_btn, {
    values$gemini_additional <- input$gemini_additional ## additional info to Gemini prompt
    values <- geminiWaitingMessage(values, input$sidebarmenu)

    # Sync: prepare images and prompt
    prep <- geminiPrepare(
      values, input$sidebarmenu, values$gemini_additional,
      values$gemini_model_parameters, input
    )

    if (!is.null(prep$error)) {
      values[[prep$field]] <- prep$error
      return()
    }

    # Snapshot for the future worker
    image <- prep$image
    prompt <- prep$prompt
    model <- prep$model
    outputSize <- prep$outputSize
    api_key <- prep$api_key
    field <- prep$field

    # Async: only the API call runs in background
    promises::future_promise({
      gemini_ai(
        image = image, prompt = prompt, model = model,
        outputSize = outputSize, api_key = api_key
      )
    }, seed = TRUE) %...>%
      (function(result) {
        values[[field]] <- result
      }) %...!%
      (function(err) {
        values[[field]] <- paste("\u26a0\ufe0f Error:", conditionMessage(err))
      })

    NULL # Don't block the observer
  })

  ### IMPORT ----
  output$runButton <- renderUI({
    if (!isTRUE(values$resetNeed)) {
      list(
        selectInput(
          "load",
          "Please, choose what to do",
          choices = c(
            " " = "null",
            "Load text files" = "import",
            "Load file from Biblioshiny" = "biblioshiny",
            "Load Tall structured files" = "load_tall",
            "Wikipedia pages" = "wiki",
            "Use a sample collection" = "demo"
          ),
          selected = "null"
        ),
        conditionalPanel(
          condition = "input.load == 'wiki'",
          textInput(
            inputId = "wikiWord",
            label = "Search Wikipedia",
            value = NULL
          ),
          sliderTextInput(
            inputId = "wikiN",
            label = "Pages",
            choices = seq(1, 20),
            selected = 1,
            animate = TRUE
          ),
          helpText(
            em(
              "By specifying a search phrase in 'Search Wikipedia',
                                         the content of up to 20 Wikipedia pages can be downloaded."
            ),
            br(),
            br(),
            em(
              "The content of each wiki page will be stored in the 'text' column.
                                         In addition, the page title, abstract and url will also be stored."
            ),
            br(),
            br(),
            em("The page title will be used as the 'doc_id'.")
          )
        ),
        conditionalPanel(
          condition = "input.load == 'import'",
          fluidRow(
            column(
              6,
              selectizeInput(
                "ext",
                label = "File format",
                choices = c(
                  "txt" = "txt",
                  "csv" = "csv",
                  "excel" = "xlsx",
                  "pdf" = "pdf"
                ),
                tags$style("height: 50px")
              )
            ),
            conditionalPanel(
              condition = "input.load == 'import' & input.ext=='csv'",
              column(
                6,
                selectizeInput(
                  "line_sep",
                  label = "CSV Separator",
                  choices = c(
                    " , " = ",",
                    " ; " = ";"
                  ),
                  tags$style("height: 50px")
                )
              )
            )
          )
        ),
        conditionalPanel(
          condition = "input.load == 'import' | input.load == 'biblioshiny'",
          uiOutput("file_rawUI"),
          uiOutput("biblioChoiceUI"), # select text field in file from bibliometrix
          uiOutput("biblioRunUI"), # apply
          uiOutput(outputId = "infoImport"),
          conditionalPanel(
            condition = "input.ext == 'xlsx' ||  input.ext =='csv'",
            uiOutput(outputId = "infoTextLabel")
          )
        ),
        conditionalPanel(
          condition = "input.load=='demo'",
          selectInput(
            "demo_file",
            label = "Select sample texts",
            choices = c(
              "BBC news" = "bbc",
              "Bibliometrix" = "bibliometrix",
              "US Airlines Tweets" = "usairlines"
            ),
            selected = "bibliometrix"
          ),
          conditionalPanel(
            condition = "input.demo_file=='bibliometrix'",
            helpText(
              em(
                "The dataset is composed of a collection of 444 scientific articles written in English
                                           in which the authors used the Bibliometrix R package to perform systematic literature reviews."
              ),
              br(),
              br(),
              em(
                "The textual data consists of the article abstracts, while the additional information includes
                                           metadata such as the list of co-authors, first author, year of publication, and journal name."
              ),
              br(),
              br(),
              em("The abstracts have already been tokenized and POS tagged.")
            )
          ),
          conditionalPanel(
            condition = "input.demo_file=='bbc'",
            helpText(
              em(
                "A collection of 386 short news stories published in the entertainment section of the BBC News website."
              ),
              br(),
              br(),
              em("The texts are in English.")
            )
          ),
          conditionalPanel(
            condition = "input.demo_file=='usairlines'",
            helpText(
              em(
                "The dataset is the 'Twitter US Airline Sentiment' collection, a publicly available and widely used dataset originally hosted on Kaggle. It contains tweets collected during February 2015 that pertain to major U.S.-based airlines."
              ),
              br(),
              br(),
              em("The texts are in English.")
            )
          )
        ),
        conditionalPanel(
          condition = "input.load == 'load_tall'",
          helpText(em("Load a collection previously exported from Tall")),
          fileInput(
            "file1",
            "Choose a file",
            multiple = FALSE,
            accept = c(
              ".tall"
            )
          )
        ),
        conditionalPanel(
          condition = "input.load != 'null'",
          div(
            align = "center",
            width = 12,
            actionButton(
              inputId = "runImport",
              label = div(
                icon("play"),
                strong("START")
              ),
              icon = NULL,
              style = "border-radius: 20px; border-width: 1px;
                                                                    font-size: 17px; color: #ffff;"
            )
          )
        )
      )
    } else {
      list(
        textAreaInput(
          inputId = "corpus_description",
          label = "Please provide a brief description of your corpus (e.g., source, type of content, domain) to improve prompts for the TALL AI Assistant:",
          placeholder = "Example: The corpus consists of 150 academic articles from biomedical journals published between 2015 and 2020...",
          value = values$corpus_description,
          rows = 8,
          width = "100%"
        ),
        div(
          align = "center",
          width = 12,
          actionButton(
            inputId = "applyCorpusDescription",
            label = div(
              icon("check"),
              strong("Update")
            ),
            icon = NULL,
            style = "border-radius: 20px; border-width: 1px;
                                                                    font-size: 17px; color: #ffff;"
          )
        ),
        br(),
        helpText(
          em(
            "To load a new text collection, it is necessary to reset the app."
          )
        ),
        br(),
        div(
          align = "center",
          width = 12,
          actionButton(
            inputId = "runReset2",
            label = div(
              icon("arrows-rotate"),
              strong("RESET")
            ),
            icon = NULL,
            style = "border-radius: 20px; border-width: 1px;
                                                                    font-size: 17px; color: #ffff;"
          )
        )
      )
    }
  })

  observeEvent(input$runImport, {
    values$corpus_description <- input$corpus_description
    updateTextAreaInput(
      session,
      "corpus_description",
      value = values$corpus_description
    )
  })

  observeEvent(input$applyCorpusDescription, {
    values$corpus_description <- input$corpus_description
    show_alert(
      title = "Corpus description updated",
      text = "The corpus description has been successfully updated.",
      type = "success"
    )
  })

  observeEvent(input$runReset2, {
    ask_confirmation(
      inputId = "reset_confirmation2",
      title = "Restart TALL",
      text = HTML(
        "Restarting TAll will result in the loss of all analyses currently in progress<br><br>
                  <b>Do you want to confirm?</b>"
      ),
      html = TRUE,
      type = "warning",
      btn_labels = c("CANCEL", "CONFIRM")
    )
  })

  observeEvent(input$reset_confirmation2, {
    if (isTRUE(input$reset_confirmation2)) {
      reset_rv(input$runReset2)
      session$reload()
      param_stay_page <<- TRUE
    }
  })

  output$file_rawUI <- renderUI({
    switch(
      input$ext,
      txt = {
        ext <- c("text/plain", ".txt", ".zip")
      },
      csv = {
        ext <- c("text/csv", ".csv", ".zip")
      },
      xlsx = {
        ext <- c("excel", ".xlsx", ".xls", ".zip")
      },
      pdf = {
        ext <- c(".pdf", ".zip")
      }
    )

    if (input$load == "biblioshiny") {
      ext <- c("text/csv", ".csv", ".zip")
    }

    fileInput(
      "file_raw",
      "Select file(s) containing text",
      multiple = TRUE,
      accept = ext,
      placeholder = "No file(s) selected"
    )
  })

  output$infoTextLabel <- renderUI({
    shinyWidgets::alert(
      icon("warning"),
      tags$b("Warning!"),
      br(),
      HTML(
        "The column including text(s) in your CSV or EXCEL file must be named <b>text</b>"
      ),
      status = "warning"
    )
    # }
  })

  ### dataImported ----
  DATAloading <- eventReactive(input$runImport, {
    switch(
      input$load,
      import = {
        if (!is.null(req(input$file_raw))) {
          file <- input$file_raw
          txt <- read_files(
            file,
            ext = input$ext,
            subfolder = FALSE,
            line_sep = input$line_sep
          )
          txt <- txt %>% clean_text() %>% trim_text_columns() ## clean text before tokenization
          values$menu <- 0
          values$custom_lists <- NULL
          values$txt <- txt %>%
            mutate(
              text = removeHTMLTags(text),
              text_original = text
            ) %>%
            arrange(doc_id)
          values$resetNeed <- TRUE
        }
      },
      biblioshiny = {
        if (!is.null(req(input$file_raw))) {
          file <- input$file_raw
          values$txt <- read_files(
            file,
            ext = "csv",
            subfolder = FALSE,
            line_sep = ","
          )
          values$biblioshiny <- TRUE
        }
      },
      load_tall = {
        req(input$file1)
        file_tall <- input$file1$datapath
        load(file_tall)
        values$menu <- menu
        values$dfTag <- dfTag
        values$txt <- rebuild_documents(dfTag)
        values$custom_lists <- custom_lists
        values$language <- language
        values$treebank <- treebank
        values$D <- D
        values$where <- where
        cd <- get0("corpus_description", ifnotfound = "")
        values$corpus_description <- if (is.null(cd) || cd == "") "" else cd

        statsValues <- updateStats(
          values$dfTag,
          term = values$generalTerm,
          statsValues
        )
        if (exists("generalTerm")) {
          values$generalTerm <- generalTerm
        }
        values$resetNeed <- TRUE
        # values$metadata <- metadata
        if (values$menu == 1) {
          updateTabItems(session, "sidebarmenu", "custTermList")
        }
        if (values$menu > 1) {
          updateTabItems(session, "sidebarmenu", "posTagSelect")
        }
        if (ncol(values$dfTag) > 1) {
          showModal(loadTallgModal(session))
        }
      },
      demo = {
        switch(
          input$demo_file,
          bibliometrix = {
            file_tall <- loadSampleCollection("bibliometrix")
            load(file_tall)
            values$menu <- menu
            values$dfTag <- dfTag
            values$txt <- rebuild_documents(dfTag)
            values$custom_lists <- custom_lists
            values$language <- language
            values$D <- D
            values$where <- where
            values$corpus_description <- "The dataset is composed of a collection of 444 scientific articles written in English in which the authors used the Bibliometrix R package to perform systematic literature reviews.\n The textual data consists of the article abstracts, while the additional information includes metadata such as the list of co-authors, the first author, the year of publication, and the journal name."
            values$resetNeed <- TRUE
            statsValues <- updateStats(
              values$dfTag,
              term = values$generalTerm,
              statsValues
            )

            if (values$menu == 1) {
              updateTabItems(session, "sidebarmenu", "custTermList")
            }
            if (values$menu > 1) {
              updateTabItems(session, "sidebarmenu", "posTagSelect")
            }
            if (ncol(values$dfTag) > 1) {
              showModal(loadTallgModal(session))
            }
          },
          bbc = {
            file_tall <- loadSampleCollection("bbc")
            files <- list(name = "bbc.zip", datapath = file_tall)
            txt <- read_files(files, ext = "txt", subfolder = FALSE)
            values$menu <- 0
            values$custom_lists <- NULL
            values$corpus_description <- "A collection of 386 short news published in the entertainment section of the BBC News website."
            values$resetNeed <- TRUE
            values$txt <- txt %>%
              mutate(text_original = text) %>%
              arrange(doc_id)
          },
          usairlines = {
            file_tall <- loadSampleCollection("usairlines")
            files <- list(name = "usairlines.zip", datapath = file_tall)
            txt <- read_files(files, ext = "csv", subfolder = FALSE)
            values$menu <- 0
            values$custom_lists <- NULL
            values$corpus_description <- paste0(
              "The dataset under analysis is the 'Twitter US Airline Sentiment' collection, a publicly available",
              " and widely used dataset originally hosted on Kaggle. It contains tweets collected during February 2015",
              " that pertain to major U.S.-based airlines. The primary objective of this dataset is to capture and analyze",
              " the sentiment expressed by travelers regarding their airline experiences. Each tweet is annotated with ",
              "a sentiment labelâ€”positive, neutral, or negativeâ€”based on the emotional tone conveyed in the text. ",
              "This collection is frequently employed in natural language processing tasks, particularly for training ",
              "and evaluating sentiment classification models. In this context, the dataset will be used to explore how ",
              "airline passengers articulated their opinions and emotions on Twitter during the specified period."
            )
            values$resetNeed <- TRUE
            values$txt <- txt %>%
              mutate(text_original = text) %>%
              arrange(doc_id)
          }
        )
      },
      wiki = {
        df <- wikiSearch(input$wikiWord, n = as.numeric(input$wikiN))
        if (is.null(df)) {
          show_alert(
            title = "No results found!",
            text = "It seems there are no Wikipedia pages matching your search.",
            type = "error"
          )
          values$resetNeed <- FALSE
        } else {
          values$menu <- 0
          values$custom_lists <- NULL
          values$txt <- wikiExtract(df) %>%
            mutate(text_original = text) %>%
            rename(
              doc_id = title,
              doc_selected = selected
            )

          values$resetNeed <- TRUE
        }
      }
    )
  })

  observeEvent(
    eventExpr = {
      values$biblioshiny
    },
    handlerExpr = {
      tallFields <- intersect(c("AB", "TI", "DE"), names(values$txt))
      output$biblioChoiceUI <- renderUI({
        selectizeInput(
          inputId = "biblioChoice",
          label = "Choose the text column",
          choices = tallFields,
          selected = NULL,
          multiple = FALSE
        )
      })
      hide("runImport")
      output$biblioRunUI <- renderUI({
        div(
          align = "center",
          width = 12,
          br(),
          actionButton(
            inputId = "biblioRun",
            label = div(
              icon("play"),
              strong("Apply")
            ),
            icon = NULL,
            style = "border-radius: 20px; border-width: 1px;
                                                                    font-size: 17px; color: #ffff;"
          )
        )
      })
    }
  )

  observeEvent(
    eventExpr = {
      input$biblioRun
    },
    handlerExpr = {
      req(input$biblioChoice)
      values$txt <- values$txt %>%
        mutate(
          text = removeHTMLTags(!!sym(input$biblioChoice)),
          text_original = text
        ) %>%
        clean_text() %>%
        trim_text_columns() %>%
        mutate(doc_selected = TRUE) %>%
        arrange(doc_id)
      values$menu <- 0
      values$custom_lists <- NULL
      values$resetNeed <- TRUE

      output$dataImported <- DT::renderDT({
        req(values$txt)
        if (nrow(values$txt) > 0) {
          col_names <- c("doc_selected", "text_original")
          # check if text exists in values$txt and then apply or not the mutate

          DTformat(
            values$txt %>%
              dplyr::filter(doc_selected) %>%
              mutate(across(
                any_of("text"),
                ~ paste0(substr(., 1, 500), "...")
              )) %>%
              select(doc_id, any_of("text"), everything()) %>%
              select(!any_of(col_names)),
            left = 3,
            nrow = 5,
            filter = "none",
            button = TRUE,
            delete = TRUE
          )
        }
      })
    }
  )

  # Trigger DATAloading when runImport is clicked
  observeEvent(input$runImport, {
    DATAloading()
  })

  output$dataImported <- DT::renderDT({
    req(values$txt)
    if (nrow(values$txt) > 0) {
      col_names <- c("doc_selected", "text_original")
      # check if text exists in values$txt and then apply or not the mutate

      DTformat(
        values$txt %>%
          dplyr::filter(doc_selected) %>%
          mutate(across(any_of("text"), ~ paste0(substr(., 1, 500), "..."))) %>%
          select(doc_id, any_of("text"), everything()) %>%
          select(!any_of(col_names)),
        left = 3,
        nrow = 5,
        filter = "none",
        button = TRUE,
        delete = TRUE
      )
    }
  })

  ### shortpath for folder path ----
  output$folder <- renderUI({
    path <- shortpath(values$path)
    if (is.null(path)) {
      path <- " --- "
    }
    HTML(paste("<pre class='tab'>", path, sep = ""))
  })

  loadTallgModal <- function(session) {
    ns <- session$ns
    modalDialog(
      title = div(
        style = "background: linear-gradient(135deg, #5cb85c 0%, #4F7942 100%);
               color: white;
               padding: 20px;
               margin: -15px -15px 20px -15px;
               border-radius: 5px 5px 0 0;",
        h3(
          icon("info-circle", lib = "font-awesome"),
          strong(" TALL Data Overview"),
          style = "margin: 0; color: white;"
        )
      ),
      div(
        style = "padding: 10px;",
        uiOutput("loadSynthesis")
      ),
      size = "l",
      easyClose = TRUE,
      footer = div(
        style = "background-color: #f8f9fa;
               padding: 15px;
               margin: 20px -15px -15px -15px;
               border-top: 2px solid #dee2e6;",
        tagList(
          actionButton(
            label = tagList(
              icon("table-list"),
              " Custom Lists"
            ),
            inputId = "modalCustomLists",
            style = "background-color: #5cb85c;
                   color: white;
                   border: none;
                   padding: 10px 20px;
                   border-radius: 5px;
                   font-weight: bold;
                   margin-right: 10px;"
          ),
          actionButton(
            label = tagList(icon("xmark"), " Close"),
            inputId = "closeModalCustomLists",
            style = "background-color: #6c757d;
                   color: white;
                   border: none;
                   padding: 10px 20px;
                   border-radius: 5px;
                   font-weight: bold;"
          )
        )
      )
    )
  }

  observeEvent(input$closeModalCustomLists, {
    removeModal(session = getDefaultReactiveDomain())
  })

  output$loadSynthesis <- renderUI({
    ndocs <- length(unique(values$dfTag$doc_id))

    # Prepare data
    txt1 <- paste0(ndocs, " documents")
    txt2 <- values$D
    txt2b_lang <- tools::toTitleCase(values$language)
    txt2b_tree <- values$treebank
    txt2c <- values$corpus_description

    if (!is.null(dim(values$custom_lists))) {
      ncust <- nrow(values$custom_lists)
      txt3 <- paste0(ncust, " words")
      custom_status <- "included"
      custom_badge_color <- "#5cb85c"
    } else {
      txt3 <- "Not included"
      custom_status <- "not-included"
      custom_badge_color <- "#6c757d"
    }

    upos <- values$dfTag %>%
      select(upos) %>%
      pull() %>%
      unique()

    if ("MULTIWORD" %in% upos) {
      txt3bis <- "Included"
      multi_badge_color <- "#5cb85c"
    } else {
      txt3bis <- "Not included"
      multi_badge_color <- "#6c757d"
    }

    items <- toupper(c(
      "email",
      "url",
      "hash",
      "emoji",
      "ip_address",
      "mention"
    ))

    if (length(intersect(items, upos)) > 0) {
      txt3ter <- paste0(
        tools::toTitleCase(tolower(intersect(items, upos))),
        collapse = ", "
      )
      entity_badge_color <- "#5cb85c"
    } else {
      txt3ter <- "Not included"
      entity_badge_color <- "#6c757d"
    }

    txt4 <- values$where

    # Create modern styled layout
    tagList(
      # Main Info Card
      div(
        style = "background: white;
               border-left: 4px solid #5cb85c;
               padding: 15px 20px;
               margin-bottom: 15px;
               box-shadow: 0 2px 4px rgba(0,0,0,0.1);
               border-radius: 5px;",
        div(
          style = "display: flex; align-items: center; margin-bottom: 10px;",
          icon(
            "file-text",
            lib = "font-awesome",
            style = "font-size: 24px; color: #5cb85c; margin-right: 10px;"
          ),
          h4(strong("Corpus Information"), style = "margin: 0; color: #2c3e50;")
        ),
        hr(style = "margin: 10px 0; border-color: #e0e0e0;"),
        div(
          style = "display: grid; grid-template-columns: 1fr 1fr; gap: 15px;",
          # Documents
          div(
            style = "padding: 10px;",
            div(
              style = "color: #6c757d; font-size: 12px; font-weight: 600; text-transform: uppercase; margin-bottom: 5px;",
              "Documents"
            ),
            div(
              style = "font-size: 24px; font-weight: bold; color: #2c3e50;",
              txt1
            )
          ),
          # Last Modified
          div(
            style = "padding: 10px;",
            div(
              style = "color: #6c757d; font-size: 12px; font-weight: 600; text-transform: uppercase; margin-bottom: 5px;",
              "Last Modified"
            ),
            div(
              style = "font-size: 16px; color: #2c3e50;",
              txt2
            )
          )
        )
      ),

      # Language & Treebank Card
      div(
        style = "background: white;
               border-left: 4px solid #17a2b8;
               padding: 15px 20px;
               margin-bottom: 15px;
               box-shadow: 0 2px 4px rgba(0,0,0,0.1);
               border-radius: 5px;",
        div(
          style = "display: flex; align-items: center; margin-bottom: 10px;",
          icon(
            "globe",
            style = "font-size: 20px; color: #17a2b8; margin-right: 10px;"
          ),
          h4(strong("Language Settings"), style = "margin: 0; color: #2c3e50;")
        ),
        hr(style = "margin: 10px 0; border-color: #e0e0e0;"),
        div(
          style = "display: grid; grid-template-columns: 1fr 1fr; gap: 15px;",
          div(
            style = "padding: 10px;",
            div(
              style = "color: #6c757d; font-size: 12px; font-weight: 600; text-transform: uppercase; margin-bottom: 5px;",
              "Language"
            ),
            span(
              txt2b_lang,
              style = "background-color: #17a2b8;
                     color: white;
                     padding: 5px 12px;
                     border-radius: 15px;
                     font-weight: 600;
                     font-size: 14px;"
            )
          ),
          div(
            style = "padding: 10px;",
            div(
              style = "color: #6c757d; font-size: 12px; font-weight: 600; text-transform: uppercase; margin-bottom: 5px;",
              "Treebank"
            ),
            span(
              txt2b_tree,
              style = "background-color: #17a2b8;
                     color: white;
                     padding: 5px 12px;
                     border-radius: 15px;
                     font-weight: 600;
                     font-size: 14px;"
            )
          )
        )
      ),

      # Description Card
      div(
        style = "background: #f8f9fa;
               border-left: 4px solid #6c757d;
               padding: 15px 20px;
               margin-bottom: 15px;
               box-shadow: 0 2px 4px rgba(0,0,0,0.1);
               border-radius: 5px;",
        div(
          style = "display: flex; align-items: center; margin-bottom: 10px;",
          icon(
            "circle-info",
            style = "font-size: 20px; color: #6c757d; margin-right: 10px;"
          ),
          h4(strong("Corpus Description"), style = "margin: 0; color: #2c3e50;")
        ),
        hr(style = "margin: 10px 0; border-color: #dee2e6;"),
        p(txt2c, style = "margin: 0; color: #495057; line-height: 1.6;")
      ),

      # Features Card
      div(
        style = "background: white;
               border-left: 4px solid #f39c12;
               padding: 15px 20px;
               margin-bottom: 15px;
               box-shadow: 0 2px 4px rgba(0,0,0,0.1);
               border-radius: 5px;",
        div(
          style = "display: flex; align-items: center; margin-bottom: 10px;",
          icon(
            "tags",
            style = "font-size: 20px; color: #f39c12; margin-right: 10px;"
          ),
          h4(
            strong("Preprocessing Features"),
            style = "margin: 0; color: #2c3e50;"
          )
        ),
        hr(style = "margin: 10px 0; border-color: #e0e0e0;"),
        div(
          style = "display: grid; grid-template-columns: 1fr 1fr; gap: 15px;",
          # Special Entities
          div(
            style = "padding: 10px;",
            div(
              style = "color: #6c757d; font-size: 12px; font-weight: 600; text-transform: uppercase; margin-bottom: 8px;",
              icon("star", style = "margin-right: 5px;"),
              "Special Entities"
            ),
            span(
              txt3ter,
              style = paste0(
                "background-color: ",
                entity_badge_color,
                ";
                           color: white;
                           padding: 5px 12px;
                           border-radius: 15px;
                           font-weight: 600;
                           font-size: 13px;
                           display: inline-block;"
              )
            )
          ),
          # Multi-Words
          div(
            style = "padding: 10px;",
            div(
              style = "color: #6c757d; font-size: 12px; font-weight: 600; text-transform: uppercase; margin-bottom: 8px;",
              icon("link", style = "margin-right: 5px;"),
              "Multi-Words"
            ),
            span(
              txt3bis,
              style = paste0(
                "background-color: ",
                multi_badge_color,
                ";
                           color: white;
                           padding: 5px 12px;
                           border-radius: 15px;
                           font-weight: 600;
                           font-size: 13px;
                           display: inline-block;"
              )
            )
          )
        ),
        div(
          style = "display: grid; grid-template-columns: 1fr 1fr; gap: 15px; margin-top: 15px;",
          # Custom Word List
          div(
            style = "padding: 10px;",
            div(
              style = "color: #6c757d; font-size: 12px; font-weight: 600; text-transform: uppercase; margin-bottom: 8px;",
              icon("list", style = "margin-right: 5px;"),
              "Custom Word List"
            ),
            span(
              txt3,
              style = paste0(
                "background-color: ",
                custom_badge_color,
                ";
                           color: white;
                           padding: 5px 12px;
                           border-radius: 15px;
                           font-weight: 600;
                           font-size: 13px;
                           display: inline-block;"
              )
            )
          ),
          # Last Step
          div(
            style = "padding: 10px;",
            div(
              style = "color: #6c757d; font-size: 12px; font-weight: 600; text-transform: uppercase; margin-bottom: 8px;",
              icon("flag", style = "margin-right: 5px;"),
              "Last Preprocessing Step"
            ),
            span(
              txt4,
              style = "background-color: #5cb85c;
                     color: white;
                     padding: 5px 12px;
                     border-radius: 15px;
                     font-weight: 600;
                     font-size: 13px;
                     display: inline-block;"
            )
          )
        )
      )
    )
  })

  observeEvent(input$modalCustomLists, {
    if (!is.null(values$custom_lists)) {
      text <- tagList(
        DTformat(
          values$custom_lists,
          left = c(1, 2),
          filename = "Custom_lists_table"
        )
      )
    } else {
      text <- tagList(
        div(
          h4(HTML("No custom lists to show.")),
          style = "text-align:left"
        )
      )
    }

    show_alert(
      title = "Custom Word Lists",
      text = text,
      type = NULL,
      size = "m",
      closeOnEsc = TRUE,
      closeOnClickOutside = TRUE,
      html = TRUE,
      showConfirmButton = TRUE,
      showCancelButton = FALSE,
      btn_labels = "OK",
      btn_colors = "#6CC283",
      timer = NULL,
      imageUrl = "",
      animation = TRUE
    )
  })

  ### Convert Raw Data in Excel functions ----

  observeEvent(
    eventExpr = {
      input$collection.save
    },
    handlerExpr = {
      file_path <- destFolder(
        paste("Tall-Export-File-", sys.time(), ".csv", sep = ""),
        values$wdTall
      )
      readr::write_csv(
        x = values$txt,
        file = file_path,
        na = "NA",
        append = FALSE,
        col_names = TRUE
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  # Back to the original import text ----
  observeEvent(input$importTextBack, {
    values$txt <- values$txt %>%
      mutate(doc_selected = TRUE)

    output$dataImported <- DT::renderDT({
      req(values$txt)
      if (nrow(values$txt) > 0) {
        col_names <- c("doc_selected", "text_original")
        DTformat(
          values$txt %>%
            dplyr::filter(doc_selected) %>%
            mutate(across(any_of("text"), ~ paste0(substr(., 1, 500), "..."))) %>%
            select(doc_id, any_of("text"), everything()) %>%
            select(!any_of(col_names)),
          left = 3,
          nrow = 5,
          filter = "none",
          button = TRUE,
          delete = TRUE,
          escape = FALSE
        )
      }
    })
  })

  output$infoGroups <- renderUI({
    if (length(input$defineGroupsList) > 1) {
      shinyWidgets::alert(
        icon("info"),
        " You need to select only one field",
        status = "danger"
      )
    }
  })
}

#### Google GEMINI API ####
gemini_ai <- function(
  image = NULL,
  prompt = "Explain these images",
  model = "2.5-flash-lite",
  type = "png",
  retry_503 = 5,
  api_key = NULL,
  outputSize = "medium"
) {
  switch(
    outputSize,
    "small" = {
      generation_config <- list(
        temperature = 1,
        maxOutputTokens = 8192,
        topP = 0.95,
        topK = 40,
        seed = 1234
      )
    },
    "medium" = {
      generation_config <- list(
        temperature = 1,
        maxOutputTokens = 16384, #8192,
        topP = 0.95,
        topK = 40,
        seed = 1234
      )
    },
    "large" = {
      generation_config <- list(
        temperature = 1,
        maxOutputTokens = 32768, #8192,
        topP = 0.95,
        topK = 40,
        seed = 1234
      )
    }
  )

  # Build URL
  model_query <- paste0("gemini-", model, ":generateContent")
  url <- paste0(
    "https://generativelanguage.googleapis.com/v1beta/models/",
    model_query
  )
  if (is.null(api_key)) {
    api_key <- Sys.getenv("GEMINI_API_KEY")
  }

  # Base structure of parts
  parts <- list(list(text = prompt))

  # Handle images if provided
  if (!is.null(image)) {
    if (!is.vector(image)) {
      image <- as.vector(image)
    }
    mime_type <- paste0("image/", type)

    for (img_path in image) {
      if (!file.exists(img_path)) {
        return(paste0("❌ Error: Image file does not exist: ", img_path))
      }

      image_data <- tryCatch(
        base64enc::base64encode(img_path),
        error = function(e) {
          return(NULL)
        }
      )

      if (is.null(image_data)) {
        return(paste0("❌ Failed to encode image: ", img_path))
      }

      parts <- append(
        parts,
        list(
          list(
            inline_data = list(
              mime_type = mime_type,
              data = image_data
            )
          )
        )
      )
    }
  }

  # Assemble request body
  request_body <- list(
    contents = list(
      parts = parts
    ),
    generationConfig = generation_config
  )

  # Retry loop
  for (attempt in seq_len(retry_503)) {
    # Build and send request
    req <- request(url) |>
      req_url_query(key = api_key) |>
      req_headers("Content-Type" = "application/json") |>
      req_body_json(request_body)

    resp <- tryCatch(
      req_perform(req),
      error = function(e) {
        return(list(
          status_code = stringr::str_extract(e$message, "(?<=HTTP )\\d+") |>
            as.numeric(),
          error = TRUE,
          message = paste("❌ Request failed with error:", e$message)
        ))
      }
    )

    # # Handle connection-level error
    # if (is.list(resp) && isTRUE(resp$error)) {
    #   return(resp$message)
    # }

    # Retry on HTTP 503 or 429
    if (resp$status_code %in% c(429, 503)) {
      if (attempt < retry_503) {
        message(paste0(
          "⚠️ HTTP 503 (Service Unavailable) - retrying in 2 seconds (attempt ",
          attempt,
          "/",
          retry_503,
          ")..."
        ))
        Sys.sleep(2)
        next
      } else {
        return(
          paste0(
            "❌ HTTP 503: Service Unavailable.\n",
            "The Google Gemini servers are currently overloaded or under maintenance.\n",
            "All retry attempts failed (",
            retry_503,
            "). Please try again in a few minutes. Alternatively, consider using a different AI model with lower latency."
          )
        )
      }
    }

    # HTTP errors
    # 400 api key not valid
    if (resp$status_code == 400) {
      msg <- tryCatch(
        {
          parsed <- jsonlite::fromJSON(httr2::resp_body_string(resp))
          parsed$error$message
        },
        error = function(e) {
          "Please check your API key. It seems to be not valid!"
        }
      )
      return(paste0("❌ HTTP ", resp$status_code, ": ", msg))
    }
    # Other HTTP errors
    if (resp$status_code != 200) {
      msg <- tryCatch(
        {
          parsed <- jsonlite::fromJSON(httr2::resp_body_string(resp))
          parsed$error$message
        },
        error = function(e) {
          "Service unavailable or unexpected error. Please check your API key and usage limit."
        }
      )

      return(paste0("❌ HTTP ", resp$status_code, ": ", msg))
    }

    # Successful response
    candidates <- httr2::resp_body_json(resp)$candidates
    outputs <- unlist(lapply(candidates, \(c) c$content$parts))
    return(outputs)
  }
}

setGeminiAPI <- function(api_key) {
  # 1. Controllo validità dell'API key
  apiCheck <- gemini_ai(
    image = NULL,
    prompt = "Hello",
    model = "2.5-flash",
    type = "png",
    retry_503 = 5,
    api_key = api_key
  )

  contains_http_error <- grepl("HTTP\\s*[1-5][0-9]{2}", apiCheck)

  if (contains_http_error) {
    return(list(
      valid = FALSE,
      message = "❌ API key seems be not valid! Please, check it or your connection."
    ))
  }

  if (is.null(api_key) || !is.character(api_key) || nchar(api_key) == 0) {
    return(list(
      valid = FALSE,
      message = "❌ API key must be a non-empty string."
    ))
  }

  if (nchar(api_key) < 10) {
    return(list(
      valid = FALSE,
      message = "❌ API key seems too short. Please verify your key."
    ))
  }

  # 2. Mostra solo gli ultimi 4 caratteri per feedback
  last_chars <- 4
  last <- substr(
    api_key,
    max(1, nchar(api_key) - last_chars + 1),
    nchar(api_key)
  )

  # 3. Imposta la variabile d'ambiente
  Sys.setenv(GEMINI_API_KEY = api_key)

  return(list(
    valid = TRUE,
    message = paste0(
      paste0(rep("*", nchar(api_key) - 4), collapse = ""),
      last,
      collapse = ""
    )
  ))
}

showGeminiAPI <- function() {
  api_key <- Sys.getenv("GEMINI_API_KEY")
  last_chars <- 4
  last <- substr(
    api_key,
    max(1, nchar(api_key) - last_chars + 1),
    nchar(api_key)
  )
  last <- paste0(
    paste0(rep("*", nchar(api_key) - 4), collapse = ""),
    last,
    collapse = ""
  )
  return(last)
}

load_api_key <- function(path = path_gemini_key) {
  if (file.exists(path)) {
    key <- readLines(path, warn = FALSE)
    if (nchar(key) >= 10) {
      Sys.setenv(GEMINI_API_KEY = key)
      return(TRUE)
    }
  }
  return(FALSE)
}

loadGeminiModel = function(file) {
  # load info about model type and output size
  if (file.exists(file)) {
    model <- readLines(file, warn = FALSE)
  } else {
    model <- c("2.5-flash", "medium")
  }
  if (length(model == 1)) {
    model <- c(model, "medium")
  }
  return(model)
}

saveGeminiModel = function(model, file) {
  if (file.exists(file)) {
    file.remove(file)
  }
  writeLines(model, file)
}

geminiOutput <- function(title = "", content = "", values) {
  if (is.null(content)) {
    content <- "Click the 'Ask TALL AI' button to analyze the visual outputs and provide an automatic interpretation of your results based on the graphs.\n
This AI-powered feature leverages Google Gemini to help you understand patterns and insights emerging from your contextual analysis.\n  \n  \n \n  \n  \n  \n"
  }
  box(
    title = title,
    width = 12,
    status = "info",
    solidHeader = TRUE,
    div(
      id = "typing-box",
      style = "white-space: pre-wrap; background-color:#f9f9f9; padding:15px; border:1px solid #ccc; border-radius:5px; max-height:400px; overflow-y: auto;",
      #HTML(text_to_html(content))
      HTML(gemini_to_html(content))
      #HTML(content)
    ),
    br(),
    em(
      "You can modify or enrich the proposed prompt with additional context or details about your analysis to help 'TALL AI' generate a more accurate and meaningful interpretation."
    ),
    textAreaInput(
      inputId = "gemini_additional",
      label = NULL,
      value = values$gemini_model_parameters,
      placeholder = "You can provide additional context or details about your analysis to help 'TALL AI' generate a more accurate and meaningful interpretation.",
      rows = 3,
      width = "100%"
    ),
    fluidRow(
      column(
        4,
        align = "center",
        actionButton(
          "gemini_btn",
          "Ask TALL AI",
          style = "color: white;",
          icon(name = "microchip", lib = "font-awesome"),
          width = "80%"
        )
      ),
      column(
        4,
        align = "center",
        actionButton(
          "copy_btn",
          "Copy",
          style = "color: white;",
          icon = icon("clipboard"),
          width = "80%"
        )
      ),
      column(
        4,
        align = "center",
        downloadButton(
          "save_btn",
          "Save",
          icon = icon("download"),
          class = "btn-default",
          style = "color: white; width: 80%;"
        )
      )
    )
  )
}

tallAiPrompts <- function(values, activeTab) {
  ## Role definition for Gemini as TALL AI assistant
  promptInitial <- paste0(
    "You are TALL AI, an expert virtual assistant specializing in text analysis and computational linguistics, integrated into the R TALL software. ",
    "Your task is to support researchers in interpreting and critically discussing the results of ",
    "their quantitative textual analyses, offering insights, contextual explanations, and guidance for data-driven interpretation. "
  )

  ## Specific prompts for each analysis
  switch(
    activeTab,
    "overview" = {
      # Build morphological features summary if available
      morph_text <- ""
      if (!is.null(values$dfTag) && "feats" %in% names(values$dfTag)) {
        morph_features <- c("Tense", "Mood", "Number", "Person", "VerbForm",
                            "Degree", "Gender", "Case", "Voice")
        morph_parts <- c()
        for (feat in morph_features) {
          pattern <- paste0("(?:^|\\|)", feat, "=([^|]+)")
          feats_col <- values$dfTag$feats[values$dfTag$docSelected]
          matches <- regmatches(feats_col, regexpr(pattern, feats_col, perl = TRUE))
          feat_vals <- sub(paste0(".*", feat, "="), "", matches)
          feat_vals <- feat_vals[nchar(feat_vals) > 0]
          if (length(feat_vals) > 0) {
            tbl <- sort(table(feat_vals), decreasing = TRUE)
            pcts <- round(100 * tbl / sum(tbl), 1)
            top_vals <- paste0(names(pcts), " (", pcts, "%)", collapse = ", ")
            morph_parts <- c(morph_parts, paste0(feat, ": ", top_vals))
          }
        }
        if (length(morph_parts) > 0) {
          morph_text <- paste0(
            " Morphological features distribution: ",
            paste(morph_parts, collapse = "; "), ". "
          )
        }
      }

      prompt <- paste0(
        "A user has analyzed a text corpus and obtained the following summary statistics. Provide a detailed interpretation of these results based on the values provided. ",
        "Corpus metrics: ",
        merge_df_to_string(values$VbData),
        morph_text,
        "Structure your response in the following way: ",
        "1.  **General Summary**: Based on the data provided, start with a paragraph that summarizes the main characteristics of the corpus. ",
        "2.  **Corpus Size and Structure**: Explain the meaning of `Documents`, `Tokens`, `Types`, `Lemma`, and `Sentences` based on their provided values. Relate the `Tokens` count to the `Types` count to give a first impression of vocabulary size. ",
        "3.  **Document and Sentence Characteristics**: Interpret the provided averages (`Avg Length`) and standard deviations (`SD Length`) for documents and sentences. Explain what the specific `SD` values imply about the consistency of lengths across the corpus (e.g., a low SD suggests uniform lengths, a high SD suggests high variability). ",
        "4.  **Lexical Richness and Variety**: Analyze the provided values for the lexical richness indices: `TTR (%)`, `Hapax (%)`, `Guiraud Index`, and `Yule's K`. For each, explain what its value indicates about the richness of the vocabulary. Clarify why multiple indices are useful (e.g., correcting for corpus length). ",
        "5.  **Stylistic and Distributional Indices**: Explain the meaning of the given values for `Lexical Density`, `Nominal Ratio`, and `Gini Index`. For each metric, interpret what the provided value suggests about the text's nature (e.g., for `Lexical Density`: informational vs. conversational; for `Nominal Ratio`: descriptive vs. narrative style; for `Gini Index`: word distribution). ",
        if (nchar(morph_text) > 0) {
          "6.  **Morphological Profile**: Interpret the morphological features distribution (Tense, Mood, Number, Person, VerbForm, etc.). Explain what the dominant morphological patterns reveal about the text's style, register, and linguistic characteristics. "
        } else { "" },
        if (nchar(morph_text) > 0) {
          "7.  **Conclusion and Potential Next Steps**: End with a summary conclusion and suggest to the user what subsequent analyses might be interesting based on the specific results they obtained."
        } else {
          "6.  **Conclusion and Potential Next Steps**: End with a summary conclusion and suggest to the user what subsequent analyses might be interesting based on the specific results they obtained."
        }
      )
    },
    "kwic" = {
      prompt <- paste0(
        "You have to provide an interpratation of a Words in Context (KWIC - Keyword in Context) analysis. ",
        "Your expertise covers concordance analysis, semantic usage patterns, contextual meaning interpretation, and discourse structure analysis. ",
        "You help users understand: ",
        " - Semantic shifts and meaning variations of target words ",
        " - Collocational patterns and linguistic relationships ",
        " - Contextual distribution and usage frequency ",
        " - Thematic contexts and discourse patterns ",
        " - Pragmatic and stylistic variations. ",
        "Always provide insights that are linguistically grounded, contextually relevant, and actionable for text analysis research."
      )
    },
    "ca" = {
      prompt <- paste0(
        "You have to interpret Correspondence Analysis (CA) results from the TALL software package in R. ",
        "Your expertise covers multivariate text analysis, dimensional interpretation, word-document associations, and visualization of textual relationships.",
        "You help users understand: ",
        "  - Dimensional structure and variance explained",
        "  - Word positioning and semantic relationships",
        "  - Document clustering and thematic groupings",
        "  - Distance interpretation and proximity patterns",
        "  - Quality of representation and contribution analysis",
        "  - Biplot interpretation and association patterns. ",
        "Always provide interpretations that are statistically sound, linguistically meaningful, and practically actionable for text analysis research."
      )
    },
    "w_networkCooc" = {
      prompt <- paste0(
        "You have to provide an interpretation of Word Co-occurrence Analysis results. ",
        "Your expertise covers network analysis, semantic relationships, collocational patterns, and lexical association interpretation. ",
        "You help users understand: ",
        " - Network topology and connectivity patterns ",
        " - Semantic clustering and thematic groupings ",
        " - Word association strengths and significance ",
        " - Community detection and lexical fields ",
        " - Centrality measures and influential words ",
        " - Network density and structural properties ",
        " - Contextual relationships and discourse patterns. ",
        "Always provide interpretations that are linguistically informed, statistically grounded, and actionable for text analysis research."
      )
    },
    "w_networkTM" = {
      prompt <- paste0(
        "You have to provide an interpretation of Thematic Map analysis. ",
        "Your expertise covers bibliometric-inspired thematic analysis, topic clustering, strategic diagram interpretation, and conceptual structure analysis. ",
        "You help users understand: ",
        " - Strategic positioning of topics (Hot, Basic, Niche, Peripheral) ",
        " - Callon Centrality and Density measures interpretation ",
        " - Topic development and maturity assessment ",
        " - Conceptual structure and thematic relationships ",
        " - Research field evolution and strategic implications ",
        " - Community detection and topic clustering ",
        " - Comparative thematic analysis across time periods or groups ",
        "Always provide interpretations that are methodologically sound, theoretically grounded, and strategically actionable for research and content analysis."
      )
    },
    "w_w2v_similarity" = {
      prompt <- paste0(
        "You have to provide an interpretation of both Word Similarity Network analysis and UMAP results from the TALL software package in R. ",
        "Your expertise covers word embedding analysis, semantic similarity interpretation, network topology analysis, and community detection in embedding spaces. ",
        "Provide analysis in TWO distinct sections: ",
        "  Section: SIMILARITY NETWORK ANALYSIS ",
        "You help users understand: ",
        " - Word2vec embedding relationships and semantic similarity patterns ",
        " - Network structure and community clusters based on cosine similarity ",
        " - Semantic neighborhoods and lexical fields ",
        " - Embedding space topology and word positioning ",
        " - Community detection results and thematic groupings. ",
        " Section 2: UMAP PROJECTION ANALYSIS ",
        "You help users understand: ",
        "  - UMAP Dimension 1 Interpretation of what semantic gradient or contrast the vertical axis represents ",
        "  - UMAP Dimension 2 Interpretation of what semantic gradient or contrast the vertical axis represents ",
        "  - Characterize the semantic themes in each quadrant of the UMAP space ",
        "  - How semantic clusters distribute across the 2D projection ",
        "  - What distances in UMAP space reveal about semantic relationships ",
        "  - Words at the edges or boundaries of the semantic space ",
        "  - What areas of high/low density indicate about vocabulary structure ",
        "Always provide interpretations that are linguistically informed, computationally grounded, and actionable for semantic analysis research. "
      )
    },
    "d_tm_estim" = {
      prompt <- paste0(
        "You have to provide an interpretation of Topic Modeling analysis results from the TALL software package in R. ",
        "Your expertise covers Latent Dirichlet Allocation (LDA), probabilistic topic modeling, beta and theta probability distributions, document-topic associations, and semantic structure discovery. ",
        "You help users understand: ",
        " - LDA model interpretation and topic coherence assessment ",
        " - Beta probabilities (word-topic distributions) and topic characterization ",
        " - Theta probabilities (document-topic distributions) and document classification ",
        " - Topic quality evaluation and model diagnostics ",
        " - Semantic structure discovery and thematic organization ",
        " - Model parameter optimization and validation ",
        "Always provide interpretations that are statistically sound, linguistically meaningful, and methodologically rigorous for probabilistic text analysis."
      )
    },
    "d_polDet" = {
      prompt <- paste0(
        "You have to provide an interpretation of Topic Modeling analysis results from the TALL software package in R. ",
        "Your expertise covers Latent Dirichlet Allocation (LDA), probabilistic topic modeling, beta and theta probability distributions, document-topic associations, and semantic structure discovery.  ",
        "You help users understand:   ",
        " - LDA model interpretation and topic coherence assessment ",
        " - Beta probabilities (word-topic distributions) and topic characterization ",
        " - Theta probabilities (document-topic distributions) and document classification ",
        " - Topic quality evaluation and model diagnostics ",
        " - Semantic structure discovery and thematic organization ",
        " - Model parameter optimization and validation. ",
        " Always provide interpretations that are statistically sound, linguistically meaningful, and methodologically rigorous for probabilistic text analysis."
      )
    },
    "d_syntactic" = {
      prompt <- paste0(
        "You have to provide an interpretation of Syntactic Complexity analysis results from the TALL software package in R. ",
        "Your expertise covers syntactic structure analysis, dependency parsing metrics, and linguistic complexity measurement. ",
        "You help users understand: ",
        " - Mean Dependency Distance (MDD) and its implications for processing difficulty ",
        " - Parse tree depth and sentence structural complexity ",
        " - Subordination ratios and clause embedding patterns ",
        " - Noun phrase complexity and modifier density ",
        " - Variation across documents and distribution patterns ",
        " - What the overall syntactic profile reveals about text register and style. ",
        "Always provide interpretations that are linguistically grounded, methodologically sound, and actionable for text analysis research."
      )
    },
    "d_svo" = {
      prompt <- paste0(
        "You have to provide an interpretation of Subject-Verb-Object (SVO) triplet extraction results from the TALL software package in R. ",
        "Your expertise covers dependency-based syntactic analysis, semantic role labeling, and argument structure interpretation. ",
        "You help users understand: ",
        " - The most frequent SVO patterns and what they reveal about the corpus content ",
        " - Subject and object distributions and their semantic roles ",
        " - Verb frequency patterns and the dominant actions or processes described ",
        " - Network structure of SVO relationships and thematic clusters ",
        " - What the SVO patterns reveal about discourse structure and argumentation. ",
        "Always provide interpretations that are linguistically informed and actionable for content analysis and discourse research."
      )
    },
    {
      prompt <- paste0(
        "Provide an interpretation of this plot creted with 'TALL R Package'"
      )
    }
  )
  prompt <- paste0(promptInitial, prompt)
  #if (!activeTab %in% c("mainInfo", "thematicMap", "trendTopic")) prompt <- paste0(prompt, " Provide also scientific references about the methodological description")
  return(prompt)
}


gemini2clip <- function(values, activeTab) {
  switch(
    activeTab,
    "kwic" = {
      values$contextGemini
    },
    "w_reinclustering" = {
      "Not yet implemented"
    },
    "ca" = {
      values$caGemini
    },
    "w_networkCooc" = {
      values$w_networkGemini
    },
    "w_networkTM" = {
      values$w_networkTMGemini
    },
    "w_w2v_similarity" = {
      values$w_w2vGemini
    },
    "d_tm_estim" = {
      values$tmGemini
    },
    "d_polDet" = {
      values$d_polDet_Gemini
    },
    "d_emo" = {
      values$d_emo_Gemini
    },
    "d_syntactic" = {
      values$d_syntactic_Gemini
    },
    "d_svo" = {
      values$d_svo_Gemini
    }
  )
}

geminiGenerate <- function(
  values,
  activeTab,
  gemini_additional,
  gemini_model_parameters,
  input
) {
  if (gemini_additional != "") {
    desc <- paste0(
      values$corpus_description,
      gemini_additional,
      gemini_model_parameters,
      collapse = ". "
    )
  } else {
    desc <- paste0(
      values$corpus_description,
      gemini_model_parameters,
      collapse = ". "
    )
  }
  prompt <- tallAiPrompts(values, activeTab)
  switch(
    activeTab,
    "overview" = {
      req(values$VbData)
      values$overviewGemini <- geminiPromptImage(
        obj = values$VbData,
        type = "text",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "kwic" = {
      req(values$contextNetwork)
      values$contextGemini <- geminiPromptImage(
        obj = values$contextNetwork,
        type = "vis",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "w_reinclustering" = {
      "Not yet implemented"
    },
    "ca" = {
      req(values$plotCA)
      values$caGemini <- geminiPromptImage(
        obj = values$plotCA,
        type = "plotly",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "w_networkCooc" = {
      req(values$netVis)
      values$w_networkGemini <- geminiPromptImage(
        obj = values$netVis,
        type = "vis",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "w_networkTM" = {
      req(values$TMmap)
      values$w_networkTMGemini <- geminiPromptImage(
        obj = plotTM(values$TM$df, size = 0.5, gemini = TRUE),
        type = "plotly",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "w_w2v_similarity" = {
      req(values$w2vNetworkPlot)
      p1 <- values$w2vNetworkPlot
      p2 <- values$w2vUMAPplot
      files <- unlist(lapply(c("similarity", "umqp"), function(x) {
        paste0(tempdir(), "/", x, ".png")
      }))
      suppressWarnings(plot2png(
        p1,
        filename = files[1],
        type = "vis",
        dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p2,
        filename = files[2],
        type = "plotly",
        dpi = 150, height = 7
      ))
      values$w_w2vGemini <- geminiPromptImage(
        obj = files,
        type = "multi",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "d_tm_estim" = {
      req(values$TMestim_result)
      tmPlot <- topicGplot(
        values$TMestim_result$beta,
        nPlot = 10,
        type = "beta"
      )
      values$tmGemini <- geminiPromptImage(
        obj = tmPlot,
        type = "ggplot2",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "d_polDet" = {
      req(values$docPolPlots)
      p1 <- values$sentimentPieChart
      p2 <- values$docPolPlots$positive %>% layout(showlegend = FALSE)
      p3 <- values$docPolPlots$negative %>% layout(showlegend = FALSE)

      files <- unlist(lapply(c("topic1", "topic2", "topic3"), function(x) {
        paste0(tempdir(), "/", x, ".png")
      }))

      suppressWarnings(plot2png(
        p1,
        filename = files[1],
        type = "plotly",
        dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p2,
        filename = files[2],
        type = "plotly",
        dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p3,
        filename = files[3],
        type = "plotly",
        dpi = 150, height = 7
      ))

      values$d_polDet_Gemini <- geminiPromptImage(
        obj = files,
        type = "multi",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "d_emo" = {
      req(values$emotionBarChart)
      p1 <- values$emotionBarChart
      p2 <- values$emotionRadarPlot
      p3 <- values$emotionHeatmap

      files <- unlist(lapply(c("emo1", "emo2", "emo3"), function(x) {
        paste0(tempdir(), "/", x, ".png")
      }))

      suppressWarnings(plot2png(
        p1,
        filename = files[1],
        type = "plotly",
        dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p2,
        filename = files[2],
        type = "plotly",
        dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p3,
        filename = files[3],
        type = "plotly",
        dpi = 150, height = 7
      ))

      values$d_emo_Gemini <- geminiPromptImage(
        obj = files,
        type = "multi",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "d_syntactic" = {
      req(values$syntacticResults)
      prompt <- paste0(
        prompt,
        " Syntactic complexity metrics (corpus-level averages): ",
        "Mean Sentence Length: ", round(mean(values$syntacticResults$mean_sent_length), 2),
        "; Mean Tree Depth: ", round(mean(values$syntacticResults$mean_tree_depth), 2),
        "; Mean Dependency Distance: ", round(mean(values$syntacticResults$mean_dep_distance), 2),
        "; Mean Clauses per Sentence: ", round(mean(values$syntacticResults$mean_clauses_per_sent), 2),
        "; Subordination Ratio: ", round(mean(values$syntacticResults$subordination_ratio), 2),
        "; Mean Branching Factor: ", round(mean(values$syntacticResults$mean_branching_factor), 2),
        ". Per-document metrics: ",
        merge_df_to_string(head(values$syntacticResults, 20))
      )
      values$d_syntactic_Gemini <- geminiPromptImage(
        obj = NULL,
        type = "text",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    },
    "d_svo" = {
      req(values$svoResults)
      top_svo <- head(values$svoResults, 30)
      verb_freq <- values$svoResults %>%
        dplyr::group_by(verb) %>%
        dplyr::summarise(freq = sum(freq), .groups = "drop") %>%
        dplyr::arrange(dplyr::desc(freq)) %>%
        head(15)
      prompt <- paste0(
        prompt,
        " Top SVO triplets (subject-verb-object with frequency): ",
        merge_df_to_string(top_svo),
        ". Most frequent verbs: ",
        paste(paste0(verb_freq$verb, " (", verb_freq$freq, ")"), collapse = ", ")
      )
      values$d_svo_Gemini <- geminiPromptImage(
        obj = NULL,
        type = "text",
        prompt = prompt,
        key = values$geminiAPI,
        desc = desc,
        values = values
      )
    }
  )
  return(values)
}

geminiParameterPrompt <- function(values, activeTab, input) {
  txt <- paste0(
    "The analysis was perfomed on ",
    values$generalTerm,
    " extracted from the original corpus. "
  )

  switch(
    activeTab,
    "overview" = {
      req(values$VbData)
      txt <- ""
    },
    "kwic" = {
      req(values$contextNetwork)
      txt <- paste0(
        txt,
        "The context windows consists of of the ",
        input$wordsContBefore,
        " words preceding and the ",
        input$wordsContAfter,
        " words following the target word."
      )
    },
    "w_reinclustering" = {
      "Not yet implemented"
    },
    "ca" = {
      req(values$plotCA)
      txt <- paste0(
        txt,
        "The occurrences of the most ",
        input$nCA,
        " frequent words were measured across ",
        input$groupCA,
        ". The words were then grouped into ",
        input$nClustersCA,
        " clusters using hierarchical clustering."
      )
    },
    "w_networkCooc" = {
      req(values$netVis)
      txt <- paste0(
        txt,
        "The co-occurrences of the most ",
        input$nMax,
        " frequent words were measured across ",
        input$w_groupNet,
        ". The co-occurrences were normalized using the ",
        input$normalizationCooc,
        " index. The words were then grouped using Louvain community detection algorithm."
      )
    },
    "w_networkTM" = {
      req(values$TMmap)
      txt <- paste0(
        txt,
        "The co-occurrences of the most ",
        input$nMaxTM,
        " frequent words were measured across ",
        input$w_groupTM,
        ". The co-occurrences were normalized using the association index. The words were then grouped using Louvain community detection algorithm."
      )
    },
    "w_w2v_similarity" = {
      req(values$w2vNetworkPlot)
      txt <- paste0(
        txt,
        "The ",
        input$w_w2v_similarityN,
        " most frequent target words were selected, and for each of them, the 10 words with the highest cosine similarity were identified."
      )
    },
    "d_tm_estim" = {
      req(values$TMestim_result)
      txt <- paste0(
        txt,
        "The Topic Model was estimated using the ",
        input$nTmEstim,
        " frequent words identified by ",
        input$top_byEstim,
        " measure."
      )
    },
    "d_polDet" = {
      req(values$docPolPlots)
      txt
    },
    "d_emo" = {
      req(values$emotionBarChart)
      txt
    },
    "d_syntactic" = {
      req(values$syntacticResults)
      txt <- paste0(txt, "Syntactic complexity was computed using dependency parsing metrics.")
    },
    "d_svo" = {
      req(values$svoResults)
      txt <- paste0(txt, "Subject-Verb-Object triplets were extracted using dependency parsing.")
    },
    {
      ""
    }
  )
  return(txt)
}

## gemini prompt for images
geminiPromptImage <- function(
  obj,
  type = "vis",
  prompt = "Explain the topics in this map",
  key,
  desc = NULL,
  values
) {
  ## Check Computer configuration to work with TALL AI
  ### Internet Connection
  if (!is_online()) {
    res <- '⚠️ **Note**: TALL AI requires an active internet connection to work.'
    return(res)
  }
  ### Chromium Browser
  if (is.null(values$Chrome_url)) {
    res <- '⚠️ **Note**: TALL AI requires a **Chrome-based browser** (such as Google Chrome or Microsoft Edge) installed on your computer to work correctly.'
    return(res)
  }
  ### Gemini API key
  if (key) {
    if (!is.null(desc)) {
      prompt <- paste0(prompt, desc, collapse = ". ")
    }
    tmpdir <- tempdir()
    owd <- setwd(tmpdir)
    on.exit(setwd(owd))
    file_path <- paste0(tempfile(), ".png", collapse = "")
    switch(
      type,
      "vis" = {
        suppressWarnings(plot2png(
          obj,
          filename = file_path,
          type = "vis",
          dpi = 150, height = 7
        ))
      },
      "plotly" = {
        suppressWarnings(plot2png(
          obj,
          filename = file_path,
          type = "plotly",
          dpi = 150, height = 7
        ))
      },
      "text" = {
        file_path <- NULL
      },
      "multi" = {
        file_path <- obj
      },
      "html" = {
        html_name <- tempfile(fileext = ".html")
        htmltools::save_html(obj, html_name)
        tallShot(html_name, zoom = 150 / 96, file = file_path)
        unlink(html_name)
        unlink(paste0(tools::file_path_sans_ext(html_name), "_files"), recursive = TRUE)
      },
      "ggplot2" = {
        ggsave(
          filename = file_path,
          plot = obj,
          dpi = 72,
          height = 7,
          width = 14,
          bg = "transparent"
        )
      }
    )

    res <- gemini_ai(
      image = file_path,
      prompt = prompt,
      model = values$gemini_api_model,
      outputSize = values$gemini_output_size
    )
  } else {
    res <- 'To access this feature, please provide a valid Gemini AI API key. You can obtain your API key by visiting the official <a href="https://aistudio.google.com/" target="_blank">Google AI Studio website</a>.'
  }

  return(res)
}

## Async-ready preparation: does everything geminiPromptImage does EXCEPT calling gemini_ai
geminiPrepareImage <- function(
  obj,
  type = "vis",
  prompt = "Explain the topics in this map",
  key,
  desc = NULL,
  values
) {
  ## Check Computer configuration to work with TALL AI
  ### Internet Connection
  if (!is_online()) {
    return(list(error = "\u26a0\ufe0f **Note**: TALL AI requires an active internet connection to work."))
  }
  ### Chromium Browser
  if (is.null(values$Chrome_url)) {
    return(list(error = "\u26a0\ufe0f **Note**: TALL AI requires a **Chrome-based browser** (such as Google Chrome or Microsoft Edge) installed on your computer to work correctly."))
  }
  ### Gemini API key
  if (!key) {
    return(list(error = 'To access this feature, please provide a valid Gemini AI API key. You can obtain your API key by visiting the official <a href="https://aistudio.google.com/" target="_blank">Google AI Studio website</a>.'))
  }

  if (!is.null(desc)) {
    prompt <- paste0(prompt, desc, collapse = ". ")
  }

  # Snapshot API key for the future worker (env vars don't propagate)
  api_key <- Sys.getenv("GEMINI_API_KEY")

  tmpdir <- tempdir()
  owd <- setwd(tmpdir)
  on.exit(setwd(owd))
  file_path <- paste0(tempfile(), ".png", collapse = "")

  switch(
    type,
    "vis" = {
      suppressWarnings(plot2png(
        obj, filename = file_path, type = "vis", dpi = 150, height = 7
      ))
    },
    "plotly" = {
      suppressWarnings(plot2png(
        obj, filename = file_path, type = "plotly", dpi = 150, height = 7
      ))
    },
    "text" = {
      file_path <- NULL
    },
    "multi" = {
      file_path <- obj
    },
    "html" = {
      html_name <- tempfile(fileext = ".html")
      htmltools::save_html(obj, html_name)
      tallShot(html_name, zoom = 150 / 96, file = file_path)
      unlink(html_name)
      unlink(paste0(tools::file_path_sans_ext(html_name), "_files"), recursive = TRUE)
    },
    "ggplot2" = {
      ggsave(
        filename = file_path, plot = obj,
        dpi = 72, height = 7, width = 14, bg = "transparent"
      )
    }
  )

  return(list(
    image = file_path,
    prompt = prompt,
    model = values$gemini_api_model,
    outputSize = values$gemini_output_size,
    api_key = api_key
  ))
}

## Async-ready preparation: builds prompt, converts plots, returns data for future
geminiPrepare <- function(
  values,
  activeTab,
  gemini_additional,
  gemini_model_parameters,
  input
) {
  # Build desc (same logic as geminiGenerate)
  if (gemini_additional != "") {
    desc <- paste0(
      values$corpus_description,
      gemini_additional,
      gemini_model_parameters,
      collapse = ". "
    )
  } else {
    desc <- paste0(
      values$corpus_description,
      gemini_model_parameters,
      collapse = ". "
    )
  }
  prompt <- tallAiPrompts(values, activeTab)

  # Map activeTab to reactive field name
  field <- switch(
    activeTab,
    "overview" = "overviewGemini",
    "kwic" = "contextGemini",
    "ca" = "caGemini",
    "w_networkCooc" = "w_networkGemini",
    "w_networkTM" = "w_networkTMGemini",
    "w_w2v_similarity" = "w_w2vGemini",
    "d_tm_estim" = "tmGemini",
    "d_polDet" = "d_polDet_Gemini",
    "d_emo" = "d_emo_Gemini",
    "d_syntactic" = "d_syntactic_Gemini",
    "d_svo" = "d_svo_Gemini",
    NULL
  )

  if (is.null(field)) {
    return(list(field = "overviewGemini", error = "Unsupported analysis tab"))
  }

  # Tab-specific image preparation
  prep <- switch(
    activeTab,
    "overview" = {
      req(values$VbData)
      geminiPrepareImage(
        obj = values$VbData, type = "text",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "kwic" = {
      req(values$contextNetwork)
      geminiPrepareImage(
        obj = values$contextNetwork, type = "vis",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "ca" = {
      req(values$plotCA)
      geminiPrepareImage(
        obj = values$plotCA, type = "plotly",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "w_networkCooc" = {
      req(values$netVis)
      geminiPrepareImage(
        obj = values$netVis, type = "vis",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "w_networkTM" = {
      req(values$TMmap)
      geminiPrepareImage(
        obj = plotTM(values$TM$df, size = 0.5, gemini = TRUE), type = "plotly",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "w_w2v_similarity" = {
      req(values$w2vNetworkPlot)
      p1 <- values$w2vNetworkPlot
      p2 <- values$w2vUMAPplot
      files <- unlist(lapply(c("similarity", "umqp"), function(x) {
        paste0(tempdir(), "/", x, ".png")
      }))
      suppressWarnings(plot2png(
        p1, filename = files[1], type = "vis", dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p2, filename = files[2], type = "plotly", dpi = 150, height = 7
      ))
      geminiPrepareImage(
        obj = files, type = "multi",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "d_tm_estim" = {
      req(values$TMestim_result)
      tmPlot <- topicGplot(
        values$TMestim_result$beta, nPlot = 10, type = "beta"
      )
      geminiPrepareImage(
        obj = tmPlot, type = "ggplot2",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "d_polDet" = {
      req(values$docPolPlots)
      p1 <- values$sentimentPieChart
      p2 <- values$docPolPlots$positive %>% layout(showlegend = FALSE)
      p3 <- values$docPolPlots$negative %>% layout(showlegend = FALSE)
      files <- unlist(lapply(c("topic1", "topic2", "topic3"), function(x) {
        paste0(tempdir(), "/", x, ".png")
      }))
      suppressWarnings(plot2png(
        p1, filename = files[1], type = "plotly", dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p2, filename = files[2], type = "plotly", dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p3, filename = files[3], type = "plotly", dpi = 150, height = 7
      ))
      geminiPrepareImage(
        obj = files, type = "multi",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "d_emo" = {
      req(values$emotionBarChart)
      p1 <- values$emotionBarChart
      p2 <- values$emotionRadarPlot
      p3 <- values$emotionHeatmap
      files <- unlist(lapply(c("emo1", "emo2", "emo3"), function(x) {
        paste0(tempdir(), "/", x, ".png")
      }))
      suppressWarnings(plot2png(
        p1, filename = files[1], type = "plotly", dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p2, filename = files[2], type = "plotly", dpi = 150, height = 7
      ))
      suppressWarnings(plot2png(
        p3, filename = files[3], type = "plotly", dpi = 150, height = 7
      ))
      geminiPrepareImage(
        obj = files, type = "multi",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "d_syntactic" = {
      req(values$syntacticResults)
      prompt <- paste0(
        prompt,
        " Syntactic complexity metrics (corpus-level averages): ",
        "Mean Sentence Length: ", round(mean(values$syntacticResults$mean_sent_length), 2),
        "; Mean Tree Depth: ", round(mean(values$syntacticResults$mean_tree_depth), 2),
        "; Mean Dependency Distance: ", round(mean(values$syntacticResults$mean_dep_distance), 2),
        "; Mean Clauses per Sentence: ", round(mean(values$syntacticResults$mean_clauses_per_sent), 2),
        "; Subordination Ratio: ", round(mean(values$syntacticResults$subordination_ratio), 2),
        "; Mean Branching Factor: ", round(mean(values$syntacticResults$mean_branching_factor), 2),
        ". Per-document metrics: ",
        merge_df_to_string(head(values$syntacticResults, 20))
      )
      geminiPrepareImage(
        obj = NULL, type = "text",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    },
    "d_svo" = {
      req(values$svoResults)
      top_svo <- head(values$svoResults, 30)
      # Build verb frequency summary
      verb_freq <- values$svoResults %>%
        dplyr::group_by(verb) %>%
        dplyr::summarise(freq = sum(freq), .groups = "drop") %>%
        dplyr::arrange(dplyr::desc(freq)) %>%
        head(15)
      prompt <- paste0(
        prompt,
        " Top SVO triplets (subject-verb-object with frequency): ",
        merge_df_to_string(top_svo),
        ". Most frequent verbs: ",
        paste(paste0(verb_freq$verb, " (", verb_freq$freq, ")"), collapse = ", ")
      )
      geminiPrepareImage(
        obj = NULL, type = "text",
        prompt = prompt, key = values$geminiAPI, desc = desc, values = values
      )
    }
  )

  # Always include field in result
  prep$field <- field
  return(prep)
}

geminiWaitingMessage <- function(values, activeTab) {
  messageTxt <- "⌛ Thinking..."

  switch(
    activeTab,
    "overview" = {
      req(values$VbData)
      values$overviewGemini <- messageTxt
    },
    "kwic" = {
      req(values$contextNetwork)
      values$contextGemini <- messageTxt
    },
    "w_reinclustering" = {
      "Not yet implemented"
    },
    "ca" = {
      req(values$plotCA)
      values$caGemini <- messageTxt
    },
    "w_networkCooc" = {
      req(values$netVis)
      values$w_networkGemini <- messageTxt
    },
    "w_networkTM" = {
      req(values$TMmap)
      values$w_networkTMGemini <- messageTxt
    },
    "w_w2v_similarity" = {
      req(values$w2vNetworkPlot)
      values$w_w2vGemini <- messageTxt
    },
    "d_tm_estim" = {
      req(values$TMestim_result)
      values$tmGemini <- messageTxt
    },
    "d_polDet" = {
      req(values$docPolPlots)
      values$d_polDet_Gemini <- messageTxt
    },
    "d_emo" = {
      req(values$emotionBarChart)
      values$d_emo_Gemini <- messageTxt
    },
    "d_syntactic" = {
      req(values$syntacticResults)
      values$d_syntactic_Gemini <- messageTxt
    },
    "d_svo" = {
      req(values$svoResults)
      values$d_svo_Gemini <- messageTxt
    }
  )
  return(values)
}

geminiSave <- geminiSave <- function(values, activeTab) {
  switch(
    activeTab,
    "overview" = {
      req(values$VbData)
      gemini <- values$overviewGemini
    },
    "kwic" = {
      req(values$contextNetwork)
      gemini <- values$contextGemini
    },
    "w_reinclustering" = {
      "Not yet implemented"
    },
    "ca" = {
      req(values$plotCA)
      gemini <- values$caGemini
    },
    "w_networkCooc" = {
      req(values$netVis)
      gemini <- values$w_networkGemini
    },
    "w_networkTM" = {
      req(values$TMmap)
      gemini <- values$w_networkTMGemini
    },
    "w_w2v_similarity" = {
      req(values$w2vNetworkPlot)
      gemini <- values$w_w2vGemini
    },
    "d_tm_estim" = {
      req(values$TMestim_result)
      gemini <- values$tmGemini
    },
    "d_polDet" = {
      req(values$docPolPlots)
      gemini <- values$d_polDet_Gemini
    },
    "d_emo" = {
      req(values$emotionBarChart)
      gemini <- values$d_emo_Gemini
    },
    "d_syntactic" = {
      req(values$syntacticResults)
      gemini <- values$d_syntactic_Gemini
    },
    "d_svo" = {
      req(values$svoResults)
      gemini <- values$d_svo_Gemini
    }
  )
  if (is.null(gemini)) {
    gemini <- "Click 'Ask TALL AI' for help. "
  }

  return(gemini)
}

merge_df_to_string <- function(df) {
  # Check if the input has at least two columns
  if (ncol(df) < 2) {
    stop("The data frame must have at least two columns.")
  }

  # Ensure the input is a data frame
  df <- as.data.frame(df)

  # Convert each row into a "param: value" format
  row_strings <- apply(df[, 1:2], 1, function(row) {
    paste0(row[1], ": ", row[2])
  })

  # Concatenate all row strings using "; " as separator
  final_string <- paste(row_strings, collapse = "; ")

  return(final_string)
}

copy_to_clipboard <- function(x) {
  # Check the operating system
  sys_info <- Sys.info()
  os_type <- tolower(sys_info["sysname"])

  # Convert the object to a string if it is not already
  if (!is.character(x)) {
    x <- capture.output(print(x))
  }

  # Copy to clipboard based on the operating system
  if (os_type == "windows") {
    writeClipboard(x)
  } else if (os_type == "darwin") {
    # macOS
    con <- pipe("pbcopy", "w")
    writeLines(x, con)
    close(con)
  } else if (os_type == "linux") {
    # Use xclip or xsel, if available
    if (nzchar(Sys.which("xclip"))) {
      con <- pipe("xclip -selection clipboard", "w")
      writeLines(x, con)
      close(con)
    } else if (nzchar(Sys.which("xsel"))) {
      con <- pipe("xsel --clipboard --input", "w")
      writeLines(x, con)
      close(con)
    } else {
      stop(
        "Neither 'xclip' nor 'xsel' are available. Please install one of them to use the clipboard on Linux."
      )
    }
  } else {
    stop("Unrecognized or unsupported operating system.")
  }
}

string_to_sentence_df <- function(input_string) {
  # Validate that the input is a character string of length 1
  if (!is.character(input_string) || nchar(input_string) < 2) {
    return(data.frame(
      TALL_AI = "TALL AI was not performed",
      stringsAsFactors = FALSE,
      row.names = NULL
    ))
  }

  # Split the string by newline character "\n"
  split_sentences <- unlist(strsplit(input_string, split = "\n", fixed = TRUE))

  # Trim whitespace from each sentence
  cleaned_sentences <- trimws(split_sentences)

  # Remove empty entries
  cleaned_sentences <- cleaned_sentences[cleaned_sentences != ""]

  # Convert to data frame
  sentence_df <- data.frame(
    TALL_AI = cleaned_sentences,
    stringsAsFactors = FALSE,
    row.names = NULL
  )

  return(sentence_df)
}


## New function to convert gemini output as HTML blocks
gemini_to_html <- function(text, font_size = "16px", type = "tall_ai") {
  # Remove original leading/trailing whitespace
  text <- trimws(text)

  # Divide text into lines
  lines <- unlist(strsplit(text, "\n", fixed = TRUE))

  # Initialize HTML output with CSS styles including font size
  if (type == "tall_ai") {
    html_lines <- c(
      paste0(
        "<div style='font-family: Arial, sans-serif; line-height: 1.4; margin: 0 auto; padding: 15px; font-size: ",
        font_size,
        ";'>"
      )
    )
  } else {
    html_lines <-
      "<div style='font-family: \"Georgia\", \"Times New Roman\", serif; line-height: 1.8; margin: 0 auto; padding: 15px; font-size: 1.1em;'>"
  }

  in_list <- FALSE
  list_type <- ""
  last_ol_number <- 0  # Track numbering for continued <ol>

  close_list <- function() {
    if (in_list) {
      if (list_type == "ul") {
        html_lines <<- c(html_lines, "</ul>")
      } else {
        html_lines <<- c(html_lines, "</ol>")
      }
      in_list <<- FALSE
    }
  }

  for (i in seq_along(lines)) {
    line <- trimws(lines[i])

    # Skip empty lines
    if (line == "") next

    # Skip horizontal rules (--- or ***)
    if (grepl("^[-*_]{3,}\\s*$", line)) next

    # Markdown header management (lines starting with #, ##, ###, ####)
    if (stringr::str_detect(line, "^#{1,4}\\s+")) {
      close_list()

      header_level <- nchar(stringr::str_extract(line, "^#+"))
      header_text <- stringr::str_replace(line, "^#+\\s+", "")
      header_text <- stringr::str_replace(header_text, "\\s*#+\\s*$", "")
      header_text <- stringr::str_replace_all(header_text, "\\*\\*([^*]+)\\*\\*", "\\1")

      h_tag <- paste0("h", min(header_level + 1, 5))
      h_size <- switch(as.character(header_level),
        "1" = "font-size: 1.3em;",
        "2" = "font-size: 1.15em;",
        "3" = "font-size: 1.05em;",
        "font-size: 1em;"
      )
      html_lines <- c(
        html_lines,
        paste0(
          "<", h_tag, " style='color: #333; border-bottom: 1px solid #007acc; padding-bottom: 4px; margin-bottom: 6px; margin-top: 14px; ", h_size, "'>",
          header_text,
          "</", h_tag, ">"
        )
      )
      next
    }

    # Title management (lines enclosed in ** only)
    if (stringr::str_detect(line, "^\\*\\*[^*]+\\*\\*$")) {
      close_list()

      title_text <- stringr::str_replace_all(line, "^\\*\\*(.+)\\*\\*$", "\\1")
      # Remove leading number+dot if present (e.g. "1. General Summary")
      title_text <- stringr::str_replace(title_text, "^\\d+\\.\\s*", "")
      html_lines <- c(
        html_lines,
        paste0(
          "<h3 style='color: #333; border-bottom: 1px solid #007acc; padding-bottom: 4px; margin-bottom: 6px; margin-top: 14px;'>",
          title_text,
          "</h3>"
        )
      )
      next
    }

    # Managing bulleted lists (starting with * or -)
    if (stringr::str_detect(line, "^\\s*[*\\-]\\s+")) {
      if (!in_list || list_type != "ul") {
        close_list()
        html_lines <- c(
          html_lines,
          "<ul style='margin-bottom: 6px; margin-top: 4px;'>"
        )
        in_list <- TRUE
        list_type <- "ul"
      }

      item_text <- stringr::str_replace(line, "^\\s*[*\\-]\\s+", "")
      item_text <- format_inline_text(item_text, type = type)

      html_lines <- c(
        html_lines,
        paste0("<li style='margin-bottom: 2px;'>", item_text, "</li>")
      )
      next
    }

    # Managing numbered lists (starting with a number followed by a period)
    if (stringr::str_detect(line, "^\\s*\\d+\\.\\s+")) {
      item_number <- as.integer(stringr::str_extract(line, "^\\s*\\d+"))

      if (!in_list || list_type != "ol") {
        close_list()
        # Use start attribute to continue numbering from where we left off
        start_attr <- if (item_number > 1) paste0(" start='", item_number, "'") else ""
        html_lines <- c(
          html_lines,
          paste0("<ol style='margin-bottom: 6px; margin-top: 4px;'", start_attr, ">")
        )
        in_list <- TRUE
        list_type <- "ol"
      }

      last_ol_number <- item_number
      item_text <- stringr::str_replace(line, "^\\s*\\d+\\.\\s+", "")
      item_text <- format_inline_text(item_text, type = type)
      html_lines <- c(
        html_lines,
        paste0("<li style='margin-bottom: 2px;'>", item_text, "</li>")
      )
      next
    }

    # Non-list, non-header line: close any open list
    close_list()

    # Normal paragraph management
    formatted_line <- format_inline_text(line, type = type)
    html_lines <- c(
      html_lines,
      paste0("<p style='margin-bottom: 5px; margin-top: 2px;'>", formatted_line, "</p>")
    )
  }

  # Close any lists that remain open
  close_list()

  # Close the container div
  html_lines <- c(html_lines, "</div>")

  # Merge all lines
  html_result <- paste(html_lines, collapse = "\n")

  return(html_result)
}

# Auxiliary function for formatting inline text
format_inline_text <- function(text, type = "tall_ai") {
  # Bold management (**text**)
  text <- stringr::str_replace_all(
    text,
    "\\*\\*([^*]+)\\*\\*",
    "<strong>\\1</strong>"
  )

  # Italic management (*text*) - only if it is not already in bold
  text <- stringr::str_replace_all(
    text,
    "(?<!\\*)\\*([^*]+)\\*(?!\\*)",
    "<em>\\1</em>"
  )

  if (type == "tall_ai") {
    # Handling text in quotation marks as inline code ("text")
    text <- stringr::str_replace_all(
      text,
      '"([^"]+)"',
      '<code style="background-color: #f4f4f4; padding: 2px 4px; border-radius: 3px; font-family: \'Courier New\', monospace;">\\1</code>'
    )

    # Handling parentheses with percentages or numerical values
    text <- stringr::str_replace_all(
      text,
      "\\(([^)]*%[^)]*)\\)",
      "<span style='color: #007acc; font-weight: bold;'>(\\1)</span>"
    )
  } else {
    # Handling text in quotation marks as inline code ("text")
    text <- stringr::str_replace_all(
      text,
      '"([^"]+)"',
      '<span style="font-weight: bold; color: #2c3e50;">"\\1"</span>'
    )

    # Handling parentheses with percentages or numerical values
    text <- stringr::str_replace_all(
      text,
      "\\(([^)]*%[^)]*)\\)",
      "<span style='font-weight: bold;'>(\\1)</span>"
    )
  }

  return(text)
}

### token counter for gemini
estimate_gemini_tokens <- function(text) {
  # Function to estimate the number of tokens for Gemini API
  # Parameter: text - text string to analyze

  if (!is.character(text) || length(text) == 0) {
    return(0)
  }

  # Combine everything into a single string if it's a vector
  if (length(text) > 1) {
    text <- paste(text, collapse = " ")
  }

  # Handle empty strings or strings with only whitespace
  text <- trimws(text)
  if (nchar(text) == 0) {
    return(0)
  }

  # Text preprocessing
  # Remove multiple spaces
  text <- gsub("\\s+", " ", text)

  # Count special characters that might be separate tokens
  special_chars <- length(gregexpr(
    "[{}()\\[\\]<>.,;:!?\"'`@#$%^&\\*\\+=|\\~-]",
    text
  )[[1]])
  if (
    special_chars == 1 &&
      attr(
        gregexpr("[{}()\\[\\]<>.,;:!?\"'`@#$%^&\\*\\+=|\\~-]", text)[[1]],
        "match.length"
      ) ==
        -1
  ) {
    special_chars <- 0
  }

  # Count numbers (often tokenized separately)
  numbers <- length(gregexpr("\\b\\d+\\b", text)[[1]])
  if (
    numbers == 1 &&
      attr(gregexpr("\\b\\d+\\b", text)[[1]], "match.length") == -1
  ) {
    numbers <- 0
  }

  # Split text into words
  words <- unlist(strsplit(text, "\\s+"))
  words <- words[words != ""] # Remove empty strings

  if (length(words) == 0) {
    return(0)
  }

  # Estimate tokens based on word length and linguistic characteristics
  token_count <- 0

  # Detect the main language of the text to apply correction factors
  is_cjk <- any(grepl(
    "[\u4e00-\u9fff\u3040-\u309f\u30a0-\u30ff\uac00-\ud7af]",
    words
  ))
  is_arabic <- any(grepl(
    "[\u0600-\u06ff\u0750-\u077f\u08a0-\u08ff\ufb50-\ufdff\ufe70-\ufeff]",
    words
  ))
  is_cyrillic <- any(grepl("[\u0400-\u04ff]", words))
  is_devanagari <- any(grepl("[\u0900-\u097f]", words))

  for (word in words) {
    # Remove punctuation from the word for calculation
    clean_word <- gsub("[[:punct:]]", "", word)
    word_length <- nchar(clean_word)

    if (word_length == 0) {
      next # Skip if word is only punctuation
    }

    # Handle languages with special writing systems
    if (is_cjk) {
      # Chinese, Japanese, Korean: often 1 character = 1 token
      cjk_chars <- length(gregexpr(
        "[\u4e00-\u9fff\u3040-\u309f\u30a0-\u30ff\uac00-\ud7af]",
        clean_word
      )[[1]])
      if (
        cjk_chars > 0 &&
          attr(
            gregexpr(
              "[\u4e00-\u9fff\u3040-\u309f\u30a0-\u30ff\uac00-\ud7af]",
              clean_word
            )[[1]],
            "match.length"
          )[1] !=
            -1
      ) {
        token_count <- token_count + cjk_chars
      } else {
        # Non-CJK characters in CJK text
        token_count <- token_count + max(1, ceiling(word_length / 4))
      }
    } else if (is_arabic) {
      # Arabic: words may be tokenized differently due to complex morphology
      token_count <- token_count + max(1, ceiling(word_length / 3))
    } else if (is_devanagari) {
      # Hindi/Sanskrit: tokenization based on morphemes
      token_count <- token_count + max(1, ceiling(word_length / 3.5))
    } else {
      # Languages with Latin, Cyrillic, and other alphabets
      if (word_length <= 2) {
        # Very short words (articles, short prepositions)
        token_count <- token_count + 1
      } else if (word_length <= 4) {
        # Short words
        token_count <- token_count + 1
      } else if (word_length <= 7) {
        # Medium words
        token_count <- token_count + 1
      } else if (word_length <= 12) {
        # Long words
        token_count <- token_count + ceiling(word_length / 5)
      } else {
        # Very long words (German compounds, technical terms, etc.)
        if (is_cyrillic) {
          token_count <- token_count + ceiling(word_length / 4)
        } else {
          token_count <- token_count + ceiling(word_length / 4.5)
        }
      }
    }
  }

  # Add tokens for special characters and numbers
  # Special characters are often separate tokens
  token_count <- token_count + ceiling(special_chars * 0.8)

  # Numbers may be tokenized in a special way
  token_count <- token_count + ceiling(numbers * 0.5)

  # Apply language-specific correction factors
  if (is_cjk) {
    # CJK languages might have more efficient tokenization
    token_count <- token_count * 0.9
  } else if (is_arabic) {
    # Arabic has complex morphology that can increase tokens
    token_count <- token_count * 1.15
  } else if (is_devanagari) {
    # Hindi and related languages
    token_count <- token_count * 1.1
  } else if (is_cyrillic) {
    # Russian and other Cyrillic languages
    token_count <- token_count * 1.05
  } else {
    # Languages with Latin alphabet (including accented characters)
    has_accents <- any(grepl(
      "[\u00c0-\u017f\u1ea0-\u1ef9]",
      text,
      ignore.case = TRUE
    ))
    if (has_accents) {
      token_count <- token_count * 1.08
    }
  }

  # Round and ensure it's at least 1 if there's any text
  token_count <- max(1, round(token_count))

  return(token_count)
}

loadLanguageModel <- function(file, model_repo = "2.15") {
  home <- homeFolder()

  # setting up the main directory
  path_tall <- file.path(home, "tall")
  path_language_model <- file.path(path_tall, "language_models")
  # check if sub directory exists
  if (file.exists(path_tall)) {
    if (!file.exists(path_language_model)) dir.create(path_language_model)
  } else {
    dir.create(path_tall)
    dir.create(path_language_model)
  }

  # check if the file model already exists
  file_model <- dir(
    path_language_model,
    pattern = paste0(file, "-ud-", model_repo, ".udpipe")
  )[1]

  if (is.na(file_model)) {
    lang_file <- tall_download_model(
      file = file,
      model_dir = path_language_model,
      model_repo = model_repo
    )
    udmodel_lang <- udpipe_load_model(file = lang_file$file_model)
  } else {
    udmodel_lang <- udpipe_load_model(
      file = paste0(path_language_model, "/", file_model)
    )
  }

  return(udmodel_lang)
}

tall_download_model <- function(
  file,
  model_dir = NULL,
  model_repo = "2.15",
  overwrite = TRUE
) {
  filename <- paste0(file, "-ud-", model_repo, ".udpipe")

  url <- file.path(
    "https://raw.githubusercontent.com/massimoaria/tall.language.models/main",
    model_repo,
    filename
  )
  to <- file.path(model_dir, filename)
  download_failed <- FALSE
  download_message <- "OK"
  dl <- suppressWarnings(try(
    utils::download.file(url = url, destfile = to, mode = "wb"),
    silent = TRUE
  ))
  if (inherits(dl, "try-error")) {
    download_failed <- TRUE
    download_message <- as.character(dl)
  } else if (inherits(dl, "integer") && dl != 0) {
    download_failed <- TRUE
    download_message <- "Download failed. Please check internet connectivity"
  }
  if (download_failed) {
    message("Something went wrong")
    message(download_message)
  } else {
    message(sprintf("Downloading finished, model stored at '%s'", to))
  }

  data.frame(
    language = file,
    file_model = to,
    url = url,
    download_failed = download_failed,
    download_message = download_message,
    stringsAsFactors = FALSE
  )
}

updateStats <- function(dfTag, term, statsValues = NULL) {
  if (!is.null(dfTag)) {
    filtered <- LemmaSelection(dfTag) %>%
      dplyr::filter(docSelected)
    statsValues$n_docs <- n_distinct(filtered$doc_id)
    statsValues$n_tokens <- nrow(filtered)
    statsValues$last_update <- Sys.time()
  } else {
    # Reactive value per le statistiche
    statsValues <- reactiveValues(
      n_docs = 0,
      n_tokens = 0,
      last_update = NULL
    )
  }
  return(statsValues)
}

## Tagging Special Entites ----

TaggingCorpusElements <- function(x) {
  if ("upos_specialentities" %in% names(x)) {
    x <- resetSpecialEntities(x)
  } else {
    x$upos_specialentities <- x$upos
  }

  regexList <- c(
    EMAIL = "(?i)([_+a-z0-9-]+(\\.[_+a-z0-9-]+)*@[a-z0-9-]+(\\.[a-z0-9-]+)*(\\.[a-z]{2,14}))",
    URL = "(?<!@)\\b(https?://[\\w.-]+\\.[a-z]{2,6}(/[\\S]*)?|[\\w.-]+\\.(com|org|net|edu|gov|it|uk)\\b)",
    HASH = "^#",
    # Regex più preciso per gli emoji - esclude i caratteri CJK
    EMOJI = "[\U0001F600-\U0001F64F]|[\U0001F300-\U0001F5FF]|[\U0001F680-\U0001F6FF]|[\U0001F1E0-\U0001F1FF]|[\U00002600-\U000026FF]|[\U00002700-\U000027BF]|[\U0001F900-\U0001F9FF]|[\U0001FA70-\U0001FAFF]|[\U00002B00-\U00002BFF]|[\U00003030\U0000303D\U00003297\U00003299]",
    IP_ADDRESS = "\\b\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\.\\d{1,3}\\b",
    MENTION = "^@"
  )
  items <- names(regexList)

  resList <- list()
  j <- 0

  for (i in 1:length(items)) {
    item <- items[i]
    results <- stringi::stri_detect_regex(x$token, regexList[[item]])
    if (sum(results) > 0) {
      j <- j + 1
      resList[[j]] <- data.frame(
        doc_id = x$doc_id[results],
        item = x$token[results],
        tag = item
      )
      x$upos[results] <- toupper(item)
      x$POSSelected[results] <- ifelse(
        x$upos[results] %in% c("HASH", "MENTION", "EMOJI"),
        TRUE,
        FALSE
      )
    }
  }

  if (length(resList) > 0) {
    resList <- dplyr::bind_rows(resList) %>%
      dplyr::filter(!is.na(item))
  } else {
    resList <- tibble::tibble(doc_id = 0, item = NA, tag = "email") %>%
      dplyr::filter(!is.na(item))
  }

  # normalize hash and email
  x <- x %>%
    mutate(
      lemma = case_when(
        upos %in% c("HASH", "EMAIL") ~ tolower(lemma),
        upos == "EMOJI" ~ trimws(lemma),
        TRUE ~ lemma
      ),
      token = case_when(
        upos %in% c("HASH", "EMAIL") ~ tolower(token),
        upos == "EMOJI" ~ trimws(token),
        TRUE ~ token
      )
    )

  if (nrow(resList) > 0) {
    resList <- resList %>%
      mutate(
        item = case_when(
          tag %in% c("HASH", "EMAIL") ~ tolower(item),
          tag == "EMOJI" ~ trimws(item),
          TRUE ~ item
        )
      )
  }

  return(list(resList = resList, x = x))
}

resetSpecialEntities <- function(x) {
  if ("upos_specialentities" %in% names(x)) {
    items <- toupper(c(
      "email",
      "url",
      "hash",
      "emoji",
      "ip_address",
      "mention"
    ))
    x <- x %>%
      mutate(upos = ifelse(upos %in% items, upos_specialentities, upos))
  } else {
    x$upos_specialentities <- x$upos
  }
  return(x)
}

summarySpecialEntities <- function(resList, type = "all") {
  data.frame(
    UPOS = toupper(c("email", "url", "hash", "emoji", "ip_address", "mention")),
    "N. of Items" = rep(0, 6),
    "N. of Docs" = rep(0, 6)
  )

  switch(
    type,
    "all" = {
      resList %>%
        group_by(tag) %>%
        summarise(items = length(unique(item))) %>%
        rename(
          UPOS = tag,
          "Frequency" = items
        ) %>%
        ungroup() %>%
        bind_rows(tibble(
          UPOS = toupper(c(
            "email",
            "url",
            "hash",
            "emoji",
            "ip_address",
            "mention"
          )),
          "Frequency" = rep(0, 6)
        )) %>%
        group_by(UPOS) %>%
        summarize_all(sum)
    },
    {
      label <- toupper(type)
      resList %>%
        rename(UPOS = tag) %>%
        filter(UPOS == label) %>%
        count(item) %>%
        arrange(desc(n))
    }
  )
}

## Custom Lists merging
mergeCustomLists <- function(df, custom_lists, term = "lemma") {
  if (!is.null(custom_lists)) {
    switch(
      term,
      "lemma" = {
        df <- df %>%
          customListsReset() %>%
          # mutate(lemma_norm = ifelse(upos!="PROPN", tolower(lemma), lemma)) %>%
          left_join(custom_lists, by = c("lemma" = "lemma")) %>%
          mutate(
            upos.x = ifelse(!is.na(upos.y), toupper(upos.y), upos.x),
            POSSelected = ifelse(
              upos.x %in%
                c("ADJ", "NOUN", "PROPN", "VERB", "HASH", "EMOJI", "MENTION"),
              TRUE,
              FALSE
            )
          ) %>%
          select(-upos.y) %>%
          rename(upos = upos.x)
      },
      "token" = {
        df <- df %>%
          customListsReset() %>%
          # mutate(lemma_norm = ifelse(upos!="PROPN", tolower(lemma), lemma)) %>%
          left_join(custom_lists, by = c("token" = "token")) %>%
          mutate(
            upos.x = ifelse(!is.na(upos.y), toupper(upos.y), upos.x),
            POSSelected = ifelse(
              upos.x %in% c("ADJ", "NOUN", "PROPN", "VERB"),
              TRUE,
              FALSE
            )
          ) %>%
          select(-upos.y) %>%
          rename(upos = upos.x)
      }
    )
  } else {
    df <- df %>% customListsReset()
  }
  return(df)
}

customListsReset <- function(df) {
  if ("upos_original_custom" %in% names(df)) {
    df <- df %>%
      mutate(upos = upos_original_custom)
  } else {
    df <- df %>%
      mutate(upos_original_custom = upos)
  }

  return(df)
}

### SYNONYMS MERGING ----
applySynonymsReplacement <- function(dfTag, synonyms_df, term_type = "lemma") {
  #' Apply Synonyms Replacement with PoS Tag Update
  #'
  #' Replaces synonym terms with target terms and updates the upos column
  #'
  #' @param dfTag Data frame containing the tagged text data
  #' @param synonyms_df Data frame with synonyms. First column is target term,
  #'        second column is upos tag to assign, remaining columns are synonyms
  #' @param term_type Character. Either "token" or "lemma"
  #'
  #' @return Modified dfTag with synonyms replaced and upos updated

  if (is.null(dfTag) || is.null(synonyms_df)) {
    return(dfTag)
  }

  if (!term_type %in% c("token", "lemma")) {
    stop("term_type must be either \'token\' or \'lemma\'")
  }

  # Get column names
  target_col <- names(synonyms_df)[1]
  upos_col <- names(synonyms_df)[2] # PoS tag to assign
  synonym_cols <- names(synonyms_df)[-(1:2)]

  # Create replacement mapping: synonym -> list(target_term, upos_tag)
  replacement_map <- list()

  for (i in seq_len(nrow(synonyms_df))) {
    target_term <- synonyms_df[[target_col]][i]
    upos_tag <- synonyms_df[[upos_col]][i]

    # Get all synonyms for this target (excluding NA values)
    synonyms <- synonyms_df[i, synonym_cols]
    synonyms <- unlist(synonyms[!is.na(synonyms) & synonyms != ""])

    # Add each synonym to the replacement map with target and upos
    for (syn in synonyms) {
      replacement_map[[tolower(as.character(syn))]] <- list(
        target = as.character(target_term),
        upos = as.character(upos_tag)
      )
    }
  }

  # Build vectorized lookup tables from replacement_map
  target_vec <- vapply(replacement_map, function(x) x$target, character(1))
  upos_vec <- vapply(replacement_map, function(x) x$upos, character(1))

  # Apply replacements to the specified column AND update upos (vectorized)
  col_name <- term_type
  lower_vals <- tolower(as.character(dfTag[[col_name]]))
  matches <- match(lower_vals, names(target_vec))
  matched_idx <- !is.na(matches)
  if (any(matched_idx)) {
    dfTag[[col_name]][matched_idx] <- target_vec[matches[matched_idx]]
    dfTag$upos[matched_idx] <- upos_vec[matches[matched_idx]]
  }

  return(dfTag)
}


### MULTI-WORD CREATION ----

# rake function to create multi-words
rakeReset <- function(x) {
  if ("upos_original" %in% names(x)) {
    x <- x %>%
      select(-"upos") %>%
      rename(upos = upos_original)
  }

  if ("lemma_original_nomultiwords" %in% names(x)) {
    x <- x %>%
      select(-"lemma") %>%
      rename(lemma = lemma_original_nomultiwords)
  }
  if ("token_original_nomultiwords" %in% names(x)) {
    x <- x %>%
      select(-"token") %>%
      rename(token = token_original_nomultiwords)
  }

  if ("ngram" %in% names(x)) {
    x <- x %>%
      select(-"ngram")
  }
  return(x)
}

#' Extract keywords using RAKE or other multiword detection methods (OPTIMIZED)
#'
#' @param x Data frame with annotated tokens
#' @param group Grouping variable (default: "doc_id")
#' @param ngram_max Maximum n-gram length (default: 5)
#' @param ngram_min Minimum n-gram length (default: 2)
#' @param relevant Relevant POS tags (default: c("PROPN", "NOUN", "ADJ", "VERB"))
#' @param freq.min Minimum frequency threshold (default: 10)
#' @param term Term type to use: "lemma" or "token" (default: "lemma")
#' @param type Type of extraction: "automatic" or other (default: "automatic")
#' @param keywordList Optional keyword list (default: NULL)
#' @param method Extraction method: "rake", "pmi", "md", "lfmd", "is" (default: "rake")
### Syntactic Complexity ----

computeSyntacticComplexity <- function(x) {
  req_cols <- c("doc_id", "sentence_id", "token_id", "head_token_id", "dep_rel", "upos")
  if (!all(req_cols %in% names(x))) return(data.frame())

  result <- tall::compute_syntactic_complexity(
    doc_id = as.character(x$doc_id),
    sent_id = as.integer(x$sentence_id),
    token_id = as.integer(x$token_id),
    head_token_id = as.integer(x$head_token_id),
    dep_rel = as.character(x$dep_rel),
    upos = as.character(x$upos)
  )

  # Round numeric columns
  num_cols <- sapply(result, is.numeric)
  result[num_cols] <- lapply(result[num_cols], round, 3)

  return(result)
}

### SVO Triplet Extraction ----

extractSVO <- function(x, term = "lemma", freq.min = 2) {
  req_cols <- c("doc_id", "sentence_id", "token_id", "head_token_id", "dep_rel", "upos", term)
  if (!all(req_cols %in% names(x))) {
    return(data.frame())
  }

  sent_factor <- as.integer(factor(paste0(x$doc_id, "_", x$sentence_id)))

  svo_raw <- tall::extract_svo_triplets(
    sent_id = sent_factor,
    token_id = as.integer(x$token_id),
    head_token_id = as.integer(x$head_token_id),
    dep_rel = as.character(x$dep_rel),
    upos = as.character(x$upos),
    terms = as.character(x[[term]])
  )

  if (nrow(svo_raw) == 0) return(data.frame())

  # Aggregate
  svo_stats <- svo_raw %>%
    group_by(subject, verb, object) %>%
    summarise(
      freq = n(),
      rel_type = first(rel_type),
      .groups = "drop"
    ) %>%
    filter(freq >= freq.min) %>%
    arrange(desc(freq))

  return(svo_stats)
}

### Noun Phrase Extraction via Dependency Parsing ----

extractNounPhrases <- function(x, term = "lemma", freq.min = 3, ngram_max = 5) {
  req_cols <- c("doc_id", "sentence_id", "token_id", "head_token_id", "dep_rel", "upos", term)
  if (!all(req_cols %in% names(x))) {
    stop("Missing required columns for dependency-based NP extraction: ",
      paste(setdiff(req_cols, names(x)), collapse = ", "))
  }

  # Create integer sentence ID for C++
  sent_factor <- as.integer(factor(paste0(x$doc_id, "_", x$sentence_id)))

  # Call C++ function
  np_raw <- tall::extract_noun_phrases(
    sent_id = sent_factor,
    token_id = as.integer(x$token_id),
    head_token_id = as.integer(x$head_token_id),
    dep_rel = as.character(x$dep_rel),
    upos = as.character(x$upos),
    terms = as.character(x[[term]]),
    ngram_max = ngram_max,
    max_gap = 3L
  )

  if (nrow(np_raw) == 0) {
    return(data.frame(keyword = character(), ngram = integer(), freq = integer(), dep_score = numeric()))
  }

  # Aggregate in R (fast on pre-extracted data)
  stats <- np_raw %>%
    group_by(keyword, ngram) %>%
    summarise(freq = n(), .groups = "drop") %>%
    filter(freq >= freq.min) %>%
    mutate(dep_score = freq * ngram) %>%
    arrange(desc(freq))

  return(stats)
}

#'
#' @return List with two elements:
#'   - stats: Data frame with keyword statistics
#'   - dfMW: Data frame with multiword annotations
#'
#' @export
rake <- function(
  x,
  group = "doc_id",
  ngram_max = 5,
  ngram_min = 2,
  relevant = c("PROPN", "NOUN", "ADJ", "VERB"),
  freq.min = 10,
  term = "lemma",
  type = "automatic",
  keywordList = NULL,
  method = "rake"
) {
  # =============================================================================
  # STEP 1: KEYWORD EXTRACTION
  # =============================================================================

  switch(
    type,
    automatic = {
      switch(
        method,
        dep = {
          # Dependency-based Noun Phrase extraction
          stats <- extractNounPhrases(
            x,
            term = term,
            freq.min = freq.min,
            ngram_max = ngram_max
          )
          stats <- stats %>%
            select("keyword", "ngram", "freq", "dep_score")
        },
        rake = {
          # rake multi-word creation
          stats <- keywords_rake(
            x = x,
            term = term,
            group = group,
            ngram_max = ngram_max,
            n_min = freq.min,
            relevant = x$upos %in% relevant
          )

          # identify ngrams>1
          stats <- stats %>%
            dplyr::filter(ngram >= ngram_min)
        },
        pmi = {
          stats <- keywords_collocation(
            x %>% filter(upos %in% relevant),
            term = term,
            group = group,
            ngram_max = ngram_max,
            n_min = freq.min,
            sep = " "
          ) %>%
            select("keyword", "ngram", "freq", "pmi")

          # identify ngrams>1
          stats <- stats %>%
            dplyr::filter(ngram >= ngram_min)
        },
        md = {
          stats <- keywords_collocation(
            x %>% filter(upos %in% relevant),
            term = term,
            group = group,
            ngram_max = ngram_max,
            n_min = freq.min,
            sep = " "
          ) %>%
            select("keyword", "ngram", "freq", "md")

          # identify ngrams>1
          stats <- stats %>%
            dplyr::filter(ngram >= ngram_min)
        },
        lfmd = {
          stats <- keywords_collocation(
            x %>% filter(upos %in% relevant),
            term = term,
            group = group,
            ngram_max = ngram_max,
            n_min = freq.min,
            sep = " "
          ) %>%
            select("keyword", "ngram", "freq", "lfmd")

          # identify ngrams>1 with reka index>reka.min
          stats <- stats %>%
            dplyr::filter(ngram >= ngram_min)
        },
        is = {
          stats <- calculate_ngram_is(
            x,
            max_ngram = ngram_max,
            term = term,
            pos = relevant,
            min_freq = freq.min
          )
          stats <- stats %>%
            mutate(
              keyword = ngram,
              ngram = n_length,
              freq = ngram_freq,
              is_norm = IS_norm
            ) %>%
            select("keyword", "ngram", "freq", "is_norm")
        }
      )
    },
    {
      stats <- keywordList %>%
        mutate(
          keyword = trimws(keyword),
          ngram = lengths(strsplit(keyword, " "))
        )
    }
  )

  # =============================================================================
  # STEP 2: FILTER ORIGINAL TOKEN DF
  # =============================================================================

  if (method %in% c("is", "dep")) {
    include_pos <- c(
      "DET",
      "NOUN",
      "PRON",
      "ADV",
      "VERB",
      "ADP",
      "AUX",
      "ADJ",
      "CCONJ",
      "INTJ",
      "PART",
      "PROPN",
      "SCONJ"
    )
    x2 <- x %>% filter(upos %in% include_pos)
  } else {
    x2 <- x %>% filter(upos %in% relevant)
  }

  # =============================================================================
  # STEP 3: MULTIWORD PROCESSING (OTTIMIZZATO)
  # =============================================================================
  # 🚀 VERSIONE OTTIMIZZATA: Elimina switch duplicato + usa C++
  # Performance: ~100-200x più veloce
  # =============================================================================

  # Pre-calcola lookup per evitare join ripetuti
  keyword_lookup <- stats$ngram
  names(keyword_lookup) <- stats$keyword

  # Seleziona la colonna corretta (evita duplicazione switch)
  term_col <- if (term == "lemma") x2$lemma else x2$token

  # 🚀 USA FUNZIONE C++ OTTIMIZZATA per identificare multiword
  # Questa è la parte che diventa ~80-150x più veloce
  multiword <- txt_recode_ngram_fast(
    x = term_col,
    compound = stats$keyword,
    ngram = stats$ngram,
    sep = " "
  )

  # Operazioni vettorizzate (molto più veloci di mutate multiple)
  upos_multiword <- ifelse(
    term_col == multiword,
    x2$upos,
    ifelse(is.na(multiword), "NGRAM_MERGED", "MULTIWORD")
  )

  # Lookup diretto invece di join (più veloce)
  ngram_values <- keyword_lookup[multiword]

  # Crea risultato direttamente (più veloce di select + mutate)
  x2 <- data.frame(
    doc_id = x2$doc_id,
    term_id = x2$term_id,
    multiword = multiword,
    upos_multiword = upos_multiword,
    ngram = unname(ngram_values),
    stringsAsFactors = FALSE
  )

  # =============================================================================
  # STEP 4: FINAL STATISTICS
  # =============================================================================

  stats <- x2 %>%
    filter(upos_multiword == "MULTIWORD", multiword %in% stats$keyword) %>%
    group_by(multiword) %>%
    select(multiword) %>%
    count() %>%
    ungroup() %>%
    rename(
      keyword = multiword,
      freq = n
    ) %>%
    right_join(
      stats %>% select(-starts_with("freq")),
      by = "keyword"
    ) %>%
    filter(freq >= freq.min) %>%
    arrange(desc(freq))

  return(list(stats = stats, dfMW = x2))
}


# ==============================================================================
# VERSIONE ALTERNATIVA: Con helper function per massima leggibilità
# ==============================================================================
# Usa questa se preferisci una versione ancora più pulita e modulare

rake_v2 <- function(
  x,
  group = "doc_id",
  ngram_max = 5,
  ngram_min = 2,
  relevant = c("PROPN", "NOUN", "ADJ", "VERB"),
  freq.min = 10,
  term = "lemma",
  type = "automatic",
  keywordList = NULL,
  method = "rake"
) {
  # Step 1: Keyword extraction (identico all'originale)
  switch(
    type,
    automatic = {
      switch(
        method,
        rake = {
          stats <- keywords_rake(
            x = x,
            term = term,
            group = group,
            ngram_max = ngram_max,
            n_min = freq.min,
            relevant = x$upos %in% relevant
          ) %>%
            dplyr::filter(ngram >= ngram_min)
        },
        pmi = {
          stats <- keywords_collocation(
            x %>% filter(upos %in% relevant),
            term = term,
            group = group,
            ngram_max = ngram_max,
            n_min = freq.min,
            sep = " "
          ) %>%
            select("keyword", "ngram", "freq", "pmi") %>%
            dplyr::filter(ngram >= ngram_min)
        },
        md = {
          stats <- keywords_collocation(
            x %>% filter(upos %in% relevant),
            term = term,
            group = group,
            ngram_max = ngram_max,
            n_min = freq.min,
            sep = " "
          ) %>%
            select("keyword", "ngram", "freq", "md") %>%
            dplyr::filter(ngram >= ngram_min)
        },
        lfmd = {
          stats <- keywords_collocation(
            x %>% filter(upos %in% relevant),
            term = term,
            group = group,
            ngram_max = ngram_max,
            n_min = freq.min,
            sep = " "
          ) %>%
            select("keyword", "ngram", "freq", "lfmd") %>%
            dplyr::filter(ngram >= ngram_min)
        },
        is = {
          stats <- calculate_ngram_is(
            x,
            max_ngram = ngram_max,
            term = term,
            pos = relevant,
            min_freq = freq.min
          ) %>%
            mutate(
              keyword = ngram,
              ngram = n_length,
              freq = ngram_freq,
              is_norm = IS_norm
            ) %>%
            select("keyword", "ngram", "freq", "is_norm")
        }
      )
    },
    {
      stats <- keywordList %>%
        mutate(
          keyword = trimws(keyword),
          ngram = lengths(strsplit(keyword, " "))
        )
    }
  )

  # Step 2: Filter tokens
  if (method == "is") {
    include_pos <- c(
      "DET",
      "NOUN",
      "PRON",
      "ADV",
      "VERB",
      "ADP",
      "AUX",
      "ADJ",
      "CCONJ",
      "INTJ",
      "PART",
      "PROPN",
      "SCONJ"
    )
    x2 <- x %>% filter(upos %in% include_pos)
  } else {
    x2 <- x %>% filter(upos %in% relevant)
  }

  # Step 3: Process multiwords using optimized helper function
  # 🚀 QUESTA È LA PARTE OTTIMIZZATA
  x2 <- process_multiwords_fast(x2, stats, term)

  # Step 4: Final statistics
  stats <- x2 %>%
    filter(upos_multiword == "MULTIWORD", multiword %in% stats$keyword) %>%
    group_by(multiword) %>%
    select(multiword) %>%
    count() %>%
    ungroup() %>%
    rename(keyword = multiword, freq = n) %>%
    right_join(stats %>% select(-starts_with("freq")), by = "keyword") %>%
    filter(freq >= freq.min) %>%
    arrange(desc(freq))

  return(list(stats = stats, dfMW = x2))
}

applyRake <- function(x, rakeResults, row_sel = NULL, term = "lemma") {
  # Remove ngram column if exists
  if ("ngram" %in% names(x)) {
    x <- x %>% select(-"ngram")
  }

  # Filter stats based on selected rows
  rakeResults$stats <- rakeResults$stats[row_sel, ]
  selected_keywords <- rakeResults$stats$keyword

  # Filter dfMW - optimized logic
  rakeResults$dfMW <- rakeResults$dfMW %>%
    filter(
      (upos_multiword == "MULTIWORD" & multiword %in% selected_keywords) |
        (upos_multiword == "NGRAM_MERGED" & multiword %in% selected_keywords) |
        (!upos_multiword %in% c("MULTIWORD", "NGRAM_MERGED"))
    )

  # Select term column once (avoid switch duplication)
  term_col <- if (term == "lemma") "lemma" else "token"
  term_original_col <- paste0(term_col, "_original_nomultiwords")

  # Join and process in one pipeline
  x <- x %>%
    left_join(rakeResults$dfMW, by = c("doc_id", "term_id")) %>%
    mutate(
      # Combine multiword operations
      multiword = ifelse(is.na(multiword), .data[[term_col]], multiword),
      upos_multiword = ifelse(is.na(upos_multiword), upos, upos_multiword),

      # Update POSSelected
      POSSelected = case_when(
        upos_multiword == "MULTIWORD" ~ TRUE,
        upos_multiword == "NGRAM_MERGED" ~ FALSE,
        TRUE ~ POSSelected
      )
    )

  # Rename upos if needed
  if (!"upos_original" %in% names(x)) {
    names(x)[names(x) == "upos"] <- "upos_original"
  }
  x <- x %>%
    select(-ends_with("upos")) %>%
    rename(upos = upos_multiword)

  # Rename term column if needed
  if (!term_original_col %in% names(x)) {
    names(x)[names(x) == term_col] <- term_original_col
  }
  x <- x %>%
    select(-ends_with(term_col)) %>%
    rename(!!term_col := multiword)

  # Calculate new start/end values for multiwords (vectorized)
  ind <- which(!is.na(x$ngram))
  if (length(ind) > 0) {
    ind2 <- ind + (x$ngram[ind] - 1)
    x$end[ind] <- x$end[ind2]
  }

  # Calculate ngram (vectorized - no group_by needed)
  x$ngram <- ifelse(
    x$upos == "MULTIWORD",
    pmax(
      lengths(strsplit(x$lemma, " ", fixed = TRUE)),
      lengths(strsplit(x$token, " ", fixed = TRUE))
    ),
    NA_integer_
  )

  return(x)
}

### POS TAG SELECTION ----

# create description for pos tag check box ----
posTagAll <- function(df) {
  posLegend <- data.frame(
    pos = c(
      "ADJ",
      "ADP",
      "ADV",
      "AUX",
      "CCONJ",
      "DET",
      "INTJ",
      "NOUN",
      "NUM",
      "PART",
      "PRON",
      "PROPN",
      "PUNCT",
      "SCONJ",
      "SYM",
      "VERB",
      "X",
      "EMAIL",
      "EMOJI",
      "HASH",
      "IP_ADDRESS",
      "MENTION",
      "URL",
      "MULTIWORD"
    ),
    description = c(
      "Adjective",
      "Adposition",
      "Adverb",
      "Auxiliary",
      "Coord. Conjunction",
      "Determiner",
      "Interjection",
      "Noun",
      "Numeral",
      "Particle",
      "Pronoun",
      "Proper Noun",
      "Punctuation",
      "Subord. Conjunction",
      "Symbol",
      "Verb",
      "Other",
      rep("Special Entity", 6),
      "Custom PoS"
    )
  )
  row.names(posLegend) <- posLegend$pos

  pos <- unique(df$upos)
  ordinaryPos <- intersect(posLegend$pos, pos)
  additionalPos <- sort(setdiff(pos, ordinaryPos))
  pos <- c(ordinaryPos, additionalPos)
  description <- c(
    posLegend$description[posLegend$pos %in% pos],
    rep("Custom PoS", length(additionalPos))
  )
  description <- paste(pos, description, sep = ": ")
  obj <- data.frame(pos = pos, description = description) %>%
    filter(!pos == "NGRAM_MERGED")
  return(obj)
}

### GROUP MENU FUNCTIONS -----
noGroupLabels <- function(label) {
  setdiff(
    label,
    c(
      "doc_id",
      "paragraph_id",
      "sentence_id",
      "sentence",
      "start",
      "end",
      "term_id",
      "noSingleChar",
      "token_id",
      "token",
      "lemma",
      "upos",
      "xpos",
      "feats",
      "head_token_id",
      "dep_rel",
      "deps",
      "misc",
      "original_doc_id",
      "ungroupDoc_id",
      "ungroupP_id",
      "ungroupS_id",
      "POSSelected",
      "token_hl",
      "start_hl",
      "end_hl",
      "sentence_hl",
      "lemma_original_nomultiwords",
      "filename",
      "upos_original",
      "folder",
      "docSelected",
      "ngram",
      "doc_selected",
      "noHapax",
      "FrequencyRange",
      "text_original",
      "doc_id_old",
      "split_word",
      "token_original_nomultiwords",
      "lemma_original",
      "upos_specialentities",
      "upos_original_custom",
      "keyness_group"
    )
  )
}

groupByMetadata <- function(dfTag, metadata) {
  if (length(metadata) == 1) {
    ## group texts by a metadata

    if (!"ungroupDoc_id" %in% names(dfTag)) {
      dfTag <- dfTag %>%
        mutate(
          ungroupDoc_id = doc_id,
          ungroupP_id = paragraph_id,
          ungroupS_id = sentence_id
        )
    }

    # newDoc_id <- sprintf(paste0("%0",nchar(length(unique(dfTag$ungroupDoc_id))),"d"), unique_identifier(dfTag, fields=metadata, start_from = 1L))
    dfTag$paragraph_id <- paste0(dfTag$ungroupDoc_id, "_", dfTag$ungroupP_id)
    dfTag$sentence_id <- paste0(dfTag$ungroupDoc_id, "_", dfTag$ungroupS_id)

    # newDoc_id <-dfTag[[metadata]]
    newDoc_id <- ifelse(
      !is.na(dfTag[[metadata]]),
      dfTag[[metadata]],
      "Not Available"
    )

    dfTag <- dfTag %>%
      mutate(doc_id = newDoc_id) %>%
      group_by(doc_id) %>%
      mutate(
        paragraph_id = unique_identifier(paragraph_id),
        sentence_id = unique_identifier(sentence_id)
      ) %>%
      ungroup() %>%
      arrange(doc_id, paragraph_id, sentence_id)
  } else {
    dfTag <- backToOriginalGroups(dfTag)
  }

  return(dfTag)
}

backToOriginalGroups <- function(dfTag) {
  # back to original ungrouped data frame
  if ("ungroupDoc_id" %in% names(dfTag)) {
    dfTag <- dfTag %>%
      mutate(
        doc_id = ungroupDoc_id,
        paragraph_id = ungroupP_id,
        sentence_id = ungroupS_id
      ) %>%
      select(-ungroupDoc_id, -ungroupP_id, -ungroupS_id) %>%
      arrange(doc_id, paragraph_id, sentence_id)
  }
  return(dfTag)
}

### OVERVIEW ----

# Vocabulary calculation

editUI <- function() {
  ## EDIT ----

  ### Split ----

  split <- tabItem(
    tabName = "split_tx",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("scissors"), strong("Split Corpus"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;"),
          br(),
        )
      )
    ),
    tabsetPanel(
      type = "tabs",
      tabPanel(
        title = "Split Results",
        icon = icon("table"),
        br(),
        fluidPage(
          fluidRow(
            column(
              9,
              shinycssloaders::withSpinner(
                DT::DTOutput("splitTextData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            column(
              3,
              fluidRow(
                box(
                  width = 12,
                  div(
                    h3(strong(em("Split texts"))),
                    style = "margin-top:-57px"
                  ),
                  hr(),
                  textInput(
                    inputId = "txSplitWord",
                    label = "Split texts by a sequence of characters (e.g. **H1**)",
                    value = NULL
                  ),
                  hr(),
                  helpText(em(
                    "The minimum sequence of characters required to split the text must consist of at least three characters.",
                    br(),
                    br(),
                    "It's important to note that the text used as a delimiter for splitting is case sensitive (e.g., 'CHAPTER' is different from 'chapter')."
                  )),
                  div(
                    fluidRow(
                      column(
                        4,
                        div(
                          align = "center",
                          title = t_run,
                          do.call(
                            "actionButton",
                            c(
                              run_bttn,
                              list(
                                inputId = "splitTextRun"
                              )
                            )
                          )
                        )
                      ),
                      column(
                        4,
                        div(
                          align = "center",
                          title = t_back,
                          do.call(
                            "actionButton",
                            c(
                              back_bttn,
                              list(
                                inputId = "splitTextBack"
                              )
                            )
                          )
                        )
                      ),
                      column(
                        4,
                        div(
                          align = "center",
                          title = t_save,
                          do.call(
                            "actionButton",
                            c(list(
                              label = NULL,
                              style = "display:block; height: 37px; width: 37px; border-radius: 50%;
                                      border: 1px; margin-top: 16px;",
                              icon = icon("floppy-disk"),
                              inputId = "splitTextSave"
                            ))
                          )
                        )
                      )
                    ),
                    style = "margin-top:-15px"
                  )
                )
              )
            )
          )
        )
      ),
      tabPanel(
        title = "Info & References",
        icon = icon("circle-info"),
        br(),
        fluidPage(
          fluidRow(
            column(1),
            column(
              10,
              br(),
              HTML(infoTexts$split)
            ),
            column(1)
          )
        )
      )
    )
  )

  ### Random selection ----

  randomText <- tabItem(
    tabName = "randomText",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("shuffle"), strong("Random Selection"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;"),
          br(),
        )
      )
    ),
    tabsetPanel(
      type = "tabs",
      tabPanel(
        title = "Random Selection",
        icon = icon("table"),
        br(),
        fluidPage(
          fluidRow(
            column(
              9,
              shinycssloaders::withSpinner(
                DT::DTOutput("randomTextData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            column(
              3,
              fluidRow(
                box(
                  width = 12,
                  div(
                    h3(strong(em("Random Text Selection"))),
                    style = "margin-top:-57px"
                  ),
                  hr(),
                  uiOutput("randomDescription"),
                  br(),
                  "Extract a random sample of texts to analyze",
                  hr(),
                  div(
                    numericInput(
                      "sampleSize",
                      "Sample Size (%)",
                      value = 30,
                      min = 1,
                      max = 100,
                      step = 1
                    ),
                    style = "margin-top:-9px"
                  ),
                  uiOutput("sampleSizeUI"),
                  fluidRow(
                    column(
                      4,
                      title = t_run,
                      do.call(
                        "actionButton",
                        c(
                          run_bttn,
                          list(
                            inputId = "randomTextRun"
                          )
                        )
                      )
                    ),
                    column(
                      4,
                      title = t_back,
                      do.call(
                        "actionButton",
                        c(
                          back_bttn,
                          list(
                            inputId = "randomTextBack"
                          )
                        )
                      )
                    ),
                    column(
                      4,
                      div(
                        align = "center",
                        title = t_save,
                        do.call(
                          "actionButton",
                          c(list(
                            label = NULL,
                            style = "display:block; height: 37px; width: 37px; border-radius: 50%;
                                                          border: 1px; margin-top: 16px;",
                            icon = icon("floppy-disk"),
                            inputId = "randomTextSave"
                          ))
                        )
                      )
                    )
                  )
                )
              )
            )
          )
        )
      ),
      tabPanel(
        title = "Info & References",
        icon = icon("circle-info"),
        br(),
        fluidPage(
          fluidRow(
            column(1),
            column(
              10,
              br(),
              HTML(infoTexts$random)
            ),
            column(1)
          )
        )
      )
    )
  )

  ### EXTERNAL INFORMATION ----

  extInfo <- tabItem(
    tabName = "extInfo",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("table-columns"), strong("External Information"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;"),
          br(),
        )
      )
    ),
    tabsetPanel(
      type = "tabs",
      tabPanel(
        title = "Corpus with External Information",
        icon = icon("table"),
        br(),
        fluidPage(
          fluidRow(
            column(
              9,
              shinycssloaders::withSpinner(
                DT::DTOutput("extInfoData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            column(
              3,
              fluidRow(
                box(
                  width = 12,
                  div(
                    h3(strong(em("Add from a file"))),
                    style = "margin-top:-57px"
                  ),
                  helpText(h5(
                    "To import external information, please make sure that the file
                           to be uploaded is in Excel format and contains a column labeled
                           'doc_id' to identify documents associated to the text(s) imported."
                  )),
                  helpText(h5(
                    "You can download the list of doc_id associated with the imported text files below."
                  )),
                  fluidRow(
                    column(
                      12,
                      div(
                        align = "center",
                        actionButton(
                          inputId = "doc_idExport",
                          label = strong("Export doc_id list"),
                          icon = NULL,
                          style = "border-radius: 15px; border-width: 1px; font-size: 15px;
                                                                    text-align: center; color: #ffff; "
                        )
                      )
                    )
                  ),
                  hr(),
                  fileInput(
                    inputId = "extInfoFile",
                    label = "Import external information",
                    multiple = FALSE,
                    accept = c(
                      ".xls",
                      ".xlsx"
                    ),
                    placeholder = "No file(s) selected"
                  ),
                  hr(),
                  div(
                    fluidRow(
                      column(
                        4,
                        div(
                          title = t_run,
                          do.call(
                            "actionButton",
                            c(
                              run_bttn,
                              list(
                                inputId = "extInfoRun"
                              )
                            )
                          ),
                          align = "center"
                        )
                      ),
                      column(
                        4,
                        div(
                          align = "center",
                          title = "Back to imported text(s) ",
                          do.call(
                            "actionButton",
                            c(
                              back_bttn,
                              list(
                                inputId = "extInfoTextBack"
                              )
                            )
                          )
                        )
                      ),
                      column(
                        4,
                        div(
                          title = t_save,
                          do.call(
                            "actionButton",
                            c(
                              list(
                                label = NULL,
                                style = "display:block; height: 37px; width: 37px; border-radius: 50%;
                                      border: 1px; margin-top: 16px;",
                                icon = icon("floppy-disk")
                              ),
                              list(
                                inputId = "extInfoSave"
                              )
                            )
                          ),
                          align = "center"
                        )
                      )
                    ),
                    style = "margin-top: -15px"
                  )
                )
              )
            )
          )
        )
      ),
      tabPanel(
        title = "Info & References",
        icon = icon("circle-info"),
        br(),
        fluidPage(
          fluidRow(
            column(1),
            column(
              10,
              br(),
              HTML(infoTexts$externalinfo)
            ),
            column(1)
          )
        )
      )
    )
  )
  return(
    list(
      split = split,
      randomText = randomText,
      extInfo = extInfo
    )
  )
}

editServer <- function(input, output, session, values, statsValues) {
  ## EDIT ----

  ### SPLIT ----

  splitDocFunc <- eventReactive(input$splitTextRun, {
    if (nchar(input$txSplitWord) < 3) {
      popUpGeneric(
        title = "Error",
        type = "error",
        color = c("#913333"),
        subtitle = "Sequence must be at least 3 characters long",
        btn_labels = "OK"
      )
    } else {
      values$txt <- splitDoc(values$txt, word = input$txSplitWord)
      popUpGeneric(
        title = paste0("Split by: '", input$txSplitWord, "'"),
        type = "success",
        color = c("#1d8fe1"),
        subtitle = paste0("Now you have ", nrow(values$txt), " documents"),
        btn_labels = "OK"
      )
    }
  })

  ## back to the original txt
  observeEvent(input$splitTextBack, {
    values$txt <- unsplitDoc(values$txt)
    popUpGeneric(
      title = "Restored",
      type = "waiting",
      color = c("#FFA800"),
      subtitle = paste0("Now you have ", nrow(values$txt), " documents"),
      btn_labels = "OK"
    )
  })

  output$splitTextData <- DT::renderDT({
    splitDocFunc()
    DTformat(
      values$txt %>%
        mutate(text = paste0(substr(text, 1, 500), "...")) %>%
        mutate(across(
          c(where(is.character), -text),
          ~ paste0(substr(., 1, 150), "...")
        )) %>%
        select(-c("text_original", ends_with("id_old"))) %>%
        filter(doc_selected) %>%
        select(-"doc_selected"),
      left = 2,
      nrow = 5,
      filter = "none",
      button = TRUE,
      delete = TRUE
    )
  })

  observeEvent(
    eventExpr = {
      input$splitTextSave
    },
    handlerExpr = {
      file_path <- destFolder(
        paste("Tall-Export-File-", sys.time(), ".csv", sep = ""),
        values$wdTall
      )
      readr::write_csv(
        x = values$txt %>%
          filter(doc_selected) %>%
          select(-c("text_original", "doc_selected", ends_with("id_old"))),
        file = file_path,
        na = "NA",
        append = FALSE,
        col_names = TRUE
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  # Back to the original import text ----
  observeEvent(input$extInfoTextBack, {
    values$txt <- values$txt %>%
      mutate(doc_selected = TRUE)
  })

  ### RANDOM SELECTION ----

  output$randomDescription <- renderUI({
    HTML(paste("Number of imported texts: <b>", nrow(values$txt), "</b>"))
  })

  randomTextFunc <- eventReactive(input$randomTextRun, {
    set.seed(values$random_seed)
    values$txt <- samplingText(
      values$txt,
      n = as.numeric(round((input$sampleSize / 100) * nrow(values$txt)), 0)
    )
  })

  output$sampleSizeUI <- renderUI({
    req(input$sampleSize)
    HTML(paste0(
      "<br><h5><em>Number of randomly selected texts: </em><b>",
      as.numeric(round((input$sampleSize / 100) * nrow(values$txt)), 0),
      "</b></h5>"
    ))
  })

  output$randomTextData <- DT::renderDT({
    randomTextFunc()
    DTformat(
      values$txt %>%
        filter(doc_selected) %>%
        select(-c("doc_selected", "text_original")) %>%
        mutate(text = paste0(substr(text, 1, 500), "...")),
      left = 2,
      nrow = 5,
      filter = "none",
      button = TRUE
    )
  })

  ## back to the original txt
  observeEvent(input$randomTextBack, {
    values$txt <- values$txt %>%
      mutate(doc_selected = TRUE)
    updateNumericInput(
      inputId = "sampleSize",
      value = 100
    )
  })

  observeEvent(
    eventExpr = {
      input$randomTextSave
    },
    handlerExpr = {
      file_path <- destFolder(
        paste("Tall-Export-File-", sys.time(), ".csv", sep = ""),
        values$wdTall
      )
      readr::write_csv(
        x = values$txt %>%
          filter(doc_selected) %>%
          select(-c("text_original", "doc_selected", ends_with("id_old"))),
        file = file_path,
        na = "NA",
        append = FALSE,
        col_names = TRUE
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ### EXTERNAL INFORMATION ----

  EXTINFOloading <- eventReactive(input$extInfoRun, {
    req(input$extInfoFile)
    file_extinfo <- input$extInfoFile$datapath
    values$txt <- loadExtInfo(file_extinfo, values$txt)
  })

  output$extInfoData <- DT::renderDT({
    EXTINFOloading()
    DTformat(
      values$txt %>%
        filter(doc_selected) %>%
        select(-c("text_original", "doc_selected")) %>%
        mutate(text = paste0(substr(text, 1, 250), "...")),
      left = 4,
      nrow = 3,
      filter = "top",
      button = TRUE,
      delete = TRUE
    )
  })

  observeEvent(
    eventExpr = {
      input$doc_idExport
    },
    handlerExpr = {
      file <- paste("DocID_List-", sys.time(), ".xlsx", sep = "")
      file <- destFolder(file, values$wdTall)
      suppressWarnings(openxlsx::write.xlsx(
        values$txt %>%
          filter(doc_selected) %>%
          select(doc_id),
        file = file
      ))
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  observeEvent(
    eventExpr = {
      input$extInfoSave
    },
    handlerExpr = {
      file_path <- destFolder(
        paste("Tall-Export-File-", sys.time(), ".csv", sep = ""),
        values$wdTall
      )
      readr::write_csv(
        x = values$txt %>%
          filter(doc_selected) %>%
          select(-c("text_original", "doc_selected")),
        file = file_path,
        na = "NA",
        append = FALSE,
        col_names = TRUE
        # quote = c("needed"),
        # escape = c("backslash"),
        # eol = "\n"
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  # Back to the original import text ----
  observeEvent(input$extInfoTextBack, {
    values$txt <- values$txt %>%
      mutate(doc_selected = TRUE)
  })
}

addDataWb <- function(list_df, wb, sheetname, startRow = 1) {
  l <- length(list_df)
  # startRow <- 1
  for (i in 1:l) {
    df <- list_df[[i]]
    n <- nrow(df)
    writeDataTable(
      wb,
      sheetname,
      df,
      startRow = startRow,
      startCol = 1,
      tableStyle = "TableStyleMedium20"
    )
    startRow <- startRow + n + 3
  }
  return(wb)
}

addDataScreenWb <- function(list_df, wb, sheetname) {
  ind <- which(regexpr(sheetname, wb$sheet_names) > -1)
  if (length(ind) > 0) {
    sheetname <- paste(sheetname, "(", length(ind) + 1, ")", sep = "")
  }
  addWorksheet(wb = wb, sheetName = sheetname, gridLines = FALSE)
  if (!is.null(list_df)) {
    addDataWb(list_df, wb, sheetname, startRow = 1)
    col <- max(unlist(lapply(list_df, ncol))) + 2
  } else {
    col <- 1
  }

  results <- list(wb = wb, col = col, sheetname = sheetname)
  return(results)
}

addGgplotsWb <- function(
  list_plot,
  wb,
  sheetname,
  col,
  width = 10,
  height = 7,
  dpi = 75,
  startRow = 1
) {
  l <- length(list_plot)
  # startRow <- 1
  for (i in 1:l) {
    fileName <- tempfile(
      pattern = "figureImage",
      fileext = ".png"
    )
    if (inherits(list_plot[[i]], "ggplot")) {
      ggsave(
        plot = list_plot[[i]],
        filename = fileName,
        width = width,
        height = height,
        units = "in",
        dpi = dpi
      )
    }
    if (inherits(list_plot[[i]], "igraph")) {
      igraph2PNG(
        x = list_plot[[i]],
        filename = fileName,
        width = width,
        height = height,
        dpi = dpi
      )
    }
    insertImage(
      wb = wb,
      sheet = sheetname,
      file = fileName,
      width = width,
      height = height,
      startRow = startRow,
      startCol = col,
      units = "in",
      dpi = dpi
    )
    startRow <- startRow + (height * 6) + 1
  }
  return(wb)
}

# from igraph to png file
igraph2PNG <- function(x, filename, width = 10, height = 7, dpi = 75) {
  V(x)$centr <- centr_betw(x)$res
  df <- data.frame(
    name = V(x)$label,
    cluster = V(x)$color,
    centr = V(x)$centr
  ) %>%
    group_by(cluster) %>%
    slice_head(n = 3)
  V(x)$label[!(V(x)$label %in% df$name)] <- ""
  png(
    filename = filename,
    width = width,
    height = height,
    unit = "in",
    res = dpi
  )
  grid::grid.draw(plot(x))
  dev.off()
}


addScreenWb <- function(df, wb, width = 14, height = 8, dpi = 75) {
  names(df) <- c("sheet", "file", "n")
  if (nrow(df) > 0) {
    sheet <- unique(df$sheet)
    for (i in 1:length(sheet)) {
      sh <- sheet[i]
      df_sh <- df %>% dplyr::filter(.data$sheet == sh)
      l <- nrow(df_sh)
      startRow <- 30
      for (j in 1:l) {
        fileName <- df_sh$file[j]
        insertImage(
          wb = wb,
          sheet = sh,
          file = fileName,
          width = width,
          height = height,
          startRow = startRow,
          startCol = df_sh$n[j],
          units = "in",
          dpi = dpi
        )
        startRow <- startRow + (height * 10) + 3
      }
    }
  }
  return(wb)
}

addSheetToReport <- function(
  list_df,
  list_plot,
  sheetname,
  wb,
  dpi = 75,
  startRow = 1
) {
  ind <- which(regexpr(sheetname, wb$sheet_names) > -1)
  if (length(ind) > 0) {
    sheetname <- paste(sheetname, "(", length(ind) + 1, ")", sep = "")
  }
  addWorksheet(wb, sheetname, gridLines = FALSE)

  if (!is.null(list_df)) {
    col <- max(unlist(lapply(list_df, ncol))) + 2
    wb <- addDataWb(list_df, wb = wb, sheetname = sheetname, startRow = 1)
  } else {
    col <- 1
  }

  if (!is.null(list_plot)) {
    wb <- addGgplotsWb(
      list_plot,
      wb = wb,
      sheetname = sheetname,
      col = col,
      dpi = dpi,
      startRow = startRow
    )
  }
  # values$sheet_name <- sheetname
  return(wb)
}

#' Extract SVO (Subject-Verb-Object) Triplets via Dependency Parsing (C++ backend)
#'
#' @param sent_id Integer vector of sentence IDs
#' @param token_id Integer vector of token IDs
#' @param head_token_id Integer vector of head token IDs
#' @param dep_rel Character vector of dependency relations
#' @param upos Character vector of universal POS tags
#' @param terms Character vector of term values (lemma or token)
#' @return A data.frame with columns subject, verb, object, rel_type
#' @export
extract_svo_triplets <- function(sent_id, token_id, head_token_id, dep_rel, upos, terms) {
  extract_svo_cpp(sent_id, token_id, head_token_id, dep_rel, upos, terms)
}

#' TALL UI
#'
#' \code{tall} performs text analysis for all.
#'
#' @param port is the TCP port that the application should listen on. If the port is not specified,
#' and the shiny.port option is set (with options(shiny.port = XX)), then that port will be used.
#' Otherwise, use a random port.
#'
#' @param launch.browser If true, the system's default web browser will be launched automatically
#' after the app is started. Defaults to true in interactive sessions only. This value of
#' this parameter can also be a function to call with the application's URL.
#'
#' @param host The IPv4 address that the application should listen on.
#' Defaults to the shiny.host option, if set, or "127.0.0.1" if not.
#'
#' @param maxUploadSize is a integer. The max upload file size argument. Default value is 1000 (megabyte)
#'
#' @return
#' No return value, called for side effects.
#'
#' @export

tall <- function(host = "127.0.0.1", port = NULL,
                 launch.browser = TRUE, maxUploadSize = 1000) {
  shinyOptions(maxUploadSize = maxUploadSize)
  suppressWarnings(
    runApp(system.file("tall", package = "tall"), launch.browser = launch.browser, port = port, host = getOption("shiny.host", host))
  )
}

### TALL TEAM CARDS ----

# Funzione helper aggiornata con titolo e affiliazione
createAuthorCard <- function(
  name,
  title,
  affiliation,
  url,
  photo,
  scholar = FALSE
) {
  tags$a(
    href = url,
    target = "_blank",
    style = "text-decoration: none;",
    div(
      style = "text-align: center; transition: transform 0.3s; cursor: pointer; padding: 15px; background: #f9f9f9; border-radius: 10px;",
      onmouseover = "this.style.transform='translateY(-5px)'; this.style.boxShadow='0 8px 20px rgba(0,0,0,0.15)';",
      onmouseout = "this.style.transform='translateY(0)'; this.style.boxShadow='none';",

      # Foto circolare con bordo
      div(
        style = sprintf(
          "width: 110px; height: 110px; border-radius: 50%%; background-image: url('%s'); background-size: cover; background-position: center; margin: 0 auto 15px; border: 4px solid #3c8dbc; box-shadow: 0 4px 15px rgba(0,0,0,0.2);",
          photo
        )
      ),

      # Nome
      div(
        name,
        style = "font-weight: bold; color: #2c3e50; font-size: 16px; margin-bottom: 8px;"
      ),

      # Titolo
      div(
        title,
        style = "color: #7f8c8d; font-size: 13px; font-weight: 500; margin-bottom: 5px;"
      ),

      # Affiliazione
      div(
        icon("university", style = "margin-right: 5px;"),
        affiliation,
        style = "color: #95a5a6; font-size: 12px; line-height: 1.4; margin-bottom: 10px;"
      ),

      # Link icon
      div(
        if (scholar) {
          tagList(
            icon("graduation-cap", style = "margin-right: 3px;"),
            "Google Scholar"
          )
        } else {
          tagList(
            icon("link", style = "margin-right: 3px;"),
            "Website"
          )
        },
        style = "color: #3c8dbc; font-size: 12px; font-weight: 600;"
      )
    )
  )
}


### UTILS functions ----
# check Internet connection

check_online <- function(
  host = "8.8.8.8",
  timeout = 5,
  # min_success = 1,
  method = "ping" # method = c("ping", "socket", "http")
) {
  #method <- match.arg(method)

  if (method == "ping") {
    # Usa solo il codice di ritorno, non analizza l'output
    ping_cmd <- if (.Platform$OS.type == "windows") {
      sprintf("ping -n 1 -w %d %s", timeout * 1000, host)
    } else {
      sprintf("ping -c 1 -W %d %s", timeout, host)
    }
    exit_code <- suppressWarnings(system(
      ping_cmd,
      ignore.stdout = TRUE,
      ignore.stderr = TRUE
    ))
    return(exit_code == 0)
  } else if (method == "socket") {
    # Connessione TCP a DNS Google (porta 53)
    tryCatch(
      {
        con <- socketConnection(
          host = host,
          port = 53,
          blocking = TRUE,
          open = "r+",
          timeout = timeout
        )
        close(con)
        return(TRUE)
      },
      error = function(e) {
        return(FALSE)
      }
    )
  } else if (method == "http") {
    # Richiesta HTTP
    tryCatch(
      {
        # check if host start with http or https and add if missing
        if (!grepl("^https?://", host)) {
          host <- paste0("https://", host)
        }

        # con <- url("https://www.google.com", open = "rb")
        con <- url(host, open = "rb")
        on.exit(close(con))
        readLines(con, n = 1, warn = FALSE)
        return(TRUE)
      },
      error = function(e) {
        return(FALSE)
      }
    )
  }
}


# Number format abbreviated
format_abbreviated <- function(x) {
  if (is.na(x)) {
    return("--")
  }
  if (x >= 1e6) {
    return(paste0(format(round(x / 1e6, 2), nsmall = 2), "M"))
  } else if (x >= 1e3) {
    return(paste0(format(round(x / 1e3, 0), nsmall = 0), "K"))
  } else {
    return(as.character(x))
  }
}

# total package download
total_downloads <- function(
  pkg_name = "tall",
  from = "2025-01-01",
  to = Sys.Date()
) {
  # Function to get total downloads of a package from CRAN logs
  # Args:
  #   pkg_name: Name of the package as a string
  # Returns:
  #   Total number of downloads as an integer

  # if (!is_Online()) {
  #   return(NA)
  # }

  #today <- Sys.Date()
  if (!is.character(pkg_name) || length(pkg_name) != 1) {
    stop("pkg_name must be a single string.")
  }

  url <- paste0(
    "https://cranlogs.r-pkg.org/downloads/total/",
    from,
    ":",
    to,
    "/",
    pkg_name
  )

  # if (!is_Online(timeout = 1, url)) {
  #   return(NA)
  # }
  if (!check_online(host = url, timeout = 1, method = "http")) {
    return(NA)
  }

  json_text <- tryCatch(
    {
      readLines(url, warn = FALSE)
    },
    error = function(e) {
      return(NA)
    }
  )

  # Se già nel tryCatch è tornato "NA", esci subito
  if (identical(json_text, "NA")) {
    return(NA)
  }

  # Extract the number manually (not robust)
  txt <- unlist(strsplit(json_text, ","))
  txt <- txt[grepl("downloads", txt)]

  if (length(txt) == 0) {
    return(NA)
  }

  downloads <- gsub("[^0-9]", "", txt)

  return(as.integer(downloads))
}

# Scroll to Top Button (Font Awesome version)

scrollToTopButton <- function() {
  tags$div(
    # CSS per il pulsante
    tags$head(
      tags$style(HTML(
        "
        #scrollToTopBtn {
          display: none;
          position: fixed;
          bottom: 20px;
          right: 20px;
          z-index: 99999;
          border: none;
          outline: none;
          background-color: rgba(68, 68, 68, 0.7);
          color: white;
          cursor: pointer;
          padding: 12px;
          border-radius: 4px;
          font-size: 16px;
          width: 45px;
          height: 45px;
          box-shadow: 0 2px 6px rgba(0, 0, 0, 0.3);
          transition: all 0.3s ease;
          backdrop-filter: blur(4px);
        }

        #scrollToTopBtn:hover {
          background-color: rgba(79, 121, 66, 0.85);
          transform: translateY(-3px);
          box-shadow: 0 4px 10px rgba(0, 0, 0, 0.4);
        }

        #scrollToTopBtn:active {
          transform: translateY(-1px);
          box-shadow: 0 2px 6px rgba(0, 0, 0, 0.3);
        }

        #scrollToTopBtn i {
          margin: 0;
          padding: 0;
          line-height: 1;
        }
      "
      ))
    ),

    # Il pulsante HTML
    tags$button(
      id = "scrollToTopBtn",
      icon("arrow-up"),
      onclick = "scrollToTop()",
      title = "Back to top"
    ),

    # JavaScript per gestire lo scroll
    tags$script(HTML(
      "
      // Funzione per mostrare/nascondere il pulsante
      window.onscroll = function() {scrollFunction()};

      function scrollFunction() {
        var btn = document.getElementById('scrollToTopBtn');
        if (document.body.scrollTop > 100 || document.documentElement.scrollTop > 100) {
          btn.style.display = 'block';
        } else {
          btn.style.display = 'none';
        }
      }

      // Funzione per scrollare in alto con animazione smooth
      function scrollToTop() {
        window.scrollTo({
          top: 0,
          behavior: 'smooth'
        });
      }
    "
    ))
  )
}

## CPU cores ------
coresCPU <- function() {
  ## set cores for parallel computing
  ncores <- max(1, parallel::detectCores() - 1)

  ## set cores for windows machines
  if (Sys.info()[["sysname"]] == "Windows") {
    cl <- parallel::makeCluster(ncores)
    doParallel::registerDoParallel(cl)
  }
  return(ncores)
}

## Check Internet connection ----
is_online <- function() {
  # Attempt to connect to a known online resource (e.g., Google's DNS server)
  check_online()
}

## clean raw text before apply tokenization ----

short2long <- function(df, myC) {
  z <- unlist(lapply(myC, function(x) {
    y <- gsub(r"{\s*\([^\)]+\)}", "", x)
    repl <- df$long[df$short == y]
    if (length(repl) == 0) return(x)
    gsub(y, repl, x)
  }))
  names(myC) <- z
  return(myC)
}


## Labels sheets Report

dfLabel <- function() {
  short <- c(
    "Empty Report",
    "Overview",
    "WordsFreq",
    "WordCloud",
    "PoSFreq",
    "Keyness",
    "KWICNetwork",
    "Reinert",
    "CorrespondenceAnalysis",
    "WordNetwork",
    "ThematicMap",
    "EmbeddingTraining",
    "EmbeddingSimilarity",
    "KChoice",
    "ModelEstim",
    "DependencyTree",
    "SyntacticComplexity",
    "SVOTriplets",
    "PolarityDetection",
    "AbstractiveSummarization",
    "ExtractiveSummarization"
  )

  long <- c(
    "Empty Report",
    "Overview",
    "Words Frequency",
    "WordCloud",
    "PoS Tag Frequency",
    "Keyness Analysis",
    "KWIC Network Analysis",
    "Reinert Clustering",
    "Correspondence Analysis",
    "Word Network Analysis",
    "Thematic Map",
    "Word Embedding Training",
    "Word Embedding Similarity",
    "TM-K choice",
    "TM-Model Estimation",
    "Dependency Tree",
    "Syntactic Complexity",
    "SVO Triplets",
    "Polarity Detection",
    "Abstractive Summarization",
    "Extractive Summarization"
  )

  data.frame(short = short, long = long)
}

## Add to Report PopUp

popUp <- function(title = NULL, type = "success", btn_labels = "OK") {
  switch(
    type,
    saved = {
      title <- title
      subtitle <- ""
      btn_colors <- "#1d8fe1"
      showButton <- TRUE
      timer <- 3000
    },
    success = {
      title <- paste(title, "\n added to report", sep = "")
      subtitle <- ""
      btn_colors <- "#1d8fe1"
      showButton <- TRUE
      timer <- 3000
    },
    error = {
      title <- "No results to add to the report "
      subtitle <- "Please Run the analysis and then Add it to the report"
      btn_colors <- "#913333"
      showButton <- TRUE
      timer <- 3000
    },
    waiting = {
      title <- "Please wait... "
      subtitle <- "Adding results to report"
      btn_colors <- "#FFA800"
      showButton <- FALSE
      btn_labels <- NA
      timer <- NA
    }
  )
  if (type == "saved") {
    type <- "success"
  }
  show_alert(
    title = title,
    text = subtitle,
    type = type,
    size = "s",
    closeOnEsc = TRUE,
    closeOnClickOutside = TRUE,
    html = FALSE,
    showConfirmButton = showButton,
    showCancelButton = FALSE,
    btn_labels = btn_labels,
    btn_colors = btn_colors,
    timer = timer,
    imageUrl = "",
    animation = TRUE
  )
}

## generic popup ###


popUpGeneric <- function(
  title = NULL,
  type = "success",
  color = c("#1d8fe1", "#913333", "#FFA800"),
  subtitle = "",
  btn_labels = "OK",
  html = FALSE
) {
  showButton <- TRUE
  timer <- NA
  show_alert(
    title = title,
    text = subtitle,
    type = type,
    size = "s",
    closeOnEsc = TRUE,
    closeOnClickOutside = TRUE,
    html = html,
    showConfirmButton = showButton,
    showCancelButton = FALSE,
    btn_labels = btn_labels,
    btn_colors = color,
    timer = timer,
    imageUrl = "",
    animation = TRUE
  )
}


### UTILITY FUNCTIONS ----
# find home folder

homeFolder <- function() {
  switch(
    Sys.info()[["sysname"]],
    Windows = {
      home <- Sys.getenv("R_USER")
    },
    Linux = {
      home <- Sys.getenv("HOME")
    },
    Darwin = {
      home <- Sys.getenv("HOME")
    }
  )
  return(home)
}
## check working folder

wdFolder <- function() {
  home <- paste0(homeFolder(), "/tall")
  wdFile <- paste0(home, "/tallWD.tall")
  wdTall <- NULL

  if (file.exists(wdFile)) {
    wdTall <- readLines(wdFile)
    if (!file.exists(wdTall)) {
      file.remove(wdFile)
      wdTall <- NULL
    }
  }
  return(wdTall)
}

# add working folder path to a file name

destFolder <- function(file, folder) {
  paste0(folder, "/", file)
}


sys.time <- function() {
  format(Sys.time(), "%Y-%m-%d_%H-%M-%S")
}

## Reset reactive values

resetValues <- function() {
  ### Initial values ----
  values <- list()
  values <- reactiveValues()

  if (inherits(try(pagedown::find_chrome(), silent = T), "try-error")) {
    Chrome_url <- NULL
  } else {
    Chrome_url <- pagedown::find_chrome()
  }

  #  Sys.setenv (CHROMOTE_CHROME = Chrome_url)

  ## chrome configuration for shinyapps server

  if (identical(Sys.getenv("R_CONFIG_ACTIVE"), "shinyapps")) {
    chromote::set_default_chromote_object(
      chromote::Chromote$new(chromote::Chrome$new(
        args = c(
          "--disable-gpu",
          "--no-sandbox",
          "--disable-dev-shm-usage", # required bc the target easily crashes
          c("--force-color-profile", "srgb")
        )
      ))
    )
  }
  ## end configuration

  ## Check if Chrome browser is installed on the computer
  if (is.null(Chrome_url)) {
    showModal(modalDialog(
      title = strong("Warning message!"),
      HTML(
        "Chrome or a Chromium-based browser is not installed on your computer.<br>
If you do not have either of these browsers installed, TALL will be unable to export graphs.<br>
To ensure the functionality of Biblioshiny,
           please download Chrome by <a href='https://www.google.com/chrome/' target='_blank' > <b>clicking here</b></a>."
      ),
      footer = modalButton("Dismiss"),
      easyClose = TRUE
    ))
  } else {
    Sys.setenv(CHROMOTE_CHROME = Chrome_url)
  }
  values$Chrome_url <- Chrome_url
  values$posSpecialSummary <- NULL
  values$biblioshiny <- NULL
  values$resetNeed <- FALSE
  values$normButton <- FALSE
  values$path <- NULL
  values$custom_lists <- NULL
  values$txt <- data.frame()
  values$corpusElements <- data.frame()
  values$txtOriginal <- data.frame()
  values$list_file <- data.frame(sheet = NULL, file = NULL, n = NULL)
  values$POSTagSelected <- ""
  values$wb <- openxlsx::createWorkbook()
  values$dfLabel <- dfLabel()
  values$posMwSel <- c("ADJ", "NOUN", "PROPN") # POS selected by default for multiword creation
  values$myChoices <- "Empty Report"
  values$generalTerm <- "lemma"

  accuracy <- model_accuracy()
  values$accuracy <- accuracy
  languages_df <- langrepo()
  values$language <- "english"
  values$languages <- languages_df
  label_lang <- unique(languages_df$language_name)
  names(label_lang) <- gsub("_", " ", label_lang)
  values$label_lang <- label_lang
  values$treebank <- languages_df$treebank[1]
  values$chapter <- languages_df$chapter
  values$flag <- "GB.svg"
  values$TMplotIndex <- 1
  values$TMdocIndex <- 1
  values$tmTopSentences <- FALSE
  values$selectedGroups <- NULL
  values$selectedFilter <- ""
  values$wdTall <- wdFolder()
  if (is.null(wdFolder())) {
    values$menu <- -2
  } else {
    values$menu <- -1
  }

  ## random seed
  values$random_seed <- 1234

  ## gemini api and model
  home <- homeFolder()
  path_gemini_key <- paste0(
    file.path(home, "tall"),
    "/.tall_gemini_key.txt",
    collapse = ""
  )
  # check if sub directory exists
  values$geminiAPI <- load_api_key(path_gemini_key)
  values$corpus_description <- NULL
  values$gemini_additional <- NULL

  path_gemini_model <- path_gemini_key <- paste0(
    file.path(home, "tall"),
    "/.tall_gemini_model.txt",
    collapse = ""
  )
  gemini_api_model <- loadGeminiModel(path_gemini_model)
  values$gemini_api_model <- gemini_api_model[1]
  values$gemini_output_size <- gemini_api_model[2]

  values$abstractivePrompt <- NULL

  # Initialize Feature Roles values
  values$timeVariable <- NULL
  values$timeAggregation <- NULL
  values$labelVariable <- NULL
  values$keynessVariable <- NULL
  values$keynessGroup1 <- NULL
  values$keynessGroup2 <- NULL
  values$keynessGroupsApplied <- FALSE

  return(values)
}


firstUp <- function(x) {
  x <- tolower(x)
  substr(x, 1, 1) <- toupper(substr(x, 1, 1))
  x
}

## Color palette for plots

colorlist <- function() {
  c(
    "#4DAF4A",
    "#E41A1C",
    "#377EB8",
    "#984EA3",
    "#FF7F00",
    "#A65628",
    "#F781BF",
    "#999999",
    "#66C2A5",
    "#FC8D62",
    "#8DA0CB",
    "#E78AC3",
    "#A6D854",
    "#FFD92F",
    "#B3B3B3",
    "#A6CEE3",
    "#1F78B4",
    "#B2DF8A",
    "#33A02C",
    "#FB9A99",
    "#E31A1C",
    "#FDBF6F",
    "#FF7F00",
    "#CAB2D6",
    "#6A3D9A",
    "#B15928",
    "#8DD3C7",
    "#BEBADA",
    "#FB8072",
    "#80B1D3",
    "#FDB462",
    "#B3DE69",
    "#D9D9D9",
    "#BC80BD",
    "#CCEBC5"
  )
}


## POS selection functions ----
# Set TRUE PoS selected ##

saveTall <- function(
  dfTag,
  custom_lists,
  language,
  treebank,
  menu,
  where,
  file,
  generalTerm,
  corpus_description
) {
  D <- date()
  D <- strsplit(gsub("\\s+", " ", D), " ")
  D <- paste(unlist(D)[c(1, 2, 3, 5)], collapse = " ")
  save(
    dfTag,
    custom_lists,
    language,
    treebank,
    menu,
    D,
    where,
    file = file,
    generalTerm,
    corpus_description
  )
}

### Export Tall analysis in .tall file ----

# SIDEBARMENU DYNAMIC ----

deleteCache <- function() {
  home <- homeFolder()

  # setting up the main directory
  path_tall <- file.path(home, "tall")
  result <- unlink(path_tall, recursive = TRUE)
  btn_labels <- "OK"

  if (result == 0) {
    subtitle <- paste0(
      "The folder '",
      path_tall,
      "' \nand files contained in it have been correctly deleted"
    )
    title <- ""
    btn_colors <- "#1d8fe1"
    showButton <- TRUE
    timer <- 3000
    type <- "success"
  } else {
    subtitle <- paste0(
      "The folder '",
      path_tall,
      "' \ndoes not exist or you don't have permissions to delete it"
    )
    title <- ""
    btn_colors <- "#913333"
    showButton <- TRUE
    timer <- 3000
    type <- "error"
  }
  show_alert(
    title = title,
    text = subtitle,
    type = type,
    size = "s",
    closeOnEsc = TRUE,
    closeOnClickOutside = TRUE,
    html = FALSE,
    showConfirmButton = showButton,
    showCancelButton = FALSE,
    btn_labels = btn_labels,
    btn_colors = btn_colors,
    timer = timer,
    imageUrl = "",
    animation = TRUE
  )
}

## RESET MODAL DIALOG INPUTS

resetModalButtons <- function(session) {
  session$sendCustomMessage("button_id", "null")
  session$sendCustomMessage("click", "null")
  session$sendCustomMessage("click-dend", "null")
  runjs("Shiny.setInputValue('plotly_click-A', null);")
  return(session)
}

#### language models repo ----

# Pacchetti in Depends (caricati automaticamente all'avvio)
#' @import shiny
#' @import shinydashboardPlus

# Pacchetti in Imports - pervasive use (full namespace)
#' @import dplyr
#' @import tidyr
#' @import purrr
#' @import stringr
#' @import Rcpp
#' @import visNetwork
#' @import udpipe

# Pacchetti in Imports - selective use
#' @importFrom base64enc dataURI
#' @importFrom ca ca
#' @importFrom RSpectra svds
#' @importFrom sparkline sparkline spk_chr
#' @importFrom textrank textrank_sentences
#' @importFrom httr2 request req_perform resp_body_json
#' @importFrom ggplot2 ggplot ggsave
#' @importFrom readr write_csv
#' @importFrom strucchange breakpoints
#' @importFrom shinyjs useShinyjs
#' @importFrom shinyWidgets alert updatePrettyCheckboxGroup prettyCheckboxGroup
#' @importFrom DT datatable formatStyle formatRound renderDT DTOutput
#' @importFrom jsonlite fromJSON
#' @importFrom future plan
#' @importFrom later later
#' @importFrom promises promise_all
#' @importFrom rlang sym
#' @importFrom igraph cluster_walktrap graph_from_data_frame membership centr_betw
#' @importFrom plotly add_annotations add_trace add_markers add_polygons
#' @importFrom plotly config event_data event_register layout
#' @importFrom plotly plot_ly plotlyOutput renderPlotly
#' @importFrom Matrix Matrix
#' @importFrom chromote default_chromote_object
#' @importFrom ranger ranger
#' @importFrom umap umap
#' @importFrom word2vec word2vec
#'
#' @import openxlsx
#' @import stm
#' @import topicmodels
#' @import pdftools
#' @import tidygraph
#' @import readxl
#' @import readtext
#' @import fontawesome
#' @import shinycssloaders
#' @import curl
#' @import pagedown
#' @import doParallel
#' @import shinyFiles
#' @import ggraph
#' @import ggwordcloud
#' @useDynLib tall
NULL

## usethis namespace: start
#' @importFrom Rcpp sourceCpp evalCpp
## usethis namespace: end
NULL

.onAttach <- function(...) {
  packageStartupMessage(
    "Please note that our software is open source and available for use, distributed under the MIT license.
\nWhen it is used in a publication, we ask that authors properly cite the following reference:

Aria, M., Spano, M., D'Aniello, L., Cuccurullo, C., & Misuraca, M. (2026).
TALL: Text analysis for all - an interactive R-shiny application for exploring, modeling, and visualizing textual data.
SoftwareX, 34, 102590.
Full paper (Open Access): https://www.sciencedirect.com/science/article/pii/S2352711026000841
Supplementary material: https://ars.els-cdn.com/content/image/1-s2.0-S2352711026000841-mmc1.pdf

Failure to properly cite the software is considered a violation of the license.
\nFor information and bug reports:
- Visit https://www.tall-app.com
- Write a post on https://github.com/massimoaria/tall/issues
\nTo start with the tall app, please digit:
tall()\n"
  )
}

## Keyness Analysis UI ----

keynessUI <- function() {
  tabItem(
    tabName = "keyness",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("key"), strong("Keyness Analysis"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        ),
        div(
          title = t_run,
          column(
            1,
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "run_keyness"
                )
              )
            )
          )
        ),
        div(
          title = t_export,
          column(
            1,
            do.call(
              "actionButton",
              c(
                export_bttn,
                list(
                  inputId = "keynessExport"
                )
              )
            )
          )
        ),
        div(
          title = t_report,
          column(
            1,
            do.call(
              "actionButton",
              c(
                report_bttn,
                list(
                  inputId = "keynessReport"
                )
              )
            )
          )
        ),
        div(
          column(
            1,
            dropdown(
              h4(strong("Options: ")),
              br(),

              # Main Configuration
              div(
                class = "config-section",
                div(
                  class = "config-section-header",
                  icon("gear"),
                  "Main Configuration"
                ),
                selectInput(
                  inputId = "keyness_approach",
                  label = "Analysis Approach:",
                  choices = c(
                    "Reference Corpus" = "reference_corpus",
                    "Two Corpus Comparison" = "two_corpus"
                  ),
                  selected = "reference_corpus"
                ),
                conditionalPanel(
                  condition = "input.keyness_approach == 'two_corpus'",
                  uiOutput("keyness_group_warning")
                ),
                numericInput(
                  inputId = "keyness_n",
                  label = "Max Number of Terms:",
                  value = 1000,
                  min = 100,
                  max = 10000,
                  step = 100
                ),
                numericInput(
                  inputId = "keyness_minchar",
                  label = "Min Character Length:",
                  value = 3,
                  min = 1,
                  max = 10,
                  step = 1
                ),
                selectInput(
                  inputId = "keyness_measure",
                  label = "Keyness Measure:",
                  choices = c(
                    "G2" = "G2",
                    "LogOddsRatio" = "LogOddsRatio",
                    "Phi" = "phi",
                    "Delta P" = "DeltaP"
                  ),
                  selected = "G2"
                )
                # ,
                # checkboxGroupInput(
                #   inputId = "keyness_upos",
                #   label = "POS Tags:",
                #   choices = c("NOUN", "VERB", "ADJ", "ADV"),
                #   selected = c("NOUN", "VERB"),
                #   inline = FALSE
                # )
              ),
              # Graphical Parameters Section
              tags$details(
                class = "advanced-section",
                tags$summary(
                  div(
                    class = "params-section-header",
                    style = "display: flex; justify-content: space-between; align-items: center;",
                    div(
                      icon("eye"),
                      " Graphical Parameters"
                    ),
                    icon(
                      "chevron-down",
                      style = "font-size: 12px;"
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  h4("Bar Plot:"),
                  numericInput(
                    "Keyness_Nbarplot",
                    label = "N. of Words",
                    value = 10,
                    min = 1,
                    max = 20,
                    step = 1
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  h4("Wordcloud:"),
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "Keyness_Nwc",
                        label = "N. of Words",
                        value = 50,
                        min = 10,
                        step = 1,
                        max = 200
                      )
                    ),
                    column(
                      6,
                      numericInput(
                        "Keyness_size",
                        label = "Label Size",
                        value = 100,
                        min = 10,
                        step = 1,
                        max = 200
                      )
                    )
                  )
                ),
                div(
                  style = "margin-top: 10px;",
                  h4("Frequency Context:"),
                  fluidRow(
                    column(
                      6,
                      numericInput(
                        "Keyness_freq_words",
                        label = "N. of Words",
                        value = 20,
                        min = 1,
                        step = 1,
                        max = 100
                      )
                    ),
                    column(
                      6,
                      numericInput(
                        "Keyness_label_spacing",
                        label = "Label spacing",
                        value = 0.15,
                        min = 0,
                        step = 0.01,
                        max = 1
                      )
                    )
                  )
                )
              ),
              style = "gradient",
              right = TRUE,
              animate = TRUE,
              circle = TRUE,
              tooltip = tooltipOptions(title = "Options"),
              icon = icon("sliders", lib = "font-awesome"),
              width = "300px"
            )
          ),
          style = style_opt
        )
      ),
      fluidRow(
        tabsetPanel(
          type = "tabs",

          tabPanel(
            br(),
            title = "Plot",
            icon = icon("chart-column"),

            shinycssloaders::withSpinner(
              plotlyOutput(
                outputId = "keyness_barplot_plotly",
                height = "75vh",
                width = "95.0%"
              ),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          tabPanel(
            br(),
            title = "WordCloud",
            icon = icon("cloud"),
            # Use renderUI to dynamically create wordcloud containers
            fluidRow(
              conditionalPanel(
                condition = "input.keyness_approach == 'reference_corpus'",
                h4(
                  "Target Corpus: High Keyness Words",
                  align = "center",
                  style = "color: #4F7942; margin-bottom: 20px;"
                )
              ),
              shinycssloaders::withSpinner(
                plotOutput(
                  outputId = "keyness_wordcloud_plot",
                  height = "600px",
                  width = "100%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            )
          ),
          tabPanel(
            br(),
            title = "Frequency Context Plot",
            icon = icon("chart-line"),
            conditionalPanel(
              condition = "input.keyness_approach == 'reference_corpus'",
              shinycssloaders::withSpinner(
                plotlyOutput(
                  outputId = "keyness_frequency_plotly",
                  height = "80vh",
                  width = "95.0%"
                ),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            conditionalPanel(
              condition = "input.keyness_approach == 'two_corpus'",
              div(
                style = "padding: 50px; text-align: center;",
                h4(
                  "Frequency Context Plot is only available for Reference Corpus approach"
                ),
                p(
                  "This plot compares your corpus against a general reference corpus.",
                  style = "color: #666; margin-top: 20px;"
                )
              )
            )
          ),
          tabPanel(
            title = "Table",
            icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("keyness_table"),
              color = getOption("spinner.color", default = "#4F7942")
            ),
            align = "center"
          ),
          tabPanel(
            title = "Info & References",
            icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  div(
                    style = "padding: 30px; background: white; border-radius: 8px;
                         box-shadow: 0 2px 4px rgba(0,0,0,0.05); margin-top: 20px;",
                    HTML(infoTexts$keyness)
                  )
                ),
                column(1)
              )
            )
          )
        )
      )
    )
  )
}

## Keyness Analysis Server ----

keynessServer <- function(input, output, session, values) {
  # Reactive counter to force re-rendering
  render_counter <- reactiveVal(0)

  # Check if keyness_group exists when two_corpus approach is selected
  output$keyness_group_warning <- renderUI({
    req(input$keyness_approach == "two_corpus")

    if (!"keyness_group" %in% names(values$dfTag)) {
      div(
        style = "padding: 10px; background-color: #fff3cd; border: 1px solid #ffc107;
                 border-radius: 4px; margin: 10px 0;",
        icon("exclamation-triangle", style = "color: #856404;"),
        span(
          style = "color: #856404; margin-left: 5px;",
          strong("Warning:"),
          " The 'keyness_group' variable is not defined. Please define it in the Feature Roles menu before running the analysis."
        )
      )
    }
  })

  # Keyness Analysis
  keyness_results <- eventReactive(
    input$run_keyness,
    {
      # Get approach
      approach <- input$keyness_approach

      available_upos <- LemmaSelection(values$dfTag) %>%
        dplyr::filter(docSelected) %>%
        distinct(upos) %>%
        pull(upos)

      if (approach == "refrence_corpus") {
        upos <- intersect(
          available_upos,
          c("NOUN", "VERB", "ADJ", "ADV")
        )
      } else {
        upos <- intersect(
          available_upos,
          c(
            "ADJ",
            "ADP",
            "ADV",
            "AUX",
            "CCONJ",
            "DET",
            "EMAIL",
            "EMOJI",
            "HASH",
            "INTJ",
            "IP_ADDRESS",
            "MENTION",
            "MULTIWORD",
            "NOUN",
            "NUM",
            "PRON",
            "PROPN",
            "PUNCT",
            "SCONJ",
            "SYM",
            "TO_REMOVE",
            "URL",
            "VERB"
          )
        )
      }

      # Increment counter to force re-rendering
      render_counter(render_counter() + 1)

      # Check if keyness_group exists for two_corpus approach
      if (
        approach == "two_corpus" && !"keyness_group" %in% names(values$dfTag)
      ) {
        showNotification(
          "Error: 'keyness_group' variable not found. Please define it in Feature Roles menu.",
          type = "error",
          duration = 10
        )
        return(NULL)
      }

      if (length(upos) == 0) {
        showNotification(
          "Error: No valid PoS Tag selected.",
          type = "error",
          duration = 10
        )
        return(NULL)
      }

      withProgress(message = 'Running Keyness Analysis...', value = 0, {
        incProgress(0.3, detail = "Calculating frequencies...")

        # Run keyness analysis with selected approach
        results <- tall_keyness_analysis(
          dfTag = values$dfTag,
          approach = approach,
          language = values$language,
          N = input$keyness_n,
          min.char = input$keyness_minchar,
          upos_list = upos,
          term = ifelse(
            approach == "reference_corpus",
            "token",
            values$generalTerm
          )
        )

        incProgress(0.7, detail = "Generating plots...")
        if (is.null(input$keyness_measure)) {
          measure = "G2"
        } else {
          measure = input$keyness_measure
        }
        if (is.null(input$Keyness_Nbarplot)) {
          Nbarplot = 10
        } else {
          Nbarplot = input$Keyness_Nbarplot
        }

        # Generate bar plot with approach parameter
        results <- c(
          results,
          plot_tall_keyness(
            results$results,
            measure = measure,
            N = Nbarplot,
            approach = approach
          )
        )

        # Generate frequency context plot only for reference_corpus approach
        if (approach == "reference_corpus") {
          plot_frequency <- frequency_context_analysis(
            results$results,
            top_n = input$Keyness_freq_words,
            g2_threshold = quantile(results$results$G2, 0.95),
            label_spacing = input$Keyness_label_spacing
          )

          results <- c(
            results,
            plot_frequency
          )
        }

        if (is.null(input$Keyness_Nwc)) {
          Nwc = 100
        } else {
          Nwc = input$Keyness_Nwc
        }

        # Generate wordcloud(s) based on approach
        if (approach == "reference_corpus") {
          # Single wordcloud for reference corpus approach
          data_target <- results$results %>%
            ungroup() %>%
            select(token, all_of(measure)) %>%
            arrange(desc(.data[[measure]])) %>%
            slice_head(n = Nwc) %>%
            rename(word = token, freq = all_of(measure))
          results <- c(
            results,
            wc_data = list(data_target)
          )
          set.seed(values$random_seed)
          values$keyness_wordcloud_plot <- wordcloud(
            data_target,
            shape = "circle",
            rot_per = 0.2,
            eccentricity = 1.3,
            colors = sample(colorlist(), nrow(data_target), replace = T),
            seed = values$random_seed,
            max_size = input$Keyness_size
          )
        } else if (approach == "two_corpus") {
          # For two corpus: create static wordcloud data

          # Try to split by measure sign first
          data_positive <- results$results %>%
            ungroup() %>%
            filter(.data[[measure]] > 0) %>%
            select(token, all_of(measure)) %>%
            arrange(desc(.data[[measure]])) %>%
            slice_head(n = Nwc)

          data_negative <- results$results %>%
            ungroup() %>%
            filter(.data[[measure]] < 0) %>%
            select(token, all_of(measure)) %>%
            arrange(.data[[measure]]) %>%
            slice_head(n = Nwc) %>%
            mutate(across(all_of(measure), abs))

          # Fallback: if one corpus is empty, split top/bottom
          if (nrow(data_positive) == 0 || nrow(data_negative) == 0) {
            all_data <- results$results %>%
              ungroup() %>%
              select(token, all_of(measure)) %>%
              arrange(desc(.data[[measure]]))

            half_n <- min(Nwc, ceiling(nrow(all_data) / 2))

            data_positive <- all_data %>%
              slice_head(n = half_n)

            data_negative <- all_data %>%
              slice_tail(n = half_n) %>%
              mutate(across(all_of(measure), abs))
          }

          # Prepare final data frames
          data_corpus1 <- data_positive %>%
            rename(word = token, freq = all_of(measure)) %>%
            mutate(Corpus = "Corpus 1")

          data_corpus2 <- data_negative %>%
            rename(word = token, freq = all_of(measure)) %>%
            mutate(Corpus = "Corpus 2")

          results <- c(
            results,
            wc1_data = list(data_corpus1),
            wc2_data = list(data_corpus2)
          )

          values$keyness_wordcloud_plot <- wordcloud(
            bind_rows(
              data_corpus1,
              data_corpus2
            ),
            shape = "circle",
            rot_per = 0.2,
            eccentricity = 1.3,
            colors = c("#4575B4", "#D73027"),
            seed = values$random_seed,
            max_size = input$Keyness_size * 0.7,
            facet_by = "Corpus",
            facet_ncol = 2
          )
        }

        # Format results table
        results$results <- results$results %>%
          rename(Word = token, Obs_Freq = O11, Exp_Freq = O12) %>%
          select(
            Word,
            Sig_corrected,
            Obs_Freq,
            Exp_Freq,
            G2,
            RDF,
            RateRatio,
            OddsRatio,
            LogOddsRatio,
            phi,
            MI,
            PMI,
            DeltaP
          ) %>%
          mutate(
            G2 = round(G2, 3),
            RDF = round(RDF, 3),
            RateRatio = round(RateRatio, 3),
            OddsRatio = round(OddsRatio, 3),
            LogOddsRatio = round(LogOddsRatio, 3),
            phi = round(phi, 3),
            MI = round(MI, 3),
            PMI = round(PMI, 3),
            DeltaP = round(DeltaP, 3)
          )
      })
      values$keyness_results <- results
    },
    ignoreNULL = TRUE
  )

  # Keyness Table Output
  output$keyness_table <- DT::renderDT({
    req(keyness_results())
    DTformat(values$keyness_results$results)
  })

  # Keyness Bar Plot plotly
  output$keyness_barplot_plotly <- plotly::renderPlotly({
    req(keyness_results())
    values$keyness_results$plot_plotly_bar
  })

  output$keyness_frequency_plotly <- plotly::renderPlotly({
    req(keyness_results())
    req(input$keyness_approach == "reference_corpus")
    values$keyness_results$plot_frequency
  })

  output$keyness_wordcloud_plot <- renderPlot(
    {
      req(keyness_results())
      req(input$keyness_approach %in% c("reference_corpus", "two_corpus"))

      if (input$keyness_approach == "reference_corpus") {
        req(values$keyness_results$wc_data)
        nrowdata <- nrow(values$keyness_results$wc_data)
      } else if (input$keyness_approach == "two_corpus") {
        req(values$keyness_results$wc1_data)
        nrowdata <- nrow(values$keyness_results$wc1_data) +
          nrow(values$keyness_results$wc2_data)
      }

      if (nrowdata == 0) {
        plot.new()
        text(
          0.5,
          0.5,
          "No keywords found for Target Corpus",
          cex = 1.5,
          col = "gray"
        )
        return()
      }

      # Use wordcloud package for static plot
      tryCatch(
        {
          values$keyness_wordcloud_plot
        },
        error = function(e) {
          plot.new()
          text(
            0.5,
            0.5,
            paste("Error rendering Corpus :\n", e$message),
            cex = 1,
            col = "red"
          )
        }
      )
    },
    bg = "white"
  )

  ## Report
  observeEvent(input$keynessReport, {
    req(keyness_results())
    if (!is.null(values$keyness_results)) {
      list_df <- list(values$keyness_results$results)

      if (input$keyness_approach == "reference_corpus") {
        list_plot <- list(
          values$keyness_results$plot_ggplot_bar,
          values$keyness_wordcloud_plot,
          values$keyness_results$ggplot_frequency
        )
      } else {
        list_plot <- list(
          values$keyness_results$plot_ggplot_bar,
          values$keyness_wordcloud_plot
        )
      }
      wb <- addSheetToReport(
        list_df,
        list_plot,
        sheetname = "Keyness",
        wb = values$wb,
        dpi = values$report_dpi
      )
      values$wb <- wb
      popUp(
        title = paste0("Keyness Analysis"),
        type = "success"
      )
      values$myChoices <- sheets(values$wb)
    } else {
      popUp(type = "error")
    }
  })

  ## Export Images
  observeEvent(
    eventExpr = {
      input$keynessExport
    },
    handlerExpr = {
      req(keyness_results())
      file1 <- paste("BarPlot-", sys.time(), ".png", sep = "")
      file1 <- destFolder(file1, values$wdTall)
      ggsave(
        filename = file1,
        plot = values$keyness_results$plot_ggplot_bar,
        dpi = values$dpi,
        height = values$h,
        width = values$h * values$aspect,
        bg = "transparent"
      )
      file2 <- paste("WordCloud-", sys.time(), ".png", sep = "")
      file2 <- destFolder(file2, values$wdTall)
      ggsave(
        filename = file2,
        plot = values$keyness_wordcloud_plot,
        dpi = values$dpi,
        height = values$h,
        width = values$h * values$aspect,
        bg = "transparent"
      )
      if (input$keyness_approach == "reference_corpus") {
        file3 <- paste("FrequencyPlot-", sys.time(), ".png", sep = "")
        file3 <- destFolder(file3, values$wdTall)
        ggsave(
          filename = file3,
          plot = values$keyness_results$ggplot_frequency,
          dpi = values$dpi,
          height = values$h,
          width = values$h * values$aspect,
          bg = "transparent"
        )
      }

      popUp(title = "Saved in your working folder", type = "saved")
    }
  )
}


## Keyness Analysis Function ----

tall_keyness_analysis <- function(
  dfTag,
  approach = c("reference_corpus", "two_corpus"),
  language = "english",
  N = 2000,
  min.char = 3,
  upos_list = c("NOUN", "VERB"),
  term = "token"
) {
  # Match the approach argument
  approach <- match.arg(approach)

  # ============================================================================
  # APPROACH 1: REFERENCE CORPUS (original implementation)
  # Compare target corpus against a reference word frequency list
  # ============================================================================
  if (approach == "reference_corpus") {
    # Load word frequency list for the specified language
    word_frequency <- tall_load_wordlist(language = language)

    # Calculate observed frequencies by filtering and aggregating tokens
    x <- dfTag %>%
      dplyr::filter(upos %in% upos_list) %>%
      mutate(token = tolower(token)) %>%
      dplyr::group_by(token) %>%
      dplyr::summarise(n = n()) %>%
      ungroup() %>%
      dplyr::filter(nchar(token) > min.char) %>%
      slice_max(order_by = n, n = N) %>%
      rename(obsFreq = n) %>%
      as_tibble()

    # Calculate total number of words (excluding numbers and punctuation)
    total_words <- nrow(
      dfTag %>%
        dplyr::filter(!upos %in% c("NUM", "PUNCT"))
    )

    # Calculate expected frequencies based on reference corpus
    df <- word_frequency %>%
      as_tibble() %>%
      mutate(token = tolower(token)) %>%
      group_by(token) %>%
      summarise(rel_freq = sum(rel_freq, na.rm = T)) %>%
      ungroup() %>%
      mutate(expFreq = round(rel_freq * total_words, 0))

    # Create frequency table by joining observed and expected frequencies
    freq_table <- x %>%
      left_join(df, by = c("token")) %>%
      select(token, obsFreq, expFreq) %>%
      distinct() %>%
      ungroup() %>%
      # replace NA in expFreq with 1
      dplyr::mutate(
        expFreq = ifelse(is.na(expFreq), 1, expFreq)
      )
  } else if (approach == "two_corpus") {
    # ============================================================================
    # APPROACH 2: TWO CORPUS COMPARISON
    # Compare corpus 1 against corpus 2 to identify distinctive features
    # ============================================================================
    # Check if keyness_group column exists
    if (!"keyness_group" %in% names(dfTag)) {
      stop(
        "Error: dfTag must contain a 'keyness_group' column with values 1 and 2 for two-corpus comparison."
      )
    }

    # Check if keyness_group has exactly two groups (1 and 2) and remove NAs
    groups <- unique(dfTag %>% drop_na(keyness_group) %>% pull(keyness_group))
    if (!all(c(1, 2) %in% groups) || length(groups) != 2) {
      stop(
        "Error: 'keyness_group' must contain exactly two groups with values 1 and 2."
      )
    }

    # Calculate total number of tokens in each corpus (filtered by upos)
    n_tokens_corpus1 <- dfTag %>%
      dplyr::filter(keyness_group == 1 & upos %in% upos_list) %>%
      nrow()

    n_tokens_corpus2 <- dfTag %>%
      dplyr::filter(keyness_group == 2 & upos %in% upos_list) %>%
      nrow()

    # Calculate normalization ratio to scale corpus 2 to corpus 1 size
    normalization_ratio <- n_tokens_corpus1 / n_tokens_corpus2

    # Print corpus sizes for user information
    message(sprintf("Corpus 1 size: %d tokens", n_tokens_corpus1))
    message(sprintf("Corpus 2 size: %d tokens", n_tokens_corpus2))
    message(sprintf("Normalization ratio: %.4f", normalization_ratio))

    # Calculate frequencies for corpus 1
    freq_corpus1 <- dfTag %>%
      dplyr::filter(keyness_group == 1 & upos %in% upos_list) %>%
      dplyr::rename(term_col = all_of(term)) %>%
      mutate(term_col = tolower(term_col)) %>%
      dplyr::group_by(term_col) %>%
      dplyr::summarise(obsFreq = n()) %>%
      ungroup() %>%
      dplyr::filter(nchar(term_col) > min.char) %>%
      dplyr::rename(token = term_col)

    # Calculate frequencies for corpus 2
    freq_corpus2 <- dfTag %>%
      dplyr::filter(keyness_group == 2 & upos %in% upos_list) %>%
      dplyr::rename(term_col = all_of(term)) %>%
      mutate(term_col = tolower(term_col)) %>%
      dplyr::group_by(term_col) %>%
      dplyr::summarise(expFreq_raw = n()) %>%
      ungroup() %>%
      dplyr::filter(nchar(term_col) > min.char) %>%
      dplyr::rename(token = term_col) %>%
      dplyr::mutate(expFreq = expFreq_raw * normalization_ratio)

    # Join the two frequency tables
    # Keep all tokens from both corpora
    freq_table <- freq_corpus1 %>%
      full_join(freq_corpus2 %>% select(token, expFreq), by = "token") %>%
      # Replace NA values with 1 to avoid calculation errors
      dplyr::mutate(
        obsFreq = ifelse(is.na(obsFreq), 1, obsFreq),
        expFreq = ifelse(is.na(expFreq), 1, expFreq)
      ) %>%
      # Select top N tokens by sum of frequencies
      dplyr::mutate(total_freq = obsFreq + expFreq) %>%
      slice_max(order_by = total_freq, n = N) %>%
      select(token, obsFreq, expFreq)
  }

  # ============================================================================
  # COMMON CALCULATIONS FOR BOTH APPROACHES
  # Calculate contingency table statistics and keyness measures
  # ============================================================================

  # Calculate contingency table statistics
  stats_tb2 <- freq_table %>%
    dplyr::mutate(
      C1 = sum(obsFreq),
      C2 = sum(expFreq),
      N = C1 + C2
    ) %>%
    dplyr::mutate(
      R1 = obsFreq + expFreq,
      R2 = N - R1,
      O11 = ifelse(obsFreq == 0, 0.1, obsFreq),
      O12 = R1 - O11,
      O21 = C1 - O11,
      O22 = C2 - O12,
      E11 = (R1 * C1) / N,
      E12 = (R1 * C2) / N,
      E21 = (R2 * C1) / N,
      E22 = (R2 * C2) / N
    ) %>%
    dplyr::select(-obsFreq, -expFreq)

  # Calculate association measures and keyness statistics
  assoc_tb3 <- stats_tb2 %>%
    dplyr::mutate(
      Rws = nrow(.),
      # Fisher's exact test (vectorized with mapply)
      p = mapply(function(a, b, c, d) {
        fisher.test(matrix(c(a, b, c, d), ncol = 2, byrow = TRUE))$p.value
      }, O11, O12, O21, O22),
      # Per thousand word frequencies
      ptw_target = O11 / C1 * 1000,
      ptw_ref = O12 / C2 * 1000,
      # Chi-square statistic
      X2 = (O11 - E11)^2 / E11 +
        (O12 - E12)^2 / E12 +
        (O21 - E21)^2 / E21 +
        (O22 - E22)^2 / E22,
      # Keyness measures
      phi = sqrt(X2 / N),
      MI = log2(O11 / E11),
      t.score = (O11 - E11) / sqrt(O11),
      PMI = log2((O11 / N) / ((O11 + O12) / N) * ((O11 + O21) / N)),
      DeltaP = (O11 / R1) - (O21 / R2),
      LogOddsRatio = log(
        ((O11 + 0.5) * (O22 + 0.5)) / ((O12 + 0.5) * (O21 + 0.5))
      ),
      G2 = 2 *
        ((O11 + 0.001) *
          log((O11 + 0.001) / E11) +
          (O12 + 0.001) * log((O12 + 0.001) / E12) +
          O21 * log(O21 / E21) +
          O22 * log(O22 / E22)),
      # Traditional keyness measures
      RateRatio = ((O11 + 0.001) / (C1 * 1000)) / ((O12 + 0.001) / (C2 * 1000)),
      RateDifference = (O11 / (C1 * 1000)) - (O12 / (C2 * 1000)),
      DifferenceCoefficient = RateDifference /
        ((O11 / (C1 * 1000)) + (O12 / (C2 * 1000))),
      OddsRatio = ((O11 + 0.5) * (O22 + 0.5)) / ((O12 + 0.5) * (O21 + 0.5)),
      LLR = 2 * (O11 * log(O11 / E11)),
      RDF = abs((O11 / C1) - (O12 / C2)),
      PDiff = abs(ptw_target - ptw_ref) / ((ptw_target + ptw_ref) / 2) * 100,
      SignedDKL = ifelse(O11 > 0, O11 * log(O11 / ((O11 + O12) / 2)), 0) -
        ifelse(O12 > 0, O12 * log(O12 / ((O11 + O12) / 2)), 0)
    ) %>%
    # Determine Bonferroni corrected significance
    dplyr::mutate(
      Sig_corrected = dplyr::case_when(
        p / Rws > .05 ~ "n.s.",
        p / Rws > .01 ~ "p < .05*",
        p / Rws > .001 ~ "p < .01**",
        p / Rws <= .001 ~ "p < .001***",
        T ~ "N.A."
      )
    ) %>%
    # Round p-value and determine type/antitype
    dplyr::mutate(
      p = round(p, 5),
      type = ifelse(E11 > O11, "antitype", "type"),
      phi = ifelse(E11 > O11, -phi, phi),
      G2 = ifelse(E11 > O11, -G2, G2)
    ) %>%
    # Filter out non-significant results
    dplyr::filter(Sig_corrected != "n.s.") %>%
    # Arrange by G2 statistic
    dplyr::arrange(-G2) %>%
    # Remove superfluous columns
    dplyr::select(
      -any_of(c(
        "TermCoocFreq",
        "AllFreq",
        "NRows",
        "R1",
        "R2",
        "C1",
        "C2",
        "E12",
        "E21",
        "E22",
        "upp",
        "low",
        "op",
        "t.score",
        "z.score",
        "Rws"
      ))
    ) %>%
    # Relocate important columns to the front
    dplyr::relocate(any_of(c(
      "token",
      "type",
      "Sig_corrected",
      "O11",
      "O12",
      "ptw_target",
      "ptw_ref",
      "G2",
      "RDF",
      "RateRatio",
      "RateDifference",
      "DifferenceCoefficient",
      "LLR",
      "SignedDKL",
      "PDiff",
      "LogOddsRatio",
      "MI",
      "PMI",
      "phi",
      "X2",
      "OddsRatio",
      "DeltaP",
      "p",
      "E11",
      "O21",
      "O22"
    )))

  # Return results with approach information
  return(list(
    results = assoc_tb3,
    approach = approach,
    corpus1_size = if (approach == "two_corpus") n_tokens_corpus1 else NULL,
    corpus2_size = if (approach == "two_corpus") n_tokens_corpus2 else NULL,
    normalization_ratio = if (approach == "two_corpus") {
      normalization_ratio
    } else {
      NULL
    }
  ))
}

### Keyness Wordlist Functions ----
tall_download_wordlist <- function(
  language,
  file_dir = NULL,
  overwrite = TRUE
) {
  filename <- paste0(language, "_word_frequency.keyness")

  if (is.null(file_dir)) {
    file_dir <- paste0(homeFolder(), "/tall/language_models")
  }

  url <- file.path(
    "https://raw.githubusercontent.com/massimoaria/tall.language.models/main/word.frequency.data",
    filename
  )
  to <- file.path(file_dir, filename)
  download_failed <- FALSE
  download_message <- "OK"
  dl <- suppressWarnings(try(
    utils::download.file(url = url, destfile = to, mode = "wb"),
    silent = TRUE
  ))
  if (inherits(dl, "try-error")) {
    download_failed <- TRUE
    download_message <- as.character(dl)
  } else if (inherits(dl, "integer") && dl != 0) {
    download_failed <- TRUE
    download_message <- "Download failed. Please check internet connectivity"
  }
  if (download_failed) {
    message("Something went wrong")
    message(download_message)
  } else {
    message(sprintf("Downloading finished, model stored at '%s'", to))
  }

  return(list(
    download_failed = download_failed,
    download_message = download_message,
    file_wordlist = to
  ))
}

tall_load_wordlist <- function(
  language,
  file_dir = NULL
) {
  wordlist_path <- paste0(homeFolder(), "/tall/language_models/")
  if (!dir.exists(wordlist_path)) {
    dir.create(wordlist_path, recursive = TRUE)
  }

  filename <- paste0(language, "_word_frequency.keyness")

  file_path <- file.path(wordlist_path, filename)

  if (!file.exists(file_path)) {
    message(sprintf(
      "Wordlist for language '%s' not found locally. Downloading...",
      language
    ))
    info <- tall_download_wordlist(
      language = language,
      file_dir = file_dir,
      overwrite = FALSE
    )
    if (info$download_failed) {
      return(NULL)
    }
  }

  load(file_path)

  return(word_frequency)
}


### Keyness Plotting Function ----
plot_tall_keyness <- function(
  assoc_tb3,
  measure = "G2",
  N = 10,
  approach = "reference_corpus"
) {
  # Get top N and bottom N keywords
  top <- assoc_tb3 %>%
    dplyr::ungroup() %>%
    dplyr::arrange(desc(.data[[measure]])) %>%
    dplyr::slice_head(n = N)

  bot <- assoc_tb3 %>%
    dplyr::ungroup() %>%
    dplyr::arrange(desc(.data[[measure]])) %>%
    dplyr::slice_tail(n = N)

  names(top)[which(names(top) %in% measure)] <- "Measure"
  names(bot)[which(names(bot) %in% measure)] <- "Measure"

  combined_data <- rbind(top, bot)

  # Define plot title based on approach
  if (approach == "two_corpus") {
    plot_title <- paste0("Top ", N, " keywords for Corpus 1 vs Corpus 2")
  } else {
    plot_title <- paste0("Top ", N, " keywords for Target vs Reference Corpus")
  }

  # Create ggplot bar plot for top/bottom keywords
  plot_gg_bar <- combined_data %>%
    ggplot(aes(
      x = reorder(token, Measure, mean),
      y = Measure,
      label = Measure,
      fill = type
    )) +
    geom_bar(stat = "identity") +
    geom_text(
      aes(
        y = ifelse(Measure > 0, Measure - 50, Measure + 50),
        label = round(Measure, 1)
      ),
      color = "white",
      size = 3
    ) +
    coord_flip() +
    theme_bw() +
    theme(legend.position = "none") +
    scale_fill_manual(values = c("antitype" = "#D73027", "type" = "#4575B4")) +
    labs(
      title = plot_title,
      x = "Keyword",
      y = paste0("Keyness (", measure, ")")
    )

  # Create plotly bar plot
  combined_data <- combined_data %>%
    dplyr::mutate(
      token = reorder(token, Measure, mean),
      color = ifelse(type == "antitype", "#D73027", "#4575B4")
    )

  plot_plotly_bar <- plotly::plot_ly(
    data = combined_data,
    y = ~token,
    x = ~Measure,
    type = "bar",
    orientation = "h",
    marker = list(color = ~color),
    text = ~ round(Measure, 1),
    textposition = "inside",
    textfont = list(color = "white", size = 12)
  ) %>%
    plotly::layout(
      title = plot_title,
      xaxis = list(title = paste0("Keyness (", measure, ")")),
      yaxis = list(
        title = "Keyword",
        tickmode = "linear",
        categoryorder = "trace",
        autorange = "reversed"
      ),
      showlegend = FALSE,
      margin = list(l = 100)
    )

  # Return results
  return(list(
    plot_ggplot_bar = plot_gg_bar,
    plot_plotly_bar = plot_plotly_bar
  ))
}

### Frequency Plotting Function ----

#' Identify and Visualize Frequency Context in Keyness Results
#'
#' This function analyzes keyness results to identify specialized terminology
#' (high keyness, low frequency) versus fundamental stylistic/thematic differences
#' (high keyness, high frequency).
#'
#' @param keyness_results A data frame containing keyness analysis results with
#'        columns: Word, G2, Obs_Freq (observed frequency in target corpus)
#' @param top_n Number of top high-frequency and low-frequency words to select (default: 15)
#' @param g2_threshold Minimum G2 score to consider (default: 10.83, p < 0.001)
#' @param title Plot title (default: "Frequency Context Analysis")
#' @param label_spacing Spacing factor for labels to avoid overlap (default: 0.08)
#' @param freq_threshold Frequency threshold to separate low/high frequency zones (default: NULL, uses median)
#' @return A plotly scatter plot object
#'
frequency_context_analysis <- function(
  keyness_results,
  top_n = 15,
  g2_threshold = 10.83,
  title = "Frequency Context Analysis",
  label_spacing = 0.08,
  freq_threshold = NULL
) {
  # --- 1. Preparazione Dati ---
  high_keyness <- keyness_results %>%
    dplyr::filter(G2 >= g2_threshold) %>%
    arrange(desc(G2)) %>%
    rename(Word = token, Obs_Freq = O11)

  high_freq_words <- high_keyness %>%
    arrange(desc(Obs_Freq)) %>%
    head(top_n) %>%
    mutate(Category = "High Frequency\n(Fundamental Differences)")

  low_freq_words <- high_keyness %>%
    arrange(Obs_Freq) %>%
    head(top_n) %>%
    anti_join(high_freq_words, by = "Word") %>%
    mutate(Category = "Low Frequency\n(Specialized Terminology)")

  selected_words <- bind_rows(high_freq_words, low_freq_words)

  if (is.null(freq_threshold)) {
    freq_threshold <- mean(c(
      min(high_freq_words$Obs_Freq),
      max(low_freq_words$Obs_Freq)
    ))
  }

  selected_words <- selected_words %>%
    mutate(log_freq = log10(Obs_Freq), log_g2 = log10(G2))

  x_min <- min(selected_words$Obs_Freq) * 0.5
  x_max <- max(selected_words$Obs_Freq) * 2
  y_min <- min(selected_words$G2) * 0.5
  y_max <- max(selected_words$G2) * 2

  # --- 2. Algoritmo di Posizionamento (Ottimale per Plotly) ---
  adjust_labels_alternating <- function(df, spacing = label_spacing) {
    df <- df %>% arrange(log_freq, log_g2)
    df$label_x <- df$log_freq
    df$label_y <- df$log_g2
    df$yanchor <- "bottom"
    df$spacing_mult <- ifelse(
      df$Category == "Low Frequency\n(Specialized Terminology)",
      2.5,
      1.0
    )

    clusters <- list()
    current_cluster <- c(1)
    for (i in 2:nrow(df)) {
      is_close <- FALSE
      for (j in current_cluster) {
        if (
          sqrt(
            (df$log_freq[i] - df$log_freq[j])^2 +
              (df$log_g2[i] - df$log_g2[j])^2
          ) <
            spacing * 2
        ) {
          is_close <- TRUE
          break
        }
      }
      if (is_close) {
        current_cluster <- c(current_cluster, i)
      } else {
        if (length(current_cluster) > 1) {
          clusters[[length(clusters) + 1]] <- current_cluster
        }
        current_cluster <- c(i)
      }
    }
    if (length(current_cluster) > 1) {
      clusters[[length(clusters) + 1]] <- current_cluster
    }

    for (cluster in clusters) {
      cluster_sorted <- cluster[order(df$log_g2[cluster])]
      for (idx in seq_along(cluster_sorted)) {
        i <- cluster_sorted[idx]
        mult <- df$spacing_mult[i]
        if (idx %% 2 == 0) {
          df$yanchor[i] <- "top"
          df$label_y[i] <- df$log_g2[i] - spacing * 0.5 * mult
        } else {
          df$yanchor[i] <- "bottom"
          df$label_y[i] <- df$log_g2[i] + spacing * 0.5 * mult
        }
      }
    }

    for (iter in 1:30) {
      moved <- FALSE
      for (i in 1:nrow(df)) {
        for (j in 1:nrow(df)) {
          if (i >= j) {
            next
          }
          dx <- df$label_x[i] - df$label_x[j]
          dy <- df$label_y[i] - df$label_y[j]
          dist <- sqrt(dx^2 + dy^2)
          min_dist <- spacing *
            0.8 *
            max(df$spacing_mult[i], df$spacing_mult[j])
          if (dist < min_dist && dist > 0) {
            df$label_x[i] <- df$label_x[i] + (dx / dist * (min_dist - dist) / 2)
            df$label_x[j] <- df$label_x[j] - (dx / dist * (min_dist - dist) / 2)
            df$label_y[i] <- df$label_y[i] + (dy / dist * (min_dist - dist) / 2)
            df$label_y[j] <- df$label_y[j] - (dy / dist * (min_dist - dist) / 2)
            moved <- TRUE
          }
        }
      }
      if (!moved) break
    }
    return(df)
  }

  selected_words <- adjust_labels_alternating(selected_words)

  # --- 3. Versione PLOTLY (Invariata) ---
  p_plotly <- plot_ly() %>%
    add_trace(
      data = selected_words %>% filter(grepl("High", Category)),
      x = ~Obs_Freq,
      y = ~G2,
      type = "scatter",
      mode = "markers",
      name = "High Frequency",
      marker = list(
        size = 12,
        color = "#FF8C00",
        line = list(color = "white", width = 1.5),
        opacity = 0.8
      ),
      text = ~Word,
      hovertemplate = "<b>%{text}</b><br>Freq: %{x}<br>G²: %{y:.2f}<extra></extra>"
    ) %>%
    add_trace(
      data = selected_words %>% filter(grepl("Low", Category)),
      x = ~Obs_Freq,
      y = ~G2,
      type = "scatter",
      mode = "markers",
      name = "Low Frequency",
      marker = list(
        size = 12,
        color = "#8B4789",
        line = list(color = "white", width = 1.5),
        opacity = 0.8
      ),
      text = ~Word,
      hovertemplate = "<b>%{text}</b><br>Freq: %{x}<br>G²: %{y:.2f}<extra></extra>"
    )

  annotations_list <- lapply(1:nrow(selected_words), function(i) {
    row <- selected_words[i, ]
    list(
      x = log10(row$Obs_Freq),
      y = log10(row$G2),
      xref = "x",
      yref = "y",
      text = row$Word,
      xanchor = "center",
      yanchor = row$yanchor,
      showarrow = TRUE,
      arrowhead = 0,
      ax = (row$label_x - log10(row$Obs_Freq)) * 100,
      ay = (row$label_y - log10(row$G2)) * 100,
      font = list(size = 10),
      bgcolor = "rgba(255,255,255,0.5)",
      bordercolor = "rgba(128,128,128,0.6)",
      borderwidth = 0.5
    )
  })

  p_plotly <- p_plotly %>%
    layout(
      xaxis = list(title = "Observed Frequency", type = "log"),
      yaxis = list(title = "G² Keyness Score", type = "log"),
      shapes = list(
        list(
          type = "rect",
          xref = "x",
          yref = "y",
          x0 = x_min,
          y0 = y_min,
          x1 = freq_threshold,
          y1 = y_max,
          fillcolor = "rgba(139, 71, 137, 0.08)",
          line = list(width = 0),
          layer = "below"
        ),
        list(
          type = "rect",
          xref = "x",
          yref = "y",
          x0 = freq_threshold,
          y0 = y_min,
          x1 = x_max,
          y1 = y_max,
          fillcolor = "rgba(255, 140, 0, 0.08)",
          line = list(width = 0),
          layer = "below"
        )
      ),
      annotations = annotations_list
    )

  # --- 4. Versione GGPLOT2 (Con contrazione segmenti) ---

  # Creiamo coordinate specifiche per ggplot riducendo la distanza dal punto del 60% (moltiplicatore 0.4)
  selected_words <- selected_words %>%
    mutate(
      gg_label_x = log_freq + (label_x - log_freq) * 0.4,
      gg_label_y = log_g2 + (label_y - log_g2) * 0.4
    )

  p_ggplot <- ggplot(selected_words, aes(x = Obs_Freq, y = G2)) +
    annotate(
      "rect",
      xmin = x_min,
      xmax = freq_threshold,
      ymin = y_min,
      ymax = y_max,
      fill = "#8B4789",
      alpha = 0.08
    ) +
    annotate(
      "rect",
      xmin = freq_threshold,
      xmax = x_max,
      ymin = y_min,
      ymax = y_max,
      fill = "#FF8C00",
      alpha = 0.08
    ) +
    # Segmenti accorciati
    geom_segment(
      aes(xend = 10^gg_label_x, yend = 10^gg_label_y),
      color = "grey80",
      size = 0.3
    ) +
    geom_point(
      aes(fill = Category),
      shape = 21,
      size = 4,
      color = "white",
      stroke = 1
    ) +
    # Etichette posizionate più vicine
    geom_label(
      aes(x = 10^gg_label_x, y = 10^gg_label_y, label = Word),
      size = 2.8,
      label.padding = unit(0.1, "lines"),
      alpha = 0.9,
      label.size = 0.1
    ) +
    scale_x_log10(limits = c(x_min, x_max), expand = c(0, 0)) +
    scale_y_log10(limits = c(y_min, y_max), expand = c(0, 0)) +
    scale_fill_manual(
      values = c(
        "High Frequency\n(Fundamental Differences)" = "#FF8C00",
        "Low Frequency\n(Specialized Terminology)" = "#8B4789"
      )
    ) +
    labs(
      title = title,
      x = "Observed Frequency (Target Corpus)",
      y = "G² Keyness Score"
    ) +
    theme_minimal() +
    theme(legend.position = "bottom", panel.grid.minor = element_blank())

  return(list(plot_frequency = p_plotly, ggplot_frequency = p_ggplot))
}

contextNetwork <- function(df, dfTag, target_word, n = 50, seed = 1234) {
  # Espandi le liste nelle colonne context_before, token, context_after, e upos
  longer_df <- df %>%
    mutate(segment_id = row_number()) %>%
    rowwise() %>%
    mutate(
      words = list(c(
        unlist(context_before),
        target_word,
        unlist(context_after)
      )),
      upos_list = list(unlist(upos))
    ) %>%
    ungroup() %>%
    select(segment_id, words, upos_list) %>%
    unnest(cols = c(words, upos_list)) %>%
    rename(token = words, upos = upos_list)

  uposSelected <- unique(LemmaSelection(dfTag) %>% select(upos) %>% pull())

  net <- network(
    longer_df %>% filter(upos %in% uposSelected),
    term = "token",
    group = c("segment_id"),
    n = n,
    minEdges = 100,
    labelsize = 3,
    opacity = 0.6,
    interLinks = TRUE,
    normalization = "association",
    remove.isolated = FALSE,
    community.repulsion = 0,
    seed = seed
  )

  net$edges <- net$edges %>%
    filter(
      !(color == "#69696920" &
        !(term_from == target_word) &
        !(term_to == target_word))
    )

  vis <- net2vis(net$nodes, net$edges, click = FALSE)
  return(vis)
}

### CLUSTERING ----
clustering <- function(
  dfTag,
  n = 50,
  group = "doc_id",
  term = "lemma",
  minEdges = 25,
  normalization = "association"
) {
  x <- dfTag # %>% dplyr::filter(POSSelected)

  cooc <- coocMatrix(x, term = term, group = group, n = n, pos = TRUE)

  edges <- cooc %>%
    data.frame() %>%
    rename(s = cooc) %>%
    mutate(
      sA = s / (s_from * s_to),
      sC = s / (sqrt(s_from * s_to)),
      sJ = s / (s_from + s_to - s)
    )

  switch(
    normalization,
    none = {
      edges$value <- edges$s
    },
    association = {
      edges$value <- edges$sA
    },
    cosine = {
      edges$value <- edges$sC
    },
    jaccard = {
      edges$value <- edges$sJ
    }
  )

  tailEdges <- quantile(edges$value, 1 - (minEdges / 100), na.rm = T)

  edges <- edges %>%
    dplyr::filter(value >= tailEdges) %>%
    select(term1, term2, value, s, sA, sC, sJ)

  wordnetwork <- graph_from_data_frame(edges %>% select(term1, term2, value))

  # Community detection via optimization of modularity score
  wordnetwork <- as.undirected(wordnetwork) # an undirected graph
  comm <- igraph::cluster_louvain(wordnetwork, weights = E(wordnetwork)$value)
  cluster <- data.frame(
    word = c(cooc$term1, cooc$term2),
    frequency = c(cooc$s_from, cooc$s_to)
  ) %>%
    distinct() %>%
    left_join(
      data.frame(word = comm$names, group = comm$membership),
      by = c("word")
    ) %>%
    drop_na() %>%
    group_by(word) %>%
    summarize(
      group = first(group),
      frequency = max(frequency)
    ) %>%
    arrange(group, desc(frequency))
  obj <- list(cluster = cluster, comm = comm)
}

dend2vis <- function(hc, labelsize, nclusters = 1, community = TRUE) {
  # community = TRUE means that hc is an igraph community detection object
  # community = FALSE mean that hc is a hclust object

  # transform and plot a community igraph object using dendrogram
  if (community) {
    hc <- as.hclust(hc, use.modularity = TRUE)
  }

  h_tail <- round((max(hc$height) * 0.12), 1)

  hc$height <- hc$height + h_tail

  if (!"group" %in% names(hc)) {
    hc$group <- nclusters + 1
  }

  if (nclusters < max(hc$group, na.rm = T)) {
    VIS <- visHclust(
      hc,
      cutree = nclusters,
      colorEdges = "grey60",
      horizontal = TRUE,
      export = FALSE
    )
  } else {
    VIS <- visHclust(
      hc,
      colorEdges = "grey60",
      horizontal = TRUE,
      export = FALSE
    )
  }

  VIS$x$edges <- data.frame(color = unique(VIS$x$edges$color)) %>%
    mutate(new_color = colorlist()[1:nrow(.)]) %>%
    right_join(VIS$x$edges, by = "color") %>%
    select(-color) %>%
    rename(color = new_color)
  VIS$x$nodes <- VIS$x$nodes %>%
    mutate(
      label = ifelse(group != "individual", NA, label),
      group = ifelse(group == "individual", "word", group),
      title = gsub("individuals", "words", title),
      value = 1,
      scaling.min = 10,
      scaling.max = 10
    )
  coords <- VIS$x$nodes %>%
    select(x, y) %>%
    as.matrix()

  edges <- VIS$x$edges
  nodes <- VIS$x$nodes %>%
    select(id, label) %>%
    dplyr::filter(label != "1")

  VIS$x$edges <- edges %>%
    select(-id) %>%
    left_join(nodes, by = c("to" = "id")) %>%
    select(-label.x) %>%
    rename(label = label.y) %>%
    mutate(
      value = 10,
      font.color = color,
      font.size = labelsize * 10,
      font.vadjust = -0.1 * font.size,
      label = ifelse(is.na(label), "", label)
    )

  VIS <- VIS %>%
    visGroups(
      groupname = "group",
      color = "gray90",
      shape = "dot",
      size = 10
    ) %>%
    visGroups(
      groupname = "word",
      font = list(size = 0),
      color = list(
        background = "white",
        border = "#80B1D3",
        highlight = "#e2e9e9",
        hover = "orange"
      ),
      shape = "box"
    ) %>%
    visNodes(font = list(align = VIS$x$nodes$font.align)) %>%
    visNetwork::visOptions(
      highlightNearest = list(
        enabled = T,
        hover = T,
        degree = list(to = 1000, from = 0),
        algorithm = "hierarchical"
      ),
      nodesIdSelection = FALSE,
      manipulation = FALSE,
      height = "100%",
      width = "100%"
    ) %>%
    visNetwork::visInteraction(
      dragNodes = FALSE,
      navigationButtons = F,
      hideEdgesOnDrag = TRUE,
      zoomSpeed = 0.4
    ) %>%
    visIgraphLayout(
      layout = "layout.norm",
      layoutMatrix = coords,
      type = "full"
    ) %>%
    visEdges(font = list(align = "top", size = VIS$x$edges$font.size))

  for (i in 1:nrow(VIS$x$nodes)) {
    if (VIS$x$nodes$group[i] == "group") {
      old_inertia <- as.character(VIS$x$nodes$inertia[i])
      inertia <- as.character(VIS$x$nodes$inertia[i] - h_tail)
      VIS$x$nodes$title[i] <- gsub(old_inertia, inertia, VIS$x$nodes$title[i])
    }
  }

  if ("dtm" %in% names(hc)) {
    k <- max(hc$group, na.rm = TRUE)
    if (nclusters < k) {
      new_groups <- cutree(hc, nclusters)
      for (i in names(new_groups)) {
        VIS$x$nodes$title[VIS$x$nodes$title == i] <- new_groups[i]
      }
    }
  }

  return(VIS)
}

### CORRESPONDENCE ANALYSIS -----

## Correspondence Analysis on Words ----
wordCA <- function(x, n = 50, term = "lemma", group = c("Documents")) {
  switch(
    group,
    Documents = {
      group <- "doc_id"
    },
    Paragraphs = {
      group <- c("doc_id", "paragraph_id")
    },
    Sentences = {
      group <- c("doc_id", "sentence_id")
    }
  )

  # x <- dfTag %>% dplyr::filter(POSSelected)

  if (length(group) > 1) {
    new_doc_id <- unique_identifier(x, fields = group)
  } else {
    new_doc_id <- x$doc_id
  }
  dtm <- document_term_frequencies(
    x %>% mutate(doc_id = new_doc_id),
    term = term
  )

  # dtm <- document_term_frequencies(x, term=term)
  mat <- document_term_matrix(dtm, weight = "freq")
  mat <- as.matrix(dtm_remove_lowfreq(mat, minfreq = 1, maxterms = n))

  res <- ca::ca(mat)

  # Contribute
  Ncol <- min(10, ncol(res$rowcoord))
  contrib <- data.frame((res$colcoord[, 1:Ncol]^2) * res$colmass)
  colnames(contrib) <- paste0("Contrib", 1:ncol(contrib))

  # Cosines squared
  cosine <- data.frame(((res$colcoord[, 1:Ncol]^2) / (res$coldist)))
  colnames(cosine) <- paste0("Cosine", 1:ncol(contrib))

  # Word Coordinates
  wordCoord <- res$colcoord[, 1:Ncol] %>%
    data.frame() %>%
    mutate(
      label = res$colnames,
      inertia = res$colinertia,
      dist = res$coldist,
      mass = res$colmass
    )

  docContrib <- data.frame((res$rowcoord[, 1:Ncol]^2) * res$rowmass)
  docCoord <- res$rowcoord[, 1:Ncol] %>%
    data.frame() %>%
    mutate(
      label = res$rownames,
      inertia = res$rowinertia,
      dist = res$rowdist,
      mass = res$rowmass
    )

  ## Benzecrì correction
  res$eigCorrected <- ((n / (n - 1))^2 * (res$sv - 1 / n)^2)
  # res$eigCorrected[res$eigCorrected<=1/length(res$eigCorrected)] <- 0
  res$eigCorrectedNorm <- res$eigCorrected / sum(res$eigCorrected) * 100

  ## result object
  results <- list(
    ca = res,
    wordCoord = wordCoord,
    contrib = contrib,
    cosine = cosine,
    docContrib = docContrib,
    docCoord = docCoord
  )

  return(results)
}


## caClustering ----
caClustering <- function(
  results,
  method = "ward.D2",
  nDim = 2,
  nclusters = 1,
  lim.contr = 2
) {
  vars <- "Dim"

  # filter by contribution
  contr <- results$contrib %>%
    select(1:nDim) %>%
    ## sum the square of all variables
    mutate(contr_tot = rowSums(across(everything()))) %>%
    filter(contr_tot <= lim.contr) %>%
    rownames_to_column() %>%
    select("rowname")
  #

  dati <- results$wordCoord %>%
    select(starts_with(vars)) %>%
    select(all_of(1:nDim)) %>%
    rownames_to_column() %>%
    filter(rowname %in% contr$rowname) %>%
    column_to_rownames()

  D <- dist(
    dati
  )
  h <- hclust(D, method = method)

  if (nclusters > 1) {
    groups <- cutree(h, k = nclusters)
  } else {
    groups <- rep(1, length(h$labels))
    names(groups) <- h$labels
  }

  # h$group <- groups
  results$clustering <- list(h = h, groups = groups)

  return(results)
}


## CA Plot ----
ca2plotly <- function(
  results,
  dimX = 1,
  dimY = 2,
  topWordPlot = Inf,
  topDocPlot = 20,
  threshold = 0.03,
  labelsize = 16,
  size = 5,
  lim.contr = 2
) {
  target_cols <- names(results$contrib)[c(dimX, dimY)]
  # filter by contribution
  contr <- results$contrib %>%
    as.data.frame() %>%
    rownames_to_column("rowname") %>%
    # Selezioniamo solo le colonne che ci interessano + il nome
    select(rowname, all_of(target_cols)) %>%
    # Calcoliamo la somma delle due colonne (senza quadrato!)
    mutate(contr_tot = rowSums(across(all_of(target_cols)))) %>%
    filter(contr_tot <= lim.contr) %>%
    select(rowname, contr_tot)
  #

  results$contrib <- results$contrib %>%
    rownames_to_column("word") %>%
    dplyr::filter(word %in% contr$rowname) %>%
    select(!word)

  results$wordCoord <- results$wordCoord %>%
    rownames_to_column() %>%
    right_join(contr, by = "rowname") %>%
    column_to_rownames()

  xlabel <- paste0("Dim", dimX)
  ylabel <- paste0("Dim", dimY)
  dimContrLabel <- paste0("Contrib", c(dimX, dimY))
  ymax <- diff(range((results$wordCoord[[ylabel]])))
  xmax <- diff(range((results$wordCoord[[xlabel]])))
  threshold2 <- threshold * mean(xmax, ymax)

  # scaled size for dots
  dotScale <- (results$contrib[, c(dimX, dimY)] * 200)
  dotScale <- ((dotScale[, 1] + dotScale[, 2]) / 2) + size

  # Threshold labels to plot
  thres <- sort(dotScale, decreasing = TRUE)[min(
    topWordPlot,
    nrow(results$wordCoord)
  )]

  Ncol <- sum(substr(names(results$wordCoord), 1, 3) == "Dim")
  # coordinates to plot
  noCol <- setdiff(1:Ncol, c(dimX, dimY))

  results$wordCoord <- results$wordCoord %>%
    select(-any_of(noCol))

  names(results$wordCoord)[1:2] <- c("Dim1", "Dim2")

  results$wordCoord <- results$wordCoord %>%
    mutate(
      dotSize = dotScale,
      groups = results$clustering$groups,
      labelToPlot = ifelse(dotSize >= thres, label, ""),
      font.color = ifelse(
        labelToPlot == "",
        NA,
        adjustcolor(colorlist()[groups], alpha.f = 0.85)
      ),
      font.size = round(dotSize * 2, 0)
    )

  ## Avoid label overlapping
  labelToRemove <- avoidOverlaps(results$wordCoord, threshold = threshold2)
  results$wordCoord <- results$wordCoord %>%
    mutate(
      labelToPlot = ifelse(labelToPlot %in% labelToRemove, "", labelToPlot)
    )

  hull_data <-
    results$wordCoord %>%
    group_by(.data$groups) %>%
    slice(chull(Dim1, Dim2)) %>%
    rename(color = font.color)

  hull_data <- hull_data %>%
    bind_rows(
      hull_data %>% group_by(groups) %>% slice_head(n = 1)
    ) %>%
    mutate(id = row_number()) %>%
    arrange(groups, id)

  hoverText <- paste0(
    " <b>",
    results$wordCoord$label,
    "</b>\n Inertia: ",
    round(results$wordCoord$inertia, 3),
    "\n Mass:   ",
    round(results$wordCoord$mass, 3),
    "\n Contribute:   ",
    round(results$wordCoord$contr_tot, 3)
  )

  ## Plot
  fig <- plot_ly(
    data = results$wordCoord,
    x = results$wordCoord$Dim1,
    y = results$wordCoord$Dim2, # customdata=results$wordCoord,
    type = "scatter",
    mode = "markers",
    marker = list(
      size = dotScale,
      color = adjustcolor(colorlist()[results$wordCoord$groups], alpha.f = 0.3), #' rgb(79, 121, 66, .5)',
      line = list(
        color = adjustcolor(
          colorlist()[results$wordCoord$groups],
          alpha.f = 0.3
        ), #' rgb(79, 121, 66, .8)',
        width = 2
      )
    ),
    text = hoverText,
    hoverinfo = "text",
    alpha = .3
  )

  fig <- fig %>%
    layout(
      yaxis = list(
        title = paste0(
          "Dim ",
          dimY,
          " (",
          round(results$ca$eigCorrectedNorm[2], 2),
          "%)"
        ),
        showgrid = TRUE,
        showline = FALSE,
        showticklabels = TRUE,
        domain = c(0, 1)
      ),
      xaxis = list(
        title = paste0(
          "Dim ",
          dimX,
          " (",
          round(results$ca$eigCorrectedNorm[1], 2),
          "%)"
        ),
        zeroline = TRUE,
        showgrid = TRUE,
        showline = FALSE,
        showticklabels = TRUE
      ),
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)",
      showlegend = F
    )

  for (i in seq_len(max(results$wordCoord$groups))) {
    w <- results$wordCoord %>%
      dplyr::filter(groups == i) %>%
      mutate(
        Dim1 = Dim1 + dotSize * 0.005,
        Dim2 = Dim2 + dotSize * 0.01
      )

    if (max(hull_data$groups > 1)) {
      hull_df <- hull_data %>% dplyr::filter(.data$groups == i)
      fig <- fig %>%
        add_polygons(
          x = hull_df$Dim1,
          y = hull_df$Dim2,
          inherit = FALSE,
          showlegend = FALSE,
          color = I(hull_df$color[1]),
          opacity = 0.3,
          line = list(width = 2),
          text = paste0("Cluster ", i),
          hoverinfo = "text",
          hoveron = "points"
        )
    }

    fig <- fig %>%
      add_annotations(
        data = w,
        x = ~Dim1,
        y = ~Dim2,
        xref = "x1",
        yref = "y",
        text = ~labelToPlot,
        font = list(
          family = "sans serif",
          size = labelsize,
          color = w$font.color[1]
        ),
        showarrow = FALSE
      )
  }

  ## Doc to plot
  if (topDocPlot > 0) {
    results$docContrib <- results$docContrib %>%
      select(-any_of(noCol))
    docContrib <- (results$docContrib[, 1] + results$docContrib[, 2]) / 2

    results$docCoord <- results$docCoord %>%
      mutate(contrib = docContrib)

    docCoord <- results$docCoord %>%
      select(all_of(c(dimX, dimY)), label, contrib) %>%
      slice_max(order_by = contrib, n = topDocPlot) %>%
      mutate(dotScaleDoc = contrib * 50 + size) %>%
      rename(
        labelToPlot = label,
        dotSize = contrib
      )

    names(docCoord)[1:2] <- c("Dim1", "Dim2")

    docLabelToRemove <- avoidOverlaps(docCoord, threshold = threshold2 * 1.5)
    docCoord <- docCoord %>%
      mutate(
        label = labelToPlot,
        labelToPlot = ifelse(
          labelToPlot %in% docLabelToRemove,
          "",
          labelToPlot
        ),
        symbol = "hexagon"
      )

    wDoc <- docCoord %>%
      mutate(
        Dim1 = Dim1 + dotScaleDoc * 0.01,
        Dim2 = Dim2 + dotScaleDoc * 0.015
      )

    fig <- fig %>%
      add_markers(
        data = docCoord,
        x = ~Dim1,
        y = ~Dim2,
        text = ~label,
        # type = "scatter", mode = "markers",
        marker = list(
          symbol = docCoord$symbol,
          size = docCoord$dotScaleDoc,
          color = adjustcolor("#6F7378", alpha.f = 0.3),
          line = list(
            color = adjustcolor("#6F7378", alpha.f = 0.3),
            width = 2
          )
        )
      ) %>%
      add_annotations(
        data = wDoc,
        x = ~Dim1,
        y = ~Dim2,
        xref = "x1",
        yref = "y",
        text = ~labelToPlot,
        font = list(
          family = "sans serif",
          size = labelsize,
          color = adjustcolor("#000000", alpha.f = 0.5)
        ), # 4C4E52
        showarrow = FALSE
      )
  }

  fig <- fig %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        #' toImage',
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )
  # %>%
  #   event_register("plotly_selecting")
  return(fig)
}

## function to avoid label overlapping ----
avoidOverlaps <- function(w, threshold = 0.10, y_scale = 2) {
  w[, 2] <- w[, 2] / y_scale

  Ds <- dist(
    w %>%
      dplyr::filter(labelToPlot != "") %>%
      select(1:2),
    method = "manhattan",
    upper = T
  ) %>%
    dist2df() %>%
    rename(
      from = row,
      to = col,
      dist = value
    ) %>%
    left_join(
      w %>% dplyr::filter(labelToPlot != "") %>% select(labelToPlot, dotSize),
      by = c("from" = "labelToPlot")
    ) %>%
    rename(w_from = dotSize) %>%
    left_join(
      w %>% dplyr::filter(labelToPlot != "") %>% select(labelToPlot, dotSize),
      by = c("to" = "labelToPlot")
    ) %>%
    rename(w_to = dotSize) %>%
    filter(dist < threshold)

  if (nrow(Ds) > 0) {
    st <- TRUE
    i <- 1
    label <- NULL
    case <- "n"

    while (isTRUE(st)) {
      if (Ds$w_from[i] > Ds$w_to[i] & Ds$dist[i] < threshold) {
        case <- "y"
        lab <- Ds$to[i]
      } else if (Ds$w_from[i] <= Ds$w_to[i] & Ds$dist[i] < threshold) {
        case <- "y"
        lab <- Ds$from[i]
      }

      switch(
        case,
        "y" = {
          Ds <- Ds[Ds$from != lab, ]
          Ds <- Ds[Ds$to != lab, ]
          label <- c(label, lab)
        },
        "n" = {
          Ds <- Ds[-1, ]
        }
      )

      if (i >= nrow(Ds)) {
        st <- FALSE
      }
      case <- "n"
      # print(nrow(Ds))
    }
  } else {
    label <- NULL
  }

  label
}

## convert a distance object into a data.frame ----
dist2df <- function(inDist) {
  if (!inherits(inDist, "dist")) {
    stop("wrong input type")
  }
  A <- attr(inDist, "Size")
  B <- if (is.null(attr(inDist, "Labels"))) {
    sequence(A)
  } else {
    attr(inDist, "Labels")
  }
  if (isTRUE(attr(inDist, "Diag"))) {
    attr(inDist, "Diag") <- FALSE
  }
  if (isTRUE(attr(inDist, "Upper"))) {
    attr(inDist, "Upper") <- FALSE
  }
  data.frame(
    row = B[unlist(lapply(sequence(A)[-1], function(x) x:A))],
    col = rep(B[-length(B)], (length(B) - 1):1),
    value = as.vector(inDist)
  )
}


### NETWORK -----

## cooccurrence matrix
coocMatrix <- function(
  x,
  term = "lemma",
  group = "doc_id",
  n = 50,
  pos = TRUE
) {
  term_old <- term
  if (pos) {
    # new_var <- paste0(term,"_upos")
    x$new_var <- paste0(x[[term]], "_", x$upos)
    term <- "new_var"
  } else {
    term <- term_old
  }

  new_doc_id <- unique_identifier(x, fields = group)
  dtm <- document_term_frequencies(
    x %>% mutate(doc_id = new_doc_id),
    term = term
  )

  dtm <- dtm %>%
    mutate(binary = 1)

  if (length(unique(new_doc_id)) == 1) {
    mat <- document_term_matrix(dtm, weight = "freq")
    lab <- colnames(mat)
    mat <- as.numeric(mat)
    names(mat) <- lab
    mat <- sort(mat, decreasing = TRUE)[1:min(n, length(mat))]
    mat <- matrix(mat, 1, length(mat), dimnames = list(x$doc_id[1], names(mat)))
  } else {
    mat <- document_term_matrix(dtm, weight = "binary")
    mat <- dtm_remove_lowfreq(mat, minfreq = 1, maxterms = n)
  }

  mat <- Matrix::crossprod(mat)
  if (sum(mat) - sum(Matrix::diag(mat)) == 0) {
    return(NA)
  }
  mat <- as_cooccurrence(mat)
  mat <- mat %>%
    group_by(term1) %>%
    mutate(s_from = max(cooc)) %>%
    ungroup() %>%
    group_by(term2) %>%
    mutate(s_to = max(cooc)) %>%
    ungroup() %>%
    filter(term1 != term2) %>%
    data.frame()

  if (pos) {
    mat <- mat %>%
      mutate(
        label1 = gsub("_.*", "", term1),
        label2 = gsub("_.*", "", term2),
        upos_from = gsub(".*_", "", term1),
        upos_to = gsub(".*_", "", term2)
      ) %>%
      select(-term1, -term2) %>%
      rename(
        term1 = label1,
        term2 = label2
      )
  } else {
    mat$upos_from <- mat$upos_to <- ""
  }
  return(mat)
}

# word frequency from cooccurence matrix
cooc_freq <- function(cooc) {
  term_freq <- data.frame(
    term = c(cooc$term1, cooc$term2),
    upos = c(cooc$upos_from, cooc$upos_to),
    n = c(cooc$s_from, cooc$s_to)
  ) %>%
    distinct() %>%
    group_by(term, upos) %>%
    summarize(n = sum(n)) %>%
    ungroup()
}

### Dependency-based Co-occurrence Matrix ----
depCoocMatrix <- function(x, term = "lemma", n = 50, dep_rel_filter = "all", dep_rel_custom = NULL) {
  # Define which dep_rel to include
  dep_rels <- switch(
    dep_rel_filter,
    "all" = c("nsubj", "obj", "iobj", "obl", "amod", "nmod", "compound",
      "flat", "flat:name", "conj", "appos", "acl", "advcl",
      "xcomp", "ccomp", "advmod", "nummod"),
    "noun_mod" = c("amod", "nmod", "compound", "flat", "flat:name", "nummod", "appos"),
    "svo" = c("nsubj", "obj", "iobj"),
    "custom" = if (!is.null(dep_rel_custom)) dep_rel_custom else c("nsubj", "obj", "amod", "nmod", "compound"),
    c("nsubj", "obj", "amod", "nmod", "compound")
  )

  # Required columns
  if (!all(c("token_id", "head_token_id", "dep_rel") %in% names(x))) {
    return(NA)
  }

  # Get term values and build sentence key
  term_vals <- x[[term]]
  tok_id <- as.integer(x$token_id)
  head_tok_id <- as.integer(x$head_token_id)
  sent_key <- paste0(x$doc_id, "_", x$sentence_id)

  # Filter to relevant dep_rel
  rel_idx <- which(x$dep_rel %in% dep_rels & head_tok_id > 0)

  if (length(rel_idx) == 0) return(NA)

  # Build head lookup: sent_key:token_id -> term value
  lookup_key <- paste0(sent_key, ":", tok_id)
  term_lookup <- term_vals
  names(term_lookup) <- lookup_key

  # For each relevant dependency, get (dependent_term, head_term) pairs
  dep_terms <- term_vals[rel_idx]
  head_keys <- paste0(sent_key[rel_idx], ":", head_tok_id[rel_idx])
  head_terms <- term_lookup[head_keys]

  # Remove pairs where head term was not found
  valid <- !is.na(head_terms) & dep_terms != head_terms
  dep_terms <- dep_terms[valid]
  head_terms <- unname(head_terms[valid])

  if (length(dep_terms) == 0) return(NA)

  # Normalize pair order (alphabetical) to avoid duplicates
  term1 <- ifelse(dep_terms < head_terms, dep_terms, head_terms)
  term2 <- ifelse(dep_terms < head_terms, head_terms, dep_terms)

  # Count co-occurrences
  pairs <- data.frame(term1 = term1, term2 = term2, stringsAsFactors = FALSE)
  cooc <- pairs %>%
    group_by(term1, term2) %>%
    summarise(cooc = n(), .groups = "drop")

  # Get top N terms by total frequency
  all_terms <- c(cooc$term1, cooc$term2)
  term_freq <- sort(table(all_terms), decreasing = TRUE)
  top_terms <- names(term_freq)[seq_len(min(n, length(term_freq)))]

  # Filter to top terms
  cooc <- cooc %>%
    filter(term1 %in% top_terms & term2 %in% top_terms)

  if (nrow(cooc) == 0) return(NA)

  # Calculate s_from and s_to (max co-occurrence per term)
  cooc <- cooc %>%
    group_by(term1) %>%
    mutate(s_from = max(cooc)) %>%
    ungroup() %>%
    group_by(term2) %>%
    mutate(s_to = max(cooc)) %>%
    ungroup() %>%
    mutate(upos_from = "", upos_to = "") %>%
    data.frame()

  return(cooc)
}

# network() function with seed fixing and improved community repulsion

network <- function(
  x,
  term = "lemma",
  group = c("doc_id", "sentence_id"),
  n,
  minEdges,
  labelsize = 4,
  opacity = 0.6,
  interLinks = FALSE,
  normalization = "none",
  remove.isolated = FALSE,
  community.repulsion = 0.5,
  seed = 123,
  cluster = "louvain",
  cooc_type = "cooc",
  dep_rel_filter = "all",
  dep_rel_custom = NULL
) {
  # size scaling
  scalemin <- 20 * (1 + labelsize / 5)
  scalemax <- 70 * (1 + labelsize / 5)

  colorlist <- colorlist()

  # params
  shape <- "dot"
  opacity.min <- 0.4

  if (cooc_type == "dep") {
    cooc <- depCoocMatrix(x, term = term, n = n, dep_rel_filter = dep_rel_filter, dep_rel_custom = dep_rel_custom)
  } else {
    cooc <- coocMatrix(x, term = term, group = group, n = n, pos = FALSE)
  }
  if (is.na(cooc)[1]) {
    obj <- list(nodes = NA, edges = NA)
    return(obj)
  }

  nodes <- cooc_freq(cooc) %>%
    mutate(
      id = row_number(),
      shape = shape,
      color = "navyblue"
    ) %>%
    rename(
      label = term,
      value = n
    )

  nodes$font.size <- log(nodes$value)
  Min <- min(nodes$font.size)
  Max <- max(nodes$font.size)
  if (Max > Min) {
    size <- (nodes$font.size - Min) / (Max - Min) * 15 * labelsize + 10
  } else {
    size <- 10 * labelsize
  }
  size[size < scalemin] <- scalemin
  size[size > scalemax] <- scalemax
  nodes$font.size <- size

  if (shape %in% c("dot", "square")) {
    nodes$font.vadjust <- -0.7 * nodes$font.size
  } else {
    nodes$font.vadjust <- 0
  }

  Normalize <- function(y) {
    if ((max(y) - min(y)) > 0) {
      y <- (y - min(y)) / (max(y) - min(y))
    } else {
      y <- rep(0.2, length(y))
    }
    return(y)
  }

  ### EDGES
  edges <- cooc %>%
    left_join(
      nodes %>% select(id, label),
      by = c("term1" = "label")
    ) %>%
    rename(from = id) %>%
    left_join(
      nodes %>% select(id, label),
      by = c("term2" = "label")
    ) %>%
    rename(
      to = id,
      s = cooc
    ) %>%
    mutate(
      sA = s / (s_from * s_to),
      sC = s / (sqrt(s_from * s_to)),
      sJ = s / (s_from + s_to - s)
    )

  edges$sNorm <- Normalize(edges$s) * 14 + 1
  edges$sANorm <- Normalize(edges$sA) * 14 + 1
  edges$sCNorm <- Normalize(edges$sC) * 14 + 1
  edges$sJNorm <- Normalize(edges$sJ) * 14 + 1

  switch(
    normalization,
    none = {
      edges$value <- edges$sNorm
    },
    association = {
      edges$value <- edges$sANorm
    },
    cosine = {
      edges$value <- edges$sCNorm
    },
    jaccard = {
      edges$value <- edges$sJNorm
    }
  )

  if (minEdges == "Auto") {
    y <- quantile(edges$value, seq(1, 0, -0.01), na.rm = T)
    x <- 1:length(y)
    res <- strucchange::breakpoints(y ~ x)
    tailEdges <- y[res$breakpoints[1]]
  } else {
    minEdges <- as.numeric(gsub("%", "", minEdges))
    tailEdges <- quantile(edges$value, 1 - (minEdges / 100), na.rm = T)
  }

  edges <- edges %>%
    dplyr::filter(value >= tailEdges) %>%
    select(term1, term2, from, to, value, s, sA, sC, sJ) %>%
    rename(
      term_from = term1,
      term_to = term2
    )

  ### remove isolated
  if (isTRUE(remove.isolated)) {
    id_remove <- setdiff(nodes$id, unique(c(edges$from, edges$to)))
    if (length(id_remove) > 0) {
      nodes <- nodes %>%
        filter(!id %in% id_remove)
    }
  }

  ### COMMUNITY DETECTION WITH SEED FIXING STRATEGY
  graph <- igraph::graph_from_data_frame(
    edges %>% select(-term_from, -term_to),
    directed = FALSE
  )

  # STRATEGY 1: SEED FIXING FOR STOCHASTIC ALGORITHMS
  # For stochastic algorithms (louvain, leiden), run multiple times and keep best result
  if (cluster %in% c("louvain", "leiden")) {
    n_runs <- 10
    best_modularity <- -Inf
    best_result <- NULL

    for (i in 1:n_runs) {
      set.seed(seed + i - 1) # Different seed for each run

      if (cluster == "louvain") {
        result <- igraph::cluster_louvain(graph)
      } else if (cluster == "leiden") {
        result <- igraph::cluster_leiden(
          graph,
          objective_function = "modularity",
          n_iterations = 3,
          resolution_parameter = 0.75
        )
      }

      current_modularity <- igraph::modularity(graph, result$membership)

      if (current_modularity > best_modularity) {
        best_modularity <- current_modularity
        best_result <- result
      }
    }

    net_groups <- best_result
  } else {
    # For other algorithms, use standard logic with set.seed
    set.seed(seed)

    switch(
      cluster,
      walktrap = {
        net_groups <- igraph::cluster_walktrap(graph)
      },
      optimal = {
        net_groups <- igraph::cluster_optimal(graph)
      },
      fast_greedy = {
        net_groups <- igraph::cluster_fast_greedy(graph)
      },
      leading_eigen = {
        net_groups <- igraph::cluster_leading_eigen(graph)
      },
      spinglass = {
        net_groups <- igraph::cluster_spinglass(graph)
      },
      infomap = {
        net_groups <- igraph::cluster_infomap(graph)
      },
      edge_betweenness = {
        net_groups <- igraph::cluster_edge_betweenness(graph)
      },
      {
        # Default to louvain
        net_groups <- igraph::cluster_louvain(graph)
      }
    )
  }

  cluster_df <- data.frame(as.list(igraph::membership(net_groups)))
  cluster_df <- as.data.frame(t(cluster_df)) %>%
    mutate(id = as.numeric(gsub("X", "", rownames(.)))) %>%
    rename(group = "V1")

  # Create group column
  nodes <- left_join(nodes, cluster_df, by = "id") %>%
    drop_na(group)

  # STRATEGY 2: IMPROVED COMMUNITY REPULSION
  if (community.repulsion > 0) {
    # Extract community structure information
    membership <- nodes$group
    names(membership) <- nodes$label
    n_communities <- length(unique(membership))
    n_nodes <- nrow(nodes)

    # Calculate statistics for adaptive normalization
    community_sizes <- table(membership)
    avg_community_size <- mean(community_sizes)

    # Calculate adaptive repulsion strength
    repulsion_strength <- adaptive_repulsion_strength(
      community.repulsion,
      n_nodes,
      n_communities,
      avg_community_size
    )

    # Get edge matrix
    row <- edges %>%
      select(term_from, term_to) %>%
      as.matrix()

    # Save original values
    original_values <- edges$value

    # Apply new weighting scheme with gradual growth
    new_values <- numeric(nrow(row))

    for (i in 1:nrow(row)) {
      node1 <- row[i, 1]
      node2 <- row[i, 2]

      comm1 <- membership[which(names(membership) == node1)]
      comm2 <- membership[which(names(membership) == node2)]

      if (comm1 == comm2) {
        # INTRA-COMMUNITY Edge
        # Moderate increase with sub-linear growth
        multiplier <- 1 + (repulsion_strength^0.7) * 1.5
        new_values[i] <- original_values[i] * multiplier
      } else {
        # INTER-COMMUNITY Edge
        # Gradual reduction with attenuated exponential function
        divisor <- 1 + exp(repulsion_strength * 1.2) - 1
        new_values[i] <- original_values[i] / divisor
      }
    }

    # Apply new values
    edges$value <- new_values
  }

  ## opacity for label
  opacity_font <- sqrt(
    (nodes$font.size - min(nodes$font.size)) / diff(range(nodes$font.size))
  ) *
    opacity +
    opacity.min +
    0.1
  if (is.nan(opacity_font[1])) {
    opacity_font <- rep(opacity.min, length(opacity_font))
  }

  if (labelsize > 0) {
    nodes$font.color <- unlist(lapply(opacity_font, function(x) {
      adjustcolor("black", alpha.f = x)
    }))
  } else {
    nodes$font.color <- adjustcolor("black", alpha.f = 0)
  }

  # node colors
  nodes$opacity.nodes <- (opacity_font - min(opacity_font)) /
    (diff(range(opacity_font))) *
    0.5 +
    opacity.min
  nodes$opacity.nodes[is.nan(nodes$opacity.nodes)] <- 0.5
  nodes$color <- paste0(
    colorlist[nodes$group],
    round(nodes$opacity.nodes, 2) * 100
  )

  if (interLinks) {
    interColor <- "#69696920"
  } else {
    interColor <- "#69696900"
  }
  edges <- edges %>%
    left_join(nodes %>% select(id, group, color), by = c("from" = "id")) %>%
    rename(group_from = group) %>%
    left_join(nodes %>% select(id, group), by = c("to" = "id")) %>%
    rename(group_to = group) %>%
    mutate(
      color = ifelse(
        group_from == group_to,
        paste0(substr(color, 1, 7), "20"),
        interColor
      )
    )

  obj <- list(nodes = nodes, edges = edges)
  return(obj)
}

# Helper function for adaptive community repulsion
adaptive_repulsion_strength <- function(
  community.repulsion,
  n_nodes,
  n_communities,
  avg_community_size
) {
  # Scale factor based on network size
  scale_factor <- log10(n_nodes + 10) / log10(100)

  # Correction factor based on the number of communities
  community_factor <- 1 + (n_communities - 2) * 0.1
  community_factor <- max(0.5, min(community_factor, 2))

  # Sigmoidal transformation of the user parameter
  x <- community.repulsion * 10
  sigmoid_transform <- x / (1 + x)

  # Combine the factors
  strength <- sigmoid_transform * scale_factor * community_factor

  return(strength)
}

net2vis <- function(nodes, edges, click = TRUE, noOverlap = FALSE) {
  if ((is.na(nodes))[1]) {
    VIS <- visNetwork::visNetwork(
      nodes = data.frame(
        id = "Empty Network",
        label = "No Connections Among Nodes",
        size = 0,
        title = "No Connections Among Nodes",
        font.size = 20
      ),
      type = "full",
      smooth = TRUE,
      physics = FALSE,
      x = 1,
      y = 1
    ) %>%
      visNetwork::visOptions(
        highlightNearest = list(enabled = T, hover = T, degree = 1),
        nodesIdSelection = F
      ) %>%
      visNetwork::visInteraction(
        dragNodes = TRUE,
        navigationButtons = F,
        hideEdgesOnDrag = TRUE,
        zoomSpeed = 0.2
      )
    return(VIS)
  }

  VIS <-
    visNetwork::visNetwork(
      nodes = nodes,
      edges = edges,
      type = "full",
      smooth = TRUE,
      physics = FALSE
    ) %>%
    visNetwork::visNodes(
      shadow = TRUE,
      shape = nodes$shape,
      font = list(
        color = nodes$font.color,
        size = nodes$font.size,
        vadjust = nodes$font.vadjust
      )
    ) %>%
    visNetwork::visIgraphLayout(
      layout = "layout_with_fr",
      weights = edges$value,
      type = "full"
    )

  # avoid overlaps among node labels
  ## avoid label overlaps
  if (noOverlap) {
    coords <- VIS$x$nodes %>%
      select(x, y)

    threshold <- 0.03
    ymax <- diff(range(coords[, 2]))
    xmax <- diff(range(coords[, 1]))
    threshold2 <- threshold * mean(xmax, ymax)
    w <- data.frame(
      x = coords[, 1],
      y = coords[, 2],
      labelToPlot = VIS$x$nodes$label,
      dotSize = VIS$x$nodes$font.size,
      row.names = VIS$x$nodes$label
    )
    labelToRemove <- avoidOverlaps(w, threshold = threshold2, y_scale = 3)
  } else {
    labelToRemove <- ""
  }

  VIS$x$nodes <- VIS$x$nodes %>%
    mutate(
      title = label,
      label = ifelse(label %in% labelToRemove, "", label)
    )

  VIS <- VIS %>%
    visNetwork::visEdges(smooth = list(type = "horizontal")) %>%
    visNetwork::visOptions(
      highlightNearest = list(enabled = T, hover = T, degree = 1),
      nodesIdSelection = T
    ) %>%
    visNetwork::visInteraction(
      dragNodes = TRUE,
      navigationButtons = F,
      hideEdgesOnDrag = TRUE,
      zoomSpeed = 0.2
    )

  if (click) {
    VIS <- VIS %>%
      visEvents(
        click = "function(nodes){
      Shiny.onInputChange('click', nodes.nodes[0]);
      ;}"
      )
  }

  VIS <- VIS %>%
    # visNetwork::visPhysics(barnesHut=list(avoidOverlap=1)) %>%
    visNetwork::visOptions(
      manipulation = FALSE,
      height = "100%",
      width = "100%"
    )
}

weight.community <- function(row, membership, weigth.within, weight.between) {
  if (
    as.numeric(membership[which(names(membership) == row[1])]) ==
      as.numeric(membership[which(names(membership) == row[2])])
  ) {
    weight <- weigth.within
  } else {
    weight <- weight.between
  }
  return(weight)
}

## function to avoid label overlapping ----
## THEMATIC MAP ----

#' Extract Noun Phrases via Dependency Parsing (C++ backend)
#'
#' @param sent_id Integer vector of sentence IDs
#' @param token_id Integer vector of token IDs
#' @param head_token_id Integer vector of head token IDs
#' @param dep_rel Character vector of dependency relations
#' @param upos Character vector of universal POS tags
#' @param terms Character vector of term values (lemma or token)
#' @param ngram_max Maximum phrase length (default 5)
#' @param max_gap Maximum gap between token positions (default 3)
#' @return A data.frame with columns keyword and ngram
#' @export
extract_noun_phrases <- function(sent_id, token_id, head_token_id, dep_rel, upos, terms, ngram_max = 5L, max_gap = 3L) {
  extract_np_cpp(sent_id, token_id, head_token_id, dep_rel, upos, terms, ngram_max, max_gap)
}

### DOCUMENT CLASSIFICATION MODULE ####
# Supervised classification with Random Forest
# Features: DTM or Word2Vec representations + Ranger RF

#### UI Function ####
docClassificationUI <- function() {
  tabItem(
    tabName = "doc_classification",
    fluidPage(
      # Page Header
      fluidRow(
        column(
          8,
          h2(icon("layer-group"), strong("Supervised Document Classification"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),

      # Main content
      fluidRow(
        # Configuration Panel
        column(
          8,
          box(
            title = strong(
              icon("gear"),
              " Configuration",
              style = "font-size: 18px;"
            ),
            width = 12,
            status = "primary",
            solidHeader = TRUE,
            collapsible = TRUE,
            collapsed = FALSE,
            id = "rf_config_box",

            # Target Variable Status
            div(
              style = "background-color: #e8f5e9; padding: 15px; margin-bottom: 20px; border-radius: 5px;",
              icon("tags", style = "color: #4CAF50;"),
              strong(" Target Variable: "),
              uiOutput("rf_target_var_display", inline = TRUE),
              br(),
              uiOutput("rf_target_var_info")
            ),

            # Feature Representation Selection
            div(
              style = "margin-bottom: 20px;",
              h4(
                icon("table-cells-large"),
                strong(" Feature Representation"),
                style = "color: #4F7942;"
              ),
              p(
                "Choose how documents will be represented as features for the classifier:",
                style = "color: #666; font-size: 14px;"
              ),
              radioGroupButtons(
                inputId = "rf_feature_type",
                label = NULL,
                choices = c(
                  "Document-Term Matrix (TF-IDF)" = "dtm",
                  "Word2Vec Document Embeddings" = "w2v"
                ),
                selected = "dtm",
                justified = TRUE,
                status = "primary",
                checkIcon = list(
                  yes = icon("check"),
                  no = icon("xmark")
                )
              )
            ),

            # DTM Options
            conditionalPanel(
              condition = "input.rf_feature_type == 'dtm'",
              div(
                style = "background-color: #f5f5f5; padding: 15px; border-radius: 5px; margin-bottom: 15px;",
                h5(strong("Document-Term Matrix Options")),
                fluidRow(
                  column(
                    4,
                    selectInput(
                      "rf_dtm_weighting",
                      "Weighting Scheme:",
                      choices = c(
                        "TF-IDF" = "tfidf",
                        "Term Frequency (TF)" = "tf",
                        "Binary" = "binary"
                      ),
                      selected = "tfidf"
                    )
                  ),
                  column(
                    4,
                    numericInput(
                      "rf_dtm_min_df",
                      "Min. Document Frequency:",
                      value = 2,
                      min = 1,
                      step = 1
                    ),
                    tags$small(
                      "Terms must appear in at least this many docs",
                      style = "color: #666;"
                    )
                  ),
                  column(
                    4,
                    sliderInput(
                      "rf_dtm_max_df",
                      "Max. Document Proportion:",
                      value = 0.95,
                      min = 0.5,
                      max = 1,
                      step = 0.05
                    ),
                    tags$small(
                      "Terms appearing in more docs are removed",
                      style = "color: #666;"
                    )
                  )
                ),
                fluidRow(
                  column(
                    6,
                    numericInput(
                      "rf_dtm_max_features",
                      "Max Features (optional):",
                      value = NA,
                      min = 10,
                      step = 10
                    ),
                    tags$small(
                      "Leave empty for all features",
                      style = "color: #666;"
                    )
                  )
                )
              )
            ),

            # Word2Vec Options
            conditionalPanel(
              condition = "input.rf_feature_type == 'w2v'",
              div(
                style = "background-color: #f5f5f5; padding: 15px; border-radius: 5px; margin-bottom: 15px;",
                h5(strong("Word2Vec Document Embeddings Options")),
                fluidRow(
                  column(
                    6,
                    radioButtons(
                      "rf_w2v_aggregation",
                      "Aggregation Method:",
                      choices = c(
                        "Simple Mean" = "mean",
                        "TF-IDF Weighted Mean" = "tfidf_weighted"
                      ),
                      selected = "mean"
                    ),
                    tags$small(
                      "How to combine word vectors into document vector",
                      style = "color: #666;"
                    )
                  ),
                  column(
                    6,
                    div(
                      style = "margin-top: 10px; padding: 10px; background-color: #fff; border: 1px solid #ddd; border-radius: 4px;",
                      h5(strong("Model Status")),
                      uiOutput("rf_w2v_model_status")
                    )
                  )
                )
              )
            ),

            hr(),

            # Random Forest Parameters
            div(
              h4(
                icon("tree"),
                strong(" Random Forest Parameters"),
                style = "color: #4F7942;"
              ),
              fluidRow(
                column(
                  4,
                  numericInput(
                    "rf_num_trees",
                    "Number of Trees:",
                    value = 500,
                    min = 100,
                    max = 2000,
                    step = 100
                  )
                ),
                column(
                  4,
                  sliderInput(
                    "rf_test_size",
                    "Test Set Proportion:",
                    value = 0.3,
                    min = 0.1,
                    max = 0.5,
                    step = 0.05
                  )
                ),
                column(
                  4,
                  div(
                    style = "margin-top: 25px;",
                    checkboxInput(
                      "rf_class_weights",
                      "Balance Class Weights",
                      value = TRUE
                    ),
                    tags$small(
                      "Recommended for imbalanced datasets",
                      style = "color: #666;"
                    )
                  )
                )
              )
            ),

            hr(),

            # Action Button
            div(
              style = "text-align: center; margin-top: 20px;",
              actionBttn(
                inputId = "rf_run",
                label = strong("Train Random Forest Model"),
                style = "pill",
                color = "success",
                size = "lg",
                icon = icon("play")
              )
            )
          )
        ),

        # Information Sidebar
        column(
          4,
          # Information Box
          box(
            title = strong(
              icon("circle-info"),
              " Information"
            ),
            width = 12,
            status = "info",
            solidHeader = TRUE,
            collapsible = TRUE,
            collapsed = FALSE,
            id = "rf_info_box",

            h5(strong("Prerequisites:")),
            tags$ul(
              tags$li(
                "Define a ",
                strong("Label Variable"),
                " in the Feature Roles menu"
              ),
              tags$li("At least 5 documents per class"),
              tags$li(
                "For Word2Vec: train embeddings in Words > Embeddings > Training"
              )
            ),

            hr(),

            h5(strong("Feature Representation:")),
            tags$ul(
              tags$li(
                strong("Document-Term Matrix:"),
                br(),
                "Traditional bag-of-words approach with TF-IDF weighting. Interpretable feature importance.",
                style = "margin-bottom: 10px;"
              ),
              tags$li(
                strong("Word2Vec Embeddings:"),
                br(),
                "Semantic document representations. Better generalization, lower dimensionality.",
                style = "margin-bottom: 10px;"
              )
            ),

            hr(),

            h5(strong("Random Forest:")),
            p(
              "Ensemble learning method using the ",
              tags$code("ranger"),
              " package for fast training.",
              style = "font-size: 13px;"
            )
          ),

          # Help Box
          box(
            title = strong(
              icon("circle-question"),
              " Need Help?"
            ),
            width = 12,
            status = "warning",
            collapsible = TRUE,
            collapsed = FALSE,
            id = "rf_help_box",
            p(
              "The Random Forest will predict the ",
              strong("Label Variable"),
              " defined in Feature Roles.",
              style = "font-size: 13px;"
            ),
            p(
              "Use ",
              strong("DTM"),
              " for interpretability and ",
              strong("Word2Vec"),
              " for better semantic understanding.",
              style = "font-size: 13px;"
            )
          )
        )
      ),

      # Results Panel (conditionally displayed)
      conditionalPanel(
        condition = "output.rf_trained",
        fluidRow(
          column(
            12,
            h3(
              icon("chart-bar"),
              strong(" Classification Results"),
              style = "color: #4F7942; margin-top: 30px; margin-bottom: 20px;"
            )
          )
        ),

        # Performance Metrics Row
        fluidRow(
          column(
            6,
            box(
              title = strong("Confusion Matrix - Test Set"),
              width = 12,
              status = "success",
              solidHeader = TRUE,
              shinycssloaders::withSpinner(
                plotOutput("rf_confusion_plot", height = "400px"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            )
          ),
          column(
            6,
            box(
              title = strong("Performance Metrics"),
              width = 12,
              status = "success",
              solidHeader = TRUE,
              div(
                style = "margin-bottom: 15px;",
                h4(
                  "Overall Accuracy: ",
                  textOutput("rf_accuracy_display", inline = TRUE),
                  style = "color: #4CAF50;"
                )
              ),
              shinycssloaders::withSpinner(
                DT::DTOutput("rf_metrics_table"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            )
          )
        ),

        # Feature Importance
        fluidRow(
          column(
            12,
            box(
              title = strong("Feature/Dimension Importance"),
              width = 12,
              status = "info",
              solidHeader = TRUE,
              shinycssloaders::withSpinner(
                plotlyOutput("rf_importance_plot", height = "600px"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            )
          )
        ),

        # Model Details
        fluidRow(
          column(
            6,
            box(
              title = strong("Training Details"),
              width = 12,
              status = "primary",
              verbatimTextOutput("rf_model_summary")
            )
          ),
          column(
            6,
            box(
              title = strong("Export Results"),
              width = 12,
              status = "primary",
              p(
                "Download classification results and model performance metrics."
              ),
              downloadButton(
                "rf_download_results",
                "Download Results (.xlsx)",
                class = "btn-info btn-block"
              ),
              hr(),
              actionBttn(
                inputId = "rf_apply_unlabeled",
                label = "Apply to Unlabeled Documents",
                style = "bordered",
                color = "primary",
                size = "md",
                block = TRUE,
                icon = icon("forward")
              )
            )
          )
        )
      )
    )
  )
}


#### SERVER FUNCTION ####
docClassificationServer <- function(input, output, session, values) {
  # Reactive values for classification results
  rf_results <- reactiveValues(
    model = NULL,
    train_data = NULL,
    test_data = NULL,
    train_pred = NULL,
    test_pred = NULL,
    feature_matrix = NULL,
    feature_type = NULL,
    term_mapping = NULL, # Mapping between original terms and sanitized names
    eval_train = NULL,
    eval_test = NULL,
    trained = FALSE
  )

  #### UI OUTPUTS ####

  # Display target variable
  output$rf_target_var_display <- renderUI({
    if (!is.null(values$labelVariable) && values$labelVariable != "") {
      tags$span(
        strong(values$labelVariable),
        style = "color: #4CAF50; font-size: 16px;"
      )
    } else {
      tags$span(
        "Not defined",
        style = "color: #d32f2f; font-size: 16px;"
      )
    }
  })

  # Target variable info
  output$rf_target_var_info <- renderUI({
    check <- checkClassificationPrereqs(values)

    if (!check$valid) {
      div(
        style = "margin-top: 10px; padding: 10px; background-color: #ffebee; border-left: 4px solid #f44336; border-radius: 4px;",
        icon("triangle-exclamation", style = "color: #d32f2f;"),
        strong(" Issues found:", style = "color: #d32f2f;"),
        tags$ul(
          lapply(check$errors, function(err) {
            tags$li(err, style = "color: #d32f2f;")
          })
        )
      )
    } else {
      # Show class distribution
      class_dist <- table(
        values$dfTag %>%
          filter(docSelected) %>%
          distinct(doc_id, .keep_all = TRUE) %>%
          pull(!!sym(values$labelVariable))
      )

      div(
        style = "margin-top: 10px;",
        tags$small(
          icon("check", style = "color: #4CAF50;"),
          strong(" Class distribution: "),
          paste(names(class_dist), "=", class_dist, collapse = " | "),
          style = "color: #666;"
        )
      )
    }
  })

  # Word2Vec model status
  output$rf_w2v_model_status <- renderUI({
    if (!is.null(values$w2v_model)) {
      div(
        icon("check", style = "color: #4CAF50;"),
        strong(" Model Available", style = "color: #4CAF50;"),
        br(),
        tags$small(
          sprintf("Dimensions: %d", ncol(as.matrix(values$w2v_model))),
          style = "color: #666;"
        )
      )
    } else {
      div(
        icon("triangle-exclamation", style = "color: #ff9800;"),
        strong(" Model Not Trained", style = "color: #ff9800;"),
        br(),
        tags$small(
          "Will be trained automatically with default parameters",
          style = "color: #666;"
        )
      )
    }
  })

  # Reactive flag for trained model
  output$rf_trained <- reactive({
    rf_results$trained
  })
  outputOptions(output, "rf_trained", suspendWhenHidden = FALSE)

  #### CLASSIFICATION WORKFLOW ####

  # Main classification event
  observeEvent(input$rf_run, {
    # Check prerequisites
    check <- checkClassificationPrereqs(values)
    if (!check$valid) {
      sendSweetAlert(
        session = session,
        title = "Cannot Train Model",
        text = HTML(paste(check$errors, collapse = "<br>")),
        html = TRUE,
        type = "error"
      )
      return(NULL)
    }

    # Show progress
    showNotification(
      "Training Random Forest model... This may take a moment.",
      type = "message",
      duration = NULL,
      id = "rf_training_notification"
    )

    tryCatch(
      {
        # Step 1: Create feature matrix
        if (input$rf_feature_type == "dtm") {
          # DTM approach
          dtm_result <- createDTM_forRF(
            dfTag = values$dfTag,
            term = values$generalTerm,
            weighting = input$rf_dtm_weighting,
            max_features = if (is.na(input$rf_dtm_max_features)) {
              NULL
            } else {
              input$rf_dtm_max_features
            },
            min_df = input$rf_dtm_min_df,
            max_df_prop = input$rf_dtm_max_df
          )

          feature_matrix <- dtm_result$matrix
          rf_results$term_mapping <- dtm_result$term_mapping
          rf_results$feature_type <- "DTM"

          # Inform user about DTM dimensions
          showNotification(
            sprintf(
              "DTM created: %d documents × %d terms",
              nrow(feature_matrix),
              ncol(feature_matrix)
            ),
            type = "message",
            duration = 3
          )
        } else {
          # Word2Vec approach
          # Check if model exists, if not train it
          if (is.null(values$w2v_model)) {
            showNotification(
              "Word2Vec model not found. Training with default parameters...",
              type = "warning",
              duration = 5
            )

            values$w2v_model <- trainWord2Vec(
              dfTag = values$dfTag,
              term = values$generalTerm,
              dim = 100,
              iter = 20
            )
          }

          # Aggregate word vectors to document vectors
          feature_matrix <- aggregateW2V_forRF(
            dfTag = values$dfTag,
            w2v_model = values$w2v_model,
            method = input$rf_w2v_aggregation,
            term = values$generalTerm
          )

          rf_results$term_mapping <- NULL # No term mapping for embeddings
          rf_results$feature_type <- "Word2Vec"
        }

        # Step 2: Get target variable at document level
        # CRITICAL: dfTag is tokenized, so we need to get document-level metadata correctly
        # We use group_by + summarize with first() to get the first non-NA value
        target_data <- values$dfTag %>%
          filter(docSelected) %>%
          group_by(doc_id) %>%
          summarise(
            target_var = first(na.omit(!!sym(values$labelVariable))),
            .groups = 'drop'
          )

        # Remove any documents where target is NA
        target_data <- target_data %>%
          filter(!is.na(target_var))

        # Merge features with target
        if (input$rf_feature_type == "dtm") {
          # For DTM, rownames are doc_ids
          feature_df <- as.data.frame(feature_matrix)
          feature_df$doc_id <- rownames(feature_matrix)

          # Ensure both doc_id columns are character type for reliable matching
          feature_df$doc_id <- as.character(feature_df$doc_id)
          target_data$doc_id <- as.character(target_data$doc_id)

          # Merge with target
          feature_df <- feature_df %>%
            inner_join(target_data, by = "doc_id")

          # Check if we have any rows after join
          if (nrow(feature_df) == 0) {
            stop(
              "No documents remain after merging features with target variable."
            )
          }

          # Check if target_var column exists
          if (!"target_var" %in% names(feature_df)) {
            stop("target_var column not found after merge!")
          }

          # Remove doc_id column
          feature_df <- feature_df[,
            names(feature_df) != "doc_id",
            drop = FALSE
          ]

          # Rename target_var to unique name that won't conflict with any term
          col_idx <- which(names(feature_df) == "target_var")
          if (length(col_idx) != 1) {
            stop(
              "Problem finding target_var column. Found ",
              length(col_idx),
              " matches."
            )
          }
          names(feature_df)[col_idx] <- ".target_class"
        } else {
          # For Word2Vec, feature_matrix already has doc_id column
          feature_df <- as.data.frame(feature_matrix)

          # Ensure doc_id is character
          feature_df$doc_id <- as.character(feature_df$doc_id)
          target_data$doc_id <- as.character(target_data$doc_id)

          feature_df <- feature_df %>%
            inner_join(target_data, by = "doc_id")

          # Check if we have any rows after join
          if (nrow(feature_df) == 0) {
            stop(
              "No documents remain after merging features with target variable."
            )
          }

          # Remove doc_id column - use base R
          feature_df <- feature_df[,
            names(feature_df) != "doc_id",
            drop = FALSE
          ]

          # Rename target_var to unique name
          col_idx <- which(names(feature_df) == "target_var")
          names(feature_df)[col_idx] <- ".target_class"
        }

        # Ensure target is factor AFTER merging
        # Use the unique column name
        feature_df$.target_class <- as.factor(feature_df$.target_class)

        # Calculate class weights if requested
        class_weights <- NULL
        if (input$rf_class_weights) {
          class_counts <- table(feature_df$.target_class)
          class_weights <- 1 / as.numeric(class_counts)
          names(class_weights) <- names(class_counts)
          class_weights <- class_weights / sum(class_weights)
        }

        # Step 3: Train Random Forest
        # Use random seed from settings
        rf_seed <- if (!is.null(values$random_seed)) values$random_seed else 123

        rf_model <- trainRangerRF(
          feature_matrix = feature_df,
          test_size = input$rf_test_size,
          num.trees = input$rf_num_trees,
          class.weights = class_weights,
          seed = rf_seed
        )

        # Step 4: Evaluate
        eval_train <- evaluateClassifier(
          true_labels = rf_model$train_data$.target_class,
          pred_probs = rf_model$train_pred$predictions
        )

        eval_test <- evaluateClassifier(
          true_labels = rf_model$test_data$.target_class,
          pred_probs = rf_model$test_pred$predictions
        )

        # Store results
        rf_results$model <- rf_model$model
        rf_results$train_data <- rf_model$train_data
        rf_results$test_data <- rf_model$test_data
        rf_results$train_pred <- rf_model$train_pred
        rf_results$test_pred <- rf_model$test_pred
        rf_results$feature_matrix <- feature_df
        rf_results$eval_train <- eval_train
        rf_results$eval_test <- eval_test
        rf_results$trained <- TRUE

        # Remove progress notification
        removeNotification(id = "rf_training_notification")

        # Collapse all configuration boxes to show results
        # Small delay to ensure DOM is ready
        shinyjs::delay(500, {
          shinyjs::runjs(
            "
          $('#rf_config_box').closest('.box').find('[data-widget=\"collapse\"]').click();
          $('#rf_info_box').closest('.box').find('[data-widget=\"collapse\"]').click();
          $('#rf_help_box').closest('.box').find('[data-widget=\"collapse\"]').click();
        "
          )
        })

        # Success message with proper formatting
        sendSweetAlert(
          session = session,
          title = "Model Trained Successfully!",
          text = paste0(
            "Test Set Accuracy: ",
            sprintf("%.2f%%", eval_test$accuracy * 100),
            "\n",
            "OOB Error: ",
            sprintf("%.2f%%", rf_model$model$prediction.error * 100)
          ),
          type = "success"
        )
      },
      error = function(e) {
        removeNotification(id = "rf_training_notification")
        sendSweetAlert(
          session = session,
          title = "Error Training Model",
          text = as.character(e$message),
          type = "error"
        )
      }
    )
  })

  #### RESULTS OUTPUTS ####

  # Accuracy display
  output$rf_accuracy_display <- renderText({
    req(rf_results$trained)
    sprintf("%.2f%%", rf_results$eval_test$accuracy * 100)
  })

  # Confusion matrix plot
  output$rf_confusion_plot <- renderPlot({
    req(rf_results$trained)
    plotConfusionMatrix(rf_results$eval_test$confusion_matrix)
  })

  # Metrics table
  output$rf_metrics_table <- DT::renderDT({
    req(rf_results$trained)

    metrics_df <- rf_results$eval_test$metrics_per_class %>%
      mutate(
        precision = round(precision, 3),
        recall = round(recall, 3),
        f1 = round(f1, 3)
      ) %>%
      rename(
        Class = class,
        Precision = precision,
        Recall = recall,
        `F1-Score` = f1
      )

    DT::datatable(
      metrics_df,
      options = list(
        pageLength = 10,
        dom = 't',
        ordering = FALSE
      ),
      rownames = FALSE
    )
  })

  # Feature importance plot
  output$rf_importance_plot <- renderPlotly({
    req(rf_results$trained)
    plotFeatureImportance(
      rf_results$model,
      feature_type = rf_results$feature_type,
      top_n = 30,
      term_mapping = rf_results$term_mapping
    )
  })

  # Model summary
  output$rf_model_summary <- renderPrint({
    req(rf_results$trained)

    cat("Random Forest Model Summary\n")
    cat("============================\n\n")
    cat(sprintf("Feature Type: %s\n", rf_results$feature_type))
    cat(sprintf("Number of Trees: %d\n", rf_results$model$num.trees))
    cat(sprintf("Number of Features: %d\n", ncol(rf_results$train_data) - 1))
    cat(sprintf("Training Set Size: %d\n", nrow(rf_results$train_data)))
    cat(sprintf("Test Set Size: %d\n", nrow(rf_results$test_data)))
    cat(sprintf(
      "\nOOB Prediction Error: %.2f%%\n",
      rf_results$model$prediction.error * 100
    ))
    cat(sprintf(
      "Train Set Accuracy: %.2f%%\n",
      rf_results$eval_train$accuracy * 100
    ))
    cat(sprintf(
      "Test Set Accuracy: %.2f%%\n",
      rf_results$eval_test$accuracy * 100
    ))
  })

  # Download results
  output$rf_download_results <- downloadHandler(
    filename = function() {
      paste0("RF_Classification_Results_", Sys.Date(), ".xlsx")
    },
    content = function(file) {
      exportClassificationResults(
        rf_results = rf_results,
        filename = file,
        label_var = values$labelVariable
      )
    }
  )

  # Apply to unlabeled documents
  observeEvent(input$rf_apply_unlabeled, {
    req(rf_results$trained)

    # TODO: Implement prediction on unlabeled documents
    sendSweetAlert(
      session = session,
      title = "Feature Coming Soon",
      text = "Prediction on unlabeled documents will be implemented in the next version.",
      type = "info"
    )
  })
}


#### HELPER FUNCTIONS ####

#' Check Classification Prerequisites
#'
#' Validates that all requirements are met before training
#'
#' @param values Shiny reactive values object
#' @return List with 'valid' (logical) and 'errors' (character vector)
checkClassificationPrereqs <- function(values) {
  errors <- c()

  # Check if dfTag exists
  if (is.null(values$dfTag)) {
    errors <- c(
      errors,
      "No data loaded. Please load and preprocess your corpus first."
    )
    return(list(valid = FALSE, errors = errors))
  }

  # Check label variable
  if (is.null(values$labelVariable) || values$labelVariable == "") {
    errors <- c(
      errors,
      "Label variable not defined. Please set it in the Feature Roles menu."
    )
  }

  # Check if label variable exists in dfTag
  if (
    !is.null(values$labelVariable) &&
      values$labelVariable != "" &&
      !values$labelVariable %in% names(values$dfTag)
  ) {
    errors <- c(
      errors,
      paste("Label variable '", values$labelVariable, "' not found in data.")
    )
  }

  # Check for selected documents
  if (
    !"docSelected" %in% names(values$dfTag) ||
      sum(values$dfTag$docSelected) == 0
  ) {
    errors <- c(
      errors,
      "No documents selected. Please select documents in the Documents menu."
    )
  }

  # Check if enough documents per class (only if label variable is valid)
  if (length(errors) == 0) {
    doc_labels <- values$dfTag %>%
      filter(docSelected) %>%
      distinct(doc_id, .keep_all = TRUE) %>%
      pull(!!sym(values$labelVariable))

    class_counts <- table(doc_labels)

    if (any(class_counts < 5)) {
      min_class <- names(class_counts)[which.min(class_counts)]
      errors <- c(
        errors,
        sprintf(
          "Class '%s' has only %d documents. Each class needs at least 5 documents for reliable training.",
          min_class,
          min(class_counts)
        )
      )
    }

    # Check for at least 2 classes
    if (length(class_counts) < 2) {
      errors <- c(errors, "At least 2 classes are required for classification.")
    }
  }

  return(list(valid = length(errors) == 0, errors = errors))
}


#' Create Document-Term Matrix for Random Forest
#'
#' Creates a DTM with various weighting schemes using udpipe
#'
#' @param dfTag Data frame with tokenized and tagged text
#' @param term Character, either "lemma" or "token"
#' @param weighting Character, one of "tf", "tfidf", "binary"
#' @param max_features Integer, maximum number of features to keep (NULL = all)
#' @param min_df Integer, minimum document frequency
#' @param max_df_prop Numeric, maximum proportion of documents a term can appear in
#' @return List with matrix and term_mapping
createDTM_forRF <- function(
  dfTag,
  term = "lemma",
  weighting = "tfidf",
  max_features = NULL,
  min_df = 2,
  max_df_prop = 0.95
) {
  # Filter selected documents and POS
  df_filtered <- dfTag %>%
    filter(docSelected, POSSelected)

  # Get unique doc_ids for later reference
  original_doc_ids <- unique(df_filtered$doc_id)

  # Select term
  if (term == "lemma") {
    df_filtered <- df_filtered %>%
      mutate(term_text = lemma)
  } else {
    df_filtered <- df_filtered %>%
      mutate(term_text = token)
  }

  # Use udpipe's document_term_frequencies
  dtf <- document_term_frequencies(
    x = df_filtered,
    document = "doc_id",
    term = "term_text"
  )

  # Create DTM using udpipe's document_term_matrix
  dtm <- document_term_matrix(dtf)

  # CRITICAL: Extract and preserve document IDs
  doc_ids <- rownames(dtm)

  # Convert to regular matrix
  dtm_matrix <- as.matrix(dtm)

  # Ensure rownames are preserved after conversion
  rownames(dtm_matrix) <- doc_ids

  # Store original number of documents
  n_docs_initial <- nrow(dtm_matrix)

  # Get document frequency (number of documents containing each term)
  doc_freq <- colSums(dtm_matrix > 0)
  n_docs <- nrow(dtm_matrix)

  # Apply document frequency filtering
  # Remove too rare terms
  keep_terms <- doc_freq >= min_df

  # Remove too frequent terms
  keep_terms <- keep_terms & (doc_freq / n_docs <= max_df_prop)

  # Filter matrix
  dtm_matrix <- dtm_matrix[, keep_terms, drop = FALSE]

  # CRITICAL: Remove documents that have no terms after filtering
  # This happens when all terms in a document were filtered out
  doc_term_counts <- rowSums(dtm_matrix)
  keep_docs <- doc_term_counts > 0

  n_docs_removed <- sum(!keep_docs)

  if (sum(keep_docs) == 0) {
    stop(
      "All documents were filtered out. Please reduce min_df or increase max_df_prop."
    )
  }

  dtm_matrix <- dtm_matrix[keep_docs, , drop = FALSE]

  # Update doc_freq after filtering documents
  doc_freq <- colSums(dtm_matrix > 0)
  n_docs <- nrow(dtm_matrix)

  # Apply weighting
  if (weighting == "tfidf") {
    # Calculate IDF
    idf <- log(n_docs / doc_freq)

    # Apply TF-IDF
    dtm_matrix <- sweep(dtm_matrix, 2, idf, "*")
  } else if (weighting == "binary") {
    dtm_matrix <- (dtm_matrix > 0) * 1
  }
  # else keep raw TF

  # Optional: keep only top features by variance
  if (!is.null(max_features) && ncol(dtm_matrix) > max_features) {
    term_variance <- apply(dtm_matrix, 2, var)
    top_terms <- order(term_variance, decreasing = TRUE)[1:max_features]
    dtm_matrix <- dtm_matrix[, top_terms, drop = FALSE]
  }

  # CRITICAL: Create mapping between original terms and sanitized column names
  # This preserves original terms for visualization
  original_terms <- colnames(dtm_matrix)
  sanitized_names <- make.names(original_terms, unique = TRUE)

  # Create term mapping data frame
  term_mapping <- data.frame(
    original_term = original_terms,
    sanitized_name = sanitized_names,
    stringsAsFactors = FALSE
  )

  # Apply sanitized names to matrix
  colnames(dtm_matrix) <- sanitized_names

  # Return both matrix and mapping
  return(list(
    matrix = dtm_matrix,
    term_mapping = term_mapping
  ))
}


#' Aggregate Word2Vec Embeddings to Document Level
#'
#' Creates document embeddings by aggregating word vectors
#'
#' @param dfTag Data frame with tokenized and tagged text
#' @param w2v_model word2vec model object
#' @param method Character, "mean" or "tfidf_weighted"
#' @param term Character, either "lemma" or "token"
#' @return Data frame with doc_id and embedding dimensions
aggregateW2V_forRF <- function(
  dfTag,
  w2v_model,
  method = "mean",
  term = "lemma"
) {
  # Get embedding matrix
  w2v_matrix <- as.matrix(w2v_model)

  # Filter and prepare data
  df_filtered <- dfTag %>%
    filter(docSelected, POSSelected) %>%
    mutate(term_text = tolower(!!sym(term)))

  # Calculate TF-IDF weights if needed
  if (method == "tfidf_weighted") {
    n_docs <- length(unique(df_filtered$doc_id))

    tfidf_weights <- df_filtered %>%
      group_by(doc_id, term_text) %>%
      summarise(tf = n(), .groups = 'drop') %>%
      group_by(term_text) %>%
      mutate(
        df = n(),
        idf = log(n_docs / df)
      ) %>%
      ungroup() %>%
      mutate(tfidf = tf * idf)
  }

  # Aggregate by document
  doc_embeddings <- df_filtered %>%
    group_by(doc_id) %>%
    group_modify(
      ~ {
        doc_terms <- .x$term_text

        # Filter terms present in w2v model
        valid_terms <- intersect(doc_terms, rownames(w2v_matrix))

        if (length(valid_terms) == 0) {
          # Return zero vector if no valid terms
          zero_vec <- rep(0, ncol(w2v_matrix))
          return(tibble(
            !!!setNames(as.list(zero_vec), paste0("dim_", 1:ncol(w2v_matrix)))
          ))
        }

        if (method == "mean") {
          # Simple mean
          doc_vec <- colMeans(w2v_matrix[valid_terms, , drop = FALSE])
        } else {
          # TF-IDF weighted mean
          doc_id_val <- .y$doc_id

          weights <- tfidf_weights %>%
            filter(doc_id == doc_id_val, term_text %in% valid_terms) %>%
            arrange(match(term_text, valid_terms)) %>%
            pull(tfidf)

          if (length(weights) == 0 || sum(weights) == 0) {
            doc_vec <- colMeans(w2v_matrix[valid_terms, , drop = FALSE])
          } else {
            # Align weights with terms
            term_weight_map <- tfidf_weights %>%
              filter(doc_id == doc_id_val, term_text %in% valid_terms) %>%
              select(term_text, tfidf)

            weights_aligned <- sapply(valid_terms, function(t) {
              w <- term_weight_map$tfidf[term_weight_map$term_text == t]
              if (length(w) == 0) 1 else w
            })

            weighted_vecs <- sweep(
              w2v_matrix[valid_terms, , drop = FALSE],
              1,
              weights_aligned,
              "*"
            )
            doc_vec <- colSums(weighted_vecs) / sum(weights_aligned)
          }
        }

        tibble(!!!setNames(as.list(doc_vec), paste0("dim_", 1:length(doc_vec))))
      }
    ) %>%
    ungroup()

  # CRITICAL: Sanitize column names to be valid R variable names
  # Ensure consistency with DTM approach
  colnames(doc_embeddings) <- make.names(
    colnames(doc_embeddings),
    unique = TRUE
  )

  return(doc_embeddings)
}


#' Train Word2Vec Model
#'
#' Wrapper function to train word2vec on the corpus
#'
#' @param dfTag Data frame with tokenized text
#' @param term Character, "lemma" or "token"
#' @param dim Integer, embedding dimension
#' @param iter Integer, number of iterations
#' @return word2vec model object
trainWord2Vec <- function(dfTag, term = "lemma", dim = 100, iter = 20) {
  # Prepare text data
  df_filtered <- dfTag %>%
    filter(docSelected, POSSelected) %>%
    select(sentence_id, !!sym(term)) %>%
    mutate(term_lower = tolower(!!sym(term)))

  # Create list of sentences
  sentence_list <- df_filtered %>%
    group_by(sentence_id) %>%
    summarise(sentence = list(term_lower), .groups = 'drop') %>%
    pull(sentence)

  # Train model
  w2v_model <- word2vec(
    x = sentence_list,
    type = "cbow",
    dim = dim,
    iter = iter,
    window = 5,
    min_count = 2,
    threads = parallel::detectCores() - 1
  )

  return(w2v_model)
}


#' Create Stratified Train/Test Split
#'
#' Creates stratified indices for train/test split
#'
#' @param y Factor or character vector of class labels
#' @param p Numeric, proportion for training set
#' @param seed Integer, random seed
#' @return Integer vector of training indices
createStratifiedSplit <- function(y, p = 0.7, seed = 123) {
  set.seed(seed)

  # Get unique classes
  classes <- unique(y)
  train_idx <- c()

  # For each class, sample proportionally
  for (cls in classes) {
    cls_idx <- which(y == cls)
    n_train <- round(length(cls_idx) * p)

    # Sample indices for this class
    cls_train_idx <- sample(cls_idx, size = n_train, replace = FALSE)
    train_idx <- c(train_idx, cls_train_idx)
  }

  return(sort(train_idx))
}


#' Train Random Forest with Ranger
#'
#' Trains a random forest classifier using ranger package
#'
#' @param feature_matrix Data frame with features and 'target' column
#' @param test_size Numeric, proportion of data for test set
#' @param num.trees Integer, number of trees
#' @param mtry Integer, number of variables to try at each split (NULL = auto)
#' @param min.node.size Integer, minimum node size (NULL = auto)
#' @param class.weights Named vector of class weights (NULL = equal weights)
#' @param seed Integer, random seed
#' @return List with model, train_data, test_data, train_pred, test_pred
trainRangerRF <- function(
  feature_matrix,
  test_size = 0.3,
  num.trees = 500,
  mtry = NULL,
  min.node.size = NULL,
  class.weights = NULL,
  seed = 123
) {
  set.seed(seed)

  # Target column is named .target_class to avoid conflicts with term columns
  if (!".target_class" %in% names(feature_matrix)) {
    stop("Target column '.target_class' not found in feature matrix")
  }

  # Target should already be a factor from the calling function
  if (!is.factor(feature_matrix$.target_class)) {
    feature_matrix$.target_class <- as.factor(feature_matrix$.target_class)
  }

  # Stratified split using custom function
  train_idx <- createStratifiedSplit(
    y = feature_matrix$.target_class,
    p = 1 - test_size,
    seed = seed
  )

  train_data <- feature_matrix[train_idx, ]
  test_data <- feature_matrix[-train_idx, ]

  # Auto-calculate mtry if not provided
  if (is.null(mtry)) {
    mtry <- floor(sqrt(ncol(train_data) - 1))
  }

  # Train model
  model <- ranger::ranger(
    formula = .target_class ~ .,
    data = train_data,
    num.trees = num.trees,
    mtry = mtry,
    min.node.size = min.node.size,
    importance = 'impurity',
    probability = TRUE,
    class.weights = class.weights,
    num.threads = parallel::detectCores() - 1,
    seed = seed
  )

  # Predictions
  train_pred <- predict(model, data = train_data)
  test_pred <- predict(model, data = test_data)

  return(list(
    model = model,
    train_data = train_data,
    test_data = test_data,
    train_pred = train_pred,
    test_pred = test_pred
  ))
}


#' Evaluate Classification Performance
#'
#' Computes confusion matrix and per-class metrics
#'
#' @param true_labels Factor, true class labels
#' @param pred_probs Matrix, predicted probabilities (rows = samples, cols = classes)
#' @return List with confusion_matrix, accuracy, metrics_per_class
evaluateClassifier <- function(true_labels, pred_probs) {
  # Get predicted classes (highest probability)
  pred_classes <- factor(
    colnames(pred_probs)[apply(pred_probs, 1, which.max)],
    levels = levels(true_labels)
  )

  # Confusion matrix
  cm <- table(Predicted = pred_classes, Actual = true_labels)

  # Overall accuracy
  accuracy <- sum(diag(cm)) / sum(cm)

  # Per-class metrics
  classes <- levels(true_labels)
  metrics_list <- lapply(classes, function(cls) {
    tp <- cm[cls, cls]
    fp <- sum(cm[cls, ]) - tp
    fn <- sum(cm[, cls]) - tp
    tn <- sum(cm) - tp - fp - fn

    precision <- ifelse(tp + fp > 0, tp / (tp + fp), 0)
    recall <- ifelse(tp + fn > 0, tp / (tp + fn), 0)
    f1 <- ifelse(
      precision + recall > 0,
      2 * precision * recall / (precision + recall),
      0
    )

    tibble(
      class = cls,
      precision = precision,
      recall = recall,
      f1 = f1
    )
  })

  metrics_df <- bind_rows(metrics_list)

  return(list(
    confusion_matrix = cm,
    accuracy = accuracy,
    metrics_per_class = metrics_df
  ))
}


#' Plot Confusion Matrix
#'
#' Creates a ggplot visualization of confusion matrix
#'
#' @param cm Confusion matrix (table object)
#' @return ggplot object
plotConfusionMatrix <- function(cm) {
  # Convert to data frame
  cm_df <- as.data.frame(cm)
  colnames(cm_df) <- c("Predicted", "Actual", "Count")

  # Calculate percentages by actual class
  cm_df <- cm_df %>%
    group_by(Actual) %>%
    mutate(
      Total = sum(Count),
      Percentage = Count / Total * 100
    ) %>%
    ungroup()

  # Create plot
  p <- ggplot(cm_df, aes(x = Actual, y = Predicted, fill = Count)) +
    geom_tile(color = "white", size = 1) +
    geom_text(
      aes(label = sprintf("%d\n(%.1f%%)", Count, Percentage)),
      color = "black",
      size = 4
    ) +
    scale_fill_gradient(
      low = "#fff7ec",
      high = "#7F0000",
      name = "Count"
    ) +
    labs(
      title = "Confusion Matrix",
      x = "Actual Class",
      y = "Predicted Class"
    ) +
    theme_minimal() +
    theme(
      plot.title = element_text(hjust = 0.5, face = "bold", size = 14),
      axis.text = element_text(size = 11),
      axis.title = element_text(size = 12, face = "bold"),
      panel.grid = element_blank()
    ) +
    coord_fixed()

  return(p)
}


#' Plot Feature Importance
#'
#' Creates an interactive plotly bar chart of feature importance
#'
#' @param model ranger model object
#' @param feature_type Character, "DTM" or "Word2Vec"
#' @param top_n Integer, number of top features to show
#' @param term_mapping Data frame with original_term and sanitized_name columns (NULL for Word2Vec)
#' @return plotly object
plotFeatureImportance <- function(
  model,
  feature_type = "DTM",
  top_n = 30,
  term_mapping = NULL
) {
  # Extract importance
  importance_values <- model$variable.importance

  # Sort and get top N
  importance_df <- tibble(
    feature = names(importance_values),
    importance = importance_values
  ) %>%
    arrange(desc(importance)) %>%
    head(top_n)

  # If DTM and term_mapping is available, restore original terms
  if (feature_type == "DTM" && !is.null(term_mapping)) {
    importance_df <- importance_df %>%
      left_join(term_mapping, by = c("feature" = "sanitized_name")) %>%
      mutate(
        display_name = ifelse(!is.na(original_term), original_term, feature)
      )
  } else {
    importance_df <- importance_df %>%
      mutate(display_name = feature)
  }

  # Create label based on feature type
  if (feature_type == "DTM") {
    x_label <- "Terms"
    title_text <- "Top Terms by Importance"
  } else {
    x_label <- "Embedding Dimensions"
    title_text <- "Embedding Dimensions by Importance"
  }

  # Create plot
  p <- plot_ly(
    data = importance_df,
    x = ~importance,
    y = ~ reorder(display_name, importance),
    type = "bar",
    orientation = 'h',
    marker = list(
      color = importance_df$importance,
      colorscale = 'Viridis',
      showscale = TRUE,
      colorbar = list(title = "Importance")
    ),
    hovertemplate = paste0(
      "<b>%{y}</b><br>",
      "Importance: %{x:.4f}<br>",
      "<extra></extra>"
    )
  ) %>%
    layout(
      title = list(text = title_text, x = 0.5, xanchor = "center"),
      xaxis = list(title = "Importance (Gini)"),
      yaxis = list(title = x_label),
      margin = list(l = 150),
      hovermode = "closest"
    )

  return(p)
}


#' Export Classification Results
#'
#' Exports results to Excel file with multiple sheets
#'
#' @param rf_results Reactive values with classification results
#' @param filename Character, output file path
#' @param label_var Character, name of label variable
exportClassificationResults <- function(rf_results, filename, label_var) {
  # Create workbook
  wb <- openxlsx::createWorkbook()

  # Sheet 1: Model Summary
  summary_data <- tibble(
    Metric = c(
      "Feature Type",
      "Number of Trees",
      "Number of Features",
      "Training Set Size",
      "Test Set Size",
      "OOB Error (%)",
      "Train Accuracy (%)",
      "Test Accuracy (%)"
    ),
    Value = c(
      rf_results$feature_type,
      rf_results$model$num.trees,
      ncol(rf_results$train_data) - 1,
      nrow(rf_results$train_data),
      nrow(rf_results$test_data),
      round(rf_results$model$prediction.error * 100, 2),
      round(rf_results$eval_train$accuracy * 100, 2),
      round(rf_results$eval_test$accuracy * 100, 2)
    )
  )

  openxlsx::addWorksheet(wb, "Model_Summary")
  openxlsx::writeData(wb, "Model_Summary", summary_data)

  # Sheet 2: Performance Metrics
  metrics_data <- rf_results$eval_test$metrics_per_class %>%
    mutate(
      precision = round(precision, 4),
      recall = round(recall, 4),
      f1 = round(f1, 4)
    )

  openxlsx::addWorksheet(wb, "Performance_Metrics")
  openxlsx::writeData(wb, "Performance_Metrics", metrics_data)

  # Sheet 3: Confusion Matrix
  cm_data <- as.data.frame.matrix(rf_results$eval_test$confusion_matrix)
  cm_data <- cbind(Predicted = rownames(cm_data), cm_data)

  openxlsx::addWorksheet(wb, "Confusion_Matrix")
  openxlsx::writeData(wb, "Confusion_Matrix", cm_data)

  # Sheet 4: Feature Importance
  importance_data <- tibble(
    Feature = names(rf_results$model$variable.importance),
    Importance = rf_results$model$variable.importance
  ) %>%
    arrange(desc(Importance))

  # If DTM with term mapping, add original terms
  if (!is.null(rf_results$term_mapping)) {
    importance_data <- importance_data %>%
      left_join(
        rf_results$term_mapping,
        by = c("Feature" = "sanitized_name")
      ) %>%
      mutate(
        Original_Term = ifelse(!is.na(original_term), original_term, Feature)
      ) %>%
      select(Original_Term, Feature, Importance)
  }

  openxlsx::addWorksheet(wb, "Feature_Importance")
  openxlsx::writeData(wb, "Feature_Importance", importance_data)

  # Sheet 5: Test Predictions
  test_pred_classes <- colnames(rf_results$test_pred$predictions)[
    apply(rf_results$test_pred$predictions, 1, which.max)
  ]

  predictions_data <- tibble(
    Actual = as.character(rf_results$test_data$target),
    Predicted = test_pred_classes,
    Correct = Actual == Predicted
  )

  # Add probabilities
  prob_df <- as.data.frame(rf_results$test_pred$predictions)
  colnames(prob_df) <- paste0("Prob_", colnames(prob_df))
  predictions_data <- cbind(predictions_data, prob_df)

  openxlsx::addWorksheet(wb, "Test_Predictions")
  openxlsx::writeData(wb, "Test_Predictions", predictions_data)

  # Save workbook
  openxlsx::saveWorkbook(wb, filename, overwrite = TRUE)
}

preprocessingUI <- function() {
  ## Tokenization & PoS Tagging -----

  tokpos <- tabItem(
    tabName = "tokPos",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("spell-check"), strong("Tokenization & PoS Tagging"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      )
    ),
    tabsetPanel(
      type = "tabs",
      tabPanel(
        title = "Annotated Text Table", icon = icon("table"),
        br(),
        fluidRow(
          column(
            8,
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("tokPosTagData"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          column(
            4,
            div(
              box(
                width = 12,
                title = tags$div(
                  style = "display: flex; align-items: center;", # Allineamento orizzontale
                  tags$h4(
                    "Language Model   ",
                    style = "margin-top:-20px; margin-right: 40px; font-weight: bold"
                  ), # Testo con margine a destra
                  uiOutput("flagUI") # Immagine SVG
                ),
                br(),
                div(
                  fluidRow(
                    column(
                      6,
                      div(
                        uiOutput("optionsTokenization")
                        # selectInput(
                        #   inputId = 'language_model', label="Language", choices = label_lang,
                        #   multiple=FALSE,
                        #   width = "100%"
                        # )
                      ),
                      style = "margin-top:-9px"
                    ),
                    column(
                      6,
                      div(
                        uiOutput("treebankSelect")
                        # selectInput("treebank", "Treebank", choices = "GUM")
                      ),
                      style = "margin-top:-9px"
                    )
                  ),
                  fluidRow(
                    column(
                      6,
                      div(
                        align = "center",
                        style = "margin-top:-15px",
                        width = 12,
                        title = t_run,
                        do.call(
                          "actionButton",
                          c(
                            run_bttn,
                            list(
                              inputId = "tokPosRun"
                            )
                          )
                        )
                      )
                    ),
                    column(
                      6,
                      div(
                        align = "center",
                        style = "margin-top:-15px",
                        width = 12,
                        title = t_save,
                        do.call(
                          "actionButton",
                          c(
                            list(
                              label = NULL,
                              style = "display:block; height: 37px; width: 37px; border-radius: 50%;
                                      border: 1px; margin-top: 15px;",
                              icon = icon("floppy-disk")
                            ),
                            list(
                              inputId = "tokPosSave"
                            )
                          )
                        )
                      )
                    )
                  ),
                  style = "margin-top:-5px"
                ),
                tags$hr(),
                uiOutput("info_treebank"),
                tags$hr(),
                div(
                  style = "margin-top: 12px; margin-bottom: 12px; padding: 15px; border-radius: 6px;",

                  # Token Normalization section (Teal theme)
                  div(
                    style = "display: flex; align-items: center; gap: 15px; padding: 12px; background: linear-gradient(135deg, #e0f7f4 0%, #e8f6f3 100%); border-radius: 6px; border-left: 4px solid #1ABC9C; margin-bottom: 10px;",
                    div(
                      style = "display: flex; align-items: center; flex-grow: 1;",
                      icon(
                        "sort-alpha-down",
                        style = "color: #1ABC9C; margin-right: 10px; font-size: 22px;"
                      ),
                      span(
                        "Token Normalization (Lowercase)",
                        style = "font-weight: bold; color: #16A085; font-size: 15px;"
                      )
                    ),
                    div(
                      style = "flex-shrink: 0;",
                      materialSwitch(
                        inputId = "token_lowercase",
                        label = NULL,
                        value = TRUE,
                        status = "success"
                      )
                    )
                  ),

                  # Lemma Normalization section (Blue theme)
                  div(
                    style = "display: flex; align-items: center; gap: 15px; padding: 12px; background: linear-gradient(135deg, #e3f2fd 0%, #e8eaf6 100%); border-radius: 6px; border-left: 4px solid #3498DB;",
                    div(
                      style = "display: flex; align-items: center; flex-grow: 1;",
                      icon(
                        "exchange-alt",
                        style = "color: #3498DB; margin-right: 10px; font-size: 22px;"
                      ),
                      span(
                        "Lemma Normalization (Lowercase)",
                        style = "font-weight: bold; color: #2980B9; font-size: 15px;"
                      )
                    ),
                    div(
                      style = "flex-shrink: 0;",
                      materialSwitch(
                        inputId = "lemma_lowercase",
                        label = NULL,
                        value = TRUE,
                        status = "primary"
                      )
                    )
                  )
                ),
                uiOutput("unitAnalysis")
              )
            ),
            style = "margin-top:40px"
          )
        )
      ),
      tabPanel(
        title = "Info & References", icon = icon("circle-info"),
        br(),
        fluidPage(
          fluidRow(
            column(1),
            column(
              10,
              br(),
              HTML(infoTexts$tokenization)
            ),
            column(1)
          )
        )
      )
    )
  )

  ### POS Special Tagging ----
  specialentities <- tabItem(
    tabName = "posSpecial",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("at"), strong("Tagging Special Entities"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      )
    ),
    br(),
    fluidRow(
      column(
        9,
        tabsetPanel(
          type = "tabs",
          # Main tab with HTML-based entity cards
          tabPanel(
            title = "Special Entities Overview", icon = icon("table"),
            br(),
            div(
              style = "padding: 20px;",
              # Custom HTML output for special entities cards
              shinycssloaders::withSpinner(
                uiOutput("posSpecialCardsUI"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            )
          ),
          # Annotated text table tab (unchanged)
          tabPanel(
            title = "Annotated Text Table", icon = icon("table"),
            br(),
            shinycssloaders::withSpinner(
              DT::DTOutput("posSpecialData"),
              color = getOption("spinner.color", default = "#4F7942")
            )
          ),
          # Info tab (unchanged)
          tabPanel(
            title = "Info & References", icon = icon("circle-info"),
            br(),
            fluidPage(
              fluidRow(
                column(1),
                column(
                  10,
                  br(),
                  HTML(infoTexts$specialentities)
                ),
                column(1)
              )
            )
          )
        )
      ),
      column(
        3,
        div(
          box(
            width = 12,
            div(
              h3(strong(em("Special Entities"))),
              style = "margin-top:-57px"
            ),
            tags$hr(),
            helpText(h5(
              "When processing text, special tags will be assigned to certain detected entities."
            )),
            helpText(h5("These include:")),
            helpText(h5("•    Email addresses: example@domain.com")),
            helpText(h5("•    URLs: https://www.example.com/path")),
            helpText(h5("•   Emojis: 😊, 🚀, ❤️")),
            helpText(h5("•    Hashtags: #ExampleTag")),
            helpText(h5("•    IP addresses: 192.168.1.1")),
            helpText(h5("•    Mentions: @username")),
            helpText(h5(
              "This ensures that these elements are identified and marked for further analysis within the text."
            )),
            style = "text-align: left; text-color: #989898",
            br(),
            div(
              hr(),
              div(
                fluidRow(
                  column(
                    4,
                    div(
                      align = "center",
                      style = "margin-top:-15px",
                      width = 12,
                      do.call(
                        "actionButton",
                        c(
                          run_bttn,
                          list(
                            inputId = "posSpecialRun"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    4,
                    div(
                      align = "center",
                      style = "margin-top:-15px",
                      width = 12,
                      title = t_back,
                      do.call(
                        "actionButton",
                        c(
                          back_bttn,
                          list(
                            inputId = "posSpecialBack"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    4,
                    div(
                      title = t_save,
                      div(
                        align = "center",
                        do.call(
                          "actionButton",
                          c(
                            save_bttn,
                            list(
                              inputId = "posSpecialSave"
                            )
                          )
                        )
                      )
                    )
                  )
                ),
                style = "margin-top: -8px"
              ),
              style = "margin-top:-15px"
            )
          )
        ),
        style = "margin-top:40px"
      )
    )
  )

  ## Custom PoS Lists -----

  custompos <- tabItem(
    tabName = "custTermList",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("list-check"), strong("Custom PoS List Loading and Merging"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      br(),
      br(),
      fluidRow(
        column(
          9,
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Pos Tagging with Custom List", icon = icon("table"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("customPosTagData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & References", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(infoTexts$customterm)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          3,
          div(
            box(
              width = 12,
              div(
                h3(strong(em("Import Custom PoS List"))),
                style = "margin-top:-57px"
              ),
              hr(),
              # selectInput("CTLterm",
              #             "Terms:",
              #             choices = c("Tokens" = "token",
              #                         "Lemma" = "lemma"),
              #             selected = "lemma"),
              helpText(h5(
                "Please ensure that the Custom PoS List is formatted as an Excel file with two columns.
                                       In the first column include the list of terms.
                                       In the second column provide any custom PoS tag (e.g., METHOD, TO_REMOVE, APPLICATION, SETTING, etc.) or standard PoS tags (NOUN, VERB, ADJ, etc.)."
              )),
              fileInput(
                "custom_lists",
                label = NULL,
                multiple = TRUE,
                accept = c(
                  ".csv",
                  ".xls",
                  ".xlsx"
                )
              ),
              hr(),
              # Download buttons section
              fluidRow(
                column(
                  6,
                  downloadButton(
                    "downloadCustomTermTemplate",
                    "Template",
                    style = "width: 100%; margin-bottom: 10px; background-color: #6CC283; border-color: #4F7942;color: #ffffff;"
                  )
                ),
                column(
                  6,
                  downloadButton(
                    "downloadCustomTermVocabulary",
                    "Vocabulary",
                    style = "width: 100%; margin-bottom: 10px; background-color: #6CC283; border-color: #4F7942;color: #ffffff;"
                  )
                )
              ),
              hr(),
              fluidRow(
                column(
                  4,
                  div(
                    align = "center",
                    style = "margin-top:-15px",
                    width = 12,
                    do.call(
                      "actionButton",
                      c(
                        run_bttn,
                        list(
                          inputId = "custTermListRun"
                        )
                      )
                    )
                  )
                ),
                column(
                  4,
                  div(
                    align = "center",
                    style = "margin-top:-15px",
                    width = 12,
                    title = t_back,
                    do.call(
                      "actionButton",
                      c(
                        back_bttn,
                        list(
                          inputId = "custTermListBack"
                        )
                      )
                    )
                  )
                ),
                column(
                  4,
                  div(
                    title = t_save,
                    div(
                      align = "center",
                      do.call(
                        "actionButton",
                        c(
                          save_bttn,
                          list(
                            inputId = "custTermSave"
                          )
                        )
                      )
                    )
                  )
                )
              ),
              hr(),
              div(
                helpText(
                  "Pressing Run Button will delete previous custom PoS"
                ),
                style = "text-align: center"
              )
            ),
            style = "margin-top:40px"
          )
        )
      )
    )
  )

  # Synonyms Management -----

  synonyms <- tabItem(
    tabName = "synonymsMgmt",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("right-left"), strong("Synonyms Merging"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      br(),
      br(),
      fluidRow(
        column(
          9,
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Processed Data with Synonyms", icon = icon("table"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("synonymsProcessedData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Synonyms List Preview", icon = icon("list"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("synonymsListPreview"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & Instructions", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(helpContent()$synonyms)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          3,
          div(
            box(
              width = 12,
              div(
                h3(strong(em("Import Synonyms File"))),
                style = "margin-top:-57px"
              ),
              hr(),
              helpText(h5(
                "Upload a CSV or Excel file with your synonyms list. Column 2 must contain the PoS tag (upos) to assign."
              )),
              fileInput(
                "synonyms_file",
                label = NULL,
                multiple = FALSE,
                accept = c(
                  ".csv",
                  ".xls",
                  ".xlsx"
                )
              ),
              hr(),
              helpText(
                uiOutput("synonymsStats")
              ),
              hr(),
              # Download buttons section
              fluidRow(
                column(
                  6,
                  downloadButton(
                    "downloadSynonymsTemplate",
                    "Template",
                    style = "width: 100%; margin-bottom: 10px; background-color: #6CC283; border-color: #4F7942;color: #ffffff;"
                  )
                ),
                column(
                  6,
                  downloadButton(
                    "downloadVocabulary",
                    "Vocabulary",
                    style = "width: 100%; margin-bottom: 10px; background-color: #6CC283; border-color: #4F7942;color: #ffffff;"
                  )
                )
              ),
              hr(),
              fluidRow(
                column(
                  4,
                  div(
                    align = "center",
                    style = "margin-top:-15px",
                    width = 12,
                    do.call(
                      "actionButton",
                      c(
                        run_bttn,
                        list(
                          inputId = "synonymsRun"
                        )
                      )
                    )
                  )
                ),
                column(
                  4,
                  div(
                    align = "center",
                    style = "margin-top:-15px",
                    width = 12,
                    title = t_back,
                    do.call(
                      "actionButton",
                      c(
                        back_bttn,
                        list(
                          inputId = "synonymsBack"
                        )
                      )
                    )
                  )
                ),
                column(
                  4,
                  div(
                    title = t_save,
                    div(
                      align = "center",
                      do.call(
                        "actionButton",
                        c(
                          save_bttn,
                          list(
                            inputId = "synonymsSave"
                          )
                        )
                      )
                    )
                  )
                )
              ),
              hr(),
              div(
                helpText(
                  "Pressing Run will replace synonyms and update their PoS tags"
                ),
                style = "text-align: center"
              )
            ),
            style = "margin-top:40px"
          )
        )
      )
    )
  )

  ## Multi-Word Creation -----

  mwcreation <- tabItem(
    tabName = "multiwordCreat",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("link"), strong("Multi-Word Creation"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      br(),
      br(),
      fluidRow(
        column(
          7,
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Multi-Word List", icon = icon("list"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("multiwordList"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Annotated Text including Multi-Word", icon = icon("table"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("multiwordData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & References", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(infoTexts$multiwordcreation)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          5,
          div(
            box(
              width = 12,
              div(
                h3(strong(em("Automatic Multi-Words"))),
                style = "margin-top:-57px"
              ),
              helpText(h5(
                "Multi-word creation extracts keywords (sequence of terms) from the text."
              )),
              helpText(h5(
                "After keywords are generated, select those you wish to include in your data from the list."
              )),
              hr(),
              style = "text-align: left; text-color: #989898",
              fluidRow(
                column(
                  12,
                  selectInput(
                    "MWmethod",
                    "Relevant Collocation Algorithm",
                    choices = c(
                      "Dependency Parsing (NP)" = "dep",
                      "Rake" = "rake",
                      "Pointwise Mutual Information" = "pmi",
                      "Mutual Dependency" = "md",
                      "Log-Frequency Biased Mutual Dependency" = "lfmd",
                      "Absorption Index (IS)" = "is"
                    ),
                    selected = "dep"
                  )
                )
              ),
              fluidRow(
                column(
                  6,
                  conditionalPanel(
                    'input.MWmethod == "dep"',
                    numericInput(
                      inputId = "ngram_max_dep",
                      label = "Max Phrase Length",
                      min = 2,
                      max = 10,
                      value = 5,
                      step = 1
                    )
                  ),
                  conditionalPanel(
                    'input.MWmethod != "dep"',
                    numericInput(
                      inputId = "ngram_max",
                      label = "Ngrams",
                      min = 2,
                      max = 10,
                      value = 4,
                      step = 1
                    )
                  )
                ),
                column(
                  6,
                  numericInput(
                    inputId = "freq_minMW",
                    label = "Freq Min",
                    min = 1,
                    max = Inf,
                    value = 3,
                    step = 1
                  )
                )
              ),
              conditionalPanel(
                'input.MWmethod == "dep"',
                div(
                  style = "padding: 10px; background-color: #e3f2fd; border-left: 4px solid #2196F3; border-radius: 4px; margin-bottom: 10px;",
                  icon("info-circle", style = "color: #1976d2;"),
                  span(
                    " Noun phrases are extracted from the syntactic dependency tree. PoS tag selection is not required.",
                    style = "color: #1976d2; font-size: 12px;"
                  )
                )
              ),
              conditionalPanel(
                'input.MWmethod != "dep" && input.MWmethod != "is"',
                fluidRow(
                  column(
                    12,
                    h5(em(strong("Multi-Words created by:")))
                  ),
                )
              ),
              conditionalPanel(
                'input.MWmethod == "is"',
                fluidRow(
                  column(
                    12,
                    h5(em(strong("Lexical Words to use:")))
                  ),
                )
              ),
              conditionalPanel(
                'input.MWmethod != "dep"',
                fluidRow(
                  column(
                    12,
                    div(
                      class = "multicol",
                      uiOutput("multiwordPosSel")
                    )
                  )
                )
              ),
              hr(),
              div(
                fluidRow(
                  column(
                    4,
                    div(
                      align = "center",
                      style = "margin-top:-15px",
                      width = 12,
                      do.call(
                        "actionButton",
                        c(
                          run_bttn,
                          list(
                            inputId = "multiwordCreatRun"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    4,
                    div(
                      align = "center",
                      style = "margin-top:-15px",
                      width = 12,
                      title = t_back,
                      do.call(
                        "actionButton",
                        c(
                          back_bttn,
                          list(
                            inputId = "multiwordCreatBack"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    4,
                    div(
                      title = t_save,
                      div(
                        align = "center",
                        do.call(
                          "actionButton",
                          c(
                            save_bttn,
                            list(
                              inputId = "multiwordCreatSave"
                            )
                          )
                        )
                      )
                    )
                  )
                ),
                style = "margin-top:-15px"
              ),
              hr(),
              # prova pulsante apply multiword
              uiOutput("multiwordCreatApplyUI")
            ),
            style = "margin-top:40px; width:100%;"
          )
        )
      )
    )
  )

  ## Multi-word By a List ----
  mwlist <- tabItem(
    tabName = "multiwordByList",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("list-ul"), strong("Multi-Word Creation by a List"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      br(),
      br(),
      fluidRow(
        column(
          9,
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Multi-Word List", icon = icon("list"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("multiwordList2"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Annotated Text including Multi-Word", icon = icon("table"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("multiwordData2"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & References", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(infoTexts$multiwordlist)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          3,
          div(
            box(
              width = 12,
              div(
                h3(strong(em("Import a Multi-Word List"))),
                style = "margin-top:-57px"
              ),
              helpText(h5(
                "Please ensure that the Multi-Word List is formatted as an Excel/CSV file with one column.
                                       Each cell of that column include a multi-word. Each term have to be separated by a single whitespace."
              )),
              hr(),
              style = "text-align: left; text-color: #989898",
              fluidRow(column(
                12,
                fileInput(
                  "multiword_lists",
                  label = NULL,
                  multiple = TRUE,
                  accept = c(
                    ".csv",
                    ".xls",
                    ".xlsx"
                  )
                )
              )),
              hr(),
              # Download buttons section
              fluidRow(
                column(
                  6,
                  downloadButton(
                    "downloadMultiwordTemplate",
                    "Template",
                    style = "width: 100%; margin-bottom: 10px; background-color: #6CC283; border-color: #4F7942;color: #ffffff;"
                  )
                ),
                column(
                  6,
                  downloadButton(
                    "downloadMultiwordVocabulary",
                    "Vocabulary",
                    style = "width: 100%; margin-bottom: 10px; background-color: #6CC283; border-color: #4F7942;color: #ffffff;"
                  )
                )
              ),
              hr(),
              div(
                fluidRow(
                  column(
                    4,
                    div(
                      align = "center",
                      style = "margin-top:-15px",
                      width = 12,
                      do.call(
                        "actionButton",
                        c(
                          run_bttn,
                          list(
                            inputId = "multiwordListRun"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    4,
                    div(
                      align = "center",
                      style = "margin-top:-15px",
                      width = 12,
                      title = t_back,
                      do.call(
                        "actionButton",
                        c(
                          back_bttn,
                          list(
                            inputId = "multiwordListBack"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    4,
                    div(
                      title = t_save,
                      div(
                        align = "center",
                        do.call(
                          "actionButton",
                          c(
                            save_bttn,
                            list(
                              inputId = "multiwordListSave"
                            )
                          )
                        )
                      )
                    )
                  )
                ),
                style = "margin-top:-15px"
              )
            ),
            style = "margin-top:40px; width:100%;"
          )
        )
      )
    )
  )

  ## PoS Tag Selection -----
  postagselection <- tabItem(
    tabName = "posTagSelect",
    fluidPage(
      fluidRow(
        column(
          8,
          h2(icon("filter"), strong("PoS Tag Selection"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;"),
          br(),
          br(),
          tabsetPanel(
            type = "tabs",
            tabPanel(
              title = "Annotated Text", icon = icon("table"),
              br(),
              shinycssloaders::withSpinner(
                DT::DTOutput("posTagSelectData"),
                color = getOption("spinner.color", default = "#4F7942")
              )
            ),
            tabPanel(
              title = "Info & References", icon = icon("circle-info"),
              br(),
              fluidPage(
                fluidRow(
                  column(1),
                  column(
                    10,
                    br(),
                    HTML(infoTexts$posselection)
                  ),
                  column(1)
                )
              )
            )
          )
        ),
        column(
          4,
          div(
            box(
              width = 12,
              style = "padding-top: 0px; margin-top: 0px;", # Azzera completamente

              fluidRow(
                style = "margin-top: 0px;", # Azzera anche la fluidRow
                column(
                  12,
                  style = "padding-top: 0px;", # Azzera anche la column

                  uiOutput("posStatsUI"),

                  # PoS Tags UI
                  div(
                    uiOutput("posTagListsUI")
                  ),

                  # Filters Box (Hapax e Single Character)
                  div(
                    style = "background: white; border: 2px solid #95a5a6; border-radius: 8px; padding: 15px; margin-top: 15px; box-shadow: 0 2px 4px rgba(0,0,0,0.05); width: 100%;",

                    # Header
                    div(
                      style = "display: flex; align-items: center; margin-bottom: 12px; padding-bottom: 8px; border-bottom: 2px solid #95a5a6;",
                      div(
                        style = "width: 35px; height: 35px; border-radius: 50%; background: #95a5a6; display: flex; align-items: center; justify-content: center; margin-right: 10px; flex-shrink: 0;",
                        icon(
                          "filter",
                          style = "color: white; font-size: 16px;"
                        )
                      ),
                      div(
                        strong(
                          "Additional Filters",
                          style = "color: #2c3e50; font-size: 15px;"
                        )
                      )
                    ),

                    # Checkboxes affiancati con column-count
                    div(
                      style = "column-count: 2; column-gap: 20px;",
                      checkboxGroupInput(
                        "posTagHapax",
                        label = NULL,
                        choices = c("Hapax"),
                        selected = "Hapax"
                      ),
                      checkboxGroupInput(
                        "posTagSingleChar",
                        label = NULL,
                        choices = c("Single Character"),
                        selected = NULL
                      )
                    )
                  )
                )
              ),
              div(
                hr(),
                div(
                  fluidRow(
                    column(
                      6,
                      div(
                        align = "center",
                        style = "margin-top:-15px",
                        width = 12,
                        do.call(
                          "actionButton",
                          c(
                            run_bttn,
                            list(
                              inputId = "posTagSelectRun"
                            )
                          )
                        )
                      )
                    ),
                    column(
                      6,
                      div(
                        title = t_save,
                        div(
                          align = "center",
                          do.call(
                            "actionButton",
                            c(
                              save_bttn,
                              list(
                                inputId = "posTagSelectSave"
                              )
                            )
                          )
                        )
                      )
                    )
                  ),
                  style = "margin-top:-1px"
                ),
                style = "margin-top:-1px"
              )
            ),
            style = "margin-top:0px"
          )
        )
      )
    )
  )

  return(list(
    tokpos = tokpos,
    specialentities = specialentities,
    custompos = custompos,
    synonyms = synonyms,
    mwcreation = mwcreation,
    mwlist = mwlist,
    postagselection = postagselection
  ))
}

preprocessingServer <- function(input, output, session, values, statsValues) {
  ### PRE-PROCESSING ----

  ## Tokenization & PoS Tagging ----

  output$flagUI <- renderUI({
    tags$img(
      src = values$flag,
      height = "25px",
      width = "40px",
      style = "margin-top:-30px;"
    )
  })

  output$optionsTokenization <- renderUI({
    selectInput(
      inputId = "language_model",
      label = "Language",
      choices = values$label_lang,
      multiple = FALSE,
      width = "100%",
      selected = values$language
    )
  })

  output$treebankSelect <- renderUI({
    selected_treebanks <- values$languages$treebank[
      values$languages$language_name == values$language
    ]
    selectInput(
      "treebank",
      "Treebank",
      choices = selected_treebanks,
      selected = values$treebank
    )
  })

  output$info_treebank <- renderUI({
    ud_description <- values$languages %>%
      filter(language_name == values$language, treebank == values$treebank) %>%
      select(description) %>%
      as.character()
    ud_info <- values$languages %>%
      filter(language_name == values$language, treebank == values$treebank) %>%
      select(tokens, words, sentences)
    ud_info <- paste0(
      "Tokens: ",
      format(as.numeric(ud_info$tokens), big.mark = ",", scientific = FALSE),
      " - Words: ",
      format(as.numeric(ud_info$words), big.mark = ",", scientific = FALSE),
      " - Sentences: ",
      format(as.numeric(ud_info$sentences), big.mark = ",", scientific = FALSE)
    )
    accuracy <- values$accuracy %>%
      filter(language_name == values$language, treebank == values$treebank) %>%
      select(Words, Lemma, Sentences, UPOS)
    ud_accuracy1 <- paste0(
      "Words: ",
      accuracy$Words,
      "% ---  Lemma: ",
      accuracy$Lemma,
      "%"
    )
    ud_accuracy2 <- paste0(
      "Sentences: ",
      accuracy$Sentences,
      "%  ---  PoS:   ",
      accuracy$UPOS,
      "%"
    )
    ud_contributors <- values$languages %>%
      filter(language_name == values$language, treebank == values$treebank) %>%
      select(contributors) %>%
      as.character()
    ud_hub_page_link <- values$languages %>%
      filter(language_name == values$language, treebank == values$treebank) %>%
      select(hub_page_link) %>%
      as.character()
    values$flag <- values$languages %>%
      filter(language_name == values$language, treebank == values$treebank) %>%
      select(flag) %>%
      as.character()
    #  HTML
    tagList(
      tags$div(
        class = "card",
        # style = "border: 1px solid #ddd; border-radius: 5px; padding: 15px; margin: 10px;",
        tags$h4(strong(em("Description")), style = "font-size: 12px;"),
        tags$p(ud_description, style = "font-size: 11px;"),
        tags$h4(strong(em("Treebank Data")), style = "font-size: 12px;"),
        tags$p(ud_info, style = "font-size: 11px;"),
        tags$h4(strong(em("Model Accuracy")), style = "font-size: 12px;"),
        tags$p(ud_accuracy1, style = "font-size: 11px;"),
        tags$p(ud_accuracy2, style = "font-size: 11px;"),
        tags$h4(strong(em("Contributors")), style = "font-size: 12px;"),
        tags$p(ud_contributors, style = "font-size: 10px;"),
        # tags$h4("Hub Page"),
        tags$a(
          href = ud_hub_page_link,
          target = "_blank",
          "UD Treebank link",
          style = "font-size: 10px; color: blue; text-decoration: underline;"
        )
      )
    )
  })

  observeEvent(input$language_model, {
    values$language <- input$language_model
    selected_treebanks <- values$languages$treebank[
      values$languages$language_name == values$language
    ]
    updateSelectInput(
      session,
      "treebank",
      choices = selected_treebanks,
      selected = values$treebank
    )
  })

  observeEvent(input$treebank, {
    values$treebank <- input$treebank
  })

  posTagging <- eventReactive(
    {
      input$tokPosRun
    },
    {
      # values$language <- input$language_model
      # values$treebank <- input$treebank
      values$language_file <- values$languages %>%
        dplyr::filter(
          language_name == values$language,
          treebank == values$treebank
        ) %>%
        select(file) %>%
        as.character()
      ## download and load model language
      udmodel_lang <- loadLanguageModel(file = values$language_file)

      # ## set cores for parallel computing
      ncores <- coresCPU()

      # Lemmatization and POS Tagging
      values$dfTag <- udpipe(
        object = udmodel_lang,
        x = values$txt %>%
          dplyr::filter(doc_selected),
        parallel.cores = ncores
      )

      # Merge metadata from the original txt object
      values$dfTag <- values$dfTag %>%
        left_join(
          values$txt %>% select(-text, -text_original),
          by = "doc_id"
        ) %>%
        filter(!is.na(upos)) %>% ##
        posSel(., c("ADJ", "NOUN", "PROPN", "VERB"))

      if (input$token_lowercase) {
        values$dfTag <- values$dfTag %>%
          mutate(
            token = tolower(token)
          )
      }
      if (input$lemma_lowercase) {
        values$dfTag <- values$dfTag %>%
          mutate(
            lemma = tolower(lemma)
          )
      }

      values$dfTag$docSelected <- TRUE
      values$menu <- 1
      statsValues <- updateStats(
        values$dfTag,
        term = values$generalTerm,
        statsValues
      )
    }
  )

  output$unitAnalysis <- renderUI({
    if (!is.null(values$dfTag)) {
      list(
        # h5(strong("Select Analysis Term: Lemma or Token")),
        # br(),
        radioGroupButtons(
          inputId = "generalTerm",
          label = "Select Analysis Term:",
          choices = c("Lemma" = "lemma", "Token" = "token"),
          selected = values$generalTerm,
          status = "primary",
          justified = TRUE
        )
      )
    }
  })

  ## Term selected on the dashboard
  output$termSelected <- renderText({
    if (!is.null(input$generalTerm)) {
      values$generalTerm <- input$generalTerm
      statsValues <- updateStats(
        values$dfTag,
        term = values$generalTerm,
        statsValues
      )
    }
    if (values$menu >= 1) {
      HTML(paste(
        "Analysis by: <b>",
        tools::toTitleCase(values$generalTerm),
        "</b>"
      ))
    }
  })

  output$tokPosTagData <- DT::renderDT({
    posTagging()

    if (!is.null(values$dfTag)) {
      DTformat(
        values$dfTag %>%
          filter(docSelected) %>%
          select(
            doc_id,
            paragraph_id,
            sentence_id,
            sentence,
            token,
            lemma,
            upos
          ) %>%
          rename(
            D_id = doc_id,
            P_id = paragraph_id,
            S_id = sentence_id,
            Sentence = sentence,
            Token = token,
            Lemma = lemma,
            "Part of Speech" = upos
          )
      )
    }
  })

  observeEvent(
    eventExpr = {
      input$tokPosSave
    },
    handlerExpr = {
      file <- paste("Tall-Export-File-", sys.time(), ".tall", sep = "")
      file_path <- destFolder(file, values$wdTall)
      saveTall(
        values$dfTag,
        values$custom_lists,
        values$language,
        values$treebank,
        values$menu,
        "Custom Term Lists",
        file_path,
        values$generalTerm,
        values$corpus_description
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## Tagging Special Entities ----
  posSpecialTagging <- eventReactive(
    {
      input$posSpecialRun
    },
    {
      res <- TaggingCorpusElements(values$dfTag)

      values$dfTag <- res$x %>% filter(!token %in% c("#", "@")) # remove empty hashs and tags
      values$posSpecialData <- res$resList %>% filter(!item %in% c("#", "@"))
      statsValues <- updateStats(
        values$dfTag,
        term = values$generalTerm,
        statsValues
      )
      rm(res)

      # Store summary data for cards
      values$posSpecialSummary <- values$posSpecialData %>%
        summarySpecialEntities(type = "all")
    },
    ignoreNULL = TRUE
  )

  ## Render HTML Cards for Special Entities (COMPACT VERSION) ----
  output$posSpecialCardsUI <- renderUI({
    posSpecialTagging()

    # Get summary data
    summary_data <- values$posSpecialSummary

    # Define entity metadata with colors, icons, and labels
    entity_metadata <- list(
      EMAIL = list(
        color = "#3498db",
        icon = "envelope",
        label = "Email Addresses",
        example = "example@domain.com"
      ),
      EMOJI = list(
        color = "#f39c12",
        icon = "smile",
        label = "Emojis",
        example = "😊 🚀 ❤️"
      ),
      HASH = list(
        color = "#9b59b6",
        icon = "hashtag",
        label = "Hashtags",
        example = "#ExampleTag"
      ),
      IP_ADDRESS = list(
        color = "#e74c3c",
        icon = "network-wired",
        label = "IP Addresses",
        example = "192.168.1.1"
      ),
      MENTION = list(
        color = "#1abc9c",
        icon = "at",
        label = "Mentions",
        example = "@username"
      ),
      URL = list(
        color = "#34495e",
        icon = "link",
        label = "URLs",
        example = "https://example.com"
      )
    )

    # Create card HTML for each entity type
    cards <- lapply(1:nrow(summary_data), function(i) {
      entity_type <- summary_data$UPOS[i]
      frequency <- summary_data$Frequency[i]
      metadata <- entity_metadata[[entity_type]]

      # Skip if metadata not found
      if (is.null(metadata)) {
        return(NULL)
      }

      # Generate unique button ID for modal
      button_id <- paste0("viewEntity_", entity_type)

      # Create compact card HTML
      div(
        style = sprintf(
          "background: white;
         border-left: 4px solid %s;
         border-radius: 8px;
         padding: 12px 16px;
         margin-bottom: 10px;
         box-shadow: 0 1px 4px rgba(0,0,0,0.08);
         transition: all 0.3s ease;
         cursor: pointer;",
          metadata$color
        ),
        class = "entity-card",

        # Card content with horizontal layout
        fluidRow(
          style = "display: flex; align-items: center; margin: 0;",

          # Icon column (reduced size)
          column(
            2,
            style = "padding: 0; display: flex; align-items: center; justify-content: center;",
            div(
              style = sprintf(
                "width: 45px;
               height: 45px;
               border-radius: 50%%;
               background: linear-gradient(135deg, %s, %s);
               display: flex;
               align-items: center;
               justify-content: center;
               box-shadow: 0 2px 6px rgba(0,0,0,0.12);",
                metadata$color,
                adjustcolor(metadata$color, alpha.f = 0.7)
              ),
              icon(metadata$icon, style = "color: white; font-size: 20px;")
            )
          ),

          # Entity info column
          column(
            6,
            style = "padding: 0 10px; display: flex; flex-direction: column; justify-content: center;",
            div(
              h5(
                strong(metadata$label),
                style = sprintf(
                  "color: %s; margin: 0 0 3px 0; font-size: 15px;",
                  metadata$color
                )
              ),
              p(
                style = "color: #95a5a6; margin: 0; font-size: 11px; line-height: 1.3;",
                sprintf("Example: %s", metadata$example)
              )
            )
          ),

          # Frequency badge and view button column (horizontal layout)
          column(
            4,
            style = "padding: 0; display: flex; align-items: center; justify-content: flex-end; gap: 8px;",

            # Frequency badge
            div(
              style = sprintf(
                "background: %s;
               color: white;
               padding: 6px 14px;
               border-radius: 16px;
               font-weight: bold;
               font-size: 14px;
               min-width: 45px;
               text-align: center;",
                metadata$color
              ),
              format(frequency, big.mark = ",")
            ),

            # View button (only if frequency > 0)
            if (frequency > 0) {
              actionButton(
                inputId = button_id,
                label = NULL,
                icon = icon("eye"),
                style = sprintf(
                  "background: %s;
                 color: white;
                 border: none;
                 padding: 6px 12px;
                 border-radius: 5px;
                 font-size: 12px;
                 cursor: pointer;
                 height: 32px;
                 min-width: 32px;",
                  metadata$color
                ),
                onclick = sprintf(
                  "Shiny.setInputValue('viewEntityModal', '%s', {priority: 'event'})",
                  entity_type
                )
              )
            }
          )
        )
      )
    })

    # Add custom CSS for hover effects
    tagList(
      tags$style(HTML(
        "
      .entity-card:hover {
        transform: translateY(-2px);
        box-shadow: 0 4px 12px rgba(0,0,0,0.15) !important;
      }
    "
      )),
      cards
    )
  })

  ## Modal for Frequency Distribution ----
  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$viewEntityModal
    },
    handlerExpr = {
      if (!is.null(input$viewEntityModal) && input$viewEntityModal != "") {
        entity_type <- input$viewEntityModal

        # Get entity metadata for modal styling
        entity_metadata <- list(
          EMAIL = list(color = "#3498db", label = "Email", icon = "envelope"),
          EMOJI = list(color = "#f39c12", label = "Emoji", icon = "smile"),
          HASH = list(color = "#9b59b6", label = "Hashtag", icon = "hashtag"),
          IP_ADDRESS = list(
            color = "#e74c3c",
            label = "IP Address",
            icon = "network-wired"
          ),
          MENTION = list(color = "#1abc9c", label = "Mention", icon = "at"),
          URL = list(color = "#34495e", label = "URL", icon = "link")
        )

        metadata <- entity_metadata[[entity_type]]

        # Create and show modal with white title text
        showModal(
          modalDialog(
            title = div(
              style = "color: white; font-weight: bold; font-size: 20px; display: flex; align-items: center; gap: 10px;",
              icon(metadata$icon, style = "font-size: 22px;"),
              sprintf("Frequency Distribution of %s Entities", metadata$label)
            ),
            # Modal content
            div(
              style = "padding: 15px;",
              shinycssloaders::withSpinner(
                htmlOutput("specialEntityFreqHTML"),
                color = metadata$color
              )
            ),
            size = "l",
            easyClose = TRUE,
            footer = tagList(
              actionButton(
                inputId = "closeModalSpecialEntities",
                label = "Close",
                icon = icon("times"),
                style = sprintf(
                  "background: %s; color: white; border: none; padding: 8px 20px; border-radius: 5px;",
                  metadata$color
                )
              )
            )
          )
        )
      }
    }
  )

  ## Close modal observer ----
  observeEvent(input$closeModalSpecialEntities, {
    removeModal()
  })

  ## Render Frequency Distribution in Modal (WITH LARGE EMOJIS) ----
  output$specialEntityFreqHTML <- renderUI({
    req(input$viewEntityModal)

    entity_type <- input$viewEntityModal

    # Get frequency data
    freq_data <- values$posSpecialData %>%
      summarySpecialEntities(type = entity_type) %>%
      rename("Frequency" = "n")

    # Get entity color
    colors <- list(
      EMAIL = "#3498db",
      EMOJI = "#f39c12",
      HASH = "#9b59b6",
      IP_ADDRESS = "#e74c3c",
      MENTION = "#1abc9c",
      URL = "#34495e"
    )
    entity_color <- colors[[entity_type]]

    # Check if we're displaying emojis (for larger font size)
    is_emoji <- (entity_type == "EMOJI")

    # Create HTML table with modern styling
    table_html <- "<div style='max-height: 500px; overflow-y: auto;'>"
    table_html <- paste0(
      table_html,
      "
    <table style='width: 100%; border-collapse: collapse; font-family: Arial, sans-serif;'>
      <thead>
        <tr style='background: linear-gradient(135deg, ",
      entity_color,
      ", ",
      adjustcolor(entity_color, alpha.f = 0.8),
      "); color: white; position: sticky; top: 0; z-index: 10;'>
          <th style='padding: 12px; text-align: left; font-weight: bold; border-bottom: 2px solid white;'>Item</th>
          <th style='padding: 12px; text-align: center; font-weight: bold; border-bottom: 2px solid white; width: 120px;'>Frequency</th>
        </tr>
      </thead>
      <tbody>
  "
    )

    # Add rows
    for (i in 1:nrow(freq_data)) {
      item <- freq_data$item[i]
      frequency <- freq_data$Frequency[i]

      # Special handling for different entity types
      if (entity_type == "URL") {
        # URLs - make them clickable
        item_display <- sprintf(
          '<a href="%s" target="_blank" style="color: %s; text-decoration: none;">%s</a>',
          item,
          entity_color,
          item
        )
      } else if (is_emoji) {
        # EMOJIS - increase font size for better visibility
        item_display <- sprintf(
          '<span style="font-size: 28px; line-height: 1.5;">%s</span>',
          htmltools::htmlEscape(item)
        )
      } else {
        # Other types - normal display
        item_display <- htmltools::htmlEscape(item)
      }

      # Alternate row colors
      bg_color <- if (i %% 2 == 0) "#f8f9fa" else "white"

      # Adjust padding for emoji rows (more vertical space needed)
      row_padding <- if (is_emoji) "14px" else "10px"

      table_html <- paste0(
        table_html,
        sprintf(
          "
      <tr style='background: %s; transition: background 0.2s;' onmouseover='this.style.background=\"#e8f4f8\"' onmouseout='this.style.background=\"%s\"'>
        <td style='padding: %s; border-bottom: 1px solid #dee2e6; vertical-align: middle;'>%s</td>
        <td style='padding: %s; text-align: center; border-bottom: 1px solid #dee2e6; vertical-align: middle;'>
          <span style='background: %s; color: white; padding: 4px 12px; border-radius: 12px; font-weight: bold; font-size: 13px;'>%s</span>
        </td>
      </tr>
    ",
          bg_color,
          bg_color,
          row_padding,
          item_display,
          row_padding,
          entity_color,
          format(frequency, big.mark = ",")
        )
      )
    }

    table_html <- paste0(
      table_html,
      "
      </tbody>
    </table>
  </div>"
    )

    # Add summary info
    summary_html <- sprintf(
      "
    <div style='margin-top: 15px; padding: 12px; background: #f8f9fa; border-radius: 8px; border-left: 4px solid %s;'>
      <strong style='color: %s;'>Total Items:</strong> %s |
      <strong style='color: %s;'>Total Occurrences:</strong> %s
    </div>
  ",
      entity_color,
      entity_color,
      format(nrow(freq_data), big.mark = ","),
      entity_color,
      format(sum(freq_data$Frequency), big.mark = ",")
    )

    HTML(paste0(table_html, summary_html))
  })

  ## Annotated Text Table (unchanged) ----
  output$posSpecialData <- DT::renderDT({
    posSpecialTagging()

    if (!is.null(values$dfTag)) {
      DTformat(
        values$dfTag %>%
          filter(docSelected) %>%
          select(
            doc_id,
            paragraph_id,
            sentence_id,
            sentence,
            token,
            lemma,
            upos
          ) %>%
          rename(
            D_id = doc_id,
            P_id = paragraph_id,
            S_id = sentence_id,
            Sentence = sentence,
            Token = token,
            Lemma = lemma,
            "Part of Speech" = upos
          ),
        col_to_remove = values$generalTerm
      )
    }
  })

  ## Back button - Reset special entities ----
  observeEvent(input$posSpecialBack, {
    values$dfTag <- resetSpecialEntities(values$dfTag)
    statsValues <- updateStats(
      values$dfTag,
      term = values$generalTerm,
      statsValues
    )

    # Reset summary data
    values$posSpecialSummary <- data.frame(
      UPOS = toupper(c(
        "email",
        "url",
        "hash",
        "emoji",
        "ip_address",
        "mention"
      )),
      Frequency = rep(0, 6)
    )

    popUpGeneric(
      title = "Special Entities Tags Removed",
      type = "waiting",
      color = c("#FFA800"),
      subtitle = paste0(
        "Now all Special Entities Tags have been removed from your documents"
      ),
      btn_labels = "OK"
    )
  })

  ## Save button ----
  observeEvent(
    eventExpr = {
      input$posSpecialSave
    },
    handlerExpr = {
      file <- paste("Tall-Export-File-", sys.time(), ".tall", sep = "")
      file_path <- destFolder(file, values$wdTall)
      saveTall(
        values$dfTag,
        values$custom_lists,
        values$language,
        values$treebank,
        values$menu,
        "POS Tag Selection",
        file_path,
        values$generalTerm,
        values$corpus_description
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  # =============================================================================
  # ADDITIONAL CSS FOR ENHANCED STYLING
  # =============================================================================
  tags$style(HTML(
    "
  /* Smooth transitions for all entity cards */
  .entity-card {
    transition: all 0.3s cubic-bezier(0.4, 0, 0.2, 1);
  }

  /* Hover effect for entity cards */
  .entity-card:hover {
    transform: translateY(-2px);
    box-shadow: 0 4px 12px rgba(0,0,0,0.15) !important;
  }

  /* Button hover effects */
  .entity-card button:hover {
    opacity: 0.9;
    transform: scale(1.05);
  }

  /* Modal backdrop styling */
  .modal-backdrop {
    background-color: rgba(0, 0, 0, 0.5);
  }

  /* Modal content styling */
  .modal-content {
    border-radius: 10px;
    box-shadow: 0 10px 40px rgba(0, 0, 0, 0.2);
  }

  /* Scrollbar styling for modal tables */
  .modal-body ::-webkit-scrollbar {
    width: 8px;
  }

  .modal-body ::-webkit-scrollbar-track {
    background: #f1f1f1;
    border-radius: 10px;
  }

  .modal-body ::-webkit-scrollbar-thumb {
    background: #888;
    border-radius: 10px;
  }

  .modal-body ::-webkit-scrollbar-thumb:hover {
    background: #555;
  }
"
  ))

  ## Custom Term List Merging ----

  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$custom_lists
    },
    handlerExpr = {
      file <- input$custom_lists
      req(file$datapath[1])
      custom_lists <- tryCatch(
        {
          parts <- lapply(file$datapath, function(x) {
            df <- read_excel(x)
            if (ncol(df) < 2) {
              stop("needs_two_cols")
            }
            df <- df %>% select(c(1, 2))
            names(df) <- c(values$generalTerm, "upos")
            df
          })
          do.call(rbind, parts)
        },
        error = function(e) e
      )
      if (inherits(custom_lists, "error")) {
        shinyWidgets::sendSweetAlert(
          session = session,
          title = "Invalid file",
          text = "The uploaded file must have at least two columns: term and POS tag. Please download and use the Custom POS List template (not the Multi-Word template).",
          type = "error"
        )
        return()
      }
      values$custom_lists <- custom_lists

      term_label <- if (identical(values$generalTerm, "lemma")) "Lemma" else "Token"
      preview_df <- custom_lists
      names(preview_df) <- c(term_label, "Part of Speech")
      preview_html <- HTML(paste0(
        "<div style='max-height:400px; overflow:auto;'>",
        paste(capture.output(print(
          knitr::kable(preview_df, format = "html", table.attr = "class='table table-striped table-sm' style='width:100%'")
        )), collapse = "\n"),
        "</div>"
      ))

      show_alert(
        title = "Custom List",
        text = preview_html,
        type = NULL,
        width = "80%",
        closeOnEsc = TRUE,
        closeOnClickOutside = TRUE,
        html = TRUE,
        showConfirmButton = TRUE,
        showCancelButton = FALSE,
        btn_labels = "OK",
        btn_colors = "#6CC283",
        timer = NULL,
        imageUrl = "",
        animation = TRUE
      )
    }
  )

  proxy1 <- dataTableProxy("posTagSelectData")

  customListMerging <- eventReactive(
    {
      input$custTermListRun
    },
    {
      # req(input$custom_lists)
      values$dfTag <- mergeCustomLists(
        values$dfTag,
        values$custom_lists,
        values$generalTerm
      )
      statsValues <- updateStats(
        values$dfTag,
        term = values$generalTerm,
        statsValues
      )
      # Update the DT proxy

      replaceData(proxy1, values$dfTag, resetPaging = FALSE)
    }
  )

  output$customPosTagData <- DT::renderDT({
    customListMerging()

    if (!is.null(values$dfTag)) {
      DTformat(
        values$dfTag %>%
          select(doc_id, sentence_id, sentence, token, lemma, upos) %>%
          rename(
            D_id = doc_id,
            S_id = sentence_id,
            Sentence = sentence,
            Token = token,
            Lemma = lemma,
            "Part of Speech" = upos
          ),
        col_to_remove = values$generalTerm
      )
    }
  })

  output$customListData <- DT::renderDT(server = FALSE, {
    term_label <- if (identical(values$generalTerm, "lemma")) "Lemma" else "Token"
    if (is.null(values$custom_lists) || !is.data.frame(values$custom_lists) || ncol(values$custom_lists) < 2) {
      return(DTformat(setNames(data.frame(a = character(), b = character()), c(term_label, "Part of Speech"))))
    }
    df_preview <- values$custom_lists[, 1:2, drop = FALSE]
    names(df_preview) <- c(term_label, "Part of Speech")
    DTformat(df_preview, col_to_remove = values$generalTerm)
  })
  outputOptions(output, "customListData", suspendWhenHidden = FALSE)

  observeEvent(
    eventExpr = {
      input$custTermSave
    },
    handlerExpr = {
      file <- paste("Tall-Export-File-", sys.time(), ".tall", sep = "")
      file_path <- destFolder(file, values$wdTall)
      saveTall(
        values$dfTag,
        values$custom_lists,
        values$language,
        values$treebank,
        values$menu,
        "Custom PoS Lists",
        file_path,
        values$generalTerm,
        values$corpus_description
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## back to the original txt
  observeEvent(input$custTermListBack, {
    values$custom_list <- NULL
    values$dfTag <- customListsReset(values$dfTag)
    statsValues <- updateStats(
      values$dfTag,
      term = values$generalTerm,
      statsValues
    )

    popUpGeneric(
      title = "Custom List Removed",
      type = "waiting",
      color = c("#FFA800"),
      subtitle = paste0(
        "Now all Custom PoS have been remove from your documents"
      ),
      btn_labels = "OK"
    )
    # Update the DT proxy
    replaceData(proxy1, values$dfTag, resetPaging = FALSE)
  })

  ## Synonyms Merging -----

  observeEvent(input$synonymsRun, {
    updateTabItems(session, "sidebarmenu", "synonymsMgmt")
  })

  observeEvent(input$synonymsBack, {
    # Restore previous state if backup exists
    if (!is.null(values$dfTag_backup_synonyms)) {
      values$dfTag <- values$dfTag_backup_synonyms
      show_alert(
        title = "Restored",
        text = "Previous state has been restored",
        type = "info",
        btn_labels = "OK",
        btn_colors = "#6CC283"
      )
    }
  })

  observeEvent(
    eventExpr = {
      input$synonymsSave
    },
    handlerExpr = {
      file <- paste("Tall-Export-File-", sys.time(), ".tall", sep = "")
      file_path <- destFolder(file, values$wdTall)
      saveTall(
        values$dfTag,
        values$custom_lists,
        values$language,
        values$treebank,
        values$menu,
        "Synonyms Merging",
        file_path,
        values$generalTerm,
        values$corpus_description
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  synonyms_data <- reactive({
    req(input$synonyms_file)

    file_path <- input$synonyms_file$datapath
    file_ext <- tools::file_ext(input$synonyms_file$name)

    tryCatch(
      {
        if (file_ext %in% c("xlsx", "xls")) {
          data <- readxl::read_excel(file_path)
        } else if (file_ext == "csv") {
          data <- read.csv(file_path, stringsAsFactors = FALSE)
        } else {
          show_alert(
            title = "Error",
            text = "Unsupported file format. Please use CSV or Excel.",
            type = "error"
          )
          return(NULL)
        }

        # Validate file structure
        if (ncol(data) < 3) {
          show_alert(
            title = "Error",
            text = "File must have at least 3 columns: target_term, upos, and at least one synonym.",
            type = "error"
          )
          return(NULL)
        }

        # Check if second column looks like upos
        valid_upos <- values$dfTag %>% distinct(upos) %>% pull(upos)

        if (!any(toupper(data[[2]]) %in% valid_upos)) {
          show_alert(
            title = "Warning",
            text = paste(
              "Column 2 should contain PoS tags (upos) like NOUN, VERB, ADJ, etc.",
              "Please verify your file structure."
            ),
            type = "warning"
          )
        }

        return(data)
      },
      error = function(e) {
        show_alert(
          title = "Error",
          text = paste("Error reading file:", e$message),
          type = "error"
        )
        return(NULL)
      }
    )
  })

  output$synonymsListPreview <- DT::renderDT({
    req(synonyms_data())

    DTformat(
      synonyms_data(),
      nrow = 20,
      filename = "synonyms_list"
    )
  })

  output$synonymsStats <- renderUI({
    if (is.null(synonyms_data())) {
      return("No file loaded")
    }

    data <- synonyms_data()
    n_targets <- nrow(data)
    n_synonyms <- sum(!is.na(data[, -(1:2)]))

    tagList(
      h5(strong("Statistics"), style = "color: #4F7942;"),
      HTML(paste0(
        "Target terms: ",
        n_targets,
        "<br/>",
        "Total synonyms: ",
        n_synonyms
      ))
    )
  })

  synonymsProcessing <- eventReactive(
    input$synonymsRun,
    {
      req(synonyms_data())

      # Create backup before processing
      values$dfTag_backup_synonyms <- values$dfTag

      show_toast(
        title = "Processing...",
        text = "Applying synonym replacements and updating PoS tags",
        type = "info",
        timer = 2000
      )

      # Apply synonyms replacement
      values$dfTag <- applySynonymsReplacement(
        dfTag = values$dfTag,
        synonyms_df = synonyms_data(),
        term_type = values$generalTerm
      )

      values$menu <- 2
      statsValues <- updateStats(
        values$dfTag,
        term = values$generalTerm,
        statsValues
      )

      show_alert(
        title = "Success",
        text = paste(
          "Synonyms replacement completed!",
          "Modified column:",
          input$synonymsTermType,
          "\\nPoS tags (upos) have been updated for replaced terms."
        ),
        type = "success",
        btn_labels = "OK",
        btn_colors = "#6CC283"
      )

      return(values$dfTag)
    }
  )

  output$synonymsProcessedData <- DT::renderDT({
    synonymsProcessing()

    if (!is.null(values$dfTag)) {
      DTformat(
        values$dfTag %>%
          select(doc_id, sentence_id, sentence, token, lemma, upos) %>%
          rename(
            D_id = doc_id,
            S_id = sentence_id,
            Sentence = sentence,
            Token = token,
            Lemma = lemma,
            PoS = upos
          ),
        nrow = 20,
        filename = "processed_with_synonyms"
      )
    }
  })

  ## Download Template Excel -----
  output$downloadSynonymsTemplate <- downloadHandler(
    filename = function() {
      paste0("synonyms_template_", Sys.Date(), ".xlsx")
    },
    content = function(file) {
      # Create template dataframe with example structure
      template_df <- data.frame(
        target_term = c(
          "machine_learning",
          "artificial_intelligence",
          "statistical_analysis",
          "",
          ""
        ),
        upos = c("NOUN", "NOUN", "NOUN", "", ""),
        synonym1 = c("ml", "ai", "statistics", "", ""),
        synonym2 = c("ML", "AI", "stats", "", ""),
        synonym3 = c("machine learning", "A.I.", "statistical method", "", ""),
        synonym4 = c("", "", "", "", ""),
        stringsAsFactors = FALSE
      )

      # Create instructions dataframe
      instructions_df <- data.frame(
        Instructions = c(
          "HOW TO USE THIS TEMPLATE:",
          "",
          "1. Column 'target_term': Enter the standardized term to use",
          "2. Column 'upos': Enter the Part-of-Speech tag (NOUN, VERB, ADJ, ADV, PROPN, etc.)",
          "3. Columns 'synonym1', 'synonym2', etc.: List all synonyms to replace",
          "",
          "IMPORTANT:",
          "- The upos value will be assigned to all occurrences of the target term",
          "- Matching is case-insensitive",
          "- Empty cells are ignored",
          "- You can add more synonym columns if needed",
          "",
          "EXAMPLE:",
          "target_term: machine_learning | upos: NOUN | synonym1: ml | synonym2: ML",
          "",
          "Valid PoS tags: NOUN, VERB, ADJ, ADV, PRON, DET, ADP, NUM, CONJ, PROPN, etc."
        )
      )

      # Write to Excel file with two sheets
      openxlsx::write.xlsx(
        list(
          Template = template_df,
          Instructions = instructions_df
        ),
        file = file
      )
    }
  )

  ## Download Vocabulary (Frequency Distribution) -----
  output$downloadVocabulary <- downloadHandler(
    filename = function() {
      paste0("vocabulary_", values$generalTerm, "_", Sys.Date(), ".xlsx")
    },
    content = function(file) {
      req(values$dfTag)

      # Get the term column based on generalTerm (token or lemma)
      term_col <- values$generalTerm

      # Calculate frequency distribution
      vocabulary <- values$dfTag %>%
        group_by(!!sym(term_col), upos) %>%
        summarise(frequency = n(), .groups = "drop") %>%
        arrange(desc(frequency)) %>%
        rename(
          term = !!sym(term_col),
          pos_tag = upos
        )

      # Create summary statistics
      summary_df <- data.frame(
        Metric = c(
          "Total unique terms",
          "Total occurrences",
          "Most frequent term",
          "Most frequent PoS tag",
          "Date generated",
          "Term type"
        ),
        Value = c(
          nrow(vocabulary),
          sum(vocabulary$frequency),
          vocabulary$term[1],
          names(sort(table(vocabulary$pos_tag), decreasing = TRUE))[1],
          as.character(Sys.Date()),
          values$generalTerm
        )
      )

      # Write to Excel file with two sheets
      openxlsx::write.xlsx(
        list(
          Vocabulary = vocabulary,
          Summary = summary_df
        ),
        file = file
      )
    }
  )

  ## Download Custom Term Template -----
  output$downloadCustomTermTemplate <- downloadHandler(
    filename = function() {
      paste0("custom_term_template_", Sys.Date(), ".xlsx")
    },
    content = function(file) {
      template_df <- data.frame(
        term = c("artificial intelligence", "deep learning", "preprocess", "dataset", "remove", ""),
        upos = c("METHOD", "METHOD", "DATA_HANDLING", "DATA_HANDLING", "TO_REMOVE", ""),
        stringsAsFactors = FALSE
      )

      instructions_df <- data.frame(
        Instructions = c(
          "HOW TO USE THIS TEMPLATE:",
          "",
          "1. Column 'term': Enter the term to tag",
          "2. Column 'upos': Enter any custom PoS tag you need for your analysis",
          "",
          "You can use ANY custom tag, for example:",
          "- Standard PoS tags: NOUN, VERB, ADJ, ADV, PROPN, etc.",
          "- Custom category tags: METHOD, APPLICATION, SETTING, DATA_HANDLING, etc.",
          "- Filtering tags: TO_REMOVE, IGNORE, STOPWORD, etc.",
          "",
          "IMPORTANT:",
          "- Matching is case-insensitive",
          "- Empty cells are ignored",
          "- Custom tags will appear in the PoS selection menus throughout the app"
        )
      )

      openxlsx::write.xlsx(
        list(
          Template = template_df,
          Instructions = instructions_df
        ),
        file = file
      )
    }
  )

  ## Download Custom Term Vocabulary -----
  output$downloadCustomTermVocabulary <- downloadHandler(
    filename = function() {
      paste0("vocabulary_custom_term_", values$generalTerm, "_", Sys.Date(), ".xlsx")
    },
    content = function(file) {
      req(values$dfTag)

      term_col <- values$generalTerm

      vocabulary <- values$dfTag %>%
        group_by(!!sym(term_col), upos) %>%
        summarise(frequency = n(), .groups = "drop") %>%
        arrange(desc(frequency)) %>%
        rename(
          term = !!sym(term_col),
          pos_tag = upos
        )

      openxlsx::write.xlsx(
        list(Vocabulary = vocabulary),
        file = file
      )
    }
  )

  ## Download Multi-Word Template -----
  output$downloadMultiwordTemplate <- downloadHandler(
    filename = function() {
      paste0("multiword_template_", Sys.Date(), ".xlsx")
    },
    content = function(file) {
      template_df <- data.frame(
        multiword = c("machine learning", "artificial intelligence", ""),
        stringsAsFactors = FALSE
      )

      instructions_df <- data.frame(
        Instructions = c(
          "HOW TO USE THIS TEMPLATE:",
          "",
          "1. Column 'multiword': Enter one multi-word per row",
          "2. Each term in the multi-word must be separated by a single whitespace",
          "",
          "EXAMPLE:",
          "machine learning",
          "artificial intelligence",
          "deep learning"
        )
      )

      openxlsx::write.xlsx(
        list(
          Template = template_df,
          Instructions = instructions_df
        ),
        file = file
      )
    }
  )

  ## Download Multi-Word Vocabulary -----
  output$downloadMultiwordVocabulary <- downloadHandler(
    filename = function() {
      paste0("vocabulary_multiword_", values$generalTerm, "_", Sys.Date(), ".xlsx")
    },
    content = function(file) {
      req(values$dfTag)

      term_col <- values$generalTerm

      vocabulary <- values$dfTag %>%
        group_by(!!sym(term_col), upos) %>%
        summarise(frequency = n(), .groups = "drop") %>%
        arrange(desc(frequency)) %>%
        rename(
          term = !!sym(term_col),
          pos_tag = upos
        )

      openxlsx::write.xlsx(
        list(Vocabulary = vocabulary),
        file = file
      )
    }
  )

  ## Multi-Word Creation ----

  output$multiwordPosSel <- renderUI({
    if (input$MWmethod != "is") {
      poslist <- posTagAll(
        values$dfTag %>%
          dplyr::filter(
            !upos %in%
              c("MULTIWORD", "NGRAM_MERGED", "PUNCT", "SYM", "X", "NUM")
          )
      )$description

      posSelected <- posTagAll(
        values$dfTag %>% dplyr::filter(upos %in% values$posMwSel)
      )$description
    } else {
      poslist <- posTagAll(
        values$dfTag %>%
          dplyr::filter(
            upos %in%
              c("NOUN", "PROPN", "ADJ", "VERB", "ADV")
          )
      )$description
      posSelected <- posTagAll(
        values$dfTag %>% dplyr::filter(upos %in% values$posMwSel)
      )$description
    }

    checkboxGroupInput(
      "multiwordPosSelGroup",
      label = NULL,
      choices = poslist,
      selected = posSelected
    )
  })

  proxy4 <- dataTableProxy("multiwordData")

  multiword <- eventReactive(
    {
      input$multiwordCreatRun
    },
    {
      ### REKA Algorithm

      values$dfTag <- rakeReset(values$dfTag) ## reset previous multiword creation steps

      if (input$MWmethod == "dep") {
        values$posMwSel <- c("NOUN", "PROPN", "ADJ")
      } else {
        values$posMwSel <- gsub(
          ":",
          "",
          gsub(":.*", "", input$multiwordPosSelGroup)
        )
      }

      ngram_max_val <- if (input$MWmethod == "dep") input$ngram_max_dep else input$ngram_max

      values$rakeResults <- rake(
        values$dfTag,
        group = "doc_id",
        ngram_max = ngram_max_val,
        relevant = values$posMwSel,
        freq.min = input$freq_minMW,
        term = values$generalTerm,
        method = input$MWmethod
      )

      values$stats <- values$rakeResults$stats

      names(values$stats) <- c(
        "Multi-Words",
        "Freq",
        "Length",
        toupper(input$MWmethod)
      )
    }
  )

  observeEvent(
    ignoreNULL = FALSE,
    eventExpr = {
      input$multiwordList_rows_selected
    },
    handlerExpr = {
      if (length(input$multiwordList_rows_selected) > 0) {
        output$multiwordCreatApplyUI <- renderUI({
          run_bttn <- list(
            label = strong("Apply List"),
            style = "border-radius: 15px; border-width: 1px; font-size: 15px; text-align: center; color: #ffff; ",
            icon = NULL
          )
          div(
            align = "center",
            style = "margin-top:-5px",
            width = 12,
            helpText(
              "Please note",
              br(),
              "pressing 'Apply List' will delete previous multiword and custom list entries"
            ),
            do.call(
              "actionButton",
              c(
                run_bttn,
                list(
                  inputId = "multiwordCreatApply"
                )
              )
            )
          )
        })
      } else {
        output$multiwordCreatApplyUI <- renderUI({})
      }
    }
  )
  output$multiwordList <- renderDT(server = FALSE, {
    multiword()
    DTformat(
      values$stats %>%
        arrange(desc(Freq), .by_group = FALSE),
      numeric = 4,
      selection = TRUE,
      nrow = 20
    )
  })

  output$multiwordData <- renderDT(server = TRUE, {
    DTformat(
      values$dfTag %>%
        dplyr::filter(docSelected) %>%
        select(doc_id, sentence, token, lemma, upos) %>%
        rename(
          D_id = doc_id,
          Sentence = sentence,
          Token = token,
          Lemma = lemma,
          "Part of Speech" = upos
        ),
      size = "80%",
      col_to_remove = values$generalTerm
    )
  })
  observeEvent(input$multiwordCreatApply, {
    row_sel <- input$multiwordList_rows_selected

    if (length(row_sel) > 0) {
      showNotification(
        ui = paste0(
          "Integrating ",
          length(row_sel),
          " Multi-Words in your corpus"
        ),
        type = "warning",
        duration = 3,
        closeButton = TRUE
      )
      values$dfTag <- applyRake(
        values$dfTag,
        rakeResults = values$rakeResults,
        row_sel = row_sel,
        term = values$generalTerm
      )

      statsValues <- updateStats(
        values$dfTag,
        term = values$generalTerm,
        statsValues
      )

      replaceData(proxy4, values$dfTag, resetPaging = FALSE)

      showNotification(
        ui = paste0(
          "Multi-Words Applied: ",
          length(row_sel),
          " multi-words have been added to your documents"
        ),
        type = "message",
        duration = 5,
        closeButton = TRUE
      )

      # show_alert(
      #   title = "Annotated Data with Multi-Words",
      #   paste0("Now Multi-Words have been added to your documents"),
      #   type = NULL,
      #   width = "50%",
      #   closeOnEsc = TRUE,
      #   closeOnClickOutside = TRUE,
      #   html = TRUE,
      #   showConfirmButton = TRUE,
      #   showCancelButton = FALSE,
      #   btn_labels = "OK",
      #   btn_colors = "#6CC283",
      #   timer = NULL,
      #   imageUrl = "",
      #   animation = TRUE
      # )
    }
  })

  observeEvent(
    eventExpr = {
      input$multiwordCreatSave
    },
    handlerExpr = {
      file <- paste("Tall-Export-File-", sys.time(), ".tall", sep = "")
      file_path <- destFolder(file, values$wdTall)
      saveTall(
        values$dfTag,
        values$stats,
        values$language,
        values$treebank,
        values$menu,
        "Multi-Word Creation",
        file_path,
        values$generalTerm,
        values$corpus_description
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## back to the original txt
  observeEvent(input$multiwordCreatBack, {
    values$dfTag <- rakeReset(values$dfTag)
    values$multiwords <- data.frame(
      keyword = "",
      ngram = NA,
      freq = NA,
      rake = NA
    )
    popUpGeneric(
      title = "Multiword Removed",
      type = "waiting",
      color = c("#FFA800"),
      subtitle = paste0(
        "Now all multiwords have been remove from your documents"
      ),
      btn_labels = "OK"
    )
  })

  ## Multi-Word by a list ----
  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$multiword_lists
    },
    handlerExpr = {
      file <- input$multiword_lists
      req(file$datapath[1])
      keyword_lists <- lapply(file$datapath, function(x) {
        x <- read_excel(x) %>% select(1)
        names(x) <- "keyword"
        return(x)
      })
      keywordList <- do.call(rbind, keyword_lists)
      values$keywordList <- keywordList
    }
  )

  multiword2 <- eventReactive(
    {
      input$multiwordListRun
    },
    {
      req(input$multiword_lists)
      # to replace with input values
      term <- values$generalTerm

      relevant <- unique(values$dfTag$upos)

      values$rakeResults <- rake(
        values$dfTag,
        relevant = relevant,
        term = term,
        freq.min = 1,
        type = "bylist",
        keywordList = values$keywordList
      )

      row_sel <- 1:nrow(values$rakeResults$stats)

      values$dfTag <- applyRake(
        values$dfTag,
        rakeResults = values$rakeResults,
        row_sel = row_sel,
        term = term
      )

      showNotification(
        ui = paste0(
          "Multi-Words Applied: ",
          length(row_sel),
          " multi-words have been added to your documents"
        ),
        type = "message",
        duration = 5,
        closeButton = TRUE
      )

      statsValues <- updateStats(
        values$dfTag,
        term = values$generalTerm,
        statsValues
      )
    }
  )

  output$multiwordData2 <- renderDT({
    multiword2()
    DTformat(
      values$dfTag %>%
        dplyr::filter(POSSelected) %>%
        group_by(doc_id, sentence_id) %>%
        select(doc_id, sentence_id, sentence, token, lemma, upos) %>%
        rename(
          D_id = doc_id,
          S_id = sentence_id,
          Sentence = sentence,
          Token = token,
          Lemma = lemma,
          "Part of Speech" = upos
        ),
      col_to_remove = values$generalTerm
    )
  })

  output$multiwordList2 <- renderDT(server = FALSE, {
    multiword2()
    DTformat(
      values$rakeResults$stats %>%
        rename(
          "Multi-Words" = keyword,
          "Lenght" = ngram,
          "Freq" = freq
        ) %>%
        arrange(desc(Freq), .by_group = FALSE)
    )
  })

  observeEvent(
    eventExpr = {
      input$multiwordListSave
    },
    handlerExpr = {
      file <- paste("Tall-Export-File-", sys.time(), ".tall", sep = "")
      file_path <- destFolder(file, values$wdTall)
      saveTall(
        values$dfTag,
        values$custom_lists,
        values$language,
        values$treebank,
        values$menu,
        "Multi-Word by a List",
        file_path,
        values$generalTerm,
        values$corpus_description
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )

  ## back to the original txt
  observeEvent(input$multiwordListBack, {
    values$dfTag <- rakeReset(values$dfTag)
    values$multiwords <- data.frame(keyword = "", ngram = NA, freq = NA)

    popUpGeneric(
      title = "Multiword Removed",
      type = "waiting",
      color = c("#FFA800"),
      subtitle = paste0(
        "Now all multiwords have been remove from your documents"
      ),
      btn_labels = "OK"
    )
  })

  ## PoS Tag Selection ----
  observe({
    output$posTagListsUI <- renderUI({
      # Ottieni tutti i PoS tags
      all_pos <- posTagAll(values$dfTag)
      selected_pos <- (posTagAll(
        values$dfTag %>% dplyr::filter(POSSelected)
      ))$description

      # Identifica dinamicamente le categorie
      classic_available <- all_pos$description[
        !grepl("Special Entity$|Custom PoS$", all_pos$description)
      ]

      special_available <- all_pos$description[
        grepl("Special Entity$", all_pos$description)
      ]

      custom_available <- all_pos$description[
        grepl("Custom PoS$", all_pos$description)
      ]

      # Funzione per creare etichette pulite
      clean_labels <- function(items, pattern) {
        labels <- gsub(pattern, "", items)
        labels <- trimws(labels)
        return(list(
          names = unname(as.list(labels)),
          values = unname(as.list(items))
        ))
      }

      special_clean <- if (length(special_available) > 0) {
        clean_labels(special_available, ": Special Entity$")
      } else {
        NULL
      }

      custom_clean <- if (length(custom_available) > 0) {
        clean_labels(custom_available, ": Custom PoS$")
      } else {
        NULL
      }

      tagList(
        # Classic PoS Card
        if (length(classic_available) > 0) {
          div(
            style = "background: white; border: 2px solid #3498db; border-radius: 8px; padding: 15px; margin-bottom: 15px; box-shadow: 0 2px 4px rgba(0,0,0,0.05); width: 100%;",

            # Header
            div(
              style = "display: flex; align-items: center; margin-bottom: 12px; padding-bottom: 8px; border-bottom: 2px solid #3498db;",
              div(
                style = "width: 35px; height: 35px; border-radius: 50%; background: #3498db; display: flex; align-items: center; justify-content: center; margin-right: 10px; flex-shrink: 0;",
                icon("tag", style = "color: white; font-size: 16px;")
              ),
              div(
                strong(
                  "Classic PoS Tags",
                  style = "color: #2c3e50; font-size: 15px;"
                ),
                div(
                  sprintf("%d tags", length(classic_available)),
                  style = "color: #7f8c8d; font-size: 11px; margin-top: 2px;"
                )
              )
            ),

            # Checkboxes in due colonne
            div(
              style = "column-count: 2; column-gap: 20px;",
              checkboxGroupInput(
                "posTagLists_classic",
                label = NULL,
                choices = classic_available,
                selected = intersect(selected_pos, classic_available)
              )
            )
          )
        },

        # Special Entities Card
        if (length(special_available) > 0) {
          div(
            style = "background: white; border: 2px solid #f39c12; border-radius: 8px; padding: 15px; margin-bottom: 15px; box-shadow: 0 2px 4px rgba(0,0,0,0.05); width: 100%;",

            # Header
            div(
              style = "display: flex; align-items: center; margin-bottom: 12px; padding-bottom: 8px; border-bottom: 2px solid #f39c12;",
              div(
                style = "width: 35px; height: 35px; border-radius: 50%; background: #f39c12; display: flex; align-items: center; justify-content: center; margin-right: 10px; flex-shrink: 0;",
                icon("star", style = "color: white; font-size: 16px;")
              ),
              div(
                strong(
                  "Special Entities",
                  style = "color: #2c3e50; font-size: 15px;"
                ),
                div(
                  sprintf("%d entities", length(special_available)),
                  style = "color: #7f8c8d; font-size: 11px; margin-top: 2px;"
                )
              )
            ),

            # Checkboxes con etichette pulite
            div(
              style = "column-count: 2; column-gap: 20px;",
              checkboxGroupInput(
                "posTagLists_special",
                label = NULL,
                choiceNames = special_clean$names,
                choiceValues = special_clean$values,
                selected = intersect(selected_pos, special_available)
              )
            )
          )
        },

        # Custom PoS Card
        if (length(custom_available) > 0) {
          div(
            style = "background: white; border: 2px solid #1abc9c; border-radius: 8px; padding: 15px; margin-bottom: 15px; box-shadow: 0 2px 4px rgba(0,0,0,0.05); width: 100%;",

            # Header
            div(
              style = "display: flex; align-items: center; margin-bottom: 12px; padding-bottom: 8px; border-bottom: 2px solid #1abc9c;",
              div(
                style = "width: 35px; height: 35px; border-radius: 50%; background: #1abc9c; display: flex; align-items: center; justify-content: center; margin-right: 10px; flex-shrink: 0;",
                icon("wrench", style = "color: white; font-size: 16px;")
              ),
              div(
                strong(
                  "Custom PoS",
                  style = "color: #2c3e50; font-size: 15px;"
                ),
                div(
                  sprintf("%d custom tags", length(custom_available)),
                  style = "color: #7f8c8d; font-size: 11px; margin-top: 2px;"
                )
              )
            ),

            # Checkboxes con etichette pulite
            div(
              style = "column-count: 2; column-gap: 20px;",
              checkboxGroupInput(
                "posTagLists_custom",
                label = NULL,
                choiceNames = custom_clean$names,
                choiceValues = custom_clean$values,
                selected = intersect(selected_pos, custom_available)
              )
            )
          )
        }
      )
    })
  })

  # Renderizza la box delle statistiche
  output$posStatsUI <- renderUI({
    # Inizializza con i valori totali se non ancora aggiornato
    n_docs <- if (statsValues$n_docs == 0) {
      length(unique(values$df$D_id))
    } else {
      statsValues$n_docs
    }

    n_tokens <- if (statsValues$n_tokens == 0) {
      nrow(values$df)
    } else {
      statsValues$n_tokens
    }

    div(
      style = "background: linear-gradient(135deg, #667eea 0%, #764ba2 100%); border-radius: 8px; padding: 20px; margin-bottom: 20px; box-shadow: 0 4px 15px rgba(102, 126, 234, 0.3);",

      # Titolo
      div(
        style = "color: white; font-size: 14px; font-weight: 600; margin-bottom: 15px; text-align: center; opacity: 0.95;",
        icon("chart-bar", style = "margin-right: 8px;"),
        "Selection Statistics"
      ),

      # Statistiche in griglia
      div(
        style = "display: grid; grid-template-columns: 1fr 1fr; gap: 15px;",

        # Documenti
        div(
          style = "background: rgba(255,255,255,0.15); border-radius: 6px; padding: 12px; text-align: center;",
          div(
            style = "color: rgba(255,255,255,0.9); font-size: 11px; margin-bottom: 5px; text-transform: uppercase; letter-spacing: 0.5px;",
            "Documents"
          ),
          div(
            style = "color: white; font-size: 24px; font-weight: bold;",
            format(n_docs, big.mark = ",")
          )
        ),

        # Token/Lemma
        div(
          style = "background: rgba(255,255,255,0.15); border-radius: 6px; padding: 12px; text-align: center;",
          div(
            style = "color: rgba(255,255,255,0.9); font-size: 11px; margin-bottom: 5px; text-transform: uppercase; letter-spacing: 0.5px;",
            ifelse(values$generalTerm == "token", "Tokens", "Lemma")
          ),
          div(
            style = "color: white; font-size: 24px; font-weight: bold;",
            format(n_tokens, big.mark = ",")
          )
        )
      ),

      # Timestamp ultimo aggiornamento (opzionale)
      if (!is.null(statsValues$last_update)) {
        div(
          style = "margin-top: 12px; padding-top: 12px; border-top: 1px solid rgba(255,255,255,0.2); text-align: center; color: rgba(255,255,255,0.8); font-size: 10px;",
          icon("sync-alt", style = "margin-right: 5px;"),
          "Updated: ",
          format(statsValues$last_update, "%H:%M:%S")
        )
      }
    )
  })

  PosFilterData <- eventReactive(
    {
      input$posTagSelectRun
    },
    {
      posTagLists_combined <- c(
        input$posTagLists_classic,
        input$posTagLists_special,
        input$posTagLists_custom
      )
      selected <- (posTagAll(values$dfTag) %>%
        dplyr::filter(description %in% (posTagLists_combined)))$pos
      values$dfTag <- removeHapaxFreq(
        values$dfTag,
        input$posTagHapax,
        input$posTagSingleChar
      )
      values$dfTag <- posSel(values$dfTag, pos = selected)
      values$menu <- 2

      statsValues <- updateStats(
        values$dfTag,
        term = values$generalTerm,
        statsValues
      )

      # Update the DT proxy
      proxy <- dataTableProxy("posTagSelectData")
      replaceData(proxy, values$dfTag, resetPaging = FALSE)
    }
  )

  output$posTagSelectData <- DT::renderDT({
    PosFilterData()

    if (!"lemma_original_nomultiwords" %in% names(values$dfTag)) {
      values$dfTag <- values$dfTag %>%
        mutate(lemma_original_nomultiwords = lemma)
    }

    DTformat(
      LemmaSelection(values$dfTag) %>%
        group_by(doc_id, sentence_id) %>%
        mutate(
          SentenceByPos = paste(lemma_original_nomultiwords, collapse = " ")
        ) %>%
        select(
          doc_id,
          sentence_id,
          sentence,
          SentenceByPos,
          token,
          lemma,
          upos
        ) %>%
        rename(
          D_id = doc_id,
          S_id = sentence_id,
          Sentence = sentence,
          Token = token,
          Lemma = lemma,
          "Part of Speech" = upos
        ) %>%
        ungroup(),
      col_to_remove = values$generalTerm
    )
  })

  observeEvent(
    eventExpr = {
      input$posTagSelectSave
    },
    handlerExpr = {
      file <- paste("Tall-Export-File-", sys.time(), ".tall", sep = "")
      file_path <- destFolder(file, values$wdTall)
      saveTall(
        values$dfTag,
        values$custom_lists,
        values$language,
        values$treebank,
        values$menu,
        "POS Tag Selection",
        file_path,
        values$generalTerm,
        values$corpus_description
      )
      popUp(title = "Saved in your working folder", type = "saved")
    }
  )
}

homeUI <- function() {
  tabItem(
    tabName = "tall",
    fluidPage(
      tags$style(HTML(
        "
      .compact-home {
        padding: 30px;
        max-width: 1300px;
        margin: 0 auto;
      }
      .header-section {
        display: flex;
        align-items: center;
        justify-content: center;
        gap: 50px;
        margin-bottom: 40px;
      }
      .logo-container img {
        height: 380px;
        width: auto;
        border-radius: 15px;
        box-shadow: 0 8px 25px rgba(0,0,0,0.15);
      }
      .title-container {
        text-align: left;
      }
      .main-title {
        font-size: 80px;
        font-weight: 700;
        color: #2c3e50;
        margin: 0 0 15px 0;
        line-height: 1;
        letter-spacing: -1px;
      }
      .subtitle {
        font-size: 28px;
        color: #7f8c8d;
        margin: 0 0 18px 0;
        font-weight: 400;
      }
      .ai-badge {
        background: linear-gradient(135deg, #4F7942 0%, #6CC283 100%);
        color: white;
        padding: 8px 20px;
        border-radius: 20px;
        font-size: 15px;
        font-weight: 700;
        display: inline-block;
        margin-bottom: 15px;
        letter-spacing: 0.5px;
      }
      .ai-description {
        font-size: 22px;
        color: #495057;
        margin-top: 10px;
        line-height: 1.5;
        font-weight: 400;
      }
      .content-grid {
        display: grid;
        grid-template-columns: 1fr 1fr;
        gap: 30px;
        margin-top: 30px;
      }
      .citation-box {
        background: linear-gradient(135deg, #4F7942 0%, #6CC283 100%);
        padding: 30px;
        border-radius: 15px;
        box-shadow: 0 6px 25px rgba(79, 121, 66, 0.3);
        color: #1a1a2e;
      }
      .citation-title {
        font-size: 20px;
        font-weight: 700;
        margin-bottom: 18px;
        display: flex;
        align-items: center;
        gap: 10px;
        color: #1a1a2e;
      }
      .citation-text {
        font-size: 16px;
        line-height: 1.8;
        background: rgba(255,255,255,0.7);
        padding: 20px;
        border-radius: 10px;
        border-left: 4px solid rgba(26, 26, 46, 0.3);
        color: #1a1a2e;
        font-weight: 400;
      }
      .citation-warning {
        margin-top: 15px;
        font-size: 15px;
        display: flex;
        align-items: center;
        gap: 8px;
        color: #1a1a2e;
        font-weight: 600;
      }
      .info-panel {
        display: flex;
        flex-direction: column;
        gap: 20px;
      }
      .info-card {
        background: #f8f9fa;
        padding: 25px;
        border-radius: 12px;
        border-left: 5px solid #4F7942;
        font-size: 16px;
        color: #495057;
        line-height: 1.6;
      }
      .info-card-title {
        font-weight: 700;
        color: #2c3e50;
        margin-bottom: 10px;
        display: flex;
        align-items: center;
        gap: 10px;
        font-size: 18px;
      }
      .website-link {
        color: #4F7942;
        font-weight: 700;
        text-decoration: none;
        border-bottom: 2px solid #4F7942;
      }
      .website-link:hover {
        color: #3a5c31;
        border-bottom-color: #3a5c31;
      }
    "
      )),

      div(
        class = "compact-home",

        # Header with logo and title side by side
        div(
          class = "header-section",
          div(
            class = "logo-container",
            img(src = "tall_logo.jpg")
          ),
          div(
            class = "title-container",
            h1(
              "TALL",
              class = "main-title"
            ),
            p("Text Analysis for All", class = "subtitle"),
            br(),
            span("TALL AI POWERED", class = "ai-badge"),
            p(
              class = "ai-description",
              "A user-friendly interface for text analysis with advanced text mining",
              br(),
              "and natural language processing methods"
            )
          )
        ),

        # Grid with citation and info
        div(
          class = "content-grid",

          # Citation Box
          div(
            class = "citation-box",
            div(
              class = "citation-title",
              icon("quote-left"),
              "How to Cite"
            ),
            div(
              class = "citation-text",
              strong(
                "Aria, M., Spano, M., D'Aniello, L., Cuccurullo, C., & Misuraca, M."
              ),
              " (2026). TALL: Text analysis for all - an interactive R-shiny application for exploring, modeling, and visualizing textual data. ",
              em("SoftwareX"),
              ", 34, 102590.",
              br(),
              br(),
              tags$a(
                icon("file-alt"),
                " Full Paper (Open Access)",
                href = "https://doi.org/10.1016/j.softx.2026.102590",
                target = "_blank",
                style = "color: #1a1a2e; font-weight: 600; text-decoration: none; margin-right: 20px;"
              ),
              tags$a(
                icon("paperclip"),
                " Supplementary Material",
                href = "https://ars.els-cdn.com/content/image/1-s2.0-S2352711026000841-mmc1.pdf",
                target = "_blank",
                style = "color: #1a1a2e; font-weight: 600; text-decoration: none;"
              )
            ),
            div(
              class = "citation-warning",
              icon("exclamation-triangle"),
              "Failure to cite is a license violation"
            )
          ),

          # Info Panel
          div(
            class = "info-panel",

            div(
              class = "info-card",
              div(
                class = "info-card-title",
                icon("certificate"),
                "Open Source & Free"
              ),
              "Distributed under the MIT license. Free for academic and commercial use."
            ),

            div(
              class = "info-card",
              div(
                class = "info-card-title",
                icon("book"),
                "Documentation"
              ),
              "Visit the ",
              tags$a(
                "TALL website",
                href = "https://www.tall-app.com",
                target = "_blank",
                class = "website-link"
              ),
              " for tutorials and examples."
            )
          )
        )
      )
    )
  )
}

homeServer <- function(input, output, session, values, statsValues) {
  ### SIDEBARMENU ----
  output$rest_of_sidebar <- renderMenu({
    if (values$menu == -2) {
      # popup to inform user that at first run you have to set the working folder
      popUpGeneric(
        title = HTML(
          "<span style='color: #4F7942;'><strong>Welcome to TALL</strong></span>"
        ),
        type = "info",
        color = c("#4F7942", "#6CC283"),
        subtitle = HTML(
          "
    <div style='text-align: left; line-height: 1.6; color: #333;'>
      <p style='margin-bottom: 15px;'>
        Before we begin, please select a <strong>default working folder</strong> where all your analyses,
        reports, and results will be saved.
      </p>
      <p style='margin-bottom: 0; color: #666; font-size: 14px;'>
        <em>You can always change this location later in Settings.</em>
      </p>
    </div>
  "
        ),
        btn_labels = "Continue",
        html = TRUE
      )
      updateTabItems(session, "sidebarmenu", "settings")
    }
    if (values$menu == 2) {
      if (length(noGroupLabels(names(values$dfTag))) > 0) {
        values$menu <- 3
      }
    }
    sidebarMenu(.list = menuList(values$menu))
  })

  observeEvent(input$workingfolder, {
    updateTabItems(session, "sidebarmenu", "settings")
  })

  observeEvent(input$runImport, {
    updateTabItems(session, "sidebarmenu", "import_tx")
  })

  observeEvent(input$biblioRun, {
    updateTabItems(session, "sidebarmenu", "import_tx")
  })

  observeEvent(input$tokPosRun, {
    updateTabItems(session, "sidebarmenu", "tokPos")
  })

  observeEvent(input$custTermListRun, {
    updateTabItems(session, "sidebarmenu", "custTermList")
  })

  observeEvent(input$posTagSelectRun, {
    updateTabItems(session, "sidebarmenu", "posTagSelect")
  })

  observeEvent(input$multiwordCreatRun, {
    updateTabItems(session, "sidebarmenu", "multiwordCreat")
  })

  observeEvent(input$multiwordCreatApply, {
    updateTabItems(session, "sidebarmenu", "multiwordCreat")
  })

  observeEvent(input$multiwordCreatBack, {
    updateTabItems(session, "sidebarmenu", "multiwordCreat")
  })

  observeEvent(input$multiwordListRun, {
    updateTabItems(session, "sidebarmenu", "multiwordByList")
  })

  observeEvent(input$multiwordListBack, {
    updateTabItems(session, "sidebarmenu", "multiwordByList")
  })

  # Settings button handler
  observeEvent(
    input$go_to_settings,
    {
      updateTabItems(session, "sidebarmenu", "settings")
    },
    ignoreInit = TRUE
  )

  # Settings button handler
  observeEvent(
    input$go_to_postagging,
    {
      updateTabItems(session, "sidebarmenu", "tokPos")
    },
    ignoreInit = TRUE
  )
}

helpContent <- function() {
  ## import ----
  importmenu <- "
  <body>

  <h3><strong>Importing Data</strong></h3>

  <p>TALL provides a versatile and user-friendly interface for importing textual data from various sources, ensuring flexibility in data handling for diverse analytical needs. The platform supports multiple file formats and structures, allowing users to seamlessly prepare their datasets for analysis.</p>
  <hr>

  <h4><strong>Supported File Formats</strong></h4>

  <h4>1. Plain Text Files (.txt)</h4>
  <p>Plain text files can be imported in three different ways, depending on the structure of the data:</p>
  <ul>
    <li><strong>Single file containing a single document:</strong> Ideal for analyzing an individual document, such as a speech transcript, literary work, or report.</li>
    <li><strong>Single file with multiple documents separated by alphanumeric codes</strong> (e.g., 'Chapter', '0001', '****'):
      <ul>
        <li>TALL automatically detects these separators, enabling structured document segmentation.</li>
        <li>Users can further refine the segmentation using the <strong>Edit → Split</strong> menu.</li>
      </ul>
    </li>
    <li><strong>Multiple .txt files, where each file represents a separate document:</strong>
      <ul>
        <li>Users can either select individual files manually or import a compressed (.zip) folder containing multiple text files.</li>
        <li>Each document will be automatically assigned an ID based on its file name, ensuring clear organization.</li>
      </ul>
    </li>
  </ul>

  <h4>2. Tabular Data (.csv, .xlsx)</h4>
  <p>Tabular formats are useful for structured datasets, such as online reviews, survey responses, or social media posts.</p>
  <ul>
    <li>The text to be analyzed must be stored in a dedicated column named <strong>'text'</strong> to ensure proper identification.</li>
    <li>Each row in the dataset is treated as an individual document.</li>
    <li>Additional metadata (e.g., timestamps, user IDs, categories) can be retained for contextual analysis.</li>
  </ul>

  <h4>3. PDF Documents (.pdf)</h4>
  <p>TALL supports the import of PDF files, facilitating the analysis of academic papers, reports, books, and other document types.</p>
  <ul>
    <li>Text extraction occurs automatically, converting the content into a format suitable for processing.</li>
    <li><strong>Limitation:</strong> At the moment, TALL can only import and process PDFs that follow a single-column formatting. PDFs with multi-column layouts, footnotes, or complex page structures may not be correctly parsed, and additional preprocessing may be required.</li>
  </ul>

  <h4>4. Biblioshiny Export Files</h4>
  <p>TALL supports the import of files exported from <strong>Biblioshiny</strong>, the graphical user interface of the <strong>Bibliometrix</strong> R package. This feature allows users to directly analyze the textual content of bibliographic metadata extracted from bibliometric databases such as Scopus or Web of Science.</p>
  <ul>
    <li>The exported file (typically in <strong>.csv</strong> format) can be loaded into TALL.</li>
    <li>Users must specify which column (e.g., <strong>Abstract</strong>, <strong>Keywords</strong>, or <strong>Title</strong>) should be used as the main textual content for analysis.</li>
    <li>Other fields (e.g., authors, year, journal) can be imported and used as metadata for document grouping or filtering.</li>
  </ul>

  <hr>
  <h4><strong>TALL Structured Files (.tall)</strong></h4>
  <p>TALL allows users to save their analysis progress in a structured format, ensuring continuity across sessions.</p>
  <ul>
    <li><strong>Save Progress:</strong> Users can export their current session as a <strong>.tall</strong> file, preserving all imported data, configurations, and analytical steps.</li>
    <li><strong>Load Saved Sessions:</strong> Previously saved <strong>.tall</strong> files can be reloaded, allowing users to resume their work seamlessly without the need to re-import or preprocess data.</li>
  </ul>

  <p>By offering flexible and structured data import capabilities, TALL streamlines the initial steps of text analysis, enabling users to focus on extracting insights efficiently.</p>
  <hr>

  <div class='references'>
    <h4><strong>References</strong></h4>
    <p><strong>Aria, M., Cuccurullo, C., D’Aniello, L., Misuraca, M., & Spano, M. (2024).</strong> <i>Breaking Barriers with TALL: A Text Analysis Shiny app for ALL</i>. In A. Dister, D. Longrée (eds.), <i>Mots competes textes déchiffrés (JADT24)</i> Presses Universitaires De Louvain Vol.1 pp.39-48.</p>
    <p><strong>Aria, M., Cuccurullo, C., D’Aniello, L., Misuraca, M., & Spano, M. (2024).</strong> <i>TALL: A New Shiny App for Text Analysis</i>. In <i>Scientific Meeting of the Italian Statistical Society</i> (pp. 64-70). Cham: Springer Nature Switzerland.</p>
    <p><strong>Aria, M., Cuccurullo, C., D'Aniello, L., Misuraca, M., & Spano, M. (2023).</strong> <i>TALL: A New Shiny App of Text Analysis for All</i>. In <i>CLiC-it</i>.</p>
  </div>

</body>
"

  ## split ----
  split <- "<body>

    <h3><strong>Splitting the Corpus</strong></h3>

    <p>TALL allows users to split textual data into smaller segments based on a specified sequence of characters. This feature is particularly useful when dealing with large documents containing multiple sections or structured content that needs to be analyzed separately.</p>
    <hr>
    <h4><strong>How It Works</strong></h4>
    <ul>
    <li>Users can define a <strong>delimiter</strong>, which is a sequence of characters used to segment the text.</li>
    <li>The delimiter must contain at least <strong>three characters</strong> to ensure accurate text splitting.</li>
    <li>The splitting process is <strong>case-sensitive</strong>, meaning that uppercase and lowercase variations are treated as distinct (e.g., <code>'CHAPTER'</code> is different from <code>'chapter'</code>).</li>
    </ul>
    <hr>
    <h4><strong>Example Use Cases</strong></h4>
    <ul>
    <li><strong>Books or Reports:</strong> Splitting a novel into chapters using <code>'CHAPTER '</code> as a delimiter.</li>
    <li><strong>Survey Responses:</strong> Separating responses when they are structured using a marker like <code>'###'</code> between answers.</li>
    <li><strong>Transcriptions:</strong> Dividing interview transcripts based on speaker labels (e.g., <code>'Speaker 1:'</code>).</li>
    </ul>

    <p>By offering a flexible splitting mechanism, TALL ensures that text segmentation aligns with the user's analytical needs, preserving the original structure for meaningful interpretation.</p>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>
      <p><strong>Aria, M., Cuccurullo, C., D'Aniello, L., Misuraca, M., &amp; Spano, M.</strong> (2024) <i>Breaking Barriers with TALL: A Text Analysis Shiny app for ALL.</i> In A. Dister, D. Longr&eacute;e (eds.), <i>Mots competes textes d&eacute;chiffr&eacute;s (JADT24)</i>, Presses Universitaires De Louvain, Vol.1, pp.39-48.</p>
    </div>

</body>"

  # random sample ----
  random <- "<body>

    <h3><strong>Random Text Selection</strong></h3>

    <p>TALL allows users to extract a random subset of imported texts for focused analysis. This feature is particularly useful when working with large corpora, enabling users to explore representative samples without processing the entire dataset.</p>
    <hr>
    <h4><strong>How It Works</strong></h4>
    <ul>
    <li>The total number of imported texts is displayed, providing an overview of the dataset size.</li>
    <li>Users can define the <strong>sample size</strong> as a percentage (%) of the total corpus.</li>
    <li>The selection process is <strong>random</strong>, ensuring an unbiased representation of the dataset.</li>
    </ul>
    <hr>
    <h4><strong>Example Use Cases</strong></h4>
    <ul>
    <li><strong>Analyzing Social Media Data:</strong> Selecting 10% of tweets from a large dataset to perform sentiment analysis.</li>
    <li><strong>Survey Research:</strong> Extracting a random subset of open-ended responses for qualitative coding.</li>
    <li><strong>Document Sampling:</strong> Reviewing a sample of reports or articles instead of analyzing the full collection.</li>
    </ul>

    <p>By enabling controlled sampling, TALL helps users balance efficiency and analytical depth, making text exploration more manageable and meaningful.</p>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>
      <p><strong>Aria, M., Cuccurullo, C., D'Aniello, L., Misuraca, M., &amp; Spano, M.</strong> (2024) <i>Breaking Barriers with TALL: A Text Analysis Shiny app for ALL.</i> In A. Dister, D. Longr&eacute;e (eds.), <i>Mots competes textes d&eacute;chiffr&eacute;s (JADT24)</i>, Presses Universitaires De Louvain, Vol.1, pp.39-48.</p>
    </div>

    </body>"

  ## external info ----
  externalinfo <- "<body>

    <h3><strong>Importing External Information</strong></h3>

    <p>TALL allows users to integrate additional information into their analysis by importing external datasets. This feature is particularly useful for enriching text data with metadata, annotations, or categorical variables, enabling a more comprehensive exploration of textual patterns.</p>
    <hr>
    <h4><strong>How to Import External Data</strong></h4>
    <ul>
    <li>The external file must be in <strong>Excel format</strong> (<code>.xlsx</code>).</li>
    <li>The dataset must include a column labeled <strong>'doc_id'</strong>, which is used to match external information with the previously imported text data.</li>
    <li>The <strong>'doc_id'</strong> values must correspond exactly to the document identifiers assigned during text import to ensure proper alignment.</li>
    </ul>
    <hr>
    <h4><strong>Using External Information</strong></h4>
    <ul>
    <li>Imported external data can be used to <strong>filter</strong> or <strong>group documents</strong> based on specific attributes (e.g., author, category, sentiment).</li>
    <li>This allows users to segment text collections efficiently, focusing on subsets relevant to their research questions.</li>
    </ul>
    <hr>
    <h4><strong>Download Document Identifiers</strong></h4>
    <p>To facilitate the integration process, users can download a list of <strong>'doc_id'</strong> values associated with the imported text files below. This ensures that external data is formatted correctly before uploading.</p>

    <p>By supporting the import of structured external data, TALL enhances text analysis capabilities, allowing users to incorporate contextual information for richer insights.</p>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>
      <p><strong>Aria, M., Cuccurullo, C., D'Aniello, L., Misuraca, M., &amp; Spano, M.</strong> (2024) <i>Breaking Barriers with TALL: A Text Analysis Shiny app for ALL.</i> In A. Dister, D. Longr&eacute;e (eds.), <i>Mots competes textes d&eacute;chiffr&eacute;s (JADT24)</i>, Presses Universitaires De Louvain, Vol.1, pp.39-48.</p>
    </div>

    </body>"

  ## tokenization ----
  tokenization <- "<body>

    <h3><strong>Tokenization, Lemmatization, and PoS Tagging</strong></h3>

    <p>TALL provides robust Natural Language Processing (NLP) capabilities for preprocessing textual data, including <strong>tokenization, lemmatization, and Part-of-Speech (PoS) tagging</strong>. These steps are essential for transforming raw text into a structured format suitable for further analysis.</p>
    <hr>
    <h4><strong>Powered by UDPipe for NLP Preprocessing</strong></h4>
    <p>TALL leverages the <strong>UDPipe</strong> library to perform tokenization, tagging, lemmatization, and dependency parsing. The <a href='https://cran.r-project.org/web/packages/udpipe/index.html' target='_blank'>udpipe R package</a> offers seamless access to pre-trained annotation models, supporting multiple languages.</p>

      <ul>
      <li><strong>Tokenization:</strong> Splits raw text into individual words or tokens.</li>
      <li><strong>Lemmatization:</strong> Converts words into their base or dictionary form (e.g., 'running' → 'run').</li>
      <li><strong>PoS Tagging:</strong> Assigns grammatical categories (e.g., noun, verb, adjective) to each word.</li>
      <li><strong>Dependency Parsing:</strong> Identifies syntactic relationships between words in a sentence.</li>
      </ul>
      <hr>
      <h4><strong>Updated Pre-trained Language Models</strong></h4>
      <p>By default, UDPipe includes models based on <a href='https://universaldependencies.org/' target='_blank'>Universal Dependencies (UD)</a> version 2.5, but these had not been updated in some time. To enhance accuracy and ensure better linguistic processing, TALL now integrates updated <strong>pre-trained NLP language models</strong> from <strong>Universal Dependencies (UD) version 2.15</strong>.</p>

        <p>These models were trained using <strong>gold standard annotated corpora</strong> from the UD project, significantly improving the quality of text analysis in TALL. The updated pre-trained models used in TALL can be accessed through our <a href='https://github.com/massimoaria/tall.language.models' target='_blank'>GitHub repository</a>.</p>
      <hr>
          <h4><strong>Applications in NLP and Text Analysis</strong></h4>
          <ul>
          <li><strong>Sentiment Analysis:</strong> Better understanding of word usage and context.</li>
          <li><strong>Topic Modeling:</strong> Improved preprocessing for cleaner topic extraction.</li>
          <li><strong>Corpus Exploration:</strong> Advanced filtering and segmentation of texts based on linguistic attributes.</li>
          </ul>

          <p>By integrating updated NLP models and leveraging powerful preprocessing techniques, TALL ensures high-quality text analysis, making it a valuable tool for researchers and practitioners in computational linguistics.</p>
         <hr>
          <div class='references'>
            <h4><strong>References</strong></h4>
            <p><strong>TALL Pre-trained Models Repository:</strong> <a href='https://github.com/massimoaria/tall.language.models' target='_blank'>GitHub repository for pre-trained models</a></p>
              <p><strong>UDPipe R Package:</strong> <a href='https://cran.r-project.org/web/packages/udpipe/index.html' target='_blank'>CRAN link to UDPipe</a></p>
                <p><strong>Universal Dependencies Repository:</strong> <a href='https://universaldependencies.org/' target='_blank'>Universal Dependencies project</a></p>
                  </div>

                  </body>"

  ## special entities ----
  specialentities <- "<body>

    <h3><strong>Tagging Special Entities</strong></h3>

    <p>TALL automatically detects and tags <strong>special entities</strong> within texts, ensuring that key non-linguistic elements are properly identified and can be leveraged in further analysis.
    <br>Recognizing these entities helps improve text preprocessing, pattern recognition, and contextual analysis.</p>
    <hr>
    <h4><strong>Detected Special Entities</strong></h4>
    <p>When processing textual data, TALL assigns specific tags to the following entities:</p>
    <ul>
    <li><strong>Email Addresses:</strong> Recognizes and tags email formats (e.g., <code>example@domain.com</code>).</li>
    <li><strong>URLs:</strong> Detects web links, ensuring they can be excluded or analyzed separately (e.g., <code>https://www.example.com/path</code>).</li>
    <li><strong>Emojis:</strong> Identifies and classifies emojis used in digital communication (e.g., 😊, 🚀, ❤️).</li>
    <li><strong>Hashtags:</strong> Extracts hashtags commonly used in social media and categorization (e.g., <code>#ExampleTag</code>).</li>
    <li><strong>IP Addresses:</strong> Detects standard IP address formats (e.g., <code>192.168.1.1</code>), which may be useful in network-related text analysis.</li>
    <li><strong>Mentions:</strong> Identifies references to usernames, particularly in social media or chat applications (e.g., <code>@username</code>).</li>
    </ul>
    <hr>
    <h4><strong>Why Special Entity Tagging Matters?</strong></h4>
    <ul>
    <li><strong>Enhanced Text Cleaning:</strong> Filtering out or isolating elements that may not contribute to linguistic analysis.</li>
    <li><strong>Social Media and Web Analysis:</strong> Extracting meaningful patterns from hashtags, mentions, and URLs.</li>
    <li><strong>Sentiment and Emotion Studies:</strong> Analyzing the role of emojis in sentiment-based communication.</li>
    <li><strong>Cybersecurity and Digital Forensics:</strong> Identifying sensitive data points such as email addresses and IP addresses.</li>
    </ul>

    <p>By integrating special entity recognition, TALL enhances the preprocessing phase, ensuring that these elements are structured for more effective text analysis.</p>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>
      <p><strong>Jurafsky, D., &amp; Martin, J.H.</strong> (2024) <i>Speech and Language Processing.</i> 3rd edition (draft). Chapter 8: Sequence Labeling for Parts of Speech and Named Entities.</p>
      <p><strong>Nadeau, D., &amp; Sekine, S.</strong> (2007) <i>A survey of named entity recognition and classification.</i> <strong>Lingvisticae Investigationes</strong>, 30(1), 3-26.</p>
    </div>

    </body>"

  ## multiword creation ----

  multiwordcreation <- "
<body>
    <h3><strong>Algorithms for Automatic Multi-Word Extraction</strong></h3>
    <p>TALL implements six methods to automatically extract multi-word expressions from a corpus. These include a <strong>syntactic approach</strong> based on dependency parsing and five <strong>statistical approaches</strong> based on co-occurrence measures.</p>

    <hr>
    <h4><strong>Syntactic Approach</strong></h4>

    <br><h4><strong>- Dependency Parsing (NP)</strong></h4>
    <p>This method extracts <strong>noun phrases</strong> by traversing the syntactic dependency tree produced during tokenization. Unlike statistical methods that rely on sequential co-occurrence patterns, this approach identifies linguistically motivated multi-word units based on their grammatical structure.</p>
    <p>The algorithm works as follows:</p>
    <ul>
      <li>For each <strong>NOUN</strong> or <strong>PROPN</strong> in the corpus, the dependency tree is traversed to find its syntactic dependents.</li>
      <li>Dependents with the following relations are collected: <strong>amod</strong> (adjective modifier, e.g., &quot;higher rate&quot;), <strong>compound</strong> (compound noun, e.g., &quot;machine learning&quot;), <strong>flat</strong> (flat multiword, e.g., &quot;New York&quot;), <strong>nmod</strong> (prepositional modifier, e.g., &quot;rate of prediabetes&quot;), and <strong>nummod</strong> (numeric modifier).</li>
      <li>For <strong>nmod</strong> dependents, the connecting preposition (<strong>case</strong> relation) is also included.</li>
      <li>The resulting tokens are sorted by position and assembled into a phrase.</li>
    </ul>
    <p><strong>Advantages over statistical methods:</strong></p>
    <ul>
      <li>Captures <strong>non-adjacent</strong> multi-word units (e.g., &quot;rate of prediabetes&quot; where &quot;of&quot; is a function word).</li>
      <li>Does not require <strong>PoS tag selection</strong> &mdash; the algorithm follows the syntactic structure automatically.</li>
      <li><strong>No false positives</strong> from coincidental co-occurrence &mdash; only grammatically related words are grouped.</li>
      <li>Implemented in <strong>C++</strong> for high performance on large corpora.</li>
    </ul>
    <p><strong>Parameters:</strong> <em>Max Phrase Length</em> (max tokens per phrase, default 5) and <em>Freq Min</em> (minimum occurrences).</p>
    <p><strong>Score:</strong> DEP = frequency &times; phrase length (longer frequent phrases score higher).</p>
    <p><strong>References:</strong><br>
    de Marneffe, M.-C., Manning, C.D., Nivre, J., &amp; Zeman, D. (2021). <em>Universal Dependencies</em>. Computational Linguistics, 47(2), 255-308.<br>
    Straka, M., &amp; Strakov&aacute;, J. (2017). <em>Tokenizing, POS Tagging, Lemmatizing and Parsing UD 2.0 with UDPipe</em>. In CoNLL Shared Task.</p>

    <hr>
    <h4><strong>Statistical Approaches</strong></h4>
    <p>The following methods identify multi-word expressions based on statistical co-occurrence patterns. They require the user to select which <strong>PoS tags</strong> to consider for candidate terms.</p>

    <br><h4><strong>- Rapid Automatic Keyword Extraction (RAKE)</strong></h4>
    <p>RAKE is a domain-independent keyword extraction algorithm that identifies key phrases by analyzing word co-occurrences within a document. It segments text into candidate keyword phrases based on stopword delimiters and then assigns scores based on word co-occurrence and frequency. Higher-scoring phrases are considered more relevant as multi-word expressions.</p>
    <p><strong>Reference:</strong><br>
    Rose, S., Engel, D., Cramer, N., &amp; Cowley, W. (2010). <em>Automatic keyword extraction from individual documents</em>. Text Mining: Applications and Theory, 1(1), 1-20.</p>

    <br><h4><strong>- Pointwise Mutual Information (PMI)</strong></h4>
    <p>PMI is a statistical measure used to assess the association strength between two words. It is defined as:</p>
    <p style='text-align: center;'>
        <em>PMI(w<sub>1</sub>, w<sub>2</sub>) = log ( P(w<sub>1</sub>, w<sub>2</sub>) / (P(w<sub>1</sub>) P(w<sub>2</sub>)) )</em>
    </p>
    <p>where P(w<sub>1</sub>, w<sub>2</sub>) is the probability of words w<sub>1</sub> and w<sub>2</sub> appearing together, and P(w<sub>1</sub>) and P(w<sub>2</sub>) are their individual probabilities. High PMI values indicate strong word associations, making the phrase a good multi-word candidate.</p>
    <p><strong>Reference:</strong><br>
    Church, K. W., &amp; Hanks, P. (1990). <em>Word association norms, mutual information, and lexicography</em>. Computational Linguistics, 16(1), 22-29.</p>

    <br><h4><strong>- Mutual Dependency (MD)</strong></h4>
    <p>Mutual Dependency extends PMI by considering the full context of a multi-word expression rather than just pairwise co-occurrence. It incorporates statistical dependency measures, ensuring that all words in a multi-word sequence contribute significantly to its overall meaning. This approach is particularly useful for identifying multi-word units beyond simple bigrams.</p>
    <p><strong>Reference:</strong><br>
    Thanopoulos, A., Fakotakis, N., &amp; Kokkinakis, G. (2002, May). <em>Comparative Evaluation of Collocation Extraction Metrics.</em> In LREC (Vol. 2, pp. 620-625).</p>

    <br><h4><strong>- Log-Frequency Biased Mutual Dependency (LF-MD)</strong></h4>
    <p>LF-MD refines the MD approach by incorporating word frequency into the dependency calculation. This method biases the selection of multi-word expressions toward frequent collocations while maintaining a balance between statistical significance and linguistic relevance.</p>
    <p><strong>Reference:</strong><br>
    Thanopoulos, A., Fakotakis, N., &amp; Kokkinakis, G. (2002, May). <em>Comparative Evaluation of Collocation Extraction Metrics.</em> In LREC (Vol. 2, pp. 620-625).</p>

    <br><h4><strong>- IS Index (Absorption Index)</strong></h4>
    <p>The IS Index, proposed by Morrone (1993), is a cohesiveness measure for word sequences that combines three key factors: word rarity, sequence frequency, and lexical density. The index is calculated as:</p>
    <p style='text-align: center;'>
        <em>IS(s) = (&Sigma; 1/freq(w<sub>i</sub>)) &times; freq(s) &times; n<sub>lexical</sub></em>
    </p>
    <p>where freq(w<sub>i</sub>) is the frequency of each word in the sequence, freq(s) is the frequency of the complete sequence, and n<sub>lexical</sub> is the number of lexical words in the sequence. The normalized version, IS<sub>norm</sub> = IS / L&sup2;, allows fair comparison between sequences of different lengths.</p>
    <p><strong>Reference:</strong><br>
    Morrone, A. (1993). <em>Alcuni criteri di valutazione della significativit&agrave; dei segmenti ripetuti</em>. In JADT (pp. 445-453).</p>

    <hr>
    <h4><strong>Choosing the Right Method</strong></h4>
    <table style='width:100%; border-collapse:collapse; margin-bottom:15px;'>
      <thead>
        <tr style='background-color:#f0f0f0; border-bottom:2px solid #ccc;'>
          <th style='padding:8px; text-align:left;'>Method</th>
          <th style='padding:8px; text-align:left;'>Basis</th>
          <th style='padding:8px; text-align:left;'>Best for</th>
          <th style='padding:8px; text-align:center;'>PoS selection</th>
        </tr>
      </thead>
      <tbody>
        <tr><td style='padding:6px;'><strong>Dep. Parsing (NP)</strong></td><td>Syntactic structure</td><td>Linguistically accurate phrases, prepositional phrases</td><td style='text-align:center;'>No</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'>RAKE</td><td>Co-occurrence + frequency</td><td>General-purpose keyword extraction</td><td style='text-align:center;'>Yes</td></tr>
        <tr><td style='padding:6px;'>PMI</td><td>Probabilistic association</td><td>Identifying strongly associated word pairs</td><td style='text-align:center;'>Yes</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'>MD / LF-MD</td><td>Mutual dependency</td><td>Balanced extraction in large corpora</td><td style='text-align:center;'>Yes</td></tr>
        <tr><td style='padding:6px;'>IS Index</td><td>Rarity + frequency + density</td><td>Terminology extraction, domain-specific terms</td><td style='text-align:center;'>No (auto)</td></tr>
      </tbody>
    </table>
</body>
"

  ## multiword list ----
  multiwordlist <- "<body>

    <h3><strong>Multi-Word Creation by a List</strong></h3>

    <p>TALL allows users to define <strong>multi-word expressions (MWEs)</strong> by importing a predefined list of multi-word terms. This feature is particularly useful for ensuring that specific phrases or domain-specific expressions are treated as single units during text processing, improving linguistic analysis.</p>
    <hr>
    <h4><strong>How to Import a Multi-Word List</strong></h4>
    <p>To integrate multi-word expressions into the analysis, users must provide a properly formatted list:</p>
    <ul>
    <li>The list must be in <strong>Excel (<code>.xlsx</code>) or CSV (<code>.csv</code>) format</strong>.</li>
    <li>The file should contain a <strong>single column</strong> where each row represents one multi-word expression.</li>
    <li><strong>Each term within a multi-word expression must be separated by a single whitespace<br></strong> (e.g., <code>machine learning</code>, <code>natural language processing</code>).</li>
    </ul>
    <hr>
    <h4><strong>Why Use Multi-Word Expressions?</strong></h4>
    <ul>
    <li><strong>Preserving Meaningful Phrases:</strong> Ensuring that key terms (e.g., <code>artificial intelligence</code>) are not split into separate words.</li>
    <li><strong>Improving Text Preprocessing:</strong> Enhancing tokenization and lemmatization by treating phrases as cohesive units.</li>
    <li><strong>Enhancing Domain-Specific Analysis:</strong> Beneficial in specialized fields such as legal, medical, or technical texts, where multi-word terms have precise meanings.</li>
    </ul>

    <p>By supporting multi-word recognition, TALL provides users with greater flexibility in structuring their text analysis and ensures that critical expressions are accurately identified and processed.</p>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>
      <p><strong>Sag, I.A., Baldwin, T., Bond, F., Copestake, A., &amp; Flickinger, D.</strong> (2002) <i>Multiword expressions: A pain in the neck for NLP.</i> In <strong>Proceedings of the Third International Conference on Intelligent Text Processing and Computational Linguistics (CICLing)</strong>, 1-15.</p>
    </div>

    </body>"

  ## custom pos ----
  customterm <- "<body>

    <h3><strong>Custom PoS List</strong></h3>

    <p>TALL allows users to define a <strong>Custom PoS List</strong>, enabling more precise control over text processing and linguistic analysis. This feature allows users to manually assign custom PoS tags to specific terms, overriding their default categorization by the language model.</p>
    <hr>
    <h4><strong>Why Use a Custom PoS List?</strong></h4>
    <ul>
    <li><strong>Highlighting Specific Concepts:</strong> Identifying key terms related to methodologies, specialized vocabulary, or domain-specific jargon.</li>
    <li><strong>Filtering Stop Words:</strong> Removing terms that are irrelevant to the analysis, ensuring a cleaner dataset.</li>
    <li><strong>Enhancing Named Entity Recognition (NER):</strong> Manually tagging specific words that the language model may misclassify.</li>
    <li><strong>Overriding Default PoS Assignments:</strong> Ensuring consistency in tagging across texts by defining a fixed categorization for certain terms.</li>
    </ul>
    <hr>
    <h4><strong>How to Import a Custom PoS List</strong></h4>
    <p>To integrate a custom list of terms, users must provide a properly formatted file:</p>
    <ul>
    <li>The list must be in <strong>Excel format (<code>.xlsx</code>)</strong>.</li>
    <li>The file should contain <strong>two columns</strong>:</li>
    <ul>
    <li><strong>First column:</strong> The list of terms to be tagged.</li>
    <li><strong>Second column:</strong> The corresponding Part-of-Speech (PoS) or user-defined category assigned to each term.</li>
    </ul>
    <li>The tags can be <strong>any custom category</strong> you define for your analysis needs (e.g., METHOD, APPLICATION, SETTING, TO_REMOVE, etc.), or standard PoS tags (NOUN, VERB, ADJ, etc.).</li>
    </ul>
    <p><strong>Note:</strong> Custom tags will appear in the PoS selection menus throughout the app, allowing you to filter, group, or exclude terms based on your custom categories.</p>
    <hr>
  <h4><strong>Example of Custom PoS List Format</strong></h4>

  <table border='1' cellspacing='0' cellpadding='5'>
  <tr>
  <th> -------- Term ---------- </th>
  <th> ------ Custom Tag ------ </th>
  </tr>
  <tr>
  <td>artificial intelligence</td>
  <td>METHOD</td>
  </tr>
  <tr>
  <td>deep learning</td>
  <td>METHOD</td>
  </tr>
  <tr>
  <td>preprocess</td>
  <td>DATA_HANDLING</td>
  </tr>
  <tr>
  <td>dataset</td>
  <td>DATA_HANDLING</td>
  </tr>
  <tr>
  <td>remove</td>
  <td>TO_REMOVE</td>
  </tr>
  <tr>
  <td>neural network</td>
  <td>APPLICATION</td>
  </tr>
  <tr>
  <td>batch size</td>
  <td>SETTING</td>
  </tr>
  </table>
<br>
    <p>By allowing users to define and control term tagging with fully custom categories, TALL provides enhanced flexibility for text analysis, making it a powerful tool for domain-specific research and refined linguistic processing.</p>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>
      <p><strong>Petrov, S., Das, D., &amp; McDonald, R.</strong> (2012) <i>A universal part-of-speech tagset.</i> In <strong>Proceedings of the Eighth International Conference on Language Resources and Evaluation (LREC)</strong>, Istanbul, Turkey.</p>
      <p><strong>de Marneffe, M.-C., Manning, C.D., Nivre, J., &amp; Zeman, D.</strong> (2021) <i>Universal Dependencies.</i> <strong>Computational Linguistics</strong>, 47(2), 255-308.</p>
    </div>

    </body>"

  ## synonyms ----
  synonyms <- "<body>

    <h3><strong>Synonyms Merging</strong></h3>

    <p>TALL allows users to merge <strong>synonyms and variant forms</strong> into standardized target terms. This feature is essential for reducing vocabulary noise, consolidating spelling variants, abbreviations, and domain-specific synonyms into a single canonical form, thereby improving the quality of downstream analyses such as topic modeling, co-occurrence networks, and frequency distributions.</p>
    <hr>
    <h4><strong>File Format Requirements</strong></h4>
    <p>The synonyms file must be in <strong>CSV</strong> or <strong>XLSX</strong> format with the following structure:</p>
    <ul>
      <li><strong>Column 1 (target_term):</strong> The standardized term that will replace all synonyms.</li>
      <li><strong>Column 2 (upos):</strong> The Part-of-Speech tag to assign to the target term (e.g., NOUN, VERB, ADJ).</li>
      <li><strong>Columns 3+ (synonym1, synonym2, ...):</strong> Alternative terms to be replaced.</li>
    </ul>
    <hr>
    <h4><strong>Example File Structure</strong></h4>

    <table style='width:100%; border-collapse:collapse; margin-bottom:15px;'>
      <thead>
        <tr style='background-color:#f0f0f0; border-bottom:2px solid #ccc;'>
          <th style='padding:8px; text-align:left;'>target_term</th>
          <th style='padding:8px; text-align:left;'>upos</th>
          <th style='padding:8px; text-align:left;'>synonym1</th>
          <th style='padding:8px; text-align:left;'>synonym2</th>
          <th style='padding:8px; text-align:left;'>synonym3</th>
        </tr>
      </thead>
      <tbody>
        <tr><td style='padding:6px;'>machine_learning</td><td>NOUN</td><td>ml</td><td>ML</td><td>machine learning</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'>artificial_intelligence</td><td>NOUN</td><td>ai</td><td>AI</td><td>A.I.</td></tr>
      </tbody>
    </table>

    <hr>
    <h4><strong>How It Works</strong></h4>
    <ol>
      <li><strong>Upload</strong> your synonyms file (CSV or Excel format).</li>
      <li><strong>Select the replacement type:</strong> Choose whether to match against <em>tokens</em> or <em>lemmas</em>.</li>
      <li><strong>Preview</strong> your synonyms list in the preview tab before applying.</li>
      <li><strong>Apply:</strong> Click Run to replace all synonyms with their target terms.</li>
      <li><strong>Review</strong> the processed data in the results tab.</li>
    </ol>
    <hr>
    <h4><strong>Processing Details</strong></h4>
    <ul>
      <li><strong>Token-based replacement:</strong> Searches and replaces synonym tokens, updating their <code>upos</code> tag.</li>
      <li><strong>Lemma-based replacement:</strong> Searches and replaces synonym lemmas, updating their <code>upos</code> tag.</li>
      <li><strong>Case-insensitive matching:</strong> Matching ignores case differences.</li>
      <li><strong>PoS tag update:</strong> When a synonym is replaced, its Part-of-Speech tag is also updated to the specified value.</li>
    </ul>
    <hr>
    <h4><strong>Important Notes</strong></h4>
    <ul>
      <li>The merging process is <strong>irreversible</strong> after saving. Always verify your synonyms list in the preview tab before applying.</li>
      <li>Ensure <code>upos</code> values correspond to valid Universal PoS tags (NOUN, VERB, ADJ, ADV, PROPN, PRON, DET, ADP, NUM, CONJ, INTJ).</li>
      <li>Empty cells in synonym columns are ignored.</li>
    </ul>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>
      <p><strong>Jurafsky, D., &amp; Martin, J.H.</strong> (2024) <i>Speech and Language Processing.</i> 3rd edition (draft). Chapter 6: Vector Semantics and Embeddings.</p>
      <p><strong>Manning, C.D., Raghavan, P., &amp; Sch&uuml;tze, H.</strong> (2008) <i>Introduction to Information Retrieval.</i> Cambridge University Press. Chapter 2: The term vocabulary and postings lists.</p>
    </div>

    </body>"

  ## pos selection ----
  posselection <- "<body>

    <h3><strong>PoS Tagging Selection</strong></h3>

    <p>TALL provides users with the flexibility to <strong>select specific Part-of-Speech (PoS) tags</strong> to be used in subsequent analyses.
    This feature allows for greater control over the linguistic elements included in text processing,
    ensuring that only relevant grammatical categories are considered.</p>
    <hr>
    <h4><strong>Why Select PoS Tags?</strong></h4>
    <ul>
    <li><strong>Filtering Out Unnecessary Elements:</strong> Excluding determiners, conjunctions, or punctuation that may not contribute to the analysis.</li>
    <li><strong>Focusing on Key Linguistic Features:</strong> Selecting only nouns and verbs for topic modeling, or adjectives and adverbs for sentiment analysis.</li>
    <li><strong>Improving Computational Efficiency:</strong> Reducing data size and processing time by analyzing only the most relevant word categories.</li>
    </ul>
    <hr>
    <h4><strong>How It Works</strong></h4>
    <ul>
    <li>Users can manually <strong>select or deselect</strong> PoS categories from a predefined list.</li>
    <li>The available PoS tags follow the <strong>Universal Dependencies (UD) annotation scheme</strong>, ensuring consistency across different languages.</li>
    </ul>
    <hr>
    <h4><strong>Default Selected PoS Tags</strong></h4>
    <p>By default, TALL selects the following PoS categories:</p>
    <ul>
    <li><strong>ADJ:</strong> Adjective – Descriptive words (e.g., 'beautiful', 'quick').</li>
    <li><strong>NOUN:</strong> Noun – Common nouns representing entities (e.g., 'dog', 'city').</li>
    <li><strong>PROPN:</strong> Proper Noun – Specific names of places, people, or organizations (e.g., 'London', 'NASA').</li>
    <li><strong>VERB:</strong> Verb – Action words representing processes (e.g., 'run', 'speak').</li>
    <li><strong>HAPAX:</strong> Words appearing only once in the text, useful for lexical richness analysis.</li>
    </ul>
    <hr>
    <h4><strong>Available PoS Categories</strong></h4>
    <table border='1' cellspacing='0' cellpadding='5'>
    <tr>
    <th>PoS Tag</th>
    <th>Description</th>
    </tr>
    <tr><td>ADJ</td><td>Adjective</td></tr>
    <tr><td>ADP</td><td>Adposition</td></tr>
    <tr><td>ADV</td><td>Adverb</td></tr>
    <tr><td>AUX</td><td>Auxiliary</td></tr>
    <tr><td>CCONJ</td><td>Coordinating Conjunction</td></tr>
    <tr><td>DET</td><td>Determiner</td></tr>
    <tr><td>INTJ</td><td>Interjection</td></tr>
    <tr><td>NOUN</td><td>Noun</td></tr>
    <tr><td>NUM</td><td>Numeral</td></tr>
    <tr><td>PART</td><td>Particle</td></tr>
    <tr><td>PRON</td><td>Pronoun</td></tr>
    <tr><td>PROPN</td><td>Proper Noun</td></tr>
    <tr><td>PUNCT</td><td>Punctuation</td></tr>
    <tr><td>SCONJ</td><td>Subordinating Conjunction</td></tr>
    <tr><td>SYM</td><td>Symbol</td></tr>
    <tr><td>VERB</td><td>Verb</td></tr>
    <tr><td>X</td><td>Other</td></tr>
    <tr><td>Hapax</td><td>Words appearing only once in the corpus</td></tr>
    <tr><td>Single Character</td><td>Individual symbols or characters</td></tr>
    </table>
    <hr>
    <h4><strong>Custom Categories</strong></h4>
    <p>In addition to predefined PoS categories, users may have also <strong>generated custom categories</strong> through the <strong>Custom List</strong> and <strong>Multi-Word</strong> menus.
    <br>These user-defined tags allow for specialized analysis by grouping specific terms under a unique classification system.</p>
    <hr>
    <h4><strong>Enhancing Analysis with PoS Selection</strong></h4>
    <p>By allowing users to choose specific PoS categories, TALL ensures that the analysis is tailored to the user's research goals.
    <br>Whether performing <strong>keyword extraction, syntactic analysis, topic modeling, or sentiment analysis</strong>,
    the ability to refine PoS selection enhances the precision and interpretability of results.</p>

</body>"

  ## features ----
  featureroles <- "<body>

  <h3><strong>Feature Roles</strong></h3>

  <p>The Feature Roles section allows users to assign specific roles to variables in their dataset. These role assignments enable sophisticated text analysis workflows across various TALL features, ensuring that the appropriate variables are used for different analytical purposes.</p>

  <hr>

  <h4><strong>Purpose</strong></h4>

  <p>By explicitly defining feature roles, TALL can automatically configure and optimize analysis parameters based on the characteristics of your data. This structured approach ensures consistency across different analytical modules and reduces the need for repetitive variable selection.</p>

  <hr>

  <h4><strong>Available Feature Roles</strong></h4>

  <h5><i class='fas fa-clock' style='color: #3f51b5;'></i> <strong>Time Variable</strong></h5>

  <p>A numeric or date variable that serves as a temporal indicator for diachronic text analysis.</p>

  <p><strong>Primary Applications:</strong></p>
  <ul>
    <li><strong>Longitudinal Topic Modeling:</strong> Track how topics evolve over time periods</li>
    <li><strong>Temporal Trend Analysis:</strong> Identify patterns and shifts in vocabulary usage across different time frames</li>
    <li><strong>Time-Series Text Mining:</strong> Analyze textual data with temporal dependencies</li>
    <li><strong>Diachronic Linguistic Studies:</strong> Examine language change and evolution</li>
  </ul>

  <p><strong>Requirements:</strong> The variable must be numeric (e.g., year, month number) or in date format (e.g., YYYY-MM-DD).</p>

  <hr>

  <h5><i class='fas fa-tag' style='color: #00bcd4;'></i> <strong>Label Variable</strong></h5>

  <p>A categorical variable representing the response or target class in supervised text classification tasks.</p>

  <p><strong>Primary Applications:</strong></p>
  <ul>
    <li><strong>Supervised Machine Learning:</strong> Train classification models such as Random Forest, Support Vector Machines (SVM), and Naive Bayes</li>
    <li><strong>Text Categorization:</strong> Automatically assign documents to predefined categories</li>
    <li><strong>Sentiment Classification:</strong> Predict sentiment labels (positive, negative, neutral)</li>
    <li><strong>Document Classification:</strong> Classify documents based on topic, genre, or other categorical attributes</li>
    <li><strong>Model Evaluation:</strong> Validate classification performance using labeled data</li>
  </ul>

  <p><strong>Requirements:</strong> The variable should contain discrete categorical values. For binary classification, two distinct categories are needed. For multi-class classification, multiple categories can be used.</p>

  <hr>

  <h5><i class='fas fa-right-left' style='color: #ff9800;'></i> <strong>Keyness Group Variable</strong></h5>

  <p>A binary or categorical variable that divides the corpus into distinct groups for comparative analysis.</p>

  <p><strong>Primary Applications:</strong></p>
  <ul>
    <li><strong>Keyness Analysis:</strong> Identify words and phrases that are statistically more characteristic of one group compared to another</li>
    <li><strong>Comparative Corpus Linguistics:</strong> Compare vocabulary usage between different subcorpora</li>
    <li><strong>Distinctive Vocabulary Identification:</strong> Discover words that differentiate groups</li>
    <li><strong>Contrastive Analysis:</strong> Examine linguistic differences between categories (e.g., male vs. female authors, different time periods, geographic regions)</li>
  </ul>

  <p><strong>Requirements:</strong> The variable should ideally contain two distinct categories for binary comparison, though categorical variables with multiple groups can also be used (with pairwise comparisons).</p>

  <hr>

  <h4><strong>Usage Guidelines</strong></h4>

  <ul>
    <li><strong>Available Features:</strong> Only metadata columns are available for role assignment. Technical columns generated during text processing (e.g., <code>token_id</code>, <code>lemma</code>, <code>upos</code>) are automatically excluded by the <code>noGroupLabels()</code> function.</li>
    <li><strong>Multiple Role Assignment:</strong> A single variable can be assigned to multiple roles if appropriate for your analysis workflow.</li>
    <li><strong>Session Persistence:</strong> Role assignments are maintained throughout your TALL session until explicitly changed or reset.</li>
    <li><strong>Flexible Configuration:</strong> Roles can be modified at any time to accommodate different analytical needs.</li>
    <li><strong>No Mandatory Assignments:</strong> Not all roles need to be assigned. Only configure the roles relevant to your specific analysis objectives.</li>
  </ul>

  <hr>

  <h4><strong>Best Practices</strong></h4>

  <ul>
    <li><strong>Data Quality:</strong> Ensure that selected variables contain valid, non-missing data appropriate for their assigned role.</li>
    <li><strong>Temporal Consistency:</strong> For time variables, verify that temporal values are consistent and properly formatted.</li>
    <li><strong>Balanced Labels:</strong> For label variables in classification tasks, consider class balance to avoid biased model performance.</li>
    <li><strong>Clear Group Definitions:</strong> For keyness analysis, ensure group categories are well-defined and meaningful for comparison.</li>
    <li><strong>Documentation:</strong> Keep track of which variables are assigned to which roles, especially in complex analytical workflows.</li>
  </ul>

  <hr>

  <h4><strong>Technical Notes</strong></h4>

  <p>Feature role assignments are stored in the reactive values object (<code>values$timeVariable</code>, <code>values$labelVariable</code>, <code>values$keynessVariable</code>) and can be accessed programmatically throughout the TALL application. These assignments inform downstream analytical functions about the appropriate variables to use for specific tasks.</p>

</body>"

  ## overview ----
  overview <- "<body>
  <div class='container'>
    <h3><strong>Corpus Metrics</strong></h3>
    <p>These metrics provide a summary of the key textual characteristics of the analyzed corpus.</p>

    <h4><strong>Corpus Size &amp; Structure</strong></h4>
    <ul>
    <li><strong>Documents →</strong> The total number of documents in the corpus.</li>
    <li><strong>Sentences →</strong> The total number of sentences in the corpus.</li>
    <li><strong>Tokens →</strong> The total number of words or linguistic units, including punctuation marks.</li>
    <li><strong>Types →</strong> The number of unique words in the corpus, representing vocabulary richness.</li>
    <li><strong>Lemma →</strong> The number of unique lemmas, considering the base form of words.</li>
    </ul>

    <h4><strong>Average Length Metrics</strong></h4>
    <ul>
    <li><strong>Doc Avg Length in Chars →</strong> The average number of characters per document.<br>
    <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
      <mrow>
      <mfrac>
      <mtext>Total Characters</mtext>
      <mtext>Number of Documents</mtext>
      </mfrac>
      </mrow>
      </math>
      </li>
      <li><strong>Doc Avg Length in Tokens →</strong> The average number of tokens per document.<br>
      <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
        <mrow>
        <mfrac>
        <mtext>Total Tokens</mtext>
        <mtext>Number of Documents</mtext>
        </mfrac>
        </mrow>
        </math>
        </li>
        <li><strong>Sent Avg Length in Chars →</strong> The average number of characters per sentence.<br>
        <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
          <mrow>
          <mfrac>
          <mtext>Total Characters</mtext>
          <mtext>Number of Sentences</mtext>
          </mfrac>
          </mrow>
          </math>
          </li>
          <li><strong>Sent Avg Length in Tokens →</strong> The average number of tokens per sentence.<br>
          <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
            <mrow>
            <mfrac>
            <mtext>Total Tokens</mtext>
            <mtext>Number of Sentences</mtext>
            </mfrac>
            </mrow>
            </math>
            </li>
            </ul>

            <h4><strong>Lexical Metrics</strong></h4>
            <ul>
            <li><strong>Type-Token Ratio (TTR) →</strong> Ratio of unique words (types) to total words (tokens). Higher values indicate greater lexical diversity.<br>
            <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
              <mrow>
              <mtext>TTR = </mtext>
                <mfrac>
                <mtext>Types</mtext>
                <mtext>Tokens</mtext>
                </mfrac>
                </mrow>
                </math>
                </li>

                <li><strong>Hapax Legomena (%) →</strong> Percentage of words that appear only once in the corpus.<br>
                <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
                  <mrow>
                  <mtext>Hapax % = </mtext>
                    <mfrac>
                    <mtext>Hapax</mtext>
                    <mtext>Types</mtext>
                    </mfrac>
                    <mo>×</mo>
                    <mn>100</mn>
                    </mrow>
                    </math>
                    </li>

                    <li><strong>Guiraud Index →</strong> Measure of lexical richness correcting for text length.<br>
                    <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
                      <mrow>
                      <mtext>Guiraud = </mtext>
                        <mfrac>
                        <mtext>Types</mtext>
                        <msqrt>
                        <mtext>Tokens</mtext>
                        </msqrt>
                        </mfrac>
                        </mrow>
                        </math>
                        </li>
                        </ul>

                        <h4><strong>Additional Lexical Measures</strong></h4>
                        <ul>
                        <li><strong>Lexical Density →</strong> Proportion of content words over total tokens.<br>
                        <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
                          <mrow>
                          <mtext>Lexical Density = </mtext>
                            <mfrac>
                            <mtext>Content Words</mtext>
                            <mtext>Total Tokens</mtext>
                            </mfrac>
                            </mrow>
                            </math>
                            </li>

                            <li><strong>Nominal Ratio →</strong> Ratio between nouns and verbs.<br>
                            <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
                              <mrow>
                              <mtext>Nominal Ratio = </mtext>
                                <mfrac>
                                <mtext>Number of Nouns</mtext>
                                <mtext>Number of Verbs</mtext>
                                </mfrac>
                                </mrow>
                                </math>
                                </li>

                                <li><strong>Gini Index →</strong> Measure of inequality in word frequency distribution. Calculated from the Lorenz curve of word frequencies.</li>

                                <li><strong>Yule’s K Index →</strong> Measure of lexical diversity based on word repetition.<br>
                                <math xmlns='http://www.w3.org/1998/Math/MathML' style='font-size: 1.5em; display: block; text-align: center; margin-top: 4px; margin-bottom: 12px;'>
                                  <mrow>
                                  <mtext>K = </mtext>
                                    <mn>10,000</mn>
                                    <mo>×</mo>
                                    <mfrac>
                                    <mrow>
                                    <mo>(</mo>
                                           <munderover>
                                           <mo>&#x2211;</mo>
                                           <mi>i</mi>
                                           <mi>n</mi>
                                           </munderover>
                                           <msup><mi>f</mi><mn>2</mn></msup>
                                           <mo>)</mo>
                                    <mo>-</mo>
                                    <mi>N</mi>
                                    </mrow>
                                    <msup><mi>N</mi><mn>2</mn></msup>
                                    </mfrac>
                                    </mrow>
                                    </math>
                                    </li>
                                    </ul>

    <h4><strong>Morphological Features</strong></h4>
    <p>The <strong>Morphological Features</strong> tab displays the distribution of grammatical features extracted from the <code>feats</code> column of the Universal Dependencies annotation. Each token is annotated with morphological properties that reveal grammatical patterns in the corpus.</p>

    <h4><em>Available Features</em></h4>
    <table style='width:100%; border-collapse:collapse; margin-bottom:15px;'>
      <thead>
        <tr style='background-color:#f0f0f0; border-bottom:2px solid #ccc;'>
          <th style='padding:8px; text-align:left;'>Feature</th>
          <th style='padding:8px; text-align:left;'>Values</th>
          <th style='padding:8px; text-align:left;'>Applies to</th>
          <th style='padding:8px; text-align:left;'>What it reveals</th>
        </tr>
      </thead>
      <tbody>
        <tr><td style='padding:6px;'><strong>Tense</strong></td><td>Past, Pres, Fut</td><td>VERB, AUX</td><td>Narrative (Past) vs argumentative (Pres) style</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><strong>Mood</strong></td><td>Ind, Sub, Imp, Cnd</td><td>VERB, AUX</td><td>Certainty (Ind) vs hedging/doubt (Sub, Cnd)</td></tr>
        <tr><td style='padding:6px;'><strong>Number</strong></td><td>Sing, Plur</td><td>NOUN, PRON, DET, ADJ, VERB</td><td>Generalization patterns</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><strong>Person</strong></td><td>1, 2, 3</td><td>VERB, AUX, PRON</td><td>Personal (1st) vs impersonal (3rd) style</td></tr>
        <tr><td style='padding:6px;'><strong>VerbForm</strong></td><td>Fin, Inf, Part, Ger, Conv</td><td>VERB, AUX</td><td>Finite vs non-finite verb usage</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><strong>Degree</strong></td><td>Pos, Cmp, Sup</td><td>ADJ, ADV</td><td>Use of comparatives and superlatives</td></tr>
        <tr><td style='padding:6px;'><strong>Gender</strong></td><td>Masc, Fem, Neut</td><td>NOUN, PRON, ADJ, DET</td><td>Grammatical gender distribution (language-dependent)</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><strong>Case</strong></td><td>Nom, Acc, Gen, Dat, ...</td><td>NOUN, PRON, ADJ</td><td>Syntactic roles (language-dependent)</td></tr>
        <tr><td style='padding:6px;'><strong>Voice</strong></td><td>Act, Pass</td><td>VERB</td><td>Active vs passive constructions</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><strong>Definite</strong></td><td>Def, Ind</td><td>DET, NOUN</td><td>Definiteness marking</td></tr>
        <tr><td style='padding:6px;'><strong>PronType</strong></td><td>Prs, Dem, Rel, Int, ...</td><td>PRON, DET</td><td>Pronoun/determiner types</td></tr>
      </tbody>
    </table>

    <p>The <strong>bar chart</strong> shows the frequency of each value for the selected feature. The <strong>cross-tabulation table</strong> shows how each feature value distributes across PoS tags (e.g., Tense=Past appears in both VERB and AUX).</p>

    <p><strong>Note:</strong> Available features depend on the language model used for tokenization. Not all features are annotated in every language.</p>

    <hr>
    <h4><strong>Dependency Tree Viewer</strong></h4>
    <p>The <strong>Dependency Tree</strong> tab provides an interactive visualization of the syntactic structure of individual sentences. Users can select any document and sentence in the corpus to inspect how the parser has analyzed its grammatical structure.</p>

    <h4><em>How to Read the Tree</em></h4>
    <ul>
      <li><strong>Tokens</strong> are displayed horizontally in sentence order, shown as labeled boxes colored by Part of Speech:
        <span style='color:#4F7942;'>NOUN/PROPN (green)</span>,
        <span style='color:#E41A1C;'>VERB/AUX (red)</span>,
        <span style='color:#377EB8;'>ADJ (blue)</span>,
        <span style='color:#FF7F00;'>ADV (orange)</span>,
        <span style='color:#984EA3;'>PRON (purple)</span>,
        <span style='color:#999999;'>DET/ADP (gray)</span>.</li>
      <li><strong>Arcs</strong> connect each token to its syntactic head. The arc height is proportional to the distance between the dependent and the head &mdash; longer arcs indicate more distant dependencies.</li>
      <li><strong>Arc labels</strong> show the dependency relation (e.g., <code>nsubj</code>, <code>obj</code>, <code>amod</code>).</li>
      <li><strong>Arc colors</strong> indicate the relation type:
        <span style='color:#E41A1C;'>subject relations (red)</span>,
        <span style='color:#377EB8;'>object relations (blue)</span>,
        <span style='color:#4DAF4A;'>nominal modifiers (green)</span>,
        <span style='color:#FF7F00;'>adverbial modifiers (orange)</span>,
        <span style='color:#984EA3;'>compounds (purple)</span>.</li>
    </ul>

    <h4><em>Token Details Table</em></h4>
    <p>Below the tree, a table shows the full annotation for each token: ID, Token, Lemma, PoS tag, Dependency Relation, Head ID, and Head Token. This allows precise inspection of the parse output.</p>

    <h4><em>Common Dependency Relations</em></h4>
    <table style='width:100%; border-collapse:collapse; margin-bottom:15px;'>
      <thead>
        <tr style='background-color:#f0f0f0; border-bottom:2px solid #ccc;'>
          <th style='padding:8px; text-align:left;'>Relation</th>
          <th style='padding:8px; text-align:left;'>Meaning</th>
          <th style='padding:8px; text-align:left;'>Example</th>
        </tr>
      </thead>
      <tbody>
        <tr><td style='padding:6px;'><code>nsubj</code></td><td>Nominal subject</td><td><em>The <u>cat</u> sat</em> (cat &rarr; sat)</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><code>obj</code></td><td>Direct object</td><td><em>She read a <u>book</u></em> (book &rarr; read)</td></tr>
        <tr><td style='padding:6px;'><code>amod</code></td><td>Adjective modifier</td><td><em>a <u>big</u> house</em> (big &rarr; house)</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><code>nmod</code></td><td>Nominal modifier (via preposition)</td><td><em>rate of <u>growth</u></em> (growth &rarr; rate)</td></tr>
        <tr><td style='padding:6px;'><code>advmod</code></td><td>Adverbial modifier</td><td><em>runs <u>quickly</u></em> (quickly &rarr; runs)</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><code>compound</code></td><td>Compound word</td><td><em><u>machine</u> learning</em> (machine &rarr; learning)</td></tr>
        <tr><td style='padding:6px;'><code>det</code></td><td>Determiner</td><td><em><u>the</u> house</em> (the &rarr; house)</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><code>case</code></td><td>Preposition/postposition</td><td><em>rate <u>of</u> growth</em> (of &rarr; growth)</td></tr>
        <tr><td style='padding:6px;'><code>conj</code></td><td>Conjunct</td><td><em>cats <u>and</u> dogs</em> (dogs &rarr; cats)</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><code>advcl</code></td><td>Adverbial clause</td><td><em>He left <u>because</u> it rained</em></td></tr>
        <tr><td style='padding:6px;'><code>acl</code></td><td>Adnominal clause</td><td><em>the man <u>who</u> came</em></td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'><code>xcomp</code></td><td>Open clausal complement</td><td><em>I want <u>to go</u></em></td></tr>
      </tbody>
    </table>

                                    <div class='references'>
                                      <h4><strong>References</strong></h4>
                                      <p><strong>Baayen, R. H.</strong> <i>The effect of lexical specialization on the growth curve of vocabulary.</i> <strong>Computational Linguistics</strong>, 22(2), 1996.</p>
                                      <p><strong>Bentz, C., Alikaniotis, D., Cysouw, M., &amp; Ferrer-i-Cancho, R.</strong> <i>The entropy of words &mdash; learnability and expressivity across more than 1000 languages.</i> <strong>Entropy</strong>, 19(6), 2017.</p>
                                      <p><strong>Biber, D.</strong> <i>Variation across speech and writing.</i> <strong>Cambridge University Press</strong>, 1988.</p>
                                      <p><strong>de Marneffe, M.-C., Manning, C.D., Nivre, J., &amp; Zeman, D.</strong> (2021) <i>Universal Dependencies.</i> <strong>Computational Linguistics</strong>, 47(2), 255-308.</p>
                                      <p><strong>Guiraud, P.</strong> <i>Les caract&egrave;res statistiques du vocabulaire.</i> <strong>Presse Universitaire de France</strong>, 1954.</p>
                                      <p><strong>Tweedie, F. J., &amp; Baayen, R. H.</strong> <i>How variable may a constant be? Measures of lexical richness in perspective.</i> <strong>Computers and the Humanities</strong>, 32(5), 323-352, 1998.</p>
                                      <p><strong>Ure, J.</strong> <i>Lexical density and register differentiation.</i> In G. Perren and J.L.M. Trim (eds), <strong>Applications of Linguistics</strong>, Cambridge University Press, 443-452, 1971.</p>
                                      <p><strong>Yule, G. U.</strong> <i>The statistical study of literary vocabulary.</i> <strong>Cambridge University Press</strong>, 1944.</p>
                                      <p><strong>Zeman, D.</strong> (2017) <i>Universal Dependencies v2: An evergrowing multilingual treebank collection.</i> In <strong>Proceedings of LREC</strong>.</p>
                                      </div>
                                      </div>
                                      </body>"

  ## word in context ----
  wordincontext <- "<body>

    <h3 style='color: #4F7942;'><strong>Words in Context</strong></h3>

    <p>The <strong>Words in Context</strong> feature in TALL allows users to analyze how specific words appear in textual data, offering valuable insights into <strong>semantic usage, contextual meaning, and discourse structure</strong>. This tool is particularly useful for <strong>qualitative text analysis, linguistic research, and content exploration</strong> in diverse domains, such as <strong>social sciences, digital humanities, marketing, and legal studies</strong>.</p>
    <hr>
    <h4 style='color: #4F7942;'><strong>How Words in Context Works</strong></h4>

    <h4 style='color: #4F7942;'><strong>1. Concordance Analysis (Keyword in Context - KWIC)</strong></h4>
    <ul>
    <li>Displays a <strong>side-by-side view of words and their surrounding textual context</strong> (left and right neighbors).</li>
    <li>Helps in identifying <strong>common phrases, recurring structures, and usage variations</strong>.</li>
    <li>Useful for <strong>studying semantic shifts, idiomatic expressions, and collocations</strong>.</li>
    </ul>

    <div class='example'>
      <strong>Example:</strong><br>
      If analyzing the term <strong>'sustainable'</strong> in a corpus of news articles, KWIC might show:<br>
      - 'sustainable <strong>development</strong> is a key focus of international policies'<br>
      - 'the company promotes <strong>sustainable</strong> and ethical supply chains'<br>
      - 'concerns over <strong>sustainable</strong> agricultural practices are increasing'<br>
      This helps in understanding <strong>how 'sustainable' is used in different thematic contexts</strong>.
    </div>

      <h4 style='color: #4F7942;'><strong>2. Context Window Customization</strong></h4>
      <ul>
      <li>Users can define the <strong>window size</strong> (number of words before and after the target term) to adjust the level of contextual information displayed.</li>
      <li>Shorter windows highlight <strong>immediate linguistic relationships</strong>, while larger windows help analyze <strong>broader semantic dependencies</strong>.</li>
      </ul>

      <div class='example'>
        <strong>Example:</strong><br>
        When studying <strong>'risk'</strong> in financial reports, adjusting the window size allows users to see if it is used in association with:<br>
        - <strong>'risk management,' 'high-risk investments'</strong> (short window)<br>
        - <strong>'the recent economic downturn has increased financial risk for small businesses'</strong> (larger window)<br>
        </div>

        <h4 style='color: #4F7942;'><strong>3. Frequency and Distribution Insights</strong></h4>
        <ul>
        <li>Words appearing in <strong>multiple contexts</strong> can be analyzed for <strong>frequency trends</strong>, helping users identify <strong>dominant themes</strong> associated with a term.</li>
        <li>Examines whether a word is <strong>evenly distributed</strong> across the corpus or <strong>clustered</strong> in specific sections/documents.</li>
        </ul>

        <div class='example'>
          <strong>Example:</strong><br>
          In a dataset of <strong>customer reviews</strong>, the word <strong>'expensive'</strong> might frequently co-occur with:<br>
          - <strong>'but worth it'</strong> in <strong>positive reviews</strong><br>
          - <strong>'not justified for the quality'</strong> in <strong>negative reviews</strong><br>
          This helps distinguish <strong>when 'expensive' has a neutral, positive, or negative connotation</strong>.
        </div>

          <p>By enabling <strong>customizable and interactive text exploration</strong>, the <strong>Words in Context</strong> tool in TALL provides users with <strong>a deeper understanding of language patterns</strong> in large textual datasets.</p>


          </body>"

  ## keyness ----
  keyness <- "<body>

    <h3><strong>Keyness Analysis</strong></h3>

    <p>Keyness analysis is a statistical technique used to identify words that are <strong>significantly more or less frequent</strong> in a target corpus compared to a reference corpus. This method enables researchers to <strong>detect distinctive vocabulary</strong> and <strong>linguistic features</strong> that characterize specific texts, genres, or discourse communities (<strong>Scott, 1997; Gabrielatos, 2018</strong>).</p>

    <p>In TALL, keyness is computed by comparing the word frequencies in your corpus against <strong>reference word frequency lists</strong> derived from large, general-purpose language datasets. This allows users to identify words that are <strong>overused</strong> or <strong>underused</strong> in their texts relative to typical language usage patterns.</p>
    <hr>

    <h4><strong>Reference Word Frequency Lists</strong></h4>

    <p>TALL uses word frequency lists calculated from the <strong>OpenSubtitles corpus</strong>, a large collection of subtitle files from movies and TV series across multiple languages. This data is sourced from the <strong>OPUS NLPL project</strong> (<strong>Tiedemann, 2012; Lison & Tiedemann, 2016</strong>), available at <a href='https://opus.nlpl.eu' target='_blank'>https://opus.nlpl.eu</a>.</p>

    <p>The OpenSubtitles corpus provides a <strong>balanced representation of everyday spoken language</strong> across diverse contexts, making it an ideal baseline for keyness analysis. By comparing specialized or domain-specific texts against this general-purpose reference, users can identify <strong>terminological distinctiveness</strong> and <strong>stylistic features</strong> that set their corpus apart.</p>
    <hr>

    <h4><strong>Supported Languages</strong></h4>

    <p>TALL currently supports keyness analysis for the following <strong>69 languages</strong>:</p>

    <ul>
      <li>Afrikaans, Albanian, Arabic, Armenian</li>
      <li>Basque, Bengali, Bosnian, Breton, Bulgarian</li>
      <li>Catalan, Chinese (Simplified), Chinese (Traditional), Chinese (English), Chinese (Pinyin), Classical Chinese, Croatian, Czech</li>
      <li>Danish, Dutch</li>
      <li>English, Esperanto, Estonian</li>
      <li>Finnish, French</li>
      <li>Galician, Georgian, German, Greek</li>
      <li>Hebrew, Hindi, Hungarian</li>
      <li>Icelandic, Indonesian, Italian</li>
      <li>Japanese</li>
      <li>Kazakh, Korean</li>
      <li>Latvian, Lithuanian</li>
      <li>Macedonian, Malay, Malayalam</li>
      <li>Norwegian</li>
      <li>Persian, Polish, Portuguese (European), Portuguese (Brazilian)</li>
      <li>Romanian, Russian</li>
      <li>Serbian, Sinhala, Slovak, Slovenian, Spanish, Swedish</li>
      <li>Tagalog, Tamil, Telugu, Turkish</li>
      <li>Ukrainian, Urdu</li>
      <li>Vietnamese</li>
    </ul>

    <p>Each language is supported with a dedicated word frequency list computed from the corresponding OpenSubtitles dataset, ensuring <strong>language-specific reference data</strong> for accurate keyness calculations.</p>
    <hr>

    <h4><strong>Output Visualizations</strong></h4>

    <p>TALL provides three complementary visualizations for keyness analysis results:</p>

    <h4><em>1. Keyness Plot</em></h4>
    <p>A horizontal bar chart displaying the <strong>Top keywords</strong> ranked by keyness scores. Words with <strong>positive keyness values</strong> (shown in blue) are overrepresented in the target corpus, while words with <strong>negative keyness values</strong> (shown in red) are underrepresented compared to the reference corpus. The length of each bar corresponds to the magnitude of the keyness score, providing an immediate visual indication of the most distinctive words.</p>

    <h4><em>2. Word Cloud</em></h4>
    <p>An interactive word cloud visualization where <strong>word size is proportional to keyness strength</strong>. This representation offers an intuitive overview of the most characteristic terms in the corpus, with larger words indicating higher keyness values. The word cloud is particularly useful for quickly identifying dominant themes and terminology.</p>

    <h4><em>3. Statistical Table</em></h4>
    <p>A comprehensive data table providing <strong>detailed statistical metrics</strong> for all analyzed words. The table includes multiple keyness measures and effect size indicators, allowing for in-depth quantitative analysis. Users can sort, filter, and export the results for further statistical processing or reporting.</p>
    <hr>

    <h4><strong>Statistical Indices in Keyness Analysis</strong></h4>

    <p>TALL computes a comprehensive set of statistical measures to assess the significance and magnitude of lexical differences between the target and reference corpora. Each index provides unique insights into word distinctiveness:</p>

    <h4><em>G² (Log-Likelihood Ratio)</em></h4>
    <p>The <strong>Log-Likelihood test (G²)</strong> is the primary keyness measure used in TALL (<strong>Dunning, 1993</strong>). It assesses whether the observed frequency difference between the target and reference corpus is statistically significant. The G² statistic follows a chi-squared distribution and provides a robust measure of keyness that is less sensitive to corpus size than alternative methods. Higher absolute values indicate stronger keyness, with positive values representing overuse and negative values representing underuse.</p>

    <h4><em>Sig_corrected (Statistical Significance)</em></h4>
    <p>The <strong>corrected significance level</strong> indicates whether the keyness difference is statistically significant after applying <strong>multiple testing corrections</strong> (e.g., Bonferroni or FDR correction). This ensures that identified keywords are truly distinctive and not the result of random variation. Typical significance thresholds include p < .001, p < .01, and p < .05.</p>

    <h4><em>Obs_Freq (Observed Frequency)</em></h4>
    <p>The <strong>observed frequency</strong> represents the actual count of the word in the target corpus. This raw frequency provides context for understanding how prevalent a term is in the analyzed texts.</p>

    <h4><em>Exp_Freq (Expected Frequency)</em></h4>
    <p>The <strong>expected frequency</strong> is the count that would be anticipated in the target corpus based on the word's frequency in the reference corpus and the relative sizes of both corpora. Substantial deviations between observed and expected frequencies indicate keyness.</p>

    <h4><em>RDF (Relative Document Frequency)</em></h4>
    <p>The <strong>Relative Document Frequency</strong> measures the proportion of documents in the target corpus that contain the word. This metric helps distinguish between words that appear frequently in a few documents versus words that are distributed across many documents, providing insights into vocabulary consistency and dispersion.</p>

    <h4><em>RateRatio</em></h4>
    <p>The <strong>Rate Ratio</strong> (also known as the <strong>Relative Risk</strong>) is the ratio of the word's frequency rate in the target corpus to its frequency rate in the reference corpus. A RateRatio > 1 indicates overuse in the target corpus, while a RateRatio < 1 indicates underuse. This effect size measure provides an intuitive interpretation of the magnitude of difference.</p>

    <h4><em>OddsRatio</em></h4>
    <p>The <strong>Odds Ratio</strong> compares the odds of a word appearing in the target corpus versus the reference corpus (<strong>Everitt, 2002</strong>). It is calculated as:</p>
    <p><code>OddsRatio = (a × d) / (b × c)</code></p>
    <p>where <em>a</em> is the word's frequency in the target corpus, <em>b</em> is the frequency of other words in the target corpus, <em>c</em> is the word's frequency in the reference corpus, and <em>d</em> is the frequency of other words in the reference corpus. Values greater than 1 indicate overuse; values less than 1 indicate underuse.</p>

    <h4><em>LogOddsRatio</em></h4>
    <p>The <strong>Log Odds Ratio</strong> is the natural logarithm of the Odds Ratio, providing a <strong>symmetric measure</strong> of effect size. This transformation makes interpretation easier, as equal magnitudes of positive and negative values represent equivalent strengths of association in opposite directions. The Log Odds Ratio is particularly useful for comparing keyness across different studies and corpora.</p>

    <h4><em>phi (Phi Coefficient)</em></h4>
    <p>The <strong>Phi coefficient</strong> is a measure of association between two binary variables, in this case, whether a word appears in the target versus reference corpus. Values range from -1 to +1, with values closer to ±1 indicating stronger associations. The phi coefficient is related to the chi-squared statistic and provides a normalized effect size measure.</p>

    <h4><em>MI (Mutual Information)</em></h4>
    <p>The <strong>Mutual Information</strong> score quantifies the amount of information shared between a word's occurrence and corpus membership (<strong>Church & Hanks, 1990</strong>). Higher MI values indicate that the presence of a word is highly informative about whether a text belongs to the target corpus. MI is particularly useful for identifying highly specific terminology but can be biased toward low-frequency words.</p>

    <h4><em>PMI (Pointwise Mutual Information)</em></h4>
    <p>The <strong>Pointwise Mutual Information</strong> is a variant of MI that measures the association strength between a specific word and the target corpus. PMI values indicate how much more likely a word is to appear in the target corpus compared to chance. Positive PMI values suggest positive association (overuse), while negative values suggest negative association (underuse).</p>

    <h4><em>DeltaP (Delta P)</em></h4>
    <p>The <strong>Delta P statistic</strong> (<strong>Gries, 2013</strong>) measures the <strong>directional association strength</strong> between a word and corpus membership. It ranges from -1 to +1, where positive values indicate attraction to the target corpus and negative values indicate repulsion. Delta P is considered a robust measure of keyness that accounts for the asymmetric nature of word-corpus associations.</p>
    <hr>

    <h4><strong>Interpreting Keyness Results</strong></h4>

    <p>When analyzing keyness results in TALL, consider the following guidelines:</p>

    <ul>
      <li><strong>Statistical Significance vs. Effect Size:</strong> A word may be statistically significant (low p-value) but have a small effect size, or vice versa. Always examine both significance tests (G², Sig_corrected) and effect size measures (RateRatio, LogOddsRatio, Delta P) for comprehensive interpretation.</li>
      <li><strong>Positive vs. Negative Keyness:</strong> Positive keyness (blue bars) indicates words that are characteristic of your corpus, while negative keyness (red bars) reveals words that are markedly absent or underused compared to general language.</li>
      <li><strong>Frequency Context:</strong> High keyness scores for low-frequency words (small Obs_Freq) may indicate specialized terminology, while high keyness for high-frequency words suggests fundamental stylistic or thematic differences.</li>
      <li><strong>Multiple Indices:</strong> Different indices may highlight different aspects of keyness. G² emphasizes statistical significance, Log Odds Ratio provides symmetric effect size, and Delta P accounts for directional associations.</li>
    </ul>
    <hr>

    <h4><strong>Applications of Keyness Analysis</strong></h4>

    <ul>
      <li><strong>Genre and Register Analysis:</strong> Identifying linguistic features that distinguish academic writing, legal texts, news articles, or social media discourse.</li>
      <li><strong>Author Attribution and Stylometry:</strong> Detecting distinctive word usage patterns that characterize individual authors or writing styles.</li>
      <li><strong>Comparative Corpus Linguistics:</strong> Comparing vocabulary across different time periods, regions, or social groups.</li>
      <li><strong>Terminology Extraction:</strong> Identifying technical terms and domain-specific vocabulary in specialized corpora.</li>
      <li><strong>Discourse Analysis:</strong> Revealing ideological or thematic emphases through the detection of overused keywords.</li>
      <li><strong>Content Analysis:</strong> Characterizing the distinctive features of different text types, publications, or communication channels.</li>
    </ul>
    <hr>

    <h4><strong>Advantages of Keyness Analysis</strong></h4>

    <ul>
      <li><strong>Multilingual Support:</strong> Provides reference frequency lists for 69 languages, enabling cross-lingual keyness studies.</li>
      <li><strong>Large-Scale Reference Data:</strong> Uses the OpenSubtitles corpus, which contains millions of words per language, ensuring robust statistical comparisons.</li>
      <li><strong>Comprehensive Statistical Measures:</strong> Computes multiple keyness indices, allowing users to select the most appropriate measure for their research questions.</li>
      <li><strong>Multiple Visualization Options:</strong> Offers plot, word cloud, and table views to suit different analytical and presentation needs.</li>
      <li><strong>Integration with TALL's NLP Pipeline:</strong> Works seamlessly with TALL's tokenization, lemmatization, and PoS-tagging infrastructure.</li>
      <li><strong>Exportable Results:</strong> All statistical results can be exported to Excel for further analysis, reporting, or integration with other tools.</li>
    </ul>

    <p>By integrating keyness analysis with <strong>OpenSubtitles-based reference data</strong> and providing <strong>comprehensive statistical measures</strong>, TALL enables researchers to conduct <strong>rigorous comparative linguistic studies</strong> across a wide range of languages and text types.</p>
    <hr>

    <div class='references'>
      <h4><strong>References</strong></h4>

      <p><strong>Church, K. W., & Hanks, P.</strong></p>
      <p><i>Word association norms, mutual information, and lexicography.</i> <strong>Computational Linguistics</strong>, 16(1), 22-29, 1990.</p>

      <p><strong>Dunning, T.</strong></p>
      <p><i>Accurate methods for the statistics of surprise and coincidence.</i> <strong>Computational Linguistics</strong>, 19(1), 61-74, 1993.</p>

      <p><strong>Everitt, B. S.</strong></p>
      <p><i>The Cambridge Dictionary of Statistics</i> (2nd ed.). Cambridge University Press, 2002.</p>

      <p><strong>Gabrielatos, C.</strong></p>
      <p><i>Keyness Analysis: Nature, metrics and techniques.</i> In C. Taylor & A. Marchi (Eds.), <strong>Corpus Approaches to Discourse: A Critical Review</strong> (pp. 225-258). Routledge, 2018.</p>

      <p><strong>Gries, S. T.</strong></p>
      <p><i>50-something years of work on collocations: What is or should be next...</i> <strong>International Journal of Corpus Linguistics</strong>, 18(1), 137-166, 2013. DOI: <a href='https://doi.org/10.1075/ijcl.18.1.09gri' target='_blank'>10.1075/ijcl.18.1.09gri</a></p>

      <p><strong>Lison, P., & Tiedemann, J.</strong></p>
      <p><i>OpenSubtitles2016: Extracting Large Parallel Corpora from Movie and TV Subtitles.</i> <strong>Proceedings of the 10th International Conference on Language Resources and Evaluation (LREC 2016)</strong>, 2016.</p>

      <p><strong>Scott, M.</strong></p>
      <p><i>PC analysis of key words—And key key words.</i> <strong>System</strong>, 25(2), 233-245, 1997. DOI: <a href='https://doi.org/10.1016/S0346-251X(97)00011-0' target='_blank'>10.1016/S0346-251X(97)00011-0</a></p>

      <p><strong>Tiedemann, J.</strong></p>
      <p><i>Parallel Data, Tools and Interfaces in OPUS.</i> <strong>Proceedings of the 8th International Conference on Language Resources and Evaluation (LREC 2012)</strong>, 2012.</p>

      <p><strong>OPUS NLPL Project:</strong> <a href='https://opus.nlpl.eu' target='_blank'>https://opus.nlpl.eu</a></p>
    </div>

  </body>"

  ## reinert ----
  reinert <- "<body>

    <h3><strong>Reinert Clustering</strong></h3>

    <p>Reinert clustering is a <strong>hierarchical descending classification method</strong> used for <strong>textual data clustering</strong>. It identifies <strong>lexically homogeneous word clusters</strong> based on the <strong>co-occurrence of terms within textual contexts</strong>. Originally developed by <strong>Max Reinert (1983, 1990)</strong>, this approach has become a core method in <strong>corpus linguistics, sociolinguistics, and content analysis</strong>.</p>

    <p>Reinert’s method is particularly effective in <strong>structuring large textual datasets</strong>, making it a powerful tool for <strong>thematic segmentation, discourse analysis, and socio-linguistic research</strong>.</p>
    <hr>
    <h4><strong>How Reinert Clustering Works</strong></h4>

    <h4>1. Text Segmentation into Context Units</h4>
    <ul>
    <li>The text is divided into <strong>small context units (CUs)</strong>, typically <strong>paragraphs or fixed-length segments</strong>, to capture local lexical co-occurrence patterns.</li>
    <li>Each CU is treated as a <strong>vector</strong> of word frequencies.</li>
    </ul>

    <h4>2. Iterative Splitting of Clusters</h4>
    <ul>
    <li>The method starts with <strong>all CUs grouped together</strong>.</li>
    <li>A <strong>first split</strong> is performed, <strong>maximizing intra-cluster homogeneity</strong> while ensuring that <strong>word distributions</strong> differ between groups.</li>
    <li>This <strong>recursive process</strong> continues until no further meaningful lexical differentiation can be achieved.</li>
    </ul>

    <h4>3. Statistical Association of Words to Clusters</h4>
    <ul>
    <li>Words are <strong>assigned probabilistic weights</strong> based on their <strong>distribution within each cluster</strong>.</li>
    <li>The <strong>most characteristic words</strong> of each cluster are identified, forming the <strong>lexical profile</strong> of the topic.</li>
    </ul>

    <h4>4. Interpretation and Thematic Analysis</h4>
    <ul>
    <li>The final clusters represent <strong>coherent thematic units</strong>.</li>
    <li>Thematic interpretation is facilitated by <strong>analyzing the most significant words in each cluster</strong>.</li>
    </ul>
    <hr>
    <h4><strong>Reinert Clustering vs. Traditional Topic Modeling</strong></h4>
    <table border='1' cellspacing='0' cellpadding='5'>
      <tr>
      <th>Feature</th>
      <th>Reinert Clustering</th>
      <th>LDA Topic Modeling</th>
      </tr>
      <tr>
      <td><strong>Method</strong></td>
      <td>Hierarchical word clustering</td>
      <td>Probabilistic word-topic assignment</td>
      </tr>
      <tr>
      <td><strong>Output</strong></td>
      <td>Discrete word clusters with distinct themes</td>
      <td>Soft assignment of words to topics</td>
      </tr>
      <tr>
      <td><strong>Context Sensitivity</strong></td>
      <td>High – Uses local lexical co-occurrence</td>
      <td>Medium – Uses global probability distributions</td>
      </tr>
      <tr>
      <td><strong>Interpretability</strong></td>
      <td>Direct thematic segmentation</td>
      <td>Requires manual topic interpretation</td>
      </tr>
      <tr>
      <td><strong>Application</strong></td>
      <td>Text segmentation, discourse analysis</td>
      <td>Thematic classification, topic inference</td>
      </tr>
      </table>
    <hr>
      <h4><strong>Implementation of Reinert Clustering</strong></h4>

      <p>The implementation of <strong>Reinert clustering in TALL</strong> was <strong>inspired by the 'rainette' package</strong> (<strong>Barnier & Privé, 2023</strong>). The original routines have been <strong>adapted to work with the TALL data structure</strong>, which includes <strong>tokenized, lemmatized, and PoS-tagged corpora</strong>.</p>

      <p>This adaptation allows:</p>
      <ul>
      <li><strong>Customization of context unit size</strong> to fit different corpus structures.</li>
      <li><strong>Compatibility with pre-processed linguistic data</strong>, ensuring greater accuracy in lexical clustering.</li>
      <li><strong>Optimized performance</strong> for large-scale text analysis, leveraging <strong>TALL’s text processing pipeline</strong>.</li>
      <li><strong>Graphical visualization</strong> of thematic structures to facilitate <strong>interpretation and reporting</strong>.</li>
      </ul>

      <p>By adapting <strong>Reinert’s methodology to TALL’s specialized NLP framework</strong>, researchers can <strong>conduct advanced text clustering analyses</strong> while maintaining compatibility with <strong>state-of-the-art linguistic preprocessing techniques</strong>.</p>
    <hr>
      <div class='references'>
        <h4><strong>References</strong></h4>

        <p><strong>Reinert, M.</strong></p>
        <p><i>Une méthode de classification descendante hiérarchique : application à l'analyse lexicale par contexte.</i> <strong>Cahiers de l'analyse des données</strong>, 8(2), 1983.</p>

        <p><strong>Reinert, M.</strong></p>
        <p><i>Alceste: Une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval.</i> <strong>Bulletin de Méthodologie Sociologique</strong>, 26(1), 1990. DOI: <a href='https://doi.org/10.1177/075910639002600103' target='_blank'>10.1177/075910639002600103</a></p>

        <p><strong>Barnier, J., & Privé, F.</strong></p>
        <p><i>rainette: The Reinert Method for Textual Data Clustering.</i> <strong>R CRAN Repository</strong>, 2023. DOI: <a href='https://doi.org/10.32614/CRAN.package.rainette' target='_blank'>10.32614/CRAN.package.rainette</a></p>
    </div>

</body>"

  ## correspondece analysis ----
  correspondenceanalysis <- "<body>

    <h3><strong>Correspondence Analysis</strong></h3>

    <p>Correspondence Analysis (<strong>CA</strong>) is a fundamental technique for exploring <strong>semantic relationships</strong> among words within a text collection (<strong>Benzécri, 1982; Lebart et al., 1997</strong>). By applying <strong>dimensionality reduction</strong>, CA represents the most relevant information in a low-rank vector space, uncovering <strong>latent structures</strong> within the data. These structures are then <strong>visualized on factorial maps</strong>, allowing users to detect associations between terms and documents effectively.</p>
    <hr>
    <h4><strong>Why Use Correspondence Analysis?</strong></h4>
    <ul>
    <li><strong>Revealing Hidden Patterns:</strong> CA captures relationships between words and documents that might not be immediately apparent.</li>
    <li><strong>Dimensionality Reduction:</strong> By projecting the data into a lower-dimensional space, CA simplifies complex text corpora while retaining key semantic information.</li>
    <li><strong>Visualization on Factorial Maps:</strong> The results are displayed on a <strong>graphical representation</strong>, enabling easy interpretation of term clusters and document similarities.</li>
    </ul>
    <hr>
    <h4><strong>Limitations of Correspondence Analysis</strong></h4>
    <p>One of the primary challenges of CA is that the <strong>new features</strong> generated through dimensionality reduction often lack <strong>direct interpretability</strong>. Since the transformation is data-driven, the factors extracted do not always correspond to clear linguistic or thematic constructs, making it more difficult to derive <strong>explicit meaning</strong> from the analysis.</p>
    <hr>
    <h4><strong>Enhancing Interpretability: The Tandem Approach</strong></h4>
    <p>To address this limitation, <strong>TALL integrates a tandem approach</strong>, which combines CA with <strong>clustering techniques</strong> to improve the interpretability of results (<strong>Misuraca & Spano, 2020</strong>). This approach follows a <strong>two-step process</strong>:</p>
    <ul>
    <li><strong>Dimensionality Reduction with CA:</strong> The text data is transformed into <strong>orthogonal and ordered features</strong>, preserving essential relationships while reducing complexity.</li>
    <li><strong>Hierarchical Clustering:</strong> Clustering is applied to the transformed data, allowing for <strong>multi-level aggregation</strong> of terms and documents. Unlike simple factor analysis, this method provides <strong>non-overlapping clusters</strong>, making the results easier to interpret.</li>
    </ul>
    <hr>
    <h4><strong>Applications of Correspondence Analysis in Text Mining</strong></h4>
    <ul>
    <li><strong>Exploring Co-occurrence Patterns:</strong> Identifying how frequently certain words appear together in a corpus.</li>
    <li><strong>Thematic Segmentation:</strong> Grouping documents based on their shared linguistic characteristics.</li>
    <li><strong>Semantic Mapping:</strong> Revealing <strong>latent structures</strong> within unstructured text data.</li>
    <li><strong>Lexical Field Analysis:</strong> Understanding how words are distributed and related within a text collection.</li>
    </ul>

    <p>By integrating <strong>Correspondence Analysis</strong> with <strong>clustering methods</strong>, TALL enhances the <strong>interpretability and usability</strong> of text mining workflows, offering a <strong>powerful framework</strong> for <strong>unsupervised exploration</strong> of large document collections.</p>
    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>
      <p><strong>Benzécri, J. P.</strong> (1982). <i>Histoire et préhistoire de l’analyse des données.</i> Paris: Dunod.</p>
      <p><strong>Lebart, L., Salem, A., & Berry, L.</strong> (1997). <i>Exploring textual data.</i> Volume 4. Springer Science & Business Media.</p>
      <p><strong>Misuraca, M., & Spano, M.</strong> (2020). <i>Unsupervised Analytic Strategies to Explore Large Document Collections.</i> Heidelberg: Springer, 06, 17-28.</p>
      </div>

      </body>"

  ## co-word analysis ----
  cowordanalysis <- "<body>

    <h3><strong>Word Network Analysis</strong></h3>

    <p>TALL provides two complementary approaches for building word networks: <strong>Co-occurrence networks</strong> based on statistical proximity, and <strong>Dependency networks</strong> based on syntactic structure. Both approaches produce interactive network visualizations with community detection for thematic clustering.</p>

    <hr>
    <h4><strong>Network Type: Co-occurrence</strong></h4>
    <p>Co-occurrence analysis is a classic <strong>network-based text mining technique</strong> that examines how often words appear together within the same context unit (<strong>Callon et al., 1983</strong>). Two words are connected if they co-occur in the same sentence, paragraph, document, or group.</p>
    <ul>
      <li><strong>Nodes</strong> represent words (terms extracted from the corpus).</li>
      <li><strong>Edges</strong> represent co-occurrence relationships based on the selected grouping level.</li>
      <li><strong>Edge weights</strong> reflect co-occurrence frequency: stronger relationships produce thicker connections.</li>
    </ul>
    <p><strong>Co-occurrence level:</strong> Users can choose the unit of analysis (Sentences, Paragraphs, Documents, or Groups). Smaller units (sentences) capture tighter semantic relationships, while larger units (documents) capture broader thematic associations.</p>

    <hr>
    <h4><strong>Network Type: Dependency</strong></h4>
    <p>Dependency networks use the <strong>syntactic dependency tree</strong> produced during tokenization to build word connections. Two words are linked only if one syntactically depends on the other in the parse tree, regardless of their linear distance in the text.</p>
    <ul>
      <li><strong>Nodes</strong> represent words, as in co-occurrence networks.</li>
      <li><strong>Edges</strong> represent <strong>grammatical relationships</strong> (e.g., a noun modified by an adjective, a verb with its subject or object).</li>
      <li><strong>Edge weights</strong> reflect how many times the syntactic relationship occurs across the corpus.</li>
    </ul>
    <p><strong>Relation filters:</strong></p>
    <ul>
      <li><strong>All syntactic</strong> &mdash; includes all major dependency relations (nsubj, obj, amod, nmod, compound, conj, advcl, etc.).</li>
      <li><strong>Noun modifiers</strong> &mdash; focuses on noun-centered relations (amod, nmod, compound, flat, nummod, appos). Reveals the descriptive structure of concepts.</li>
      <li><strong>Subject-Verb-Object</strong> &mdash; focuses on core argument structure (nsubj, obj, iobj). Shows &quot;who does what to whom&quot;.</li>
      <li><strong>Custom</strong> &mdash; lets users select specific dependency relations.</li>
    </ul>
    <p>For a complete list and description of all syntactic dependency relations, see the <strong>Dependency</strong> tab in the <strong>Overview</strong> menu.</p>

    <hr>
    <h4><strong>Co-occurrence vs Dependency: When to Use Which</strong></h4>
    <table style='width:100%; border-collapse:collapse; margin-bottom:15px;'>
      <thead>
        <tr style='background-color:#f0f0f0; border-bottom:2px solid #ccc;'>
          <th style='padding:8px; text-align:left;'>Aspect</th>
          <th style='padding:8px; text-align:left;'>Co-occurrence</th>
          <th style='padding:8px; text-align:left;'>Dependency</th>
        </tr>
      </thead>
      <tbody>
        <tr><td style='padding:6px;'>Connection basis</td><td>Proximity (same context unit)</td><td>Grammatical relation</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'>Noise level</td><td>Higher (coincidental proximity)</td><td>Lower (only structural links)</td></tr>
        <tr><td style='padding:6px;'>Distant relations</td><td>Missed if outside window</td><td>Captured if syntactically linked</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'>Interpretability</td><td>Thematic association</td><td>Grammatical roles</td></tr>
        <tr><td style='padding:6px;'>Best for</td><td>Broad thematic mapping</td><td>Precise semantic relationships</td></tr>
      </tbody>
    </table>

    <hr>
    <h4><strong>Normalization Measures</strong></h4>
    <p>Both network types support the same normalization measures to control for term frequency bias (<strong>Eck &amp; Waltman, 2009</strong>):</p>

    <h4><em>Association Index</em></h4>
    <p>Normalizes co-occurrence relative to expected frequency: <code>AI<sub>ij</sub> = C<sub>ij</sub> / (C<sub>i</sub> &times; C<sub>j</sub>)</code></p>

    <h4><em>Cosine Similarity</em></h4>
    <p>Geometric similarity measure: <code>cos(&theta;) = C<sub>ij</sub> / sqrt(C<sub>i</sub> &times; C<sub>j</sub>)</code></p>

    <h4><em>Jaccard Similarity</em></h4>
    <p>Set-based overlap measure: <code>J<sub>ij</sub> = C<sub>ij</sub> / (C<sub>i</sub> + C<sub>j</sub> - C<sub>ij</sub>)</code></p>

    <hr>
    <h4><strong>Community Detection</strong></h4>
    <p>TALL applies the <strong>Louvain algorithm</strong> (<strong>Blondel et al., 2008</strong>) for community detection, which optimizes modularity to find dense clusters of connected words:</p>
    <ul>
      <li>Runs <strong>10 iterations</strong> with different random seeds and selects the solution with highest modularity.</li>
      <li>Groups words into <strong>non-overlapping clusters</strong> representing latent topics or conceptual domains.</li>
      <li>Applies <strong>community repulsion</strong> to spatially separate clusters in the visualization.</li>
    </ul>

    <hr>
    <h4><strong>Applications</strong></h4>
    <ul>
      <li><strong>Thematic mapping:</strong> Identifying research trends and conceptual structures in academic literature.</li>
      <li><strong>Topic detection:</strong> Extracting underlying themes from news, reports, or social media.</li>
      <li><strong>Semantic role analysis</strong> (dependency mode): Understanding &quot;who does what&quot; patterns in a corpus.</li>
      <li><strong>Terminology extraction:</strong> Discovering domain-specific concept networks.</li>
    </ul>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>

      <p><strong>Callon, M., Courtial, J.-P., Turner, W.A., &amp; Bauin, S.</strong> (1983) <i>From translations to problematic networks: An introduction to co-word analysis.</i> <strong>Social Science Information</strong>, 22(2), 191-235.</p>

      <p><strong>Eck, N.J.V., &amp; Waltman, L.</strong> (2009) <i>How to normalize co-occurrence data? An analysis of some well-known similarity measures.</i> <strong>Journal of the American Society for Information Science and Technology</strong>, 60(8), 1635-1651.</p>

      <p><strong>Blondel, V.D., Guillaume, J.-L., Lambiotte, R., &amp; Lefebvre, E.</strong> (2008) <i>Fast unfolding of communities in large networks.</i> <strong>Journal of Statistical Mechanics</strong>, 2008(10), P10008.</p>

      <p><strong>de Marneffe, M.-C., Manning, C.D., Nivre, J., &amp; Zeman, D.</strong> (2021) <i>Universal Dependencies.</i> <strong>Computational Linguistics</strong>, 47(2), 255-308.</p>

      <p><strong>Fortunato, S., &amp; Hric, D.</strong> (2016) <i>Community detection in networks: A user guide.</i> <strong>Physics Reports</strong>, 659, 1-44.</p>
    </div>

    </body>"

  ## thematic map ----
  thematicmap <- "
  <body>

  <h3><strong>Thematic Map</strong></h3>

  <p>
  The <strong>Thematic Map</strong> feature in TALL enables users to explore the conceptual structure of a text corpus by visually mapping the most relevant topics. It is based on an unsupervised, network-based method designed to extract, cluster, and characterize groups of words representing distinct semantic areas within the analyzed texts. This approach has been successfully applied in bibliometric research and adapted in TALL for general-purpose text analysis.
</p>

  <hr>

  <h4><strong>Methodological Framework</strong></h4>

  <p>
  Thematic mapping starts with the construction of a <strong>co-occurrence matrix</strong> from the pre-processed text corpus. The association strength between terms is then calculated to normalize the raw co-occurrence frequencies:
  </p>

  <math xmlns=,http://www.w3.org/1998/Math/MathML, style=,font-size: 1.1em; display: block; text-align: center; margin: 10px 0;,>
    <mrow>
    <mtext>AS</mtext><msub><mi>jj'</mi></msub><mo>=</mo>
      <mfrac>
        <msub><mi>a</mi><mi>jj'</mi></msub>
    <mrow>
    <msub><mi>a</mi><mi>jj</mi></msub><mo>&#x22C5;</mo><msub><mi>a</mi><mi>j'j'</mi></msub>
    </mrow>
    </mfrac>
    </mrow>
    </math>

    <p>
    where <em>AS<sub>jj'</sub></em> is the association strength between terms <em>j</em> and <em>j'</em>, and <em>a<sub>jj'</sub></em> is their observed co-occurrence. This metric expresses the semantic relatedness of term pairs.
  </p>

  <p>
    A <strong>community detection algorithm</strong> (WalkTrap) is then applied to the normalized network to identify clusters of terms (i.e., topics). Each cluster is projected onto a two-dimensional plane using two dimensions:
  </p>

  <ul>
    <li><strong>Callon Centrality (CC):</strong> measures a topic’s interaction with others, indicating its <em>relevance</em> in the corpus.</li>
    <li><strong>Callon Density (CD):</strong> measures the internal cohesion of the topic, reflecting its <em>development</em>.</li>
  </ul>

  <p>Each topic is placed on a <strong>strategic diagram</strong> based on its centrality and density values:</p>

  <ul>
    <li><strong>Upper-right (Hot Topics):</strong> High centrality and high density – well-developed and important.</li>
    <li><strong>Lower-right (Basic Topics):</strong> High centrality and low density – important but still under development.</li>
    <li><strong>Upper-left (Niche Topics):</strong> Low centrality and high density – well developed but marginal.</li>
    <li><strong>Lower-left (Peripheral Topics):</strong> Low centrality and low density – weakly developed and marginal.</li>
  </ul>

  <hr>

  <h4><strong>Features</strong></h4>

  <ul>
    <li>Users can generate thematic maps from any textual dataset preprocessed and tokenized in TALL.</li>
    <li>The algorithm works automatically and does <strong>not require setting the number of topics</strong> in advance.</li>
    <li>Topics are labeled by the most frequent keywords within each cluster.</li>
    <li>Topic size (i.e., the size of the bubble) represents the number of terms in the cluster.</li>
    <li>The user can select specific time slices or metadata filters to perform <strong>comparative thematic analysis</strong> across groups or periods.</li>
  </ul>

  <p>
    Thematic maps offer a rich, interpretable representation of discourse structure and are particularly effective for exploratory text mining and culturomic studies.
  </p>

  <hr>

  <div class='references'>
    <h4><strong>References</strong></h4>
    <p><strong>Aria, M., Cuccurullo, C., D’Aniello, L., Misuraca, M., & Spano, M. (2022).</strong> <i>Thematic Analysis as a New Culturomic Tool: The Social Media Coverage on COVID-19 Pandemic in Italy.</i> <strong>Sustainability</strong>, 14(6), 3643. https://doi.org/10.3390/su14063643</p>
    <p><strong>Cobo, M.J., López-Herrera, A.G., Herrera-Viedma, E., & Herrera, F. (2011).</strong> <i>An approach for detecting, quantifying, and visualising the evolution of a research field: A practical application to the fuzzy sets theory field.</i> <strong>Journal of Informetrics</strong>, 5(1), 146–166.</p>
  </div>

</body>

  "

  ## embedding training ----
  embeddingtrain <- "

    <body>
    <h3><strong>Training Word Embeddings</strong></h3>
    <p>
    The <strong>Training</strong> module in TALL enables users to generate <strong>custom word embeddings</strong> from their own corpus using the <strong>word2vec algorithm</strong>,
  which includes both the <strong>Continuous Bag-of-Words (CBOW)</strong> and <strong>Skip-gram</strong> architectures.
  These models create dense vector representations that capture semantic and syntactic relationships among words based on their distributional context.
  </p>

    <hr>
    <h4><strong>Available Architectures</strong></h4>
    <ul>
    <li><strong>CBOW:</strong> Predicts a word from its surrounding context. It is faster and works well with frequent words.</li>
    <li><strong>Skip-gram:</strong> Predicts surrounding context words from a target word. It is slower but performs better with infrequent words.</li>
    </ul>

    <hr>
    <h4><strong>How It Works</strong></h4>
    <ul>
    <li>Text data is lemmatized and filtered to exclude non-informative tokens (e.g., punctuation, auxiliaries, determiners).</li>
    <li>Training is performed at the sentence level to preserve local context.</li>
    <li>Stopwords are automatically identified and excluded.</li>
    <li>Parameters such as <code>dimensionality</code>, <code>number of iterations</code>, and <code>architecture (CBOW/Skip-gram)</code> can be configured.</li>
    </ul>

    <hr>
    <h4><strong>Outputs</strong></h4>
    <ul>
    <li>Word embedding matrix.</li>
    <li>Descriptive statistics for each vector dimension (mean, SD, skewness, kurtosis).</li>
    <li>PCA analysis to evaluate variance explained by each component.</li>
    <li>Cosine similarity and Euclidean distance metrics for quality assessment.</li>
    </ul>

    <hr>
    <h4><strong>Example</strong></h4>
    <p>
    Training a word2vec model on a corpus of product reviews may reveal that terms like <code>“delivery”</code> and <code>“shipping”</code> appear close in vector space,
  indicating their semantic similarity within that context.
  </p>

    <hr>
    <h4><strong>References</strong></h4>
    <ul>
    <li>
    Mikolov, T., Chen, K., Corrado, G., & Dean, J. (2013).
  <i>Efficient Estimation of Word Representations in Vector Space</i>.
  <a href='https://arxiv.org/abs/1301.3781' target='_blank'>arXiv:1301.3781</a>
    </li>
    <li>
    Mikolov, T., Sutskever, I., Chen, K., Corrado, G., & Dean, J. (2013).
  <i>Distributed Representations of Words and Phrases and their Compositionality</i>.
  <a href='https://arxiv.org/abs/1310.4546' target='_blank'>arXiv:1310.4546</a>
    </li>
    </ul>
    </body>
  "

  ## embedding similarity ----
  embeddingsimilarity <- "<body>
    <h3><strong>Word Similarity Network</strong></h3>
    <p>
    The <strong>Similarity</strong> module in TALL allows users to explore semantic relationships between words through an interactive <strong>similarity network</strong> generated from word embeddings trained in the <strong>Training</strong> tab. These embeddings are built using the <strong>word2vec</strong> algorithm (either CBOW or Skip-gram).
  </p>

    <hr>
    <h4><strong>How It Works</strong></h4>
    <ul>
    <li>
    TALL selects the <strong>top 100 most frequent content words</strong> in the corpus (restricted to POS: NOUN, PROPN, ADJ).
  </li>
    <li>
    For each of these 100 terms, the system computes the <strong>10 most similar words</strong> based on <strong>cosine similarity</strong> in the embedding space.
  </li>
    <li>
    The resulting network is composed of:
    <ul>
    <li><strong>Nodes:</strong> the 100 target words (triangles) and their similar terms (dots).</li>
    <li><strong>Edges:</strong> connections representing semantic similarity scores (cosine similarity ≥ 0.5), with width proportional to similarity.</li>
    </ul>
    </li>
    <li>
    The network also undergoes <strong>community detection</strong> using the Walktrap algorithm to highlight thematic clusters.
  </li>
    </ul>

    <hr>
    <h4><strong>Visualization Tools</strong></h4>
    <ul>
    <li><strong>UMAP projection:</strong> two-dimensional semantic mapping of all words in the embedding matrix.</li>
    <li><strong>Overlap reduction:</strong> improves readability by adjusting label positions and opacity in dense areas.</li>
    <li><strong>Interactive display:</strong> with zoom, node highlighting, and draggable layout via <code>visNetwork</code>.</li>
    </ul>

    <hr>
    <h4><strong>Example</strong></h4>
    <p>
    After training on a corpus of scientific publications, the similarity network might display <code>“method”</code>, <code>“approach”</code>, and <code>“model”</code> as top frequent terms, each connected to semantically related concepts such as <code>“algorithm”</code>, <code>“technique”</code>, or <code>“framework”</code>.
  </p>

    <hr>
<h4><strong>References</strong></h4>
<ul>
  <li>
    Mikolov, T., Sutskever, I., Chen, K., Corrado, G. S., & Dean, J. (2013).
    <i>Distributed Representations of Words and Phrases and their Compositionality.</i>
    In *Advances in Neural Information Processing Systems* (NeurIPS 2013), 26, 3111–3119.
    <a href='https://papers.nips.cc/paper_files/paper/2013/file/9aa42b31882ec039965f3c4923ce901b-Paper.pdf' target='_blank'>
      [View PDF]
    </a>
  </li>
  <li>
    Mikolov, T., Chen, K., Corrado, G., & Dean, J. (2013).
    <i>Efficient Estimation of Word Representations in Vector Space.</i>
    <a href='https://arxiv.org/abs/1301.3781' target='_blank'>arXiv:1301.3781</a>
  </li>
</ul>
    </body>

    "

  ## tm chioice ----
  tmkchoice <- "<body>

  <h3><strong>Topic Modeling: K Selection</strong></h3>

  <p>Determining the <strong>optimal number of topics (K)</strong> is one of the most critical steps in topic modeling. TALL provides a comprehensive K selection framework that supports three model types: <strong>LDA</strong>, <strong>CTM</strong>, and <strong>STM</strong>.</p>

  <hr>
  <h4><strong>Why is K Selection Important?</strong></h4>
  <ul>
    <li>A <strong>too small K</strong> may <strong>merge distinct topics</strong>, reducing the model's ability to separate different thematic structures.</li>
    <li>A <strong>too large K</strong> may <strong>fragment coherent topics</strong>, introducing unnecessary complexity and reducing interpretability.</li>
    <li>The <strong>correct K</strong> ensures that topics are <strong>coherent, interpretable, and representative</strong> of the dataset.</li>
  </ul>

  <hr>
  <h4><strong>Supported Models</strong></h4>
  <p>TALL supports K selection for three topic modeling approaches:</p>
  <ul>
    <li><strong>LDA (Latent Dirichlet Allocation)</strong> – The classic probabilistic topic model (Blei et al., 2003). Topics are assumed to be independent.</li>
    <li><strong>CTM (Correlated Topic Model)</strong> – An extension of LDA that allows topics to be correlated with each other (Blei & Lafferty, 2007), using a logistic normal distribution instead of Dirichlet.</li>
    <li><strong>STM (Structural Topic Model)</strong> – Allows external covariates (e.g., time, source, metadata) to influence topic prevalence and content (Roberts et al., 2019).</li>
  </ul>

  <hr>
  <h4><strong>Metrics for LDA and CTM</strong></h4>
  <p>For LDA and CTM, TALL computes four standard metrics across the K range and identifies the optimal K using the <strong>elbow method</strong> (maximum distance from line):</p>

  <h4><em>Cao et al. (2009) – Topic Coherence</em></h4>
  <p>Computes the <strong>average pairwise cosine similarity</strong> between topics. The optimal K minimizes inter-topic similarity, ensuring well-separated topics. <strong>Lower values are better.</strong></p>

  <h4><em>Arun et al. (2010) – KL Divergence</em></h4>
  <p>Compares the <strong>word-topic distribution</strong> (via SVD) and <strong>document-topic distribution</strong> using <strong>symmetric Kullback-Leibler divergence</strong>. The optimal K is where divergence stabilizes. <strong>Lower values are better.</strong></p>

  <h4><em>Deveaud et al. (2014) – Jensen-Shannon Divergence</em></h4>
  <p>Measures <strong>pairwise Jensen-Shannon divergence</strong> between topic distributions. Balances topic coherence and diversity. <strong>Lower values indicate more separated topics.</strong></p>

  <h4><em>Perplexity (Blei et al., 2003)</em></h4>
  <p>A <strong>likelihood-based metric</strong> measuring how well the model generalizes. Defined as the inverse geometric mean of the likelihood. <strong>Lower values indicate better generalization.</strong></p>

  <hr>
  <h4><strong>Metrics for STM</strong></h4>
  <p>For STM, TALL uses <code>stm::searchK()</code> which provides STM-specific quality metrics:</p>

  <h4><em>Exclusivity</em></h4>
  <p>Measures how <strong>exclusive</strong> the top words of each topic are. High exclusivity means topic words are not shared across topics. <strong>Higher values are better.</strong></p>

  <h4><em>Semantic Coherence (Mimno et al., 2011)</em></h4>
  <p>Measures how often the <strong>top words of a topic co-occur</strong> within documents. Highly coherent topics are easier to interpret. <strong>Higher (less negative) values are better.</strong></p>

  <h4><em>Combined Score (Exclusivity + Coherence)</em></h4>
  <p>A synthetic metric that sums exclusivity and semantic coherence, balancing both dimensions. <strong>Higher values indicate a better trade-off.</strong></p>

  <h4><em>Lower Bound</em></h4>
  <p>The variational <strong>lower bound on the log-likelihood</strong>. Higher values indicate better model fit.</p>

  <hr>
  <h4><strong>Analysis Tabs</strong></h4>
  <ul>
    <li><strong>Tuning Plot</strong> – Interactive plot of the selected metric across K values. The optimal K (elbow point) is highlighted in red.</li>
    <li><strong>Multi-Metric Comparison</strong> – All four metrics normalized to [0, 1] and plotted together. Each metric's elbow point is marked with a diamond. This allows visual comparison of where different metrics agree or disagree.</li>
    <li><strong>K Recommendation</strong> – A consensus panel showing the optimal K suggested by each metric and the <strong>overall recommended K</strong> (mode of all suggestions). The consensus approach is more robust than relying on a single metric.</li>
    <li><strong>Table</strong> – Full table with raw and normalized metric values for all K values.</li>
  </ul>

  <hr>
  <h4><strong>Integration with Model Estimation</strong></h4>
  <p>When K selection is completed, the recommended K is <strong>automatically transferred</strong> to the Model Estimation panel. Users can accept the recommendation or adjust K manually based on domain knowledge.</p>

  <hr>
  <h4><strong>Practical Guidelines</strong></h4>
  <ul>
    <li><strong>For exploratory research:</strong> Start with <strong>low K</strong> values (e.g., <strong>5-20 topics</strong>) to gain an overview of broad themes.</li>
    <li><strong>For fine-grained analysis:</strong> Use <strong>higher K values</strong> (e.g., <strong>30-100 topics</strong>) to capture more nuanced subtopics.</li>
    <li><strong>For benchmarking:</strong> Compare different K values using the Multi-Metric Comparison tab.</li>
    <li><strong>For STM:</strong> Look for K values that maximize both exclusivity and semantic coherence simultaneously (upper-right quadrant in the Model Diagnostics scatter plot).</li>
  </ul>

  <hr>
  <div class='references'>
    <h4><strong>References</strong></h4>

    <p><strong>Blei, D.M., Ng, A.Y., & Jordan, M.I.</strong> (2003) <i>Latent Dirichlet Allocation.</i> <strong>Journal of Machine Learning Research</strong>, 3, 993-1022.</p>

    <p><strong>Blei, D.M. & Lafferty, J.D.</strong> (2007) <i>A correlated topic model of Science.</i> <strong>The Annals of Applied Statistics</strong>, 1(1), 17-35.</p>

    <p><strong>Cao, J., Xia, T., Li, J., Zhang, Y., & Tang, S.</strong> (2009) <i>A density-based method for adaptive LDA model selection.</i> <strong>Neurocomputing</strong>, 72(7), 1775-1781.</p>

    <p><strong>Arun, R., Suresh, V., Veni Madhavan, C.E., & Narasimha Murthy, M.N.</strong> (2010) <i>On finding the natural number of topics with latent Dirichlet allocation: Some observations.</i> In <strong>Advances in Knowledge Discovery and Data Mining</strong> (pp. 391-402). Springer.</p>

    <p><strong>Mimno, D., Wallach, H.M., Talley, E., Leenders, M., & McCallum, A.</strong> (2011) <i>Optimizing semantic coherence in topic models.</i> In <strong>Proceedings of EMNLP</strong> (pp. 262-272).</p>

    <p><strong>Deveaud, R., Sanjuan, E., & Bellot, P.</strong> (2014) <i>Accurate and effective latent concept modeling for ad hoc information retrieval.</i> <strong>Document Numerique</strong>, 17, 61-84.</p>

    <p><strong>Roberts, M.E., Stewart, B.M., & Tingley, D.</strong> (2019) <i>stm: An R package for structural topic models.</i> <strong>Journal of Statistical Software</strong>, 91(2), 1-40.</p>
  </div>

  </body>"

  ## topic model estimation ----
  tmmodelestimation <- "<body>

    <h3><strong>Topic Modeling: Model Estimation</strong></h3>

    <p>TALL implements three topic modeling approaches, each suited to different analytical needs. All three produce <strong>beta</strong> (term-topic) and <strong>theta</strong> (document-topic) probability matrices that can be explored through interactive visualizations.</p>

    <hr>
    <h4><strong>LDA – Latent Dirichlet Allocation</strong></h4>
    <p><strong>Blei et al. (2003)</strong> introduced LDA as a Bayesian generative model where:</p>
    <ul>
      <li>Each document is a <strong>mixture of topics</strong> with different proportions (theta).</li>
      <li>Each topic is a <strong>probability distribution over words</strong> (beta).</li>
      <li>Topics are assumed to be <strong>independent</strong> of each other.</li>
    </ul>
    <p>TALL estimates LDA using <strong>Gibbs sampling</strong> (500 iterations), which tends to produce more stable results than variational inference for small-to-medium corpora.</p>

    <hr>
    <h4><strong>CTM – Correlated Topic Model</strong></h4>
    <p><strong>Blei & Lafferty (2007)</strong> extended LDA by replacing the Dirichlet prior on topic proportions with a <strong>logistic normal distribution</strong>, which allows topics to be correlated. This is useful when:</p>
    <ul>
      <li>Topics are naturally <strong>related</strong> (e.g., &quot;politics&quot; and &quot;economics&quot; often co-occur).</li>
      <li>You want to <strong>model inter-topic dependencies</strong> rather than assuming independence.</li>
    </ul>
    <p>The <strong>Topic Correlation</strong> tab is particularly informative with CTM, as correlations reflect the model's learned structure rather than post-hoc observations.</p>

    <hr>
    <h4><strong>STM – Structural Topic Model</strong></h4>
    <p><strong>Roberts, Stewart & Tingley (2019)</strong> developed STM to incorporate <strong>document-level metadata</strong> (covariates) into the topic model. STM allows:</p>
    <ul>
      <li><strong>Prevalence covariates</strong> – External variables (e.g., publication year, source, author) that influence <strong>how much</strong> each topic appears in a document.</li>
      <li><strong>Correlated topics</strong> – Like CTM, STM allows topic correlations.</li>
    </ul>
    <p>Covariates are selected in the Options panel. After estimation, the <strong>Covariate Effects</strong> tab shows how each covariate influences topic prevalence, with regression coefficients and effect plots.</p>

    <hr>
    <h4><strong>Analysis Tabs</strong></h4>
    <ul>
      <li><strong>Topic by Words Plot</strong> – Top words per topic ranked by beta probability, displayed in groups of three for easy comparison. Navigate with arrow buttons.</li>
      <li><strong>Topic by Docs Plot</strong> – Documents most associated with each topic, ranked by theta probability.</li>
      <li><strong>Beta Probability</strong> – Full term-topic probability table.</li>
      <li><strong>Theta Probability</strong> – Full document-topic probability table.</li>
      <li><strong>Topic Correlation</strong> – Heatmap of correlations between topics based on their word distributions, with embedded mini-scatterplots showing the data distribution.</li>
      <li><strong>Model Diagnostics</strong> – Global quality metrics and per-topic indicators:
        <ul>
          <li><strong>LDA/CTM:</strong> Log-Likelihood, topic share, word entropy, and top word probability per topic.</li>
          <li><strong>STM:</strong> Variational lower bound, semantic coherence, exclusivity per topic, plus a scatter plot of coherence vs. exclusivity (ideal topics are in the upper-right quadrant).</li>
        </ul>
      </li>
      <li><strong>Covariate Effects</strong> (STM only) – For each prevalence covariate:
        <ul>
          <li><strong>Effect plots</strong> showing how the covariate influences each topic's prevalence (continuous variables show trend lines with 95% CI; categorical variables show point estimates).</li>
          <li><strong>Regression coefficients table</strong> with estimates, standard errors, t-values, and p-values per topic.</li>
        </ul>
      </li>
    </ul>

    <hr>
    <h4><strong>Choosing the Right Model</strong></h4>
    <table style='width:100%; border-collapse:collapse; margin-bottom:15px;'>
      <thead>
        <tr style='background-color:#f0f0f0; border-bottom:2px solid #ccc;'>
          <th style='padding:8px; text-align:left;'>Feature</th>
          <th style='padding:8px; text-align:center;'>LDA</th>
          <th style='padding:8px; text-align:center;'>CTM</th>
          <th style='padding:8px; text-align:center;'>STM</th>
        </tr>
      </thead>
      <tbody>
        <tr><td style='padding:6px;'>Topic independence</td><td style='text-align:center;'>Assumed</td><td style='text-align:center;'>Correlated</td><td style='text-align:center;'>Correlated</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'>External covariates</td><td style='text-align:center;'>No</td><td style='text-align:center;'>No</td><td style='text-align:center;'>Yes</td></tr>
        <tr><td style='padding:6px;'>Inference method</td><td style='text-align:center;'>Gibbs sampling</td><td style='text-align:center;'>Variational EM</td><td style='text-align:center;'>Variational EM</td></tr>
        <tr style='background-color:#f8f8f8;'><td style='padding:6px;'>Best for</td><td style='text-align:center;'>General use</td><td style='text-align:center;'>Correlated themes</td><td style='text-align:center;'>Metadata-rich corpora</td></tr>
        <tr><td style='padding:6px;'>Speed</td><td style='text-align:center;'>Moderate</td><td style='text-align:center;'>Fast</td><td style='text-align:center;'>Moderate</td></tr>
      </tbody>
    </table>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>

      <p><strong>Blei, D.M., Ng, A.Y., & Jordan, M.I.</strong> (2003) <i>Latent Dirichlet Allocation.</i> <strong>Journal of Machine Learning Research</strong>, 3, 993-1022.</p>

      <p><strong>Blei, D.M. & Lafferty, J.D.</strong> (2007) <i>A correlated topic model of Science.</i> <strong>The Annals of Applied Statistics</strong>, 1(1), 17-35.</p>

      <p><strong>Mimno, D., Wallach, H.M., Talley, E., Leenders, M., & McCallum, A.</strong> (2011) <i>Optimizing semantic coherence in topic models.</i> In <strong>Proceedings of EMNLP</strong> (pp. 262-272).</p>

      <p><strong>Roberts, M.E., Stewart, B.M., & Tingley, D.</strong> (2019) <i>stm: An R package for structural topic models.</i> <strong>Journal of Statistical Software</strong>, 91(2), 1-40.</p>
    </div>

    </body>"

  ## polarity detection ----
  ## syntactic complexity ----
  syntacticcomplexity <- "<body>

    <h3><strong>Syntactic Complexity Analysis</strong></h3>

    <p>Syntactic complexity analysis quantifies the structural properties of sentences in a corpus using <strong>dependency parsing</strong>. For each document, TALL computes a set of metrics that characterize writing style, readability, and linguistic sophistication.</p>

    <hr>
    <h4><strong>Metrics</strong></h4>

    <h4><em>Mean Sentence Length</em></h4>
    <p>Average number of content words per sentence (excluding punctuation). Longer sentences generally indicate more complex writing.</p>

    <h4><em>Mean Tree Depth</em></h4>
    <p>Average maximum depth of the dependency tree across sentences. Deeper trees indicate more levels of syntactic embedding (e.g., nested relative clauses, complex noun phrases). Based on BFS traversal from the root.</p>
    <div style='text-align: center; margin: 20px 0;'>
      <img src='dep_tree_example.png' style='max-width: 100%; border: 1px solid #ddd; border-radius: 8px; box-shadow: 0 2px 8px rgba(0,0,0,0.1);' alt='Dependency tree example' />
      <p style='color: #888; font-size: 12px; margin-top: 8px;'><em>Example of a dependency tree showing syntactic relations between tokens. Tree depth is the longest path from root to leaf.</em></p>
    </div>

    <h4><em>Mean Dependency Distance</em></h4>
    <p>Average absolute distance (in token positions) between each word and its syntactic head (<strong>Liu, 2008</strong>). Longer distances indicate more complex processing demands. Languages and styles with more head-final or head-initial constructions tend to show different profiles.</p>

    <h4><em>Clauses per Sentence</em></h4>
    <p>Average number of clauses per sentence, computed as: 1 (main clause) + subordinate clauses + coordinate verb clauses. Higher values indicate more complex multi-clause constructions.</p>

    <h4><em>Subordination Ratio</em></h4>
    <p>Proportion of subordinate clauses (advcl, acl, ccomp, xcomp, csubj) relative to total clauses. Higher ratios indicate more hypotactic writing style (subordination), while lower ratios suggest paratactic style (coordination). A key indicator of academic vs. informal writing.</p>

    <h4><em>Branching Factor</em></h4>
    <p>Average number of direct dependents per non-leaf node in the dependency tree. Higher branching indicates flatter but wider syntactic structures (e.g., enumerations, multiple modifiers).</p>

    <hr>
    <h4><strong>Analysis Tabs</strong></h4>
    <ul>
      <li><strong>Document Metrics</strong> &mdash; Full table with all metrics per document. Sortable and filterable.</li>
      <li><strong>Corpus Summary</strong> &mdash; Six summary cards showing corpus-level averages for each metric.</li>
      <li><strong>Distributions</strong> &mdash; Overlaid histograms of all metrics (toggle visibility via legend). Shows how documents are distributed along each complexity dimension.</li>
    </ul>

    <hr>
    <h4><strong>Applications</strong></h4>
    <ul>
      <li><strong>Readability assessment:</strong> Higher tree depth and dependency distance correlate with lower readability.</li>
      <li><strong>Authorship analysis:</strong> Different authors exhibit characteristic syntactic profiles.</li>
      <li><strong>Genre comparison:</strong> Academic writing typically shows higher subordination ratios than journalistic or informal text.</li>
      <li><strong>Language development:</strong> Syntactic complexity increases with proficiency in L2 learner corpora.</li>
      <li><strong>Temporal analysis:</strong> Track how writing complexity evolves over time in a corpus.</li>
    </ul>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>

      <p><strong>Liu, H.</strong> (2008) <i>Dependency distance as a metric of language comprehension difficulty.</i> <strong>Journal of Cognitive Science</strong>, 9(2), 159-191.</p>

      <p><strong>Lu, X.</strong> (2010) <i>Automatic analysis of syntactic complexity in second language writing.</i> <strong>International Journal of Corpus Linguistics</strong>, 15(4), 474-496.</p>

      <p><strong>de Marneffe, M.-C., Manning, C.D., Nivre, J., &amp; Zeman, D.</strong> (2021) <i>Universal Dependencies.</i> <strong>Computational Linguistics</strong>, 47(2), 255-308.</p>
    </div>

    </body>"

  ## SVO triplets ----
  svo <- "<body>

    <h3><strong>SVO Triplet Extraction</strong></h3>

    <p>SVO (Subject-Verb-Object) triplet extraction uses <strong>dependency parsing</strong> to identify structured semantic relationships in text. For each verb in the corpus, the algorithm extracts its syntactic subject and object, producing triplets that answer the question: <strong>who does what to whom?</strong></p>

    <hr>
    <h4><strong>How It Works</strong></h4>
    <p>The algorithm traverses the dependency tree of each sentence and, for every verb (VERB/AUX), identifies:</p>
    <ul>
      <li><strong>Subject</strong> &mdash; tokens with dependency relation <code>nsubj</code> (nominal subject), <code>nsubj:pass</code> (passive subject), or <code>csubj</code> (clausal subject).</li>
      <li><strong>Object</strong> &mdash; tokens with relation <code>obj</code> (direct object), <code>iobj</code> (indirect object), <code>obl</code> (oblique argument), <code>xcomp</code> or <code>ccomp</code> (clausal complement).</li>
    </ul>
    <p>For transitive verbs, all subject-object pairs are generated. For intransitive verbs, subject-verb pairs are emitted with an empty object field. The extraction is implemented in <strong>C++</strong> for high performance.</p>

    <hr>
    <h4><strong>Analysis Tabs</strong></h4>
    <ul>
      <li><strong>SVO Table</strong> &mdash; Full table of extracted triplets with frequency counts. Sortable and filterable by subject, verb, object, or relation type.</li>
      <li><strong>SVO Network</strong> &mdash; Sankey diagram showing the flow from subjects (blue) through verbs (green) to objects (red). Edge width reflects frequency. Reveals which actors perform which actions on which targets.</li>
      <li><strong>Verb Frequency</strong> &mdash; Bar chart of the most frequent verbs in SVO triplets. Highlights the dominant actions in the corpus.</li>
    </ul>

    <hr>
    <h4><strong>Parameters</strong></h4>
    <ul>
      <li><strong>Min. Frequency</strong> &mdash; Minimum number of occurrences for a triplet to be included (default: 2). Increase to focus on recurring patterns.</li>
      <li><strong>Top N Triplets</strong> &mdash; Number of top triplets to display in the Sankey network (default: 50). Controls visual complexity.</li>
    </ul>

    <hr>
    <h4><strong>Applications</strong></h4>
    <ul>
      <li><strong>Content analysis:</strong> Identify dominant actors, actions, and targets in a corpus (e.g., &quot;government implements policy&quot;).</li>
      <li><strong>Comparative studies:</strong> Compare SVO patterns across groups, time periods, or sources.</li>
      <li><strong>Event extraction:</strong> Discover recurring events described in news, scientific abstracts, or reports.</li>
      <li><strong>Narrative analysis:</strong> Map the structure of narratives by examining who does what.</li>
    </ul>

    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>

      <p><strong>de Marneffe, M.-C., Manning, C.D., Nivre, J., &amp; Zeman, D.</strong> (2021) <i>Universal Dependencies.</i> <strong>Computational Linguistics</strong>, 47(2), 255-308.</p>

      <p><strong>Jurafsky, D., &amp; Martin, J.H.</strong> (2024) <i>Speech and Language Processing.</i> 3rd edition (draft). Chapter 18: Information Extraction.</p>
    </div>

    </body>"

  polaritydetection <- "<body>

    <h3><strong>Polarity Detection</strong></h3>

    <p>Polarity detection is a fundamental <strong>sentiment analysis technique</strong> used to determine whether a document expresses a <strong>positive, negative, or neutral</strong> sentiment. This process is essential in analyzing <strong>consumer feedback, financial reports, product reviews, and social media discussions</strong>, where understanding sentiment trends can provide valuable insights into public opinion and decision-making processes.</p>
    <hr>
    <h4><strong>How Polarity Detection Works</strong></h4>
    <p>TALL calculates <strong>document polarity</strong> using a <strong>lexicon-based approach</strong>, incorporating <strong>contextual adjustments</strong> to refine sentiment scoring. The methodology follows three key steps:</p>

    <h4>1. Lexicon-Based Sentiment Scoring</h4>
    <ul>
    <li>Each word in the text is assigned a <strong>polarity score</strong> based on its presence in <strong>sentiment lexicons</strong>.</li>
    <li><strong>Positive words</strong> (e.g., 'excellent,' 'happy') are assigned <strong>+1</strong>, while <strong>negative words</strong> (e.g., 'bad,' 'fail') receive <strong>-1</strong>.</li>
    <li>Words <strong>not found in sentiment lexicons</strong> are considered neutral and assigned a score of <strong>0</strong>.</li>
    </ul>

    <h4>2. Contextual Modifications Using Valence Shifters</h4>
    <ul>
    <li><strong>Negators:</strong> Words like “not,” “never,” or “no” <strong>invert the polarity</strong> of a nearby sentiment word (e.g., 'not happy' changes from <strong>+1 to -1</strong>).</li>
    <li><strong>Amplifiers:</strong> Words such as 'very,' 'extremely,' and 'highly' <strong>increase the intensity</strong> of a sentiment (e.g., 'very good' is weighted more than 'good').</li>
    <li><strong>De-amplifiers (Diminishers):</strong> Terms like 'slightly' or 'somewhat' <strong>reduce sentiment intensity</strong> (e.g., 'slightly disappointing' has a weaker negative score than 'disappointing').</li>
    </ul>

    <h4>3. Aggregation and Normalization</h4>
    <ul>
    <li>Sentiment scores are <strong>summed across the document</strong> to obtain an <strong>overall polarity score</strong>.</li>
    <li>An <strong>optional normalization step</strong> scales the final score within the <strong>[-1, 1] range</strong>, ensuring comparability across different text lengths.</li>
    <li>Documents with scores near <strong>0</strong> are classified as <strong>neutral</strong>, indicating a balanced mix of sentiment or the absence of strong emotions.</li>
    </ul>
    <hr>
    <h4><strong>Sentiment Lexicons Used</strong></h4>

    <h4>1. Hu and Liu (2004) - Opinion Lexicon</h4>
    <ul>
    <li>Designed for analyzing <strong>consumer reviews</strong>, categorizing words into <strong>positive and negative</strong> classes.</li>
    <li>Particularly useful for <strong>e-commerce platforms, review aggregation sites, and user-generated feedback</strong>.</li>
    <li><strong>Language:</strong> English</li>
    </ul>

    <h4>2. Loughran and McDonald (2016) - Financial Sentiment Dictionary</h4>
    <ul>
    <li>Developed for <strong>financial and accounting texts</strong>, including categories such as <strong>“positive,” “negative,” “uncertainty,” “litigious,” and “constraining”</strong>.</li>
    <li>Widely used in <strong>financial risk assessment, investor sentiment analysis, and stock market forecasting</strong>.</li>
    <li><strong>Language:</strong> English</li>
    </ul>

    <h4>3. NRC Emotion Lexicon (Mohammad & Turney, 2010)</h4>
    <ul>
    <li>Captures emotions beyond basic polarity, categorizing words into <strong>eight primary emotions</strong>: Joy, Sadness, Anger, Fear, Surprise, Disgust, Trust, and Anticipation.</li>
    <li>Useful for <strong>social media mining, psychological studies, and literary analysis</strong>.</li>
    <li><strong>Language:</strong> Multilingual</li>
    </ul>
    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>

      <p><strong>Hu, M., & Liu, B.</strong></p>
      <p><i>Mining and summarizing customer reviews.</i> <strong>Proceedings of the Tenth ACM SIGKDD International Conference on Knowledge Discovery and Data Mining</strong>, KDD ’04, 168-177. New York, NY, USA: Association for Computing Machinery.</p>

      <p><strong>Loughran, T., & McDonald, B.</strong></p>
      <p><i>Textual analysis in accounting and finance: A survey.</i> <strong>Journal of Accounting Research</strong>, 54(4), 1187-1230.</p>

      <p><strong>Mohammad, S., & Turney, P.</strong></p>
      <p><i>Emotions evoked by common words and phrases: Using Mechanical Turk to create an emotion lexicon.</i> In <strong>Proceedings of the NAACL HLT 2010 Workshop on Computational Approaches to Analysis and Generation of Emotion in Text</strong>, 26-34. Los Angeles, CA: Association for Computational Linguistics.</p>
      </div>

      </body>"

  ## emotion analysis ----
  emotionanalysis <- "<body>

    <h3><strong>Emotion Detection</strong></h3>

    <p>Emotion analysis goes beyond simple positive/negative polarity by identifying <strong>specific emotions</strong> expressed in text.
    TALL uses the <strong>NRC Word-Emotion Association Lexicon (EmoLex)</strong> to detect eight primary emotions:
    <strong>Anger, Anticipation, Disgust, Fear, Joy, Sadness, Surprise, and Trust</strong>.</p>
    <hr>
    <h4><strong>How Emotion Detection Works</strong></h4>

    <h4>1. Lexicon-Based Emotion Detection</h4>
    <ul>
    <li>Each lemma in the text is matched against the <strong>NRC EmoLex</strong>, which provides binary associations between words and eight emotions.</li>
    <li>A single word can be associated with <strong>multiple emotions</strong> (e.g., 'abandon' may evoke both <strong>fear</strong> and <strong>sadness</strong>).</li>
    </ul>

    <h4>2. Document-Level Aggregation</h4>
    <ul>
    <li>For each document, emotion word counts are <strong>summed across all matched lemmas</strong>.</li>
    <li>Counts are <strong>normalized to proportions</strong>, allowing comparison across documents of different lengths.</li>
    </ul>

    <h4>3. Corpus-Level Summary</h4>
    <ul>
    <li>Emotion counts are aggregated across all documents to provide a <strong>corpus-level emotion profile</strong>.</li>
    <li>The <strong>Emotion Distribution</strong> chart shows the relative prevalence of each emotion in the corpus.</li>
    </ul>

    <h4>4. Word-Level Analysis</h4>
    <ul>
    <li>For each emotion, the <strong>most frequent contributing words</strong> are identified.</li>
    <li>This helps understand <strong>which vocabulary drives</strong> each emotion category.</li>
    </ul>
    <hr>
    <h4><strong>The NRC EmoLex</strong></h4>
    <ul>
    <li>Created by <strong>Saif Mohammad and Peter Turney</strong> using crowdsourcing via Amazon Mechanical Turk.</li>
    <li>Contains approximately <strong>14,000 words</strong> annotated with eight emotion categories.</li>
    <li>Available in <strong>multiple languages</strong> through automatic translation from the English source.</li>
    <li>Widely used in <strong>social media analysis, literary studies, psychological research, and opinion mining</strong>.</li>
    </ul>
    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>

      <p><strong>Mohammad, S.M., & Turney, P.D.</strong></p>
      <p><i>Crowdsourcing a word-emotion association lexicon.</i> <strong>Computational Intelligence</strong>, 29(3), 436-465, 2013.</p>

      <p><strong>Mohammad, S.M., & Turney, P.D.</strong></p>
      <p><i>Emotions evoked by common words and phrases: Using Mechanical Turk to create an emotion lexicon.</i>
      In <strong>Proceedings of the NAACL HLT 2010 Workshop on Computational Approaches to Analysis and Generation of Emotion in Text</strong>,
      26-34. Los Angeles, CA: Association for Computational Linguistics.</p>
      </div>

      </body>"

  ## summarzation ----
  summarization <- "<body>

    <h3><strong>Summarization</strong></h3>

    <p>Summarization is a <strong>key technique in text analysis</strong> that allows users to extract the <strong>most relevant information</strong> from a document
    while maintaining its core meaning.
    <br><strong>TALL implements extractive summarization</strong>, a method that selects and reorders the most important sentences <strong>directly from the original text</strong>
    to generate a <strong>coherent, condensed version</strong> of the content.</p>

    <p>Unlike <strong>abstractive summarization</strong>, which rephrases content using deep learning models, <strong>extractive summarization</strong> ensures that
    the summary remains <strong>factually consistent</strong> with the input document, making it a <strong>reliable method for automated text compression</strong>.</p>
    <hr>
    <h4><strong>How Summarization Works</strong></h4>

    <h4>1. Sentence Tokenization and Preprocessing</h4>
    <ul>
    <li>The text is split into <strong>individual sentences</strong> to form the basis of the summarization process.</li>
    <li>Sentences are <strong>preprocessed</strong>, removing unnecessary punctuation and stopwords to enhance <strong>semantic clarity</strong>.</li>
    </ul>

    <h4>2. Graph Construction Using Sentence Similarity</h4>
    <ul>
    <li>A <strong>graph-based representation</strong> of the document is created, where:</li>
    <ul>
    <li><strong>Nodes</strong> represent sentences.</li>
    <li><strong>Edges</strong> connect sentences based on their <strong>semantic similarity</strong> (measured using cosine similarity or word overlap).</li>
    </ul>
    <li>Sentences that share a <strong>high degree of lexical similarity</strong> are considered <strong>strongly connected</strong> in the graph.</li>
    </ul>

    <h4>3. Application of TextRank Algorithm</h4>
    <ul>
    <li>The <strong>TextRank algorithm</strong> assigns an <strong>importance score</strong> to each sentence based on its <strong>connectivity</strong> within the graph.</li>
    <li>Sentences with the <strong>highest PageRank scores</strong> are deemed <strong>the most representative</strong> of the overall document.</li>
    </ul>

    <h4>4. Sentence Selection and Ordering</h4>
    <ul>
    <li>The <strong>top-ranked sentences</strong> are selected for the summary.</li>
    <li>A <strong>reordering step</strong> ensures that sentences are presented in a <strong>logical and coherent structure</strong>, preserving the original document’s flow.</li>
    </ul>
    <hr>
    <h4><strong>Advantages of Summarization</strong></h4>
    <ul>
    <li><strong>Extractive and Factually Consistent</strong> – Ensures that summaries are directly sourced from the original text,
    <br>reducing the risk of hallucinations or misinterpretations.</li>
    <li><strong>Graph-Based Ranking for Objective Selection</strong> – Uses <strong>TextRank</strong>, an unsupervised method that
    <br><strong>ranks sentences purely based on semantic importance</strong>, eliminating bias.</li>
    <li><strong>Efficient and Scalable</strong> – Processes <strong>large documents quickly</strong>, making it ideal for summarizing
    <br><strong>research papers, news articles, legal documents, and reviews</strong>.</li>
    <li><strong>No Need for Pre-Trained Models</strong> – Unlike abstractive methods that require deep learning models,
    <br><strong>extractive summarization works effectively on any text without additional training</strong>.</li>
    <li><strong>Customizable Summary Length</strong> – Users can <strong>adjust the number of extracted sentences</strong> to control the
    <br><strong>level of detail</strong> in the summary.</li>
    </ul>
    <hr>
    <h4><strong>Implementation of Summarization</strong></h4>

    <p>TALL’s <strong>summarization routines</strong> are built upon the <strong>TextRank algorithm</strong>, with optimizations for handling
    <strong>preprocessed and structured corpora</strong>:</p>

    <ul>
    <li><strong>Customized Text Preprocessing</strong> – The system operates on <strong>tokenized, lemmatized, and PoS-tagged corpora</strong>,
    ensuring better sentence representation.</li>
    <li><strong>Sentence Similarity Based on Multiple Metrics</strong> – Supports <strong>TF-IDF, cosine similarity, and word embeddings</strong> for improved ranking.</li>
    <li><strong>Multi-Document Summarization (Future Work)</strong> – The framework is being expanded to support <strong>multi-document summarization</strong>,
    allowing users to extract summaries from <strong>multiple related texts</strong>.</li>
    </ul>

    <p>By integrating <strong>unsupervised graph-based techniques</strong>, TALL provides users with a <strong>robust and efficient summarization tool</strong>,
    ideal for <strong>academic, business, and legal applications</strong>.</p>
    <hr>
    <div class='references'>
      <h4><strong>References</strong></h4>

      <p><strong>Mihalcea, R., & Tarau, P.</strong></p>
      <p><i>TextRank: Bringing order into text.</i> <strong>Proceedings of the 2004 Conference on Empirical Methods in Natural Language Processing</strong>,
      404-411, Barcelona, Spain, July. Association for Computational Linguistics.</p>

      <p><strong>Page, L., Brin, S., Motwani, R., & Winograd, T.</strong></p>
      <p><i>The PageRank Citation Ranking: Bringing Order to the Web.</i> <strong>Technical report, Stanford Digital Library Technologies Project</strong>, 1998.</p>
      </div>

      </body>"

  return(list(
    importmenu = importmenu,
    split = split,
    random = random,
    externalinfo = externalinfo,
    tokenization = tokenization,
    specialentities = specialentities,
    multiwordcreation = multiwordcreation,
    multiwordlist = multiwordlist,
    customterm = customterm,
    synonyms = synonyms,
    posselection = posselection,
    featureroles = featureroles,
    overview = overview,
    keyness = keyness,
    wordincontext = wordincontext,
    reinert = reinert,
    correspondenceanalysis = correspondenceanalysis,
    cowordanalysis = cowordanalysis,
    thematicmap = thematicmap,
    embeddingtrain = embeddingtrain,
    embeddingsimilarity = embeddingsimilarity,
    tmkchoice = tmkchoice,
    tmmodelestimation = tmmodelestimation,
    syntacticcomplexity = syntacticcomplexity,
    svo = svo,
    polaritydetection = polaritydetection,
    emotionanalysis = emotionanalysis,
    summarization = summarization
  ))
}

CaoJuan2009 <- function(models) {
  metrics <- sapply(models, function(model) {
    # topic-word matrix
    m1 <- exp(model@beta)
    # pair-wise cosine distance
    pairs <- utils::combn(nrow(m1), 2)
    cos.dist <- apply(pairs, 2, function(pair) {
      x <- m1[pair[1], ]
      y <- m1[pair[2], ]
      # dist <- lsa::cosine(x, y)
      dist <- crossprod(x, y) / sqrt(crossprod(x) * crossprod(y))
      return(dist)
    })
    # metric
    metric <- sum(cos.dist) / (model@k * (model@k - 1) / 2)
    return(metric)
  })
  return(metrics)
}

Arun2010 <- function(models, dtm) {
  # length of documents (count of words)
  len <- Matrix::rowSums(dtm)
  # evaluate metrics
  metrics <- sapply(models, FUN = function(model) {
    # matrix M1 topic-word
    m1 <- exp(model@beta) # rowSums(m1) == 1
    m1.svd <- svd(m1)
    cm1 <- as.matrix(m1.svd$d)
    # matrix M2 document-topic
    m2 <- model@gamma # rowSums(m2) == 1
    cm2 <- len %*% m2 # crossprod(len, m2)
    norm <- norm(as.matrix(len), type = "m")
    cm2 <- as.vector(cm2 / norm)
    # symmetric Kullback-Leibler divergence
    divergence <- sum(cm1 * log(cm1 / cm2)) + sum(cm2 * log(cm2 / cm1))
    return(divergence)
  })
  return(metrics)
}

Deveaud2014 <- function(models) {
  metrics <- sapply(models, function(model) {
    # topic-word matrix
    m1 <- exp(model@beta)
    # prevent NaN
    if (any(m1 == 0)) {
      m1 <- m1 + .Machine$double.xmin
    }
    # pair-wise Jensen-Shannon divergence
    pairs <- utils::combn(nrow(m1), 2)
    jsd <- apply(pairs, 2, function(pair) {
      x <- m1[pair[1], ]
      y <- m1[pair[2], ]
      ### divergence by Deveaud2014
      jsd <- 0.5 * sum(x * log(x / y)) + 0.5 * sum(y * log(y / x))
      return(jsd)
    })

    # metric
    metric <- sum(jsd) / (model@k * (model@k - 1))
    return(metric)
  })
  return(metrics)
}

# Funzione per valutare un singolo modello
evaluate_single_k <- function(k, dtm, seed = 1234, method = "LDA") {
  if (method == "CTM") {
    model <- CTM(dtm, k = k, control = list(seed = seed))
  } else {
    model <- LDA(dtm, k = k, method = "VEM", control = list(seed = seed))
  }
  log_lik <- logLik(model)
  perp <- perplexity(model, newdata = dtm)
  return(list(k = k, logLik = log_lik, Perplexity = perp, model = model))
}

# Funzione parallela completa
evaluate_tm_parallel <- function(
  dtm,
  k_seq = 2:20,
  seed = 1234,
  n_cores = detectCores() - 1,
  method = "LDA"
) {
  cl <- parallel::makeCluster(n_cores)
  parallel::clusterEvalQ(cl, {
    library(topicmodels)
  })
  parallel::clusterExport(
    cl,
    varlist = c("dtm", "seed", "method", "evaluate_single_k"),
    envir = environment()
  )

  results <- parallel::parLapply(cl, k_seq, function(k) {
    evaluate_single_k(k, dtm, seed, method)
  })
  parallel::stopCluster(cl)

  # Estrai metriche e modelli
  metrics <- do.call(
    rbind,
    lapply(results, function(l) {
      data.frame(
        k = l$k,
        logLik = as.numeric(l$logLik),
        Perplexity = l$Perplexity
      )
    })
  )
  models <- setNames(lapply(results, function(x) x$model), paste0("k_", k_seq))

  metrics$CaoJuan2009 <- CaoJuan2009(models)
  metrics$Arun2010 <- Arun2010(models, dtm)
  metrics$Deveaud2014 <- Deveaud2014(models)

  return(list(metrics = metrics, models = models))
}

find_elbow <- function(k, metric, decreasing = TRUE, plot = TRUE) {
  # Normalizza i dati
  x <- as.numeric(scale(k))
  y <- as.numeric(scale(if (decreasing) -metric else metric))

  # Calcola distanza punto-linea per ogni punto
  point1 <- c(x[1], y[1])
  point2 <- c(x[length(x)], y[length(y)])
  line_vec <- point2 - point1
  line_vec_norm <- line_vec / sqrt(sum(line_vec^2))

  distances <- sapply(1:length(x), function(i) {
    p <- c(x[i], y[i])
    vec_from_line <- p - point1
    proj_len <- sum(vec_from_line * line_vec_norm)
    proj_point <- point1 + proj_len * line_vec_norm
    dist <- sqrt(sum((p - proj_point)^2))
    return(dist)
  })

  # Trova il massimo della distanza: è il gomito
  elbow_idx <- which.max(distances)
  elbow_k <- k[elbow_idx]

  if (plot) {
    plot(
      k,
      metric,
      type = "b",
      pch = 16,
      xlab = "Number of Topics (k)",
      ylab = "Metric",
      main = "Elbow Method"
    )
    points(k[elbow_idx], metric[elbow_idx], col = "red", pch = 19, cex = 1.5)
    legend("topright", legend = paste("k =", elbow_k), col = "red", pch = 19)
  }

  return(elbow_k)
}

tmTuning <- function(
  x,
  group = c("doc_id", "sentence_id"),
  term = "lemma",
  metric = c("CaoJuan2009", "Deveaud2014", "Arun2010", "Perplexity"),
  n = 100,
  top_by = c("freq", "tfidf"),
  minK = 2,
  maxK = 20,
  Kby = 1,
  method = "LDA",
  prevalence = NULL,
  seed = 1234
) {
  ## check min and max K
  ClusterRange <- sort(c(minK, maxK))
  minK <- ClusterRange[1]
  maxK <- ClusterRange[2]
  minK <- max(minK, 1)
  maxK <- min(maxK, length(unique(x$doc_id)))
  ###

  x$topic_level_id <- unique_identifier(x, fields = group)

  dtf <- document_term_frequencies(
    x,
    document = "topic_level_id",
    term = "lemma"
  )

  dtm <- document_term_matrix(x = dtf)

  if (method == "STM") {
    ## STM tuning using stm::searchK
    switch(
      top_by,
      freq = {
        dtm <- dtm_remove_lowfreq(dtm, minfreq = 1, maxterms = n)
      },
      tfidf = {
        dtm <- dtm_remove_tfidf(dtm, top = n)
      }
    )
    result <- stmTuning(
      x = x, dtm = dtm, group = group,
      prevalence = prevalence,
      minK = minK, maxK = maxK, Kby = Kby,
      seed = seed
    )
  } else {
    ## LDA / CTM tuning
    switch(
      top_by,
      freq = {
        dtm <- dtm_remove_lowfreq(dtm, minfreq = 1, maxterms = n)
        dtm <- tm::as.DocumentTermMatrix(dtm, weighting = tm::weightTf)
      },
      tfidf = {
        dtm <- dtm_remove_tfidf(dtm, top = n)
        dtm <- tm::as.DocumentTermMatrix(dtm, weighting = tm::weightTfIdf)
      }
    )
    result <- evaluate_tm_parallel(
      dtm,
      k_seq = seq(from = minK, to = maxK, by = Kby),
      seed = seed,
      n_cores = coresCPU(),
      method = method
    )
  }

  return(result)
}

tmTuningPlot <- function(result, metric) {
  df <- result$metrics %>%
    rename(topics = k)

  switch(
    metric,
    CaoJuan2009 = {
      bestT <- find_elbow(
        df$topics,
        df$CaoJuan2009,
        decreasing = TRUE,
        plot = FALSE
      )
    },
    Arun2010 = {
      bestT <- find_elbow(
        df$topics,
        df$Arun2010,
        decreasing = FALSE,
        plot = FALSE
      )
    },
    Deveaud2014 = {
      bestT <- find_elbow(
        df$topics,
        df$Deveaud2014,
        decreasing = TRUE,
        plot = FALSE
      )
    },
    Perplexity = {
      bestT <- find_elbow(
        df$topics,
        df$Perplexity,
        decreasing = TRUE,
        plot = FALSE
      )
    }
  )
  df <- df %>%
    select("topics", any_of(metric))
  names(df) <- c("x", "y")
  df <- df %>%
    mutate(
      y = (y - min(y)) / diff(range(y)),
      color = ifelse(x == bestT, "#cb453e", "#6CC283"),
      size = ifelse(x == bestT, 20, 10)
    )

  hoverText <- paste(
    " <b>Topic ",
    df$x,
    "</b>\n ",
    metric,
    ": ",
    round(df$y, 2),
    sep = ""
  )

  fig <- plot_ly(
    df,
    x = ~x,
    y = ~y,
    type = "scatter",
    mode = "markers+lines",
    line = list(color = "#6CC28360", width = 2),
    marker = list(
      size = ~size,
      color = ~color, #"#6CC283",
      line = list(
        color = ~color, #color = "#6CC283",
        width = 2
      )
    ),
    text = hoverText,
    hoverinfo = "text"
  ) %>%
    layout(
      annotations = list(
        text = paste0(
          "K selection by Elbow Method with ",
          metric,
          " metric: Optimal K: ",
          bestT
        ),
        xref = "paper",
        x = 0.5,
        yref = "paper",
        y = 1,
        yshift = 30,
        showarrow = FALSE,
        font = list(size = 24, color = "gray30")
      ),
      # title = paste0("K selection by ",metric," metric"),
      paper_bgcolor = "rgb(255,255,255)",
      plot_bgcolor = "rgb(255,255,255)",
      xaxis = list(
        title = "Topics",
        gridcolor = "rgb(229,229,229)",
        showgrid = TRUE,
        showline = FALSE,
        showticklabels = TRUE,
        tickcolor = "rgb(229,229,229)",
        ticks = "outside",
        zeroline = TRUE,
        range = c(0, max(df$x) + 1),
        dtick = 1,
        tick0 = 0
      ),
      yaxis = list(
        title = metric,
        gridcolor = "rgb(229,229,229)",
        showgrid = FALSE,
        showline = FALSE,
        showticklabels = TRUE,
        tickcolor = "rgb(229,229,229)",
        ticks = "inside",
        zeroline = TRUE,
        range = c(-0.02, 1.05),
        dtick = 0.20,
        tick0 = 0
      ),
      showlegend = FALSE
    )

  fig <- fig %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        #' toImage',
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )
  # %>%
  #   event_register("plotly_selecting")

  return(fig)
}

### Consensus K from multiple metrics ----
tmConsensusK <- function(metrics_df, method = "LDA") {
  if (method == "STM") {
    elbow_list <- list(
      list(col = "CaoJuan2009", decreasing = TRUE),
      list(col = "Arun2010", decreasing = FALSE),
      list(col = "Deveaud2014", decreasing = TRUE),
      list(col = "logLik", decreasing = FALSE)
    )
  } else {
    elbow_list <- list(
      list(col = "CaoJuan2009", decreasing = TRUE),
      list(col = "Arun2010", decreasing = FALSE),
      list(col = "Deveaud2014", decreasing = TRUE),
      list(col = "Perplexity", decreasing = TRUE)
    )
  }

  k_votes <- sapply(elbow_list, function(m) {
    if (!m$col %in% names(metrics_df)) return(NA)
    find_elbow(metrics_df$k, metrics_df[[m$col]], decreasing = m$decreasing, plot = FALSE)
  })
  k_votes <- k_votes[!is.na(k_votes)]

  # Mode (most frequent K)
  k_table <- sort(table(k_votes), decreasing = TRUE)
  as.integer(names(k_table)[1])
}

### Multi-Metric Comparison Plot ----

tmMultiMetricPlot <- function(result, method = "LDA") {
  df <- result$metrics %>% rename(topics = k)

  if (method == "STM") {
    metric_info <- list(
      list(col = "CaoJuan2009", label = "Exclusivity", color = "#e74c3c", decreasing = TRUE),
      list(col = "Arun2010", label = "Semantic Coherence", color = "#3498db", decreasing = FALSE),
      list(col = "Deveaud2014", label = "Excl. + Coherence", color = "#2ecc71", decreasing = TRUE),
      list(col = "logLik", label = "Lower Bound", color = "#9b59b6", decreasing = FALSE)
    )
  } else {
    metric_info <- list(
      list(col = "CaoJuan2009", label = "CaoJuan 2009", color = "#e74c3c", decreasing = TRUE),
      list(col = "Arun2010", label = "Arun 2010", color = "#3498db", decreasing = FALSE),
      list(col = "Deveaud2014", label = "Deveaud 2014", color = "#2ecc71", decreasing = TRUE),
      list(col = "Perplexity", label = "Perplexity", color = "#9b59b6", decreasing = TRUE)
    )
  }

  fig <- plot_ly()

  for (m in metric_info) {
    if (!m$col %in% names(df)) next
    vals <- df[[m$col]]
    # Min-max normalize to [0, 1]; invert if lower = better
    if (m$decreasing) {
      norm_vals <- (vals - min(vals)) / (max(vals) - min(vals) + .Machine$double.eps)
    } else {
      norm_vals <- 1 - (vals - min(vals)) / (max(vals) - min(vals) + .Machine$double.eps)
    }

    # Find elbow for this metric
    k_opt <- find_elbow(df$topics, vals, decreasing = m$decreasing, plot = FALSE)
    opt_idx <- which(df$topics == k_opt)

    fig <- fig %>%
      add_trace(
        x = df$topics,
        y = norm_vals,
        type = "scatter",
        mode = "lines+markers",
        name = m$label,
        line = list(color = m$color, width = 2),
        marker = list(size = 8, color = m$color),
        hovertemplate = paste0(
          "<b>", m$label, "</b><br>",
          "K = %{x}<br>",
          "Normalized: %{y:.3f}<br>",
          "<extra></extra>"
        )
      ) %>%
      add_trace(
        x = df$topics[opt_idx],
        y = norm_vals[opt_idx],
        type = "scatter",
        mode = "markers",
        marker = list(
          size = 16,
          color = m$color,
          line = list(color = "white", width = 3),
          symbol = "diamond"
        ),
        name = paste0(m$label, ": K=", k_opt),
        showlegend = FALSE,
        hovertemplate = paste0(
          "<b>", m$label, " (Optimal)</b><br>",
          "K = ", k_opt, "<br>",
          "<extra></extra>"
        )
      )
  }

  fig <- fig %>%
    layout(
      title = list(
        text = "<b>Multi-Metric Comparison</b>",
        font = list(size = 18, color = "gray30"),
        x = 0.5
      ),
      paper_bgcolor = "white",
      plot_bgcolor = "white",
      xaxis = list(
        title = "Number of Topics (K)",
        gridcolor = "rgb(229,229,229)",
        dtick = 1,
        showgrid = TRUE
      ),
      yaxis = list(
        title = "Normalized Score (0 = worst, 1 = best)",
        gridcolor = "rgb(229,229,229)",
        range = c(-0.05, 1.08),
        showgrid = TRUE
      ),
      legend = list(
        orientation = "h",
        x = 0.5,
        xanchor = "center",
        y = -0.12
      ),
      hovermode = "x unified"
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        "sendDataToCloud", "pan2d", "select2d", "lasso2d",
        "toggleSpikelines", "hoverClosestCartesian", "hoverCompareCartesian"
      )
    )

  return(fig)
}

### model estimation

tmEstimate <- function(
  x,
  K,
  group = c("doc_id", "sentence_id"),
  term = "lemma",
  n = 100,
  top_by = c("freq", "tfidf"),
  method = c("LDA", "CTM", "STM"),
  prevalence = NULL,
  seed = 1234
) {
  method <- match.arg(method)

  x$topic_level_id <- unique_identifier(x, fields = group)

  dtf <- document_term_frequencies(x, document = "topic_level_id", term = term)

  dtm <- document_term_matrix(x = dtf)

  switch(
    top_by,
    freq = {
      dtm <- dtm_remove_lowfreq(dtm, minfreq = 1, maxterms = n)
    },
    tfidf = {
      dtm <- dtm_remove_tfidf(dtm, top = n)
    }
  )

  if (method == "STM") {
    return(stmEstimate(x, dtm, K, group, prevalence, seed = seed))
  }

  # LDA / CTM
  if (method == "LDA") {
    topicModel <- LDA(dtm, K, method = "Gibbs", control = list(iter = 500, seed = seed))
  } else {
    topicModel <- CTM(dtm, K, control = list(seed = seed))
  }

  # Posterior distributions
  tmResult <- posterior(topicModel)

  # topics are probability distributions over the entire vocabulary
  beta <- tmResult$terms

  beta_norm <- beta / matrix(colSums(beta), K, ncol(beta), byrow = TRUE)
  beta_norm <- t(beta_norm) %>%
    as.data.frame() %>%
    mutate(word = colnames(beta_norm))

  variables <- as.character(1:K)
  beta <- t(beta) %>%
    as.data.frame() %>%
    mutate(word = colnames(beta)) %>%
    select(word, all_of(variables))

  # for every document we have a probability distribution of its contained topics
  row_label <- unique(x$doc_id)[as.numeric(row.names(tmResult$topics))]
  theta <- tmResult$topics %>%
    as.data.frame() %>%
    mutate(doc = row_label) %>%
    select(doc, all_of(variables))

  results <- list(
    topicModel = topicModel,
    tmResult = tmResult,
    beta = beta,
    beta_norm = beta_norm,
    theta = theta
  )

  return(results)
}


### STM (Structural Topic Model) functions ----

# Build prevalence metadata for STM from tokenized data
stmBuildMeta <- function(x, group, prevalence) {
  # Aggregate metadata at document level
  meta <- x %>%
    group_by(across(all_of(group))) %>%
    summarise(
      across(all_of(prevalence), ~ first(.x)),
      .groups = "drop"
    )
  # Ensure row order matches DTM (topic_level_id order)
  meta <- meta %>% arrange(across(all_of(group)))
  as.data.frame(meta)
}

# Self-contained LDA/CTM tuning for use inside a future worker.
# Runs sequentially (no parallel clusters) since it already runs in background.
# Only depends on package functions: topicmodels, slam, utils, stats.
tmTuningAsync <- function(dtm, k_seq, seed, method) {
  # Ensure topicmodels is fully loaded so S4 methods (logLik, perplexity) are registered
  requireNamespace("topicmodels", quietly = TRUE)
  library(topicmodels, quietly = TRUE)

  models_list <- list()
  metrics_rows <- list()

  for (i in seq_along(k_seq)) {
    k <- k_seq[i]
    if (method == "CTM") {
      model <- topicmodels::CTM(dtm, k = k, control = list(seed = seed))
    } else {
      model <- topicmodels::LDA(dtm, k = k, method = "VEM", control = list(seed = seed))
    }
    models_list[[paste0("k_", k)]] <- model
    metrics_rows[[i]] <- data.frame(
      k = k,
      logLik = as.numeric(logLik(model)),
      Perplexity = perplexity(model, newdata = dtm)
    )
  }

  metrics <- do.call(rbind, metrics_rows)

  # CaoJuan2009
  metrics$CaoJuan2009 <- sapply(models_list, function(model) {
    m1 <- exp(model@beta)
    pairs <- utils::combn(nrow(m1), 2)
    cos_dist <- apply(pairs, 2, function(pair) {
      x <- m1[pair[1], ]; y <- m1[pair[2], ]
      as.numeric(crossprod(x, y) / sqrt(crossprod(x) * crossprod(y)))
    })
    sum(cos_dist) / (model@k * (model@k - 1) / 2)
  })

  # Arun2010
  len <- if (inherits(dtm, "simple_triplet_matrix")) {
    slam::row_sums(dtm)
  } else {
    Matrix::rowSums(dtm)
  }
  metrics$Arun2010 <- sapply(models_list, function(model) {
    m1 <- exp(model@beta)
    m1_svd <- svd(m1)
    cm1 <- as.matrix(m1_svd$d)
    m2 <- model@gamma
    cm2 <- len %*% m2
    norm_val <- norm(as.matrix(len), type = "m")
    cm2 <- as.vector(cm2 / norm_val)
    sum(cm1 * log(cm1 / cm2)) + sum(cm2 * log(cm2 / cm1))
  })

  # Deveaud2014
  metrics$Deveaud2014 <- sapply(models_list, function(model) {
    m1 <- exp(model@beta)
    if (any(m1 == 0)) m1 <- m1 + .Machine$double.xmin
    pairs <- utils::combn(nrow(m1), 2)
    jsd <- apply(pairs, 2, function(pair) {
      x <- m1[pair[1], ]; y <- m1[pair[2], ]
      0.5 * sum(x * log(x / y)) + 0.5 * sum(y * log(y / x))
    })
    sum(jsd) / (model@k * (model@k - 1))
  })

  list(metrics = metrics, models = models_list)
}

# Self-contained STM tuning for use inside a future worker.
# Only depends on package functions: stm, tm.
stmTuningAsync <- function(dtm, minK, maxK, Kby, seed) {
  dtm_tm <- tm::as.DocumentTermMatrix(dtm, weighting = tm::weightTf)
  stm_data <- stm::readCorpus(dtm_tm, type = "dtm")
  k_seq <- seq(from = minK, to = maxK, by = Kby)

  search_result <- stm::searchK(
    documents = stm_data$documents,
    vocab = stm_data$vocab,
    K = k_seq,
    seed = seed,
    verbose = FALSE
  )

  sr <- search_result$results
  metrics <- data.frame(
    k = unlist(sr$K),
    logLik = unlist(sr$lbound),
    Perplexity = -unlist(sr$lbound),
    CaoJuan2009 = unlist(sr$exclus),
    Arun2010 = -unlist(sr$semcoh),
    Deveaud2014 = unlist(sr$exclus) + unlist(sr$semcoh)
  )

  list(metrics = metrics, models = NULL)
}

# STM tuning: uses stm::searchK
stmTuning <- function(
  x, dtm, group,
  prevalence = NULL,
  minK = 2, maxK = 20, Kby = 1,
  seed = 1234
) {
  # Convert DTM to DocumentTermMatrix (slam-based) then to stm format
  dtm_tm <- tm::as.DocumentTermMatrix(dtm, weighting = tm::weightTf)
  row_tot <- slam::row_sums(dtm_tm)
  keep_docs <- rownames(dtm_tm)[row_tot > 0]
  dtm_tm <- dtm_tm[row_tot > 0, ]
  stm_data <- stm::readCorpus(dtm_tm, type = "dtm")

  # Build prevalence formula and metadata
  prev_formula <- NULL
  meta <- NULL
  if (!is.null(prevalence) && length(prevalence) > 0) {
    meta <- stmBuildMeta(x, "topic_level_id", prevalence)
    meta <- meta[as.character(meta$topic_level_id) %in% keep_docs, , drop = FALSE]
    meta <- meta[match(keep_docs, as.character(meta$topic_level_id)), , drop = FALSE]
    prev_formula <- as.formula(paste("~", paste(paste0("`", prevalence, "`"), collapse = " + ")))
  }

  k_seq <- seq(from = minK, to = maxK, by = Kby)

  # Use stm::searchK for model selection
  search_result <- stm::searchK(
    documents = stm_data$documents,
    vocab = stm_data$vocab,
    K = k_seq,
    prevalence = prev_formula,
    data = meta,
    seed = seed,
    verbose = FALSE
  )

  # Extract metrics into a compatible format
  sr <- search_result$results
  metrics <- data.frame(
    k = unlist(sr$K),
    logLik = unlist(sr$lbound),
    Perplexity = -unlist(sr$lbound),
    CaoJuan2009 = unlist(sr$exclus),
    Arun2010 = -unlist(sr$semcoh),
    Deveaud2014 = unlist(sr$exclus) + unlist(sr$semcoh)
  )

  # Normalize to make compatible with existing plotting code
  # CaoJuan2009 -> use exclusivity (higher = better, elbow finds minimum so invert)
  # Arun2010 -> use -semantic coherence (lower = better)
  # Deveaud2014 -> use combined (higher = better)

  return(list(metrics = metrics, models = NULL))
}

# STM estimation
stmEstimate <- function(x, dtm, K, group, prevalence = NULL, seed = 1234) {
  # Convert DTM to DocumentTermMatrix (slam-based) then to stm format
  dtm_tm <- tm::as.DocumentTermMatrix(dtm, weighting = tm::weightTf)
  # Drop empty documents so meta and docs stay aligned with stm
  row_tot <- slam::row_sums(dtm_tm)
  keep_docs <- rownames(dtm_tm)[row_tot > 0]
  dtm_tm <- dtm_tm[row_tot > 0, ]
  stm_data <- stm::readCorpus(dtm_tm, type = "dtm")

  # Build prevalence formula and metadata
  prev_formula <- NULL
  meta <- NULL
  if (!is.null(prevalence) && length(prevalence) > 0) {
    meta <- stmBuildMeta(x, "topic_level_id", prevalence)
    meta <- meta[as.character(meta$topic_level_id) %in% keep_docs, , drop = FALSE]
    meta <- meta[match(keep_docs, as.character(meta$topic_level_id)), , drop = FALSE]
    prev_formula <- as.formula(paste("~", paste(paste0("`", prevalence, "`"), collapse = " + ")))
  }

  # Fit STM model
  topicModel <- stm::stm(
    documents = stm_data$documents,
    vocab = stm_data$vocab,
    K = K,
    prevalence = prev_formula,
    data = meta,
    seed = seed,
    verbose = FALSE
  )

  # Extract beta (term-topic probabilities)
  # stm stores log beta in model$beta$logbeta
  beta_raw <- exp(topicModel$beta$logbeta[[1]])
  # beta_raw: K rows x V cols
  colnames(beta_raw) <- stm_data$vocab
  rownames(beta_raw) <- as.character(1:K)

  variables <- as.character(1:K)

  # beta: words x topics (same format as LDA output)
  beta <- as.data.frame(t(beta_raw))
  colnames(beta) <- variables
  beta$word <- stm_data$vocab
  beta <- beta %>% select(word, all_of(variables))

  # beta_norm: normalized version
  col_sums <- colSums(beta[, variables, drop = FALSE])
  beta_norm <- beta
  beta_norm[, variables] <- sweep(beta[, variables, drop = FALSE], 2, col_sums, "/")

  # Extract theta (document-topic probabilities)
  theta_raw <- topicModel$theta
  # Get document labels
  doc_ids <- x %>%
    distinct(topic_level_id, doc_id) %>%
    arrange(topic_level_id)
  row_label <- doc_ids$doc_id[seq_len(nrow(theta_raw))]

  theta <- as.data.frame(theta_raw)
  colnames(theta) <- variables
  theta$doc <- row_label
  theta <- theta %>% select(doc, all_of(variables))

  results <- list(
    topicModel = topicModel,
    tmResult = list(terms = beta_raw, topics = theta_raw),
    beta = beta,
    beta_norm = beta_norm,
    theta = theta,
    stm_meta = meta,
    stm_documents = stm_data$documents
  )

  return(results)
}


## hellinger distance ----
hellinger <- function(beta) {
  beta <- sqrt(beta)
  B <- matrix(NA, ncol(beta), ncol(beta))
  for (i in 1:ncol(beta)) {
    for (j in i:ncol(beta)) {
      B[i, j] <- sum((beta[, i] - beta[, j])^2)
    }
  }

  H <- sqrt(B) * (1 / sqrt(2))
}

tmNetwork <- function(beta, minEdge) {
  beta <- as.matrix(results$beta[, -1])

  H <- 1 - hellinger(beta)
  diag(H) <- NA

  topics <- paste0("Topic_", seq(1, nrow(H)))

  H <- data.frame(H)
  colnames(H) <- topics
  H$from <- topics

  H <- H %>%
    pivot_longer(
      cols = 1:length(topics),
      names_to = "to",
      values_to = "size"
    ) %>%
    drop_na()
  edges <- H %>%
    mutate(size = size * 10) %>%
    filter(size > minEdge * 10) %>%
    drop_na()

  nodes <- data.frame(
    id = topics,
    size = 10,
    color = "#4F7942",
    title = topics,
    label = topics,
    font.color = adjustcolor("black", alpha.f = 0.6)
  )

  VIS <- visNetwork::visNetwork(
    nodes = nodes,
    edges = edges,
    type = "full",
    smooth = TRUE,
    physics = FALSE
  ) %>%
    visNetwork::visIgraphLayout(layout = "layout_nicely", type = "full") %>%
    visNetwork::visEdges(smooth = list(type = "horizontal")) %>%
    visNetwork::visOptions(
      highlightNearest = list(enabled = T, hover = T, degree = 1),
      nodesIdSelection = T
    ) %>%
    visNetwork::visInteraction(
      dragNodes = TRUE,
      navigationButtons = F,
      hideEdgesOnDrag = TRUE,
      zoomSpeed = 0.2
    ) %>%
    visNetwork::visOptions(
      manipulation = FALSE,
      height = "100%",
      width = "100%"
    )

  results <- list(H = H %>% rename(value = size), VIS = VIS)
  return(results)
}

tmHeatmap <- function(beta) {
  # Calculate correlation matrix
  data <- cor(as.matrix(beta[, -1]))
  diag(data) <- 0

  df <- data.frame(data)

  # Create topic labels - ordinamento naturale
  n_topics <- nrow(data)
  id <- sprintf(paste0("%0", nchar(n_topics), "d"), 1:n_topics)
  x <- colnames(df) <- paste0("topic ", id)
  y <- row.names(df) <- paste0("topic ", id)

  # Prepare data for plotting
  df_long <- df %>%
    rownames_to_column("y") %>%
    pivot_longer(
      cols = starts_with("topic "),
      names_to = "variable",
      values_to = "value"
    ) %>%
    mutate(
      value = round(value, 3),
      abs_value = abs(value),
      topic_y_num = as.numeric(gsub("topic ", "", y)),
      topic_x_num = as.numeric(gsub("topic ", "", variable))
    )

  # Create mini scatterplot data for each cell
  scatter_data_list <- list()
  beta_matrix <- as.matrix(beta[, -1])
  n_topics_beta <- ncol(beta_matrix)

  for (i in 1:nrow(df_long)) {
    row_info <- df_long[i, ]
    topic_y <- row_info$topic_y_num
    topic_x <- row_info$topic_x_num

    if (topic_x != topic_y) {
      # Sample points for scatter (max 50 points per cell)
      n_points <- min(50, nrow(beta_matrix))
      sample_idx <- sample(1:nrow(beta_matrix), n_points)

      scatter_data_list[[i]] <- data.frame(
        cell_id = i,
        x_pos = row_info$variable,
        y_pos = row_info$y,
        x_cat_num = row_info$topic_x_num,
        y_cat_num = row_info$topic_y_num,
        scatter_x = beta_matrix[sample_idx, topic_x],
        scatter_y = beta_matrix[sample_idx, topic_y],
        correlation = row_info$value
      )
    }
  }

  scatter_data <- bind_rows(scatter_data_list)

  # Modern color palette
  pal <- colorRampPalette(c(
    "#d73027",
    "#f46d43",
    "#fdae61",
    "#fee090",
    "#ffffbf",
    "#e0f3f8",
    "#abd9e9",
    "#74add1",
    "#4575b4"
  ))(100)

  # PLOTLY VERSION - Interactive with borders and scatter
  Hplot <- plot_ly()

  # Add heatmap base - topic 1 at top
  Hplot <- Hplot %>%
    add_trace(
      data = df_long,
      x = ~ topic_x_num - 1,
      y = ~ topic_y_num - 1,
      z = ~value,
      type = "heatmap",
      colorscale = list(
        list(0, "#d73027"),
        list(0.25, "#fdae61"),
        list(0.5, "#ffffbf"),
        list(0.75, "#abd9e9"),
        list(1, "#4575b4")
      ),
      zmin = -1,
      zmax = 1,
      hovertemplate = "Correlation: %{z:.3f}<extra></extra>",
      # hovertemplate = paste(
      #   "<b>%{customdata[0]} vs %{customdata[1]}</b><br>",
      #   "Correlation: %{z:.3f}<br>",
      #   "<extra></extra>"
      # ),
      customdata = ~ cbind(y, variable),
      showscale = TRUE,
      xgap = 3,
      ygap = 3,
      colorbar = list(
        title = "Correlation",
        titleside = "right",
        tickmode = "linear",
        tick0 = -1,
        dtick = 0.5,
        len = 0.7,
        thickness = 15,
        x = 1.02
      )
    )

  # Add scatter points for each non-diagonal cell
  if (nrow(scatter_data) > 0) {
    for (cell_id in unique(scatter_data$cell_id)) {
      cell_scatter <- scatter_data[scatter_data$cell_id == cell_id, ]

      # Get cell position
      x_pos <- unique(cell_scatter$x_cat_num) - 1
      y_pos <- unique(cell_scatter$y_cat_num) - 1

      # Normalize scatter values to fit within cell (±0.35 offset)
      x_vals <- cell_scatter$scatter_x
      if (length(unique(x_vals)) > 1) {
        x_norm <- (x_vals - mean(x_vals)) / (max(x_vals) - min(x_vals))
        x_norm <- x_norm * 0.35
      } else {
        x_norm <- rep(0, length(x_vals))
      }

      y_vals <- cell_scatter$scatter_y
      if (length(unique(y_vals)) > 1) {
        y_norm <- (y_vals - mean(y_vals)) / (max(y_vals) - min(y_vals))
        y_norm <- y_norm * 0.35
      } else {
        y_norm <- rep(0, length(y_vals))
      }

      # Add scatter trace
      Hplot <- Hplot %>%
        add_markers(
          x = x_pos + x_norm,
          y = y_pos + y_norm,
          marker = list(
            size = 4,
            color = "rgba(50, 50, 50, 0.3)",
            line = list(width = 0)
          ),
          hoverinfo = "skip",
          showlegend = FALSE
        )
    }
  }

  # Add text annotations - POSITIONED AT TOP OF CELL
  Hplot <- Hplot %>%
    add_annotations(
      data = df_long,
      x = ~ topic_x_num - 1,
      y = ~ topic_y_num - 1,
      text = ~ ifelse(value == 0, "—", as.character(value)),
      xref = "x",
      yref = "y",
      yshift = 25,
      showarrow = FALSE,
      font = list(
        color = ~ ifelse(abs(value) > 0.5, "white", "black"),
        size = 14,
        family = "Arial, sans-serif",
        weight = "bold"
      )
    ) %>%
    layout(
      title = list(
        text = "<b>Topic Correlation Matrix</b>",
        font = list(size = 18, color = "#2c3e50"),
        x = 0.5,
        xanchor = "center"
      ),
      xaxis = list(
        title = "",
        tickmode = "array",
        tickvals = 0:(n_topics - 1),
        ticktext = x,
        tickangle = -45,
        tickfont = list(
          size = 11,
          family = "Arial, sans-serif",
          color = "black"
        ),
        side = "bottom",
        showgrid = FALSE,
        zeroline = FALSE
      ),
      yaxis = list(
        title = "",
        tickmode = "array",
        tickvals = 0:(n_topics - 1),
        ticktext = y, # NOT reversed - use natural order
        tickfont = list(
          size = 11,
          family = "Arial, sans-serif",
          color = "black"
        ),
        showgrid = FALSE,
        zeroline = FALSE,
        autorange = "reversed" # This makes y=0 appear at top
      ),
      plot_bgcolor = "#f8f9fa",
      paper_bgcolor = "white",
      margin = list(l = 100, r = 120, t = 80, b = 100),
      hovermode = "closest"
    ) %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian",
        "autoScale2d",
        "zoom2d"
      )
    )

  # GGPLOT2 VERSION - Static with scatterplots in cells
  df_long$y <- factor(df_long$y, levels = rev(y))
  df_long$variable <- factor(df_long$variable, levels = x)

  if (nrow(scatter_data) > 0) {
    scatter_data$y_pos <- factor(scatter_data$y_pos, levels = rev(y))
    scatter_data$x_pos <- factor(scatter_data$x_pos, levels = x)
  }

  HplotStatic <- ggplot() +
    # Base heatmap
    geom_tile(
      data = df_long,
      aes(x = variable, y = y, fill = value),
      color = "white",
      size = 1.5
    ) +
    # Add mini scatterplots in each cell
    {
      if (nrow(scatter_data) > 0) {
        geom_point(
          data = scatter_data,
          aes(
            x = as.numeric(x_pos) +
              (scatter_x - mean(scatter_x)) /
                (max(scatter_x) - min(scatter_x) + 0.001) *
                0.35,
            y = as.numeric(y_pos) +
              (scatter_y - mean(scatter_y)) /
                (max(scatter_y) - min(scatter_y) + 0.001) *
                0.35
          ),
          alpha = 0.3,
          size = 0.8,
          color = "gray30"
        )
      }
    } +
    # Add text annotations - POSITIONED AT TOP OF CELL
    geom_text(
      data = df_long,
      aes(
        x = variable,
        y = y,
        label = ifelse(value == 0, "—", value),
        color = abs(value) > 0.5
      ),
      size = 4.5,
      fontface = "bold",
      family = "sans",
      vjust = -0.5
    ) +
    # Color scales
    scale_fill_gradientn(
      colors = c("#d73027", "#fdae61", "#ffffbf", "#abd9e9", "#4575b4"),
      limits = c(-1, 1),
      breaks = seq(-1, 1, 0.5),
      guide = guide_colorbar(
        title = "Correlation",
        title.position = "top",
        title.hjust = 0.5,
        barwidth = 1,
        barheight = 15,
        frame.colour = "black",
        ticks.colour = "black"
      )
    ) +
    scale_color_manual(
      values = c("TRUE" = "white", "FALSE" = "black"),
      guide = "none"
    ) +
    # Theme and styling
    theme_minimal(base_size = 12) +
    theme(
      plot.title = element_text(
        hjust = 0.5,
        face = "bold",
        size = 16,
        color = "#2c3e50",
        margin = margin(b = 15)
      ),
      axis.text.x = element_text(
        angle = 45,
        hjust = 1,
        vjust = 1,
        size = 10,
        face = "bold"
      ),
      axis.text.y = element_text(
        size = 10,
        face = "bold"
      ),
      axis.title = element_blank(),
      panel.grid = element_blank(),
      panel.border = element_blank(),
      plot.background = element_rect(fill = "white", color = NA),
      panel.background = element_rect(fill = "#f8f9fa", color = NA),
      legend.position = "right",
      legend.title = element_text(size = 11, face = "bold"),
      legend.text = element_text(size = 9),
      plot.margin = margin(20, 20, 20, 20)
    ) +
    labs(
      title = "Topic Correlation Matrix with Data Distribution"
    ) +
    coord_fixed(ratio = 1)

  return(list(
    Hplot = Hplot,
    HplotStatic = HplotStatic
  ))
}

tmTopicPlot <- function(beta, topic = 1, nPlot = 10) {
  dfPlot <- beta %>%
    select(word, any_of(as.character(topic)))
  names(dfPlot)[2] <- "y"
  dfPlot <- dfPlot %>%
    arrange(desc(y)) %>%
    slice_max(y, n = nPlot, with_ties = FALSE) %>%
    mutate(
      y = y + runif(nPlot, 0, 1) / (10^7),
      word = factor(word, levels = unique(word)[order(y, decreasing = FALSE)])
    )

  fig <- freqPlotly(
    dfPlot,
    x = "y",
    y = "word",
    n = nPlot,
    ylabel = "Words",
    xlabel = "Beta Probability",
    scale = "identity",
    topicmodel = TRUE,
    colorlist()[topic],
    decimal = 4
  ) %>%
    layout(
      annotations = list(
        text = paste0("Topic ", topic),
        xref = "paper",
        x = 0.5,
        yref = "paper",
        y = 1,
        yshift = 30,
        showarrow = FALSE,
        font = list(size = 24, color = "gray30")
      )
    )

  fig <- fig %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        #' toImage',
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )
  # %>%
  #   event_register("plotly_selecting")

  return(fig)
}

tmDocPlot <- function(theta, topic = 1, nPlot = 10) {
  nPlot <- min(nPlot, nrow(theta))
  dfPlot <- theta %>%
    select(doc, any_of(as.character(topic)))
  names(dfPlot)[2] <- "y"
  dfPlot <- dfPlot %>%
    arrange(desc(y)) %>%
    slice_max(y, n = nPlot, with_ties = FALSE) %>%
    mutate(
      y = y + runif(nPlot, 0, 1) / (10^7),
      doc = factor(doc, levels = unique(doc)[order(y, decreasing = FALSE)])
    )

  fig <- freqPlotly(
    dfPlot,
    x = "y",
    y = "doc",
    n = nPlot,
    ylabel = "Documents",
    xlabel = "Theta Probability",
    scale = "identity",
    topicmodel = TRUE,
    colorlist()[topic],
    decimal = 4
  ) %>%
    layout(
      annotations = list(
        text = paste0("Topic ", topic),
        xref = "paper",
        x = 0.5,
        yref = "paper",
        y = 1,
        yshift = 30,
        showarrow = FALSE,
        font = list(size = 24, color = "gray30")
      )
    )

  fig <- fig %>%
    config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        #' toImage',
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    ) %>%
    event_register("plotly_selecting")

  return(fig)
}

### POLARITY DETECTION ----

# download sentiment lexicons

utils::globalVariables(c(
  "IS",
  "IS_norm",
  "n_length",
  "n_lexical",
  "ngram",
  "pos_sequence",
  "sentence_id",
  "sum_reciprocal_freq",
  "term_col",
  "token_id",
  "token_id words",
  "words"
))

#' Calculate IS index for n-grams
#'
#' This function calculates the IS (Absorption Index) from Morrone (1996)
#' for all n-grams in the corpus. Only n-grams that start AND end with
#' lexical words are considered.
#'
#' @param dfTag A data frame with tagged text data containing columns: doc_id,
#'   sentence_id, token_id, lemma/token, upos
#' @param max_ngram Maximum length of n-grams to generate (default: 5)
#' @param term Character string indicating which column to use: "lemma" or "token" (default: "lemma")
#' @param pos Character vector of POS tags considered lexical (default: c("NOUN", "ADJ", "ADV", "VERB"))
#' @param min_freq Minimum frequency threshold for n-grams (default: 1)
#' @param min_IS_norm Minimum normalized IS threshold for n-grams (default: 0)
#'
#' @return A tibble with columns: ngram, n_length, ngram_freq, n_lexical, IS, IS_norm
#'
#' @details
#' The IS index is calculated as: IS = (sum 1/freq_i) × freq_ngram × n_lexical
#' where freq_i is the frequency of each word in the n-gram, freq_ngram is the
#' frequency of the n-gram, and n_lexical is the number of lexical words.
#' IS_norm is the normalized version: IS / L^2 where L is the n-gram length.
#'
#' OPTIMIZATION: Only n-grams that start AND end with lexical words (as defined by
#' the 'pos' parameter) are generated, significantly reducing computation time.
#'
#' @export
#' @examples
#' \dontrun{
#' IS <- calculate_ngram_is(dfTag, max_ngram = 4, term = "lemma", min_freq = 2)
#' head(IS)
#' }
calculate_ngram_is <- function(
  dfTag,
  max_ngram = 5,
  term = "lemma",
  pos = c("NOUN", "ADJ", "ADV", "VERB"),
  min_freq = 1,
  min_IS_norm = 0
) {
  include_pos <- c(
    "DET",
    "NOUN",
    "PRON",
    "ADV",
    "VERB",
    "ADP",
    "AUX",
    "ADJ",
    "CCONJ",
    "INTJ",
    "PART",
    "PROPN",
    "SCONJ"
  )

  # Filter valid tokens
  df_filtered <- dfTag %>%
    dplyr::filter(upos %in% include_pos, !is.na(!!sym(term))) %>%
    select(doc_id, sentence_id, token_id, term_col = !!sym(term), upos) %>%
    arrange(doc_id, sentence_id, as.numeric(token_id))

  # Calculate global word frequencies
  word_freq_df <- df_filtered %>%
    count(term_col, name = "word_freq")

  # Prepare sentences as lists
  sentences <- df_filtered %>%
    group_by(doc_id, sentence_id) %>%
    summarise(
      terms = list(term_col),
      pos_tags = list(upos),
      .groups = "drop"
    )

  # Generate n-grams using C++ function
  # OPTIMIZATION: Pass lexical_pos to filter n-grams at generation time
  ngrams_list <- generate_ngrams_cpp(
    sentences_terms = sentences$terms,
    sentences_pos = sentences$pos_tags,
    max_ngram = max_ngram,
    lexical_pos = pos # Pass lexical POS for filtering
  )

  # Convert to tibble
  ngrams_df <- tibble(
    ngram = ngrams_list$ngram,
    words = ngrams_list$words,
    pos_sequence = ngrams_list$pos_sequence,
    n_length = ngrams_list$n_length
  )

  # Calculate frequencies and filter
  ngram_counts <- ngrams_df %>%
    count(ngram, n_length, name = "ngram_freq") %>%
    dplyr::filter(ngram_freq >= min_freq)

  if (nrow(ngram_counts) == 0) {
    return(tibble(
      ngram = character(),
      n_length = integer(),
      ngram_freq = integer(),
      n_lexical = integer(),
      IS = numeric(),
      IS_norm = numeric()
    ))
  }

  # Get metadata
  ngram_metadata <- ngrams_df %>%
    distinct(ngram, n_length, .keep_all = TRUE) %>%
    select(ngram, n_length, words, pos_sequence)

  ngram_freq <- ngram_counts %>%
    left_join(ngram_metadata, by = c("ngram", "n_length"))

  # Calculate IS components using C++ functions
  ngram_freq$n_lexical <- count_lexical_cpp(
    ngram_pos = ngram_freq$pos_sequence,
    lexical_pos = pos
  )

  ngram_freq$sum_reciprocal_freq <- calc_reciprocal_sum_cpp(
    ngram_words = ngram_freq$words,
    word_freq_names = word_freq_df$term_col,
    word_freq_values = word_freq_df$word_freq
  )

  # Final calculations
  ngram_freq <- ngram_freq %>%
    mutate(
      IS = sum_reciprocal_freq * ngram_freq * n_lexical,
      IS_norm = IS / (n_length^2)
    ) %>%
    dplyr::filter(IS_norm >= min_IS_norm) %>%
    select(ngram, n_length, ngram_freq, n_lexical, IS, IS_norm) %>%
    arrange(desc(IS_norm))

  return(ngram_freq)
}

test_that("calculate_ngram_is returns expected structure", {
  data(mobydick, package = "tall")
  result <- calculate_ngram_is(mobydick, max_ngram = 3, term = "lemma", min_freq = 2)

  expect_s3_class(result, "tbl_df")
  expect_named(result, c("ngram", "n_length", "ngram_freq", "n_lexical", "IS", "IS_norm"))
  expect_true(nrow(result) > 0)
  expect_true(all(result$ngram_freq >= 2))
  expect_true(all(result$n_length <= 3))
  expect_true(all(result$IS_norm >= 0))
})

test_that("calculate_ngram_is respects min_freq filter", {
  data(mobydick, package = "tall")
  result_low <- calculate_ngram_is(mobydick, max_ngram = 2, min_freq = 1)
  result_high <- calculate_ngram_is(mobydick, max_ngram = 2, min_freq = 5)

  expect_true(nrow(result_low) >= nrow(result_high))
  if (nrow(result_high) > 0) {
    expect_true(all(result_high$ngram_freq >= 5))
  }
})

test_that("calculate_ngram_is respects max_ngram", {
  data(mobydick, package = "tall")
  result <- calculate_ngram_is(mobydick, max_ngram = 2, min_freq = 2)

  expect_true(all(result$n_length <= 2))
})

test_that("calculate_ngram_is returns empty tibble with impossible filter", {
  data(mobydick, package = "tall")
  result <- calculate_ngram_is(mobydick, max_ngram = 2, min_freq = 999999)

  expect_s3_class(result, "tbl_df")
  expect_equal(nrow(result), 0)
  expect_named(result, c("ngram", "n_length", "ngram_freq", "n_lexical", "IS", "IS_norm"))
})

w2vTraining <- function(x, term = "lemma", dim = 100, iter = 20) {
  # Filter tokens excluding those with upos PUNCT or X
  x <- x %>%
    filter(!upos %in% c("PUNCT", "X")) %>%
    mutate(id = paste0(doc_id, "_", sentence_id))

  stopwords <- x %>%
    filter(upos %in% c("AUX", "DET", "ADP", "CCONJ", "SCONJ", "INTJ")) %>%
    pull(!!sym(term)) %>%
    unique()

  # Group by sentence_id and create a list of token/lemma vectors.
  word_list <- split(x %>% select(any_of(term)) %>% pull() %>% tolower(), x$id)

  w2v_model <- word2vec(
    x = word_list,
    type = "cbow",
    dim = dim,
    iter = iter,
    stopwords = stopwords,
    threads = coresCPU()
  )
  return(w2v_model)
}

summary_stats_embeddings <- function(embedding_matrix, as_tibble = TRUE) {
  # Skewness personalizzata (corretta per bias)
  skewness_custom <- function(x) {
    m <- mean(x)
    s <- sd(x)
    n <- length(x)
    if (s == 0) {
      return(0)
    }
    sum(((x - m) / s)^3) * (n / ((n - 1) * (n - 2)))
  }

  # Kurtosis personalizzata (excess kurtosis corretta per bias)
  kurtosis_custom <- function(x) {
    m <- mean(x)
    s <- sd(x)
    n <- length(x)
    if (s == 0) {
      return(0)
    }
    term1 <- sum(((x - m) / s)^4) *
      (n * (n + 1)) /
      ((n - 1) * (n - 2) * (n - 3))
    term2 <- (3 * (n - 1)^2) / ((n - 2) * (n - 3))
    return(term1 - term2)
  }

  stats <- data.frame(
    Mean = apply(embedding_matrix, 2, mean),
    Median = apply(embedding_matrix, 2, median),
    SD = apply(embedding_matrix, 2, sd),
    Min = apply(embedding_matrix, 2, min),
    P25 = apply(embedding_matrix, 2, quantile, probs = 0.25),
    P75 = apply(embedding_matrix, 2, quantile, probs = 0.75),
    Max = apply(embedding_matrix, 2, max),
    Range = apply(embedding_matrix, 2, function(x) max(x) - min(x)),
    Skewness = apply(embedding_matrix, 2, skewness_custom),
    Kurtosis = apply(embedding_matrix, 2, kurtosis_custom)
  )

  # Se richiesto, restituisci una tibble (se disponibile)
  if (as_tibble && "tibble" %in% rownames(installed.packages())) {
    stats <- tibble::as_tibble(stats, rownames = "Dimension")
  } else {
    stats$Dimension <- rownames(stats)
    stats <- stats[, c("Dimension", setdiff(names(stats), "Dimension"))]
  }

  return(stats)
}


## cosine among matrix vectors
distance_similarity_stats <- function(embedding_matrix) {
  n <- nrow(embedding_matrix)

  # --- Distanza euclidea ---
  dist_euclidean <- dist(embedding_matrix, method = "euclidean")
  mean_euclidean <- mean(as.vector(dist_euclidean))

  # --- Similarità coseno manuale ---
  cosine_similarity_matrix <- function(mat) {
    # Numeratore: prodotto scalare tra righe
    dot_products <- mat %*% t(mat)
    # Denominatore: norme
    norms <- sqrt(rowSums(mat^2))
    denom <- outer(norms, norms)
    sim <- dot_products / denom
    sim[is.na(sim)] <- 0 # nel caso ci siano divisioni per zero
    return(sim)
  }

  cosine_matrix <- cosine_similarity_matrix(embedding_matrix)
  # Consideriamo solo i valori nella parte superiore senza la diagonale
  cosine_values <- cosine_matrix[upper.tri(cosine_matrix)]
  mean_cosine <- mean(cosine_values, na.rm = TRUE)

  return(list(
    Mean_Euclidean_Distance = mean_euclidean,
    Mean_Cosine_Similarity = mean_cosine
  ))
}


pca_analysis_embeddings <- function(embedding_matrix) {
  pca <- prcomp(embedding_matrix, center = TRUE, scale. = TRUE)
  var_explained <- summary(pca)$importance[2, ]
  return(var_explained)
}

## WORD EMBEDDING SIMILARITY ----
w2vNetwork <- function(w2v_model, dfTag, term, n = 100) {
  w2v_matrix <- as.matrix(w2v_model)

  ## similarity
  top_words <- dfTag %>%
    filter(docSelected) %>%
    filter(upos %in% c("NOUN", "PROPN", "ADJ")) %>%
    group_by(!!sym(term)) %>%
    summarize(n = n()) %>%
    arrange(desc(n)) %>%
    slice_head(n = n) %>%
    pull(!!sym(term)) %>%
    tolower()

  # remove top_words felt in the stop_word list
  top_words <- intersect(top_words, row.names(as.matrix(w2v_model)))

  similarity <- predict(w2v_model, newdata = top_words)
  df_similarity <- bind_rows(similarity) %>%
    select(-rank) %>%
    rename(from = term1, to = term2)

  # Nodi unici
  nodes <- data.frame(
    id = as.character(unique(c(df_similarity$from, df_similarity$to)))
  ) %>%
    mutate(
      label = id,
      shape = ifelse(id %in% top_words, "triangle", "dot"),
      size = 10,
      font.size = 35
    ) %>%
    arrange(id)

  # Edges
  edges <- df_similarity %>%
    filter(similarity >= 0.5) %>%
    mutate(width = similarity * 10)

  ### COMMUNITY DETECTION
  graph <- igraph::graph_from_data_frame(edges, directed = FALSE)
  cluster <- igraph::cluster_louvain(graph)
  cluster_df <- data.frame(as.list(igraph::membership(cluster)))
  cluster_df <- as.data.frame(t(cluster_df)) %>%
    rownames_to_column(var = "id") %>%
    rename(group = "V1")

  # Create group column
  nodes <- left_join(nodes, cluster_df, by = "id") %>%
    drop_na(group)

  return(list(nodes = nodes, edges = edges, top_words = top_words))
}

apply_horizontal_transparency <- function(
  nodes,
  x_col = "x",
  y_col = "y",
  threshold = 30,
  type = "transparency"
) {
  # Check that the specified x and y coordinate columns exist in the data frame
  if (!all(c(x_col, y_col) %in% names(nodes))) {
    stop("The specified coordinate columns are not found in the data frame.")
  }

  # Extract the x and y coordinates from the node data frame
  x_coords <- nodes[[x_col]]
  y_coords <- nodes[[y_col]]

  # Compute the absolute pairwise horizontal (x-axis) distances between all nodes
  distance_matrix_x <- outer(x_coords, x_coords, FUN = function(a, b) {
    abs(a - b)
  })

  # Compute vertical (y-axis) distances, but consider only values below a small threshold (0.05),
  # otherwise assign Inf to ignore vertical mismatches
  distance_matrix_y <- outer(y_coords, y_coords, FUN = function(a, b) {
    ifelse(abs(a - b) < 0.05, abs(a - b), Inf)
  })

  # Set the diagonal elements to Inf to avoid self-comparisons
  diag(distance_matrix_x) <- Inf
  diag(distance_matrix_y) <- Inf

  # Identify nodes whose combined horizontal + (negligible) vertical distance is less than the threshold
  overlapping_nodes <- apply(
    distance_matrix_x + distance_matrix_y,
    1,
    function(row) any(row < threshold)
  )

  # Assign font color based on overlap:
  # use semi-transparent black (alpha = 0.3) for overlapping nodes,
  # and fully opaque black (alpha = 0.9) otherwise
  switch(
    type,
    "transparency" = {
      nodes$font.color <- ifelse(
        overlapping_nodes,
        "rgba(0, 0, 0, 0.3)", # semi-transparent label
        "rgba(0, 0, 0, 0.9)"
      ) # fully opaque label
    },
    hide = {
      nodes <- nodes %>%
        mutate(
          overlapping_nodes = overlapping_nodes,
          label = ifelse(overlapping_nodes, "", label)
        ) %>%
        select(-overlapping_nodes)
    }
  )

  return(nodes)
}

w2v2Vis <- function(
  nodes,
  edges,
  layout = "layout_nicely",
  size = 20,
  labelsize = 35,
  overlap = "none"
) {
  nodes$font.size <- labelsize * 2.5
  nodes$size <- round(labelsize / 1.2, 0)
  nodes$font.vadjust = -20
  K <- max(nodes$group)
  colors <- paste0((rep(colorlist(), ceiling(K / 35))), "70", sep = "")
  nodes$color = colors[nodes$group]

  VIS <- visNetwork(
    nodes,
    edges,
    type = "full",
    smooth = TRUE,
    physics = FALSE,
    x = 1,
    y = 1
  ) %>%
    visEdges(smooth = TRUE) %>%
    visOptions(
      highlightNearest = list(enabled = TRUE, degree = 1, hover = TRUE),
      nodesIdSelection = list(enabled = FALSE)
    ) %>%
    visIgraphLayout(layout = "layout_nicely", type = "full") %>%
    visNetwork::visInteraction(
      dragNodes = TRUE,
      navigationButtons = F,
      hideEdgesOnDrag = TRUE,
      zoomSpeed = 0.2
    )

  if (overlap != "none") {
    VIS$x$nodes <- apply_horizontal_transparency(
      VIS$x$nodes,
      threshold = labelsize * 0.0035714,
      type = overlap
    )
  }

  VIS$x$nodes <- VIS$x$nodes %>%
    mutate(title = id)

  return(VIS)
}

w2vUMAP <- function(w2v_model, top_words) {
  cbow_embedding <- as.matrix(w2v_model)
  visualization <- umap(cbow_embedding, n_neighbors = 15, n_threads = 2)

  df <- data.frame(
    word = rownames(cbow_embedding),
    x = visualization$layout[, 1],
    y = visualization$layout[, 2],
    stringsAsFactors = FALSE
  ) %>%
    filter(word %in% top_words)
  return(df)
}

reduce_overlap <- function(df, jitter_amount = 0.05, min_dist = 0.05) {
  df_sorted <- df[order(df$x, df$y), ] # ordinamento spaziale semplice
  for (i in 2:nrow(df_sorted)) {
    dx <- df_sorted$x[i] - df_sorted$x[i - 1]
    dy <- df_sorted$y[i] - df_sorted$y[i - 1]
    dist <- sqrt(dx^2 + dy^2)
    if (dist < min_dist) {
      df_sorted$y[i] <- df_sorted$y[i] + runif(1, -jitter_amount, jitter_amount)
      df_sorted$x[i] <- df_sorted$x[i] + runif(1, -jitter_amount, jitter_amount)
    }
  }
  return(df_sorted)
}


adjust_labels_iterative_with_opacity <- function(
  df,
  min_dist = 0.03,
  max_iter = 50,
  shift_step = 0.05,
  alpha_low = 0.4
) {
  df$opacity_val <- rep(0.9, nrow(df)) # inizialmente opacità massima

  for (iter in seq_len(max_iter)) {
    overlap_found <- FALSE
    for (i in 1:(nrow(df) - 1)) {
      for (j in (i + 1):nrow(df)) {
        dx <- df$x[i] - df$x[j]
        dy <- df$y[i] - df$y[j]
        dist <- sqrt(dx^2 + dy^2)
        if (dist < min_dist) {
          overlap_found <- TRUE
          angle <- atan2(dy, dx) + pi / 2
          df$x[i] <- df$x[i] + shift_step * cos(angle)
          df$y[i] <- df$y[i] + shift_step * sin(angle)
          df$x[j] <- df$x[j] - shift_step * cos(angle)
          df$y[j] <- df$y[j] - shift_step * sin(angle)

          # Diminuzione dell'opacità per j (o entrambi, a scelta)
          df$opacity_val[j] <- min(df$opacity_val[j], alpha_low)
        }
      }
    }
    if (!overlap_found) break
  }

  # Assegnazione colore con trasparenza RGBA
  base_rgb <- "79,121,66" # colore #4F7942 in formato RGB
  df$text_color <- paste0("rgba(", base_rgb, ",", df$opacity_val, ")")
  return(df)
}


#### TOPIC MODELING ----

### model tuning

trim_text_columns <- function(df) {
  df %>%
    dplyr::mutate(dplyr::across(
      where(is.character),
      ~ trimws(.x)
    ))
}


clean_text <- function(
  df,
  text_column = "text",
  add_space = TRUE,
  remove_quotes = TRUE,
  punctuation_marks = c(
    ",",
    ";",
    "!",
    "_",
    "»",
    "«",
    "&",
    "(",
    ")",
    "--",
    "..",
    "...",
    "....",
    "--",
    "---",
    ".#",
    "“",
    "”",
    "??",
    "???"
  )
) {
  # Improved emoji regex pattern to capture Unicode emojis
  EMOJI <- "[\U0001F600-\U0001F64F\U0001F300-\U0001F5FF\U0001F680-\U0001F6FF\U0001F700-\U0001F77F\U0001F780-\U0001F7FF\U0001F800-\U0001F8FF\U0001F900-\U0001F9FF\U0001FA00-\U0001FA6F\U0001FA70-\U0001FAFF\U00002702-\U000027B0\U000024C2-\U0001F251]"
  # Sort punctuation marks by length (longest first) to prioritize sequences
  punctuation_marks <- punctuation_marks[order(
    nchar(punctuation_marks),
    decreasing = TRUE
  )]
  # Escape special regex characters
  punctuation_marks <- sapply(punctuation_marks, function(x) {
    gsub("([\\^$.|?*+(){}])", "\\\\\\1", x)
  })
  # Create a regex pattern for punctuation sequences
  punctuation_regex <- paste0(
    "(",
    paste0(punctuation_marks, collapse = "|"),
    ")"
  )

  df %>%
    mutate(
      !!text_column := case_when(
        add_space & remove_quotes ~
          {
            text_cleaned <- stringr::str_replace_all(
              .data[[text_column]],
              "[‘’]", "'"
            ) # Normalize curly single quotes to straight apostrophes
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              '[\"]',
              ""
            ) # Remove double quotes
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              "([a-z])\\.([A-Z])",
              "\\1. \\2"
            ) # Add space between lowercase.Uppercase
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              punctuation_regex,
              " \\1 "
            ) # Add spaces around punctuation
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              EMOJI,
              " \\0 "
            ) # Add spaces around emojis

            # Remove extra spaces but preserve newlines
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              "[ \t]+",
              " "
            ) # Replace multiple spaces/tabs with single space
            stringr::str_replace_all(text_cleaned, "[ \t]*\n[ \t]*", "\n") # Clean spaces around newlines
          },
        add_space ~
          {
            text_cleaned <- stringr::str_replace_all(
              .data[[text_column]],
              "[‘’]", "'"
            ) # Normalize curly single quotes to straight apostrophes
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              "([a-z])\\.([A-Z])",
              "\\1. \\2"
            ) # Add space between lowercase.Uppercase
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              punctuation_regex,
              " \\1 "
            )
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              EMOJI,
              " \\0 "
            )
            # Remove extra spaces but preserve newlines
            text_cleaned <- stringr::str_replace_all(
              text_cleaned,
              "[ \t]+",
              " "
            ) # Replace multiple spaces/tabs with single space
            stringr::str_replace_all(text_cleaned, "[ \t]*\n[ \t]*", "\n") # Clean spaces around newlines
          },
        remove_quotes ~
          {
            stringr::str_replace_all(.data[[text_column]], '[\"|\']', "")
          },
        TRUE ~ .data[[text_column]]
      )
    )
}

### DATA ----
# IMPORT TEXT FUNCTIONS ----

getFileNameExtension <- function(fn) {
  # remove a path
  splitted <- strsplit(x = fn, split = "/")[[1]]
  # or use .Platform$file.sep in stead of '/'
  fn <- splitted[length(splitted)]
  ext <- ""
  splitted <- strsplit(x = fn, split = "\\.")[[1]]
  l <- length(splitted)
  if (l > 1 && sum(splitted[1:(l - 1)] != "")) {
    ext <- splitted[l]
  }
  # the extention must be the suffix of a non-empty name
  ext
}

read_files <- function(
  files,
  ext = c("txt", "csv", "xlsx", "pdf"),
  subfolder = TRUE,
  line_sep = ","
) {
  # files <- list.files(path=path, pattern = paste0(".",ext,"$"), recursive = subfolder)
  if (is.null(files)) {
    return(data.frame(doc_id = NA, text = NA, folder = NA))
  }

  if ("datapath" %in% names(files)) {
    doc_id <- files$name
    file <- files$datapath
    folder <- NA
  }

  if (getFileNameExtension(file[1]) == "zip") {
    owd <- setwd(tempdir())
    on.exit(setwd(owd))
    zip_file <- unzip(file[1])
    zip_file <- zip_file[
      (substr(zip_file, nchar(zip_file) - nchar(ext) + 1, nchar(zip_file)) ==
        ext)
    ]
    file <- zip_file[regexpr("__MACOSX", zip_file) == -1]
    doc_id <- unlist(lapply(strsplit(file, "/"), function(l) {
      l[length(l)]
    }))
    folder <- unlist(lapply(strsplit(file, "/"), function(l) {
      l[length(l) - 1]
    }))
  }

  switch(
    ext,
    txt = {
      ## detect text encoding for each file
      df <- readtext::readtext(file)
      encod <- suppressMessages(readtext::encoding(df, verbose = FALSE)$all)
      ## read txt files using the right encoding
      df <- data.frame(
        doc_id = doc_id,
        text = NA,
        folder = folder,
        file = file,
        encod = encod
      ) %>%
        group_by(doc_id) %>%
        mutate(
          text = readtext::readtext(file, encoding = encod, verbosity = 0)$text
        ) %>%
        select(-c(file, encod))
    },
    csv = {
      listdf <- list()
      for (i in seq_len(length(file))) {
        # listdf[[i]] <- readtext::readtext(file[i], fill=TRUE, text_field="text", quote='"') %>%
        #   mutate(doc_id = doc_id[i])
        listdf[[i]] <- read_delim(file[i], delim = line_sep, quote = '"') %>%
          mutate(filename = doc_id[i])
      }

      df <- do.call(rbind, listdf)
    },
    xlsx = {
      listdf <- list()
      for (i in seq_len(length(file))) {
        # listdf[[i]] <- readtext::readtext(file[i], fill=TRUE, text_field="text", quote='"') %>%
        #   mutate(doc_id = doc_id[i])
        listdf[[i]] <- readxl::read_excel(file[i], col_types = "text") %>%
          mutate(filename = doc_id[i])
      }
      df <- do.call(rbind, listdf)
    },
    pdf = {
      listdf <- list()
      for (i in seq_len(length(file))) {
        # listdf[[i]] <- readtext::readtext(file[i], fill=TRUE, text_field="text", quote='"') %>%
        #   mutate(doc_id = doc_id[i])
        listdf[[i]] <- data.frame(
          text = pdf2txt_auto(file[i]),
          filename = doc_id[i]
        )
      }
      df <- do.call(rbind, listdf)
    }
  )
  if ("doc_id" %in% names(df)) {
    if (sum(duplicated(df$doc_id), na.rm = T) > 0) {
      num <- sprintf(paste0("%0", nchar(nrow(df)), "d"), 1:nrow(df))
      df <- df %>%
        mutate(
          original_doc_id = doc_id,
          doc_id = paste0("doc_", num)
        ) %>%
        select(doc_id, everything())
    }
  } else {
    num <- sprintf(paste0("%0", nchar(nrow(df)), "d"), 1:nrow(df))
    df <- df %>%
      mutate(doc_id = paste0("doc_", num)) %>%
      select(doc_id, everything())
  }

  df <- df %>%
    mutate(doc_selected = TRUE)

  return(df)
}

## PDF to TEXT ----
pdf2txt_multicolumn_safe <- function(
  file,
  column_threshold = NULL,
  preserve_structure = TRUE
) {
  # Check compatibility for poppler_config
  has_poppler_config <- exists(
    "poppler_config",
    where = asNamespace("pdftools"),
    mode = "function"
  )

  if (has_poppler_config) {
    if (!pdftools::poppler_config()$has_pdf_data) {
      message(
        "Pdf import feature requires a recent version of libpoppler. Please install it."
      )
      return(NA)
    }
  }

  # Try using pdf_data to handle multi-column documents
  tryCatch(
    {
      # Extract data with positions
      data_list <- pdftools::pdf_data(file)

      all_text <- c()

      for (page_num in seq_along(data_list)) {
        page_data <- data_list[[page_num]]

        if (nrow(page_data) == 0) {
          next
        }

        # Automatically determine column threshold if not specified
        if (is.null(column_threshold)) {
          # Analyze distribution of x positions to detect multiple columns
          x_positions <- page_data$x

          # If there are enough different positions, try to detect multiple columns
          if (length(unique(x_positions)) > 20) {
            # Use clustering to identify potential columns
            tryCatch(
              {
                clusters <- kmeans(x_positions, centers = 2, nstart = 10)
                cluster_centers <- sort(clusters$centers[, 1])

                # Use midpoint between cluster centers as threshold
                column_threshold <- mean(cluster_centers)
              },
              error = function(e) {
                # If clustering fails, use page midpoint
                column_threshold <- (max(page_data$x) + min(page_data$x)) / 2
              }
            )
          } else {
            # For documents with few different x positions, use page midpoint
            column_threshold <- (max(page_data$x) + min(page_data$x)) / 2
          }
        }

        # Separate columns based on x position
        left_column <- page_data[page_data$x < column_threshold, ]
        right_column <- page_data[page_data$x >= column_threshold, ]

        # Check if there are actually two significant columns
        if (nrow(left_column) < 5 || nrow(right_column) < 5) {
          # Probably single-column document, treat everything together
          page_data <- page_data[order(page_data$y, page_data$x), ]
          page_text <- reconstruct_text_structured(
            page_data,
            preserve_structure
          )
        } else {
          # Sort by y (top to bottom) and then by x within each column
          left_column <- left_column[order(left_column$y, left_column$x), ]
          right_column <- right_column[order(right_column$y, right_column$x), ]

          # Reconstruct text for each column maintaining structure
          left_text <- reconstruct_text_structured(
            left_column,
            preserve_structure
          )
          right_text <- reconstruct_text_structured(
            right_column,
            preserve_structure
          )

          # Combine columns
          if (preserve_structure) {
            page_text <- paste(left_text, right_text, sep = "\n\n")
          } else {
            page_text <- paste(left_text, right_text, sep = " ")
          }
        }

        all_text <- c(all_text, page_text)
      }

      # Final post-processing
      if (preserve_structure) {
        txt <- paste(all_text, collapse = "\n\n")

        # Improve detection of paragraphs and sections
        # Identify numbered titles like "1. Introduction", "2.1. Internal processing"
        txt <- gsub(
          "([0-9]+(?:\\.[0-9]+)*\\.\\s+[A-Z][A-Za-z\\s]{3,50})",
          "\n\n\\1",
          txt
        )

        # Identify isolated uppercase sections
        txt <- gsub("\\s+([A-Z][A-Z\\s]{10,60})\\s+", "\n\n\\1\n\n", txt)

        # Separate paragraphs after complete sentences
        txt <- gsub("([.!?])\\s+([A-Z][a-z])", "\\1\n\n\\2", txt)

        # Clean triple \n
        txt <- gsub("\\n{3,}", "\n\n", txt)
      } else {
        txt <- paste(all_text, collapse = " ")
      }

      # Remove line-ending hyphens but preserve structure
      txt <- gsub("-\\s*\n", "", txt)
      txt <- gsub("-\\s+", "", txt)

      # Normalize spaces
      if (preserve_structure) {
        txt <- gsub("[ \t]+", " ", txt) # Normalize spaces and tabs
        txt <- gsub("\\n ", "\n", txt) # Remove spaces after \n
      } else {
        txt <- gsub("\\s+", " ", txt)
      }

      # Remove leading and trailing spaces
      txt <- trimws(txt)

      return(txt)
    },
    error = function(e) {
      message("pdf_data failed, falling back to pdf_text method: ", e$message)

      # Fallback to original function with structural improvements
      pages <- pdftools::pdf_length(file)
      txt <- pdftools::pdf_text(file)

      if (preserve_structure) {
        # Improve structure preservation in fallback
        # Identify numbered titles
        txt <- gsub(
          "([0-9]+(?:\\.[0-9]+)*\\.\\s+[A-Z][A-Za-z\\s]{3,50})",
          "\n\n\\1",
          txt
        )

        # Preserve existing paragraphs
        txt <- gsub("\\n\\s*\\n", "\n\n", txt)

        # Separate sentences into new paragraphs when appropriate
        txt <- gsub("([.!?])\\s*\n\\s*([A-Z])", "\\1\n\n\\2", txt)

        # Remove \n that are not at sentence endings but preserve structure
        txt <- gsub("(?<![.!?\\n])\\n(?![A-Z0-9\\n])", " ", txt, perl = TRUE)

        # Combine pages
        txt <- paste(txt, collapse = "\n\n")
      } else {
        # Original version
        txt <- gsub("(?<![\\s\\.])\\n(?!\\s)", " ", txt, perl = TRUE)
        txt <- gsub("\n  ", "\n\n", txt)
        txt <- paste(txt, collapse = " ")
      }

      # Remove word separation hyphens
      txt <- gsub("-\\s", "", txt)

      return(txt)
    }
  )
}

# Simplified helper function that doesn't use non-existent columns
reconstruct_text_structured <- function(
  column_data,
  preserve_structure = TRUE
) {
  if (nrow(column_data) == 0) {
    return("")
  }

  # Group words by line based on similar y positions
  tolerance <- 4
  column_data$line <- round(column_data$y / tolerance) * tolerance

  # Check which columns actually exist
  available_cols <- names(column_data)

  # Use only available information to identify structures
  if ("height" %in% available_cols) {
    column_data$font_size <- column_data$height
  } else {
    column_data$font_size <- 12 # default
  }

  # Split data by line
  lines <- split(column_data, column_data$line)

  # Reconstruct each line
  line_results <- lapply(lines, function(line) {
    line <- line[order(line$x), ] # Sort by x position

    # Reconstruct line text
    line_text <- paste(line$text, collapse = " ")
    line_text <- trimws(line_text)

    # Line characteristics to identify titles/special structures
    avg_font_size <- mean(line$font_size, na.rm = TRUE)
    is_short <- nchar(line_text) < 80
    is_caps <- grepl("^[A-Z\\s\\d\\.\\-]+$", line_text)
    starts_with_number <- grepl("^\\d+\\.", line_text)
    starts_with_section <- grepl("^\\d+\\.\\d+", line_text)

    # Identify potential titles or special structures
    is_title <- (is_short &&
      (is_caps || starts_with_number || starts_with_section))

    return(list(
      text = line_text,
      y = min(line$y),
      is_title = is_title,
      font_size = avg_font_size,
      starts_with_number = starts_with_number
    ))
  })

  # Remove empty lines
  line_results <- line_results[sapply(line_results, function(x) {
    nchar(x$text) > 0
  })]

  # Sort lines from top to bottom
  line_results <- line_results[order(sapply(line_results, function(x) x$y))]

  if (!preserve_structure) {
    # Simple mode: join everything with spaces
    result <- paste(sapply(line_results, function(x) x$text), collapse = " ")
  } else {
    # Structured mode: preserve paragraphs and titles
    result_parts <- c()

    for (i in seq_along(line_results)) {
      current_line <- line_results[[i]]
      line_text <- current_line$text

      if (nchar(line_text) == 0) {
        next
      }

      # Add appropriate separators
      if (i == 1) {
        # First line
        result_parts <- c(result_parts, line_text)
      } else {
        prev_line <- line_results[[i - 1]]

        # Determine separator type
        if (current_line$is_title) {
          # Titles: double \n before
          result_parts <- c(result_parts, "\n\n", line_text)
        } else if (prev_line$is_title) {
          # After a title: double \n
          result_parts <- c(result_parts, "\n\n", line_text)
        } else if (
          grepl("[.!?]\\s*$", prev_line$text) &&
            grepl("^[A-Z]", line_text) &&
            !grepl("^[A-Z][a-z]+\\s+[a-z]", line_text)
        ) {
          # End of sentence + start with capital (but not normal continuation)
          result_parts <- c(result_parts, "\n\n", line_text)
        } else {
          # Continuation: single space
          result_parts <- c(result_parts, " ", line_text)
        }
      }
    }

    result <- paste(result_parts, collapse = "")

    # Post-processing to improve structure
    result <- gsub("\\s+", " ", result) # Normalize spaces
    result <- gsub("\\n\\s+", "\n", result) # Remove spaces after \n
    result <- gsub("\\n{3,}", "\n\n", result) # Max 2 consecutive \n
  }

  result <- trimws(result)
  return(result)
}

# Automatic wrapper function
pdf2txt_auto <- function(file, preserve_structure = TRUE) {
  # First try multi-column method
  result <- pdf2txt_multicolumn_safe(
    file,
    preserve_structure = preserve_structure
  )

  # If result seems malformed, try original method
  if (is.na(result) || nchar(result) < 100) {
    message(
      "Multi-column method failed or returned short text, trying original method..."
    )

    # Complete fallback to original method with improvements
    tryCatch(
      {
        pages <- pdftools::pdf_length(file)
        txt <- pdftools::pdf_text(file)

        if (preserve_structure) {
          # Better structure preservation in fallback
          # Identify numbered titles
          txt <- gsub(
            "([0-9]+(?:\\.[0-9]+)*\\.\\s+[A-Za-z][A-Za-z\\s]{3,50})",
            "\n\n\\1\n\n",
            txt,
            perl = TRUE
          )

          # Preserve existing paragraphs
          txt <- gsub("\\n\\s*\\n", "\n\n", txt)

          # Separate sentences into paragraphs when they start with capital
          txt <- gsub(
            "([.!?])\\s*\n\\s*([A-Z][a-z])",
            "\\1\n\n\\2",
            txt,
            perl = TRUE
          )

          # Remove internal \n but preserve structure
          txt <- gsub("(?<![.!?\\n])\\n(?![A-Z0-9\\n])", " ", txt, perl = TRUE)

          txt <- paste(txt, collapse = "\n\n")
        } else {
          txt <- gsub("(?<![\\s\\.])\\n(?!\\s)", " ", txt, perl = TRUE)
          txt <- gsub("\n  ", "\n\n", txt)
          txt <- paste(txt, collapse = " ")
        }

        txt <- gsub("-\\s", "", txt)
        return(txt)
      },
      error = function(e) {
        message("All methods failed: ", e$message)
        return(NA)
      }
    )
  }

  return(result)
}

## Remove HTML TAGS ----

removeHTMLTags <- function(text) {
  text <- text %>%
    gsub("&nbsp;|&amp;|&current;|&trade;", " ", .) %>%
    trimws() %>%
    gsub("<br>", "\\\n", .) %>%
    gsub("</p>", "\\\n", .) %>%
    gsub("<.*?>", "", .)
}

### download sample data
loadSampleCollection <- function(sampleName) {
  home <- homeFolder()

  # setting up the main directory
  path_tall <- file.path(home, "tall")
  path_language_model <- file.path(path_tall, "language_models")
  # check if sub directory exists
  if (file.exists(path_tall)) {
    if (!file.exists(path_language_model)) dir.create(path_language_model)
  } else {
    dir.create(path_tall)
    dir.create(path_language_model)
  }

  switch(
    sampleName,
    bibliometrix = {
      # check if the file model already exists
      file_lang <- dir(path_language_model, pattern = "tall_bibliometrix.tall")[
        1
      ]
      url <- paste0(
        "https://raw.githubusercontent.com/massimoaria/tall.language.models/main/sample.data/tall_bibliometrix.tall"
      )

      destfile <- paste0(path_language_model, "/tall_bibliometrix.tall")
      file <- paste0(path_language_model, "/tall_bibliometrix.tall")
    },
    bbc = {
      file_lang <- dir(path_language_model, pattern = "bbc.zip")[1]
      url <- paste0(
        "https://raw.githubusercontent.com/massimoaria/tall.language.models/main/sample.data/bbc.zip"
      )
      destfile <- paste0(path_language_model, "/bbc.zip")
      file <- paste0(path_language_model, "/bbc.zip")
    },
    usairlines = {
      file_lang <- dir(path_language_model, pattern = "usairlines.zip")[1]
      url <- paste0(
        "https://raw.githubusercontent.com/massimoaria/tall.language.models/main/sample.data/usairlines.zip"
      )
      destfile <- paste0(path_language_model, "/usairlines.zip")
      file <- paste0(path_language_model, "/usairlines.zip")
    }
  )

  if (is.na(file_lang)) {
    switch(
      Sys.info()[["sysname"]],
      Windows = {
        download.file(url = url, destfile = destfile, mode = "wb")
      },
      {
        download.file(url = url, destfile = destfile)
      }
    )
  }

  return(file)
}

## Wikipedia download sample pages ----
wikiSearch <- function(term, n = 10) {
  if (!inherits(n, "numeric")) {
    n <- 10
  }

  # select n of results
  search_url <- paste0(
    "https://en.wikipedia.org/w/api.php?action=query&generator=search&gsrlimit=",
    n,
    "&prop=extracts&exintro&explaintext&exlimit=max&format=json&gsrsearch="
  )

  term <- gsub("–", " ", gsub("\\s", "_", term))

  result <- jsonlite::fromJSON(paste0(
    search_url,
    term
  ))

  if (!"query" %in% names(result)) {
    return(NULL)
  }

  title <- c()
  abstract <- c()
  url <- c()

  for (i in 1:length(result$query$pages)) {
    wikidata <- result$query$pages[i]
    names(wikidata) <- "content"
    page_abstract <- gsub("<.*?>", "", wikidata$content$extract)

    abstract <- append(abstract, page_abstract)
    title <- append(title, wikidata$content$title)
    url <- append(
      url,
      paste0(
        "https://en.wikipedia.org/wiki/",
        gsub("\\s", "_", wikidata$content$title)
      )
    )
  }
  url <-
    paste0(
      '<a href=\"',
      url,
      '\" target=\"_blank\">',
      url,
      "</a>"
    )

  df <- data.frame(title, abstract, url, text = NA, selected = TRUE)
  return(df)
}

wikiExtract <- function(df) {
  items <- gsub("\\s", "_", df$title)
  for (i in 1:length(items)) {
    if (df$selected[i]) {
      title <- items[i]
      # print(title)
      result <- jsonlite::fromJSON(paste0(
        "https://en.wikipedia.org/w/api.php?action=query&titles=",
        title,
        "&prop=extracts&redirects=&format=json"
      ))
      names(result$query$pages) <- "content"
      text <- result$query$pages$content$extract
      ## remove Latex equations and textstyle
      df$text[i] <- gsub("<.*?>", "", text) %>%
        gsub("\\{\\\\displaystyle.*?\\}\\n", "", .) %>%
        gsub("\\\\textstyle", " ", .)
    }
  }
  return(df)
}

### REBUILD ORIGINAL DOCUMENTS -----

rebuild_documents <- function(df) {
  columns <- intersect(
    names(df),
    c(
      "start",
      "end",
      "term_id",
      "token",
      "token_id",
      "lemma",
      "token_original_nomultiwords",
      "upos_original",
      "feats",
      "head_token_id",
      "dep_rel",
      "deps",
      "POSSelected",
      "upos",
      "ngram",
      "sentence_hl",
      "xpos",
      "noHapax",
      "noSingleChar",
      "lemma_original_nomultiwords",
      "token_hl",
      "start_hl",
      "end_hl",
      "misc",
      "upos_specialentities"
    )
  )

  df <- df %>%
    select(!all_of(columns)) %>%
    distinct(., doc_id, paragraph_id, sentence_id, .keep_all = TRUE)

  # Combine sentences into paragraphs
  paragraphs <- df %>%
    group_by(doc_id, paragraph_id) %>%
    arrange(sentence_id) %>%
    summarise(
      paragraph_text = paste(sentence, collapse = " "),
      .groups = "drop"
    )

  # Combine paragraphs into full documents
  documents <- paragraphs %>%
    group_by(doc_id) %>%
    arrange(paragraph_id) %>%
    summarise(
      text = paste(paragraph_text, collapse = "\n\n"),
      .groups = "drop"
    ) %>%
    left_join(
      df %>%
        select(
          !c("paragraph_id", "sentence_id", "sentence", "doc_selected")
        ) %>%
        distinct(., doc_id, .keep_all = TRUE),
      by = c("doc_id")
    ) %>%
    rename("doc_selected" = "docSelected") %>%
    select(!starts_with("upos")) %>%
    mutate(text_original = text)

  return(documents)
}

### SPLIT TEXT INTO SUB-DOCS
splitDoc <- function(df, word) {
  if (nchar(word) <= 3) {
    return(df)
  }

  df <- df %>% filter(doc_selected)

  if ("split_word" %in% names(df)) {
    df <- df %>% select(-split_word)
  }
  if ("doc_id_old" %in% names(df)) {
    df <- df %>% select(-doc_id_old)
  }

  df_splitted <- list()
  n <- nrow(df)

  for (i in seq_len(n)) {
    testo <- df$text[i]
    delimiter <- "\u0001SPLIT_HERE\u0001"
    testo_marked <- gsub(word, paste0(delimiter, word), testo, fixed = TRUE)
    parti <- unlist(strsplit(testo_marked, delimiter, fixed = TRUE))
    parti <- parti[nchar(trimws(parti)) > 0]
    df_splitted[[i]] <- parti
  }

  doc_id_old <- rep(df$doc_id, lengths(df_splitted))
  total_docs <- sum(lengths(df_splitted))
  padding <- nchar(as.character(total_docs))

  df_result <- data.frame(
    doc_id = paste0("doc_", sprintf(paste0("%0", padding, "d"), 1:total_docs)),
    text = unlist(df_splitted),
    doc_id_old = doc_id_old,
    doc_selected = TRUE,
    stringsAsFactors = FALSE
  )

  df_for_join <- df %>% select(-c(text, doc_selected))

  df_result <- df_result %>%
    left_join(df_for_join, by = c("doc_id_old" = "doc_id")) %>%
    mutate(split_word = word)

  return(df_result)
}

unsplitDoc <- function(df) {
  if ("doc_id_old" %in% names(df)) {
    text_rejoined <- df %>%
      group_by(doc_id_old) %>%
      summarise(text = paste(text, collapse = ""), .groups = "drop")

    cols_to_exclude <- c("doc_id", "text", "doc_selected", "split_word")

    other_columns <- df %>%
      group_by(doc_id_old) %>%
      slice(1) %>%
      ungroup() %>%
      select(-all_of(cols_to_exclude))

    df <- text_rejoined %>%
      left_join(other_columns, by = "doc_id_old") %>%
      mutate(doc_id = doc_id_old, doc_selected = TRUE) %>%
      select(-doc_id_old) %>%
      arrange(doc_id)
  }

  return(df)
}

### TEXT SAMPLING ----
samplingText <- function(txt, n) {
  id <- sample(txt$doc_id, n)
  txt$doc_selected <- (txt$doc_id %in% id)
  return(txt)
}


### EXTERNAL INFORMATION ----
loadExtInfo <- function(file, txt) {
  df <- readxl::read_excel(file)

  txt <- txt %>%
    left_join(df, by = "doc_id")

  return(txt)
}

### PRE_PROCESSING ----

restoreText <- function(x) {
  x <- x %>%
    mutate(text = text_original)
}

### 1. TOKENIZATION ----

filters_groupsUI <- function() {
  ### FILTER ----
  filters <- tabItem(
    tabName = "filter_text",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("filter"), strong("Filter docs by available external information"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      fluidRow(
        column(
          9,
          shinycssloaders::withSpinner(
            DT::DTOutput("filterData"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        column(
          3,
          fluidRow(
            box(
              width = 12,
              div(h3(strong(em("Filter by"))), style = "margin-top:-57px"),
              hr(),
              helpText(h5("Select an external information to filter docs:")),
              uiOutput("filterListUI"),
              uiOutput("filterValue"),
              hr(),
              div(
                fluidRow(
                  column(
                    6,
                    div(
                      align = "center",
                      title = t_run,
                      do.call(
                        "actionButton",
                        c(
                          run_bttn,
                          list(
                            inputId = "filterRun"
                          )
                        )
                      )
                    )
                  ),
                  column(
                    6,
                    div(
                      align = "center",
                      title = t_back,
                      do.call(
                        "actionButton",
                        c(
                          back_bttn,
                          list(
                            inputId = "filterBack"
                          )
                        )
                      )
                    )
                  )
                ),
                style = "margin-top:-15px"
              ),
              br(),
              htmlOutput("filterSummary")
            )
          )
        )
      )
    )
  )

  ### GROUPS ----
  groups <- tabItem(
    tabName = "defineGroups",
    fluidPage(
      fluidRow(
        column(
          12,
          h2(icon("object-group"), strong("Define groups by available external information"), style = "color: #4F7942; text-align: center; margin-bottom: 20px;")
        )
      ),
      fluidRow(
        column(
          9,
          shinycssloaders::withSpinner(
            DT::DTOutput("defineGroupsData"),
            color = getOption("spinner.color", default = "#4F7942")
          )
        ),
        column(
          3,
          fluidRow(
            box(
              width = 12,
              div(
                h3(strong(em("Select external information"))),
                style = "margin-top:-57px"
              ),
              hr(),
              helpText(h5(
                "Select an external information to define new document groups:"
              )),
              uiOutput("defineGroupsListUI"),
              uiOutput(outputId = "infoGroups"),
              hr(),
              div(
                fluidRow(
                  div(
                    align = "center",
                    title = t_run,
                    do.call(
                      "actionButton",
                      c(
                        run_bttn,
                        list(
                          inputId = "defineGroupsRun"
                        )
                      )
                    )
                  )
                ),
                style = "margin-top:-15px"
              )
            )
          )
        )
      )
    )
  )

  return(list(filters = filters, groups = groups))
}

filters_groupsServer <- function(input, output, session, values, statsValues) {
  ## FILTER ----
  output$filterListUI <- renderUI({
    label <- c("", sort(noGroupLabels(names(values$dfTag))))
    selectInput(
      inputId = "filterList",
      label = NULL,
      choices = label,
      selected = values$selectedFilter,
      multiple = TRUE,
      width = "100%"
    )
  })

  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$filterList
    },
    handlerExpr = {
      if (length(input$filterList) > 0) {
        filtervalues <- LemmaSelection(values$dfTag) %>%
          select(all_of(input$filterList)) %>%
          distinct()
        values$filtervalues <- sort(filtervalues[[1]])
        values$selectedFilter <- input$filterList
      }
    }
  )

  output$filterValue <- renderUI({
    req(input$filterList)

    lapply(input$filterList, function(var) {
      if (!is.null(values[[paste0("filter_", var)]])) {
        sel_value <- values[[paste0("filter_", var)]]
      } else {
        sel_value <- ""
      }
      if (is.factor(values$dfTag[[var]])) {
        selectInput(
          inputId = paste0("filter_", var),
          label = paste("Filter", var),
          choices = levels(values$dfTag[[var]]),
          selected = sel_value,
          multiple = TRUE
        )
      } else if (is.numeric(values$dfTag[[var]])) {
        if (sel_value[1] == "") {
          sel_value <- range(values$dfTag[[var]], na.rm = TRUE)
        }
        sliderInput(
          inputId = paste0("filter_", var),
          label = paste("Filter", var),
          min = min(values$dfTag[[var]], na.rm = TRUE),
          max = max(values$dfTag[[var]], na.rm = TRUE),
          value = sel_value
        )
      } else if (is.character(values$dfTag[[var]])) {
        selectInput(
          inputId = paste0("filter_", var),
          label = paste("Filter", var),
          choices = sort(unique(values$dfTag[[var]])),
          selected = sel_value,
          multiple = TRUE
        )
      }
    })
  })

  # Eagerly update docSelected when filter is applied
  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = input$filterRun,
    handlerExpr = {
      if (!is.null(input$filterList)) {
        docSelected <- data.frame(matrix(ncol = 0, nrow = nrow(values$dfTag)))
        for (var in input$filterList) {
          input_id <- paste0("filter_", var)
          filter_value <- input[[input_id]]
          values[[input_id]] <- filter_value

          if (is.factor(values$dfTag[[var]])) {
            docSelected[[var]] <- ifelse(
              values$dfTag[[var]] %in% filter_value,
              TRUE,
              FALSE
            )
          } else if (is.character(values$dfTag[[var]])) {
            docSelected[[var]] <- ifelse(
              values$dfTag[[var]] %in% filter_value,
              TRUE,
              FALSE
            )
          } else if (is.numeric(values$dfTag[[var]])) {
            docSelected[[var]] <- ifelse(
              values$dfTag[[var]] >= filter_value[1] &
                values$dfTag[[var]] <= filter_value[2],
              TRUE,
              FALSE
            )
          } else {
            docSelected[[var]] <- TRUE
          }
        }
        df <- values$dfTag
        df$docSelected <- ifelse(
          rowSums(docSelected) >= ncol(docSelected),
          TRUE,
          FALSE
        )
        values$dfTag <- df
      } else {
        df <- values$dfTag
        df$docSelected <- TRUE
        values$dfTag <- df
      }

      # Update filter summary
      output$filterSummary <- renderUI({
        filtered <- LemmaSelection(values$dfTag) %>% filter(docSelected)
        num_docs <- n_distinct(filtered$doc_id)
        num_paragraphs <- sum(
          filtered %>%
            group_by(doc_id) %>%
            summarize(sent = max(paragraph_id)) %>%
            pull()
        )
        num_sentences <- sum(
          filtered %>%
            group_by(doc_id) %>%
            summarize(sent = max(sentence_id)) %>%
            pull()
        )
        num_tokens <- nrow(filtered)

        HTML(paste(
          "<div style='border: 1px solid #ddd; padding: 10px; border-radius: 5px; background-color: #f9f9f9;'>",
          "<h4><strong>Filter Summary</strong></h4>",
          "<br>",
          "<p>Number of Documents: <strong>",
          num_docs,
          "</strong></p>",
          "<p>Number of Paragraphs: <strong>",
          num_paragraphs,
          "</strong></p>",
          "<p>Number of Sentences: <strong>",
          num_sentences,
          "</strong></p>",
          "<p>Number of Tokens: <strong>",
          num_tokens,
          "</strong></p>",
          "</div>"
        ))
      })
    }
  )

  output$filterData <- renderDT({
    req(values$dfTag)
    input$filterRun
    DTformat(
      LemmaSelection(values$dfTag) %>%
        dplyr::filter(docSelected) %>%
        select("doc_id", "sentence", "token", "lemma", "upos"),
      nrow = 3,
      size = "100%",
      title = paste0(
        "Filtered Data by ",
        paste0(input$filterList, collapse = ", ")
      ),
      col_to_remove = values$generalTerm
    )
  })

  ## Data filtered by dynamic text on dashboardHeader

  output$dataFilteredBy <- renderText({
    if (!is.null(input$filterList)) {
      req(input$filterRun)
      HTML(paste(
        "Documents filtered by: <b>",
        paste0(input$filterList, collapse = ", "),
        "</b>"
      ))
    } else {
      HTML("")
    }
  })

  observeEvent(
    ignoreNULL = TRUE,
    eventExpr = {
      input$filterBack
    },
    handlerExpr = {
      df <- values$dfTag
      df$docSelected <- TRUE
      values$dfTag <- df
      lapply(input$filterList, function(var) {
        removeUI(paste0("filter_", var))
        if (is.factor(values$dfTag[[var]])) {
          values[[paste0("filter_", var)]] <- NULL
        } else if (is.numeric(values$dfTag[[var]])) {
          values[[paste0("filter_", var)]] <- range(
            values$dfTag[[var]],
            na.rm = TRUE
          )
        } else if (is.character(values$dfTag[[var]])) {
          values[[paste0("filter_", var)]] <- NULL
        }
      })
      updateSelectInput(
        inputId = "filterList",
        session = session,
        selected = ""
      )
    }
  )

  ## GROUPS ----

  ### Define groups ----

  output$defineGroupsListUI <- renderUI({
    label <- noGroupLabels(names(values$dfTag))
    multiInput(
      inputId = "defineGroupsList",
      label = NULL,
      choices = label,
      selected = values$selectedGroups,
      width = "100%"
    )
  })

  output$infoGroups <- renderUI({
    if (length(input$defineGroupsList) > 1) {
      shinyWidgets::alert(
        icon("info"),
        " You need to select only one field",
        status = "danger"
      )
    }
  })

  groupMetadata <- eventReactive(
    ignoreNULL = TRUE,
    eventExpr = {
      input$defineGroupsRun
    },
    valueExpr = {
      values$selectedGroups <- input$defineGroupsList
      values$dfTag <- groupByMetadata(
        values$dfTag,
        metadata = input$defineGroupsList
      )
      if (length(input$defineGroupsList) == 1) {
        showModal(groupModal(session))
      } else {
        showModal(ungroupModal(session))
      }
    }
  )

  ## Data grouped by dynamic text on dashboardHeader
  output$dataGroupedBy <- renderText({
    if (length(input$defineGroupsList) == 1) {
      req(input$defineGroupsRun)
      HTML(paste("Documents grouped by: <b>", input$defineGroupsList, "</b>"))
    } else {
      HTML("")
    }
  })

  output$defineGroupsData <- renderDT({
    groupMetadata()
    DTformat(
      values$dfTag %>% filter(docSelected),
      nrow = 3,
      size = "100%",
      title = "Data Grouped By External Information",
      col_to_remove = values$generalTerm
    )
  })

  groupModal <- function(session) {
    ns <- session$ns
    values$newGr <- values$dfTag %>%
      filter(docSelected) %>%
      count(doc_id, ungroupDoc_id) %>%
      group_by(doc_id) %>%
      count()
    names(values$newGr) <- c(input$defineGroupsList, "N. of Docs")
    txt <- paste0(
      "<hr><br><br>The original <b>",
      sum(values$newGr[, 2]),
      "</b> documents have been partitioned into <b>",
      nrow(values$newGr),
      "</b> groups <br><br>"
    )
    modalDialog(
      h3(strong(paste0("Documents grouped by ", input$defineGroupsList))),
      h4(HTML(txt)),
      br(),
      DTOutput(ns("groupData")),
      size = "m",
      easyClose = TRUE,
      footer = tagList(
        modalButton("Close")
      ),
    )
  }

  ungroupModal <- function(session) {
    ns <- session$ns
    txt <- paste0(
      "<hr><br><br>The original partitioning<br>into documents has been correctly restored.<br><br><hr>"
    )
    modalDialog(
      h3(HTML(txt)),
      size = "m",
      easyClose = TRUE,
      footer = tagList(
        modalButton("Close")
      ),
    )
  }

  output$groupData <- renderDT(server = FALSE, {
    DTformat(
      values$newGr,
      nrow = 10,
      size = "100%",
      title = "Groups By External Information",
      left = 1,
      col_to_remove = values$generalTerm
    )
  })
}

wordcloud <- function(
  data,
  col_names = NULL,
  max_size = 20,
  min_size = 2,
  eccentricity = 0.65,
  shape = "circle",
  rm_outside = FALSE,
  rot_per = 0.1,
  colors = "black",
  color_by_size = FALSE,
  color_gradient_low = "darkblue",
  color_gradient_high = "lightblue",
  seed = NA,
  area_corr = TRUE,
  rstep = 0.01,
  tstep = 0.02,
  grid_size = 4,
  max_steps = 10,
  xlim = c(NA, NA),
  ylim = c(NA, NA),
  mask = NA,
  show_boxes = FALSE,
  background_color = "white",
  facet_by = NULL,
  facet_ncol = NULL,
  facet_nrow = NULL,
  facet_scales = "fixed",
  facet_labeller = "label_value",
  facet_text_size = 16,
  facet_text_face = "bold",
  facet_text_color = "black",
  ...
) {
  # Validate input
  if (!is.data.frame(data)) {
    stop("'data' must be a data frame")
  }

  if (ncol(data) < 2) {
    stop("'data' must have at least 2 columns")
  }

  # Handle column names
  if (is.null(col_names)) {
    words_col <- names(data)[1]
    size_col <- names(data)[2]
  } else {
    if (length(col_names) != 2) {
      stop("'col_names' must be a character vector of length 2")
    }
    if (!all(col_names %in% names(data))) {
      stop("Column names specified in 'col_names' not found in data")
    }
    words_col <- col_names[1]
    size_col <- col_names[2]
  }

  # Validate facet_by if provided
  if (!is.null(facet_by)) {
    if (!facet_by %in% names(data)) {
      stop("Column '", facet_by, "' specified in 'facet_by' not found in data")
    }
  }

  # Prepare data
  plot_data <- data.frame(
    label = as.character(data[[words_col]]),
    size = as.numeric(data[[size_col]]),
    stringsAsFactors = FALSE
  )

  # Add facet variable if specified
  if (!is.null(facet_by)) {
    plot_data$facet_var <- data[[facet_by]]
  }

  # Remove NA values
  plot_data <- plot_data[!is.na(plot_data$label) & !is.na(plot_data$size), ]

  if (nrow(plot_data) == 0) {
    stop("No valid data to plot after removing NAs")
  }

  # Add rotation angle if rot_per > 0
  if (rot_per > 0) {
    n <- nrow(plot_data)
    plot_data$angle <- 90 *
      sample(c(0, 1), n, replace = TRUE, prob = c(1 - rot_per, rot_per))
  } else {
    plot_data$angle <- 0
  }

  # Handle colors
  if (color_by_size) {
    # Use size for color gradient
    plot_data$color_var <- plot_data$size
  } else {
    # Check if we have faceting and colors match number of facets
    if (!is.null(facet_by) && length(colors) > 1) {
      # Get unique facets
      unique_facets <- unique(plot_data$facet_var)
      n_facets <- length(unique_facets)

      # If number of colors equals number of facets, assign one color per facet
      if (length(colors) == n_facets) {
        # Create a named vector mapping facets to colors
        facet_colors <- setNames(colors, unique_facets)
        # Assign colors based on facet
        plot_data$color_var <- facet_colors[as.character(plot_data$facet_var)]
      } else {
        # Otherwise, assign colors to individual words
        plot_data$color_var <- rep_len(colors, nrow(plot_data))
      }
    } else {
      # No faceting or single color
      if (length(colors) == 1) {
        plot_data$color_var <- colors
      } else {
        # If colors vector is shorter than data, recycle it
        plot_data$color_var <- rep_len(colors, nrow(plot_data))
      }
    }
  }

  # Set seed if provided
  if (!is.na(seed)) {
    set.seed(seed)
  }

  # Build the base plot with color aesthetic
  p <- ggplot2::ggplot(
    plot_data,
    ggplot2::aes(label = label, size = size, angle = angle, color = color_var)
  )

  # Choose geom based on area_corr
  if (area_corr) {
    p <- p +
      ggwordcloud::geom_text_wordcloud_area(
        eccentricity = eccentricity,
        shape = shape,
        rm_outside = rm_outside,
        rstep = rstep,
        tstep = tstep,
        grid_size = grid_size,
        max_steps = max_steps,
        xlim = xlim,
        ylim = ylim,
        mask = mask,
        area_corr = TRUE,
        show_boxes = show_boxes,
        ...
      )
  } else {
    p <- p +
      ggwordcloud::geom_text_wordcloud(
        eccentricity = eccentricity,
        shape = shape,
        rm_outside = rm_outside,
        rstep = rstep,
        tstep = tstep,
        grid_size = grid_size,
        max_steps = max_steps,
        xlim = xlim,
        ylim = ylim,
        mask = mask,
        area_corr = FALSE,
        show_boxes = show_boxes,
        ...
      )
  }

  # Add size scale
  p <- p + ggplot2::scale_size_area(max_size = max_size)

  # Add color scale
  if (color_by_size) {
    p <- p +
      ggplot2::scale_color_gradient(
        low = color_gradient_low,
        high = color_gradient_high,
        guide = "none"
      )
  } else {
    if (length(colors) == 1) {
      p <- p + ggplot2::scale_color_identity()
    } else {
      p <- p + ggplot2::scale_color_identity()
    }
  }

  # Add faceting if specified
  if (!is.null(facet_by)) {
    p <- p +
      ggplot2::facet_wrap(
        ~facet_var,
        ncol = facet_ncol,
        nrow = facet_nrow,
        scales = facet_scales,
        labeller = facet_labeller
      )
  }

  # Add theme
  p <- p +
    ggplot2::theme_minimal() +
    ggplot2::theme(
      panel.background = ggplot2::element_rect(
        fill = background_color,
        color = NA
      ),
      plot.background = ggplot2::element_rect(
        fill = background_color,
        color = NA
      ),
      panel.grid = ggplot2::element_blank(),
      axis.text = ggplot2::element_blank(),
      axis.title = ggplot2::element_blank(),
      axis.ticks = ggplot2::element_blank(),
      legend.position = "none",
      strip.text = ggplot2::element_text(
        size = facet_text_size,
        face = facet_text_face,
        color = facet_text_color
      )
    )

  return(p)
}

### WORD IN CONTEXT ----
get_context_window <- function(
  df,
  target_word,
  n_left = 5,
  n_right = 5,
  term = "token"
) {
  # Ensure term is correctly set
  if (!term %in% c("token", "lemma")) {
    stop("The 'term' argument must be either 'token' or 'lemma'.")
  }

  # Exclude irrelevant POS tags
  no_upos <- c(
    "NGRAM_MERGED",
    "X",
    "PUNCT",
    "SYM",
    "URL",
    "IP_ADDRESS",
    "EMAIL",
    "DET",
    "CCONJ"
  )

  df <- df %>%
    filter(!upos %in% no_upos) %>%
    group_by(doc_id) %>%
    mutate(term_id = row_number()) %>%
    ungroup() %>%
    select(doc_id, term_id, token, lemma, upos) %>%
    mutate(
      token = tolower(token),
      lemma = tolower(lemma) # Ensure consistency in case sensitivity
    )

  # Select target column based on 'term' argument
  target_column <- if (term == "token") "token" else "lemma"

  target_rows <- df %>% filter(!!sym(target_column) == target_word)

  # Initialize list to store context windows
  context_list <- vector("list", length = nrow(target_rows))

  for (i in seq_len(nrow(target_rows))) {
    row <- target_rows[i, ] # Specific occurrence of the target word
    doc_subset <- df %>% filter(doc_id == row$doc_id)

    middle <- row$term_id
    start <- max(1, row$term_id - n_left) # Ensure it doesn't go below 1
    end <- min(max(doc_subset$term_id), row$term_id + n_right) # Ensure it doesn't exceed document length

    # Extract words in left and right context based on 'term'
    context_left <- doc_subset %>%
      filter(term_id >= start & term_id < middle) %>%
      pull(!!sym(term))

    context_right <- doc_subset %>%
      filter(term_id > middle & term_id <= end) %>%
      pull(!!sym(term))

    # Extract POS tags for the context
    context_upos <- doc_subset %>%
      filter(term_id >= start & term_id <= end) %>%
      pull(upos)

    # Store results in a tibble while keeping lists separate
    context_list[[i]] <- tibble(
      doc_id = row$doc_id,
      context_before = list(context_left),
      target_word = row[[target_column]], # Use the correct target reference
      context_after = list(context_right),
      upos = list(context_upos)
    )
  }

  # Combine all tibbles into a single dataframe
  context_df <- bind_rows(context_list)

  return(context_df)
}

## Context network

tallThematicmap <- function(
  dfTag,
  term = "lemma",
  group = "doc_id",
  n = 100,
  labelsize = 10,
  n.labels = 1,
  opacity = 0.8,
  seed = 1234
) {
  net <- network(
    LemmaSelection(dfTag) %>% dplyr::filter(docSelected),
    term = term,
    group = group,
    n = n,
    minEdges = "Auto",
    labelsize = labelsize,
    opacity = opacity,
    interLinks = FALSE,
    normalization = "association",
    remove.isolated = FALSE,
    community.repulsion = 0.5,
    seed = seed,
    cluster = "louvain"
  )
  nodes <- net$nodes
  edges <- net$edges %>%
    mutate(sE = sC^2)

  # centrality and density calculation
  thematicIndices <- edges %>%
    mutate(ext = ifelse(group_from != group_to, 1, 0)) %>%
    # filter(group_from == group_to) %>%
    group_by(group_from) %>%
    mutate(k = n()) %>%
    reframe(
      n = max(k),
      centrality = sum(sE * ext) * 10,
      density = sum(sE * (1 - ext) / k) * 100
    ) %>%
    rename(group = group_from) %>%
    left_join(
      nodes %>%
        select(group, label, value) %>%
        group_by(group) %>%
        slice_max(order_by = value, n = 1),
      by = "group"
    ) %>%
    mutate(
      rcentrality = rank(centrality),
      rdensity = rank(density)
    )

  df <- nodes %>%
    group_by(group) %>% # dplyr::filter(sC>1) %>%
    arrange(desc(value), .by_group = TRUE) %>%
    mutate(freq = sum(value)) %>%
    slice_max(n = 10, value, with_ties = FALSE) %>%
    reframe(
      wordlist = paste(label, value, collapse = "\n"),
      name_full = paste(label[1:min(n.labels, n())], collapse = "\n"),
      name_full_gemini = paste(label[1:min(10, n())], collapse = "\n"),
      color = color[1],
      freq = max(freq)
    ) %>%
    right_join(., thematicIndices, by = "group") %>%
    rename(
      name = label,
      groups = group,
      words = wordlist
    ) %>%
    select(-value)

  df_lab <- nodes %>%
    rename(
      Words = label,
      Occurrences = value,
      Cluster = group,
      Color = color
    ) %>%
    group_by(Cluster) %>%
    arrange(desc(Occurrences), .by_group = TRUE) %>%
    mutate(
      Cluster_Label = Words[1],
      Cluster_Frequency = sum(Occurrences)
    ) %>%
    ungroup() %>%
    select(
      "Occurrences",
      "Words",
      "Cluster",
      "Color",
      "Cluster_Label",
      "Cluster_Frequency"
    )

  return(list(net = net, df = df, df_lab = df_lab))
}

plotTM <- function(df, size = 0.5, gemini = FALSE) {
  if (gemini) {
    df$name_full <- df$name_full_gemini
  }
  meandens <- mean(df$rdensity)
  meancentr <- mean(df$rcentrality)
  xlimits <- c(0, max(df$rcentrality) + 1)
  ylimits <- c(0, max(df$rdensity) + 1)
  # size <- 0.5
  df_labels <- df[df$freq > 1, ]
  df_labels <- df_labels %>%
    mutate(size = log(as.numeric(freq)) * (5 + size))

  if (nrow(df_labels) > 1) {
    df_labels <- adjust_positions_oblique(
      df_labels,
      xvar = "rcentrality",
      yvar = "rdensity",
      min_dist = 1
    )
  }

  annotations <- data.frame(
    xpos = sort(c(xlimits, xlimits)),
    ypos = c(ylimits, ylimits),
    words = c(
      "Peripheral Topics",
      "Niche Topics",
      "Basic Topics ",
      "Hot Topics "
    ),
    hjustvar = c(0, 0, 1, 1),
    vjustvar = c(0, 1.0, 0, 1)
  )

  # 1. Crea i punti con hover e dimensione proporzionale
  fig <- plot_ly(
    data = df_labels,
    x = ~rcentrality,
    y = ~rdensity,
    type = "scatter",
    mode = "markers",
    marker = list(
      size = ~size,
      color = ~color,
      opacity = 0.5
    ),
    hovertext = ~words,
    hovertemplate = "%{hovertext}<extra></extra>"
    # hoverinfo = 'hovertext'
  )

  # 2. Aggiungi le etichette (solo se freq > 1)

  fig <- fig %>%
    add_trace(
      data = df_labels,
      x = ~rcentrality,
      y = ~rdensity,
      type = "scatter",
      mode = "text",
      text = ~ tolower(name_full),
      textfont = list(
        color = "rgba(0,0,0,0.7)",
        size = 12 * (1 + size)
      ),
      # hoverinfo = 'none',
      showlegend = FALSE
    )

  # 3. Aggiungi linee medie
  fig <- fig %>%
    add_trace(
      data = NULL,
      inherit = FALSE,
      type = "scatter",
      mode = "lines",
      x = c(xlimits[1], xlimits[2]),
      y = c(meandens, meandens),
      line = list(dash = "dash", color = "rgba(0,0,0,0.7)"),
      marker = list(opacity = 0),
      showlegend = FALSE,
      hoverinfo = "none"
    ) %>%
    add_trace(
      data = NULL,
      inherit = FALSE,
      type = "scatter",
      mode = "lines",
      x = c(meancentr, meancentr),
      y = c(ylimits[1], ylimits[2]),
      line = list(dash = "dash", color = "rgba(0,0,0,0.7)"),
      marker = list(opacity = 0),
      showlegend = FALSE,
      hoverinfo = "none"
    )

  # 4. Eventuali annotazioni (puoi rimuovere questo blocco se non ne hai)
  if (exists("annotations")) {
    k <- 0
    for (i in 1:nrow(annotations)) {
      if (i > 2) {
        k <- max(xlimits) * 0.05
      }
      fig <- fig %>%
        add_annotations(
          x = annotations$xpos[i] - k,
          y = annotations$ypos[i],
          text = annotations$words[i],
          showarrow = FALSE,
          font = list(
            color = "rgba(50,50,50,0.5)",
            size = 8 * (1 + size * 2)
          ),
          xanchor = "center",
          yanchor = "middle"
        )
    }
  }

  # 5. Layout
  fig <- fig %>%
    layout(
      xaxis = list(
        title = "Relevance degree<br>(Centrality)",
        range = xlimits,
        showticklabels = FALSE,
        showline = FALSE,
        zeroline = FALSE,
        ticks = "",
        showgrid = TRUE
      ),
      yaxis = list(
        title = "Development degree<br>(Density)",
        range = ylimits,
        showticklabels = FALSE,
        showline = FALSE,
        zeroline = FALSE,
        ticks = "",
        showgrid = TRUE
      ),
      plot_bgcolor = "#FFFFFF",
      showlegend = FALSE
    )

  return(fig)
}

adjust_positions_oblique <- function(
  df,
  xvar = "rcentrality",
  yvar = "rdensity",
  min_dist = 0.5,
  max_iter = 100,
  step_factor = 0.5,
  jitter_strength = 0.1
) {
  df_adj <- df

  for (iter in 1:max_iter) {
    moved <- FALSE
    for (i in 1:(nrow(df_adj) - 1)) {
      for (j in (i + 1):nrow(df_adj)) {
        xi <- df_adj[[xvar]][i]
        yi <- df_adj[[yvar]][i]
        xj <- df_adj[[xvar]][j]
        yj <- df_adj[[yvar]][j]

        dx <- xi - xj
        dy <- yi - yj
        dist <- sqrt(dx^2 + dy^2)

        # Se perfettamente sovrapposti, applica jitter obliquo casuale
        if (dist == 0) {
          jitter_angle <- runif(1, 0, 2 * pi)
          offset <- jitter_strength

          df_adj[[xvar]][i] <- xi + cos(jitter_angle) * offset
          df_adj[[yvar]][i] <- yi + sin(jitter_angle) * offset
          df_adj[[xvar]][j] <- xj - cos(jitter_angle) * offset
          df_adj[[yvar]][j] <- yj - sin(jitter_angle) * offset

          moved <- TRUE
        } else if (dist < min_dist) {
          angle <- atan2(dy, dx)
          offset <- (min_dist - dist) * step_factor

          df_adj[[xvar]][i] <- xi + cos(angle) * offset
          df_adj[[yvar]][i] <- yi + sin(angle) * offset
          df_adj[[xvar]][j] <- xj - cos(angle) * offset
          df_adj[[yvar]][j] <- yj - sin(angle) * offset

          moved <- TRUE
        }
      }
    }
    if (!moved) break
  }

  return(df_adj)
}


## WORD EMBEDDING TRAINING ----

menuList <- function(menu) {
  CORPUS <- tags$div(
    id = "corpus-header",
    style = "display: flex;
        align-items: center;
        font-size: 14px;
        font-weight: 600;
        color: #FFFFFF;
        background: rgba(255,255,255,0.1);
        padding: 10px 10px;
        margin: 15px 8px 8px 8px;
        border-radius: 6px;
        border-left: 3px solid #26A69A;
        letter-spacing: 0.8px;",
    tags$span(
      style = "background: #26A69A;
          padding: 4px 8px;
          border-radius: 4px;
          margin-right: 10px;
          font-size: 12px;",
      icon("database")
    ),
    "CORPUS"
  )

  PREPROCESSING <- tags$div(
    id = "preprocessing-header",
    style = "display: flex;
        align-items: center;
        justify-content: space-between;
        font-size: 14px;
        font-weight: 600;
        color: #FFFFFF;
        background: rgba(255,255,255,0.1);
        padding: 10px 10px;
        margin: 15px 8px 8px 8px;
        border-radius: 6px;
        border-left: 3px solid #9C27B0;
        letter-spacing: 0.8px;",
    tags$div(
      style = "display: flex; align-items: center;",
      tags$span(
        style = "background: #9C27B0;
            padding: 4px 8px;
            border-radius: 4px;
            margin-right: 10px;
            font-size: 12px;",
        icon("wand-magic-sparkles")
      ),
      "PREPROCESSING"
    )
  )

  FEATURES <- tags$div(
    id = "features-header",
    style = "display: flex;
        align-items: center;
        font-size: 14px;
        font-weight: 600;
        color: #FFFFFF;
        background: rgba(255,255,255,0.1);
        padding: 10px 10px;
        margin: 15px 8px 8px 8px;
        border-radius: 6px;
        border-left: 3px solid #5C6BC0;
        letter-spacing: 0.8px;",
    tags$span(
      style = "background: #5C6BC0;
          padding: 4px 8px;
          border-radius: 4px;
          margin-right: 10px;
          font-size: 12px;",
      icon("puzzle-piece")
    ),
    "FEATURES"
  )

  ANALYSIS <- tags$div(
    id = "analysis-header",
    style = "display: flex;
        align-items: center;
        font-size: 14px;
        font-weight: 600;
        color: #FFFFFF;
        background: rgba(255,255,255,0.1);
        padding: 10px 10px;
        margin: 15px 8px 8px 8px;
        border-radius: 6px;
        border-left: 3px solid #FF5722;
        letter-spacing: 0.8px;",
    tags$span(
      style = "background: #FF5722;
          padding: 4px 8px;
          border-radius: 4px;
          margin-right: 10px;
          font-size: 12px;",
      icon("microscope")
    ),
    "ANALYSIS"
  )

  import_menu <- menuItem(
    "Import",
    tabName = "import_tx",
    icon = icon("folder-open")
  )

  edit_menu <- menuItem(
    "Edit",
    tabName = "edit_tx",
    icon = icon("pen-to-square"),
    menuSubItem(
      "Split",
      tabName = "split_tx",
      icon = icon("chevron-right")
    ),
    menuSubItem(
      "Random Selection",
      tabName = "randomText",
      icon = icon("chevron-right")
    ),
    menuSubItem(
      "External Information",
      tabName = "extInfo",
      icon = icon("chevron-right")
    )
  )

  preprocessing_menu0 <- menuItem(
    "Pre-processing",
    tabName = "prePro",
    icon = icon("indent"),
    startExpanded = TRUE,
    menuSubItem(
      "Tokenization & PoS Tagging",
      tabName = "tokPos",
      icon = icon("chevron-right"),
      selected = TRUE
    )
  )

  preprocessing_menu1 <- menuItem(
    "Pre-processing",
    tabName = "prePro",
    icon = icon("indent"),
    startExpanded = TRUE,
    menuSubItem(
      "Tokenization & PoS Tagging",
      tabName = "tokPos",
      icon = icon("chevron-right")
    ),
    menuSubItem(
      "Tagging Special Entities",
      tabName = "posSpecial",
      icon = icon("chevron-right")
    ),
    menuItem(
      "Multi-Word",
      tabName = "multiword",
      icon = icon("chevron-right"),
      startExpanded = TRUE,
      menuSubItem(
        "Automatic",
        tabName = "multiwordCreat",
        icon = icon("chevron-right")
      ),
      menuSubItem(
        "By a List",
        tabName = "multiwordByList",
        icon = icon("chevron-right")
      )
    ),
    menuSubItem(
      "Custom PoS List",
      tabName = "custTermList",
      icon = icon("chevron-right"),
      selected = TRUE
    ),
    menuSubItem(
      "Synonyms Merging",
      tabName = "synonymsMgmt",
      icon = icon("chevron-right")
    ),
    menuSubItem(
      "PoS Tag Selection",
      tabName = "posTagSelect",
      icon = icon("chevron-right")
    )
  )

  overview_menu <- menuItem(
    "Overview",
    tabName = "overview",
    icon = icon("magnifying-glass")
  )

  keyness_menu <- menuItem(
    "Keyness",
    tabName = "keyness",
    icon = icon("key")
  )

  kwic_menu <- menuItem(
    "KWIC",
    tabName = "kwic",
    icon = icon("align-center")
  )

  word_menu <- menuItem(
    "Words",
    tabName = "words",
    icon = icon("font"),
    # menuSubItem(
    #   "KWIC",
    #   tabName = "wordCont",
    #   icon = icon("chevron-right")
    # ),
    # menuSubItem("Clustering", tabName = "w_clustering", icon = icon("chevron-right")),
    menuSubItem(
      "Reinert Clustering",
      tabName = "w_reinclustering",
      icon = icon("chevron-right")
    ),
    menuSubItem(
      "Correspondence Analysis",
      tabName = "ca",
      icon = icon("chevron-right")
    ),
    menuItem(
      "Network",
      tabName = "w_network",
      icon = icon("chevron-right"),
      menuSubItem(
        "Word Network",
        tabName = "w_networkCooc",
        icon = icon("chevron-right")
      ),
      menuSubItem(
        "Thematic Map",
        tabName = "w_networkTM",
        icon = icon("chevron-right")
      )
    ),
    menuSubItem(
      "SVO Triplets",
      tabName = "d_svo",
      icon = icon("chevron-right")
    ),
    menuItem(
      "Word Embeddings",
      tabName = "w_embeddings",
      icon = icon("chevron-right"),
      menuSubItem(
        "Training",
        tabName = "w_word2vec",
        icon = icon("chevron-right")
      ),
      # if (embedding){
      menuSubItem(
        "Similarity",
        tabName = "w_w2v_similarity",
        icon = icon("chevron-right")
      )
      # }
    )
  )
  document_menu <- menuItem(
    "Documents",
    tabName = "documents",
    icon = icon("clone"),
    menuSubItem(
      "Syntactic Complexity",
      tabName = "d_syntactic",
      icon = icon("chevron-right")
    ),
    menuItem(
      "Topic Modeling",
      tabName = "d_topicMod",
      icon = icon("chevron-right"),
      menuSubItem(
        "K choice",
        tabName = "d_tm_select",
        icon = icon("chevron-right")
      ),
      menuSubItem(
        "Model Estimation",
        tabName = "d_tm_estim",
        icon = icon("chevron-right")
      )
    ),
    menuSubItem(
      "Supervised Classification",
      tabName = "doc_classification",
      icon = icon("chevron-right")
    ),
    menuItem(
      "Sentiment Analysis",
      tabName = "d_sentiment",
      icon = icon("chevron-right"),
      menuSubItem(
        "Polarity Detection",
        tabName = "d_polDet",
        icon = icon("chevron-right")
      ),
      menuSubItem(
        "Emotion Detection",
        tabName = "d_emo",
        icon = icon("chevron-right")
      )
    ),
    menuItem(
      "Summarization",
      tabName = "summarization",
      icon = icon("chevron-right"),
      menuSubItem(
        "Abstractive",
        tabName = "d_astractive",
        icon = icon("chevron-right")
      ),
      menuSubItem(
        "Extractive",
        tabName = "d_summarization",
        icon = icon("chevron-right")
      )
    )
  )

  features_menu <- list(
    menuItem(
      "Filter",
      tabName = "filter_text",
      icon = icon("filter")
    ),
    menuItem(
      "Groups",
      tabName = "defineGroups",
      icon = icon("table-cells")
    ),
    menuItem(
      "Feature Roles",
      tabName = "feature_roles",
      icon = icon("tags")
    )
  )

  setting_menu <- tags$div(
    style = "display: none;",
    menuItem("Settings", tabName = "settings", icon = icon("tasks"))
  )

  switch(
    as.character(menu),
    "-2" = {
      list(
        CORPUS,
        PREPROCESSING,
        FEATURES,
        ANALYSIS,
        setting_menu
      )
    },
    "0" = {
      list(
        CORPUS,
        import_menu,
        edit_menu,
        PREPROCESSING,
        preprocessing_menu0,
        FEATURES,
        ANALYSIS,
        setting_menu
      )
    },
    "1" = {
      list(
        CORPUS,
        import_menu,
        edit_menu,
        PREPROCESSING,
        preprocessing_menu1,
        FEATURES,
        ANALYSIS,
        setting_menu
      )
    },
    "2" = {
      list(
        CORPUS,
        import_menu,
        edit_menu,
        PREPROCESSING,
        preprocessing_menu1,
        FEATURES,
        ANALYSIS,
        overview_menu,
        keyness_menu,
        kwic_menu,
        word_menu,
        document_menu,
        tags$div(style = "margin-top: 20px;"),
        menuItem("Report", tabName = "report", icon = icon("list-alt")),
        setting_menu
      )
    },
    "3" = {
      list(
        CORPUS,
        import_menu,
        edit_menu,
        PREPROCESSING,
        preprocessing_menu1,
        FEATURES,
        features_menu,
        ANALYSIS,
        overview_menu,
        keyness_menu,
        kwic_menu,
        word_menu,
        document_menu,
        tags$div(style = "margin-top: 20px;"),
        menuItem("Report", tabName = "report", icon = icon("list-alt")),
        setting_menu
      )
    },
    {
      list(
        CORPUS,
        import_menu,
        PREPROCESSING,
        FEATURES,
        ANALYSIS,
        setting_menu
      )
    }
  )
}

# DATA TABLE FORMAT ----
DTformat <- function(
  df,
  nrow = 10,
  filename = "Table",
  pagelength = TRUE,
  left = NULL,
  right = NULL,
  numeric = NULL,
  dom = TRUE,
  size = "85%",
  filter = "top",
  columnShort = NULL,
  columnSmall = NULL,
  round = 2,
  title = "",
  button = FALSE,
  delete = FALSE,
  escape = FALSE,
  selection = FALSE,
  specialtags = FALSE,
  col_to_remove = NULL
) {
  tall_primary <- "#4a7c59"
  tall_secondary <- "#5a9269"
  tall_accent <- "#F7F9FA"
  tall_text <- "#2D3748"
  tall_border <- "#E2E8F0"
  tall_white <- "#FFFFFF"

  if (!is.null(col_to_remove)) {
    col_to_remove_lower <- tolower(col_to_remove)

    if (col_to_remove_lower == "lemma") {
      if ("token" %in% names(df)) {
        df <- df %>% select(-token)
      }
      if ("Token" %in% names(df)) {
        df <- df %>% select(-Token)
      }
    } else if (col_to_remove_lower == "token") {
      if ("lemma" %in% names(df)) {
        df <- df %>% select(-lemma)
      }
      if ("Lemma" %in% names(df)) {
        df <- df %>% select(-Lemma)
      }
    }
  }

  if ("text" %in% names(df)) {
    df <- df %>%
      mutate(text = gsub("<|>", "", text))
  }

  if (length(columnShort) > 0) {
    columnDefs <- list(
      list(
        className = "dt-center",
        targets = 0:(length(names(df)) - 1)
      ),
      list(
        targets = columnShort - 1,
        render = JS(
          "function(data, type, row, meta) {",
          "return type === 'display' && data.length > 500 ?",
          "'<span title=\"' + data + '\">' + data.substr(0, 500) + '...</span>' : data;",
          "}"
        )
      )
    )
  } else {
    columnDefs <- list(list(
      className = "dt-center",
      targets = 0:(length(names(df)) - 1)
    ))
  }

  if (isTRUE(pagelength)) {
    buttons <- list(
      list(extend = "pageLength"),
      list(
        extend = "excel",
        filename = paste0(filename, "_tall_", sys.time()),
        text = '<i class="fa fa-file-excel"></i> Excel',
        className = 'btn btn-success',
        header = TRUE,
        exportOptions = list(
          modifier = list(page = "all")
        )
      )
    )
  } else {
    buttons <- list(
      list(
        extend = "excel",
        filename = paste0(filename, "_tall_", sys.time()),
        text = '<i class="fa fa-file-excel"></i> Excel',
        className = 'btn btn-success',
        header = TRUE,
        exportOptions = list(
          modifier = list(page = "all")
        )
      )
    )
  }

  if (isTRUE(dom)) {
    dom <- "Brtip"
  } else {
    dom <- "Bt"
  }

  if (nchar(title) > 0) {
    caption <- htmltools::tags$caption(
      style = "caption-side: top; text-align: center; color:black;  font-size:140% ;",
      title
    )
  } else {
    caption <- htmltools::tags$caption(
      style = "caption-side: top; text-align: center; color:black;  font-size:140% ;",
      ""
    )
  }

  if (isTRUE(button)) {
    df <- df %>%
      mutate(
        Document = paste0(
          '<button id="custom_btn" onclick="Shiny.onInputChange(\'button_id\', \'',
          doc_id,
          '\')">View</button>'
        )
      ) %>%
      select(Document, everything())
  }

  if (isTRUE(specialtags)) {
    df <- df %>%
      rename("Special Entity" = "UPOS") %>%
      mutate(
        "Frequency Distribution" = paste0(
          '<button id2="custom_btn" onclick="Shiny.onInputChange(\'button_id2\', \'',
          `Special Entity`,
          '\')">View</button>'
        )
      ) %>%
      select("Frequency Distribution", everything())
  }

  if (isTRUE(delete)) {
    df <- df %>%
      mutate(
        Remove = paste0(
          '<button id="custom_btn_del" onclick="Shiny.onInputChange(\'button_id_del\', \'',
          doc_id,
          '\')">Remove</button>'
        )
      ) %>%
      select(Document, Remove, everything())
  }

  if (isTRUE(selection)) {
    extensions <- c("Buttons", "Select", "ColReorder")
    buttons <- c(buttons, c("selectAll", "selectNone"))
    select <- list(style = "os", items = "row")
  } else {
    extensions <- c("Buttons", "ColReorder")
    select <- NULL
  }

  tab <- DT::datatable(
    df,
    escape = escape,
    rownames = FALSE,
    caption = caption,
    selection = "none",
    extensions = extensions,
    filter = filter,
    options = list(
      colReorder = TRUE,
      fixedHeader = FALSE,
      pageLength = nrow,
      autoWidth = FALSE,
      scrollX = TRUE,
      dom = dom,
      buttons = buttons,
      select = select,
      lengthMenu = list(
        c(10, 25, 50, -1),
        c("10 rows", "25 rows", "50 rows", "Show all")
      ),
      columnDefs = columnDefs,
      initComplete = DT::JS(
        paste0(
          "function(settings, json) {
          var api = this.api();

          $(api.table().header()).css({
            'background': '",
          tall_primary,
          "',
            'color': 'white',
            'font-weight': '600',
            'font-size': '13px',
            'padding': '12px 8px',
            'border-bottom': '2px solid ",
          tall_secondary,
          "',
            'text-transform': 'uppercase',
            'letter-spacing': '0.5px'
          });

          $(api.table().container()).find('.dataTables_wrapper').css({
            'font-family': '-apple-system, BlinkMacSystemFont, \"Segoe UI\", Roboto, \"Helvetica Neue\", Arial, sans-serif'
          });

          $(api.table().container()).find('.dataTables_filter input').css({
            'border': '1px solid ",
          tall_border,
          "',
            'border-radius': '4px',
            'padding': '6px 12px',
            'margin-left': '8px',
            'font-size': '13px',
            'outline': 'none'
          });

          $(api.table().container()).find('.dataTables_length select').css({
            'border': '1px solid ",
          tall_border,
          "',
            'border-radius': '4px',
            'padding': '4px 8px',
            'margin': '0 8px',
            'font-size': '13px'
          });

          $(api.table().container()).find('.dt-buttons .dt-button').css({
            'background': '",
          tall_primary,
          "',
            'color': 'white',
            'border': 'none',
            'border-radius': '4px',
            'padding': '6px 14px',
            'margin-right': '6px',
            'font-size': '13px',
            'font-weight': '500',
            'cursor': 'pointer',
            'box-shadow': '0 1px 3px rgba(0,0,0,0.1)'
          });

          $(api.table().container()).find('.dataTables_paginate .paginate_button').css({
            'border-radius': '4px',
            'padding': '6px 12px',
            'margin': '0 2px',
            'border': '1px solid ",
          tall_border,
          "',
            'font-size': '13px'
          });

          $(api.table().container()).find('.dataTables_paginate .paginate_button.current').css({
            'background': '",
          tall_primary,
          "',
            'color': 'white',
            'border-color': '",
          tall_primary,
          "'
          });
        }"
        )
      ),
      rowCallback = DT::JS(
        paste0(
          "function(row, data, index) {
          if(index % 2 === 0) {
            $(row).css('background-color', '",
          tall_accent,
          "');
          } else {
            $(row).css('background-color', '",
          tall_white,
          "');
          }
        }"
        )
      )
    ),
    class = "cell-border stripe hover"
  ) %>%
    DT::formatStyle(
      names(df),
      textAlign = "center",
      fontSize = size,
      fontFamily = '-apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, "Helvetica Neue", Arial, sans-serif',
      color = tall_text,
      padding = '10px 8px'
    )

  if (!is.null(left)) {
    tab <- tab %>%
      DT::formatStyle(
        names(df)[left],
        textAlign = "left"
      )
  }

  if (!is.null(right)) {
    tab <- tab %>%
      DT::formatStyle(
        names(df)[right],
        textAlign = "right"
      )
  }

  if (!is.null(numeric)) {
    tab <- tab %>%
      formatRound(names(df)[c(numeric)], digits = round)
  }

  tab
}


### FUNCTIONS FOR EXPORTING PLOTS ----


plot2png <- function(p, filename, type = "vis", dpi = 300, height = 7) {
  zoom <- dpi / 96
  vwidth  <- round(height * 2 * 96)
  vheight <- round(height * 96)

  filename_html <- tools::file_path_sans_ext(filename)
  html_name <- paste0(filename_html, ".html")
  switch(
    type,
    vis = {
      visSave(p, html_name)
    },
    plotly = {
      htmlwidgets::saveWidget(p, file = html_name, selfcontained = FALSE)
    }
  )
  tallShot(url = html_name, zoom = zoom, vwidth = vwidth, vheight = vheight, file = filename)

  # Clean up temporary HTML and associated files
  unlink(html_name)
  unlink(paste0(filename_html, "_files"), recursive = TRUE)

  invisible(normalizePath(filename, mustWork = FALSE))
}

## freqGgplot ----
## ggplot for frequency plots to download

freqGgplot <- function(df, x = 2, y = 1, n = 20, title = "NOUN Frequency") {
  df <- df %>%
    dplyr::slice_head(n = n)

  col_x <- names(df)[x]
  col_y <- names(df)[y]

  max_val <- max(df[[col_x]], na.rm = TRUE)
  limit_x <- max_val + (max_val * 0.06)

  g <- ggplot(
    df,
    aes(x = .data[[col_x]], y = reorder(.data[[col_y]], .data[[col_x]]))
  ) +
    geom_col(color = "#c3d1be", fill = "#96af8e") +
    geom_text(
      aes(label = .data[[col_x]]),
      hjust = -0.2,
      color = "#4f7942",
      size = 3.7
    ) +
    labs(title = title, y = "", x = "Frequency") +
    scale_x_continuous(
      limits = c(0, limit_x),
      expand = c(0, 0)
    ) +
    theme(
      axis.text.y = element_text(size = 9),
      axis.text.x = element_text(size = 10),
      panel.grid.major = element_blank(),
      panel.grid.minor = element_blank(),
      panel.background = element_blank()
    )

  return(g)
}

topicGplot <- function(x, nPlot = 10, type = "beta") {
  # Identify ID column based on type
  id_col <- if (type == "beta") "word" else "doc"
  topic_names <- setdiff(colnames(x), id_col)

  # Reshape and filter top n per topic
  long_data <- x %>%
    pivot_longer(
      cols = all_of(topic_names),
      names_to = "topic",
      values_to = "probability"
    ) %>%
    group_by(topic) %>%
    slice_max(order_by = probability, n = nPlot) %>%
    arrange(desc(probability), .by_group = TRUE) %>%
    ungroup() %>%
    mutate(topic = paste0("Topic ", topic))

  # Reorder factor levels for y-axis
  long_data <- long_data %>%
    group_by(topic) %>%
    mutate(
      label = factor(.data[[id_col]], levels = rev(unique(.data[[id_col]])))
    ) %>%
    ungroup()

  # Select the required number of colors
  unique_topics <- unique(long_data$topic)
  colors <- colorlist()[seq_along(unique_topics)]

  # Build the plot
  g <- ggplot(long_data, aes(x = probability, y = label, fill = topic)) +
    geom_col(show.legend = FALSE) +
    facet_wrap(~topic, scales = "free") +
    scale_fill_manual(values = setNames(colors, unique_topics)) +
    theme_minimal(base_size = 11) +
    theme(
      axis.text.y = element_text(angle = 0, hjust = 0),
      panel.grid.major.y = element_blank()
    ) +
    labs(y = ifelse(type == "beta", "Word", "Document"), x = "Probability")

  return(g)
}

### deleteCache ------

#' Fast text recoding (Rcpp version)
#'
#' @description
#' Efficiently recodes text values using C++ hash tables. This is a drop-in
#' replacement for \code{txt_recode} but significantly faster for large vectors.
#'
#' @param x A character vector to recode
#' @param from A character vector with values of \code{x} which you want to recode
#' @param to A character vector with values you want to use to recode to
#' @param na.rm Logical, if set to TRUE, will put all values of \code{x} which
#'   have no matching value in \code{from} to NA. Defaults to FALSE
#'
#' @return A character vector of the same length as \code{x} where values
#'   matching \code{from} are replaced by corresponding values in \code{to}
#'
#' @details
#' This function uses C++ hash tables for O(1) lookup time, making it much
#' faster than the pure R implementation, especially for large datasets.
#'
#' Performance improvement: ~50-100x faster than base R \code{txt_recode}
#' for vectors with 100K+ elements.
#'
#' @examples
#' x <- c("NOUN", "VERB", "NOUN", "ADV")
#' txt_recode_fast(x,
#'   from = c("VERB", "ADV"),
#'   to = c("conjugated verb", "adverb")
#' )
#'
#' @export
txt_recode_fast <- function(x, from = c(), to = c(), na.rm = FALSE) {
  if (length(x) == 0) {
    return(x)
  }

  stopifnot(length(from) == length(to))

  # Chiama la funzione C++
  txt_recode_cpp(x = x, from = from, to = to, na_rm = na.rm)
}


#' Fast n-gram recoding for multiword detection
#'
#' @description
#' Efficiently combines consecutive tokens into multiword expressions using C++.
#' This function scans text sequentially to identify and merge n-gram patterns.
#'
#' @param x Character vector of tokens (e.g., lemmas or tokens)
#' @param compound Character vector of multiword expressions to match
#' @param ngram Integer vector indicating the length of each compound
#' @param sep String separator to use when joining tokens (default: " ")
#'
#' @return Character vector where matched n-grams are combined and subsequent
#'   tokens (that were merged) are set to NA
#'
#' @details
#' When a multiword match is found:
#' \itemize{
#'   \item The first position gets the combined multiword expression
#'   \item Subsequent positions that were merged are set to NA
#' }
#'
#' The function checks n-grams from longest to shortest to prioritize longer matches.
#'
#' Performance: ~80-150x faster than pure R implementation for typical text data.
#'
#' @examples
#' tokens <- c("machine", "learning", "is", "cool", "machine", "learning")
#' compounds <- c("machine learning")
#' ngrams <- c(2)
#' txt_recode_ngram_fast(tokens, compounds, ngrams, " ")
#' # Returns: c("machine learning", NA, "is", "cool", "machine learning", NA)
#'
#' @export
txt_recode_ngram_fast <- function(x, compound, ngram, sep = " ") {
  if (length(x) == 0 || length(compound) == 0) {
    return(x)
  }

  # Chiama la funzione C++
  txt_recode_ngram_cpp(x = x, compound = compound, ngram = ngram, sep = sep)
}


#' Batch text recoding (internal C++ function)
#'
#' @description
#' Internal C++ function for batch recoding multiple character vectors.
#' Efficiently recodes multiple text columns using a single hash table lookup.
#'
#' @param x List of character vectors to recode
#' @param from Character vector of values to match
#' @param to Character vector of replacement values
#' @param na_rm Logical, whether to set unmatched values to NA (default: FALSE)
#'
#' @return List of character vectors with recoded values
#'
#' @details
#' This is an internal C++ function exported for advanced users.
#' The hash table is constructed once and reused for all vectors in the list,
#' making it very efficient for batch operations.
#'
#' @keywords internal
#' @name txt_recode_batch_cpp
NULL


#' Internal C++ text recoding function
#'
#' @description
#' Internal C++ implementation of text recoding using hash tables.
#' Users should call \code{\link{txt_recode_fast}} instead.
#'
#' @param x Character vector to recode
#' @param from Character vector of values to match
#' @param to Character vector of replacement values
#' @param na_rm Logical, whether to set unmatched values to NA
#'
#' @return Character vector with recoded values
#'
#' @keywords internal
#' @name txt_recode_cpp
NULL


#' Internal C++ n-gram recoding function
#'
#' @description
#' Internal C++ implementation of n-gram recoding for multiword detection.
#' Users should call \code{\link{txt_recode_ngram_fast}} instead.
#'
#' @param x Character vector of tokens
#' @param compound Character vector of multiword expressions to match
#' @param ngram Integer vector indicating the length of each compound
#' @param sep String separator to use when joining tokens
#'
#' @return Character vector where matched n-grams are combined
#'
#' @keywords internal
#' @name txt_recode_ngram_cpp
NULL


#' Optimized multiword processing workflow
#'
#' @description
#' Complete optimized workflow for multiword detection and processing.
#' Uses C++ functions and data.table for maximum performance.
#'
#' @param x2 Data frame with token information
#' @param stats Data frame with multiword statistics (keyword, ngram columns)
#' @param term Type of term to process: "lemma" or "token"
#'
#' @return Data frame with columns: doc_id, term_id, multiword, upos_multiword, ngram
#'
#' @details
#' This function replaces the original switch block with an optimized version
#' that uses:
#' \itemize{
#'   \item C++ functions for text recoding
#'   \item Vectorized operations instead of multiple mutate calls
#'   \item Pre-computed lookups to avoid repeated joins
#' }
#'
#' @examples
#' \dontrun{
#' result <- process_multiwords_fast(dfTag, multiword_stats, term = "lemma")
#' }
#'
#' @export
process_multiwords_fast <- function(x2, stats, term = c("lemma", "token")) {
  term <- match.arg(term)

  # Pre-calcola lookup per evitare join ripetuti
  keyword_lookup <- stats$ngram
  names(keyword_lookup) <- stats$keyword

  # Seleziona la colonna corretta
  term_col <- if (term == "lemma") x2$lemma else x2$token

  # Usa la funzione C++ per identificare multiword
  multiword <- txt_recode_ngram_fast(
    x = term_col,
    compound = stats$keyword,
    ngram = stats$ngram,
    sep = " "
  )

  # Operazioni vettorizzate (molto più veloci di mutate multiple)
  upos_multiword <- ifelse(
    term_col == multiword,
    x2$upos,
    ifelse(is.na(multiword), "NGRAM_MERGED", "MULTIWORD")
  )

  # Lookup diretto invece di join
  ngram_values <- keyword_lookup[multiword]

  # Crea risultato direttamente (più veloce di select + mutate)
  data.frame(
    doc_id = x2$doc_id,
    term_id = x2$term_id,
    multiword = multiword,
    upos_multiword = upos_multiword,
    ngram = unname(ngram_values),
    stringsAsFactors = FALSE
  )
}

test_that("txt_recode_fast replaces matching values", {
  x <- c("NOUN", "VERB", "NOUN", "ADV")
  result <- txt_recode_fast(x, from = c("VERB", "ADV"), to = c("v", "a"))
  expect_equal(result, c("NOUN", "v", "NOUN", "a"))
})

test_that("txt_recode_fast returns unchanged vector when no matches", {
  x <- c("NOUN", "VERB")
  result <- txt_recode_fast(x, from = c("ADJ"), to = c("adj"))
  expect_equal(result, c("NOUN", "VERB"))
})

test_that("txt_recode_fast handles empty vector", {
  result <- txt_recode_fast(character(0), from = c("A"), to = c("B"))
  expect_equal(result, character(0))
})

test_that("txt_recode_fast handles empty from/to", {
  x <- c("A", "B", "C")
  result <- txt_recode_fast(x, from = character(0), to = character(0))
  expect_equal(result, x)
})

test_that("txt_recode_fast with na.rm sets unmatched to NA", {
  x <- c("NOUN", "VERB", "ADJ")
  result <- txt_recode_fast(x, from = c("NOUN"), to = c("n"), na.rm = TRUE)
  expect_equal(result, c("n", NA, NA))
})

test_that("txt_recode_fast errors when from/to lengths differ", {
  expect_error(
    txt_recode_fast(c("A"), from = c("A", "B"), to = c("a")),
    "length\\(from\\) == length\\(to\\)"
  )
})

test_that("txt_recode_ngram_fast identifies multiword expressions", {
  tokens <- c("machine", "learning", "is", "cool", "machine", "learning")
  compounds <- c("machine learning")
  ngrams <- c(2L)
  result <- txt_recode_ngram_fast(tokens, compounds, ngrams, " ")

  expect_equal(result[1], "machine learning")
  expect_true(is.na(result[2]))
  expect_equal(result[3], "is")
  expect_equal(result[4], "cool")
  expect_equal(result[5], "machine learning")
  expect_true(is.na(result[6]))
})

test_that("txt_recode_ngram_fast handles no matches", {
  tokens <- c("hello", "world")
  compounds <- c("machine learning")
  ngrams <- c(2L)
  result <- txt_recode_ngram_fast(tokens, compounds, ngrams, " ")

  expect_equal(result, c("hello", "world"))
})

test_that("txt_recode_ngram_fast handles empty input", {
  result <- txt_recode_ngram_fast(character(0), character(0), integer(0), " ")
  expect_equal(result, character(0))
})

test_that("process_multiwords_fast returns expected structure", {
  x2 <- data.frame(
    doc_id = rep("doc1", 6),
    term_id = 1:6,
    token = c("machine", "learning", "is", "very", "deep", "learning"),
    lemma = c("machine", "learning", "be", "very", "deep", "learning"),
    upos = c("NOUN", "NOUN", "AUX", "ADV", "ADJ", "NOUN"),
    stringsAsFactors = FALSE
  )
  stats <- data.frame(
    keyword = c("machine learning"),
    ngram = c(2L),
    stringsAsFactors = FALSE
  )

  result <- process_multiwords_fast(x2, stats, term = "lemma")

  expect_s3_class(result, "data.frame")
  expect_named(result, c("doc_id", "term_id", "multiword", "upos_multiword", "ngram"))
  expect_equal(nrow(result), 6)
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

cpp_chisq <- function(T1, T2, Tsum) {
    .Call('_tall_cpp_chisq', PACKAGE = 'tall', T1, T2, Tsum)
}

cpp_split_tab <- function(tab, indices) {
    .Call('_tall_cpp_split_tab', PACKAGE = 'tall', tab, indices)
}

cpp_switch_docs <- function(tab1, tab2) {
    .Call('_tall_cpp_switch_docs', PACKAGE = 'tall', tab1, tab2)
}

extract_np_cpp <- function(sent_id, token_id, head_token_id, dep_rel, upos, terms, ngram_max = 5L, max_gap = 3L) {
    .Call('_tall_extract_np_cpp', PACKAGE = 'tall', sent_id, token_id, head_token_id, dep_rel, upos, terms, ngram_max, max_gap)
}

extract_svo_cpp <- function(sent_id, token_id, head_token_id, dep_rel, upos, terms) {
    .Call('_tall_extract_svo_cpp', PACKAGE = 'tall', sent_id, token_id, head_token_id, dep_rel, upos, terms)
}

calc_reciprocal_sum_cpp <- function(ngram_words, word_freq_names, word_freq_values) {
    .Call('_tall_calc_reciprocal_sum_cpp', PACKAGE = 'tall', ngram_words, word_freq_names, word_freq_values)
}

count_lexical_cpp <- function(ngram_pos, lexical_pos) {
    .Call('_tall_count_lexical_cpp', PACKAGE = 'tall', ngram_pos, lexical_pos)
}

generate_ngrams_cpp <- function(sentences_terms, sentences_pos, max_ngram, lexical_pos) {
    .Call('_tall_generate_ngrams_cpp', PACKAGE = 'tall', sentences_terms, sentences_pos, max_ngram, lexical_pos)
}

syntactic_complexity_cpp <- function(doc_id, sent_id, token_id, head_token_id, dep_rel, upos) {
    .Call('_tall_syntactic_complexity_cpp', PACKAGE = 'tall', doc_id, sent_id, token_id, head_token_id, dep_rel, upos)
}

txt_recode_cpp <- function(x, from, to, na_rm = FALSE) {
    .Call('_tall_txt_recode_cpp', PACKAGE = 'tall', x, from, to, na_rm)
}

txt_recode_ngram_cpp <- function(x, compound, ngram, sep = " ") {
    .Call('_tall_txt_recode_ngram_cpp', PACKAGE = 'tall', x, compound, ngram, sep)
}

txt_recode_batch_cpp <- function(x, from, to, na_rm = FALSE) {
    .Call('_tall_txt_recode_batch_cpp', PACKAGE = 'tall', x, from, to, na_rm)
}

#' Lemmatized Text of Moby-Dick (Chapters 1-10)
#'
#' This dataset contains the lemmatized version of the first 10 chapters of the novel Moby-Dick by Herman Melville.
#' The data is structured as a dataframe with multiple linguistic annotations.
#'
#' @format A dataframe with multiple rows and 26 columns:
#' \describe{
#'   \item{doc_id}{Character: Unique document identifier}
#'   \item{paragraph_id}{Integer: Paragraph index within the document}
#'   \item{sentence_id}{Integer: Sentence index within the paragraph}
#'   \item{sentence}{Character: Original sentence text}
#'   \item{start}{Integer: Start position of the token in the sentence}
#'   \item{end}{Integer: End position of the token in the sentence}
#'   \item{term_id}{Integer: Unique term identifier}
#'   \item{token_id}{Integer: Token index in the sentence}
#'   \item{token}{Character: Original token (word)}
#'   \item{lemma}{Character: Lemmatized form of the token}
#'   \item{upos}{Character: Universal POS tag}
#'   \item{xpos}{Character: Language-specific POS tag}
#'   \item{feats}{Character: Morphological features}
#'   \item{head_token_id}{Integer: Head token in dependency tree}
#'   \item{dep_rel}{Character: Dependency relation label}
#'   \item{deps}{Character: Enhanced dependency relations}
#'   \item{misc}{Character: Additional information}
#'   \item{folder}{Character: Folder containing the document}
#'   \item{split_word}{Character: The word used to separate the chapters in the original book}
#'   \item{filename}{Character: Source file name}
#'   \item{doc_selected}{Logical: Whether the document is selected}
#'   \item{POSSelected}{Logical: Whether POS was selected}
#'   \item{sentence_hl}{Character: Highlighted sentence}
#'   \item{docSelected}{Logical: Whether the document was manually selected}
#'   \item{noHapax}{Logical: Whether hapax legomena were removed}
#'   \item{noSingleChar}{Logical: Whether single-character words were removed}
#'   \item{lemma_original_nomultiwords}{Character: Lemmatized form without multi-word units}
#' }
#'
#' @usage data(mobydick)
#' @source Extracted and processed from the text of Moby-Dick by Herman Melville.
#' @examples
#' data(mobydick)
#' head(mobydick)
"mobydick"

utils::globalVariables(c(
  "doc_id",
  "uc",
  "uce",
  "segment_size",
  "upos",
  "noSingleChar",
  "token",
  "lemma",
  "freq",
  "chi_square",
  "cluster",
  "negative",
  "positive",
  "term",
  "freq_true",
  "indep",
  "p_value",
  ".",
  "segment",
  "seg_doc",
  "prev_doc"
))


#' Segment clustering based on the Reinert method - Simple clustering
#'
#' @param x tall data frame of documents
#' @param k maximum number of clusters to compute
#' @param term indicates the type of form "lemma" or "token". Default value is term = "lemma".
#' @param segment_size number of forms by document. Default value is segment_size = 40
#' @param min_segment_size minimum number of forms by document. Default value is min_segment_size = 5
#' @param min_split_members minimum number of segment in a cluster
#' @param cc_test contingency coefficient value for feature selection
#' @param tsj minimum frequency value for feature selection
#'
#' @details
#' See the references for original articles on the method.
#' Special thanks to the authors of the rainette package (https://github.com/juba/rainette)
#' for inspiring the coding approach used in this function.
#'
#'
#' @return
#' The result is a list of both class `hclust` and `reinert_tall`.
#'
#' @references
#'
#' - Reinert M, Une methode de classification descendante hierarchique: application à l'analyse lexicale par contexte, Cahiers de l'analyse des donnees, Volume 8, Numéro 2, 1983. <https://www.numdam.org/item/?id=CAD_1983__8_2_187_0>
#' - Reinert M., Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval, Bulletin de Methodologie Sociologique, Volume 26, Numero 1, 1990. \doi{10.1177/075910639002600103}
#' - Barnier J., Privé F., rainette: The Reinert Method for Textual Data Clustering, 2023, \doi{10.32614/CRAN.package.rainette}
#'
#' @examples
#' \donttest{
#' data(mobydick)
#' res <- reinert(
#'   x = mobydick,
#'   k = 10,
#'   term = "token",
#'   segment_size = 40,
#'   min_segment_size = 5,
#'   min_split_members = 10,
#'   cc_test = 0.3,
#'   tsj = 3
#' )
#' }
#'
#' @export
#'

reinert <- function(
  x,
  k = 10,
  term = "token",
  segment_size = 40,
  min_segment_size = 3,
  min_split_members = 5,
  cc_test = 0.3,
  tsj = 3
) {
  if (min_split_members < 3) {
    warning("min_split_members has been set to 3, its smallest possible value.")
    min_split_members <- 3
  }

  uposList <- c(
    "ADP",
    "DET",
    "INTJ",
    "NUM",
    "PART",
    "PRON",
    "PUNCT",
    "SCONJ",
    "SYM",
    "X",
    "EMAIL",
    "EMOJI",
    "HASH",
    "IP_ADDRESS",
    "MENTION",
    "URL"
  )

  ## Create DTM from TAll Data Frame
  switch(
    term,
    token = {
      x <- x %>%
        dplyr::filter(!upos %in% uposList & noSingleChar) %>%
        mutate(token_rainette = tolower(token))
      colTerm <- "token_rainette"
    },
    lemma = {
      x <- x %>%
        dplyr::filter(!upos %in% uposList & noSingleChar) %>%
        mutate(lemma_rainette = tolower(lemma))
      colTerm <- "lemma_rainette"
    }
  )

  ## Create segments by segment_size
  idTable <- split_segments(x$doc_id, segment_size)

  ## Correspondance table between uce and uc
  corresp_uce_uc_full <- merge_small_segments(
    idTable,
    min_length = min_segment_size
  )

  x <- x %>% bind_cols(corresp_uce_uc_full %>% select(-"doc_id"))

  corresp_uce_uc_full$term <- x[[colTerm]]

  ## Table with assosacions among uc and uce
  corresp_uce_uc <- corresp_uce_uc_full %>%
    select(-"doc_id", -"term") %>%
    distinct() %>%
    as.data.frame()

  ## Table with segments
  corresp_uce_uc_full <- corresp_uce_uc_full %>%
    group_by(doc_id, uc) %>%
    summarize(segment = paste0(term, collapse = " ")) %>%
    as.data.frame()

  corresp_uce_uc_full <- corresp_uce_uc_full %>%
    group_by(uc) %>%
    reframe(
      doc_id = doc_id[1],
      segment = paste0(segment, collapse = " ")
    ) %>%
    distinct() %>%
    data.frame()

  row.names(corresp_uce_uc_full) <- corresp_uce_uc_full$uc

  dtf <- document_term_frequencies(x, document = "uc", term = colTerm)

  ## Apply binary weighting
  dtf <- dtf %>%
    mutate(freq = freq^0)

  ## create DTM
  dtmOriginal <- document_term_matrix(dtf, weight = "freq")

  dtmOriginal <- dtm_remove_lowfreq(dtmOriginal, minfreq = 3)
  ## Remove low frequency terms
  dtm <- as.matrix(dtmOriginal)

  ## Remove empty strings to avoid subcript out of bounds errors
  ind <- rowSums(dtm) > 0
  dtm <- dtm[ind, ]

  message("  Clustering...")

  ## Initialize results list with first dtm
  res <- list(list(tabs = list(dtm)))

  exclusion_list <- NULL

  i <- 1

  while (i < k) {
    ## Split the biggest group
    nrows <- purrr::map(res[[i]]$tabs, nrow)
    names(nrows) <- 1:length(nrows)
    # nrows <- lapply(nrows, function(x) if (is.null(x)) 0 else x)
    # biggest_group <- which.max(nrows)
    if (length(exclusion_list) > 0) {
      biggest_group <- as.numeric(names(which.max(nrows[-exclusion_list])))
    } else {
      biggest_group <- as.numeric(names(which.max(nrows)))
    }

    if (nrow(res[[i]]$tabs[[biggest_group]]) < min_split_members) {
      message(
        "! No more group bigger than min_split_members. Stopping after iteration ",
        i,
        "."
      )
      k <- i
      break
    }
    tab <- res[[i]]$tabs[[biggest_group]]
    ## textmodel_ca only works if nrow >= 3 and ncol >= 3
    if (nrow(tab) < 3 || ncol(tab) < 3) {
      message(
        "! Tab to be splitted is not big enough. Stopping after iteration ",
        i,
        "."
      )
      k <- i
      break
    }

    ## Remove documents and features with zero occurrences
    tab <- tab[rowSums(tab) > 0, colSums(tab) > 0]

    clusters <- cluster_tab(tab, cc_test = cc_test, tsj = tsj)
    if (
      "matrix" %in%
        class(clusters$tabs[[1]]) &
        "matrix" %in% class(clusters$tabs[[2]])
    ) {
      ## Populate results
      res[[i + 1]] <- list()
      res[[i + 1]]$height <- clusters$height
      res[[i + 1]]$splitted <- c(biggest_group, biggest_group + 1)
      res[[i + 1]]$tabs <- append(
        res[[i]]$tabs[-biggest_group],
        clusters$tabs,
        after = biggest_group - 1
      )
      res[[i + 1]]$groups <- append(
        res[[i]]$groups[-biggest_group],
        clusters$groups,
        after = biggest_group - 1
      )
      i <- i + 1
    } else {
      exclusion_list <- c(exclusion_list, biggest_group)
    }
  }
  # })

  if (k == 1) {
    message("! No computed clusters. Returning NULL.")
    return(NULL)
  }

  res <- res[-1]

  ## Compute the merge element of resulting hclust result
  groups <- 1:k
  merge <- matrix(nrow = 0, ncol = 2)
  for (i in (k - 1):1) {
    split <- res[[i]]$splitted
    merge <- rbind(merge, -groups[split])
    groups <- groups[-split[2]]
    groups[split[1]] <- -(k - i)
  }

  # Compute groups by uce at each k
  uce_groups <- list()
  for (i in 1:(k - 1)) {
    # group <- rep(NA, nrow(corresp_uce_uc))
    group <- rep(NA, nrow(corresp_uce_uc_full))
    indices <- res[[i]]$groups
    for (group_index in seq_along(indices)) {
      group[corresp_uce_uc_full[indices[[group_index]], "uc"]] <- group_index
      # group[corresp_uce_uc_full$uc %in% as.numeric(indices[[group_index]])] <- group_index
    }
    uce_groups[[i]] <- group
  }

  ## Get the final group element of resulting hclust result
  ## by uce
  group <- uce_groups[[k - 1]]

  message("  Done.")

  ## Compute and return hclust-class result
  hres <- list(
    method = "reinert",
    call = match.call(),
    height = cumsum(rev(purrr::map_dbl(res, ~ .x$height))),
    order = 1:k,
    labels = as.character(1:k),
    merge = merge,
    group = group,
    uce_groups = uce_groups,
    corresp_uce_uc = corresp_uce_uc,
    corresp_uce_uc_full = corresp_uce_uc_full,
    dtmOriginal = dtmOriginal
  )

  class(hres) <- c("reinert_tall", "hclust")
  hres
}


order_docs <- function(m) {
  ## Compute first factor of CA on DTM
  ## Code taken from getAnywhere(textmodel_ca.dfm)
  p <- m / sum(m)
  rm <- rowSums(p)
  cm <- colSums(p)
  ep <- tcrossprod(rm, cm)
  s <- (p - ep) / sqrt(ep)
  dec <- RSpectra::svds(s, k = 1, nv = 0)
  coord <- dec$u[, 1] / sqrt(rm)

  ## Order documents by their first factor coordinates
  indices <- (seq_along(coord))[order(coord)]

  return(indices)
}


switch_docs <- function(m, indices, max_index, max_chisq) {
  ## Group indices and tabs
  group1 <- indices[1:which(indices == max_index)]
  group2 <- indices[(which(indices == max_index) + 1):length(indices)]

  switched <- TRUE

  ## Run while points are switched
  while (switched) {
    switched <- FALSE

    tab1 <- m[group1, , drop = FALSE]
    tab2 <- m[group2, , drop = FALSE]

    chisq_values <- cpp_switch_docs(tab1, tab2)
    current_max <- max(chisq_values, na.rm = TRUE)

    if (current_max > max_chisq) {
      switched <- TRUE
      to_switch <- indices[which.max(chisq_values)]
      if (to_switch %in% group1) {
        group1 <- group1[group1 != to_switch]
        group2 <- c(group2, to_switch)
      } else {
        group2 <- group2[group2 != to_switch]
        group1 <- c(group1, to_switch)
      }
      max_chisq <- current_max
    }
  }

  return(list(
    indices1 = group1,
    indices2 = group2,
    chisq = max_chisq
  ))
}


select_features <- function(m, indices1, indices2, cc_test = 0.3, tsj = 3) {
  ## features count for each group
  tab1 <- colSums(m[indices1, , drop = FALSE])
  tab2 <- colSums(m[indices2, , drop = FALSE])
  ## Total number of features in each group
  nfeat_group1 <- sum(tab1)
  nfeat_group2 <- sum(tab2)
  ## Observed frequency of features
  observed <- rbind(tab1, tab2)
  ## Expected frequency of features
  expected_prop <- (tab1 + tab2) / sum(tab1 + tab2)
  expected <- rbind(expected_prop * nfeat_group1, expected_prop * nfeat_group2)
  ## Chi2 and contingency coefficients for each feature
  feat_chisq <- colSums((observed - expected)^2 / expected)
  ## C contingency coefficient, sqrt(khi2 / (khi2 + N))
  feat_cc <- sqrt(feat_chisq / (feat_chisq + colSums(observed)))

  ## Features selection
  cols1 <- character()
  cols2 <- character()

  for (i in seq_along(feat_cc)) {
    cc <- feat_cc[i]
    name <- names(feat_cc)[i]
    ## Keep feature if cc <= cc_test and frequency > 0
    if (cc <= cc_test && tab1[i] >= tsj) {
      cols1 <- c(cols1, name)
    }
    if (cc <= cc_test && tab2[i] >= tsj) {
      cols2 <- c(cols2, name)
    }
    ## If cc > cc_test, only keep feature in the group
    ## where observed frequency > expected frequency
    if (cc > cc_test) {
      if (tab1[i] > expected[1, i] && tab1[i] >= tsj) {
        cols1 <- c(cols1, name)
      }
      if (tab2[i] > expected[2, i] && tab2[i] >= tsj) {
        cols2 <- c(cols2, name)
      }
    }
  }

  return(list(
    cols1 = cols1,
    cols2 = cols2
  ))
}


cluster_tab <- function(m, cc_test = 0.3, tsj = 3) {
  uc <- row.names(m)
  ## Remove features with zero
  storage.mode(m) <- "integer"

  ## First step : CA partition

  indices <- order_docs(m)
  res <- cpp_split_tab(m, indices)
  max_index <- res$max_index
  max_chisq <- res$max_chisq

  ## Second step : switching docs

  res <- switch_docs(m, indices, max_index, max_chisq)
  indices1 <- res$indices1
  indices2 <- res$indices2
  chisq <- res$chisq

  ## Third step : features selection

  res <- select_features(m, indices1, indices2, cc_test, tsj)
  cols1 <- res$cols1
  cols2 <- res$cols2

  return(list(
    groups = list(
      rainette_uc_id = uc[indices1],
      rainette_uc_id = uc[indices2]
    ),
    tabs = list(
      m[indices1, cols1],
      m[indices2, cols2]
    ),
    height = chisq
  ))
}


## Split segments
split_segments <- function(doc_id, segment_size) {
  segment_id <- integer(length(doc_id))
  current_segment <- 1
  count <- 0

  for (i in seq_along(doc_id)) {
    count <- count + 1
    segment_id[i] <- current_segment

    # Controllo fine del segmento o cambio di documento
    if (
      count == segment_size ||
        (i < length(doc_id) && doc_id[i] != doc_id[i + 1])
    ) {
      current_segment <- current_segment + 1
      count <- 0
    }
  }

  return(data.frame(doc_id = doc_id, uce = segment_id))
}

merge_small_segments <- function(idTable, min_length = 5) {
  # Compute segment sizes and first doc_id per segment
  seg_info <- idTable %>%
    group_by(uce) %>%
    mutate(segment_size = n()) %>%
    ungroup()

  # Get the doc_id of the first row of each segment
  seg_doc <- seg_info %>%
    group_by(uce) %>%
    summarize(seg_doc = doc_id[1], .groups = "drop")

  # Build lookup for previous segment's doc
  seg_doc <- seg_doc %>%
    mutate(prev_doc = dplyr::lag(seg_doc, default = ""))

  seg_info <- seg_info %>%
    left_join(seg_doc, by = "uce") %>%
    mutate(
      # Only merge with previous segment if same document
      uc = if_else(segment_size < min_length & seg_doc == prev_doc, uce - 1, uce)
    ) %>%
    select(-segment_size, -seg_doc, -prev_doc)

  seg_info
}

#' Extract Terms and Segments for Document Clusters
#'
#' This function processes the results of a document clustering algorithm based on the Reinert method.
#' It computes the terms and their significance for each cluster, as well as the associated document segments.
#'
#' @param res A list containing the results of the Reinert clustering algorithm. Must include at least `dtm` (a document-term matrix) and `corresp_uce_uc_full` (a correspondence between segments and clusters).
#' @param cutree A custom cutree structure. If `NULL`, the default `cutree_reinart` is used to determine cluster membership.
#' @param k A vector of integers specifying the clusters to analyze. Default is `1`.
#' @param negative Logical. If `TRUE`, include negative terms in the results. If `FALSE`, exclude them. Default is `TRUE`.
#'
#' @return A list with the following components:
#' \item{terms}{A data frame of significant terms for each cluster. Columns include:
#'   \itemize{
#'     \item \code{chi_square}: Chi-squared statistic for the term.
#'     \item \code{p_value}: P-value of the chi-squared test.
#'     \item \code{sign}: Significance of the term (\code{positive}, \code{negative}, or \code{none}).
#'     \item \code{term}: The term itself.
#'     \item \code{freq}: Observed frequency of the term in the cluster.
#'     \item \code{indep}: Expected frequency of the term under independence.
#'     \item \code{cluster}: The cluster ID.
#'   }
#' }
#' \item{segments}{A data frame of document segments associated with each cluster. Columns include:
#'   \itemize{
#'     \item \code{uc}: Unique segment identifier.
#'     \item \code{doc_id}: Document ID for the segment.
#'     \item \code{cluster}: Cluster ID.
#'     \item \code{segment}: The text content of each segment.
#'   }
#' }
#'
#' @details The function integrates document-term matrix rows for missing segments, calculates term statistics for each cluster,
#' and filters terms based on their significance. Terms can be excluded based on their significance (\code{signExcluded}).
#'
#' @examples
#' \donttest{
#' data(mobydick)
#' res <- reinert(
#'   x = mobydick,
#'   k = 10,
#'   term = "token",
#'   segment_size = 40,
#'   min_segment_size = 5,
#'   min_split_members = 10,
#'   cc_test = 0.3,
#'   tsj = 3
#' )
#'
#' tc <- term_per_cluster(res, cutree = NULL, k = 1:10, negative = FALSE)
#'
#' head(tc$segments, 10)
#'
#' head(tc$terms, 10)
#' }
#'
#' @export

term_per_cluster <- function(res, cutree = NULL, k = 1, negative = TRUE) {
  k <- k[!is.na(k)]
  dtm <- res$dtmOriginal
  groups <- cutree_reinart(res, cutree)
  terms <- colnames(dtm)

  ## integrate dtm with removed segments (by full-zero rows)
  label <- row.names(dtm)
  max_ind <- max(res$corresp_uce_uc_full$uc)

  ind <- setdiff(as.character(1:max_ind), label)

  if (length(ind) > 0) {
    m <- Matrix::Matrix(0, nrow = length(ind), ncol = ncol(dtm), sparse = TRUE)
    rownames(m) <- ind
    colnames(m) <- colnames(dtm)
    dtm <- rbind(dtm, m)
    dtm <- dtm[order(as.numeric(rownames(dtm))), ]
  }

  terms_list <- list()
  segments_list <- list()
  K <- k
  for (i in 1:length(K)) {
    k <- K[i]
    ## list of segments following into the cluster k
    select <- (groups == k & !is.na(groups))
    segments <- row.names(dtm)[select]
    segments_df <- tibble(uc = as.numeric(segments)) %>%
      left_join(res$corresp_uce_uc_full, by = "uc") %>%
      mutate(cluster = k)
    segments_list[[k]] <- segments_df

    chi_res_df <- dtm_chisq(dtm, groups = select) %>%
      mutate(indep = sum(freq_true) / sum(freq)) %>% # expected proportion for independence
      mutate(cluster = k) %>%
      rename(
        "chi_square" = "chisq",
        "p_value" = "p.value"
      )

    terms_list[[k]] <- chi_res_df
  }

  if (isTRUE(negative)) {
    signExcluded <- c("none")
  } else {
    signExcluded <- c("none", "negative")
  }

  terms_list <- bind_rows(terms_list) %>%
    group_by(term) %>%
    mutate(freq = sum(freq_true)) %>%
    ungroup() %>%
    filter(p_value <= 0.001) %>%
    mutate(
      freq = freq_true / freq,
      sign = ifelse(freq > indep, "positive", "negative")
    ) %>%
    filter(!sign %in% signExcluded)

  segments_list <- bind_rows(segments_list) %>% drop_na("doc_id")

  return(list(terms = terms_list, segments = segments_list))
}

# plot for terms by cluster

#' Plot Terms by Cluster
#'
#' This function creates a horizontal bar plot to visualize the most significant terms
#' for each cluster, based on their Chi-squared statistics.
#'
#' @param terms A data frame containing terms and their associated statistics, such as Chi-squared values,
#' generated by the `term_per_cluster` function. The data frame must include the following columns:
#' \itemize{
#'   \item \code{term}: The term to plot.
#'   \item \code{chi_square}: The Chi-squared statistic associated with the term.
#'   \item \code{sign}: The sign of the term (\code{"positive"} or \code{"negative"}).
#' }
#' @param nPlot Integer. The number of top terms to plot for each sign (\code{"positive"} and \code{"negative"}). Default is 10.
#'
#' @return An interactive horizontal bar plot (using `plotly`) displaying the top terms for each cluster. The plot includes:
#' \itemize{
#'   \item Bars representing the Chi-squared values of terms.
#'   \item Hover information displaying the term and its Chi-squared value.
#' }
#'
#' @details The function organizes the input data by Chi-squared values and selects the top terms for each sign.
#' The plot uses different colors for positive and negative terms, with hover tooltips providing detailed information.
#'
#' @seealso \code{\link{term_per_cluster}}
#'
#' @examples
#' \dontrun{
#' data(mobydick)
#' res <- reinert(
#'   x = mobydick,
#'   k = 10,
#'   term = "token",
#'   segment_size = 40,
#'   min_segment_size = 5,
#'   min_split_members = 10,
#'   cc_test = 0.3,
#'   tsj = 3
#' )
#'
#' tc <- term_per_cluster(res, cutree = NULL, k = 1, negative = FALSE)
#'
#' fig <- reinPlot(tc$terms, nPlot = 10)
#' }
#'
#' @export
#'
reinPlot <- function(terms, nPlot = 10) {
  # Separate positive and negative terms
  dfPositive <- terms %>%
    filter(sign == "positive") %>%
    ungroup() %>%
    arrange(desc(chi_square)) %>%
    slice_head(n = nPlot)

  dfNegative <- terms %>%
    filter(sign == "negative") %>%
    ungroup() %>%
    arrange(desc(chi_square)) %>%
    slice_head(n = nPlot)

  # Combine positive and negative terms
  dfPlot <- bind_rows(dfPositive, dfNegative) %>%
    mutate(
      term = factor(term, levels = rev(c(dfPositive$term, dfNegative$term))) # Invertire l'ordine
    )

  # Assign colors
  color <- colorlist()

  # build the plot
  fig1 <- plot_ly(
    data = dfPlot,
    x = ~chi_square,
    y = ~term,
    type = "bar",
    orientation = "h",
    marker = list(color = ~ ifelse(sign == "positive", color[1], color[2])),
    hovertemplate = "<b><i>Term: %{y}</i></b> <br> <b><i>Chi2: %{x}</i></b><extra></extra>"
  )

  # layout
  fig1 <- fig1 %>%
    plotly::layout(
      yaxis = list(
        title = "Terms",
        showgrid = FALSE,
        showline = FALSE,
        showticklabels = TRUE,
        domain = c(0, 1)
      ),
      xaxis = list(
        title = "Chi2",
        zeroline = FALSE,
        showline = FALSE,
        showticklabels = TRUE,
        showgrid = FALSE
      ),
      plot_bgcolor = "rgba(0, 0, 0, 0)",
      paper_bgcolor = "rgba(0, 0, 0, 0)"
    ) %>%
    plotly::config(
      displaylogo = FALSE,
      modeBarButtonsToRemove = c(
        "sendDataToCloud",
        "pan2d",
        "select2d",
        "lasso2d",
        "toggleSpikelines",
        "hoverClosestCartesian",
        "hoverCompareCartesian"
      )
    )

  return(fig1)
}

# summarize Reinert Results

#' Summarize Reinert Clustering Results
#'
#' This function summarizes the results of the Reinert clustering algorithm, including the most frequent documents and significant terms for each cluster.
#' The input is the result returned by the `term_per_cluster` function.
#'
#' @param tc A list returned by the \code{term_per_cluster} function. The list includes:
#' \itemize{
#'   \item \code{segments}: A data frame with segments information, including \code{cluster} and \code{doc_id}.
#'   \item \code{terms}: A data frame with terms information, including \code{cluster}, \code{sign}, \code{chi_square}, and \code{term}.
#' }
#' @param n Integer. The number of top terms (based on Chi-squared value) to include in the summary for each cluster and sign. Default is 10.
#'
#' @return A data frame summarizing the clustering results. The table includes:
#' \itemize{
#'   \item \code{cluster}: The cluster ID.
#'   \item \code{Positive terms}: The top \code{n} positive terms for each cluster, concatenated into a single string.
#'   \item \code{Negative terms}: The top \code{n} negative terms for each cluster, concatenated into a single string.
#'   \item \code{Most frequent document}: The document ID that appears most frequently in each cluster.
#'   \item \code{N. of Documents per Cluster}: The number of documents in each cluster.
#' }
#'
#' @details This function performs the following steps:
#' \enumerate{
#'   \item Extracts the most frequent document for each cluster.
#'   \item Summarizes the number of documents per cluster.
#'   \item Selects the top \code{n} terms for each cluster, separated by positive and negative signs.
#'   \item Combines the terms and segment information into a final summary table.
#' }
#'
#'
#' @seealso \code{\link{term_per_cluster}}, \code{\link{reinPlot}}
#'
#' @examples
#' \donttest{
#' data(mobydick)
#' res <- reinert(
#'   x = mobydick,
#'   k = 10,
#'   term = "token",
#'   segment_size = 40,
#'   min_segment_size = 5,
#'   min_split_members = 10,
#'   cc_test = 0.3,
#'   tsj = 3
#' )
#'
#' tc <- term_per_cluster(res, cutree = NULL, k = 1:10, negative = FALSE)
#'
#' S <- reinSummary(tc, n = 10)
#'
#' head(S, 10)
#' }
#'
#' @export
#'
reinSummary <- function(tc, n = 10) {
  segments <- tc$segments %>%
    group_by(cluster) %>%
    # summarise("Number of Segments" = n()) %>%
    # group_by(cluster) %>%
    summarize(
      "Most frequent document" = names(which.max(table(doc_id))),
      "N. of Segments per Cluster" = n()
    )

  terms <- tc$terms %>%
    group_by(cluster, sign) %>%
    slice_max(order_by = chi_square, n = 10, with_ties = TRUE) %>%
    summarize(terms = paste0(term, collapse = "; "), .groups = "drop") %>%
    pivot_wider(
      names_from = sign,
      values_from = terms,
      names_prefix = "",
      values_fill = list(terms = NA) # Riempie con NA se mancano valori
    )

  if (!"positive" %in% names(terms)) {
    terms$positive <- ""
  } else {
    terms$positive[is.na(terms$positive)] <- ""
  }
  if (!"negative" %in% names(terms)) {
    terms$negative <- ""
  } else {
    terms$negative[is.na(terms$negative)] <- ""
  }

  terms <- terms %>%
    rename(
      "Positively Associated Terms" = positive,
      "Negatively Associated Terms" = negative
    )

  summaryTable <- terms %>%
    left_join(segments, by = "cluster")

  return(summaryTable)
}

cutree_reinart <- function(res, k = NULL) {
  if (!is.null(k)) {
    res$uce_groups[[min(c(k - 1, length(res$uce_groups)))]]
  } else {
    res$uce_groups[[length(res$uce_groups)]]
  }
}

## Color palette for plots
colorlist <- function() {
  c(
    "#4DAF4A",
    "#E41A1C",
    "#377EB8",
    "#984EA3",
    "#FF7F00",
    "#A65628",
    "#F781BF",
    "#999999",
    "#66C2A5",
    "#FC8D62",
    "#8DA0CB",
    "#E78AC3",
    "#A6D854",
    "#FFD92F",
    "#B3B3B3",
    "#A6CEE3",
    "#1F78B4",
    "#B2DF8A",
    "#33A02C",
    "#FB9A99",
    "#E31A1C",
    "#FDBF6F",
    "#FF7F00",
    "#CAB2D6",
    "#6A3D9A",
    "#B15928",
    "#8DD3C7",
    "#BEBADA",
    "#FB8072",
    "#80B1D3",
    "#FDB462",
    "#B3DE69",
    "#D9D9D9",
    "#BC80BD",
    "#CCEBC5"
  )
}

test_that("reinert returns correct class", {
  data(mobydick, package = "tall")
  res <- reinert(
    x = mobydick,
    k = 5,
    term = "token",
    segment_size = 40,
    min_segment_size = 5,
    min_split_members = 10,
    cc_test = 0.3,
    tsj = 3
  )

  expect_s3_class(res, "reinert_tall")
  expect_s3_class(res, "hclust")
  expect_true("group" %in% names(res))
  expect_true("dtmOriginal" %in% names(res))
  expect_true("corresp_uce_uc_full" %in% names(res))
})

test_that("term_per_cluster returns terms and segments", {
  data(mobydick, package = "tall")
  res <- reinert(
    x = mobydick,
    k = 5,
    term = "token",
    segment_size = 40,
    min_segment_size = 5,
    min_split_members = 10,
    cc_test = 0.3,
    tsj = 3
  )

  tc <- term_per_cluster(res, cutree = NULL, k = 1:3, negative = FALSE)

  expect_type(tc, "list")
  expect_named(tc, c("terms", "segments"))
  expect_s3_class(tc$terms, "data.frame")
  expect_s3_class(tc$segments, "data.frame")
  expect_true("cluster" %in% names(tc$terms))
  expect_true("cluster" %in% names(tc$segments))
})

test_that("reinSummary returns a summary table", {
  data(mobydick, package = "tall")
  res <- reinert(
    x = mobydick,
    k = 5,
    term = "token",
    segment_size = 40,
    min_segment_size = 5,
    min_split_members = 10,
    cc_test = 0.3,
    tsj = 3
  )

  tc <- term_per_cluster(res, cutree = NULL, k = 1:3, negative = FALSE)
  S <- reinSummary(tc, n = 10)

  expect_s3_class(S, "data.frame")
  expect_true("cluster" %in% names(S))
  expect_true("Positively Associated Terms" %in% names(S))
  expect_true(nrow(S) > 0)
})

#' Compute Syntactic Complexity Metrics per Document (C++ backend)
#'
#' @param doc_id Character vector of document IDs
#' @param sent_id Integer vector of sentence IDs
#' @param token_id Integer vector of token IDs
#' @param head_token_id Integer vector of head token IDs
#' @param dep_rel Character vector of dependency relations
#' @param upos Character vector of universal POS tags
#' @return A data.frame with one row per document and complexity metrics
#' @export
compute_syntactic_complexity <- function(doc_id, sent_id, token_id, head_token_id, dep_rel, upos) {
  syntactic_complexity_cpp(doc_id, sent_id, token_id, head_token_id, dep_rel, upos)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mobydick.R
\docType{data}
\name{mobydick}
\alias{mobydick}
\title{Lemmatized Text of Moby-Dick (Chapters 1-10)}
\format{
A dataframe with multiple rows and 26 columns:
\describe{
\item{doc_id}{Character: Unique document identifier}
\item{paragraph_id}{Integer: Paragraph index within the document}
\item{sentence_id}{Integer: Sentence index within the paragraph}
\item{sentence}{Character: Original sentence text}
\item{start}{Integer: Start position of the token in the sentence}
\item{end}{Integer: End position of the token in the sentence}
\item{term_id}{Integer: Unique term identifier}
\item{token_id}{Integer: Token index in the sentence}
\item{token}{Character: Original token (word)}
\item{lemma}{Character: Lemmatized form of the token}
\item{upos}{Character: Universal POS tag}
\item{xpos}{Character: Language-specific POS tag}
\item{feats}{Character: Morphological features}
\item{head_token_id}{Integer: Head token in dependency tree}
\item{dep_rel}{Character: Dependency relation label}
\item{deps}{Character: Enhanced dependency relations}
\item{misc}{Character: Additional information}
\item{folder}{Character: Folder containing the document}
\item{split_word}{Character: The word used to separate the chapters in the original book}
\item{filename}{Character: Source file name}
\item{doc_selected}{Logical: Whether the document is selected}
\item{POSSelected}{Logical: Whether POS was selected}
\item{sentence_hl}{Character: Highlighted sentence}
\item{docSelected}{Logical: Whether the document was manually selected}
\item{noHapax}{Logical: Whether hapax legomena were removed}
\item{noSingleChar}{Logical: Whether single-character words were removed}
\item{lemma_original_nomultiwords}{Character: Lemmatized form without multi-word units}
}
}
\source{
Extracted and processed from the text of Moby-Dick by Herman Melville.
}
\usage{
data(mobydick)
}
\description{
This dataset contains the lemmatized version of the first 10 chapters of the novel Moby-Dick by Herman Melville.
The data is structured as a dataframe with multiple linguistic annotations.
}
\examples{
data(mobydick)
head(mobydick)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reinert.R
\name{term_per_cluster}
\alias{term_per_cluster}
\title{Extract Terms and Segments for Document Clusters}
\usage{
term_per_cluster(res, cutree = NULL, k = 1, negative = TRUE)
}
\arguments{
\item{res}{A list containing the results of the Reinert clustering algorithm. Must include at least \code{dtm} (a document-term matrix) and \code{corresp_uce_uc_full} (a correspondence between segments and clusters).}

\item{cutree}{A custom cutree structure. If \code{NULL}, the default \code{cutree_reinart} is used to determine cluster membership.}

\item{k}{A vector of integers specifying the clusters to analyze. Default is \code{1}.}

\item{negative}{Logical. If \code{TRUE}, include negative terms in the results. If \code{FALSE}, exclude them. Default is \code{TRUE}.}
}
\value{
A list with the following components:
\item{terms}{A data frame of significant terms for each cluster. Columns include:
\itemize{
\item \code{chi_square}: Chi-squared statistic for the term.
\item \code{p_value}: P-value of the chi-squared test.
\item \code{sign}: Significance of the term (\code{positive}, \code{negative}, or \code{none}).
\item \code{term}: The term itself.
\item \code{freq}: Observed frequency of the term in the cluster.
\item \code{indep}: Expected frequency of the term under independence.
\item \code{cluster}: The cluster ID.
}
}
\item{segments}{A data frame of document segments associated with each cluster. Columns include:
\itemize{
\item \code{uc}: Unique segment identifier.
\item \code{doc_id}: Document ID for the segment.
\item \code{cluster}: Cluster ID.
\item \code{segment}: The text content of each segment.
}
}
}
\description{
This function processes the results of a document clustering algorithm based on the Reinert method.
It computes the terms and their significance for each cluster, as well as the associated document segments.
}
\details{
The function integrates document-term matrix rows for missing segments, calculates term statistics for each cluster,
and filters terms based on their significance. Terms can be excluded based on their significance (\code{signExcluded}).
}
\examples{
\donttest{
data(mobydick)
res <- reinert(
  x = mobydick,
  k = 10,
  term = "token",
  segment_size = 40,
  min_segment_size = 5,
  min_split_members = 10,
  cc_test = 0.3,
  tsj = 3
)

tc <- term_per_cluster(res, cutree = NULL, k = 1:10, negative = FALSE)

head(tc$segments, 10)

head(tc$terms, 10)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/txt_recode_fast.R
\name{process_multiwords_fast}
\alias{process_multiwords_fast}
\title{Optimized multiword processing workflow}
\usage{
process_multiwords_fast(x2, stats, term = c("lemma", "token"))
}
\arguments{
\item{x2}{Data frame with token information}

\item{stats}{Data frame with multiword statistics (keyword, ngram columns)}

\item{term}{Type of term to process: "lemma" or "token"}
}
\value{
Data frame with columns: doc_id, term_id, multiword, upos_multiword, ngram
}
\description{
Complete optimized workflow for multiword detection and processing.
Uses C++ functions and data.table for maximum performance.
}
\details{
This function replaces the original switch block with an optimized version
that uses:
\itemize{
\item C++ functions for text recoding
\item Vectorized operations instead of multiple mutate calls
\item Pre-computed lookups to avoid repeated joins
}
}
\examples{
\dontrun{
result <- process_multiwords_fast(dfTag, multiword_stats, term = "lemma")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_svo.R
\name{extract_svo_triplets}
\alias{extract_svo_triplets}
\title{Extract SVO (Subject-Verb-Object) Triplets via Dependency Parsing (C++ backend)}
\usage{
extract_svo_triplets(sent_id, token_id, head_token_id, dep_rel, upos, terms)
}
\arguments{
\item{sent_id}{Integer vector of sentence IDs}

\item{token_id}{Integer vector of token IDs}

\item{head_token_id}{Integer vector of head token IDs}

\item{dep_rel}{Character vector of dependency relations}

\item{upos}{Character vector of universal POS tags}

\item{terms}{Character vector of term values (lemma or token)}
}
\value{
A data.frame with columns subject, verb, object, rel_type
}
\description{
Extract SVO (Subject-Verb-Object) Triplets via Dependency Parsing (C++ backend)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tall.R
\name{tall}
\alias{tall}
\title{TALL UI}
\usage{
tall(
  host = "127.0.0.1",
  port = NULL,
  launch.browser = TRUE,
  maxUploadSize = 1000
)
}
\arguments{
\item{host}{The IPv4 address that the application should listen on.
Defaults to the shiny.host option, if set, or "127.0.0.1" if not.}

\item{port}{is the TCP port that the application should listen on. If the port is not specified,
and the shiny.port option is set (with options(shiny.port = XX)), then that port will be used.
Otherwise, use a random port.}

\item{launch.browser}{If true, the system's default web browser will be launched automatically
after the app is started. Defaults to true in interactive sessions only. This value of
this parameter can also be a function to call with the application's URL.}

\item{maxUploadSize}{is a integer. The max upload file size argument. Default value is 1000 (megabyte)}
}
\value{
No return value, called for side effects.
}
\description{
\code{tall} performs text analysis for all.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/txt_recode_fast.R
\name{txt_recode_ngram_fast}
\alias{txt_recode_ngram_fast}
\title{Fast n-gram recoding for multiword detection}
\usage{
txt_recode_ngram_fast(x, compound, ngram, sep = " ")
}
\arguments{
\item{x}{Character vector of tokens (e.g., lemmas or tokens)}

\item{compound}{Character vector of multiword expressions to match}

\item{ngram}{Integer vector indicating the length of each compound}

\item{sep}{String separator to use when joining tokens (default: " ")}
}
\value{
Character vector where matched n-grams are combined and subsequent
tokens (that were merged) are set to NA
}
\description{
Efficiently combines consecutive tokens into multiword expressions using C++.
This function scans text sequentially to identify and merge n-gram patterns.
}
\details{
When a multiword match is found:
\itemize{
\item The first position gets the combined multiword expression
\item Subsequent positions that were merged are set to NA
}

The function checks n-grams from longest to shortest to prioritize longer matches.

Performance: ~80-150x faster than pure R implementation for typical text data.
}
\examples{
tokens <- c("machine", "learning", "is", "cool", "machine", "learning")
compounds <- c("machine learning")
ngrams <- c(2)
txt_recode_ngram_fast(tokens, compounds, ngrams, " ")
# Returns: c("machine learning", NA, "is", "cool", "machine learning", NA)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reinert.R
\name{reinPlot}
\alias{reinPlot}
\title{Plot Terms by Cluster}
\usage{
reinPlot(terms, nPlot = 10)
}
\arguments{
\item{terms}{A data frame containing terms and their associated statistics, such as Chi-squared values,
generated by the \code{term_per_cluster} function. The data frame must include the following columns:
\itemize{
\item \code{term}: The term to plot.
\item \code{chi_square}: The Chi-squared statistic associated with the term.
\item \code{sign}: The sign of the term (\code{"positive"} or \code{"negative"}).
}}

\item{nPlot}{Integer. The number of top terms to plot for each sign (\code{"positive"} and \code{"negative"}). Default is 10.}
}
\value{
An interactive horizontal bar plot (using \code{plotly}) displaying the top terms for each cluster. The plot includes:
\itemize{
\item Bars representing the Chi-squared values of terms.
\item Hover information displaying the term and its Chi-squared value.
}
}
\description{
This function creates a horizontal bar plot to visualize the most significant terms
for each cluster, based on their Chi-squared statistics.
}
\details{
The function organizes the input data by Chi-squared values and selects the top terms for each sign.
The plot uses different colors for positive and negative terms, with hover tooltips providing detailed information.
}
\examples{
\dontrun{
data(mobydick)
res <- reinert(
  x = mobydick,
  k = 10,
  term = "token",
  segment_size = 40,
  min_segment_size = 5,
  min_split_members = 10,
  cc_test = 0.3,
  tsj = 3
)

tc <- term_per_cluster(res, cutree = NULL, k = 1, negative = FALSE)

fig <- reinPlot(tc$terms, nPlot = 10)
}

}
\seealso{
\code{\link{term_per_cluster}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/txt_recode_fast.R
\name{txt_recode_fast}
\alias{txt_recode_fast}
\title{Fast text recoding (Rcpp version)}
\usage{
txt_recode_fast(x, from = c(), to = c(), na.rm = FALSE)
}
\arguments{
\item{x}{A character vector to recode}

\item{from}{A character vector with values of \code{x} which you want to recode}

\item{to}{A character vector with values you want to use to recode to}

\item{na.rm}{Logical, if set to TRUE, will put all values of \code{x} which
have no matching value in \code{from} to NA. Defaults to FALSE}
}
\value{
A character vector of the same length as \code{x} where values
matching \code{from} are replaced by corresponding values in \code{to}
}
\description{
Efficiently recodes text values using C++ hash tables. This is a drop-in
replacement for \code{txt_recode} but significantly faster for large vectors.
}
\details{
This function uses C++ hash tables for O(1) lookup time, making it much
faster than the pure R implementation, especially for large datasets.

Performance improvement: ~50-100x faster than base R \code{txt_recode}
for vectors with 100K+ elements.
}
\examples{
x <- c("NOUN", "VERB", "NOUN", "ADV")
txt_recode_fast(x,
  from = c("VERB", "ADV"),
  to = c("conjugated verb", "adverb")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/syntactic_complexity.R
\name{compute_syntactic_complexity}
\alias{compute_syntactic_complexity}
\title{Compute Syntactic Complexity Metrics per Document (C++ backend)}
\usage{
compute_syntactic_complexity(
  doc_id,
  sent_id,
  token_id,
  head_token_id,
  dep_rel,
  upos
)
}
\arguments{
\item{doc_id}{Character vector of document IDs}

\item{sent_id}{Integer vector of sentence IDs}

\item{token_id}{Integer vector of token IDs}

\item{head_token_id}{Integer vector of head token IDs}

\item{dep_rel}{Character vector of dependency relations}

\item{upos}{Character vector of universal POS tags}
}
\value{
A data.frame with one row per document and complexity metrics
}
\description{
Compute Syntactic Complexity Metrics per Document (C++ backend)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reinert.R
\name{reinert}
\alias{reinert}
\title{Segment clustering based on the Reinert method - Simple clustering}
\usage{
reinert(
  x,
  k = 10,
  term = "token",
  segment_size = 40,
  min_segment_size = 3,
  min_split_members = 5,
  cc_test = 0.3,
  tsj = 3
)
}
\arguments{
\item{x}{tall data frame of documents}

\item{k}{maximum number of clusters to compute}

\item{term}{indicates the type of form "lemma" or "token". Default value is term = "lemma".}

\item{segment_size}{number of forms by document. Default value is segment_size = 40}

\item{min_segment_size}{minimum number of forms by document. Default value is min_segment_size = 5}

\item{min_split_members}{minimum number of segment in a cluster}

\item{cc_test}{contingency coefficient value for feature selection}

\item{tsj}{minimum frequency value for feature selection}
}
\value{
The result is a list of both class \code{hclust} and \code{reinert_tall}.
}
\description{
Segment clustering based on the Reinert method - Simple clustering
}
\details{
See the references for original articles on the method.
Special thanks to the authors of the rainette package (https://github.com/juba/rainette)
for inspiring the coding approach used in this function.
}
\examples{
\donttest{
data(mobydick)
res <- reinert(
  x = mobydick,
  k = 10,
  term = "token",
  segment_size = 40,
  min_segment_size = 5,
  min_split_members = 10,
  cc_test = 0.3,
  tsj = 3
)
}

}
\references{
\itemize{
\item Reinert M, Une methode de classification descendante hierarchique: application à l'analyse lexicale par contexte, Cahiers de l'analyse des donnees, Volume 8, Numéro 2, 1983. \url{https://www.numdam.org/item/?id=CAD_1983__8_2_187_0}
\item Reinert M., Alceste une méthodologie d'analyse des données textuelles et une application: Aurelia De Gerard De Nerval, Bulletin de Methodologie Sociologique, Volume 26, Numero 1, 1990. \doi{10.1177/075910639002600103}
\item Barnier J., Privé F., rainette: The Reinert Method for Textual Data Clustering, 2023, \doi{10.32614/CRAN.package.rainette}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extract_np.R
\name{extract_noun_phrases}
\alias{extract_noun_phrases}
\title{Extract Noun Phrases via Dependency Parsing (C++ backend)}
\usage{
extract_noun_phrases(
  sent_id,
  token_id,
  head_token_id,
  dep_rel,
  upos,
  terms,
  ngram_max = 5L,
  max_gap = 3L
)
}
\arguments{
\item{sent_id}{Integer vector of sentence IDs}

\item{token_id}{Integer vector of token IDs}

\item{head_token_id}{Integer vector of head token IDs}

\item{dep_rel}{Character vector of dependency relations}

\item{upos}{Character vector of universal POS tags}

\item{terms}{Character vector of term values (lemma or token)}

\item{ngram_max}{Maximum phrase length (default 5)}

\item{max_gap}{Maximum gap between token positions (default 3)}
}
\value{
A data.frame with columns keyword and ngram
}
\description{
Extract Noun Phrases via Dependency Parsing (C++ backend)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/txt_recode_fast.R
\name{txt_recode_ngram_cpp}
\alias{txt_recode_ngram_cpp}
\title{Internal C++ n-gram recoding function}
\arguments{
\item{x}{Character vector of tokens}

\item{compound}{Character vector of multiword expressions to match}

\item{ngram}{Integer vector indicating the length of each compound}

\item{sep}{String separator to use when joining tokens}
}
\value{
Character vector where matched n-grams are combined
}
\description{
Internal C++ implementation of n-gram recoding for multiword detection.
Users should call \code{\link{txt_recode_ngram_fast}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/txt_recode_fast.R
\name{txt_recode_cpp}
\alias{txt_recode_cpp}
\title{Internal C++ text recoding function}
\arguments{
\item{x}{Character vector to recode}

\item{from}{Character vector of values to match}

\item{to}{Character vector of replacement values}

\item{na_rm}{Logical, whether to set unmatched values to NA}
}
\value{
Character vector with recoded values
}
\description{
Internal C++ implementation of text recoding using hash tables.
Users should call \code{\link{txt_recode_fast}} instead.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calculate_ngram_is.R
\name{calculate_ngram_is}
\alias{calculate_ngram_is}
\title{Calculate IS index for n-grams}
\usage{
calculate_ngram_is(
  dfTag,
  max_ngram = 5,
  term = "lemma",
  pos = c("NOUN", "ADJ", "ADV", "VERB"),
  min_freq = 1,
  min_IS_norm = 0
)
}
\arguments{
\item{dfTag}{A data frame with tagged text data containing columns: doc_id,
sentence_id, token_id, lemma/token, upos}

\item{max_ngram}{Maximum length of n-grams to generate (default: 5)}

\item{term}{Character string indicating which column to use: "lemma" or "token" (default: "lemma")}

\item{pos}{Character vector of POS tags considered lexical (default: c("NOUN", "ADJ", "ADV", "VERB"))}

\item{min_freq}{Minimum frequency threshold for n-grams (default: 1)}

\item{min_IS_norm}{Minimum normalized IS threshold for n-grams (default: 0)}
}
\value{
A tibble with columns: ngram, n_length, ngram_freq, n_lexical, IS, IS_norm
}
\description{
This function calculates the IS (Absorption Index) from Morrone (1996)
for all n-grams in the corpus. Only n-grams that start AND end with
lexical words are considered.
}
\details{
The IS index is calculated as: IS = (sum 1/freq_i) × freq_ngram × n_lexical
where freq_i is the frequency of each word in the n-gram, freq_ngram is the
frequency of the n-gram, and n_lexical is the number of lexical words.
IS_norm is the normalized version: IS / L^2 where L is the n-gram length.

OPTIMIZATION: Only n-grams that start AND end with lexical words (as defined by
the 'pos' parameter) are generated, significantly reducing computation time.
}
\examples{
\dontrun{
IS <- calculate_ngram_is(dfTag, max_ngram = 4, term = "lemma", min_freq = 2)
head(IS)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/txt_recode_fast.R
\name{txt_recode_batch_cpp}
\alias{txt_recode_batch_cpp}
\title{Batch text recoding (internal C++ function)}
\arguments{
\item{x}{List of character vectors to recode}

\item{from}{Character vector of values to match}

\item{to}{Character vector of replacement values}

\item{na_rm}{Logical, whether to set unmatched values to NA (default: FALSE)}
}
\value{
List of character vectors with recoded values
}
\description{
Internal C++ function for batch recoding multiple character vectors.
Efficiently recodes multiple text columns using a single hash table lookup.
}
\details{
This is an internal C++ function exported for advanced users.
The hash table is constructed once and reused for all vectors in the list,
making it very efficient for batch operations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reinert.R
\name{reinSummary}
\alias{reinSummary}
\title{Summarize Reinert Clustering Results}
\usage{
reinSummary(tc, n = 10)
}
\arguments{
\item{tc}{A list returned by the \code{term_per_cluster} function. The list includes:
\itemize{
\item \code{segments}: A data frame with segments information, including \code{cluster} and \code{doc_id}.
\item \code{terms}: A data frame with terms information, including \code{cluster}, \code{sign}, \code{chi_square}, and \code{term}.
}}

\item{n}{Integer. The number of top terms (based on Chi-squared value) to include in the summary for each cluster and sign. Default is 10.}
}
\value{
A data frame summarizing the clustering results. The table includes:
\itemize{
\item \code{cluster}: The cluster ID.
\item \code{Positive terms}: The top \code{n} positive terms for each cluster, concatenated into a single string.
\item \code{Negative terms}: The top \code{n} negative terms for each cluster, concatenated into a single string.
\item \code{Most frequent document}: The document ID that appears most frequently in each cluster.
\item \code{N. of Documents per Cluster}: The number of documents in each cluster.
}
}
\description{
This function summarizes the results of the Reinert clustering algorithm, including the most frequent documents and significant terms for each cluster.
The input is the result returned by the \code{term_per_cluster} function.
}
\details{
This function performs the following steps:
\enumerate{
\item Extracts the most frequent document for each cluster.
\item Summarizes the number of documents per cluster.
\item Selects the top \code{n} terms for each cluster, separated by positive and negative signs.
\item Combines the terms and segment information into a final summary table.
}
}
\examples{
\donttest{
data(mobydick)
res <- reinert(
  x = mobydick,
  k = 10,
  term = "token",
  segment_size = 40,
  min_segment_size = 5,
  min_split_members = 10,
  cc_test = 0.3,
  tsj = 3
)

tc <- term_per_cluster(res, cutree = NULL, k = 1:10, negative = FALSE)

S <- reinSummary(tc, n = 10)

head(S, 10)
}

}
\seealso{
\code{\link{term_per_cluster}}, \code{\link{reinPlot}}
}
